//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODAUDIOSOURCEAC3_H__
#define __MODAUDIOSOURCEAC3_H__

#include "AudioSource.h"
#include <vd2/system/File64.h>
#include "ModList.h"
#include "ModClassInterfaces.h"


// WAVE wFormatTag field values (WORD) :
//   0x2000 for AC3


// ===================================================================
// According to the specs (see below), an AC3 frame is composed of 5 main parts :
//  - SI  (Synchonisation Information header) (5 bytes)
//  - BSI (Bit Stream Information header)
//  - 6 coded audio blocks
//  - AUX (auxiliary data block) (optional)
//  - CRC (error check field)
//
// The SI block begins with a synchonisation word (0x0B77), contains a CRC
// field, and two fields used to determine the sample rate, the bitrate and the
// size (in words - 2 bytes) of the frame (including the SI block).
// The biggest frame size is 1920 words (see table below) = 3840 bytes.
//
// Each coded audio block represents 256 new audio samples
// , i.e. there is 256*6=1536 samples in a frame.
//
// The BSI block can contain timecode that can have a reolution of 1/64th of a frame.
//====================================================================

// Synch word for a frame is 0B77
// Frame valuable information (for parsing) are stored in the first 7 bytes
#define AC3_HEADER_LEN		7
#define AC3_SYNCWORD_LEN	2
#define AC3_SAMPLESPERFRAME	1536
#define AC3_MAX_LENGTH		3840


// Some of the AC3 specifications
// Found on http://www.atsc.org, standards section, document A/52
// Digital Audio Compression (AC-3) Standard


// Samplerates ([SampleRate flags])
static sint32 AC3_SampleRate[4] = { 48000, 44100, 32000, -1 };

// Bitrates ([Bitrate flags])
static sint32 AC3_Bitrate[38] =
{ 32,  32,  40,  40,  48,  48,  56,  56,  64,  64,  80,  80,  96,  96
, 112, 112, 128, 128, 160, 160, 192, 192, 224, 224, 256, 256, 320, 320
, 384, 384, 448, 448, 512, 512, 576, 576, 640, 640 };

// Size of a frame in words (2 bytes) ([Bitrate flags][Samplerate flags])
static sint32 AC3_FrameLength[38][3] =
{
	{ 64,   69,   96},
	{ 64,   70,   96},
	{ 80,   87,   120},
	{ 80,   88,   120},
	{ 96,   104,  144},
	{ 96,   105,  144},
	{ 112,  121,  168},
	{ 112,  122,  168},
	{ 128,  139,  192},
	{ 128,  140,  192},
	{ 160,  174,  240},
	{ 160,  175,  240},
	{ 192,  208,  288},
	{ 192,  209,  288},
	{ 224,  243,  336},
	{ 224,  244,  336},
	{ 256,  278,  384},
	{ 256,  279,  384},
	{ 320,  348,  480},
	{ 320,  349,  480},
	{ 384,  417,  576},
	{ 384,  418,  576},
	{ 448,  487,  672},
	{ 448,  488,  672},
	{ 512,  557,  768},
	{ 512,  558,  768},
	{ 640,  696,  960},
	{ 640,  697,  960},
	{ 768,  835,  1152},
	{ 768,  836,  1152},
	{ 896,  975,  1344},
	{ 896,  976,  1344},
	{ 1024, 1114, 1536},
	{ 1024, 1115, 1536},
	{ 1152, 1253, 1728},
	{ 1152, 1254, 1728},
	{ 1280, 1393, 1920},
	{ 1280, 1394, 1920}
};

// Number of channels ([Channels flags])
static sint32 AC3_NbChannels[8] = { 2, 1, 2, 3, 3, 4, 4, 5};

class AC3Frame : public ModListNode<AC3Frame>
{
public:
	sint64	m_FileOffset;	// Position of the frame in the file
	uint32	m_Flags;		// Information on the frame
};
// Flags for corrupte frame : set bitrate to 'bad'
#define CORRUPTED_AC3_FRAME 0xFF

// Some options
struct AC3Options
{
	bool	m_bKeepCorruptedData;
	bool	m_bFrameAccuracy;
};


// Some general information
struct AC3StreamInfo
{
	char	m_Buffer[AC3_HEADER_LEN];
	sint16	m_Version;
	sint16	m_SampleRateFlags;
	sint32	m_SampleRate;
	sint16	m_Channels;
	sint32	m_NbChannels;
	sint32	m_LFE;
	sint32	m_Bitrate;
	sint32	m_FrameLength;
	sint64	m_Frames;
	ModList<AC3Frame> m_FrameList;
	sint64	m_Bytes;
	sint64	m_BytesSkipped;
};

class ModAudioSourceAC3 : public AudioSource, public IFramedSource {
protected:
	VDFile			m_VDFile;
	bool			m_bHasRIFFHeader;
	sint64			m_RIFFHeaderLength;
	sint64			m_FileSize;
	AC3StreamInfo	m_AC3Info;
	sint32			m_MaxFrameLength;
	bool			m_bVariableFrameLength;
	AC3Options		m_AC3Options;
	sint32			m_BlockAlign;
	sint64			m_NodeSample;
	sint32			m_NodeFrame;
	sint64			m_NodeByte;
	sint32			m_FrameCount;
	sint64			m_LastSample;
	sint64			m_TotalBytes;
	ModList<AC3Frame>::iterator m_Iterator;
	//
	bool			m_bVBRHandling;
	sint32			m_GroupFrames;

	void	CheckTags(void);
	void	AddAC3Frame(void);
	void	AddCorruptedPart(sint64 startCorrupted, sint64 bytesCorrupted);
	void	UpdateFrameCount(void);
	bool	GetHeader(bool ensureConsistancy);
	bool	GetNextFrame(bool ensureConsistancy, sint64& bytesSkipped);
	sint32	GetFrameLength(AC3Frame *node);
	sint32	ComputeFrameLength(AC3Frame *previous, AC3Frame *current, AC3Frame *next);
	bool	SeekSample(LONG lStart, sint32& framesLength, sint32& frameNumber, sint32& frameLength);
	void	ComputeFramesInNode(void);

public:
	ModAudioSourceAC3(const wchar_t *szFileName, AC3Options options);
	virtual ~ModAudioSourceAC3();

	bool			init();
	void			extend(bool bVBRHandling, sint32 groupFrames);
	virtual int		_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);
	virtual sint64	sampleOffset(sint64 lSample);
	virtual sint32	samplesLag(sint64 lStart);

	VDStringW		writeReport(void);

	bool			QueryInterface(InterfaceID ID, void** destination);
	bool			UseFrameMode(void);
};

#endif // __MODAUDIOSOURCEAC3_H__