//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODAUDIOSOURCEDTS_H__
#define __MODAUDIOSOURCEDTS_H__

#include "AudioSource.h"
#include <vd2/system/File64.h>
#include "ModList.h"
#include "ModClassInterfaces.h"


// WAVE wFormatTag field values (WORD) :
//   0x2001 for DTS (*hmm*, not really official :p)


// Synch word for a frame is 7FFE8001
// Frame valuable information (for parsing) are stored in the first 10 bytes
#define DTS_HEADER_LEN		10
#define DTS_SYNCWORD_LEN	4
#define DTS_MAX_LENGTH		16383
#define DTS_SAMPLESPERBLOCK	32


// ===================================================================
// Some of the DTS specifications
// According to some specs here is the 'Sub-sampled audio decoding
//  for standard sampling rates' table :
//	+---------------+-------------------+-------------------+
//	| Core Sampling	| Hardware Sampling	| Required			|
//	| Frequency		| Frequency			| Filtering			|
//	+===============+===================+===================+
//	| 8 kHz			| 32 kHz			| 4  Interpolation	|
//	| 16 kHz		| 32 kHz			| 2  Interpolation	|
//	| 32 kHz		| 32 kHz			| none				|
//	| 11 kHz		| 44,1 kHz			| 4  Interpolation	|
//	| 22,05 kHz		| 44,1 kHz			| 2  Interpolation	|
//	| 44,1 kHz		| 44,1 kHz			| none				|
//	| 12 kHz		| 48 kHz			| 4  Interpolation	|
//	| 24 kHz		| 48 kHz			| 2  Interpolation	|
//	| 48 kHz		| 48 kHz			| none				|
//	+---------------+-------------------+-------------------+
// ===================================================================

// Samplerates ([SampleRate flags])
static sint32 DTS_SampleRate[16] = {
			// core frequency
	-1,		// invalid
	32000,	// 8000
	32000,	// 16000
	32000,	// 32000
	-1,		// invalid
	-1,		// invalid
	44100,	// 11025
	44100,	// 22050
	44100,	// 44100
	-1,		// invalid
	-1,		// invalid
	48000,	// 12000
	48000,	// 24000
	48000,	// 48000
	-1,		// invalid
	-1		// invalid
};

// Bitrates ([Bitrate flags])
static sint32 DTS_Bitrate[32] = {
	32000,
	56000,
	64000,
	96000,
	112000,
	128000,
	192000,
	224000,
	256000,
	320000,
	384000,
	448000,
	512000,
	576000,
	640000,
	768000,
	960000,
	1024000,
	1152000,
	1280000,
	1344000,
	1408000,
	1411200,
	1472000,
	1536000,
	1920000,
	2048000,
	3072000,
	3840000,
	0,	// open
	0,	// Variable
	0	// Loss-less
};


// Number of channels ([Channels flags])
static sint32 DTS_NbChannels[16] =
	{ 1, 2, 2, 2, 2, 3, 3, 4, 4, 5, 6, 6, 6, 7, 8, 8};

class DTSFrame : public ModListNode<DTSFrame>
{
public:
	sint64	m_FileOffset;	// Position of the frame in the file
	uint8	m_Short;
	uint8	m_Blocks;
	uint16	m_Length;
};
// Flags for corrupted frame : set m_Short to 0xFF
#define CORRUPTED_DTS_FRAME 0xFF

// Some options
struct DTSOptions
{
	bool	m_bKeepCorruptedData;
	bool	m_bFrameAccuracy;
};


// Some general information
struct DTSStreamInfo
{
	char	m_Buffer[DTS_HEADER_LEN];
	sint16	m_SampleRateFlags;
	sint32	m_SampleRate;
	sint16	m_Channels;
	sint32	m_NbChannels;
	sint32	m_Bitrate;
	sint32	m_FrameLength;
	sint32	m_NbBlocks;
	sint32	m_SampleDeficit;
	sint32	m_Samples;
	sint64	m_TotalSamples;
	sint64	m_Frames;
	ModList<DTSFrame> m_FrameList;
	sint64	m_Bytes;
	sint64	m_BytesSkipped;
};

class ModAudioSourceDTS : public AudioSource, public IFramedSource {
protected:
	VDFile			m_VDFile;
	bool			m_bHasRIFFHeader;
	sint64			m_RIFFHeaderLength;
	sint64			m_FileSize;
	DTSStreamInfo	m_DTSInfo;
	sint32			m_MaxFrameLength;
	bool			m_bVariableFrameLength;
	DTSOptions		m_DTSOptions;
	sint32			m_BlockAlign;
	sint64			m_NodeSample;
	sint32			m_NodeFrame;
	sint64			m_NodeByte;
	sint32			m_FrameCount;
	sint32			m_SampleCount;
	sint64			m_LastSample;
	sint64			m_TotalBytes;
	ModList<DTSFrame>::iterator m_Iterator;
	//
	bool			m_bVBRHandling;
	sint32			m_GroupFrames;

	void	CheckTags(void);
	void	AddDTSFrame(void);
	void	AddCorruptedPart(sint64 startCorrupted, sint64 bytesCorrupted);
	void	UpdateFrameCount(void);
	bool	GetHeader(bool ensureConsistancy);
	bool	GetNextFrame(bool ensureConsistancy, sint64& bytesSkipped);
	sint32	GetFrameLength(DTSFrame *node);
	sint32	ComputeFrameLength(DTSFrame *previous, DTSFrame *current, DTSFrame *next);
	sint32	GetFrameSize(DTSFrame *node);
	sint32	ComputeFrameSize(DTSFrame *previous, DTSFrame *current, DTSFrame *next);
	bool	SeekSample(LONG lStart, sint32& framesLength, sint32& frameNumber, sint32& frameLength);
	void	ComputeFramesInNode(void);

public:
	ModAudioSourceDTS(const wchar_t *szFileName, DTSOptions options);
	~ModAudioSourceDTS();

	bool			init();
	void			extend(bool bVBRHandling, sint32 groupFrames);
	virtual int		_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);
	virtual sint64	sampleOffset(sint64 lSample);
	virtual sint32	samplesLag(sint64 lStart);

	VDStringW		writeReport(void);

	bool			QueryInterface(InterfaceID ID, void** destination);
	bool			UseFrameMode(void);
};


#endif // __MODAUDIOSOURCEDTS_H__
