//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODAUDIOSOURCEMP3_H__
#define __MODAUDIOSOURCEMP3_H__

#define _USE_COMPACTLIST_MPA

#include "AudioSource.h"
#include <vd2/system/File64.h>
#ifdef _USE_COMPACTLIST_MPA
#include "ModCompactList.h"
#else // !_USE_COMPACTLIST_MPA
#include "ModList.h"
#endif //_USE_COMPACTLIST_MPA
#include "ModClassInterfaces.h"


// WAVE wFormatTag field values (WORD) :
//   WAVE_FORMAT_MPEG        0x0050  // MPEG Layer 1,2
//   WAVE_FORMAT_MPEGLAYER3  0x0055  // MPEG Layer 3


// ===================================================================
// Mpeg Audio source
//
// For Layer I files use this formula:
//
//   FrameLengthInBytes = (12 * BitRate / SampleRate + Padding) * 4
//
// For Layer II & III files use this formula:
//
//   FrameLengthInBytes = 144 * BitRate / SampleRate + Padding
//
// For Mpeg2/2.5-Layer III use thif formula:
//
//   FrameLengthInBytes = 72 * BitRate / SampleRate + Padding
//
// For Layer I :
//   max BitRate is 448000 for MPEG V1, 256000 for MPEG V2/2.5
//   min SampleRate is 32000 for MPEG V1, and 8000 for MPEG V2/2.5
//   so max size is 676 for MPEG V1, and 1540 for MPEG V2/2.5
// For Layer II/III, it's :
//   384000 and 160000
//   still 32000 and 8000
//   so max sizes are 1729 and 2881
//
// Layer I : 1 frame = 384 samples
// Layer II/III : 1 frame = 1152 samples
// Mpeg2/2.5-Layer III : 1 frame = 576 samples
// ===================================================================

// Synch word for a frame is FFE0 (first 11 bits must be set)
// Frame valuable information (for parsing) are stored in the first 4 bytes
#define MPA_HEADER_LEN		4
#define MPA_SYNCWORD_LEN	2
#define MPA_MAX_LENGTH		2881

// The MPEG versions
enum MPA_Version
{
	MPEG_V1 = 0,
	MPEG_V2,
	MPEG_V25,
	MPEG_RESERVED
};

// The layer versions
enum MPA_Layer
{
	LAYER_1 = 0,
	LAYER_2,
	LAYER_3,
	LAYER_RESERVED
};

// The frame size ([Layer][MPEG Version])
static sint32 MPA_FrameSize[4][4] =
{ 
	{	384,	384,	384,	0},
	{	1152,	1152,	1152,	0},
	{	1152,	576,	576,	0},
	{	0,		0,		0,		0}
};

// The bitrates for MPEG Version 1 ([Layer][Bitrate flags])
static sint32 MPA_Bitrate_V1[4][16] =
{
	{	0,	32,	64,	96,	128,	160,	192,	224,	256,	288,	320,	352,	384,	416,	448,	-1},
	{	0,	32,	48,	56,	64,		80,		96,		112,	128,	160,	192,	224,	256,	320,	384,	-1},
	{	0,	32,	40,	48,	56,		64,		80,		96,		112,	128,	160,	192,	224,	256,	320,	-1},
	{	-1,	-1,	-1,	-1,	-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1}
};

// The bitrates for MPEG version 2 & 2.5 ([Layer][Bitrate flags])
static sint32 MPA_Bitrate_V2[4][16] =
{
	{	0,	32,	48,	56,	64,	80,	96,	112,	128,	144,	160,	176,	192,	224,	256,	-1},
	{	0,	8,	16,	24,	32,	40,	48,	56,		64,		80,		96,		112,	128,	144,	160,	-1},
	{	0,	8,	16,	24,	32,	40,	48,	56,		64,		80,		96,		112,	128,	144,	160,	-1},
	{	-1,	-1,	-1,	-1,	-1,	-1,	-1,	-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1,		-1}
};

// The samplerates ([SampleRate flags][MPEG version])
static sint32 MPA_SampleRate[4][4] =
{
	{	44100,	22050,	11025,	-1},
	{	48000,	24000,	12000,	-1},
	{	32000,	16000,	8000,	-1},
	{	-1,		-1,		-1,		-1}
};

// Channel modes
enum MPA_Channels
{
	SINGLE_CHANNEL = 0,
	DUAL_CHANNEL,
	JOINT_STEREO,
	STEREO
};

// Number of channels ([Channels mode])
static sint32 MPA_NbChannels[4] = { 1, 2, 2, 2};

// We will store information in this kind of structure
// A file which size is less than 2^32 bits (so needing a 32-bits field) could
// store up to 21 hours of MP3 (considering the highest bitrate allowed).
// Even so we will take 64-bits :P (what if I decide to record myself during a
// whole day heh !)
#ifdef _USE_COMPACTLIST_MPA
struct MPAFrame
{
#else // !_USE_COMPACTLIST_MPA
class MPAFrame : public ModListNode<MPAFrame>
{
public:
#endif // _USE_COMPACTLIST_MPA
	sint64	m_FileOffset;	// Position of the frame in the file
	uint8	m_Flags;		// Information on the frame (bitrate & padding)
};
// Flags for corrupte frame : set bitrate to 'bad'
#define CORRUPTED_MPA_FRAME 0xF0

// Some options
struct MPAOptions
{
	bool	m_bKeepTags;
	bool	m_bKeepCorruptedData;
	bool	m_bNandubTrick;
};

// Some general information
struct MPAStreamInfo
{
	char			m_Buffer[MPA_HEADER_LEN];
	MPA_Version		m_Version;
	MPA_Layer		m_Layer;
	sint32			m_SampleRate;
	MPA_Channels	m_Channels;
	sint32			m_NbChannels;
	sint32			m_Bitrate;
	uint16			m_HeadMode;
	uint16			m_HeadModeExt;
	uint8			m_Emphasis;
	bool			m_bProtected;
	bool			m_bPrivate;
	bool			m_bCopyrighted;
	bool			m_bOriginal;
	sint32			m_Padding;
	sint32			m_SamplesPerFrame;
	sint32			m_FrameLength;
	sint64			m_Frames;
#ifdef _USE_COMPACTLIST_MPA
	ModCompactList<MPAFrame> m_FrameList;
#else // !_USE_COMPACTLIST_MPA
	ModList<MPAFrame> m_FrameList;
#endif // _USE_COMPACTLIST_MPA
	sint64			m_Bytes;
	sint64			m_BytesSkipped;
};

class ModAudioSourceMP3 : public AudioSource, public IMP3Source {
protected:
	VDFile			m_VDFile;
	bool			m_bHasRIFFHeader;
	sint64			m_RIFFHeaderLength;
	bool			m_bHasTagV1;		// Tag ID3v1(.1) -> 128 bytes at the end of the file
	bool			m_bHasTagV2_Start;	// Tag ID3v2 -> variable length at the beginning of the file
	bool			m_bHasTagV2_End;	// Tag ID3v2 -> variable length at the end of the file
	sint64			m_TagV2StartLength;
	sint64			m_TagV2EndLength;
	sint64			m_FileSize;
	sint64			m_EndData;
	MPAStreamInfo	m_MPAInfo;
	sint32			m_MaxFrameLength;
	bool			m_bVariableFrameLength;
	MPAOptions		m_MPAOptions;
	sint32			m_BlockAlign;
	sint64			m_NodeSample;
	sint32			m_NodeFrame;
	sint64			m_NodeByte;
	sint32			m_FrameCount;
	sint64			m_LastSample;
	sint64			m_TotalBytes;
#ifdef _USE_COMPACTLIST_MPA
	ModCompactList<MPAFrame>::iterator m_Iterator;
#else // !_USE_COMPACTLIST_MPA
	ModList<MPAFrame>::iterator m_Iterator;
#endif // _USE_COMPACTLIST_MPA
	bool			m_bVBRHandling;
	sint32			m_GroupFrames;
	bool			m_bIsVBR;
	bool			m_bShutUp;
	bool			m_bEmphasisChanged;
	bool			m_bProtectedChanged;
	bool			m_bPrivateChanged;
	bool			m_bCopyrightedChanged;
	bool			m_bOriginalChanged;
	bool			m_bPaddingChanged;

	void	CheckTags(void);
	void	AddMPAFrame(void);
	void	AddCorruptedPart(sint64 startCorrupted, sint64 bytesCorrupted);
	void	UpdateFrameCount(void);
	bool	GetNextFrame(bool ensureConsistancy, sint64& bytesSkipped);
	bool	GetHeader(bool ensureConsistancy);
	sint32	GetFrameLength(MPAFrame *node);
	sint32	ComputeFrameLength(MPAFrame *previous, MPAFrame *current, MPAFrame *next);
	bool	SeekSample(LONG lStart, sint32& framesLength, sint32& frameNumber, sint32& frameLength);
	void	ComputeFramesInNode(void);

public:
	ModAudioSourceMP3(const wchar_t *szFileName, MPAOptions options, bool bShutUp);
	~ModAudioSourceMP3();

	bool			init();
	void			extend(bool bVBRHandling, sint32 groupFrames);
	virtual int		_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);
	virtual sint64	sampleOffset(sint64 lSample);
	virtual sint32	samplesLag(sint64 lStart);

	VDStringW		writeReport(void);

	bool			QueryInterface(InterfaceID ID, void** destination);
	bool			UseVBRalaNandub(void);
	bool			IsVBR(void);
	bool			FixVBR(bool bFix);
};

#endif // __MODAUDIOSOURCEMP3_H__