//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODLIST_H__
#define __MODLIST_H__

#include <crtdbg.h>

// Highly inspired by List definitions in VirtualDub (see list.h)
// On the contrary to VirtualDub lists this one doesn't use dummy
// head or tail because there were cases where this trick was an
// issue for me.
// As for VirtualDub lists the good point here is that to make a
// list with an object you just need to use the classe below as
// base class and your new class will only require the size of
// your data plus the size for the backward and forward pointers :)
//
// I made some speed tests to compare this list with VirtualDub
// one using a simple object (2 fields for a total of 9 bytes).
// I built each kind of list (VirtualDub one / this one) and built
// 1000000 nodes I added in each list. I repeated this a hundred
// times and made an average.
// The results were about the same : around 750 ms to proceed on
// my computer.

// Nb : the class functions must be defined here since we use templates
// We could do this in ModList.cpp but then we would have to include
// ModList.h _and_ ModList.cpp to use this list.

template<class _N> class ModList;


// ModListNode definition

template<class _N>
class ModListNode {
	friend ModList<_N>;
protected:
	_N *m_Previous;
	_N *m_Next;

public:
	ModListNode<_N>(void);
	//void Remove(void);
	_N *GetPrevious(void);
	_N *GetNext(void);
};


template<class _N>
ModListNode<_N>::ModListNode<_N>(void) {
	m_Previous = NULL;
	m_Next = NULL;
}

/*template<class _N>
void ModListNode<_N>::Remove(void) {
	if(m_Previous)
		m_Previous->m_Next = m_Next;
	if(m_Next)
		m_Next->m_Previous = m_Previous;
#ifdef _DEBUG
	m_Previous = NULL;
	m_Next = NULL;
#endif
}*/

template<class _N>
_N *ModListNode<_N>::GetPrevious(void) {
	return m_Previous;
}

template<class _N>
_N *ModListNode<_N>::GetNext(void) {
	return m_Next;
}


// ModList definition

template<class _N>
class ModList {
protected:
	size_t m_Count;
	bool m_bFree;
	_N *m_FirstNode;
	_N *m_LastNode;

	virtual void DoDeleteNode(_N *node);

public:
	typedef _N Node;
	ModList<_N>(void);
	ModList<_N>(ModList<_N>& src);
	virtual ~ModList<_N>(void);
	ModList<_N>& operator =(ModList<_N>& src);
	_N *GetFirst(void);
	_N *GetLast(void);
	_N *AddLast(void);
	size_t GetCount(void);
	_N *operator [](size_t index);
	sint32 IndexOf(_N *node);
	void Append(_N *node);
	void Insert(_N *node);
	void Insert(_N *node, size_t index);
	void Swap(_N *node1, _N *node2);
	void DetachNode(_N *node);
	void DeleteNode(_N *node);
	void DeleteContents(bool bFree);
	void Clear(void);

	class iterator_safer;
	class iterator {
	friend class iterator_safer;
	protected:
		ModList<_N> *m_List;
		_N *m_Node;

	public:
		iterator(void) {
			m_List = NULL;
			m_Node = NULL;
		};
		iterator(ModList<_N>& list) {
			m_List = &list;
			m_Node = list.GetFirst();
		};
		iterator(const iterator& src) throw() {
			m_List = src.m_List;
			m_Node = src.m_Node;
		};

		void begin(ModList<_N>& list) {
			m_List = &list;
			m_Node = list.GetFirst();
		};
		void end(ModList<_N>& list) {
			m_List = &list;
			m_Node = list.GetLast();
		};
		_N *previous(void)
			{ return m_Node->GetPrevious(); };
		_N *next(void)
			{ return m_Node->GetNext(); };
		iterator& operator =(const iterator& src) throw() {
			m_List = src.m_List;
			m_Node = src.m_Node;
			return *this;
		};
		bool operator!() const throw()
			{ return (m_Node==NULL); };
		operator bool() const throw()
			{ return (m_Node!=NULL); };
		_N *operator *() const throw()
			{ return m_Node; };
		iterator& operator++() throw() {
			m_Node = m_Node->GetNext();
			return *this;
		};
		void operator++(int) throw()
			{ ++*this; };
		iterator& operator--() throw() {
			m_Node = m_Node->GetPrevious();
			return *this;
		};
		void operator--(int) throw()
			{ --*this; };
	};

	class iterator_safer {
	protected:
		ModList<_N> *m_List;
		_N *m_Previous;
		_N *m_Node;
		_N *m_Next;

	public:
		iterator_safer(void) {
			m_List = NULL;
			m_Previous = NULL;
			m_Node = NULL;
			m_Next = NULL;
		};
		iterator_safer(ModList<_N>& list) {
			m_List = &list;
			m_Node = list.GetFirst();
			if(m_Node) {
				m_Previous = m_Node->GetPrevious();
				m_Next = m_Node->GetNext();
			} else {
				m_Previous = NULL;
				m_Next = NULL;
			}
		};
		iterator_safer(const iterator_safer& src) throw() {
			m_List = src.m_List;
			m_Previous = src.m_Previous;
			m_Node = src.m_Node;
			m_Next = src.m_Next;
		};

		void begin(ModList<_N>& list) {
			m_List = &list;
			m_Node = list.GetFirst();
			if(m_Node) {
				m_Previous = m_Node->GetPrevious();
				m_Next = m_Node->GetNext();
			} else {
				m_Previous = NULL;
				m_Next = NULL;
			}
		};
		void end(ModList<_N>& list) {
			m_List = &list;
			m_Node = list.GetLast();
			if(m_Node) {
				m_Previous = m_Node->GetPrevious();
				m_Next = m_Node->GetNext();
			} else {
				m_Previous = NULL;
				m_Next = NULL;
			}
		};
		_N *previous(void)
			{ return m_Previous; };
		_N *next(void)
			{ return m_Next; };
		iterator_safer& operator =(const iterator& src) throw() {
			m_List = src.m_List;
			m_Node = src.m_Node;
			if(m_Node) {
				m_Previous = m_Node->GetPrevious();
				m_Next = m_Node->GetNext();
			} else {
				m_Previous = NULL;
				m_Next = NULL;
			}
			return *this;
		};
		bool operator!() const throw()
			{ return (m_Node==NULL); };
		operator bool() const throw()
			{ return (m_Node!=NULL); };
		_N *operator *() const throw()
			{ return m_Node; };
		iterator_safer& operator++() throw() {
			m_Node = m_Next;
			if(m_Node) {
				m_Previous = m_Node->GetPrevious();
				m_Next = m_Node->GetNext();
			} else {
				m_Previous = NULL;
				m_Next = NULL;
			}
			return *this;
		};
		void operator++(int) throw()
			{ ++*this; };
		iterator_safer& operator--() throw() {
			m_Node = m_Previous;
			if(m_Node) {
				m_Previous = m_Node->GetPrevious();
				m_Next = m_Node->GetNext();
			} else {
				m_Previous = NULL;
				m_Next = NULL;
			}
			return *this;
		};
		void operator--(int) throw()
			{ --*this; };
	};
};


template<class _N>
ModList<_N>::ModList<_N>(void) {
	m_Count = 0;
	m_bFree = true;
	m_FirstNode = NULL;
	m_LastNode = NULL;
}

template<class _N>
ModList<_N>::ModList<_N>(ModList<_N>& src) {
	m_Count = 0;
	m_bFree = true;
	m_FirstNode = NULL;
	m_LastNode = NULL;
	*this = src;
}

template<class _N>
ModList<_N>::~ModList<_N>(void) {
	Clear();
}

template<class _N>
ModList<_N>& ModList<_N>::operator =(ModList<_N>& src) {
	DeleteContents(true);
	Clear();
	CopyModList<_N>(src, *this);
	return *this;
}

template<class _N>
_N *ModList<_N>::GetFirst(void) {
	return m_FirstNode;
}

template<class _N>
_N *ModList<_N>::GetLast(void) {
	return m_LastNode;
}

template<class _N>
_N *ModList<_N>::AddLast(void) {
	_N *newNode = new _N;
	Append(newNode);
	return newNode;
}

template<class _N>
size_t ModList<_N>::GetCount(void) {
	return m_Count;
}

template<class _N>
_N *ModList<_N>::operator [](size_t index) {
	_N *node = m_FirstNode;
	while(node && index--)
		node = node->m_Next;
	return node;
}

template<class _N>
void ModList<_N>::Swap(_N *node1, _N *node2) {
	_N *tmp;
	if(node2 == node1->m_Previous) {
		if(node1->m_Next)
			node1->m_Next->m_Previous = node2;
		if(node2->m_Previous)
			node2->m_Previous->m_Next = node1;
		node1->m_Previous = node2->m_Previous;
		node2->m_Previous = node1;
		node2->m_Next = node1->m_Next;
		node1->m_Next = node2;
	} else if(node2 == node1->m_Next) {
		if(node1->m_Previous)
			node1->m_Previous->m_Next = node2;
		if(node2->m_Next)
			node2->m_Next->m_Previous = node1;
		node2->m_Previous = node1->m_Previous;
		node1->m_Previous = node2;
		node1->m_Next = node2->m_Next;
		node2->m_Next = node1;
	} else {
		if(node1->m_Previous)
			node1->m_Previous->m_Next = node2;
		if(node1->m_Next)
			node1->m_Next->m_Previous = node2;
		if(node2->m_Previous)
			node2->m_Previous->m_Next = node1;
		if(node2->m_Next)
			node2->m_Next->m_Previous = node1;
		tmp = node1->m_Previous;
		node1->m_Previous = node2->m_Previous;
		node2->m_Previous = tmp;
		tmp = node1->m_Next;
		node1->m_Next = node2->m_Next;
		node2->m_Next = tmp;
	}
	if(m_FirstNode == node1)
		m_FirstNode = node2;
	else if(m_FirstNode == node2)
		m_FirstNode = node1;
	if(m_LastNode == node1)
		m_LastNode = node2;
	else if(m_LastNode == node2)
		m_LastNode = node1;
}

template<class _N>
sint32 ModList<_N>::IndexOf(_N *node) {
	if(!m_FirstNode || !node)
		return -1;
	sint32 index = 0;
	_N *cur = m_FirstNode;
	while(cur && (cur!=node)) {
		cur = cur->m_Next;
		index++;
	}
	if(node != cur)
		return -1;
	else
		return index;
}

template<class _N>
void ModList<_N>::Append(_N *node) {
	node->m_Previous = m_LastNode;
	node->m_Next = NULL;
	if(m_LastNode)
		m_LastNode->m_Next = node;
	else
		m_FirstNode = node;
	m_LastNode = node;
	m_Count++;
}

template<class _N>
void ModList<_N>::Insert(_N *node) {
	node->m_Previous = NULL;
	node->m_Next = m_FirstNode;
	if(m_FirstNode)
		m_FirstNode->m_Previous = node;
	else
		m_LastNode = node;
	m_FirstNode = node;
	m_Count++;
}

template<class _N>
void ModList<_N>::Insert(_N *node, size_t index) {
	if(!index) {
		Insert(node);
		return;
	}
	_N *next = this->operator [](index);
	if(!next) {
		Append(node);
		return;
	}
	node->m_Previous = next->m_Previous;
	next->m_Previous->m_Next = node;
	next->m_Previous = node;
	node->m_Next = next;
	m_Count++;
}

template<class _N>
void ModList<_N>::DetachNode(_N *node) {
#ifdef _DEBUG
	if(!node)
		VDDEBUG("ModList::DetachNode detaching NULL Node\n");
#endif
	if(node->m_Previous)
		node->m_Previous->m_Next = node->m_Next;
	else
		m_FirstNode = node->m_Next;
	if(node->m_Next)
		node->m_Next->m_Previous = node->m_Previous;
	else
		m_LastNode = node->m_Previous;
	m_Count--;
}

template<class _N>
void ModList<_N>::DeleteNode(_N *node) {
	DetachNode(node);
	DoDeleteNode(node);
}

template<class _N>
inline void ModList<_N>::DoDeleteNode(_N *node) {
	if(m_bFree)
		delete node;
}

template<class _N>
void ModList<_N>::DeleteContents(bool bFree) {
	m_bFree = bFree;
}

template<class _N>
void ModList<_N>::Clear(void) {
	_N *current = m_FirstNode, *next;
	while(current) {
		next = current->GetNext();
		DoDeleteNode(current);
		current = next;
	}
	m_FirstNode = m_LastNode = NULL;
	m_Count = 0;
}


template<class _N>
void CopyModList(ModList<_N>& src, ModList<_N>& dst) {
	ModList<_N>::iterator it(src);
	while(it) {
		_N *newNode = new _N();
		*newNode = *(*it);
		dst.Append(newNode);
		it++;
	}
}

#endif // __MODLIST_H__