//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODSTREAM_H__
#define __MODSTREAM_H__

#include <tchar.h>
#include <vd2/system/vdtypes.h>
#include <vd2/system/VDString.h>
#include <vd2/system/Fraction.h>
#include "DubInfo.h"
#include "DubSource.h"
#include "AudioSource.h"
#include "VideoSource.h"
#include "AudioFilterSystem.h"
#include "ModList.h"


namespace nsModStream {
	enum eCommentType {
		cVoid			= 0x0000FFFF,
		cSpecific		= 0x00000000,
		cGeneric		= 0x00000001,
		cSoftware		= 0x00000002,
		cTitle			= 0x00000003,
		cDescription	= 0x00000004,
		cAuthor			= 0x00000005,
		cCopyright		= 0x00000006,
		cLicense		= 0x00000007,
		cCommissioned	= 0x00000008,
		cComment		= 0x00000009,
		cGenre			= 0x0000000A,
		cKeywords		= 0x0000000B,
		cDate			= 0x0000000C,
		cVersion		= 0x0000000D,
		cPerformer		= 0x0000000E,
		cAlbum			= 0x0000000F,
		cTrackNumber	= 0x00000010,
		cOrganisation	= 0x00000011,
		cLocation		= 0x00000012,
		cContact		= 0x00000013,
		cISRC			= 0x00000014,
		cURL			= 0x00000015,
		cName			= 0x00000016,
		cLanguage		= 0x00010000
	};

	struct cTypeName {
		eCommentType	m_Type;
		wchar_t			*m_Name;
		bool			m_bCanUse;
		bool			m_bCanEditTag;
		bool			m_bCanEditValue;
		wchar_t			*m_Tip;
	};

	static struct cTypeName s_CommentTypes[] = {
		{ cSpecific,		L"Specific",		false,	true,	true,	L"Comment specific to a container"},
		{ cGeneric,			L"Generic",			true,	true,	true,	L"Generic comment"},
		{ cSoftware,		L"Software",		false,	false,	false,	L"Software used to make the file"},
		{ cTitle,			L"Title",			true,	false,	true,	L"Title of the clip/track"},
		{ cDescription,		L"Description",		true,	false,	true,	L"Subject of the clip/track"},
		{ cLanguage,		L"Language",		true,	false,	true,	L"Language of the clip/track"},
		{ cAuthor,			L"Author",			true,	false,	true,	L"Author of the clip/track"},
		{ cCopyright,		L"Copyright",		true,	false,	true,	L"Copyright"},
		{ cLicense,			L"License",			true,	false,	true,	L"License"},
		{ cCommissioned,	L"Commissioner",	true,	false,	true,	L"Person or organisation that commissioned the subject of the file"},
		{ cComment,			L"Comment",			true,	false,	true,	L"Comment on the clip/track"},
		{ cGenre,			L"Genre",			true,	false,	true,	L"Genre"},
		{ cKeywords,		L"Keywords",		true,	false,	true,	L"Keywords refering to the file or its subject"},
		{ cDate,			L"Date",			true,	false,	true,	L"Creation date"},
		{ cVersion,			L"Version",			true,	false,	true,	L"Version"},
		{ cPerformer,		L"Performer",		true,	false,	true,	L"Performer"},
		{ cAlbum,			L"Album",			true,	false,	true,	L"Album"},
		{ cTrackNumber,		L"Track Number",	true,	false,	true,	L"Track Number"},
		{ cOrganisation,	L"Organisation",	true,	false,	true,	L"Organisation"},
		{ cLocation,		L"Location",		true,	false,	true,	L"Location"},
		{ cContact,			L"Contact",			true,	false,	true,	L"Contact"},
		{ cISRC,			L"ISRC",			true,	false,	true,	L"ISRC"},
		{ cURL,				L"URL",				true,	false,	true,	L"URL"},
		{ cName,			L"Name",			true,	false,	true,	L"Track name"},
		{ cVoid,			NULL,				false,	false,	false,	NULL}
	};

	sint16	GetCommentIndex(eCommentType type);
	wchar_t *GetCommentName(eCommentType type);
	wchar_t *GetCommentTip(eCommentType type);
	bool	CanUseComment(eCommentType type);
	bool	CanEditCommentTag(eCommentType type);
	bool	CanEditCommentValue(eCommentType type);
};


class ModStreamComment : public ModListNode<ModStreamComment> {
public:
	nsModStream::eCommentType	m_Type;
	VDStringW	m_Tag;
	VDStringW	m_Value;

	ModStreamComment(void);
	ModStreamComment(ModStreamComment& otherComment);
	ModStreamComment& operator =(ModStreamComment& otherComment);
	bool operator ==(ModStreamComment& otherComment);
	bool operator !=(ModStreamComment& otherComment);
};

void RemoveDuplicata(ModList<ModStreamComment>& comments);

class ModStreamChapter {
protected:
	uint32		m_TimeMS;
	VDStringW	m_Name;

public:
	ModStreamChapter(void);
	ModStreamChapter(const ModStreamChapter& otherChapter);

	uint32		GetTimeMS(void) const;
	VDStringW	GetTime(void) const;
	const wchar_t	*GetName(void) const;
	void		SetTimeMS(uint32 iTimeMS);
	bool		SetTime(const char *time);
	bool		SetTime(const wchar_t *time);
	void		SetName(const wchar_t *wszName);

	ModStreamChapter& operator =(const ModStreamChapter& otherChapter);
	bool operator==(const ModStreamChapter& otherChapter) const;
	bool operator!=(const ModStreamChapter& otherChapter) const;
	bool SeemsSimilar(const ModStreamChapter& otherChapter) const;
};

namespace std {
	template<> struct greater<ModStreamChapter> {
		bool operator()(ModStreamChapter& chapter_1, ModStreamChapter& chapter_2) const {
			return (chapter_1.GetTimeMS() < chapter_2.GetTimeMS());
		}
	};
}

class ModChapterList {
public:
	typedef std::list<ModStreamChapter>		tChapterList;
	typedef ModStreamChapter				tChapterNode;
	typedef tChapterList::value_type		value_type;
	typedef tChapterList::iterator			iterator;
	typedef tChapterList::const_iterator	const_iterator;
	typedef tChapterList::reference			reference;
	typedef tChapterList::const_reference	const_reference;
	typedef tChapterList::difference_type	difference_type;

	ModChapterList(void);
	ModChapterList(const ModChapterList& otherList);

	ModChapterList& operator =(const tChapterList& otherList);
	ModChapterList& operator =(const ModChapterList& otherList);

	tChapterNode&	AddChapter(void);
	tChapterNode&	AddChapter(const char *szTime, const wchar_t *wszName);
	tChapterNode&	AddChapter(uint32 iTimeMS, const wchar_t *wszName);
	tChapterNode&	AddChapter(tChapterNode& chapter);

	std::list<tChapterNode *>	GetPtrs(void);
	bool	GetAdjustChapters(void);
	void	SetAdjustChapters(bool bScale);

	void			Sort(void);

	size_t					size() const		{ return m_Chapters.size(); };
	void					clear()				{ m_Chapters.clear(); };
	bool					empty() const		{ return m_Chapters.empty(); };
	iterator				begin()				{ return m_Chapters.begin(); };
	const_iterator			begin() const		{ return m_Chapters.begin(); };
	iterator				end()				{ return m_Chapters.end(); };
	const_iterator			end() const			{ return m_Chapters.end(); };
	reference				front()				{ return m_Chapters.front(); };
	const_reference			front() const		{ return m_Chapters.front(); };
	reference				back()				{ return m_Chapters.back(); };
	const_reference			back() const		{ return m_Chapters.back(); };

	iterator erase(iterator it) { return m_Chapters.erase(it); };
	iterator erase(iterator it1, iterator it2) { return m_Chapters.erase(it1, it2); };
	void insert(iterator it, const value_type& v) { m_Chapters.insert(it, v); };

protected:
	tChapterList	m_Chapters;
	bool			m_bAdjustChapters;	// Wheter or not subsets influence chapters
};

bool ImportModStreamChapters(VDStringW fileName, ModChapterList& chapters);

//class AudioStream;

class ModStream : public ModListNode<ModStream> {
public:
	bool		m_bEnabled;
	bool		m_bToDemux;
	bool		m_bToPreview;
	vdrefptr<AudioSource>	m_Source;
	bool		m_bInputFile;
	ModList<ModStreamComment>	m_Comments;
	ModChapterList	m_Chapters;
	//bool		m_bIsNandubMP3VBR;	// Well I think you understand this one :P
	VDAudioFilterGraph	m_AudioFilterGraph;
	DubAudioOptions	m_Opts;		// Options
	WAVEFORMATEX	*m_ACompressionFormat;
	DWORD			m_ACompressionFormatSize;
	DWORD			m_StreamType;
	sint32			m_StreamIndex;

	ModStream(void);
	virtual ~ModStream(void);
	void Reset(bool bFreeSource, bool bFreeData, bool bResetValues);
};

class ModVideoStream {
public:
	IVDVideoSource	*m_Source;
	ModList<ModStreamComment>	m_Comments;
	ModChapterList	m_Chapters;
	DubVideoOptions	m_Opts;
	VDFraction	m_DAR;
	bool		m_bDARScaleUp;

	ModVideoStream(void);
	~ModVideoStream(void);
};


#endif // __MODSTREAM_H__