//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "MatroskaOutputFile.h"

#include <vd2/system/error.h>
#include <vd2/system/File64.h>
#include <vd2/system/log.h>
#include <vd2/Dita/resources.h>
#include "MatroskaReadHandler.h"
#include "MatroskaComments.h"
#include "FastWriteStream.h"
#include "ModStream.h"
#include "IDGenerator.h"
#include "ISO-639.h"
#include "resource.h"
#include "prefs.h"
#include <limits.h>
#include <time.h>


#define MATROSKA_WRITE_NEWTAGS 1

namespace {
	enum { kVDST_MatroskaOutputFile = 152 };
	
	enum {
		kVDM_UnexpectedBFrame,
		kVDM_PendingBFrames,
		kVDM_OutOfRangeTimecode,
		kVDM_SeekHead_NotEnoughSpaceReserved,
		kVDM_SeekHead_OutOfLuck
	};
}

using namespace LIBMATROSKA_NAMESPACE;

extern HINSTANCE g_hInst;
extern "C" unsigned long version_num;

#ifdef _DEBUG
sint32 ReferencedKaxCluster::m_TotalHandlers = 0;
#endif

MatroskaVideoOutputStream::MatroskaVideoOutputStream(MatroskaOutputFile *parent)
: AVIOutputStream(NULL) {
	m_Parent = parent;
	m_TrackNode = NULL;
	m_SamplesProcessed = 0;
	m_bEndReached = false;
}

MatroskaVideoOutputStream::~MatroskaVideoOutputStream(void) {
} 

void MatroskaVideoOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	if(cbBuffer)
		m_Parent->AddFrame(
			m_TrackNode
			, m_SampleRate.scale64ir(VD64(1000)*m_SamplesProcessed)
			, (binary *)pBuffer
			, cbBuffer
			, (flags & AVIIF_KEYFRAME ? true : false)
			, (flags & AVIIF_BFRAME ? true : false)
			, 0);
	m_SamplesProcessed += samples;
}

void MatroskaVideoOutputStream::setEndOfStream(void) {
	m_bEndReached = true;
}

void MatroskaVideoOutputStream::setStreamLag(sint32 samplesLag) {
	throw MyError("Calling setStreamLag for the video stream");
	//m_SamplesProcessed += samplesLag;
}

sint64 MatroskaVideoOutputStream::GetCurrentTimeMS(void) {
	return m_SampleRate.scale64ir(VD64(1000)*m_SamplesProcessed);
}

bool MatroskaVideoOutputStream::IsEnd(void) {
	return m_bEndReached;
}

//////////////////////////////////////////////////////////////////////

MatroskaOutputStream::MatroskaOutputStream(MatroskaOutputFile *parent)
: AVIOutputStream(NULL) {
	m_Parent = parent;
	m_TrackNode = NULL;
	m_SamplesProcessed = 0;
	m_bEndReached = false;
}

MatroskaOutputStream::~MatroskaOutputStream(void) {
}

void MatroskaOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	if(cbBuffer && (m_SamplesProcessed >= 0))
		m_Parent->AddFrame(m_TrackNode, m_SampleRate.scale64ir(VD64(1000)*m_SamplesProcessed), (binary *)pBuffer, cbBuffer, (flags & AVIIF_KEYFRAME ? true : false), false, 0);
	m_SamplesProcessed += samples;
}

void MatroskaOutputStream::setEndOfStream(void) {
	m_bEndReached = true;
}

void MatroskaOutputStream::setStreamLag(sint32 samplesLag) {
	m_SamplesProcessed += samplesLag;
}

sint64 MatroskaOutputStream::GetCurrentTimeMS(void) {
	return m_SampleRate.scale64ir(VD64(1000)*m_SamplesProcessed);
}

void MatroskaOutputStream::AddLagTimeUS(sint64 lagUS) {
	long lagSamples = VDRoundToLong((double)m_SampleRate.scale64r(lagUS) / 1000000.);
	m_SamplesProcessed += lagSamples;
}

void MatroskaOutputStream::AddLagSamples(sint64 lagSamples) {
	m_SamplesProcessed += lagSamples;
}

bool MatroskaOutputStream::IsEnd(void) {
	return m_bEndReached;
}

//////////////////////////////////////////////////////////////////////

MatroskaSubtitleOutputStream::MatroskaSubtitleOutputStream(MatroskaOutputFile *parent, eTextType inputType)
: MatroskaOutputStream(parent)
, m_TextUtil(inputType, kTextType_UTF8, false, false) {
}

MatroskaSubtitleOutputStream::~MatroskaSubtitleOutputStream(void) {
}

bool MatroskaSubtitleOutputStream::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IOverlappingOutput::GetInterfaceID()) {
		*destination = static_cast<IOverlappingOutput *>(this);
		return true;
	}
	return MatroskaOutputStream::QueryInterface(ID, destination);
}

void MatroskaSubtitleOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	if( (m_SamplesProcessed<0) && (m_SamplesProcessed+samples>0) ) {
		samples += m_SamplesProcessed;
		m_SamplesProcessed = 0;
	}
	VDStringW subtitle = m_TextUtil.GetText((void *)pBuffer, cbBuffer);
	if(subtitle.length() && (m_SamplesProcessed >= 0)) {
		VDStringA subUTF8 = VDTextWToU8(subtitle);
		m_Parent->AddFrame(m_TrackNode, m_SampleRate.scale64ir(VD64(1000)*m_SamplesProcessed), (binary *)subUTF8.c_str(), subUTF8.length()+1, true, false, m_SampleRate.scale64ir(VD64(1000)*samples));
	}
	m_SamplesProcessed += samples;
}

BOOL MatroskaSubtitleOutputStream::WriteData(LONG dwIndexFlags, LPVOID lpBuffer, LONG cbBuffer, LONG lStart, LONG lSamples) {
	BOOL bResult = TRUE;
	if(lStart < 0) {
		lSamples += lStart;
		lStart = 0;
	}
	VDStringW subtitle = m_TextUtil.GetText(lpBuffer, cbBuffer);
	if(subtitle.length() && (lSamples > 0)) {
		VDStringA subUTF8 = VDTextWToU8(subtitle);
		bResult = m_Parent->AddFrame(m_TrackNode, m_SampleRate.scale64ir(VD64(1000)*lStart), (binary *)subUTF8.c_str(), subUTF8.length()+1, true, false, m_SampleRate.scale64ir(VD64(1000)*lSamples));
	}
	m_SamplesProcessed = 0;
	return bResult;
}

//////////////////////////////////////////////////////////////////////

char MatroskaOutputFile::szME[] = "MatroskaOutputFile";

MatroskaOutputFile::MatroskaOutputFile(void) {
	m_VideoStream		= NULL;
	m_VideoSerial		= 1;
	m_CurrentRefCluster	= NULL;
	m_File				= NULL;
	m_HeadSize			= 0;
	m_SegmentSize		= 0;
	m_Infos				= NULL;
	m_InfoSize			= 0;
	m_Duration			= NULL;
	m_Tracks			= NULL;
	m_TracksSize		= 0;
	m_CueSize			= 0;
	// Reserve room for a minimal KaxSeekHead element :
	m_ReservedMetaSeekSize =
		KaxSeekHead::ClassInfos.GlobalId.Length
		+ KaxSeek::ClassInfos.GlobalId.Length + 1
		+ KaxSeekID::ClassInfos.GlobalId.Length + 1
		+ KaxSeekPosition::ClassInfos.GlobalId.Length + 1
		// ^--- Room reserved for the headers
		+ KaxSeekHead::ClassInfos.GlobalId.Length
		+ 8
		// ^--- Room reserved for the element ID and position in file
		- EbmlVoid::ClassInfos.GlobalId.Length
		// Don't count the EbmlVoid header size which will be reserved by default :p
		+ 2
		// This way we are sure there is enough place for the element
		// and a filler EbmlVoid element just after
		;
	m_VoidElementSize	= 0;
	m_MetaSeekSize		= 0;
	m_TagsSize			= 0;
	m_ChaptersSize		= 0;
	m_TotalClusterSize	= 0;
	m_LastClusterSize	= 0;
	m_ClusterSize		= 0;
	m_ClusterBaseTimeCode = 0;
	m_ClusterTimeLimit	= g_MKVPrefs.m_ClusterTimeLimit;
	m_TimecodeScale		= 1000000;
	m_iHintDurationMS	= 0;
	m_bWriteDefaultValues = false;
}

MatroskaOutputFile::~MatroskaOutputFile(void) {
	m_TrackList.DeleteContents(true);
	m_TrackList.Clear();
	m_Outputs.DeleteContents(true);
	m_Outputs.Clear();
	if(m_File)
		delete m_File;
}

IVDMediaOutputStream *MatroskaOutputFile::createVideoStream() {
	VDASSERT(!videoOut);
	if (!(videoOut = new_nothrow MatroskaVideoOutputStream(this)))
		throw MyMemoryError();

	return videoOut;
}

IVDMediaOutputStream *MatroskaOutputFile::createAudioStream() {
	VDASSERT(!audioOut);
	if (!(audioOut = new_nothrow MatroskaOutputStream(this)))
		throw MyMemoryError();

	return audioOut;
}

bool MatroskaOutputFile::initOutputStreams(ModVideoStream& videoStream, ModList<ModDubStream>& streams) {
	if(!(videoOut = new MatroskaVideoOutputStream(this)))	return false;

	m_VideoStream = &videoStream;

	sint32 serial = 2;
	ModList<ModDubStream>::iterator it(streams);
	while(it) {
		ModDubStream *curInStream = *it;
		MatroskaOutputStreamNode *curOutStream = m_Outputs.AddLast();
		ISubtitleHandler *pSubH;
		if(curInStream->m_Source->QueryInterface(ISubtitleHandler::GetInterfaceID(), (void **)&pSubH)) {
			// Output subtitles in UTF-8
			//pSubH->FormatAsSubtiDS(false);
			//pSubH->OutputInUTF8(false);
			curOutStream->m_Output = new MatroskaSubtitleOutputStream(this, pSubH->GetOutputType());
		} else {
			if(curInStream->m_Source->getSourceMainType() == SOURCE_TEXT)
				throw MyError("Subtitle stream %ld doesn't expose the ISubtitleHandler interface!"
					, serial);
			curOutStream->m_Output = new MatroskaOutputStream(this);
		}
		if(!curOutStream->m_Output)
			return FALSE;
		curOutStream->m_Input = curInStream;
		curOutStream->m_Comments = curInStream->m_Comments;
		curOutStream->m_Chapters = curInStream->m_Chapters;
		curInStream->m_Serial = serial;
		curOutStream->m_Serial = serial;
		serial++;
		it++;
	}

	return true;
}

IVDMediaOutputStream *MatroskaOutputFile::getOutputStream(uint32 serial) {
	m_OutputIterator.begin(m_Outputs);
	while(m_OutputIterator) {
		if((*m_OutputIterator)->m_Serial == serial)
			return (*m_OutputIterator)->m_Output;
		m_OutputIterator++;
	}
	throw MyError("Attempt to access unavailable output stream (%lu) in Matroska", serial);
}

void MatroskaOutputFile::setHintDuration(uint64 iDurationMS) {
	m_iHintDurationMS = iDurationMS;
}

bool MatroskaOutputFile::init(const wchar_t *szFile) {

	m_File = new WinIOCallback(szFile, MODE_CREATE, FILE_FLAG_SEQUENTIAL_SCAN);

	//Fill the EBML header
	EbmlHead FileHead;

	EDocType& MyDocType = GetChild<EDocType>(FileHead);
	*static_cast<EbmlString *>(&MyDocType) = "matroska";

	EDocTypeVersion& MyDocTypeVer = GetChild<EDocTypeVersion>(FileHead);
	*(static_cast<EbmlUInteger *>(&MyDocTypeVer)) = 1;

	EDocTypeReadVersion& MyDocTypeReadVer = GetChild<EDocTypeReadVersion>(FileHead);
	*(static_cast<EbmlUInteger *>(&MyDocTypeReadVer)) = 1;

	m_HeadSize = FileHead.Render(*m_File);

	//m_SegmentSize = m_FileSegment.Render(*m_File);
	m_SegmentSize = m_FileSegment.WriteHead(*m_File, 8, m_bWriteDefaultValues);

	m_Infos = &GetChild<KaxInfo>(m_FileSegment);
	m_Infos->EnableChecksum();
	m_MetaSeek.EnableChecksum();

	// Who decided to make a 128-bits UID *g*
	uint32 segmentUID[4];
	IDGenerator specialGen(true);
	segmentUID[0] = specialGen.GenerateID(false);
	segmentUID[1] = specialGen.GenerateID(false);
	segmentUID[2] = specialGen.GenerateID(false);
	segmentUID[3] = specialGen.GenerateID(false);
	KaxSegmentUID& MySegmentUID = GetChild<KaxSegmentUID>(*m_Infos);
	MySegmentUID.CopyBuffer((const binary *)segmentUID, sizeof segmentUID);

	KaxTimecodeScale& MyScale = GetChild<KaxTimecodeScale>(*m_Infos);
	*(static_cast<EbmlUInteger *>(&MyScale)) = m_TimecodeScale;	// Timecodes expressed in ms

	m_Duration = &GetChild<KaxDuration>(*m_Infos);
	*(static_cast<EbmlFloat *>(m_Duration)) = 0.;

	KaxDateUTC& MyDateUTC = GetChild<KaxDateUTC>(*m_Infos);
	MyDateUTC.SetEpochDate(time(NULL));

	if(videoOut && m_VideoStream) {
		ModList<ModStreamComment>::iterator it(m_VideoStream->m_Comments);
		while(it) {
			ModStreamComment *curComment = *it;
			if(curComment->m_Type == nsModStream::cTitle) {
				KaxTitle& MyTitle = GetChild<KaxTitle>(*m_Infos);
				*(static_cast<EbmlUnicodeString *>(&MyTitle)) = curComment->m_Value.c_str();
				break;
			}
			it++;
		}
	}

	VDStringW myText;
	myText = L"libebml-";
	myText += VDFastTextAToW(LIBEBML_NAMESPACE::EbmlCodeVersion.c_str());
	myText += L" & libmatroska-";
	myText += VDFastTextAToW(LIBMATROSKA_NAMESPACE::KaxCodeVersion.c_str());
	KaxMuxingApp& MyMuxingApp = GetChild<KaxMuxingApp>(*m_Infos);
	*(static_cast<EbmlUnicodeString *>(&MyMuxingApp)) = myText.c_str();

	char szSoft[256];
	memset(szSoft, 0, sizeof(szSoft));
	LoadString(g_hInst, IDS_TITLE_NOFILE, szSoft, sizeof(szSoft) - 1);
	char szSoftware[1024];
	memset(szSoftware, 0, sizeof(szSoftware));
	_snprintf(szSoftware, sizeof(szSoftware)-1, "%s (build %d/%s)", szSoft, version_num,
#ifdef _DEBUG
		"debug"
#else
		"release"
#endif
	);
	myText = VDTextAToW(szSoftware, -1);
	KaxWritingApp& MyWritingApp = GetChild<KaxWritingApp>(*m_Infos);
	*(static_cast<EbmlUnicodeString *>(&MyWritingApp)) = myText.c_str();

	m_InfoSize = m_Infos->Render(*m_File);
	m_MetaSeek.IndexThis(*m_Infos, m_FileSegment);

	// Reference the KaxInfo, KaxTracks, KaxCues, KasTags, KaxChapters and KaxSeekHead elements
	uint64 estimatedNbElements = 6;
	uint64 estimatedNbClusters = 0;

	if(g_MKVPrefs.m_IndexClusters && (m_ClusterTimeLimit>0) ) {
		uint64 nbClusters = m_iHintDurationMS / m_ClusterTimeLimit + 1;
		if(nbClusters > estimatedNbClusters)
			estimatedNbClusters = nbClusters;
	}
	// Assume a max error of 5% in the estimation of the number of clusters
	estimatedNbClusters = (estimatedNbClusters * 105) / 100;
	estimatedNbElements += estimatedNbClusters;
	uint64 averageSeekElementSize =
		KaxSeek::ClassInfos.GlobalId.Length + 1
		+ KaxSeekID::ClassInfos.GlobalId.Length + 1
		+ KaxSeekPosition::ClassInfos.GlobalId.Length + 1
		// ^--- Room reserved for the headers
		+ 4
		+ 5
		// ^--- Room reserved for the element ID and position in file
		// ID length : 4 bytes
		// File position coded on 5 bytes (i.e. assuming filesize < 35GB)
		;
	uint64 newReservedSize = estimatedNbElements * averageSeekElementSize + 2;
	newReservedSize += KaxSeekHead::ClassInfos.GlobalId.Length;
	if(newReservedSize >= EbmlVoid::ClassInfos.GlobalId.Length)
		newReservedSize -= EbmlVoid::ClassInfos.GlobalId.Length;
	else
		newReservedSize = 0;
	if(newReservedSize > m_ReservedMetaSeekSize)
		m_ReservedMetaSeekSize = newReservedSize;

	m_VoidElement.SetSize(m_ReservedMetaSeekSize);
	m_VoidElementSize = m_VoidElement.Render(*m_File);

	// Prepare the tracks
	m_Tracks = &GetChild<KaxTracks>(m_FileSegment);

	m_Tracks->EnableChecksum();
	m_Cues.EnableChecksum();

	MatroskaTrackNode *newTrackNode = NULL;

	m_FileName = szFile;

	KaxTrackEntry *lastEntry = NULL;

	bool bTitleSet = false;
	bool bLanguageSet = false;

	IDGenerator idGen(true);

	if(videoOut) {
		AVIStreamHeader_fixed infos = videoOut->getStreamInfo();
		BITMAPINFOHEADER *bi = (BITMAPINFOHEADER *)videoOut->getFormat();

		newTrackNode = new MatroskaTrackNode();
		newTrackNode->m_Serial = m_VideoSerial;
		newTrackNode->m_bDisableLacing = true;
		newTrackNode->m_bIndexFrames = true;
		newTrackNode->m_bHasBFrames = false;

		MatroskaVideoOutputStream *matroskaVideoOut = (MatroskaVideoOutputStream *)videoOut;
		if(lastEntry)
			newTrackNode->m_TrackEntry = &GetNextChild<KaxTrackEntry>(*m_Tracks, *lastEntry);
		else
			newTrackNode->m_TrackEntry = &GetChild<KaxTrackEntry>(*m_Tracks);
		lastEntry = newTrackNode->m_TrackEntry;

		matroskaVideoOut->m_TrackNode = newTrackNode;
		newTrackNode->m_SampleRate = VDFraction(infos.dwRate, infos.dwScale);
		matroskaVideoOut->m_SampleRate = newTrackNode->m_SampleRate;

		lastEntry->SetGlobalTimecodeScale(m_TimecodeScale);

		KaxTrackNumber& MyTrackNumber = GetChild<KaxTrackNumber>(*lastEntry);
		*(static_cast<EbmlUInteger *>(&MyTrackNumber)) = m_VideoSerial;

		uint32 newUID = idGen.GenerateID(true);
		KaxTrackUID& MyTrackUID = GetChild<KaxTrackUID>(*lastEntry);
		*(static_cast<EbmlUInteger *>(&MyTrackUID)) = newUID;

		SetTags(SOURCE_VIDEO, newUID, m_VideoStream->m_Comments, m_VideoStream->m_Source->asStream()->getSourceContainer() == SOURCE_MKV);

		*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(*lastEntry))) = track_video;

		KaxTrackFlagDefault& MyTrackDefault = GetChild<KaxTrackFlagDefault>(*lastEntry);
		*(static_cast<EbmlUInteger *>(&MyTrackDefault)) = 1;

		bool bRecompress = !(m_VideoStream->m_Opts.mode == DubVideoOptions::M_NONE);
		std::string codecID;
		std::vector<uint8> codecPrivate;
		IMatroskaStream *mStream;
		uint32 displayWidth = bi->biWidth;
		uint32 displayHeight = bi->biHeight;
		bool bRedoAR = true;
		if(!bRecompress && m_VideoStream->m_Source->asStream()->QueryInterface(IMatroskaStream::GetInterfaceID(), (void **)&mStream)) {
			codecID = mStream->GetCodecID();
			codecPrivate = mStream->GetCodecPrivate();
			/*KaxTrackEntry* aEntry = (KaxTrackEntry *)mStream->GetKaxTrackEntry();
			if(aEntry) {
				EbmlElement* aElement = aEntry->FindElt(KaxTrackVideo::ClassInfos);
				if(aElement) {
					KaxTrackVideo& MyVideo = *static_cast<KaxTrackVideo *>(aElement);
					EbmlElement* aDisplayWidth = MyVideo.FindElt(KaxVideoDisplayWidth::ClassInfos);
					EbmlElement* aDisplayHeight = MyVideo.FindElt(KaxVideoDisplayHeight::ClassInfos);
					if(aDisplayWidth && aDisplayHeight) {
						bRedoAR = false;
						displayWidth = LIBEBML_uint32(*static_cast<EbmlUInteger *>(aDisplayWidth));
						displayHeight = LIBEBML_uint32(*static_cast<EbmlUInteger *>(aDisplayHeight));
					}
				}
			}*/
		}
	
		if(bRedoAR) {
			VDFraction dar(m_VideoStream->m_DAR);
			VDFraction par(displayWidth, displayHeight);

			bool bResizeWidth = false;
			bool bResizeHeight = false;
			if(dar.getHi()) {
				if(dar >= par)
					bResizeWidth = true;
				else
					bResizeHeight = true;
				if(!m_VideoStream->m_bDARScaleUp) {
					bResizeWidth = !bResizeWidth;
					bResizeHeight = !bResizeHeight;
				}
			}

			if(bResizeWidth)
				displayWidth = dar.scale64r(displayHeight);
			else if(bResizeHeight)
				displayHeight = dar.scale64ir(displayWidth);
		}

		KaxCodecID& MyTrackCodecID = GetChild<KaxCodecID>(*lastEntry);
		if(codecID.size()) {
			*(static_cast<EbmlString *>(&MyTrackCodecID)) = codecID.c_str();

			if(codecPrivate.size()) {
				KaxCodecPrivate& MyPrivate = GetChild<KaxCodecPrivate>(*lastEntry);
				MyPrivate.CopyBuffer((binary*)&codecPrivate[0],codecPrivate.size());
			}
		} else {
			*(static_cast<EbmlString *>(&MyTrackCodecID)) = "V_MS/VFW/FOURCC";
			
			KaxCodecPrivate& MyPrivate = GetChild<KaxCodecPrivate>(*lastEntry);
			binary *cpriv = (binary *)malloc(bi->biSize);
			memcpy(cpriv, bi, bi->biSize);
			MyPrivate.SetBuffer((binary*)cpriv, bi->biSize);
		}

		lastEntry->EnableLacing(!newTrackNode->m_bDisableLacing);

		bTitleSet = false;
		bLanguageSet = false;
		if(m_VideoStream) {
			ModList<ModStreamComment>::iterator commentIt(m_VideoStream->m_Comments);
			while(commentIt) {
				ModStreamComment *curComment = *commentIt;
				if((curComment->m_Type == nsModStream::cName) && !bTitleSet) {
					bTitleSet = true;
					KaxTrackName& MyName = GetChild<KaxTrackName>(*lastEntry);
					*(static_cast<EbmlUnicodeString *>(&MyName)) = curComment->m_Value.c_str();
				} else if((curComment->m_Type > nsModStream::cLanguage) && !bLanguageSet) {
					int tableIndex = (curComment->m_Type - nsModStream::cLanguage) - 1;

					VDASSERT(ISO639_IsValidIndex_2(tableIndex));

					if(ISO639_IsValidIndex_2(tableIndex)) {
						bLanguageSet = true;
						KaxTrackLanguage& MyLanguage = GetChild<KaxTrackLanguage>(*lastEntry);
						*(static_cast<EbmlString *>(&MyLanguage)) = VDTextWToA(s_LanguageList[tableIndex].m_ISO639_2_Code, -1).c_str();
					}
				}
				commentIt++;
			}
		}

		uint32 minCache = 1;
		uint32 maxCache = 0/*1*/;
		if(!bRecompress) {
			if(m_VideoStream->m_Source->hasBFrames()) {
				newTrackNode->m_bHasBFrames = true;
				minCache = 2;
				//maxCache = 2;
			} else if(m_VideoStream->m_Source->isKeyframeOnly()) {
				minCache = 0;
				maxCache = 0;
			}
		}
		KaxTrackMinCache& MyMinCache = GetChild<KaxTrackMinCache>(*lastEntry);
		*(static_cast<EbmlUInteger *>(&MyMinCache)) = minCache;

		if(maxCache) {
			KaxTrackMaxCache& MyMaxCache = GetChild<KaxTrackMaxCache>(*lastEntry);
			*(static_cast<EbmlUInteger *>(&MyMaxCache)) = maxCache;
		}

		KaxTrackDefaultDuration& MyDefaultDuration = GetChild<KaxTrackDefaultDuration>(*lastEntry);
		*(static_cast<EbmlUInteger *>(&MyDefaultDuration)) = matroskaVideoOut->m_SampleRate.scale64ir(VD64(1000000000));

		KaxTrackVideo& MyTrackVideo = GetChild<KaxTrackVideo>(*lastEntry);

		//KaxVideoFrameRate& MyFramerate = GetChild<KaxVideoFrameRate>(MyTrackVideo);
		//*(static_cast<EbmlFloat *>(&MyFramerate)) = (double)matroskaVideoOut->m_SampleRate;

		KaxVideoPixelWidth& MyTrackPWidth = GetChild<KaxVideoPixelWidth>(MyTrackVideo);
		*(static_cast<EbmlUInteger *>(&MyTrackPWidth)) = bi->biWidth;

		KaxVideoPixelHeight& MyTrackPHeight = GetChild<KaxVideoPixelHeight>(MyTrackVideo);
		*(static_cast<EbmlUInteger *>(&MyTrackPHeight)) = bi->biHeight;

		KaxVideoDisplayWidth& MyTrackDWidth = GetChild<KaxVideoDisplayWidth>(MyTrackVideo);
		*(static_cast<EbmlUInteger *>(&MyTrackDWidth)) = displayWidth;
		MyTrackDWidth.SetDefaultSize(4);

		KaxVideoDisplayHeight& MyTrackDHeight = GetChild<KaxVideoDisplayHeight>(MyTrackVideo);
		*(static_cast<EbmlUInteger *>(&MyTrackDHeight)) = displayHeight;
		MyTrackDHeight.SetDefaultSize(4);

		//KaxVideoDisplayUnit& MyTrackDUnit = GetChild<KaxVideoDisplayUnit>(MyTrackVideo);
		//*(static_cast<EbmlUInteger *>(&MyTrackDUnit)) = 0;	// dimension given in pixels

		m_TrackList.Append(newTrackNode);
	}

	if(m_Outputs.GetCount()) {
		bool bFoundAudio = false;
		bool bFoundText = false;

		ModList<MatroskaOutputStreamNode>::iterator it(m_Outputs);
		while(it) {
			MatroskaOutputStreamNode *curNode = *it;
			WAVEFORMATEX *wfe = curNode->m_Input->m_AudioStream->GetFormat();

			newTrackNode = new MatroskaTrackNode();
			newTrackNode->m_Serial = curNode->m_Serial;
			newTrackNode->m_bDisableLacing = !g_MKVPrefs.m_bEnableLacing;
			if(!newTrackNode->m_bDisableLacing)
				newTrackNode->m_bDisableLacing = (curNode->m_Input->m_Source->getSourceMainType()==SOURCE_TEXT);
			newTrackNode->m_bIndexFrames = !g_MKVPrefs.m_CueOnlyVideo;
			newTrackNode->m_bHasBFrames = false;

			if(lastEntry)
				newTrackNode->m_TrackEntry = &GetNextChild<KaxTrackEntry>(*m_Tracks, *lastEntry);
			else
				newTrackNode->m_TrackEntry = &GetChild<KaxTrackEntry>(*m_Tracks);
			lastEntry = newTrackNode->m_TrackEntry;

			curNode->m_Output->m_TrackNode = newTrackNode;
			newTrackNode->m_SampleRate = VDFraction(wfe->nSamplesPerSec);
			curNode->m_Output->m_SampleRate = newTrackNode->m_SampleRate;

			lastEntry->SetGlobalTimecodeScale(m_TimecodeScale);

			KaxTrackNumber& MyTrackNumber = GetChild<KaxTrackNumber>(*lastEntry);
			*(static_cast<EbmlUInteger *>(&MyTrackNumber)) = curNode->m_Serial;

			uint32 newUID = idGen.GenerateID(true);
			KaxTrackUID& MyTrackUID = GetChild<KaxTrackUID>(*lastEntry);
			*(static_cast<EbmlUInteger *>(&MyTrackUID)) = newUID;

			SetTags(curNode->m_Input->m_Source->getSourceType(), newUID, curNode->m_Comments, curNode->m_Input->m_Source->getSourceContainer() == SOURCE_MKV);

			lastEntry->EnableLacing(!newTrackNode->m_bDisableLacing);

			bTitleSet = false;
			bLanguageSet = false;
			ModList<ModStreamComment>::iterator commentIt(curNode->m_Comments);
			while(commentIt) {
				ModStreamComment *curComment = *commentIt;
				if((curComment->m_Type == nsModStream::cName) && !bTitleSet) {
					bTitleSet = true;
					KaxTrackName& MyName = GetChild<KaxTrackName>(*lastEntry);
					*(static_cast<EbmlUnicodeString *>(&MyName)) = curComment->m_Value.c_str();
				} else if((curComment->m_Type > nsModStream::cLanguage) && !bLanguageSet) {
					int tableIndex = (curComment->m_Type - nsModStream::cLanguage) - 1;

					VDASSERT(ISO639_IsValidIndex_2(tableIndex));

					if(ISO639_IsValidIndex_2(tableIndex)) {
						bLanguageSet = true;
						KaxTrackLanguage& MyLanguage = GetChild<KaxTrackLanguage>(*lastEntry);
						*(static_cast<EbmlString *>(&MyLanguage)) = VDTextWToA(s_LanguageList[tableIndex].m_ISO639_2_Code, -1).c_str();
					}
				}
				commentIt++;
			}

			uint64 bitDepth = 16;
			bool bSetBitDepth = false;

			bool bComesFromAVI = (curNode->m_Input->m_Source->getSourceContainer() == SOURCE_AVI);
			bool bComesFromStdFile = (curNode->m_Input->m_Source->getSourceContainer() == SOURCE_STD_FILE);
			bool bRecompress = !(curNode->m_Input->m_Opts.mode == DubAudioOptions::M_NONE);
			std::string codecID;
			std::vector<uint8> codecPrivate;
			KaxTrackEntry *sourceTrackEntry = NULL;
			IMatroskaStream *mStream;
			if(!bRecompress && curNode->m_Input->m_Source->QueryInterface(IMatroskaStream::GetInterfaceID(), (void **)&mStream)) {
				codecID = mStream->GetCodecID();
				codecPrivate = mStream->GetCodecPrivate();
				sourceTrackEntry = static_cast<KaxTrackEntry *>(mStream->GetKaxTrackEntry());
			}

			// VBR handling, don't group frames
			if(!bRecompress) {
				curNode->m_Input->m_GroupFrames = 0;
				((AudioSource *)curNode->m_Input->m_Source)->extend(true, 0);
			}

			if(curNode->m_Input->m_Source->getSourceMainType() != SOURCE_TEXT) {
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(*lastEntry))) = track_audio;

				KaxTrackFlagDefault& MyTrackDefault = GetChild<KaxTrackFlagDefault>(*lastEntry);
				*(static_cast<EbmlUInteger *>(&MyTrackDefault)) = (bFoundAudio ? 0 : 1);
				bFoundAudio = true;
				
				KaxCodecID& MyTrackCodecID = GetChild<KaxCodecID>(*lastEntry);
				bool bUseCompatMode = true;
				if(codecID.size()) {
					bUseCompatMode = false;
					*(static_cast<EbmlString *>(&MyTrackCodecID)) = codecID.c_str();
					if(codecPrivate.size()) {
						KaxCodecPrivate& MyPrivate = GetChild<KaxCodecPrivate>(*lastEntry);
						MyPrivate.CopyBuffer((binary*)&codecPrivate[0],codecPrivate.size());
					}
					if(sourceTrackEntry) {
						EbmlElement *anElement = sourceTrackEntry->FindElt(KaxTrackAudio::ClassInfos);
						if(anElement) {
							KaxTrackAudio& MyAudioTrack = *static_cast<KaxTrackAudio*>(anElement);
							anElement = MyAudioTrack.FindElt(KaxAudioBitDepth::ClassInfos);
							if(anElement) {
								KaxAudioBitDepth& MyAudioDepth = *static_cast<KaxAudioBitDepth*>(anElement);
								bSetBitDepth = true;
								bitDepth = LIBEBML_uint32(MyAudioDepth);
							}
						}
					}
				} else
				switch(wfe->wFormatTag) {
				case WAVE_FORMAT_PCM:
					{
						*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_PCM/INT/LIT";
						bSetBitDepth = true;
						bitDepth = wfe->wBitsPerSample;
					}
					break;

				case WAVE_FORMAT_IEEE_FLOAT:
					{
						*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_PCM/FLOAT/IEEE";
						bSetBitDepth = true;
						bitDepth = wfe->wBitsPerSample;
					}
					break;

				case WAVE_FORMAT_MPEG:
					if(!bRecompress && (bComesFromStdFile || curNode->m_Input->m_bIsNandubMP3VBR)) {
						bUseCompatMode = false;
						if(wfe->cbSize >= sizeof(MPEG1WAVEFORMAT) - sizeof(WAVEFORMATEX) ) {
							MPEG1WAVEFORMAT *mpgformat = (MPEG1WAVEFORMAT *)wfe;
							if(mpgformat->fwHeadLayer & ACM_MPEG_LAYER3)
								*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MPEG/L3";
							else if(mpgformat->fwHeadLayer & ACM_MPEG_LAYER2)
								*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MPEG/L2";
							else if(mpgformat->fwHeadLayer & ACM_MPEG_LAYER1)
								*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MPEG/L1";
							else
								*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MPEG/L2";
						} else
							*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MPEG/L2";
					}
					break;

				case WAVE_FORMAT_MPEGLAYER3:
					if(!bRecompress && (bComesFromStdFile || curNode->m_Input->m_bIsNandubMP3VBR)) {
						bUseCompatMode = false;
						*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MPEG/L3";
					}
					break;

				case 0x2000:
					if(!bRecompress && bComesFromStdFile) {
						bUseCompatMode = false;
						*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_AC3";
					}
					break;

				case 0x2001:
					if(!bRecompress && bComesFromStdFile) {
						bUseCompatMode = false;
						*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_DTS";
					}
					break;

				case WAVE_FORMAT_EXTENSIBLE:
					if( (curNode->m_Input->m_Source->getSourceType() == SOURCE_AUDIO_VORBIS)
					&& !bRecompress) {
						bUseCompatMode = false;
						*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_VORBIS";
						IOggStream *pOgg;
						if(!curNode->m_Input->m_Source->QueryInterface(IOggStream::GetInterfaceID(), (void **)&pOgg))
							return FALSE;
						if(pOgg->GetNbHeaderPackets() != 3)
							return FALSE;
						KaxCodecPrivate& MyPrivate = GetChild<KaxCodecPrivate>(*lastEntry);
						binary *cpriv = (binary *)malloc(1);
						if(!cpriv)
							throw MyMemoryError();
						binary *cpriv_pos = cpriv;
						sint32 cpriv_size = 1;
						int i, nbHeaders;
						nbHeaders = pOgg->GetNbHeaderPackets();
						if(nbHeaders > 0x0100)
							throw MyError("Found %d Packet headers (maximum supported in matroska : 256) in an Ogg stream", nbHeaders);
						*cpriv_pos = (unsigned char)(nbHeaders-1);
						cpriv_pos++;
						for(i=0 ; i<nbHeaders ; i++) {
							ogg_packet *packet = pOgg->GetHeaderPacket(i);
							sint32 packetSize = packet->bytes;
							if(i+1<nbHeaders)
								cpriv_size += (packetSize / 255) + 1 + packetSize;
							else
								cpriv_size += packetSize;
							sint32 save_pos = cpriv_pos - cpriv;
							cpriv = (binary *)realloc(cpriv, cpriv_size);
							if(!cpriv)
								throw MyMemoryError();
							cpriv_pos = cpriv + save_pos;
							if(i+1<nbHeaders) {
								while(packetSize > 255) {
									*cpriv_pos = 0xFF;
									packetSize -= 255;
									cpriv_pos++;
								}
								*cpriv_pos = (unsigned char)packetSize;
								cpriv_pos++;
							}
						}
						for(i=0 ; i<nbHeaders ; i++) {
							ogg_packet *packet = pOgg->GetHeaderPacket(i);
							memcpy(cpriv_pos, packet->packet, packet->bytes);
							cpriv_pos += packet->bytes;
						}
						MyPrivate.SetBuffer((binary*)cpriv, cpriv_size);
					}
					break;

				default:
					bUseCompatMode = true;
					break;
				}
				if(bUseCompatMode) {
					*(static_cast<EbmlString *>(&MyTrackCodecID)) = "A_MS/ACM";

					KaxCodecPrivate& MyPrivate = GetChild<KaxCodecPrivate>(*lastEntry);
					binary *cpriv = (binary *)malloc(sizeof(WAVEFORMATEX)+wfe->cbSize);
					memcpy(cpriv, wfe, sizeof(WAVEFORMATEX)+wfe->cbSize);
					MyPrivate.SetBuffer((binary*)cpriv, sizeof(WAVEFORMATEX)+wfe->cbSize);
				}

				KaxTrackAudio& MyTrackAudio = GetChild<KaxTrackAudio>(*lastEntry);

				KaxAudioSamplingFreq& MyTrackFreq = GetChild<KaxAudioSamplingFreq>(MyTrackAudio);
				*(static_cast<EbmlFloat *>(&MyTrackFreq)) = (double)curNode->m_Output->m_SampleRate;
				MyTrackFreq.ValidateSize();

				KaxAudioChannels& MyTrackChannels = GetChild<KaxAudioChannels>(MyTrackAudio);
				*(static_cast<EbmlUInteger *>(&MyTrackChannels)) = wfe->nChannels;

				if(bSetBitDepth) {
					KaxAudioBitDepth& MyBitDepth = GetChild<KaxAudioBitDepth>(MyTrackAudio);
					*(static_cast<EbmlUInteger *>(&MyBitDepth)) = bitDepth;
				}
			} else {
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(*lastEntry))) = track_subtitle;

				KaxTrackFlagDefault& MyTrackDefault = GetChild<KaxTrackFlagDefault>(*lastEntry);
				*(static_cast<EbmlUInteger *>(&MyTrackDefault)) = (bFoundText ? 0 : 1);
				bFoundText = true;
				
				bool bRecompress = !(curNode->m_Input->m_Opts.mode == DubAudioOptions::M_NONE);
				std::string codecID;
				std::vector<uint8> codecPrivate;
				IMatroskaStream *mStream;
				if(!bRecompress && curNode->m_Input->m_Source->QueryInterface(IMatroskaStream::GetInterfaceID(), (void **)&mStream)) {
					codecID = mStream->GetCodecID();
					codecPrivate = mStream->GetCodecPrivate();
				}

				KaxCodecID& MyTrackCodecID = GetChild<KaxCodecID>(*lastEntry);

				if(codecID.size()) {
					*(static_cast<EbmlString *>(&MyTrackCodecID)) = codecID.c_str();
					if(codecPrivate.size()) {
						KaxCodecPrivate& MyPrivate = GetChild<KaxCodecPrivate>(*lastEntry);
						MyPrivate.CopyBuffer((binary*)&codecPrivate[0],codecPrivate.size());
					}
				} else
					*(static_cast<EbmlString *>(&MyTrackCodecID)) = "S_TEXT/UTF8";
			}

			if(curNode->m_Input->m_OffsetSamples < 0)
				curNode->m_Output->AddLagSamples(curNode->m_Input->m_SamplesPerBlock.scale64r(-curNode->m_Input->m_OffsetSamples));
			m_TrackList.Append(newTrackNode);
			it++;
		}
	} else if(audioOut) {
		delete audioOut;
		audioOut = NULL;
	}

	m_TracksSize = m_Tracks->Render(*m_File);
	m_MetaSeek.IndexThis(*m_Tracks, m_FileSegment);

	if(videoOut && m_VideoStream && m_VideoStream->m_Chapters.size()) {
		m_Chapters.EnableChecksum();

		KaxEditionEntry& MyEditionEntry = GetChild<KaxEditionEntry>(m_Chapters);

		IDGenerator chapterGen(true);

		ModChapterList::iterator chapterIt;
		for(chapterIt=m_VideoStream->m_Chapters.begin() ; chapterIt!=m_VideoStream->m_Chapters.end() ; ) {
			ModChapterList::tChapterNode *curChapter = &(*chapterIt);
			chapterIt++;
			ModChapterList::tChapterNode *nextChapter = NULL;
			if(chapterIt != m_VideoStream->m_Chapters.end())
				nextChapter = &(*chapterIt);

			KaxChapterAtom& MyAtom = AddNewChild<KaxChapterAtom>(MyEditionEntry);
			KaxChapterUID& MyUID = GetChild<KaxChapterUID>(MyAtom);
			*static_cast<EbmlUInteger *>(&MyUID) = chapterGen.GenerateID(true);

			uint64 chapterStart = m_TimecodeScale * curChapter->GetTimeMS();
			uint64 chapterEnd = (nextChapter ? m_TimecodeScale * nextChapter->GetTimeMS() : uint64(-1));

			KaxChapterTimeStart& MyChapStart = GetChild<KaxChapterTimeStart>(MyAtom);
			*static_cast<EbmlUInteger *>(&MyChapStart) = chapterStart;

			KaxChapterTimeEnd& MyChapEnd = GetChild<KaxChapterTimeEnd>(MyAtom);
			*static_cast<EbmlUInteger *>(&MyChapEnd) = chapterEnd;

			KaxChapterDisplay & MyDisplay = GetChild<KaxChapterDisplay>(MyAtom);
			KaxChapterString & MyChapString = GetChild<KaxChapterString>(MyDisplay);
			*static_cast<EbmlUnicodeString *>(&MyChapString) = curChapter->GetName();
		
			//KaxChapterLanguage& MyChapLang = GetChild<KaxChapterLanguage>(MyDisplay);
			//*static_cast<EbmlString *>(&MyChapLang) = "eng";
		}

		m_ChaptersSize = m_Chapters.Render(*m_File);
	}

	m_Cues.SetGlobalTimecodeScale(m_TimecodeScale);

	return TRUE;
}

void Matroska_RemoveEntries(KaxSeekHead& seekHead, const EbmlId& id) {
	int iElements = seekHead.ListSize();
	int iIndex = 0;
	for(iIndex=0 ; iIndex<iElements ; iIndex++) {
		EbmlElement *pChild = seekHead[iIndex];
		if(pChild && (pChild->Generic().GlobalId== KaxSeek::ClassInfos.GlobalId)) {
			KaxSeek& aSeekEntry = *static_cast<KaxSeek *>(pChild);
			if(aSeekEntry.IsEbmlId(id)) {
				seekHead.Remove(iIndex);
				iElements--;
				iIndex--;
			}
		}
	}
}

void MatroskaOutputFile::finalize(void) {
	ModList<MatroskaOutputStreamNode>::iterator it;

	FlushCluster(false);
	m_CueSize = m_Cues.Render(*m_File);
	m_MetaSeek.IndexThis(m_Cues, m_FileSegment);

	m_TagsSize = m_Tags.Render(*m_File);
	m_MetaSeek.IndexThis(m_Tags, m_FileSegment);

	// Update the duration information now that we know it
	sint64 maxDuration = 0;
	if(videoOut)
		maxDuration = ((MatroskaVideoOutputStream *)videoOut)->GetCurrentTimeMS();
	it.begin(m_Outputs);
	while(it) {
		sint64 trackDuration = (*it)->m_Output->GetCurrentTimeMS();
		if(trackDuration > maxDuration)
			maxDuration = trackDuration;
		it++;
	}
	*(static_cast<EbmlFloat *>(m_Duration)) = (double)maxDuration;
	m_Infos->UpdateSize();
	if(m_Infos->ElementSize() == m_InfoSize) {
		uint64 currentFilePos = m_File->getFilePointer();
		m_File->setFilePointer(m_Infos->GetElementPosition(), seek_beginning);
		m_Infos->Render(*m_File);
		m_File->setFilePointer(currentFilePos, seek_beginning);
	} else
		VDDEBUG("Warning : could't rewrite the KaxInfo element because its size changed (%lu became %I64u bytes)\n"
			, m_InfoSize, m_Infos->UpdateSize());

	if(m_ChaptersSize) {
		EbmlVoid voidElement;

		voidElement.Overwrite(m_Chapters, *m_File, true);

		EbmlElement *anElement = m_Chapters.FindElt(KaxEditionEntry::ClassInfos);
		if(anElement) {
			KaxEditionEntry& MyEditionEntry = *static_cast<KaxEditionEntry *>(anElement);

			uint32 iIndex = MyEditionEntry.ListSize();
			while(iIndex--) {
				bool bDeleteElement = false;
				anElement = MyEditionEntry[iIndex];
				if(anElement->Generic().GlobalId == KaxChapterAtom::ClassInfos.GlobalId) {
					KaxChapterAtom& MyAtom = *static_cast<KaxChapterAtom *>(anElement);

					anElement = MyAtom.FindElt(KaxChapterTimeStart::ClassInfos);
					if(anElement) {
						uint64 chapterStart = uint64(*static_cast<KaxChapterTimeStart *>(anElement));
						chapterStart /=  m_TimecodeScale;
						if(chapterStart >= maxDuration)
							bDeleteElement = true;
					}
					anElement = MyAtom.FindElt(KaxChapterTimeEnd::ClassInfos);
					if(anElement && !bDeleteElement) {
						uint64 chapterEnd = uint64(*static_cast<KaxChapterTimeEnd *>(anElement));
						if(chapterEnd == -1)
							*static_cast<EbmlUInteger *>(anElement) = maxDuration * m_TimecodeScale;
					}
				}

				if(bDeleteElement)
					MyEditionEntry.Remove(iIndex);
			}
		}

		voidElement.ReplaceWith(m_Chapters, *m_File, true);

		m_MetaSeek.IndexThis(m_Chapters, m_FileSegment);
	}

	if(m_MetaSeekSize = m_VoidElement.ReplaceWith(m_MetaSeek, *m_File, true))
		m_VoidElementSize = 0;
	else {
		VDLogAppMessage(kVDLogWarning, kVDST_MatroskaOutputFile, kVDM_SeekHead_NotEnoughSpaceReserved, 1, &m_VoidElementSize);
		uint64 iSeekPosition = m_File->getFilePointer();
		int counter = 7;
		while(--counter) {
			KaxSeekHead LastHopeSeekElement;
			LastHopeSeekElement.EnableChecksum();
			if((counter >= 6) && m_ChaptersSize)
				LastHopeSeekElement.IndexThis(m_Chapters, m_FileSegment);
			if(counter >= 5)
				LastHopeSeekElement.IndexThis(*m_Infos, m_FileSegment);
			if(counter >= 4)
				LastHopeSeekElement.IndexThis(*m_Tracks, m_FileSegment);
			if(counter >= 3)
				LastHopeSeekElement.IndexThis(m_Tags, m_FileSegment);
			if(counter >= 2)
				LastHopeSeekElement.IndexThis(m_Cues, m_FileSegment);
			LastHopeSeekElement.IndexThis(m_MetaSeek, m_FileSegment);
			KaxSeek *pSeekEntry = LastHopeSeekElement.FindFirstOf(KaxSeekHead::ClassInfos);
			if(pSeekEntry)
				*static_cast<EbmlUInteger *>(&GetChild<KaxSeekPosition>(*pSeekEntry)) = m_FileSegment.GetRelativePosition(iSeekPosition);
			if(m_VoidElement.ReplaceWith(LastHopeSeekElement, *m_File, true))
				break;
		}
		// Remove the elements in the SeekHead at the end of the file
		if(counter) {
			if((counter >= 6) && m_ChaptersSize)
				Matroska_RemoveEntries(m_MetaSeek, KaxChapters::ClassInfos.GlobalId);
			if(counter >= 5)
				Matroska_RemoveEntries(m_MetaSeek, KaxInfo::ClassInfos.GlobalId);
			if(counter >= 4)
				Matroska_RemoveEntries(m_MetaSeek, KaxTracks::ClassInfos.GlobalId);
			if(counter >= 3)
				Matroska_RemoveEntries(m_MetaSeek, KaxTags::ClassInfos.GlobalId);
			if(counter >= 2)
				Matroska_RemoveEntries(m_MetaSeek, KaxCues::ClassInfos.GlobalId);
		}
		m_MetaSeekSize = m_MetaSeek.Render(*m_File);
		if(!counter)
			VDLogAppMessage(kVDLogWarning, kVDST_MatroskaOutputFile, kVDM_SeekHead_OutOfLuck);
	}

	if(m_FileSegment.ForceSize(m_InfoSize + m_VoidElementSize + m_MetaSeekSize + m_TracksSize + m_ChaptersSize + m_TotalClusterSize + m_CueSize + m_TagsSize + m_SegmentSize - m_FileSegment.HeadSize()))
		m_FileSegment.OverwriteHead(*m_File);

	m_File->close();
}

MatroskaTrackNode *MatroskaOutputFile::GetTrackNode(sint32 serial) {
	m_TrackIterator.begin(m_TrackList);
	while(m_TrackIterator) {
		if((*m_TrackIterator)->m_Serial == serial)
			return (*m_TrackIterator);
		m_TrackIterator++;
	}
	return NULL;
}

bool MatroskaOutputFile::ChangeCluster(uint64 timeCode) {
	bool bNeedNewCluster = false;
	uint64 newBaseTimeCode = timeCode;
	sint64 iDiffTime = timeCode - m_ClusterBaseTimeCode;
	if(iDiffTime < 0)
		iDiffTime = -iDiffTime;
	if(!m_CurrentRefCluster
	|| (timeCode - m_ClusterBaseTimeCode > m_ClusterTimeLimit)
	|| (iDiffTime > 0x7FFF)
	//|| (m_ClusterSize > 2*1024*1024)
	)
		bNeedNewCluster = true;
	if(bNeedNewCluster) {
		sint64 currentTime = 0;
		if(videoOut && !((MatroskaVideoOutputStream *)videoOut)->IsEnd()) {
			currentTime = ((MatroskaVideoOutputStream *)videoOut)->GetCurrentTimeMS();
			if((currentTime > 0) && (currentTime < newBaseTimeCode))
				newBaseTimeCode = currentTime;
		}
		m_OutputIterator.begin(m_Outputs);
		while(m_OutputIterator) {
			MatroskaOutputStreamNode *curOutput = *m_OutputIterator;
			if(!curOutput->m_Output->IsEnd()) {
				currentTime = curOutput->m_Output->GetCurrentTimeMS();
				if((currentTime > 0) && (currentTime < newBaseTimeCode) && !curOutput->m_Output->IsEnd())
					newBaseTimeCode = currentTime;
			}
			m_OutputIterator++;
		}
		bool bHasPreviousCluster = false;
		if(m_CurrentRefCluster) {
			bHasPreviousCluster = true;
			FlushCluster(true);
		}
		m_CurrentRefCluster = new ReferencedKaxCluster(new KaxCluster());
		(**m_CurrentRefCluster).SetParent(m_FileSegment);
		(**m_CurrentRefCluster).EnableChecksum();
		if(bHasPreviousCluster && g_MKVPrefs.m_SetPreviousClusterSize) {
			KaxClusterPrevSize& MyPreviousSize = GetChild<KaxClusterPrevSize>(**m_CurrentRefCluster);
			*(static_cast<EbmlUInteger *>(&MyPreviousSize)) = m_LastClusterSize;
		}
		(**m_CurrentRefCluster).SetPreviousTimecode(newBaseTimeCode, m_TimecodeScale);
		(**m_CurrentRefCluster).InitTimecode(newBaseTimeCode, m_TimecodeScale);
		m_ClusterBaseTimeCode = newBaseTimeCode;

		VDASSERT(timeCode - newBaseTimeCode <= m_ClusterTimeLimit);

		m_ClusterSize = 0;
	}
	return bNeedNewCluster;
}

void MatroskaOutputFile::FlushCluster(bool bUpdateReferences) {
	if(m_CurrentRefCluster) {
		m_TrackIterator.begin(m_TrackList);
		while(m_TrackIterator) {
			MatroskaTrackNode *curNode = *m_TrackIterator;
			SetBlockInformation(curNode);
			if(curNode->m_BlockGroup)
				curNode->UpdateLastBackwardRefBlock(curNode->m_BlockGroup, m_CurrentRefCluster);
			m_TrackIterator++;
		}
		m_LastClusterSize = (**m_CurrentRefCluster).Render(*m_File, m_Cues);
		m_TotalClusterSize += m_LastClusterSize;
		if(g_MKVPrefs.m_IndexClusters)
			m_MetaSeek.IndexThis(**m_CurrentRefCluster, m_FileSegment);
		(**m_CurrentRefCluster).ReleaseFrames();
		std::list<ReferencedKaxCluster *>::const_iterator refIt;
		for(refIt=m_ReferencedClusters.begin() ; refIt!=m_ReferencedClusters.end() ; refIt++)
			(*refIt)->Release();
		m_ReferencedClusters.clear();
		if(bUpdateReferences) {
			m_TrackIterator.begin(m_TrackList);
			while(m_TrackIterator) {
				MatroskaTrackNode *curNode = *m_TrackIterator;
				curNode->m_BlockGroup = NULL;
				if(curNode->m_LastBackwardRefCluster) {
					curNode->m_LastBackwardRefCluster->AddRef();
					m_ReferencedClusters.push_back(curNode->m_LastBackwardRefCluster);
				}
				m_TrackIterator++;
			}
		}
		m_CurrentRefCluster->Release();
		m_CurrentRefCluster = NULL;
	}
}

bool MatroskaOutputFile::AddFrame(MatroskaTrackNode *trackNode, uint64 timeCode, binary *frameData, uint32 frameLength, bool bKeyFrame, bool bBFrame, uint64 duration) {

	VDASSERT(trackNode && frameData);

	if(bBFrame && !trackNode->m_bHasBFrames) {
		VDLogAppMessage(kVDLogError, kVDST_MatroskaOutputFile, kVDM_UnexpectedBFrame);
		return false;
	}

    try {
		KaxBlockGroup *newBlock = NULL;
		binary *aCopy = new binary[frameLength];
		memcpy(aCopy, frameData, frameLength);
		DataBuffer *data = new SimpleDataBuffer(aCopy, frameLength, 0);
		if(bBFrame) {
			MatroskaBFrame *newBFrame = new MatroskaBFrame(trackNode, data, timeCode, duration);
			trackNode->m_BFramesPending.push_back(newBFrame);
			return true;
		}
		bool bNeedNewBlock = !trackNode->m_BlockGroup;
		if(!bNeedNewBlock && trackNode->m_bDisableLacing) {
			bNeedNewBlock = true;
			if(m_CurrentRefCluster) {
				//if(!FlushBFrames(trackNode))
				//	return false;
				m_CurrentRefCluster->AddRef();
				m_ReferencedClusters.push_back(m_CurrentRefCluster);
				trackNode->UpdateLastBackwardRefBlock(trackNode->m_BlockGroup, m_CurrentRefCluster);
				if(trackNode->m_bIndexFrames &&
				(!g_MKVPrefs.m_CueSystem || trackNode->m_bLastFrameWasKey) )
					m_Cues.AddBlockGroup(*trackNode->m_BlockGroup);
			}
		}
		bNeedNewBlock |= ChangeCluster(timeCode);
		sint64 iDiffTime = timeCode - m_ClusterBaseTimeCode;
		if(iDiffTime < 0)
			iDiffTime = -iDiffTime;
		if(iDiffTime > 0x7FFF)
			VDLogAppMessage(kVDLogError, kVDST_MatroskaOutputFile, kVDM_OutOfRangeTimecode, 3, &timeCode, &trackNode->m_Serial, &m_ClusterBaseTimeCode);
		if(bNeedNewBlock) {
			SetBlockInformation(trackNode);
			trackNode->m_BlockGroup = &(**m_CurrentRefCluster).GetNewBlock();
		}
		trackNode->m_BlockDuration += duration;
		if(bKeyFrame) {
			if(trackNode->m_bHasBFrames && bNeedNewBlock)
				trackNode->m_BlockGroupPriority = 2;
			if(!trackNode->m_BlockGroup->AddFrame(*trackNode->m_TrackEntry, m_TimecodeScale * timeCode, *data)) {
				SetBlockInformation(trackNode);
				if(!FlushBFrames(trackNode))
					return false;
				m_CurrentRefCluster->AddRef();
				m_ReferencedClusters.push_back(m_CurrentRefCluster);
				trackNode->UpdateLastBackwardRefBlock(trackNode->m_BlockGroup, m_CurrentRefCluster);
				if(trackNode->m_bIndexFrames)
					m_Cues.AddBlockGroup(*trackNode->m_BlockGroup);
				trackNode->m_BlockGroup = NULL;
			} else if(!FlushBFrames(trackNode))
				return false;
			//(**m_CurrentRefCluster).AddFrame(*trackNode->m_TrackEntry, m_TimecodeScale * timeCode, *data, newBlock);
			//if(newBlock) {
			//	SetBlockDuration(trackNode);
			//	m_CurrentRefCluster->AddRef();
			//	m_ReferencedClusters.push_back(m_CurrentRefCluster);
			//	m_Cues.AddBlockGroup(*newBlock);
			//	trackNode->UpdateLastBackwardRefBlock(newBlock, m_CurrentRefCluster);
			//	trackNode->m_BlockGroup = newBlock;
			//}
			trackNode->m_bLastFrameWasKey = true;
		} else {
			if(trackNode->m_bHasBFrames && bNeedNewBlock)
				trackNode->m_BlockGroupPriority = 2;
			if(!trackNode->m_LastBackwardBlockGroup) {
				VDLog(kVDLogError, VDStringW(L"No frame reference found for a P frame"));
				return false;
			}
			bool bBlockFull = !trackNode->m_BlockGroup->AddFrame(*trackNode->m_TrackEntry, m_TimecodeScale * timeCode, *data, *trackNode->m_LastBackwardBlockGroup);
			if(bBlockFull) {
				SetBlockInformation(trackNode);
				if(!FlushBFrames(trackNode))
					return false;
				m_CurrentRefCluster->AddRef();
				m_ReferencedClusters.push_back(m_CurrentRefCluster);
				trackNode->UpdateLastBackwardRefBlock(trackNode->m_BlockGroup, m_CurrentRefCluster);
				if(trackNode->m_bIndexFrames && !g_MKVPrefs.m_CueSystem)
					m_Cues.AddBlockGroup(*trackNode->m_BlockGroup);
				trackNode->m_BlockGroup = NULL;
			} else if(!FlushBFrames(trackNode))
				return false;
			//(**m_CurrentRefCluster).AddFrame(*trackNode->m_TrackEntry, m_TimecodeScale * timeCode, *data, newBlock, *trackNode->m_LastBackwardBlockGroup);
			//if(newBlock) {
			//	SetBlockDuration(trackNode);
			//	m_CurrentRefCluster->AddRef();
			//	m_ReferencedClusters.push_back(m_CurrentRefCluster);
			//	m_Cues.AddBlockGroup(*newBlock);
			//	trackNode->UpdateLastBackwardRefBlock(newBlock, m_CurrentRefCluster);
			//	trackNode->m_BlockGroup = newBlock;
			//}
			trackNode->m_bLastFrameWasKey = false;
		}
		m_ClusterSize += frameLength;
	} catch (exception & Ex) {
		const char *error = Ex.what();
		VDLogAppMessage(kVDLogError, kVDST_MatroskaReadHandler, kVDM_LibError, 1, &error);
		return false;
	}
	return true;
}

void MatroskaOutputFile::SetBlockInformation(MatroskaTrackNode *trackNode) {
	if(trackNode->m_BlockGroup) {
		if(trackNode->m_BlockDuration) {
			trackNode->m_BlockGroup->SetBlockDuration(m_TimecodeScale * trackNode->m_BlockDuration);
			trackNode->m_BlockDuration = 0;
		}
		if(trackNode->m_bHasBFrames && trackNode->m_BlockGroupPriority) {
			KaxReferencePriority& RefPriority = GetChild<KaxReferencePriority>(*trackNode->m_BlockGroup);
			*(static_cast<EbmlUInteger *>(&RefPriority)) = trackNode->m_BlockGroupPriority;
		}
	}
}

bool MatroskaOutputFile::FlushBFrames(MatroskaTrackNode *trackNode) {
	if(!trackNode->m_bHasBFrames || !trackNode->m_BFramesPending.size())
		return true;

	if(!trackNode->m_LastBackwardBlockGroup || !trackNode->m_BlockGroup) {
		VDLog(kVDLogError, VDStringW(L"No frame reference found for a B frame"));
		return false;
	}
	KaxBlockGroup *newBlockGroup = &(**m_CurrentRefCluster).GetNewBlock();
	uint64 duration = 0;
	int nbFrames = trackNode->m_BFramesPending.size();
	while(nbFrames--) {
		MatroskaBFrame *frame = trackNode->m_BFramesPending.front();
		trackNode->m_BFramesPending.pop_front();

		duration += frame->m_Duration;
		if(!newBlockGroup->AddFrame(*trackNode->m_TrackEntry, m_TimecodeScale * frame->m_Timecode, *frame->m_Buffer, *trackNode->m_LastBackwardBlockGroup, *trackNode->m_BlockGroup)
		|| trackNode->m_bDisableLacing || !nbFrames) {
			if(duration) {
				newBlockGroup->SetBlockDuration(m_TimecodeScale * duration);
				duration = 0;
			}
			m_CurrentRefCluster->AddRef();
			m_ReferencedClusters.push_back(m_CurrentRefCluster);
			if(trackNode->m_bIndexFrames && !g_MKVPrefs.m_CueSystem)
				m_Cues.AddBlockGroup(*newBlockGroup);
			if(nbFrames)
				newBlockGroup = &(**m_CurrentRefCluster).GetNewBlock();
		}

		delete frame;
	}
	return true;
}

void MatroskaOutputFile::SetTags(uint32 type, uint32 UID, ModList<ModStreamComment>& comments, bool bSourceIsMatroska) {
	if(!comments.GetCount())
		return;

	KaxTag& MyTag = AddNewChild<KaxTag>(m_Tags);

	KaxTagTargets& MyTarget = GetChild<KaxTagTargets>(MyTag);
	*(static_cast<EbmlUInteger *>(&GetChild<KaxTagTrackUID>(MyTarget))) = UID;

	bool bTitleSet = false;
	ModList<ModStreamComment>::iterator commentIt(comments);
	while(commentIt) {
		ModStreamComment *curComment = *commentIt;
#if MATROSKA_WRITE_NEWTAGS
		VDString sTag;
		VDStringW sValue;
		bool bSkip = false;
		/* The name of the track is stored in the track info */
		if(curComment->m_Type == cName)
			bSkip = true;
		/* The title is stored in the segment info */
		if((MAIN_TYPE(type) == SOURCE_VIDEO) && (curComment->m_Type == cTitle) && !bTitleSet) {
			bTitleSet = true;
			bSkip = true;
		}
		/* Skip unwated comments (Software) */
		if(!CanUseComment(curComment->m_Type))
			bSkip = true;
		if((curComment->m_Type!=nsModStream::cGeneric)
			&& !GetMatroskaCommentTag(curComment->m_Type)
			&& !bSourceIsMatroska)
			bSkip = true;
		if(bSkip) {
			commentIt++;
			continue;
		}
		if(curComment->m_Type > cLanguage) {
			int tableIndex = (curComment->m_Type - cLanguage) - 1;

			VDASSERT(ISO639_IsValidIndex_2(tableIndex));

			sValue = s_LanguageList[tableIndex].m_ISO639_2_Code;
		} else
			sValue = curComment->m_Value;
		const wchar_t *commentTag = GetMatroskaCommentTag(curComment->m_Type);
		/* Special treatment for genres */
		if(curComment->m_Type == cGenre) {
			if(MAIN_TYPE(type) == SOURCE_VIDEO)
				commentTag = L"VIDEO_GENRE";
			else
				commentTag = L"AUDIO_GENRE";
		}
		if(!commentTag && curComment->m_Tag.size())
			commentTag = curComment->m_Tag.c_str();
		if(!commentTag)
			commentTag = GetMatroskaCommentTag(cComment);;
		if(commentTag) {
			KaxTagSimple& MySimpleTag = AddNewChild<KaxTagSimple>(MyTag);
			*(static_cast<EbmlUnicodeString *>(&AddNewChild<KaxTagName>(MySimpleTag))) = commentTag;
			*(static_cast<EbmlUnicodeString *>(&AddNewChild<KaxTagString>(MySimpleTag))) = sValue.c_str();
		}
#else /* 'old' tags */
		switch(curComment->m_Type) {
		case nsModStream::cKeywords :
			{
				KaxTagGeneral& MyGeneral = GetChild<KaxTagGeneral>(MyTag);
				*(static_cast<EbmlUnicodeString *>(&AddNewChild<KaxTagKeywords>(MyGeneral))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cDescription :
			{
				KaxTagGeneral& MyGeneral = GetChild<KaxTagGeneral>(MyTag);
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagSubject>(MyGeneral))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cGenre :
			{
				KaxTagGenres& MyGenre = GetChild<KaxTagGenres>(MyTag);
				if(MAIN_TYPE(type) == SOURCE_VIDEO) {
					KaxTagVideoGenre& MyVideoGenre = AddNewChild<KaxTagVideoGenre>(MyGenre);
					MyVideoGenre.CopyBuffer((const binary *)"\0\0", 2);
					*(static_cast<EbmlString *>(&GetChild<KaxTagSubGenre>(MyGenre))) = VDTextWToA(curComment->m_Value).c_str();
				} else
					*(static_cast<EbmlString *>(&AddNewChild<KaxTagAudioGenre>(MyGenre))) = VDTextWToA(curComment->m_Value).c_str();
			}
			break;
		case nsModStream::cGeneric :
			{
				KaxTagMultiComment& MyMultiComment = AddNewChild<KaxTagMultiComment>(MyTag);
				VDStringW tag = curComment->m_Tag;
				if(!ModTextUtil::IsBlank(tag))
					*(static_cast<EbmlString *>(&GetChild<KaxTagMultiCommentName>(MyMultiComment))) = VDTextWToA(tag).c_str();
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiCommentComments>(MyMultiComment))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cComment :
			{
				KaxTagMultiComment& MyMultiComment = AddNewChild<KaxTagMultiComment>(MyTag);
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiCommentComments>(MyMultiComment))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cISRC :
			{
				KaxTagMultiIdentifier& MyMultiIdentifier = GetChild<KaxTagMultiIdentifier>(MyTag);
				KaxTagIdentifier& MyIdentifier = AddNewChild<KaxTagIdentifier>(MyMultiIdentifier);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiIdentifierType>(MyIdentifier))) = KaxTagMultiIdentifierType_ISRC;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiIdentifierString>(MyIdentifier))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cCommissioned :
			{
				KaxTagMultiEntity& MyMultiEntity = GetChild<KaxTagMultiEntity>(MyTag);
				KaxTagEntity& MyEntity = AddNewChild<KaxTagEntity>(MyMultiEntity);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiEntityType>(MyEntity))) = KaxTagMultiEntitiesType_CommissionedBy;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiEntityName>(MyEntity))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cAuthor :
			{
				KaxTagMultiEntity& MyMultiEntity = GetChild<KaxTagMultiEntity>(MyTag);
				KaxTagEntity& MyEntity = AddNewChild<KaxTagEntity>(MyMultiEntity);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiEntityType>(MyEntity))) = KaxTagMultiEntitiesType_Composer;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiEntityName>(MyEntity))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cPerformer :
			{
				KaxTagMultiEntity& MyMultiEntity = GetChild<KaxTagMultiEntity>(MyTag);
				KaxTagEntity& MyEntity = AddNewChild<KaxTagEntity>(MyMultiEntity);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiEntityType>(MyEntity))) = KaxTagMultiEntitiesType_LeadPerformerSoloist;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiEntityName>(MyEntity))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cCopyright :
			{
				KaxTagMultiLegal& MyMultiLegal = GetChild<KaxTagMultiLegal>(MyTag);
				KaxTagLegal& MyLegal = AddNewChild<KaxTagLegal>(MyMultiLegal);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiLegalType>(MyLegal))) = KaxTagMultiLegalType_Copyright;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiLegalContent>(MyLegal))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cLicense :
			{
				KaxTagMultiLegal& MyMultiLegal = GetChild<KaxTagMultiLegal>(MyTag);
				KaxTagLegal& MyLegal = AddNewChild<KaxTagLegal>(MyMultiLegal);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiLegalType>(MyLegal))) = KaxTagMultiLegalType_TermsOfUse;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiLegalContent>(MyLegal))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cAlbum :
			{
				KaxTagMultiTitle& MyMultiTitle = GetChild<KaxTagMultiTitle>(MyTag);
				KaxTagTitle& MyTitle = AddNewChild<KaxTagTitle>(MyMultiTitle);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiTitleType>(MyTitle))) = KaxTagMultiTitleType_AlbumMovieShowTitle;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiTitleName>(MyTitle))) = curComment->m_Value.c_str();
			}
			break;
		case nsModStream::cTitle :
			{
				KaxTagMultiTitle& MyMultiTitle = GetChild<KaxTagMultiTitle>(MyTag);
				KaxTagTitle& MyTitle = AddNewChild<KaxTagTitle>(MyMultiTitle);
				*(static_cast<EbmlUInteger *>(&GetChild<KaxTagMultiTitleType>(MyTitle))) = KaxTagMultiTitleType_TrackTitle;
				*(static_cast<EbmlUnicodeString *>(&GetChild<KaxTagMultiTitleName>(MyTitle))) = curComment->m_Value.c_str();
			}
			break;
		}
		if(curComment->m_Type > nsModStream::cLanguage) {
			int tableIndex = (curComment->m_Type - nsModStream::cLanguage) - 1;

			VDASSERT(ISO639_IsValidIndex_2(tableIndex));

			if(ISO639_IsValidIndex_2(tableIndex)) {
				KaxTagGeneral& MyGeneral = GetChild<KaxTagGeneral>(MyTag);
				*(static_cast<EbmlString *>(&GetChild<KaxTagLanguage>(MyGeneral))) = VDTextWToA(s_LanguageList[tableIndex].m_ISO639_2_Code, -1).c_str();
			}
		}
#endif
		commentIt++;
	}
}

//////////////////////////////////////////////////////////////////////

MatroskaTrackNode::~MatroskaTrackNode(void) {
	if(m_LastBackwardRefCluster)
		m_LastBackwardRefCluster->Release();

	int nbElements = m_BFramesPending.size();
	if(nbElements)
		VDLogAppMessage(kVDLogWarning, kVDST_MatroskaOutputFile, kVDM_PendingBFrames, 1, &nbElements);
	while(nbElements--) {
		MatroskaBFrame *frame = m_BFramesPending.front();
		m_BFramesPending.pop_front();
		frame->m_Buffer->FreeBuffer(*frame->m_Buffer);
		delete frame->m_Buffer;
		delete frame;
	}
}