//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MATROSKAOUTPUTFILE_H__
#define __MATROSKAOUTPUTFILE_H__

#include "AVIOutput.h"
#include "ModList.h"
#include "ModTextUtil.h"
#include <list>

// There are conflicting typedefs between VirtualDub and libebml
// Fool the compiler and make libebml/libmatroska typedef to
// LIBEBML_int8/16/32/64 instead of int8/16/32/64 that are already
// defined by VirtualDub
#define	uint8	LIBEBML_uint8
#define	int8	LIBEBML_int8
#define	uint16	LIBEBML_uint16
#define	int16	LIBEBML_int16
#define	uint32	LIBEBML_uint32
#define	int32	LIBEBML_int32
#define	uint64	LIBEBML_uint64
#define	int64	LIBEBML_int64

#include "src/platform/win32/WinIOCallback.h"
#include "ebml/EbmlHead.h"
#include "ebml/EbmlSubHead.h"
#include "ebml/EbmlVoid.h"
#include "matroska/FileKax.h"
#include "matroska/KaxSegment.h"
#include "matroska/KaxTracks.h"
#include "matroska/KaxTrackEntryData.h"
#include "matroska/KaxTrackAudio.h"
#include "matroska/KaxTrackVideo.h"
#include "matroska/KaxCluster.h"
#include "matroska/KaxClusterData.h"
#include "matroska/KaxBlock.h"
#include "matroska/KaxInfo.h"
#include "matroska/KaxInfoData.h"
#include "matroska/KaxSeekHead.h"
#include "matroska/KaxTags.h"
#include "matroska/KaxTag.h"
#include "matroska/KaxTagMulti.h"
#include "matroska/KaxChapters.h"
#include "ebml/EbmlVersion.h"
#include "matroska/KaxVersion.h"

#undef uint8
#undef int8
#undef uint16
#undef int16
#undef uint32
#undef int32
#undef uint64
#undef int64

using namespace LIBMATROSKA_NAMESPACE;
using namespace LIBEBML_NAMESPACE;

class ReferencedKaxCluster {
protected:
#ifdef _DEBUG
	static sint32	m_TotalHandlers;
#endif
	KaxCluster	*m_Cluster;
	sint32		m_References;

public:
	ReferencedKaxCluster(KaxCluster *cluster) {
#ifdef _DEBUG
		m_TotalHandlers++;
		VDDEBUG("Total ReferencedKaxCluster : %d\n", m_TotalHandlers);
#endif
		m_Cluster = cluster;
		m_References = 1;
	}

	virtual ~ReferencedKaxCluster(void) {
#ifdef _DEBUG
		m_TotalHandlers--;
		VDDEBUG("Total ReferencedKaxCluster : %d\n", m_TotalHandlers);
		if(m_References > 0)
			VDDEBUG("Deleting a ReferencedKaxCluster still being referenced !!\n");
#endif
		if(m_Cluster)
			delete m_Cluster;
	}

	KaxCluster *GetCluster(void) {
		return m_Cluster;
	}

	sint32 AddRef(void) {
		return ++m_References;
	}

	sint32 Release(void) {
		int refs = --m_References;
		if(!m_References)
			delete this;
		return refs;
	}

	KaxCluster& operator *(void) {
		return *m_Cluster;
	}
};

class MatroskaOutputFile;
class MatroskaTrackNode;

class MatroskaVideoOutputStream : public AVIOutputStream {
friend MatroskaOutputFile;
protected:
	sint64				m_SamplesProcessed;
	bool				m_bEndReached;
	MatroskaOutputFile	*m_Parent;
	MatroskaTrackNode	*m_TrackNode;
	VDFraction			m_SampleRate;

public:
	MatroskaVideoOutputStream(MatroskaOutputFile *parent);
	~MatroskaVideoOutputStream(void);
	void write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples);
	void setEndOfStream(void);
	void setStreamLag(sint32 samplesLag);

	sint64 GetCurrentTimeMS(void);
	bool IsEnd(void);
};


class MatroskaOutputStream : public AVIOutputStream {
friend MatroskaOutputFile;
protected:
	sint64				m_SamplesProcessed;
	bool				m_bEndReached;
	MatroskaOutputFile	*m_Parent;
	MatroskaTrackNode	*m_TrackNode;
	VDFraction			m_SampleRate;

public:
	MatroskaOutputStream(MatroskaOutputFile *parent);
	~MatroskaOutputStream(void);
	void write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples);
	void setEndOfStream(void);
	void setStreamLag(sint32 samplesLag);

	sint64 GetCurrentTimeMS(void);
	void AddLagTimeUS(sint64 lagUS);
	void AddLagSamples(sint64 lagSamples);
	bool IsEnd(void);
};


class MatroskaSubtitleOutputStream : public MatroskaOutputStream, IOverlappingOutput {
friend MatroskaOutputFile;
protected:
	ModTextUtil	m_TextUtil;

public:
	MatroskaSubtitleOutputStream(MatroskaOutputFile *parent, eTextType inputType);
	~MatroskaSubtitleOutputStream(void);

	bool	QueryInterface(InterfaceID ID, void **destination);

	void write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples);
	BOOL WriteData(LONG dwIndexFlags, LPVOID lpBuffer, LONG cbBuffer, LONG lStart, LONG lSamples);
};


class MatroskaOutputStreamNode : public ModListNode<MatroskaOutputStreamNode> {
public:
	sint32			m_Serial;
	ModDubStream	*m_Input;
	MatroskaOutputStream	*m_Output;
	ModList<ModStreamComment>	m_Comments;
	ModChapterList	m_Chapters;

	MatroskaOutputStreamNode(void) {
		m_Serial = -1;
		m_Input = NULL;
		m_Output = NULL;
	};

	virtual ~MatroskaOutputStreamNode(void) {
		if(m_Output)
			delete m_Output;
		m_Comments.DeleteContents(true);
		m_Comments.Clear();
		m_Chapters.clear();
	};
};

class FastWriteStream;

class MatroskaBFrame {
public:
	MatroskaTrackNode	*m_Track;
	DataBuffer			*m_Buffer;
	uint64				m_Timecode;
	uint64				m_Duration;

	MatroskaBFrame(MatroskaTrackNode *trackNode=NULL, DataBuffer *buffer=NULL, uint64 timeCode=0, uint64 duration=0)
	: m_Track(trackNode)
	, m_Buffer(buffer)
	, m_Timecode(timeCode)
	, m_Duration(duration) {
	};
};

class MatroskaTrackNode : public ModListNode<MatroskaTrackNode> {
public:
	sint32				m_Serial;
	KaxBlockGroup		*m_BlockGroup;
	uint32				m_BlockGroupPriority;
	KaxTrackEntry		*m_TrackEntry;
	VDFraction			m_SampleRate;
	bool				m_bDisableLacing;
	bool				m_bIndexFrames;
	bool				m_bLastFrameWasKey;
	uint64				m_BlockDuration;
	KaxBlockGroup			*m_LastBackwardBlockGroup;
	ReferencedKaxCluster	*m_LastBackwardRefCluster;
	bool				m_bHasBFrames;
	std::list<MatroskaBFrame *>	m_BFramesPending;

	MatroskaTrackNode(void) {
		m_Serial = -1;
		m_BlockGroup = NULL;
		m_BlockGroupPriority = 0;
		m_TrackEntry = NULL;
		m_bDisableLacing = true;
		m_BlockDuration = 0;
		m_LastBackwardBlockGroup = NULL;
		m_LastBackwardRefCluster = NULL;
		m_bIndexFrames = false;
		m_bLastFrameWasKey = false;
		m_bHasBFrames = false;
	};

	~MatroskaTrackNode(void);
	
	void UpdateLastBackwardRefBlock(KaxBlockGroup *blockGroup, ReferencedKaxCluster *refCluster) {
		// Add our reference now in case we are still in the same cluster
		refCluster->AddRef();
		if(m_LastBackwardRefCluster)
			m_LastBackwardRefCluster->Release();
		m_LastBackwardRefCluster = refCluster;
		m_LastBackwardBlockGroup = blockGroup;
	};
};


class MatroskaOutputFile : public AVIOutput {
friend MatroskaVideoOutputStream;
friend MatroskaOutputStream;
friend MatroskaSubtitleOutputStream;
protected:
	static char szME[];
	IOCallback		*m_File;
	uint32			m_HeadSize;
	KaxSegment		m_FileSegment;
	uint32			m_SegmentSize;
	KaxInfo			*m_Infos;
	uint32			m_InfoSize;
	KaxDuration		*m_Duration;
	KaxTracks		*m_Tracks;
	uint32			m_TracksSize;
	std::list<ReferencedKaxCluster *>	m_ReferencedClusters;
	uint64			m_ReservedMetaSeekSize;
	EbmlVoid		m_VoidElement;
	uint32			m_VoidElementSize;
	KaxCues			m_Cues;
	uint32			m_CueSize;
	KaxSeekHead		m_MetaSeek;
	uint32			m_MetaSeekSize;
	KaxTags			m_Tags;
	uint32			m_TagsSize;
	KaxChapters		m_Chapters;
	uint32			m_ChaptersSize;
	ReferencedKaxCluster	*m_CurrentRefCluster;
	uint64			m_TotalClusterSize;
	uint64			m_LastClusterSize;
	uint64			m_ClusterSize;
	uint64			m_ClusterTimeLimit;
	uint64			m_ClusterBaseTimeCode;
	ModVideoStream		*m_VideoStream;
	sint32				m_VideoSerial;
	ModList<MatroskaOutputStreamNode>	m_Outputs;
	ModList<MatroskaOutputStreamNode>::iterator	m_OutputIterator;
	ModList<MatroskaTrackNode>	m_TrackList;
	ModList<MatroskaTrackNode>::iterator	m_TrackIterator;
	VDStringW			m_FileName;
	uint64			m_TimecodeScale;
	uint64			m_iHintDurationMS;
	bool			m_bWriteDefaultValues;

	MatroskaTrackNode	*GetTrackNode(sint32 serial);
	bool				ChangeCluster(uint64 timeCode);
	bool				AddFrame(MatroskaTrackNode* trackNode, uint64 timeCode, binary *frameData, uint32 frameLength, bool bKeyFrame, bool bBFrame, uint64 duration);
	void				SetBlockInformation(MatroskaTrackNode *trackNode);
	bool				FlushBFrames(MatroskaTrackNode *trackNode);
	void				FlushCluster(bool bUpdateReferences);
	void				SetTags(uint32 type, uint32 UID, ModList<ModStreamComment>& comments, bool bSourceIsMatroska);
	//void				SetChapters(uint32 UID, ModList<ModStreamComment>& comments);

public:
	MatroskaOutputFile(void);
	virtual ~MatroskaOutputFile(void);

	IVDMediaOutputStream *createVideoStream();
	IVDMediaOutputStream *createAudioStream();
	bool initOutputStreams(ModVideoStream& videoStream, ModList<ModDubStream>& streams);
	IVDMediaOutputStream *getOutputStream(uint32 serial);
	void setHintDuration(uint64 iDurationMS);
	bool init(const wchar_t *szFile);
	void finalize(void);
};

#endif // __MATROSKAOUTPUTFILE_H__