//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "MatroskaOutputFileSystem.h"
#include "MatroskaOutputFile.h"


MatroskaOutputFileSystem::MatroskaOutputFileSystem()
	: mbAllowCaching(false)
	, mCurrentSegment(0)
{
	m_VideoStream = NULL;
	m_bAcceptsVideo = true;
	m_bLastSegment = false;
}

MatroskaOutputFileSystem::~MatroskaOutputFileSystem() {
	m_Streams.DeleteContents(true);
	m_Streams.Clear();
}

AVIOutput *MatroskaOutputFileSystem::CreateSegment() {
	vdautoptr<MatroskaOutputFile> pOutput(new MatroskaOutputFile);

	VDStringW s(mSegmentBaseName);

	if (mSegmentDigits) {
		s += VDFastTextPrintfW(L".%0*d", mSegmentDigits, mCurrentSegment++);
		VDFastTextFree();
		s += mSegmentExt;

		//pOutput->setSegmentHintBlock(true, NULL, 1);
	}

	ModList<ModDubStream> tmpList;
	tmpList.DeleteContents(false);
	ModList<ModAVIDubStream>::iterator it(m_Streams);
	while(it) {
		tmpList.Append((*it)->m_Stream);
		it++;
	}
	pOutput->initOutputStreams(*m_VideoStream, tmpList);

	uint64 iHintDurationMS = 0;

	if(!mVideoFormat.empty()) {
		IVDMediaOutputStream *pVideoOut = pOutput->getVideoOutput();
		pVideoOut->setFormat(&mVideoFormat[0], mVideoFormat.size());
		pVideoOut->setStreamInfo(mVideoStreamInfo);
		AVIStreamHeader_fixed& streamInfo = mVideoStreamInfo;
		uint64 iMS = (VD64(1000) * streamInfo.dwLength * streamInfo.dwScale) / streamInfo.dwRate + 1;
		if(iMS > iHintDurationMS)
			iHintDurationMS = iMS;
	}

	it.begin(m_Streams);
	ModAVIDubStream *curStream = NULL;
	while(it) {
		curStream = *it;
		IVDMediaOutputStream *streamOut = pOutput->getOutputStream(curStream->m_Stream->m_Serial);
		if(!streamOut)
			throw MyError("Unable to retrieve output stream %lu", curStream->m_Stream->m_Serial);
		if (!curStream->m_Format.empty()) {
			streamOut->setFormat(&curStream->m_Format[0], curStream->m_Format.size());
			streamOut->setStreamInfo(curStream->m_StreamInfo);
			AVIStreamHeader_fixed& streamInfo = curStream->m_StreamInfo;
			uint64 iMS = (VD64(1000) * (curStream->m_Stream->m_AudioStream ? curStream->m_Stream->m_AudioStream->GetLength() : streamInfo.dwLength) * streamInfo.dwScale) / streamInfo.dwRate + 1;
			if(iMS > iHintDurationMS)
				iHintDurationMS = iMS;
		}
		it++;
	}

	pOutput->setHintDuration(iHintDurationMS);
	pOutput->init(s.c_str());

	return pOutput.release();
}

void MatroskaOutputFileSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	/*AVIOutputFile *pFile = (AVIOutputFile *)pSegment;
	if (mSegmentDigits)
		pFile->setSegmentHintBlock(bLast & m_bLastSegment, NULL, 1);*/
	pSegment->finalize();
	delete pSegment;
}

void MatroskaOutputFileSystem::SetCaching(bool bAllowOSCaching) {
	mbAllowCaching = bAllowOSCaching;
}

void MatroskaOutputFileSystem::SetBuffer(int bufferSize) {
	mBufferSize = bufferSize;
}

void MatroskaOutputFileSystem::SetFilename(const wchar_t *pszFilename) {
	mSegmentBaseName	= pszFilename;
	mSegmentDigits		= 0;
}

void MatroskaOutputFileSystem::SetFilenamePattern(const wchar_t *pszFilename, const wchar_t *pszExt, int nMinimumDigits, int nCurrentSegment, bool bLastSegment) {
	mSegmentBaseName	= pszFilename;
	mSegmentExt			= pszExt;
	mSegmentDigits		= nMinimumDigits;
	mCurrentSegment		= nCurrentSegment;
	m_bLastSegment		= bLastSegment;
}

void MatroskaOutputFileSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void MatroskaOutputFileSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool MatroskaOutputFileSystem::AcceptsVideo() {
	return m_bAcceptsVideo;
}

void MatroskaOutputFileSystem::SetAcceptsVideo(bool bAccept) {
	m_bAcceptsVideo = bAccept;
}

bool MatroskaOutputFileSystem::AcceptsAudio() {
	return true;
}

void MatroskaOutputFileSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	m_VideoStream = &stream;
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void MatroskaOutputFileSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	ModAVIDubStream *newStream = m_Streams.AddLast();
	newStream->m_Stream = &stream;
	newStream->m_StreamInfo = asi;
	newStream->m_Format.resize(cbFormat);
	memcpy(&newStream->m_Format[0], pFormat, cbFormat);
}

bool MatroskaOutputFileSystem::AcceptsText() {
	return true;
}

bool MatroskaOutputFileSystem::IsRealTime() {
	return false;
}