//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <malloc.h>
#include <vd2/system/VDString.h>
#include <vd2/system/log.h>
#include <vd2/Dita/resources.h>
#include <vd2/system/text.h>
#include <vd2/system/Error.h>
#include "ProgressDialog.h"
#include "MatroskaReadHandler.h"
#include "MatroskaComments.h"
#include "DubSource.h"
#include "ModStream.h"
#include "ogg&co/OggDS.h"
#include "ISO-639.h"
#include "ModStream.h"
#include "prefs.h"

using namespace LIBMATROSKA_NAMESPACE;
using namespace LIBEBML_NAMESPACE;

extern CRITICAL_SECTION g_diskcs;
extern bool g_disklockinited;

// Attempting to realloc memory by bigger blocks (to call realloc less often)
// and finally shrink the buffer with a realloc/_expand of the exact size actually
// consumes more memory, so realloc for each new element
#define MATROSKA_REALLOC_BLOCKGROUP_SIZE	1

#define MATROSKA_PARSE_PARTIAL 1

MatroskaReadStream::MatroskaReadStream(MatroskaReadNode *readNode) {

	VDASSERT(readNode);

	m_ReadNode = readNode;
	m_ReadNode->AddRef();
	m_ReadNode->RegisterReadStream(this);
	m_bVBRHandling = true;
	m_BlockAlign = 1;
	m_TimecodeScale = 1000000;

	SeekFirstStreamSegment();

	m_Format = NULL;
	m_FormatSize = 0;
	m_LastFilePos = m_ReadNode->m_LastFilePos;

	if(!(long)m_Stream->m_SampleRate)
		throw MyError("Matroska stream %ld doesn't have a samplerate", m_Stream->m_Serial);

	m_bIsVideo = (MAIN_TYPE(m_Stream->m_StreamType) == SOURCE_VIDEO);
	m_bIsText = m_Stream->m_bIsText;
	m_bCanTrimPacket = !m_bIsVideo;

	m_Length = m_ReadNode->m_End - m_ReadNode->m_Start;

	double duration = (double)m_ReadNode->m_NbSamples / (double)m_Stream->m_SampleRate;

	sint32 byteRate = VDRoundToLong((double)m_ReadNode->m_Bytes / duration);

	switch(m_Stream->m_StreamType) {
	case SOURCE_VIDEO:
		{
			//if(std::string(*(static_cast<EbmlString*>(m_Stream->m_CodecID))) != "V_MS/VFW/FOURCC")
			//	throw MyError("Only V_MS/VFW/FOURCC video ID is handled");
			if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "V_MS/VFW/FOURCC") {
				m_FormatSize = m_Stream->m_CodecPrivate->GetSize();
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memcpy(m_Format, m_Stream->m_CodecPrivate->GetBuffer(), m_FormatSize);
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "V_MPEG4/IS0/ASP") {
				m_FormatSize = sizeof(BITMAPINFOHEADER);
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memset(m_Format, 0, m_FormatSize);
				BITMAPINFOHEADER *bih = (BITMAPINFOHEADER *)m_Format;
				bih->biSize = m_FormatSize;
				// Set biBitCount ?
				bih->biWidth = m_Stream->m_Specific.m_Video.m_Width;
				bih->biHeight = m_Stream->m_Specific.m_Video.m_Height;
				bih->biSizeImage = (bih->biWidth * bih->biHeight * bih->biBitCount) / 8;
#pragma message(__TODO__ "Let the user choose the codec to use")
				// Use XviD as decoder
				bih->biCompression = 'DIVX';
			} else {
				//throw MyError("The video ID [%s] isn't handled", (std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID))).c_str());
				m_FormatSize = sizeof(BITMAPINFOHEADER);
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memset(m_Format, 0, m_FormatSize);
				BITMAPINFOHEADER *bih = (BITMAPINFOHEADER *)m_Format;
				bih->biSize = m_FormatSize;
				// Set biBitCount ?
				bih->biWidth = m_Stream->m_Specific.m_Video.m_Width;
				bih->biHeight = m_Stream->m_Specific.m_Video.m_Height;
				bih->biSizeImage = (bih->biWidth * bih->biHeight * bih->biBitCount) / 8;
				bih->biCompression = 0xFFFFFFFF;
			}
		}
		break;

	case SOURCE_AUDIO:
		{
			bool bUseWAVEFORMATEX = true;
			WORD wBitsPerSample = 16;
			WORD nBlockAlign = m_BlockAlign = 1;
			WORD wFormatTag = WAVE_FORMAT_PCM;
			if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_MS/ACM") {
				bUseWAVEFORMATEX = false;
				m_FormatSize = m_Stream->m_CodecPrivate->GetSize();
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memcpy(m_Format, m_Stream->m_CodecPrivate->GetBuffer(), m_FormatSize);
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_PCM/INT/LIT") {
				bUseWAVEFORMATEX = true;
				wBitsPerSample = m_Stream->m_Specific.m_Audio.m_BitDepth;
				wFormatTag = WAVE_FORMAT_PCM;
				if(wBitsPerSample)
					nBlockAlign = (wBitsPerSample * m_Stream->m_Specific.m_Audio.m_NbChannels) >> 3;
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_PCM/INT/BIG") {
				bUseWAVEFORMATEX = false;
				throw MyError("PCM Big Endian not handled");
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_PCM/FLOAT/IEEE") {
				bUseWAVEFORMATEX = true;
				wBitsPerSample = m_Stream->m_Specific.m_Audio.m_BitDepth;
				wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
				if(wBitsPerSample)
					nBlockAlign = (wBitsPerSample * m_Stream->m_Specific.m_Audio.m_NbChannels) >> 3;
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_MPEG/L1") {
				bUseWAVEFORMATEX = false;
				m_FormatSize = sizeof(MPEG1WAVEFORMAT);
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memset(m_Format, 0, m_FormatSize);
				MPEG1WAVEFORMAT *mpaformat = (MPEG1WAVEFORMAT *)m_Format;
				mpaformat->wfx.cbSize = sizeof(MPEG1WAVEFORMAT) - sizeof(WAVEFORMATEX);
				mpaformat->wfx.nBlockAlign = 1;
				mpaformat->wfx.nChannels = m_Stream->m_Specific.m_Audio.m_NbChannels;
				mpaformat->wfx.nSamplesPerSec = m_Stream->m_SampleRate;
				mpaformat->wfx.wBitsPerSample = 0;
				mpaformat->wfx.wFormatTag = WAVE_FORMAT_MPEG;
				mpaformat->wfx.nAvgBytesPerSec = byteRate;
				mpaformat->fwHeadLayer = ACM_MPEG_LAYER1;
				switch(mpaformat->wfx.nSamplesPerSec) {
					// MPEG-1
				case 48000:
				case 32000:
					mpaformat->fwHeadMode = ACM_MPEG_ID_MPEG1;
					// MPEG-2
				case 24000:
				case 16000:
					// MPEG-2.5
				case 12000:
				case 8000:
					break;
				default:
					mpaformat->fwHeadMode = ACM_MPEG_ID_MPEG1;
					break;
				}
				switch(mpaformat->wfx.nChannels) {
				case 2:
					mpaformat->fwHeadMode	= ACM_MPEG_STEREO;
					break;
				case 1:
				default:
					mpaformat->fwHeadMode	= ACM_MPEG_SINGLECHANNEL;
					break;
				}
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_MPEG/L2") {
				bUseWAVEFORMATEX = false;
				m_FormatSize = sizeof(MPEG1WAVEFORMAT);
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memset(m_Format, 0, m_FormatSize);
				MPEG1WAVEFORMAT *mpaformat = (MPEG1WAVEFORMAT *)m_Format;
				mpaformat->wfx.cbSize = sizeof(MPEG1WAVEFORMAT) - sizeof(WAVEFORMATEX);
				mpaformat->wfx.nBlockAlign = 1;
				mpaformat->wfx.nChannels = m_Stream->m_Specific.m_Audio.m_NbChannels;
				mpaformat->wfx.nSamplesPerSec = m_Stream->m_SampleRate;
				mpaformat->wfx.wBitsPerSample = 0;
				mpaformat->wfx.wFormatTag = WAVE_FORMAT_MPEG;
				mpaformat->wfx.nAvgBytesPerSec = byteRate;
				mpaformat->fwHeadLayer = ACM_MPEG_LAYER2;
				switch(mpaformat->wfx.nSamplesPerSec) {
					// MPEG-1
				case 48000:
				case 32000:
					mpaformat->fwHeadMode = ACM_MPEG_ID_MPEG1;
					// MPEG-2
				case 24000:
				case 16000:
					// MPEG-2.5
				case 12000:
				case 8000:
					break;
				default:
					mpaformat->fwHeadMode = ACM_MPEG_ID_MPEG1;
					break;
				}
				switch(mpaformat->wfx.nChannels) {
				case 2:
					mpaformat->fwHeadMode	= ACM_MPEG_STEREO;
					break;
				case 1:
				default:
					mpaformat->fwHeadMode	= ACM_MPEG_SINGLECHANNEL;
					break;
				}
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_MPEG/L3") {
				bUseWAVEFORMATEX = false;
				m_FormatSize = sizeof(MPEGLAYER3WAVEFORMAT);
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				memset(m_Format, 0, m_FormatSize);
				MPEGLAYER3WAVEFORMAT *mp3format = (MPEGLAYER3WAVEFORMAT *)m_Format;
				mp3format->wfx.cbSize = MPEGLAYER3_WFX_EXTRA_BYTES;
				mp3format->wfx.nBlockAlign = 1;
				mp3format->wfx.nChannels = m_Stream->m_Specific.m_Audio.m_NbChannels;
				mp3format->wfx.nSamplesPerSec = m_Stream->m_SampleRate;
				mp3format->wfx.wBitsPerSample = 16;
				mp3format->wfx.wFormatTag = WAVE_FORMAT_MPEGLAYER3;
				mp3format->wfx.nAvgBytesPerSec = byteRate;
				mp3format->nFramesPerBlock = 1;
				mp3format->nCodecDelay = 0;
				if(mp3format->wfx.nSamplesPerSec < 32000)	// MPEG-2 ?
					mp3format->nBlockSize = (m_ReadNode->m_Bytes * 576) / m_ReadNode->m_NbSamples;
				else	// MPEG-1
					mp3format->nBlockSize = (m_ReadNode->m_Bytes * 1152) / m_ReadNode->m_NbSamples;
				if(mp3format->nBlockSize <= 0)
					mp3format->nBlockSize = 1;
				mp3format->wID = MPEGLAYER3_ID_MPEG;
				mp3format->fdwFlags = MPEGLAYER3_FLAG_PADDING_ISO;
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_AC3") {
				bUseWAVEFORMATEX = true;
				wBitsPerSample = 16;
				wFormatTag = 0x2000;
			} else if(std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID)) == "A_DTS") {
				bUseWAVEFORMATEX = true;
				wBitsPerSample = 16;
				wFormatTag = 0x2001;
			} else {
				VDDEBUG("Format [%s] not handled, DirectStream copied", (std::string(*static_cast<EbmlString*>(m_Stream->m_CodecID))).c_str());
				bUseWAVEFORMATEX = true;
				wBitsPerSample = 16;
				wFormatTag = WAVE_FORMAT_UNKNOWN;
			}
			if(bUseWAVEFORMATEX) {
				m_FormatSize = sizeof(WAVEFORMATEX);
				m_Format = malloc(m_FormatSize);
				if(!m_Format)
					throw MyMemoryError();
				WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;
				wf->cbSize = 0;
				wf->nBlockAlign = m_BlockAlign = nBlockAlign;
				wf->nChannels = m_Stream->m_Specific.m_Audio.m_NbChannels;
				wf->nSamplesPerSec = m_Stream->m_SampleRate;
				wf->wBitsPerSample = wBitsPerSample;
				wf->wFormatTag = wFormatTag;
				wf->nAvgBytesPerSec = byteRate;
			}
		}
		break;

	case SOURCE_AUDIO_VORBIS:
		{
			m_FormatSize = sizeof(WAVEFORMATEXTENSIBLE);
			m_Format = malloc(m_FormatSize);
			if(!m_Format)
				throw MyMemoryError();

			memset(m_Format, 0, m_FormatSize);

			WAVEFORMATEXTENSIBLE *wfe = (WAVEFORMATEXTENSIBLE *)m_Format;
			wfe->Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
			wfe->Format.cbSize = sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX);

			wfe->Format.nAvgBytesPerSec = byteRate;
			wfe->Format.nBlockAlign = m_BlockAlign = 1;
			wfe->Format.nChannels = m_Stream->m_Specific.m_Audio.m_NbChannels;
			wfe->Format.nSamplesPerSec = m_Stream->m_SampleRate;
			wfe->Samples.wReserved = 0;
			switch(wfe->Format.nChannels) {
				case 1:
					wfe->dwChannelMask = SPEAKER_FRONT_CENTER;
					break;
				case 2:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT;
					break;
				case 4:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT
						| SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT;
					break;
				case 5:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT
						| SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT
						| SPEAKER_LOW_FREQUENCY;
					break;
				case 6:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT
						| SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT
						| SPEAKER_LOW_FREQUENCY | SPEAKER_FRONT_CENTER;
					break;
				default:
					VDNEVERHERE;
					break;
			}
			wfe->SubFormat = MEDIASUBTYPE_Vorbis;
		}
		break;

	case SOURCE_TEXT:
	case SOURCE_TEXT_SSA:
		{
			if(memcmp(std::string(*(static_cast<EbmlString*>(m_Stream->m_CodecID))).c_str(), "S_TEXT/", 7))
				throw MyError("Only handle [S_TEXT/] subs for the moment");
			m_bCanTrimPacket = true;
			// Let's use this text stream as an audio one :p
			m_FormatSize = sizeof(WAVEFORMATEX);
			m_Format = malloc(m_FormatSize);
			if(!m_Format)
				throw MyMemoryError();

			memset(m_Format, 0, m_FormatSize);

			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;
			wf->cbSize = 0;
			// Since we may need to convert from ASCII to Unicode/UTF8, compensate here
			if(m_Stream->m_MaxFrameSize > byteRate)
				wf->nAvgBytesPerSec = 40 * m_Stream->m_MaxFrameSize;
			else if(byteRate > 0)
				wf->nAvgBytesPerSec = 40 * byteRate;
			else
				wf->nAvgBytesPerSec = 40;
			wf->nBlockAlign = 1;
			wf->nChannels = 1;
			wf->nSamplesPerSec = 1000;
			wf->wBitsPerSample = 0;
			wf->wFormatTag = 0x0000;
		}
		break;

	default:
		VDNEVERHERE;
		break;
	}
}

MatroskaReadStream::~MatroskaReadStream() {
	m_ReadNode->UnregisterReadStream(this);
	m_ReadNode->Release();
	if(m_Format)
		delete m_Format;
}

void MatroskaReadStream::Reinit(void) {
	SeekFirstStreamSegment();

	m_LastFilePos = m_ReadNode->m_LastFilePos;
	m_Length = m_ReadNode->m_End - m_ReadNode->m_Start;

	double duration = (double)m_ReadNode->m_NbSamples / (double)m_Stream->m_SampleRate;

	sint32 byteRate = VDRoundToLong((double)m_ReadNode->m_Bytes / duration);

	switch(m_Stream->m_StreamType) {
	case SOURCE_AUDIO:
	case SOURCE_AUDIO_VORBIS:
		{
			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;
			wf->nAvgBytesPerSec = byteRate;
			//wf->nBlockAlign = m_BlockAlign = 1;
			if(wf->wFormatTag == WAVE_FORMAT_MPEGLAYER3) {
				MPEGLAYER3WAVEFORMAT *mp3format = (MPEGLAYER3WAVEFORMAT *)m_Format;
				mp3format->nBlockSize = (m_ReadNode->m_Bytes * 1152) / m_ReadNode->m_NbSamples;
				if(mp3format->nBlockSize <= 0)
					mp3format->nBlockSize = 1;
			}
		}
		break;

	case SOURCE_TEXT:
	case SOURCE_TEXT_SSA:
		{
			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;

			if(m_Stream->m_MaxFrameSize > byteRate)
				wf->nAvgBytesPerSec = 40 * m_Stream->m_MaxFrameSize;
			else if(byteRate > 0)
				wf->nAvgBytesPerSec = 40 * byteRate;
			else
				wf->nAvgBytesPerSec = 40;
		}
		break;
	}
}

std::string MatroskaReadStream::GetCodecID(void) {
	SeekFirstStreamSegment();

	std::string codecID;
	codecID = *(static_cast<EbmlString*>(m_Stream->m_CodecID));
	return codecID;
}

std::vector<uint8> MatroskaReadStream::GetCodecPrivate(void) {
	SeekFirstStreamSegment();

	std::vector<uint8> codecPrivate;
	if(m_Stream->m_CodecPrivate) {
		int size = m_Stream->m_CodecPrivate->GetSize();
		if(size > 0) {
			codecPrivate.resize(size);
			memcpy(&codecPrivate[0], m_Stream->m_CodecPrivate->GetBuffer(), size);
		}
	}
	return codecPrivate;
}

void *MatroskaReadStream::GetKaxTrackEntry(void) {
	SeekFirstStreamSegment();

	return m_Stream->m_TrackEntry;
}

VDFraction MatroskaReadStream::getDAR(void) {
	VDFraction dar(0, 1);

	SeekFirstStreamSegment();

	if(!m_bIsVideo || !m_Format)
		return dar;

	BITMAPINFOHEADER *pBIH = (BITMAPINFOHEADER *)m_Format;

	uint32 iDisplayWidth = pBIH->biWidth;
	uint32 iDisplayHeight = pBIH->biHeight;

	if(m_Stream->m_TrackEntry) {
		EbmlElement* aElement = m_Stream->m_TrackEntry->FindElt(KaxTrackVideo::ClassInfos);
		if(aElement) {
			KaxTrackVideo& MyVideo = *static_cast<KaxTrackVideo *>(aElement);
			EbmlElement* aDisplayWidth = MyVideo.FindElt(KaxVideoDisplayWidth::ClassInfos);
			EbmlElement* aDisplayHeight = MyVideo.FindElt(KaxVideoDisplayHeight::ClassInfos);
			if(aDisplayWidth)
				iDisplayWidth = LIBEBML_uint32(*static_cast<EbmlUInteger *>(aDisplayWidth));
			if(aDisplayHeight)
				iDisplayHeight = LIBEBML_uint32(*static_cast<EbmlUInteger *>(aDisplayHeight));
		}
	}

	dar = VDFraction(iDisplayWidth, iDisplayHeight);

	return dar;
}

HRESULT MatroskaReadStream::BeginStreaming(long lStart, long lEnd, long lRate) {
	return 0;
}

HRESULT MatroskaReadStream::EndStreaming() {
	return 0;
}

HRESULT MatroskaReadStream::Info(AVISTREAMINFO *pasi, long lSize) {
	AVISTREAMINFO asi;

	memset(&asi, 0, sizeof asi);

	double duration = (double)m_ReadNode->m_NbSamples / (double)m_Stream->m_SampleRate;

	sint32 byteRate = VDRoundToLong((double)m_ReadNode->m_Bytes / duration);

	asi.fccType				= (m_bIsVideo ? streamtypeVIDEO
		: MAIN_TYPE(m_Stream->m_StreamType)==SOURCE_TEXT ? streamtypeTEXT
		: streamtypeAUDIO);
	if(m_bIsVideo) {
		asi.fccHandler = ((BITMAPINFOHEADER *)m_Format)->biCompression;
		// The XviD codec is selected as decompressor when fccHandler is set to DX50
		// but the XviD codec can't decode all versions of DivX (resulting in error -100
		// in VirtualDub(Mod) and no video in the preview windows).
		// So force the 'correct' decoder handle.
		switch(asi.fccHandler) {
		// DivX 4.x/5.x
		case 'xvid':
		case 'XVID':
		case '05xd':
		case '05XD':
			// Use the DivX decoder
			asi.fccHandler = 'xvid';
			break;
		// XviD
		case 'divx':
		case 'DIVX':
			// Use the XviD decoder
			asi.fccHandler = 'divx';
			break;
		}
	}
	asi.dwFlags				= 0;
	asi.wPriority			= 0;
	asi.wLanguage			= 0;
	asi.dwSampleSize		= 0;
	if(m_bIsVideo) {
		asi.dwScale				= m_Stream->m_SampleRate.getLo();
		asi.dwRate				= m_Stream->m_SampleRate.getHi();
	} else if(MAIN_TYPE(m_Stream->m_StreamType) == SOURCE_TEXT) {
		asi.dwScale				= 1;
		asi.dwRate				= 1000;
	} else {
		if(m_bVBRHandling) {
			asi.dwSampleSize		= 0;
			asi.dwScale				= 1;
			asi.dwRate				= (long)m_Stream->m_SampleRate;
			asi.dwLength			= m_Length;
		} else {
			asi.dwSampleSize		= m_BlockAlign;
			asi.dwScale				= m_BlockAlign;
			asi.dwRate				= byteRate;
			if(m_BlockAlign)
				asi.dwLength			= m_Length / m_BlockAlign;
			else
				asi.dwLength			= m_Length;
		}
	}
	//asi.dwStart				= 0;
	asi.dwStart = m_ReadNode->m_Start;
	asi.dwInitialFrames		= 0;
	asi.dwSuggestedBufferSize = m_ReadNode->m_MaxFrameSize;
	asi.dwQuality			= 0xFFFFFFFF;
	if(m_bIsVideo) {
		asi.rcFrame.top			= 0;
		asi.rcFrame.left		= 0;
		asi.rcFrame.right		= m_Stream->m_Specific.m_Video.m_Width;
		asi.rcFrame.bottom		= m_Stream->m_Specific.m_Video.m_Height;
	}

	if (lSize < sizeof asi)
		memcpy(pasi, &asi, lSize);
	else {
		memcpy(pasi, &asi, sizeof asi);
		memset((char *)pasi + sizeof asi, 0, lSize - sizeof asi);
	}

	return 0;
}

void MatroskaReadStream::SetVBRHandling(bool bVBRHandling) {
	if(m_Stream->m_StreamType == SOURCE_AUDIO_VORBIS)
		bVBRHandling = true;
	m_bVBRHandling = bVBRHandling;
	m_ReadNode->UpdateState(bVBRHandling);
	Reinit();
}

inline void MatroskaReadStream::SynchSegmentInformation(void) {

	VDASSERT(m_CurrentStreamSegment);
	VDASSERT(m_Stream != m_CurrentStreamSegment->m_MatroskaStream);

	m_Parent = m_CurrentStreamSegment->m_ReadHandler;
	m_EbmlStream = m_CurrentStreamSegment->m_ReadHandler->m_EbmlStream;
	m_Stream = m_CurrentStreamSegment->m_MatroskaStream;
	if(m_Stream->m_KeyIndex.size()) {
		m_KeyFrames = &m_Stream->m_KeyIndex.front();
		m_KeyFramesEnd = &m_Stream->m_KeyIndex.back() + 1;
		if(m_Stream->m_BIndex.size()) {
			m_BFrames = &m_Stream->m_BIndex.front();
			m_BFramesEnd = &m_Stream->m_BIndex.back() + 1;
		} else {
			m_BFrames = NULL;
			m_BFramesEnd = NULL;
		}
	} else {
		m_KeyFrames = NULL;
		m_KeyFramesEnd = NULL;
		m_BFrames = NULL;
		m_BFramesEnd = NULL;
	}
	m_ClusterIterator.begin(m_Stream->m_ClusterList);
	m_BlockByte = 0;
	m_BlockIndex = 0;
	m_TimecodeScale = m_Stream->m_TimecodeScale;
	m_TimeFactor = m_Stream->m_SampleRate / VDFraction(1000000000, m_TimecodeScale);
	m_TimeFactor.reduce();
}

bool MatroskaReadStream::SeekPreviousSegment(void) {
	MatroskaStreamSegment *previousSegment = m_CurrentStreamSegment->GetPrevious();
	if(previousSegment) {
		m_CurrentStreamSegment = previousSegment;
		SynchSegmentInformation();
		return true;
	} else
		return false;
}

bool MatroskaReadStream::SeekNextSegment(void) {
	MatroskaStreamSegment *nextSegment = m_CurrentStreamSegment->GetNext();
	if(nextSegment) {
		m_CurrentStreamSegment = nextSegment;
		SynchSegmentInformation();
		return true;
	} else
		return false;
}

void MatroskaReadStream::SeekFirstStreamSegment(void) {
	m_CurrentStreamSegment = m_ReadNode->m_Segments.GetFirst();
	SynchSegmentInformation();
}

void MatroskaReadStream::SeekLastStreamSegment(void) {
	m_CurrentStreamSegment = m_ReadNode->m_Segments.GetLast();
	SynchSegmentInformation();
}

long MatroskaReadStream::FindStreamSegment(long lFrame) {
	if( (lFrame<0) || (lFrame>=m_Length) )
		return -1;
	if(lFrame < m_CurrentStreamSegment->m_Start) {
		do {
			m_CurrentStreamSegment = m_CurrentStreamSegment->GetPrevious();

			VDDEBUG("Accessing previous Matroska segment\n");
			VDASSERT(m_CurrentStreamSegment);

		} while(lFrame < m_CurrentStreamSegment->m_Start);
		SynchSegmentInformation();
	} else if(lFrame >= m_CurrentStreamSegment->m_End) {
		do {
			m_CurrentStreamSegment = m_CurrentStreamSegment->GetNext();

			VDDEBUG("Accessing next Matroska segment\n");
			VDASSERT(m_CurrentStreamSegment);

		} while(lFrame >= m_CurrentStreamSegment->m_End);
		SynchSegmentInformation();
	}

	VDASSERT(m_CurrentStreamSegment && (lFrame>=m_CurrentStreamSegment->m_Start) && (lFrame<m_CurrentStreamSegment->m_End) );

	return _GetIndexFromDisplayIndex(lFrame - m_CurrentStreamSegment->m_Start);
}

long MatroskaReadStream::_GetIndexFromDisplayIndex(long lFrame) {
	if(!m_BFrames)
		return lFrame;

	if(_IsBFrame(lFrame)) {
		if(_IsBFrame(lFrame+1))
			return lFrame+1;
		while(lFrame--) {
			if(!_IsBFrame(lFrame))
				return lFrame;
		}
		return lFrame;
	} else
		return (_IsBFrame(lFrame+1) ? lFrame+1 : lFrame);
}

void MatroskaReadStream::GetFramesToDecode(long lDesiredFrame, std::vector<ModFrameIndex>& listFrames, long lLastReadFrame) {
	if( (lDesiredFrame<0) || (lDesiredFrame>=m_Length) )
		return;

	long lCurrentFrame = NearestKeyFrame(lDesiredFrame);
	if(lCurrentFrame == -1)
		return;

	// Seems like the codec need to be given the i-frames twice :/
	/*if( (lLastReadFrame > lCurrentFrame) && (lLastReadFrame < lDesiredFrame) ) {
		lCurrentFrame = lLastReadFrame;
		while(IsBFrame(++lCurrentFrame));
	} else */if(lCurrentFrame == lDesiredFrame)
		listFrames.push_back(ModFrameIndex(lCurrentFrame, false));
	for(;;) {
		/*if(lLastReadFrame == lCurrentFrame)
			listFrames.clear();
		else*/
			listFrames.push_back(ModFrameIndex(lCurrentFrame, false));
		if(lCurrentFrame == lDesiredFrame) {
			// When seeking backward you need to give the last P-frame twices
			// so that it is decoded by XviD / DivX 5
			if((lDesiredFrame<lLastReadFrame) && !IsKeyFrame(lCurrentFrame))
				listFrames.push_back(ModFrameIndex(lCurrentFrame, false));
			break;
		} else if(lCurrentFrame > lDesiredFrame) {
			listFrames.push_back(ModFrameIndex(lDesiredFrame, true));
			break;
		}
		while(IsBFrame(++lCurrentFrame));
	}
}

bool MatroskaReadStream::_IsBFrame(long lFrame) {
	if(!m_BFrames)
		return false;
	if( (lFrame<0) || (lFrame>m_CurrentStreamSegment->m_End-m_CurrentStreamSegment->m_Start) )
		return false;
	// & 7 == % 8
	return !!(m_BFrames[lFrame>>3] & (0x80>>(lFrame&7)));
}

long MatroskaReadStream::GetIndexFromDisplayIndex(long lFrame) {
	if((lFrame=FindStreamSegment(lFrame))<0)
		return -1;
	return (m_CurrentStreamSegment->m_Start + lFrame);
}

bool MatroskaReadStream::IsKeyFrame(long lFrame) {
	if((lFrame=FindStreamSegment(lFrame))<0)
		return false;
	if(!m_KeyFrames)
		return true;
	return !!(m_KeyFrames[lFrame>>3] & (0x80>>(lFrame&7)));
}

bool MatroskaReadStream::IsBFrame(long lFrame) {
	if((lFrame=FindStreamSegment(lFrame))<0)
		return false;
	if(!m_BFrames)
		return false;
	return !!(m_BFrames[lFrame>>3] & (0x80>>(lFrame&7)));
}

sint64 MatroskaReadStream::sampleOffset(sint64 lStart) {
	if(lStart < 0)
		return 0;
	if(lStart >= m_Length)
		return m_LastFilePos;
	if(!m_bVBRHandling)
		return lStart * m_BlockAlign;

	if((lStart=FindStreamSegment(lStart))<0)
		return 0;

	while(m_ClusterIterator.previous() && ((*m_ClusterIterator)->m_Blocks[0].m_Granulepos>lStart) )
		m_ClusterIterator--;
	while(m_ClusterIterator.next() && (m_ClusterIterator.next()->m_Blocks[0].m_Granulepos<=lStart) )
		m_ClusterIterator++;

	m_Cluster = *m_ClusterIterator;
	sint64 pos = m_Cluster->m_ClusterPosition;
	uint16 blockNum = 0;
	while( (blockNum+1<m_Cluster->m_NbBlocks)
		&& (m_Cluster->m_Blocks[blockNum+1].m_Granulepos <= lStart) )
		blockNum++;

	return MKV_BLOCKPOS(m_Cluster->m_Blocks[blockNum].m_BlockPosition);
}

HRESULT MatroskaReadStream::ReadSubtitle(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples) {
	if((lStart=FindStreamSegment(lStart))<0) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return AVIERR_OK;
	}

	VDStringW subtitle;
	sint32 durationMS = 0;

	if(!m_Stream->m_SubtitleProvider.GetSubtitle_Linear(subtitle, lStart, durationMS)) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return AVIERR_OK;
	}

	VDStringA subUTF8 = VDTextWToU8(subtitle);

	if(plBytes)
		*plBytes = subUTF8.length() + 1;
	if(plSamples)
		*plSamples = durationMS;

	if(lpBuffer && (cbBuffer < subUTF8.length() + 1) )
		return AVIERR_BUFFERTOOSMALL;

	if(!lpBuffer)
		return AVIERR_OK;

	memcpy(lpBuffer, subUTF8.c_str(), subUTF8.length() + 1);

	return AVIERR_OK;
}

HRESULT MatroskaReadStream::Read(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples) {
	if(!m_bVBRHandling && !m_bIsVideo)
		return ReadCBR(lStart, lSamples, lpBuffer, cbBuffer, plBytes, plSamples);
	if(m_bIsText)
		return ReadSubtitle(lStart, lSamples, lpBuffer, cbBuffer, plBytes, plSamples);

	long lStartWanted = lStart;
	if((lStart=FindStreamSegment(lStart))<0) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return AVIERR_OK;
	}

	while(m_ClusterIterator.previous() && ((*m_ClusterIterator)->m_Blocks[0].m_Granulepos>lStart) )
		m_ClusterIterator--;
	while(m_ClusterIterator.next() && (m_ClusterIterator.next()->m_Blocks[0].m_Granulepos<=lStart) )
		m_ClusterIterator++;

	m_Cluster = *m_ClusterIterator;

	int samplesToRead = lSamples;
	if(samplesToRead == AVISTREAMREAD_CONVENIENT)
		samplesToRead = 1;

	long nbBytes = 0;
	long samplesToRead2 = samplesToRead;
	int blocksToSkip = 0;
	while((blocksToSkip+1 < m_Cluster->m_NbBlocks) && (m_Cluster->m_Blocks[blocksToSkip+1].m_Granulepos<=lStart))
		blocksToSkip++;
	VDFraction estimatedSamplesPerFrame(1);
	if(blocksToSkip+1 < m_Cluster->m_NbBlocks)
		estimatedSamplesPerFrame = VDFraction(m_Cluster->m_Blocks[blocksToSkip+1].m_Granulepos - m_Cluster->m_Blocks[blocksToSkip].m_Granulepos, m_Cluster->m_Blocks[blocksToSkip].m_NbFrames);
	else if(m_ClusterIterator.next())
		estimatedSamplesPerFrame = VDFraction(m_ClusterIterator.next()->m_Blocks[0].m_Granulepos - m_Cluster->m_Blocks[blocksToSkip].m_Granulepos, m_Cluster->m_Blocks[blocksToSkip].m_NbFrames);
	else
		estimatedSamplesPerFrame = VDFraction(m_Stream->m_EstimatedSamplesInLastBlock, m_Cluster->m_Blocks[blocksToSkip].m_NbFrames);
	int framesToSkip = 0;
	while((framesToSkip < m_Cluster->m_Blocks[blocksToSkip].m_NbFrames) && (m_Cluster->m_Blocks[blocksToSkip].m_Granulepos + estimatedSamplesPerFrame.scale64r(framesToSkip+1)<=lStart))
		framesToSkip++;

	int samplesShorten = 0;
	if(m_bCanTrimPacket)
		samplesShorten = lStart -
			(m_Cluster->m_Blocks[blocksToSkip].m_Granulepos + estimatedSamplesPerFrame.scale64r(framesToSkip));

	int idx1 = blocksToSkip;
	int idx2 = framesToSkip;
	m_TmpIterator = m_ClusterIterator;
	while(samplesToRead2>0) {
		if(!m_TmpIterator)
			break;
		m_Cluster = *m_TmpIterator;
		while((samplesToRead2>0) && (idx1<m_Cluster->m_NbBlocks)) {
			int nbFrames = 0;
			VDFraction estimatedSamplesPerFrame2((unsigned long)m_Stream->m_EstimatedSamplesInLastBlock);
			if(idx1+1 < m_Cluster->m_NbBlocks)
				estimatedSamplesPerFrame2 = VDFraction(m_Cluster->m_Blocks[idx1+1].m_Granulepos - m_Cluster->m_Blocks[idx1].m_Granulepos, m_Cluster->m_Blocks[idx1].m_NbFrames);
			else if(m_TmpIterator.next())
				estimatedSamplesPerFrame2 = VDFraction(m_TmpIterator.next()->m_Blocks[0].m_Granulepos - m_Cluster->m_Blocks[idx1].m_Granulepos, m_Cluster->m_Blocks[idx1].m_NbFrames);
			else
				estimatedSamplesPerFrame2 = VDFraction(m_Stream->m_EstimatedSamplesInLastBlock, m_Cluster->m_Blocks[idx1].m_NbFrames);
			while((samplesToRead2 - estimatedSamplesPerFrame2.scale64r(nbFrames)>0) && (idx2<m_Cluster->m_Blocks[idx1].m_NbFrames)) {
				nbBytes += m_Cluster->m_Blocks[idx1].m_NbBytes[idx2];
				nbFrames++;
				idx2++;
			}
			samplesToRead2 -= estimatedSamplesPerFrame2.scale64r(nbFrames);
			idx1++;
			idx2 = 0;
		}
		m_TmpIterator++;
		idx1 = 0;
	}
	m_Cluster = *m_ClusterIterator;

	if(m_bIsVideo && (m_Cluster->m_Blocks[blocksToSkip].m_Granulepos != lStart)) {
		// Either someone used lacing on the video stream, or this is a dropped frame
		sint64 start = sint64(m_Cluster->m_Blocks[blocksToSkip].m_Granulepos);
		if( (lStart>start) && (lStart-start<m_Cluster->m_Blocks[blocksToSkip].m_NbFrames) ) {

			VDASSERT(framesToSkip == lStart - start);

			framesToSkip = lStart - start;
		} else {
			if(plBytes)
				*plBytes = 0;
			if(plSamples)
				*plSamples = 1;
			return AVIERR_OK;
		}
	}

	if(lpBuffer && (cbBuffer < nbBytes) ) {
		return AVIERR_BUFFERTOOSMALL;
	}

	if(!lpBuffer) {
		if(plBytes)
			*plBytes = nbBytes;
		if(plSamples) {
			if(m_bIsVideo && (lSamples==AVISTREAMREAD_CONVENIENT) )
				*plSamples = 1;
			else
				*plSamples = (samplesToRead-samplesToRead2) - samplesShorten;
		}

		return AVIERR_OK;
	}

	VDDEBUG("Stream %ld : reading Frame %d (%u bytes) of Block %d (%d Frames, granulepos %I64u) of Cluster at byte %I64u (%d Blocks)\n"
		, m_Stream->m_Serial
		, framesToSkip, m_Cluster->m_Blocks[blocksToSkip].m_NbBytes[framesToSkip], blocksToSkip
		, (int)m_Cluster->m_Blocks[blocksToSkip].m_NbFrames, m_Cluster->m_Blocks[blocksToSkip].m_Granulepos
		, MKV_CLUSTERPOS(m_Cluster->m_ClusterPosition), (int)m_Cluster->m_NbBlocks);

	// Set position in file
	m_Parent->SeekFile(MKV_CLUSTERPOS(m_Cluster->m_ClusterPosition));
	long bytesRead = 0;
	long samplesRead = 0;
	char *lpBufferPos = (char *)lpBuffer;
	UpperElementLevel = 0;
	bool bFinished = false;
	bool bError = false;
	uint64 granuleposToFind = m_Cluster->m_Blocks[blocksToSkip].m_Granulepos + m_Stream->m_StartGranulepos;

	if(m_bIsVideo && m_BFrames)
		granuleposToFind = lStartWanted - m_CurrentStreamSegment->m_Start;

	try {
		ElementLevel1 = m_EbmlStream->FindNextElement(KaxSegment_Context, UpperElementLevel, uint64(0xFFFFFFFFFFFFFFFF), false);
		if (ElementLevel1 && (EbmlId(*ElementLevel1)==KaxCluster::ClassInfos.GlobalId)) {
			// Found a cluster, get level 2 elements
			KaxCluster * SegmentCluster = static_cast<KaxCluster *>(ElementLevel1);
			ElementLevel2 = m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), false);
			while(ElementLevel2) {
				if(UpperElementLevel > 0)	break;
				if(UpperElementLevel < 0)	UpperElementLevel = 0;
				if (EbmlId(*ElementLevel2) == KaxClusterTimecode::ClassInfos.GlobalId) {
					KaxClusterTimecode & ClusterTime = *static_cast<KaxClusterTimecode*>(ElementLevel2);
					ClusterTime.ReadData(m_EbmlStream->I_O());
					SegmentCluster->InitTimecode(ClusterTime, m_TimecodeScale);

					m_Parent->SeekFile(MKV_BLOCKPOS(m_Cluster->m_Blocks[blocksToSkip].m_BlockPosition));
					ElementLevel3 = m_EbmlStream->FindNextElement(KaxBlockGroup::ClassInfos.Context, UpperElementLevel, uint64(0xFFFFFFFFFFFFFFFF), false);
					if(ElementLevel3 && (EbmlId(*ElementLevel3) == KaxBlock::ClassInfos.GlobalId)) {
						KaxBlock & DataBlock = *static_cast<KaxBlock*>(ElementLevel3);
						DataBlock.ReadData(m_EbmlStream->I_O());
						DataBlock.SetParent(*SegmentCluster);
						if(granuleposToFind != m_TimeFactor.scale64r(DataBlock.GlobalTimecode() / m_TimecodeScale)) {
							bError = true;
							uint64 actualTimecode = DataBlock.GlobalTimecode();
							uint64 actualGranulepos = m_TimeFactor.scale64r(actualTimecode / m_TimecodeScale);
							VDLogAppMessage(kVDLogError, kVDST_MatroskaReadHandler, kVDM_Seek_BadGranulepos, 3, &granuleposToFind, &actualGranulepos, &actualTimecode);
						} else {
							long dataSize = DataBlock.GetBuffer(framesToSkip).Size();
							memcpy(lpBufferPos, DataBlock.GetBuffer(framesToSkip).Buffer(), dataSize);
							lpBufferPos += dataSize;
							bytesRead += dataSize;
							samplesRead = estimatedSamplesPerFrame.scale64r(framesToSkip+1)-estimatedSamplesPerFrame.scale64r(framesToSkip);
							bFinished = true;
						}
						delete ElementLevel3;
					} else {
						bError = true;
						VDLogAppMessage(kVDLogError, kVDST_MatroskaReadHandler, kVDM_Seek_BadBlock);
					}
				}
				if(bFinished || bError) {
					delete ElementLevel2;
					break;
				} else if (UpperElementLevel > 0) {
					UpperElementLevel--;
					delete ElementLevel2;
					ElementLevel2 = ElementLevel3;
					if(UpperElementLevel > 0)
						break;
				} else {
					ElementLevel2->SkipData(*m_EbmlStream, ElementLevel2->Generic().Context);
					delete ElementLevel2;
					ElementLevel2 = m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), false);
				}
			}
			delete ElementLevel1;
		}
		if(!bFinished && !bError) {
			bError = true;
			VDLogAppMessage(kVDLogError, kVDST_MatroskaReadHandler, kVDM_Seek_BadCluster);
		}
	} catch(exception & Ex) {
		bError = true;
		const char *error = Ex.what();
		VDLogAppMessage(kVDLogError, kVDST_MatroskaReadHandler, kVDM_LibError, 1, &error);
	}

	if(bError) {
		if(plBytes)		*plBytes = 0;
		if(plSamples)	*plSamples = 0;
		return AVIERR_FILEREAD;
	}

	if(plBytes)
		*plBytes = bytesRead;
	if(plSamples)
		*plSamples = samplesRead - samplesShorten;

	VDASSERT(samplesRead - samplesShorten > 0);

	return AVIERR_OK;
}

HRESULT MatroskaReadStream::ReadCBR(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples) {
	if((lStart=FindStreamSegment(lStart * m_BlockAlign))<0) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return AVIERR_OK;
	}

	m_Cluster = *m_ClusterIterator;
	uint32 count;
	sint64 nextBlockByte;
	if(m_BlockByte > lStart) {
		do {
			if(m_BlockIndex)
				m_BlockIndex--;
			else if(m_ClusterIterator.previous()) {
				m_ClusterIterator--;
				m_Cluster = *m_ClusterIterator;
				m_BlockIndex = m_Cluster->m_NbBlocks-1;
			} else
				break;
			if(count = m_Cluster->m_Blocks[m_BlockIndex].m_NbFrames) {
				while(count--)
					m_BlockByte -= m_Cluster->m_Blocks[m_BlockIndex].m_NbBytes[count];
			}
		} while(m_BlockByte > lStart);
	} else {
		nextBlockByte = m_BlockByte;
		do {
			if(count = m_Cluster->m_Blocks[m_BlockIndex].m_NbFrames) {
				while(count--)
					nextBlockByte += m_Cluster->m_Blocks[m_BlockIndex].m_NbBytes[count];
			}
			if( ((m_BlockIndex+1>=m_Cluster->m_NbBlocks) && !m_ClusterIterator.next()) || (nextBlockByte>lStart))
				break;
			if(m_BlockIndex+1<m_Cluster->m_NbBlocks)
				m_BlockIndex++;
			else {
				m_ClusterIterator++;
				m_Cluster = *m_ClusterIterator;
				m_BlockIndex = 0;
			}
			m_BlockByte = nextBlockByte;
		} while(true);
	}

	int framesToSkip = 0;
	sint64 bytePos = m_BlockByte;
	while((framesToSkip < m_Cluster->m_Blocks[m_BlockIndex].m_NbFrames) && (bytePos + m_Cluster->m_Blocks[m_BlockIndex].m_NbBytes[framesToSkip]<=lStart)) {
		bytePos += m_Cluster->m_Blocks[m_BlockIndex].m_NbBytes[framesToSkip];
		framesToSkip++;
	}

	long bytesShorten = lStart - bytePos;

	VDASSERT((bytesShorten >= 0) && !(bytesShorten % m_BlockAlign));

	int bytesToRead = lSamples * m_BlockAlign;

	m_TmpIterator = m_ClusterIterator;
	m_Cluster = *m_ClusterIterator;
	sint32 nbBytes = 0;
	int bytesToRead2 = bytesToRead;
	m_Cluster = *m_TmpIterator;
	int idx1 = m_BlockIndex;
	int idx2 = framesToSkip;
	if(lSamples == AVISTREAMREAD_CONVENIENT) {
		bytesToRead2 = 0;
		while(true) {
			if(idx2 < m_Cluster->m_Blocks[idx1].m_NbFrames) {
				nbBytes += m_Cluster->m_Blocks[idx1].m_NbBytes[idx2];
				bytesToRead = m_Cluster->m_Blocks[idx1].m_NbBytes[idx2];
				idx2++;
				break;
			} else {
				idx2 = 0;
				if(idx1+1 < m_Cluster->m_NbBlocks)
					idx1++;
				else if(m_TmpIterator.next()) {
					m_TmpIterator++;
					m_Cluster = *m_TmpIterator;
					idx1 = 0;
				} else
					break;
			}
		}
	} else {
		while(bytesToRead2>0) {
			if(idx2 < m_Cluster->m_Blocks[idx1].m_NbFrames) {
				long iTmpBytes = m_Cluster->m_Blocks[idx1].m_NbBytes[idx2];
				if(iTmpBytes > bytesToRead2)
					iTmpBytes = bytesToRead2;
				nbBytes += iTmpBytes;
				bytesToRead2 -= iTmpBytes;
				idx2++;
			} else {
				idx2 = 0;
				if(idx1+1 < m_Cluster->m_NbBlocks)
					idx1++;
				else if(m_TmpIterator.next()) {
					m_TmpIterator++;
					m_Cluster = *m_TmpIterator;
					idx1 = 0;
				} else
					break;
			}
		}
	}

	if(!lpBuffer || (cbBuffer < nbBytes) ) {
		if(plBytes)
			*plBytes = nbBytes;
		if(plSamples)
			*plSamples = VDRoundToLong(double((bytesToRead-bytesToRead2) - bytesShorten) / m_BlockAlign);
		if(!lpBuffer)
			return AVIERR_OK;
		return AVIERR_BUFFERTOOSMALL;
	}

	m_Cluster = *m_ClusterIterator;

	long bytesLeft = bytesToRead;

	VDDEBUG("Stream %ld : Reading %ld bytes in Frame %d (%lu bytes) starting at byte %I64u+%ld in Block %d (%d Frames, granulepos %I64d) of Cluster at byte %I64u (%d Blocks)\n"
		, m_Stream->m_Serial , bytesLeft, framesToSkip
		, m_Cluster->m_Blocks[m_BlockIndex].m_NbBytes[framesToSkip]
		, m_BlockByte, bytesShorten, m_BlockIndex
		, (int)m_Cluster->m_Blocks[m_BlockIndex].m_NbFrames, m_Cluster->m_Blocks[m_BlockIndex].m_Granulepos
		, MKV_CLUSTERPOS(m_Cluster->m_ClusterPosition), (int)m_Cluster->m_NbBlocks);

	// Set position in file
	m_Parent->SeekFile(MKV_CLUSTERPOS(m_Cluster->m_ClusterPosition));
	long bytesRead = 0;
	UpperElementLevel = 0;
	bool bFinished = false;
	bool bSeekOnFirstBlock = true;
	long toRead;
	sint32 frameCount;
	int blocksToSkip = m_BlockIndex;
	char *lpBufferPos = (char *)lpBuffer;
	try {
		ElementLevel1 = m_EbmlStream->FindNextElement(KaxSegment_Context, UpperElementLevel, uint64(0xFFFFFFFFFFFFFFFF), false);
		while(ElementLevel1) {
			if(UpperElementLevel > 0)	break;
			if(UpperElementLevel < 0)	UpperElementLevel = 0;
			if (EbmlId(*ElementLevel1) == KaxCluster::ClassInfos.GlobalId) {
				// Found a cluster, get level 2 elements
				KaxCluster * SegmentCluster = static_cast<KaxCluster *>(ElementLevel1);
				ElementLevel2 = m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), false);
				while(ElementLevel2) {
					if(UpperElementLevel > 0)	break;
					if(UpperElementLevel < 0)	UpperElementLevel = 0;
					bool bDeleteLevel2 = true;
					if (EbmlId(*ElementLevel2) == KaxClusterTimecode::ClassInfos.GlobalId) {
						KaxClusterTimecode & ClusterTime = *static_cast<KaxClusterTimecode*>(ElementLevel2);
						ClusterTime.ReadData(m_EbmlStream->I_O());
						SegmentCluster->InitTimecode(ClusterTime, m_TimecodeScale);
						if(bSeekOnFirstBlock) {
							bSeekOnFirstBlock = false;
							m_Parent->SeekFile(MKV_BLOCKPOS(m_Cluster->m_Blocks[blocksToSkip].m_BlockPosition));
							goto mod_faster_mkv_read_access;
						}
					} else  if (EbmlId(*ElementLevel2) == KaxBlockGroup::ClassInfos.GlobalId) {
						// Found blocks, get level 3 elements
mod_faster_mkv_read_access:
						ElementLevel3 = m_EbmlStream->FindNextElement(KaxBlockGroup::ClassInfos.Context, UpperElementLevel, uint64(0xFFFFFFFFFFFFFFFF), false);
						while(ElementLevel3) {
							if(UpperElementLevel > 0)	break;
							if(UpperElementLevel < 0)	UpperElementLevel = 0;
							if( (EbmlId(*ElementLevel3) == KaxBlock::ClassInfos.GlobalId)) {
								/*if(blocksToSkip) {
									KaxBlock & DataBlock = *static_cast<KaxBlock*>(ElementLevel3);
									DataBlock.ReadData(m_EbmlStream->I_O());
									DataBlock.SetParent(*SegmentCluster);
									if(DataBlock.TrackNum() == m_Stream->m_Serial)
										blocksToSkip--;
								} else */{
									KaxBlock & DataBlock = *static_cast<KaxBlock*>(ElementLevel3);
									DataBlock.ReadData(m_EbmlStream->I_O());
									DataBlock.SetParent(*SegmentCluster);
									if(DataBlock.TrackNum() == m_Stream->m_Serial) {
										frameCount = DataBlock.NumberFrames();
										while(!bFinished && (framesToSkip<frameCount) ) {
											long dataSize = DataBlock.GetBuffer(framesToSkip).Size();
											toRead = dataSize-bytesShorten;
											if(toRead > bytesLeft)
												toRead = bytesLeft;

											VDASSERT(toRead > 0);

											memcpy(lpBufferPos, DataBlock.GetBuffer(framesToSkip).Buffer()+bytesShorten, toRead);

											bytesShorten = 0;
											bytesRead += toRead;
											lpBufferPos += toRead;
											bytesLeft -= toRead;
											bFinished = (bytesLeft <= 0);
											framesToSkip++;
										}
										framesToSkip = 0;
									}
								}
							}
							if(bFinished) {
								delete ElementLevel3;
								break;
							} else if (UpperElementLevel > 0) {
								UpperElementLevel--;
								delete ElementLevel3;
								ElementLevel3 = ElementLevel4;
								if(UpperElementLevel > 0)
									break;
							} else {
								ElementLevel3->SkipData(*m_EbmlStream, ElementLevel3->Generic().Context);
								delete ElementLevel3;
								ElementLevel3 = m_EbmlStream->FindNextElement(KaxBlockGroup::ClassInfos.Context, UpperElementLevel, uint64(0xFFFFFFFFFFFFFFFF), false);
							}
						}
					}
					if(bFinished) {
						delete ElementLevel2;
						break;
					} else if (UpperElementLevel > 0) {
						UpperElementLevel--;
						delete ElementLevel2;
						ElementLevel2 = ElementLevel3;
						if(UpperElementLevel > 0)
							break;
					} else {
						ElementLevel2->SkipData(*m_EbmlStream, ElementLevel2->Generic().Context);
						delete ElementLevel2;
						ElementLevel2 = m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), false);
					}
				}
			}
			if(bFinished) {
				delete ElementLevel1;
				break;
			} else if (UpperElementLevel > 0) {
				UpperElementLevel--;
				delete ElementLevel1;
				ElementLevel1 = ElementLevel2;
				if(UpperElementLevel > 0)
					break;
			} else {
				ElementLevel1->SkipData(*m_EbmlStream, ElementLevel1->Generic().Context);
				delete ElementLevel1;
				//ElementLevel1 = m_EbmlStream->FindNextElement(ElementLevel0->Generic().Context, UpperElementLevel, uint64(0xFFFFFFFFFFFFFFFF), false);
				ElementLevel1 = NULL;
			}
		}
		if(!bFinished && !SeekNextSegment()) {
			//VDLogAppMessage(kVDLogError, kVDST_MatroskaReadHandler, kVDM_Seek_BadCluster);
			if(plBytes)		*plBytes = bytesRead;
			if(plSamples)	*plSamples = VDRoundToLong(double(bytesRead - bytesShorten) / m_BlockAlign);
			return AVIERR_OK;
		}
	} catch(exception & Ex) {
		throw MyError(Ex.what());
    }

	VDASSERT(bytesLeft == 0);

	if(plBytes)
		*plBytes = bytesRead;
	if(plSamples) {
		*plSamples = VDRoundToLong(double(bytesRead - bytesShorten) / m_BlockAlign);
		VDASSERT(*plSamples > 0);
	}

	return AVIERR_OK;
}

long MatroskaReadStream::Start() {
	return 0;
}

long MatroskaReadStream::End() {
	return m_Length;
}

long MatroskaReadStream::PrevKeyFrame(long lFrame) {
	if(lFrame <= 0)
		return -1;
	if(lFrame >= m_Length) {
		SeekLastStreamSegment();
		if(!m_KeyFrames)
			return m_Length-1;
		lFrame = m_CurrentStreamSegment->m_End - m_CurrentStreamSegment->m_Start;
	} else if((lFrame=FindStreamSegment(lFrame))<0)
		return -1;

	if(!m_KeyFrames)
		return (m_CurrentStreamSegment->m_Start+lFrame-1);

	do {
		uint8 *keys = m_KeyFrames + (lFrame>>3);
		sint8 offset = lFrame & 7;
		offset--;
		while(keys>=m_KeyFrames) {
			while(*keys && (offset>=0)) {
				if(((*keys)<<offset)&0x80) {
					long lResult = ((keys-m_KeyFrames)<<3) + offset;
					while(_IsBFrame(lResult+1))
						lResult++;
					return (m_CurrentStreamSegment->m_Start + lResult);
				}
				offset--;
			}
			offset = 7;
			keys--;
		}
		if(!SeekPreviousSegment())
			break;
		lFrame = m_CurrentStreamSegment->m_End;
	} while(true);

	return -1;
}

long MatroskaReadStream::NextKeyFrame(long lFrame) {
	if(lFrame >= m_Length)
		return -1;
	if(lFrame < 0) {
		return m_ReadNode->m_FirstKeyFrame;
	} else if((lFrame=FindStreamSegment(lFrame))<0)
		return -1;

	if(!m_KeyFrames)
		return (m_CurrentStreamSegment->m_Start+lFrame+1);

	do {
		uint8 *keys = m_KeyFrames + (lFrame>>3);
		sint8 offset = lFrame & 7;
		offset++;
		while(keys < m_KeyFramesEnd) {
			while(*keys && (offset<=7)) {
				if(((*keys)<<offset)&0x80) {
					long lResult = ((keys-m_KeyFrames)<<3) + offset;
					while(_IsBFrame(lResult+1))
						lResult++;
					return (m_CurrentStreamSegment->m_Start + lResult);
				}
				offset++;
			}
			offset = 0;
			keys++;
		}
		if(!SeekNextSegment())
			return -1;
		if((*m_KeyFrames) & 0x80)
			return m_CurrentStreamSegment->m_Start;
		lFrame = 0;
	} while(true);
	return -1;
}

// This function must return the nearest (previous) key-frame (and not the absolute
// nearest one)
long MatroskaReadStream::NearestKeyFrame(long lFrame) {
	if(lFrame <= 0) {
		//if(m_ReadNode->m_FirstKeyFrame)
		//	return -1;
		return 0;
	}
	if(IsKeyFrame(lFrame))
		return lFrame;

	long lprev = PrevKeyFrame(lFrame);

	if(lprev < 0)
		return -1;
	else
		return lprev;
}

HRESULT MatroskaReadStream::FormatSize(long lFrame, long *plSize) {
	if(plSize)
		*plSize = m_FormatSize;
	return 0;
}

HRESULT MatroskaReadStream::ReadFormat(long lFrame, void *pFormat, long *plSize) {
	if(!pFormat)
		*plSize = m_FormatSize;

	if(*plSize < m_FormatSize)
		memcpy(pFormat, m_Format, *plSize);
	else {
		memcpy(pFormat, m_Format, m_FormatSize);
		*plSize = m_FormatSize;
	}

	return 0;
}

bool MatroskaReadStream::isStreaming() {
	return false;
}

bool MatroskaReadStream::isKeyframeOnly() {
	return m_ReadNode->m_bKeyFrameOnly;
}

bool MatroskaReadStream::hasBFrames(void) {
	return m_ReadNode->m_bHasBFrames;
}

bool MatroskaReadStream::getVBRInfo(double& bitrate_mean, double& bitrate_stddev, double& maxdev) {
	// No there are no VBR stream in this file :p
	return false;
}

void MatroskaReadStream::fixVBRAudio(bool fix, double _bitrate_mean) {
}

sint32 MatroskaReadStream::samplesLag(sint64 lStart) {
	if(!m_bVBRHandling)
		return 0;
	if((lStart=FindStreamSegment(lStart))<0)
		return 0;
	if(MAIN_TYPE(m_Stream->m_StreamType)==SOURCE_TEXT)
		return 0;

	while(m_ClusterIterator.previous() && ((*m_ClusterIterator)->m_Blocks[0].m_Granulepos>lStart) )
		m_ClusterIterator--;
	while(m_ClusterIterator.next() && (m_ClusterIterator.next()->m_Blocks[0].m_Granulepos<=lStart) )
		m_ClusterIterator++;

	m_Cluster = *m_ClusterIterator;

	int blocksToSkip = 0;
	while((blocksToSkip+1 < m_Cluster->m_NbBlocks) && (m_Cluster->m_Blocks[blocksToSkip+1].m_Granulepos<=lStart))
		blocksToSkip++;
	VDFraction estimatedSamplesPerFrame(1);
	if(blocksToSkip+1 < m_Cluster->m_NbBlocks)
		estimatedSamplesPerFrame = VDFraction(m_Cluster->m_Blocks[blocksToSkip+1].m_Granulepos - m_Cluster->m_Blocks[blocksToSkip].m_Granulepos, m_Cluster->m_Blocks[blocksToSkip].m_NbFrames);
	else if(m_ClusterIterator.next())
		estimatedSamplesPerFrame = VDFraction(m_ClusterIterator.next()->m_Blocks[0].m_Granulepos - m_Cluster->m_Blocks[blocksToSkip].m_Granulepos, m_Cluster->m_Blocks[blocksToSkip].m_NbFrames);
	else
		estimatedSamplesPerFrame = VDFraction(m_Stream->m_EstimatedSamplesInLastBlock, m_Cluster->m_Blocks[blocksToSkip].m_NbFrames);
	int framesToSkip = 0;
	while((framesToSkip < m_Cluster->m_Blocks[blocksToSkip].m_NbFrames) && (m_Cluster->m_Blocks[blocksToSkip].m_Granulepos + estimatedSamplesPerFrame.scale64r(framesToSkip+1)<=lStart))
		framesToSkip++;

	sint64 granulepos = m_Cluster->m_Blocks[blocksToSkip].m_Granulepos + estimatedSamplesPerFrame.scale64r(framesToSkip);

	return granulepos-lStart;
}

sint32 MatroskaReadStream::sourceOffsetMS(void) {
	return m_Stream->m_StartOffsetMS;
}

void MatroskaReadStream::GetCommentList(ModList<ModStreamComment>& list) {
	MatroskaStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	if(m_bIsVideo) {
		ModStreamComment *newComment = NULL;
		EbmlElement *anElement = NULL;
		if(m_Parent->SegmentInfo) {
			anElement = m_Parent->SegmentInfo->FindElt(KaxTitle::ClassInfos);
			if(anElement) {
				KaxTitle& MyTitle = *(static_cast<KaxTitle *>(anElement));
				newComment = list.AddLast();
				newComment->m_Type = nsModStream::cTitle;
				newComment->m_Value = UTFstring(MyTitle).c_str();
			}
			anElement = m_Parent->SegmentInfo->FindElt(KaxWritingApp::ClassInfos);
			if(anElement) {
				KaxWritingApp& MyWritingApp = *(static_cast<KaxWritingApp *>(anElement));
				newComment = list.AddLast();
				newComment->m_Type = nsModStream::cSoftware;
				newComment->m_Value = UTFstring(MyWritingApp).c_str();
			}
			anElement = m_Parent->SegmentInfo->FindElt(KaxMuxingApp::ClassInfos);
			if(anElement) {
				KaxMuxingApp& MyMuxingApp = *(static_cast<KaxMuxingApp *>(anElement));
				newComment = list.AddLast();
				newComment->m_Type = nsModStream::cSoftware;
				newComment->m_Value = UTFstring(MyMuxingApp).c_str();
			}
		}
	}
	while(streamNode) {
		streamNode->/*m_MatroskaStream->*/GetCommentList(list);
		streamNode = streamNode->GetNext();
	}
	ModList<ModStreamComment> copyList;
	copyList = list;
	list.DeleteContents(true);
	list.Clear();
	ModList<ModStreamComment>::iterator srcIt(copyList);
	ModList<ModStreamComment>::iterator dstIt;
	while(srcIt) {
		bool bAddComment = true;
		dstIt.begin(list);
		while(dstIt) {
			if((**srcIt) == (**dstIt)) {
				bAddComment = false;
				break;
			}
			dstIt++;
		}
		if(bAddComment) {
			ModStreamComment *newComment = list.AddLast();
			*newComment = *(*srcIt);
		}
		srcIt++;
	}
	copyList.DeleteContents(true);
	copyList.Clear();
}

void MatroskaReadStream::GetChapterList(ModChapterList& list) {
	MatroskaStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	while(streamNode) {
		streamNode->/*m_MatroskaStream->*/GetChapterList(list, (unsigned long)streamNode->m_MatroskaStream->m_SampleRate.scale64ir(1000 * streamNode->m_Start));
		streamNode = streamNode->GetNext();
	}
	ModChapterList copyList;
	copyList = list;
	list.clear();
	ModChapterList::iterator srcIt;
	ModChapterList::iterator dstIt;
	for(srcIt=copyList.begin() ; srcIt!=copyList.end() ; srcIt++) {
		ModChapterList::tChapterNode& srcChapter = (*srcIt);
		bool bAddChapter = true;
		for(dstIt=list.begin() ; dstIt!=list.end() ; dstIt++) {
			ModChapterList::tChapterNode& dstChapter = (*dstIt);
			if( (srcChapter==dstChapter) || (srcChapter.SeemsSimilar(dstChapter)) ) {
				bAddChapter = false;
				break;
			}
		}
		if(bAddChapter)
			list.AddChapter(srcChapter);
	}
	list.Sort();
	copyList.clear();
}

uint32 MatroskaReadStream::GetStreamType(void) {
	MatroskaStreamSegment *firstNode = m_ReadNode->m_Segments.GetFirst();
	return firstNode->m_MatroskaStream->m_StreamType;
}

uint16 MatroskaReadStream::GetNbHeaderPackets(void) {
	MatroskaStream *curStream = m_ReadNode->m_Segments.GetFirst()->m_MatroskaStream;

	VDASSERT(curStream->m_StreamType == SOURCE_AUDIO_VORBIS);

	return curStream->m_OggHeaderPacketList.size();
}

ogg_packet *MatroskaReadStream::GetHeaderPacket(uint16 index) {
	MatroskaStream *curStream = m_ReadNode->m_Segments.GetFirst()->m_MatroskaStream;

	VDASSERT(curStream->m_StreamType == SOURCE_AUDIO_VORBIS);

	if(index >= curStream->m_OggHeaderPacketList.size())
		return NULL;
	return curStream->m_OggHeaderPacketList[index];
}

long MatroskaReadStream::GetStreamIndexByType(void) {
	MatroskaStreamSegment *firstNode = m_ReadNode->m_Segments.GetFirst();
	return firstNode->m_ReadHandler->GetStreamIndexByType(m_ReadNode);
}

VDStringW MatroskaReadStream::writeReport(void) {
	SeekFirstStreamSegment();
	VDStringW report;
	switch(m_Stream->m_StreamType) {
	case SOURCE_VIDEO:
		report = L"Video stream";
		break;
	case SOURCE_TEXT:
	case SOURCE_TEXT_SSA:
		report = L"Text stream";
		break;
	case SOURCE_AUDIO:
		report = L"Audio stream";
		break;
	case SOURCE_AUDIO_VORBIS:
		report = L"Vorbis stream";
		break;
	default:
		report = L"Other stream";
		break;
	}
	MatroskaStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	while(streamNode) {
		report += VDStringW::setf(L"\r\nStream report for file \"%s\"", streamNode->m_ReadHandler->GetFileName().c_str());
		report += streamNode->m_MatroskaStream->writeReport();
		streamNode = streamNode->GetNext();
	}
	return report;
}

ModSubtitleProvider *MatroskaReadStream::GetSubtitleProvider(void) {
	if(!m_bIsText)
		return NULL;

	m_SubtitleProvider.Clear();

	sint64 iOffset = 0;
	long iLastSegmentDuration = -1;
	int iSegment = 0;
	MatroskaStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	while(streamNode) {
		if(iLastSegmentDuration > 0)
			iOffset += iLastSegmentDuration;
		else
			iOffset = streamNode->m_Start;
		ModSubtitleProvider& currentProvider = streamNode->m_MatroskaStream->m_SubtitleProvider;

		std::list<SubtitleFrame>& currentList = currentProvider.GetSourceSubtitles();
		std::list<SubtitleFrame>::iterator subIt;
		for(subIt=currentList.begin() ; subIt!=currentList.end() ; subIt++) {
			SubtitleFrame *currentSub = &(*subIt);

			m_SubtitleProvider.AddSubtitle(currentSub->m_Subtitle, iOffset + currentSub->m_StartMS, currentSub->m_DurationMS);
		}

		iLastSegmentDuration = m_Parent->GetSegmentDuration(iSegment++);
		streamNode = streamNode->GetNext();
	}

	if(m_SubtitleProvider.HasCollisions())
		return &m_SubtitleProvider;
	return NULL;
}

//////////////////////////////////////////////////////////////////////

MatroskaStream::MatroskaStream(void)
: m_TextUtil(kTextType_UTF8, kTextType_UTF8, false, false)
{
	m_RefCount				= 1;
	m_Serial				= -1;
	m_UID					= 0;
	m_TimecodeScale			= 1000000;
	m_DefaultDuration		= 0;
	m_SampleRate			= VDFraction(0);
	m_EstimatedSamplesInLastBlock = 0;
	m_CurrentCluster		= NULL;
	m_CurrentBlock			= NULL;
	m_bGranuleposComputed	= false;
	m_TrackEntry			= NULL;
	m_CodecID				= NULL;
	m_CodecPrivate			= NULL;
	memset(&m_Specific, 0, sizeof(m_Specific));
	m_StreamType			= 0;
	m_bIsVideo				= false;
	m_NbClusters			= 0;
	m_NbClustersCorrupted	= 0;
	m_NbBlocks				= 0;
	m_NbFrames				= 0;
	m_MaxFrameSize			= 0;
	m_StartGranulepos		= 0;
	m_StartOffsetMS			= 0;
	m_Bytes					= 0;
	m_LastGranulepos		= 0;
	m_NbSamples				= 0;
	m_bIndexBuilt			= false;
	m_bIsText				= false;
}

MatroskaStream::~MatroskaStream(void) {
	m_ClusterList.Clear();
	m_KeyIndex.clear();
	m_BIndex.clear();
	while(!m_OggHeaderPacketList.empty()) {
		ogg_packet *packet = m_OggHeaderPacketList.back();
		m_OggHeaderPacketList.pop_back();
		if(packet->packet) {
			delete[] packet->packet;
			packet->packet = NULL;
		}
		delete packet;
	}
	if(m_TrackEntry)
		delete m_TrackEntry;
	else {
		if(m_CodecID)
			delete m_CodecID;
		if(m_CodecPrivate)
			delete m_CodecPrivate;
	}
}

void MatroskaStream::AddRef(void) {
	m_RefCount++;
}

void MatroskaStream::Release(void) {
	if(!--m_RefCount)
		delete this;
}

void MatroskaStream::AddOggHeaderPacket(binary *data, uint32 dataSize) {
	ogg_packet *new_packet = new ogg_packet;
	memset(new_packet, 0, sizeof(ogg_packet));
	if(dataSize > 0) {
		new_packet->bytes = dataSize;
		new_packet->packet = new unsigned char[dataSize];
		memcpy(new_packet->packet, data, dataSize);
	} else {
		new_packet->bytes = 0;
		new_packet->packet = NULL;
	}
	new_packet->packetno = m_OggHeaderPacketList.size();
	if(!new_packet->packetno)
		new_packet->b_o_s = 1;
	m_OggHeaderPacketList.push_back(new_packet);
}

VDStringW MatroskaStream::writeReport(void) {
	VDStringW report;
	return report;
}

/*void MatroskaStream::GetCommentList(ModList<ModStreamComment>& list) {
	if(m_TrackEntry) {
		EbmlElement *anElement = NULL;
		anElement = m_TrackEntry->FindElt(KaxTrackLanguage::ClassInfos);
		if(anElement) {
			KaxTrackLanguage& MyLanguage = *static_cast<KaxTrackLanguage*>(anElement);
			std::string lang = std::string(MyLanguage);
			ModStreamComment *newModComment = list.AddLast();
			newModComment->m_Value = VDTextAToW(lang.c_str(), -1);
			sint32 langIndex = 0;
			VDStringW tmpValue(newModComment->m_Value.c_str());
			langIndex = ISO639_GetIndex_2(tmpValue.c_str()) + 1;
			if(!langIndex)
				langIndex = ISO639_GetIndex(tmpValue.c_str()) + 1;
			newModComment->m_Type = (nsModStream::eCommentType)(nsModStream::cLanguage + langIndex);
		}
		anElement = m_TrackEntry->FindElt(KaxTrackName::ClassInfos);
		if(anElement) {
			KaxTrackName& MyName = *static_cast<KaxTrackName*>(anElement);
			ModStreamComment *newModComment = list.AddLast();
			newModComment->m_Type = nsModStream::cTitle;
			newModComment->m_Value = UTFstring(MyName).c_str();
		}
	}
}

void MatroskaStream::GetChapterList(ModChapterList& list, uint32 offsetMS) {
}*/

//////////////////////////////////////////////////////////////////////

MatroskaStreamSegment::MatroskaStreamSegment(void) {
	m_ReadHandler	= NULL;
	m_MatroskaStream = NULL;
	m_Start			= 0;
	m_End			= 0;
	m_LastFilePos	= 0;
}

MatroskaStreamSegment::~MatroskaStreamSegment(void) {
	if(m_MatroskaStream)
		m_MatroskaStream->Release();
}

void MatroskaStreamSegment::UpdateState(bool bVBRHandling) {
	m_Start = 0;
	bool bIsVideo = (MAIN_TYPE(m_MatroskaStream->m_StreamType) == SOURCE_VIDEO);

	VDASSERT(bIsVideo == m_MatroskaStream->m_bIsVideo);

	if((m_MatroskaStream->m_StreamType==SOURCE_AUDIO_VORBIS) && (!m_MatroskaStream->m_OggHeaderPacketList.size())) {

		VDASSERT(m_MatroskaStream->m_CodecPrivate && m_MatroskaStream->m_CodecPrivate->GetSize());

		if(m_MatroskaStream->m_CodecPrivate && m_MatroskaStream->m_CodecPrivate->GetSize()) {
			binary *codecPrivate = m_MatroskaStream->m_CodecPrivate->GetBuffer();
			uint32 lastHeaderSize = m_MatroskaStream->m_CodecPrivate->GetSize() - 1;
			uint16 nbHeaders = (*(uint8 *)codecPrivate);

			std::list<uint32> headerSizeList;
			codecPrivate++;
			while(nbHeaders--) {
				uint32 currentHeaderSize = 0;
				do {
					currentHeaderSize += *(uint8 *)codecPrivate;
					lastHeaderSize--;
				} while((*codecPrivate++) == 0xFF);
				lastHeaderSize -= currentHeaderSize;
				headerSizeList.push_back(currentHeaderSize);
			}
			headerSizeList.push_back(lastHeaderSize);
			while(!headerSizeList.empty()) {
				uint32 currentHeaderSize = headerSizeList.front();
				headerSizeList.pop_front();
				m_MatroskaStream->AddOggHeaderPacket(codecPrivate, currentHeaderSize);
				codecPrivate += currentHeaderSize;
			}
		}
	}

	ModCompactList<MatroskaCluster>::iterator clusterIt;

	if(!m_MatroskaStream->m_SampleRate.getHi())
		return;

	if(!m_MatroskaStream->m_bGranuleposComputed) {
		VDFraction factor = m_MatroskaStream->m_SampleRate / VDFraction(1000000000, m_MatroskaStream->m_TimecodeScale);
		VDFraction factor_2 = VDFraction::reduce64(m_MatroskaStream->m_TimecodeScale, 1000000);
		VDFraction factor_3 = VDFraction::reduce64(VD64(1000) * m_MatroskaStream->m_SampleRate.getLo() , m_MatroskaStream->m_SampleRate.getHi());
		factor.reduce();
		factor_2.reduce();
		factor_3.reduce();
		m_MatroskaStream->m_StartOffsetMS = 0;
		m_MatroskaStream->m_StartGranulepos = 0;
		bool bStartGranuleposComputed = false;
		uint64 computedGranulepos = 0;
		sint64 iFirstDiff = 0;
		bool bFirstDiff = false;
		bool bMaybeVFR = false;
		clusterIt.begin(m_MatroskaStream->m_ClusterList);
		while(clusterIt) {
			MatroskaCluster *curCluster = *clusterIt;
			if(!bStartGranuleposComputed && !bIsVideo && !m_MatroskaStream->m_bIsText && curCluster->m_NbBlocks) {
				m_MatroskaStream->m_StartGranulepos = factor.scale64r(MKV_GRANULEPOS(curCluster->m_Blocks[0].m_Granulepos));
				m_MatroskaStream->m_StartOffsetMS = m_MatroskaStream->m_SampleRate.scale64ir(VD64(1000)*m_MatroskaStream->m_StartGranulepos);
				bStartGranuleposComputed = true;
			}
			for(uint16 i=0 ; i<curCluster->m_NbBlocks ; i++) {
				uint64 flags = MKV_GETFLAGS(curCluster->m_Blocks[i].m_Granulepos);
				computedGranulepos = factor.scale64r(MKV_GRANULEPOS(curCluster->m_Blocks[i].m_Granulepos));
				if(bIsVideo && !bMaybeVFR) {
					if(!bFirstDiff) {
						if(MKV_GRANULEPOS(curCluster->m_Blocks[i].m_Granulepos)) {
							iFirstDiff = factor_2.scale64r(MKV_GRANULEPOS(curCluster->m_Blocks[i].m_Granulepos));
							iFirstDiff -= factor_3.scale64r(computedGranulepos);
							bFirstDiff = true;
						}
					} else {
						sint64 iDiff = factor_2.scale64r(MKV_GRANULEPOS(curCluster->m_Blocks[i].m_Granulepos));
						iDiff -= factor_3.scale64r(computedGranulepos);
						if((iDiff - iFirstDiff < -2) || (iDiff - iFirstDiff > 2))
							bMaybeVFR = true;
					}
				}

				if(!bIsVideo) {
					VDASSERT(computedGranulepos >= m_MatroskaStream->m_StartGranulepos);

					if(computedGranulepos >= m_MatroskaStream->m_StartGranulepos)
						computedGranulepos -= m_MatroskaStream->m_StartGranulepos;
					else
						computedGranulepos = 0;
				}
				curCluster->m_Blocks[i].m_Granulepos = computedGranulepos | flags;
			}
			clusterIt++;
		}
		if(bMaybeVFR)
			VDLogAppMessage(kVDLogWarning, kVDST_MatroskaReadHandler, kVDM_MaybeVFR);
		m_MatroskaStream->m_LastGranulepos = 0;
		clusterIt.end(m_MatroskaStream->m_ClusterList);
		while(clusterIt) {
			MatroskaCluster *curCluster = *clusterIt;
			if(curCluster->m_NbBlocks) {
				m_MatroskaStream->m_LastGranulepos = MKV_GRANULEPOS(curCluster->m_Blocks[curCluster->m_NbBlocks-1].m_Granulepos);
				break;
			}
			clusterIt--;
		}
		if(m_MatroskaStream->m_EstimatedSamplesInLastBlock > 0)
			m_MatroskaStream->m_EstimatedSamplesInLastBlock = factor.scale64r(m_MatroskaStream->m_EstimatedSamplesInLastBlock);

		m_MatroskaStream->m_bGranuleposComputed = true;
	}

	if(bIsVideo)
		m_MatroskaStream->m_EstimatedSamplesInLastBlock = 1;
	else {
		if(m_MatroskaStream->m_bIsText) {
			SubtitleFrame *lastFrame = NULL;
			if(m_MatroskaStream->m_SubtitleProvider.GetSourceSubtitles().size() != 0)
				lastFrame = &(m_MatroskaStream->m_SubtitleProvider.GetSourceSubtitles().back());
			m_MatroskaStream->m_EstimatedSamplesInLastBlock = (lastFrame ? lastFrame->m_DurationMS : 0);
		} else if(m_MatroskaStream->m_EstimatedSamplesInLastBlock <= 0)
			m_MatroskaStream->m_EstimatedSamplesInLastBlock = VDRoundToLong((double)m_MatroskaStream->m_LastGranulepos / (m_MatroskaStream->m_NbBlocks-1));
		if(m_MatroskaStream->m_EstimatedSamplesInLastBlock<=0)
			m_MatroskaStream->m_EstimatedSamplesInLastBlock = 1;
	}

	m_MatroskaStream->m_NbSamples = m_MatroskaStream->m_LastGranulepos + m_MatroskaStream->m_EstimatedSamplesInLastBlock;

	if(m_MatroskaStream->m_bIsText) {
		SubtitleFrame *lastFrame = NULL;
		if(m_MatroskaStream->m_SubtitleProvider.GetSourceSubtitles().size() != 0)
			lastFrame = &(m_MatroskaStream->m_SubtitleProvider.GetSourceSubtitles().back());
		m_End = (lastFrame ? lastFrame->m_StartMS+lastFrame->m_DurationMS : 0);
	} else if(bVBRHandling)
		m_End = m_MatroskaStream->m_NbSamples;
	else
		m_End = m_MatroskaStream->m_Bytes;

	// Build a compact key index :)
	if(bIsVideo && !m_MatroskaStream->m_bIndexBuilt) {
		uint32 nbChars = (m_MatroskaStream->m_NbSamples+7)>>3;

		m_MatroskaStream->m_KeyIndex.resize(nbChars);
		m_MatroskaStream->m_BIndex.resize(nbChars);

		bool bOnlyKeys = true;
		bool bNoB = true;
		uint8 *curPos_Key = &m_MatroskaStream->m_KeyIndex.front();
		uint8 *endPos_Key = &m_MatroskaStream->m_KeyIndex.back() + 1;
		uint8 *curPos_B = &m_MatroskaStream->m_BIndex.front();
		uint8 counter = 8;
		uint8 vKeys = 0;
		uint8 vB = 0;
		MatroskaBlock *curBlock;
		clusterIt.begin(m_MatroskaStream->m_ClusterList);
		// See what frame is the first one
		if(clusterIt && (*clusterIt)->m_NbBlocks) {
			long firstFrame = (*clusterIt)->m_Blocks[0].m_Granulepos;
			curPos_Key += (firstFrame >> 3);
			curPos_B += (firstFrame >> 3);
			counter -= firstFrame & 7;
			bOnlyKeys = false;
		}
		while(clusterIt) {
			MatroskaCluster *curCluster = *clusterIt;
			uint16 blockNum = curCluster->m_NbBlocks;
			uint8 nbFrames;
			curBlock = curCluster->m_Blocks;
			while(blockNum--) {
				bool bKey = !!MKV_GETFLAG_KEY(curBlock->m_Granulepos);
				bool bB = !!MKV_GETFLAG_B(curBlock->m_Granulepos);
				curBlock->m_Granulepos = MKV_GRANULEPOS(curBlock->m_Granulepos);
				nbFrames = curBlock->m_NbFrames;
				sint32 nbSamples = 0;
				if(blockNum)
					nbSamples = MKV_GRANULEPOS((curBlock+1)->m_Granulepos) - curBlock->m_Granulepos;
				else if(clusterIt.next() && clusterIt.next()->m_NbBlocks)
					nbSamples = MKV_GRANULEPOS(clusterIt.next()->m_Blocks[0].m_Granulepos) - curBlock->m_Granulepos;
				else
					nbSamples = m_MatroskaStream->m_EstimatedSamplesInLastBlock;
				if(nbSamples < 0)
					nbSamples = 0;
				if(!nbFrames) {
					bKey = false;
					bB = false;
				}
#ifdef _DEBUG
				VDASSERT(nbSamples >= nbFrames);
				VDASSERT(nbFrames == 1);

				if(nbSamples > nbFrames)
					VDDEBUG("Found %ld dropped frames after frame %ld\n"
					, nbSamples - nbFrames, (sint32(curPos_Key-&m_MatroskaStream->m_KeyIndex.front())<<3)+(8-counter));
				if(nbSamples < nbFrames)
					VDDEBUG("Found %ld frames (max %ld according to timestamps) after frame %ld\n"
					, (sint32)nbFrames, nbSamples, (sint32(curPos_Key-&m_MatroskaStream->m_KeyIndex.front())<<3)+(8-counter));
#endif
				if(nbSamples > 0)
				while( ((nbFrames)&&(nbFrames--)) || (nbSamples>0) ) {
					nbSamples--;
					vKeys <<= 1;
					vB <<= 1;
					if(bKey)
						vKeys |= 1;
					else {
						bOnlyKeys = false;
						if(bB) {
							vB |= 1;
							bNoB = false;
						}
					}
					if(!--counter) {
						*curPos_Key = vKeys;
						*curPos_B = vB;
						vKeys = 0;
						vB = 0;
						curPos_Key++;
						curPos_B++;
						counter = 8;
						VDASSERT( (curPos_Key<endPos_Key)
							|| ((blockNum==0) && (!clusterIt.next() || !clusterIt.next()->m_NbBlocks) ));
					}
					if(!nbFrames) {
						bKey = false;
						bB = false;
					}
				}
				curBlock++;
			}
			clusterIt++;
		}
		if(counter < 8) {
			vKeys <<= counter;
			vB <<= counter;
			*curPos_Key = vKeys;
			*curPos_B = vB;
			VDASSERT( (curPos_Key==&m_MatroskaStream->m_KeyIndex.back()) && (counter==8-((m_End-m_Start)%8)) );
		} else
			VDASSERT(curPos_Key==endPos_Key);
		if(bOnlyKeys) {
			VDDEBUG("Video stream only contains keyframes\n");
			m_MatroskaStream->m_KeyIndex.clear();
		}
		if(bNoB) {
			VDDEBUG("Video stream doesn't contain B-frames\n");
			m_MatroskaStream->m_BIndex.clear();
		}
		m_MatroskaStream->m_bIndexBuilt = true;
	}
}

void MatroskaStreamSegment::GetCommentList(ModList<ModStreamComment>& list) {
	EbmlElement *anElement = NULL;
	ModStreamComment *newComment = NULL;
	ModStreamComment *languageComment = NULL;
	if(m_MatroskaStream->m_TrackEntry) {
		anElement = m_MatroskaStream->m_TrackEntry->FindElt(KaxTrackLanguage::ClassInfos);
		if(anElement) {
			KaxTrackLanguage& MyLanguage = *static_cast<KaxTrackLanguage*>(anElement);
			std::string lang = std::string(MyLanguage);
			languageComment = list.AddLast();
			languageComment->m_Type = nsModStream::cLanguage;
			languageComment->m_Value = VDTextAToW(lang.c_str(), -1);
		}
		anElement = m_MatroskaStream->m_TrackEntry->FindElt(KaxTrackName::ClassInfos);
		if(anElement) {
			KaxTrackName& MyName = *static_cast<KaxTrackName*>(anElement);
			newComment = list.AddLast();
			newComment->m_Type = nsModStream::cName;
			newComment->m_Value = UTFstring(MyName).c_str();
		}
	}
	KaxTags *tags = (m_ReadHandler ? m_ReadHandler->GetTags() : NULL);
	if(tags) {
		EbmlElement *aTag = tags->FindFirstElt(KaxTag::ClassInfos, false);
		while(aTag) {
			/* Generic process */
			KaxTag& MyTag = *static_cast<KaxTag *>(aTag);
			anElement = MyTag.FindElt(KaxTagTargets::ClassInfos);
			bool bProcess = !anElement;
			if(anElement) {
				KaxTagTargets& MyTargets = *static_cast<KaxTagTargets *>(anElement);
				EbmlElement *aTarget = MyTargets.FindFirstElt(KaxTagTrackUID::ClassInfos, false);
				while(aTarget) {
					if(LIBEBML_uint32(*static_cast<EbmlUInteger *>(aTarget)) == m_MatroskaStream->m_UID) {
						bProcess = true;
						break;
					}
					aTarget = MyTargets.FindNextElt(*aTarget, false);
				}
			}
			if(bProcess) {
				/* 'New' tags */
				EbmlElement *aSimpleTag = MyTag.FindFirstElt(KaxTagSimple::ClassInfos);
				while(aSimpleTag) {
					KaxTagSimple& MySimpleTag = *static_cast<KaxTagSimple *>(aSimpleTag);
					anElement = MySimpleTag.FindElt(KaxTagName::ClassInfos);
					if(anElement) {
						VDStringW tagName;
						tagName = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
						eCommentType tagType = cSpecific;
						bool bOK = true;
						if(GetMatroskaCommentType(tagName.c_str(), tagType) && (tagType==cLanguage))
							bOK = !languageComment;
						anElement = MySimpleTag.FindElt(KaxTagString::ClassInfos);
						if(bOK && anElement) {
							newComment = list.AddLast();
							newComment->m_Type = tagType;
							newComment->m_Tag = tagName;
							newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
							if(tagType == cLanguage)
								languageComment = newComment;
						}
					}
					aSimpleTag = MyTag.FindNextElt(*aSimpleTag, false);
				}
				/* 'Old' tags */
				anElement = MyTag.FindElt(KaxTagGeneral::ClassInfos);
				if(anElement) {
					KaxTagGeneral& MyGeneral = *static_cast<KaxTagGeneral *>(anElement);
					anElement = MyGeneral.FindElt(KaxTagSubject::ClassInfos);
					if(anElement) {
						newComment = list.AddLast();
						newComment->m_Type = nsModStream::cDescription;
						newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
					}
					anElement = MyGeneral.FindElt(KaxTagLanguage::ClassInfos);
					if(anElement) {
						if(!languageComment)
							languageComment = list.AddLast();
						languageComment->m_Type = nsModStream::cLanguage;
						languageComment->m_Value = VDTextAToW(std::string(*static_cast<EbmlString *>(anElement)).c_str(), -1);
					}
					anElement = MyGeneral.FindFirstElt(KaxTagKeywords::ClassInfos, false);
					while(anElement) {
						newComment = list.AddLast();
						newComment->m_Type = nsModStream::cKeywords;
						newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
						anElement = MyGeneral.FindNextElt(*anElement, false);
					}
				}
				anElement = MyTag.FindElt(KaxTagGenres::ClassInfos);
				if(anElement) {
					KaxTagGenres& MyGenres = *static_cast<KaxTagGenres *>(anElement);
					anElement = MyGenres.FindFirstElt(KaxTagAudioGenre::ClassInfos, false);
					while(anElement) {
						newComment = list.AddLast();
						newComment->m_Type = nsModStream::cGenre;
						newComment->m_Value = VDTextAToW(std::string(*static_cast<EbmlString *>(anElement)).c_str(), -1);
						anElement = MyGenres.FindNextElt(*anElement, false);
					}
					anElement = MyGenres.FindElt(KaxTagSubGenre::ClassInfos);
					if(anElement) {
						newComment = list.AddLast();
						newComment->m_Type = nsModStream::cGenre;
						newComment->m_Value = VDTextAToW(std::string(*static_cast<EbmlString *>(anElement)).c_str(), -1);
					}
				}
				EbmlElement *aComment = MyTag.FindFirstElt(KaxTagMultiComment::ClassInfos, false);
				while(aComment) {
					KaxTagMultiComment& MyMultiComment = *static_cast<KaxTagMultiComment *>(aComment);
					newComment = list.AddLast();
					newComment->m_Type = nsModStream::cComment;
					anElement = MyMultiComment.FindElt(KaxTagMultiCommentName::ClassInfos);
					if(anElement) {
						newComment->m_Type = nsModStream::cGeneric;
						newComment->m_Tag = VDTextAToW(std::string(*static_cast<EbmlString *>(anElement)).c_str(), -1);
					}
					anElement = MyMultiComment.FindElt(KaxTagMultiCommentComments::ClassInfos);
					if(anElement)
						newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
					aComment = MyTag.FindNextElt(*aComment, false);
				}
				anElement = MyTag.FindElt(KaxTagMultiIdentifier::ClassInfos);
				if(anElement) {
					KaxTagMultiIdentifier& MyMultiIdentifier = *static_cast<KaxTagMultiIdentifier *>(anElement);
					EbmlElement *aIdentifier = MyMultiIdentifier.FindFirstElt(KaxTagIdentifier::ClassInfos, false);
					while(aIdentifier) {
						KaxTagIdentifier& MyIdentifier = *static_cast<KaxTagIdentifier *>(aIdentifier);
						anElement = MyIdentifier.FindElt(KaxTagMultiIdentifierType::ClassInfos);
						if(anElement) {
							uint32 type = LIBEBML_uint32(*static_cast<EbmlUInteger *>(anElement));
							switch(type) {
							case KaxTagMultiIdentifierType_ISRC :
								{
									newComment = list.AddLast();
									newComment->m_Type = nsModStream::cISRC;
									anElement = MyIdentifier.FindElt(KaxTagMultiIdentifierString::ClassInfos);
									if(anElement)
										newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
								}
								break;
							}
						}
						aIdentifier = MyMultiIdentifier.FindNextElt(*aIdentifier, false);
					}
				}
				anElement = MyTag.FindElt(KaxTagMultiEntity::ClassInfos);
				if(anElement) {
					KaxTagMultiEntity& MyMultiEntity = *static_cast<KaxTagMultiEntity *>(anElement);
					EbmlElement *aEntity = MyMultiEntity.FindFirstElt(KaxTagEntity::ClassInfos, false);
					while(aEntity) {
						KaxTagEntity& MyEntity = *static_cast<KaxTagEntity *>(aEntity);
						anElement = MyEntity.FindElt(KaxTagMultiEntityType::ClassInfos);
						if(anElement) {
							uint32 type = LIBEBML_uint32(*static_cast<EbmlUInteger *>(anElement));
							bool bAdd = false;
							nsModStream::eCommentType cType;
							switch(type) {
							case KaxTagMultiEntitiesType_CommissionedBy :
								bAdd = true;
								cType = nsModStream::cCommissioned;
								break;
							case KaxTagMultiEntitiesType_Composer :
								bAdd = true;
								cType = nsModStream::cAuthor;
								break;
							case KaxTagMultiEntitiesType_LeadPerformerSoloist :
								bAdd = true;
								cType = nsModStream::cPerformer;
								break;
							}
							if(bAdd) {
								newComment = list.AddLast();
								newComment->m_Type = cType;
								anElement = MyEntity.FindElt(KaxTagMultiEntityName::ClassInfos);
								if(anElement)
									newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
							}
						}
						aEntity = MyMultiEntity.FindNextElt(*aEntity, false);
					}
				}
				anElement = MyTag.FindElt(KaxTagMultiLegal::ClassInfos);
				if(anElement) {
					KaxTagMultiLegal& MyMultiLegal = *static_cast<KaxTagMultiLegal *>(anElement);
					EbmlElement *aLegal = MyMultiLegal.FindFirstElt(KaxTagLegal::ClassInfos, false);
					while(aLegal) {
						KaxTagLegal& MyLegal = *static_cast<KaxTagLegal *>(aLegal);
						anElement = MyLegal.FindElt(KaxTagMultiLegalType::ClassInfos);
						if(anElement) {
							uint32 type = LIBEBML_uint32(*static_cast<EbmlUInteger *>(anElement));
							bool bAdd = false;
							nsModStream::eCommentType cType;
							switch(type) {
							case KaxTagMultiLegalType_Copyright :
								bAdd = true;
								cType = nsModStream::cCopyright;
								break;
							case KaxTagMultiLegalType_TermsOfUse :
								bAdd = true;
								cType = nsModStream::cLicense;
								break;
							}
							if(bAdd) {
								newComment = list.AddLast();
								newComment->m_Type = cType;
								anElement = MyLegal.FindElt(KaxTagMultiLegalContent::ClassInfos);
								if(anElement)
									newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
							}
						}
						aLegal = MyMultiLegal.FindNextElt(*aLegal, false);
					}
				}
				anElement = MyTag.FindElt(KaxTagMultiTitle::ClassInfos);
				if(anElement) {
					KaxTagMultiTitle& MyMultiTitle = *static_cast<KaxTagMultiTitle *>(anElement);
					EbmlElement *aTitle = MyMultiTitle.FindFirstElt(KaxTagTitle::ClassInfos, false);
					while(aTitle) {
						KaxTagTitle& MyTitle = *static_cast<KaxTagTitle *>(aTitle);
						anElement = MyTitle.FindElt(KaxTagMultiTitleType::ClassInfos);
						if(anElement) {
							uint32 type = LIBEBML_uint32(*static_cast<EbmlUInteger *>(anElement));
							bool bAdd = false;
							nsModStream::eCommentType cType;
							switch(type) {
							case KaxTagMultiTitleType_AlbumMovieShowTitle :
								bAdd = true;
								cType = nsModStream::cAlbum;
								break;
							case KaxTagMultiTitleType_TrackTitle :
								bAdd = true;
								cType = nsModStream::cTitle;
								break;
							}
							if(bAdd) {
								newComment = list.AddLast();
								newComment->m_Type = cType;
								anElement = MyTitle.FindElt(KaxTagMultiTitleName::ClassInfos);
								if(anElement)
									newComment->m_Value = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
							}
						}
						aTitle = MyMultiTitle.FindNextElt(*aTitle, false);
					}
				}
			}
			aTag = tags->FindNextElt(*aTag, false);
		}
	}
	if(languageComment) {
		sint32 langIndex = 0;
		VDStringW tmpValue(languageComment->m_Value.c_str());
		langIndex = ISO639_GetIndex_2(tmpValue.c_str()) + 1;
		if(!langIndex)
			langIndex = ISO639_GetIndex(tmpValue.c_str()) + 1;
		languageComment->m_Type = (nsModStream::eCommentType)(nsModStream::cLanguage + langIndex);
	}
	RemoveDuplicata(list);
}

void MatroskaStreamSegment::GetChapterList(ModChapterList& list, uint32 offsetMS) {
	EbmlElement *anElement = NULL;
	KaxChapters *chapters = (m_ReadHandler ? m_ReadHandler->GetChapters() : NULL);
	if(chapters) {
		EbmlElement *aEntry = chapters->FindFirstElt(KaxEditionEntry::ClassInfos, false);
		while(aEntry) {
			KaxEditionEntry& MyEntry = *static_cast<KaxEditionEntry *>(aEntry);
			EbmlElement *aAtom = MyEntry.FindFirstElt(KaxChapterAtom::ClassInfos, false);
			while(aAtom) {
				KaxChapterAtom& MyAtom = *static_cast<KaxChapterAtom *>(aAtom);
				anElement = MyAtom.FindElt(KaxChapterTrack::ClassInfos);
				bool bProcess = !anElement;
				if(anElement) {
					KaxChapterTrack& MyTracks = *static_cast<KaxChapterTrack *>(anElement);
					EbmlElement *aTrack = MyTracks.FindFirstElt(KaxChapterTrackNumber::ClassInfos, false);
					while(aTrack) {
						if(LIBEBML_uint32(*static_cast<EbmlUInteger *>(aTrack)) == m_MatroskaStream->m_UID) {
							bProcess = true;
							break;
						}
						aTrack = MyTracks.FindNextElt(*aTrack, false);
					}
				}
				if(bProcess) {
					uint64 chapterStart = 0;
					VDStringW chapterName;
					anElement = MyAtom.FindElt(KaxChapterTimeStart::ClassInfos);
					if(anElement)
						chapterStart = LIBEBML_uint64(*static_cast<EbmlUInteger *>(anElement));
					anElement = MyAtom.FindFirstElt(KaxChapterDisplay::ClassInfos, false);
					if(anElement) {
						KaxChapterDisplay& MyDisplay = *static_cast<KaxChapterDisplay *>(anElement);
						anElement = MyDisplay.FindElt(KaxChapterString::ClassInfos);
						if(anElement)
							chapterName = UTFstring(*static_cast<EbmlUnicodeString *>(anElement)).c_str();
						list.AddChapter(VDRoundToLong(double(chapterStart) / 1000000.), chapterName.c_str());
					}
				}
				aAtom = MyEntry.FindNextElt(*aAtom, false);
			}
			aEntry = chapters->FindNextElt(*aEntry, false);
		}
	}
}

//////////////////////////////////////////////////////////////////////

MatroskaReadNode::MatroskaReadNode(void) {
	m_Start			= 0;
	m_End			= 0;
	m_LastFilePos	= 0;
	m_NbSamples		= 0;
	m_Bytes			= 0;
	m_FirstKeyFrame	= 0;
	m_BlockAlign	= 1;
	m_MaxFrameSize	= 0;
	m_bKeyFrameOnly	= false;
	m_bHasBFrames	= false;
}

MatroskaReadNode::~MatroskaReadNode(void) {
	m_Segments.DeleteContents(true);
	m_Segments.Clear();
}

void MatroskaReadNode::UpdateState(bool bVBRHandling) {
	m_Start			= 0;
	m_End			= 0;
	m_LastFilePos	= 0;
	m_NbSamples		= 0;
	m_Bytes			= 0;
	m_FirstKeyFrame	= 0;
	m_BlockAlign	= 1;
	m_MaxFrameSize	= 0;
	m_bKeyFrameOnly	= !!m_Segments.GetCount();
	m_bHasBFrames	= false;
	bool bHaveSameBlockAlign = true;
	ModList<MatroskaStreamSegment>::iterator it(m_Segments);
	MatroskaStreamSegment *curSegment = NULL;
	while(it) {
		curSegment = *it;
		curSegment->UpdateState(bVBRHandling);

		sint64 segmentLength = curSegment->m_End - curSegment->m_Start;

		curSegment->m_Start = m_End;
		curSegment->m_End =
		m_End = curSegment->m_Start + segmentLength;
		m_LastFilePos += curSegment->m_LastFilePos;
		m_NbSamples += curSegment->m_MatroskaStream->m_NbSamples;
		m_Bytes += curSegment->m_MatroskaStream->m_Bytes;
		if(curSegment->m_MatroskaStream->m_MaxFrameSize > m_MaxFrameSize)
			m_MaxFrameSize = curSegment->m_MatroskaStream->m_MaxFrameSize;

		if(curSegment->m_MatroskaStream->m_KeyIndex.size())
			m_bKeyFrameOnly = false;
		if(curSegment->m_MatroskaStream->m_BIndex.size())
			m_bHasBFrames = true;

		it++;
	}

	curSegment = m_Segments.GetFirst();
	bool bFound = false;
	while(curSegment && curSegment->m_MatroskaStream->m_KeyIndex.size() && !bFound) {
		if(curSegment->m_MatroskaStream->m_KeyIndex[0] & 0x80) {
			bFound = true;
			break;
		}
		uint8 *keys = &curSegment->m_MatroskaStream->m_KeyIndex.front();
		uint8 *keys_end = &curSegment->m_MatroskaStream->m_KeyIndex.back() + 1;
		sint8 offset = 0;
		offset++;
		while(!bFound && (keys < keys_end) ) {
			while(*keys && (offset<=7)) {
				if(((*keys)<<offset)&0x80) {
					m_FirstKeyFrame = (curSegment->m_Start + ((keys-&curSegment->m_MatroskaStream->m_KeyIndex.front())<<3) + offset);
					bFound = true;
					break;
				}
				offset++;
			}
			offset = 0;
			keys++;
		}
		if(bFound)
			break;
		if(!(curSegment = curSegment->GetNext()))
			break;
		if(curSegment->m_MatroskaStream->m_KeyIndex[0] & 0x80) {
			m_FirstKeyFrame = curSegment->m_Start;
			bFound = true;
			break;
		}
	}
}

void MatroskaReadNode::AppendSegment(MatroskaStreamSegment *segment) {
	if(!segment)
		return;

	MatroskaStreamSegment *newSegment = m_Segments.AddLast();

	newSegment->m_ReadHandler = segment->m_ReadHandler;
	newSegment->m_MatroskaStream = segment->m_MatroskaStream;
	newSegment->m_Start = m_End;
	newSegment->m_End =
	m_End = newSegment->m_Start + (segment->m_End - segment->m_Start);
	newSegment->m_LastFilePos = segment->m_LastFilePos;
	newSegment->m_ReadHandler->AddRef();
	newSegment->m_MatroskaStream->AddRef();

	UpdateState(true);
}

void MatroskaReadNode::AddRef(void) {
	ModList<MatroskaStreamSegment>::iterator it(m_Segments);
	while(it) {
		(*it)->m_ReadHandler->AddRef();
		it++;
	}
}

void MatroskaReadNode::Release(void) {
	std::list<MatroskaReadHandler *> tmpList;
	ModList<MatroskaStreamSegment>::iterator it(m_Segments);
	while(it) {
		tmpList.push_back((*it)->m_ReadHandler);
		it++;
	}
	std::list<MatroskaReadHandler *>::const_iterator listIt;
	MatroskaReadHandler *readHandler;
	while(!tmpList.empty()) {
		readHandler = tmpList.front();
		readHandler->Release();
		tmpList.pop_front();
	}
}

void MatroskaReadNode::RegisterReadStream(MatroskaReadStream *readStream) {
	std::list<MatroskaReadStream *>::const_iterator it;
	for(it=m_RegisteredReadStreams.begin() ; it!=m_RegisteredReadStreams.end() ; it++) {
		if((*it) == readStream)
			return;
	}
	m_RegisteredReadStreams.push_back(readStream);
}

void MatroskaReadNode::UnregisterReadStream(MatroskaReadStream *readStream) {
	m_RegisteredReadStreams.remove(readStream);
}

void MatroskaReadNode::ReinitReadStreams(void) {
	std::list<MatroskaReadStream *>::const_iterator it;
	for(it=m_RegisteredReadStreams.begin() ; it!=m_RegisteredReadStreams.end() ; it++)
		(*it)->Reinit();
}

//////////////////////////////////////////////////////////////////////

MatroskaReadHandler::MatroskaReadHandler(const wchar_t *pszFile)
{
	m_RefCount = 1;
	m_FileSize = 0;
	m_File = NULL;
	m_EbmlStream = NULL;
	m_File = new WinIOCallback(pszFile, MODE_READ, FILE_FLAG_SEQUENTIAL_SCAN);
	if(m_File)
		m_EbmlStream = new EbmlStream(*m_File);
	ElementLevel0 = NULL;
	ElementLevel1 = NULL;
	ElementLevel2 = NULL;
	ElementLevel3 = NULL;
	ElementLevel4 = NULL;
	SegmentInfo = NULL;
	SegmentTags = NULL;
	SegmentChapters = NULL;
	SegmentCluster = NULL;
	m_NbClusters = 0;
	m_TimecodeScale = 1000000;

	if (!g_disklockinited) {
		g_disklockinited=true;
		InitializeCriticalSection(&g_diskcs);
	}

	try {
		//_crtBreakAlloc = 1776;
		_construct(pszFile);
	} catch(...) {
		_destruct();
		throw;
	}
}

MatroskaReadHandler::~MatroskaReadHandler(void) {
	_destruct();
}

void MatroskaReadHandler::AddRef(void) {
	m_RefCount++;
}

void MatroskaReadHandler::Release(void) {
	if(!--m_RefCount)
		delete this;
}

VDStringW MatroskaReadHandler::GetFileName(void) {
	return m_FileName;
}

void MatroskaReadHandler::_destruct(void) {
	m_StreamList.DeleteContents(true);
	m_StreamList.Clear();
	if(SegmentInfo)
		delete SegmentInfo;
	if(SegmentTags)
		delete SegmentTags;
	if(SegmentChapters)
		delete SegmentChapters;
	if(m_EbmlStream)
		delete m_EbmlStream;
	if(m_File)
		delete m_File;
}

void MatroskaReadHandler::_setTimecodeScale(void) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		curStream->m_TrackEntry->SetGlobalTimecodeScale(m_TimecodeScale);
		curStream->m_TimecodeScale = m_TimecodeScale;
		m_StreamIterator++;
	}
}

void MatroskaReadHandler::_updateStreamStates(void) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		(*m_StreamIterator)->UpdateState(true);
		m_StreamIterator++;
	}
}

void MatroskaReadHandler::_construct(const wchar_t *pszFile) {
	m_CurrentStream = NULL;
	m_BytesRead = 0;
	m_BytesSkipped = 0;

	m_FileName = pszFile;

	if(m_File && m_EbmlStream) {
		_parseFile();

		delete m_EbmlStream;
		m_EbmlStream = NULL;
		delete m_File;
		m_File = NULL;
		// Better performances without the random access flag (or sequential one)
		m_File = new WinIOCallback((wchar_t *)pszFile, MODE_READ/*, FILE_FLAG_RANDOM_ACCESS*/);
		if(m_File)
			m_EbmlStream = new EbmlStream(*m_File);
		if(m_EbmlStream)
			_updateStreamStates();
		else {
			// Arf, couldn't reopen the file ?
			// too bad, just make as if no streams were found :'(
			m_StreamList.DeleteContents(true);
			m_StreamList.Clear();
		}
	}
}

MatroskaStream *MatroskaReadHandler::_addStream(KaxTrackEntry *trackEntry) {
	MatroskaReadNode *newNode = m_StreamList.AddLast();
	MatroskaStreamSegment *newStreamSegment = newNode->m_Segments.AddLast();
	newStreamSegment->m_ReadHandler = this;
	newStreamSegment->m_LastFilePos = m_FileSize;
	newStreamSegment->m_MatroskaStream = new MatroskaStream();
	newStreamSegment->m_MatroskaStream->m_TrackEntry = trackEntry;
	return newStreamSegment->m_MatroskaStream;
}

MatroskaStream *MatroskaReadHandler::_getStream(sint32 serial) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		if(curStream->m_Serial == serial)
			return curStream;
		m_StreamIterator++;
	}
	return NULL;
}

MatroskaStream *MatroskaReadHandler::_getStream(KaxTrackEntry *trackEntry) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		if(curStream->m_TrackEntry == trackEntry)
			return curStream;
		m_StreamIterator++;
	}
	return NULL;
}

void MatroskaReadHandler::_parseFile() {
	m_File->setFilePointer(0, seek_end);
	m_FileSize = m_File->getFilePointer();
	m_File->setFilePointer(0, seek_beginning);

	ProgressDialog pd(NULL, "Matroska File Import Filter", "Parsing Matroska File", long(m_FileSize/VD64(1024)), true);
	pd.setValueFormat("%ldK of %ldK");

	UpperElementLevel = 0;

	ElementLevel0 = m_EbmlStream->FindNextID(EbmlHead::ClassInfos, m_FileSize);
	if(ElementLevel0 != NULL) {
		ElementLevel0->SkipData(*m_EbmlStream, EbmlHead_Context);
		delete ElementLevel0;
	}

	bool bAllowDummy = true;
	
	// Get level 0 element : KaxSegment
	ElementLevel0 = m_EbmlStream->FindNextID(KaxSegment::ClassInfos, /*m_FileSize*/uint64(0xFFFFFFFFFFFFFFFF));
	if(ElementLevel0 != NULL) {
		bool bDeleteLevel0 = true;
		if(EbmlId(*ElementLevel0) == KaxSegment::ClassInfos.GlobalId) {
			// Get level 1 elements
			ElementLevel1 = m_EbmlStream->FindNextElement(ElementLevel0->Generic().Context, UpperElementLevel, ElementLevel0->GetSize(), bAllowDummy);
			while(ElementLevel1) {
				if(UpperElementLevel > 0)	break;
				if(UpperElementLevel < 0)	UpperElementLevel = 0;
				bool bDeleteLevel1 = true;
				if(EbmlId(*ElementLevel1) == KaxTracks::ClassInfos.GlobalId) {
					bDeleteLevel1 = true;
					// Found tracks entries, get level 2 elements
					ElementLevel2=m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), bAllowDummy);
					while(ElementLevel2) {
						if(UpperElementLevel > 0)	break;
						if(UpperElementLevel < 0)	UpperElementLevel = 0;
						bool bDeleteLevel2 = true;
						if(EbmlId(*ElementLevel2) == KaxTrackEntry::ClassInfos.GlobalId) {
							bDeleteLevel2 = false;
							MatroskaStream *mstream = _addStream(static_cast<KaxTrackEntry*>(ElementLevel2));
							mstream->m_TrackEntry->SetGlobalTimecodeScale(m_TimecodeScale);
							mstream->m_TimecodeScale = m_TimecodeScale;
							// Found a track, get level 3 elements
							///////////////////////////////////////
							mstream->m_TrackEntry->Read(*m_EbmlStream, KaxTrackEntry::ClassInfos.Context, UpperElementLevel, ElementLevel3, bAllowDummy);

							EbmlElement *anElement = mstream->m_TrackEntry->FindElt(KaxTrackNumber::ClassInfos);
							if(anElement) {
								KaxTrackNumber& MyTrackNumber = *static_cast<KaxTrackNumber*>(anElement);
								mstream->m_Serial = (sint32)LIBEBML_uint32(MyTrackNumber);
							}
							anElement = mstream->m_TrackEntry->FindElt(KaxTrackUID::ClassInfos);
							if(anElement) {
								KaxTrackUID& MyTrackUID = *static_cast<KaxTrackUID*>(anElement);
								mstream->m_UID = (uint64)LIBEBML_uint64(MyTrackUID);
							}
							anElement = mstream->m_TrackEntry->FindElt(KaxTrackType::ClassInfos);
							if(anElement) {
								KaxTrackType& MyTrackType = *static_cast<KaxTrackType*>(anElement);
								switch(LIBEBML_uint8(MyTrackType)) {
								case track_video:
									mstream->m_StreamType = SOURCE_VIDEO;
									mstream->m_bIsVideo = true;
									break;
								case track_audio:
									mstream->m_StreamType = SOURCE_AUDIO;
									mstream->m_bIsVideo = false;
									//if(!(long)mstream->m_SampleRate)
										mstream->m_SampleRate = VDFraction(8000);
									//if(!mstream->m_Specific.m_Audio.m_NbChannels)
										mstream->m_Specific.m_Audio.m_NbChannels = 1;
									break;
								case track_subtitle:
									mstream->m_StreamType = SOURCE_TEXT;
									mstream->m_bIsVideo = false;
									//if(!(long)mstream->m_SampleRate)
										mstream->m_SampleRate = VDFraction(1000);
									break;
								default:
									mstream->m_bIsVideo = false;
									break;
								}
							}
							anElement = mstream->m_TrackEntry->FindElt(KaxCodecID::ClassInfos);
							if(anElement) {
								mstream->m_CodecID = static_cast<KaxCodecID*>(anElement);
								if(std::string(*(static_cast<EbmlString*>(mstream->m_CodecID))) == "S_ASS")
									*(static_cast<EbmlString*>(mstream->m_CodecID)) = "S_TEXT/ASS";
								else if(std::string(*(static_cast<EbmlString*>(mstream->m_CodecID))) == "S_SSA")
									*(static_cast<EbmlString*>(mstream->m_CodecID)) = "S_TEXT/SSA";
								if(!memcmp(std::string(*(static_cast<EbmlString*>(mstream->m_CodecID))).c_str(), "S_TEXT/", 7))
									mstream->m_bIsText = true;
								if(std::string(*(static_cast<EbmlString*>(mstream->m_CodecID))) == "A_VORBIS") {

									VDASSERT(mstream->m_StreamType == SOURCE_AUDIO);

									mstream->m_StreamType = SOURCE_AUDIO_VORBIS;
								} else if(std::string(*(static_cast<EbmlString*>(mstream->m_CodecID))) == "S_TEXT/SSA") {

									VDASSERT(mstream->m_StreamType == SOURCE_TEXT);

									mstream->m_StreamType = SOURCE_TEXT_SSA;
								} else if(std::string(*(static_cast<EbmlString*>(mstream->m_CodecID))) == "S_TEXT/ASS") {

									VDASSERT(mstream->m_StreamType == SOURCE_TEXT);

									mstream->m_StreamType = SOURCE_TEXT_SSA;
								}
							}
							anElement = mstream->m_TrackEntry->FindElt(KaxCodecPrivate::ClassInfos);
							if(anElement)
								mstream->m_CodecPrivate = static_cast<KaxCodecPrivate*>(anElement);
							anElement = mstream->m_TrackEntry->FindElt(KaxTrackDefaultDuration::ClassInfos);
							if(anElement) {
								KaxTrackDefaultDuration& MyDefaultDuration = *static_cast<KaxTrackDefaultDuration*>(anElement);
								if(mstream->m_DefaultDuration = uint64(MyDefaultDuration))
									mstream->m_SampleRate = VDFraction(1000000000, mstream->m_DefaultDuration);
							}
							anElement = mstream->m_TrackEntry->FindElt(KaxTrackVideo::ClassInfos);
							if(anElement) {
								KaxTrackVideo& MyVideoTrack = *static_cast<KaxTrackVideo*>(anElement);
								anElement = MyVideoTrack.FindElt(KaxVideoPixelHeight::ClassInfos);
								if(anElement) {
									KaxVideoPixelHeight& MyVideoHeight = *static_cast<KaxVideoPixelHeight*>(anElement);
									mstream->m_Specific.m_Video.m_Height = LIBEBML_uint32(MyVideoHeight);
								}
								anElement = MyVideoTrack.FindElt(KaxVideoPixelWidth::ClassInfos);
								if(anElement) {
									KaxVideoPixelWidth& MyVideoWidth = *static_cast<KaxVideoPixelWidth*>(anElement);
									mstream->m_Specific.m_Video.m_Width = LIBEBML_uint32(MyVideoWidth);
								}
								anElement = MyVideoTrack.FindElt(KaxVideoFrameRate::ClassInfos);
								if(anElement && !mstream->m_SampleRate.getHi()) {
									KaxVideoFrameRate& MyVideoRate = *static_cast<KaxVideoFrameRate*>(anElement);
									mstream->m_SampleRate = VDFraction(double(MyVideoRate));
								}
							}
							anElement = mstream->m_TrackEntry->FindElt(KaxTrackAudio::ClassInfos);
							if(anElement) {
								KaxTrackAudio& MyAudioTrack = *static_cast<KaxTrackAudio*>(anElement);
								anElement = MyAudioTrack.FindElt(KaxAudioChannels::ClassInfos);
								if(anElement) {
									KaxAudioChannels& MyChannels = *static_cast<KaxAudioChannels*>(anElement);
									mstream->m_Specific.m_Audio.m_NbChannels = LIBEBML_uint32(MyChannels);
								}
								anElement = MyAudioTrack.FindElt(KaxAudioSamplingFreq::ClassInfos);
								if(anElement) {
									KaxAudioSamplingFreq& MyAudioRate = *static_cast<KaxAudioSamplingFreq*>(anElement);
									mstream->m_SampleRate = VDFraction(double(MyAudioRate));
								}
								anElement = MyAudioTrack.FindElt(KaxAudioBitDepth::ClassInfos);
								if(anElement) {
									KaxAudioBitDepth& MyAudioDepth = *static_cast<KaxAudioBitDepth*>(anElement);
									mstream->m_Specific.m_Audio.m_BitDepth = LIBEBML_uint32(MyAudioDepth);
								}
							}
						}
						if (UpperElementLevel > 0) {
							UpperElementLevel--;
							if(bDeleteLevel2)
								delete ElementLevel2;
							ElementLevel2 = ElementLevel3;
							if(UpperElementLevel > 0)
								break;
						} else {
							ElementLevel2->SkipData(*m_EbmlStream, ElementLevel2->Generic().Context);
							if(bDeleteLevel2)
								delete ElementLevel2;
							ElementLevel2 = m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), bAllowDummy);
						}
					}
				} else if (EbmlId(*ElementLevel1) == KaxInfo::ClassInfos.GlobalId) {
					bDeleteLevel1 = !!SegmentInfo;
					if(!SegmentInfo) {
						SegmentInfo = static_cast<KaxInfo *>(ElementLevel1);
						SegmentInfo->Read(*m_EbmlStream, KaxInfo::ClassInfos.Context, UpperElementLevel, ElementLevel2, bAllowDummy);

						EbmlElement *anElement = SegmentInfo->FindElt(KaxTimecodeScale::ClassInfos);
						if(anElement) {
							KaxTimecodeScale& TimecodeScale = *static_cast<KaxTimecodeScale*>(anElement);
							m_TimecodeScale = uint64(TimecodeScale);
							_setTimecodeScale();
						}
					}
				} else if (EbmlId(*ElementLevel1) == KaxTags::ClassInfos.GlobalId) {
					bDeleteLevel1 = !!SegmentTags;
					if(!SegmentTags) {
						SegmentTags = static_cast<KaxTags *>(ElementLevel1);
						SegmentTags->Read(*m_EbmlStream, KaxTags::ClassInfos.Context, UpperElementLevel, ElementLevel2, bAllowDummy);
					}
				} else if (EbmlId(*ElementLevel1) == KaxChapters::ClassInfos.GlobalId) {
					bDeleteLevel1 = !!SegmentChapters;
					if(!SegmentChapters) {
						SegmentChapters = static_cast<KaxChapters *>(ElementLevel1);
						SegmentChapters->Read(*m_EbmlStream, KaxChapters::ClassInfos.Context, UpperElementLevel, ElementLevel2, bAllowDummy);
					}
				} else if (EbmlId(*ElementLevel1) == KaxCluster::ClassInfos.GlobalId) {
					// Found a cluster, get level 2 elements
					bDeleteLevel1 = true;
					SegmentCluster = static_cast<KaxCluster *>(ElementLevel1);
					_addCluster(SegmentCluster);
					uint64 ClusterTimecode;
					ElementLevel2=m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), bAllowDummy);
					while(ElementLevel2) {
						if(UpperElementLevel > 0)	break;
						if(UpperElementLevel < 0)	UpperElementLevel = 0;
						bool bDeleteLevel2 = true;
						if (EbmlId(*ElementLevel2) == KaxClusterTimecode::ClassInfos.GlobalId) {
							KaxClusterTimecode & ClusterTime = *static_cast<KaxClusterTimecode*>(ElementLevel2);
							ClusterTime.ReadData(m_EbmlStream->I_O());
							ClusterTimecode = uint64(ClusterTime);
							SegmentCluster->InitTimecode(ClusterTimecode, m_TimecodeScale);
						} else  if (EbmlId(*ElementLevel2) == KaxBlockGroup::ClassInfos.GlobalId) {
							pd.advance(long(m_File->getFilePointer()/VD64(1024)));
							pd.check();
							KaxBlockGroup & DataBlockGroup = *static_cast<KaxBlockGroup*>(ElementLevel2);
							// Found blocks, get level 3 elements
							uint64 blockTime = ClusterTimecode;
							uint64 blockDuration = 0;
							VDStringW subtitle;
							bool bTime = false;
							bool bDuration = false;
							bool bBRef = false;
							bool bFRef = false;
							bool bKey = true;
							bool bB = false;
							MatroskaStream *mstream = NULL;
							ElementLevel3=m_EbmlStream->FindNextElement(ElementLevel2->Generic().Context, UpperElementLevel, ElementLevel2->GetSize(), bAllowDummy);
							while(ElementLevel3) {
								if(UpperElementLevel > 0)	break;
								if(UpperElementLevel < 0)	UpperElementLevel = 0;
								bool bDeleteLevel3 = true;
								if (EbmlId(*ElementLevel3) == KaxBlockDuration::ClassInfos.GlobalId) {
									KaxBlockDuration & BlockDuration = *static_cast<KaxBlockDuration*>(ElementLevel3);
									BlockDuration.ReadData(m_EbmlStream->I_O());
									blockDuration = uint64(*(static_cast<EbmlUInteger *>(&BlockDuration)));
									bDuration = true;
								} else if (EbmlId(*ElementLevel3) == KaxBlock::ClassInfos.GlobalId) {
									KaxBlock & DataBlock = *static_cast<KaxBlock*>(ElementLevel3);
#if MATROSKA_PARSE_PARTIAL
									sint64 iSavedPos = m_EbmlStream->I_O().getFilePointer();
									DataBlock.ReadData(m_EbmlStream->I_O(), SCOPE_PARTIAL_DATA);
#else
									DataBlock.ReadData(m_EbmlStream->I_O(), SCOPE_ALL_DATA);
#endif
									DataBlock.SetParent(*SegmentCluster);
									mstream = _getStream(DataBlock.TrackNum());
									if(mstream) {
										blockTime /*+*/= DataBlock.GlobalTimecode() / mstream->m_TimecodeScale;
										if(mstream->m_bIsText) {

											VDASSERT(DataBlock.NumberFrames() == 1);

#if MATROSKA_PARSE_PARTIAL
											m_EbmlStream->I_O().setFilePointer(iSavedPos);
											DataBlock.ReadData(m_EbmlStream->I_O(), SCOPE_ALL_DATA);
#endif

											subtitle = mstream->m_TextUtil.GetText(DataBlock.GetBuffer(0).Buffer(), DataBlock.GetBuffer(0).Size());
										}
										_addBlock(mstream, DataBlock);
										bTime = true;
									}
#if MATROSKA_VERSION >= 2
								} else if (EbmlId(*ElementLevel3) == KaxBlockVirtual::ClassInfos.GlobalId) {
									throw MyError("Virtual block not handled (position: %I64u)", ElementLevel3->GetElementPosition());
									printf(" Virtual Block\n");
								//} else if (EbmlId(*ElementLevel3) == KaxReferencePriority::ClassInfos.GlobalId) {
#endif
								} else if (EbmlId(*ElementLevel3) == KaxReferenceBlock::ClassInfos.GlobalId) {
									KaxReferenceBlock& MyRef = *static_cast<KaxReferenceBlock*>(ElementLevel3);
									MyRef.ReadData(m_EbmlStream->I_O());
									int64 ref = int64(*(static_cast<EbmlSInteger *>(&MyRef)));
									if(ref < 0) {

										VDASSERT(!bBRef);
										if(bBRef)
											throw MyError("Found a frame with many backward refs, unhandled!");

										bBRef = true;
										bKey = false;
									} else if(ref > 0) {

										VDASSERT(!bFRef);
										if(bFRef)
											throw MyError("Found a frame with many forward refs, unhandled!");

										bFRef = true;
										bKey = false;
										bB = true;
									}
#if MATROSKA_VERSION >= 2
								} else if (EbmlId(*ElementLevel3) == KaxReferenceVirtual::ClassInfos.GlobalId) {
									throw MyError("Virtual reference not handled");
#endif
								}
								if (UpperElementLevel > 0) {
									_ASSERT(0 == 1);
									UpperElementLevel--;
									if(bDeleteLevel3)
										delete ElementLevel3;
									ElementLevel3 = ElementLevel4;
									if(UpperElementLevel > 0)
										break;
								} else {
									ElementLevel3->SkipData(*m_EbmlStream, ElementLevel3->Generic().Context);
									if(bDeleteLevel3)
										delete ElementLevel3;
									ElementLevel3 = m_EbmlStream->FindNextElement(ElementLevel2->Generic().Context, UpperElementLevel, ElementLevel2->GetSize(), bAllowDummy);
								}
							}
							if(!mstream)
								VDDEBUG("Found data for a track without track entry");
							else if(!bTime)
								VDLogAppMessage(kVDLogWarning, kVDST_MatroskaReadHandler, kVDM_BlockWithoutTimecode, 1, &mstream->m_Serial);
							else {
								if(bFRef && !bBRef)
									VDLogAppMessage(kVDLogWarning, kVDST_MatroskaReadHandler, kVDM_OnlyForwardRef, 2, &mstream->m_Serial, &blockTime);
								mstream->m_EstimatedSamplesInLastBlock = blockDuration;
								_setBlockInfo(mstream, blockTime, bKey, bB);
								if(mstream->m_bIsText) {
									if(!bDuration)
										VDLogAppMessage(kVDLogWarning, kVDST_MatroskaReadHandler, kVDM_SubtitleWithoutDuration, 1, &mstream->m_Serial);
										//throw MyError("No duration found for a subtitle");
									else if(!mstream->m_SubtitleProvider.AddSubtitle(
										subtitle
										, (blockTime * mstream->m_TimecodeScale) / 1000000
										, (blockDuration * mstream->m_TimecodeScale) / 1000000))
										VDDEBUG("One subtitle coming from stream %d wasn't kept.\n"
										, mstream->m_Serial);
								}
							}
						}
						if (UpperElementLevel > 0) {
							UpperElementLevel--;
							if(bDeleteLevel2)
								delete ElementLevel2;
							ElementLevel2 = ElementLevel3;
							if(UpperElementLevel > 0)
								break;
						} else {
							ElementLevel2->SkipData(*m_EbmlStream, ElementLevel2->Generic().Context);
							if(bDeleteLevel2)
								delete ElementLevel2;
							ElementLevel2 = m_EbmlStream->FindNextElement(ElementLevel1->Generic().Context, UpperElementLevel, ElementLevel1->GetSize(), bAllowDummy);
						}
					}
				}
				if (UpperElementLevel > 0) {
					UpperElementLevel--;
					if(bDeleteLevel1)
						delete ElementLevel1;
					ElementLevel1 = ElementLevel2;
					if(UpperElementLevel > 0)
						break;
				} else {
					ElementLevel1->SkipData(*m_EbmlStream, ElementLevel1->Generic().Context);
					if(bDeleteLevel1)
						delete ElementLevel1;
					ElementLevel1 = m_EbmlStream->FindNextElement(ElementLevel0->Generic().Context, UpperElementLevel, ElementLevel0->GetSize(), bAllowDummy);
				}
			}
		}
		if(bDeleteLevel0)
			delete ElementLevel0;
	}

	// Remove last added Cluster if empty
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		MatroskaCluster *lastCluster = curStream->m_ClusterList.GetLast();
		if(lastCluster && !lastCluster->m_NbBlocks) {
			curStream->m_ClusterList.RemoveLast();
			curStream->m_CurrentCluster = curStream->m_ClusterList.GetLast();
		} else if(lastCluster && (lastCluster->m_NbBlocks % MATROSKA_REALLOC_BLOCKGROUP_SIZE) )
			_trimClusterBlockAlloc(lastCluster);

		// Order frames in display order
		if(curStream->m_bIsVideo) {
			ModCompactList<MatroskaCluster>::iterator clusterIt;
			long nbBlocks = 0;
			int nbRun = 0;
			while(true) {
				nbRun++;
				bool bReordered = false;
				clusterIt.begin(curStream->m_ClusterList);
				//MatroskaBlock tmpBlock;
				MatroskaBlock *curBlock = NULL;
				MatroskaBlock *nextBlock = NULL;
				uint64 gpos;
				if(clusterIt) {
					if(nbBlocks = (*clusterIt)->m_NbBlocks)
						curBlock = &(*clusterIt)->m_Blocks[0];
				}
				while(curBlock) {
					nextBlock = NULL;
					if(--nbBlocks)
						nextBlock = curBlock + 1;
					else if(clusterIt.next()) {
						clusterIt++;
						if(nbBlocks = (*clusterIt)->m_NbBlocks)
							nextBlock = &(*clusterIt)->m_Blocks[0];
					}
					if(nextBlock
					&& ( MKV_GRANULEPOS(nextBlock->m_Granulepos) < MKV_GRANULEPOS(curBlock->m_Granulepos))) {
						bReordered = true;
						gpos = MKV_GRANULEPOS(curBlock->m_Granulepos);
						curBlock->m_Granulepos &= MKV_FLAGS;
						curBlock->m_Granulepos |= MKV_GRANULEPOS(nextBlock->m_Granulepos);
						nextBlock->m_Granulepos &= MKV_FLAGS;
						nextBlock->m_Granulepos |= gpos;
						/*tmpBlock = *curBlock;
						*curBlock = *nextBlock;
						*nextBlock = tmpBlock;*/
					}
					curBlock = nextBlock;
				}
				/*tmpBlock.m_NbFrames = 0;
				tmpBlock.m_NbBytes = NULL;*/
				if(!bReordered)
					break;
				if(nbRun == 2)
					VDLogAppMessage(kVDLogWarning, kVDST_MatroskaReadHandler, kVDM_WeirdFrameOrder);
			}
		}

		// Search for overlapping subtitles
		if(curStream->m_bIsText && curStream->m_SubtitleProvider.HasCollisions())
			VDLogAppMessage(kVDLogWarning, kVDST_MatroskaReadHandler, kVDM_OverlappingSubtitles, 1, &curStream->m_Serial);

		m_StreamIterator++;
	}
}

IAVIReadStream *MatroskaReadHandler::GetStream(DWORD fccType, LONG lParam) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		switch(MAIN_TYPE(curStream->m_StreamType)) {
		case SOURCE_VIDEO:
			if( (fccType==streamtypeVIDEO) && (!lParam--) )
				return new MatroskaReadStream(*m_StreamIterator);
			break;
		case SOURCE_AUDIO:
			if( (fccType==streamtypeAUDIO) && (!lParam--) )
				return new MatroskaReadStream(*m_StreamIterator);
			break;
		case SOURCE_TEXT:
			if( (fccType==streamtypeAUDIO) && (!lParam--) )
				return new MatroskaReadStream(*m_StreamIterator);
			break;
		}
		m_StreamIterator++;
	}
	return NULL;
}

IAVIReadStream *MatroskaReadHandler::GetStream(LONG lParam) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		if(!lParam--)
			return new MatroskaReadStream(*m_StreamIterator);
		m_StreamIterator++;
	}
	return NULL;
}

long MatroskaReadHandler::GetStreamIndexByType(MatroskaReadNode *stream) {
	
	VDASSERT(stream);
	
	long index = 0;
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaReadNode *curStream = *m_StreamIterator;
		if(curStream == stream)
			return index;
		if(MAIN_TYPE(curStream->m_Segments.GetFirst()->m_MatroskaStream->m_StreamType) == MAIN_TYPE(stream->m_Segments.GetFirst()->m_MatroskaStream->m_StreamType) )
			index++;
		m_StreamIterator++;
	}
	return -1;
}

ModList<MatroskaReadNode>& MatroskaReadHandler::GetStreamList(void) {
	return m_StreamList;
}

void MatroskaReadHandler::EnableFastIO(bool bEnable) {
}

bool MatroskaReadHandler::isOptimizedForRealtime(void) {
	return false;
}

bool MatroskaReadHandler::isStreaming(void) {
	return false;
}

bool MatroskaReadHandler::isIndexFabricated(void) {
	return false;
}

bool MatroskaReadHandler::AppendFile(const wchar_t *pszFile) {
	if(!pszFile)
		return false;
	MatroskaReadHandler *nextSegment = new MatroskaReadHandler(pszFile);
	if(!nextSegment)
		throw MyMemoryError();

	if(m_StreamList.GetCount() != nextSegment->GetStreamList().GetCount())
		throw MyError("Cannot append segment: The segment has a different number of streams.");

	ModList<MatroskaReadNode>::iterator firstSegmentIt(m_StreamList);
	ModList<MatroskaReadNode>::iterator nextSegmentIt(nextSegment->GetStreamList());
	while(firstSegmentIt) {
		MatroskaStream *firstStreamSegment = (*firstSegmentIt)->m_Segments.GetFirst()->m_MatroskaStream;
		MatroskaStream *nextStreamSegment = (*nextSegmentIt)->m_Segments.GetFirst()->m_MatroskaStream;

		if(firstStreamSegment->m_StreamType != nextStreamSegment->m_StreamType)
			throw MyError("Cannot append segment: The stream types do not match.");

		/*if(firstStreamSegment->m_StreamHeader->size != nextStreamSegment->m_StreamHeader->size)
			throw MyError("Cannot append segment: The header sizes do not match (%d vs %d)."
				, firstStreamSegment->m_StreamHeader->size
				, nextStreamSegment->m_StreamHeader->size);*/

		if(firstStreamSegment->m_SampleRate != nextStreamSegment->m_SampleRate)
			throw MyError("Cannot append segment: The samplerates do not match (%.5f vs %.5f)."
				, (double)firstStreamSegment->m_SampleRate
				, (double)nextStreamSegment->m_SampleRate);

		switch(MAIN_TYPE(firstStreamSegment->m_StreamType)) {
		case SOURCE_VIDEO:
			if( (firstStreamSegment->m_Specific.m_Video.m_Width!=nextStreamSegment->m_Specific.m_Video.m_Width)
			|| (firstStreamSegment->m_Specific.m_Video.m_Height!=nextStreamSegment->m_Specific.m_Video.m_Height) )
				throw MyError("Cannot append segment: The video size do not match (%ldx%ld vs %ldx%ld)."
					, firstStreamSegment->m_Specific.m_Video.m_Width, firstStreamSegment->m_Specific.m_Video.m_Height
					, nextStreamSegment->m_Specific.m_Video.m_Width, nextStreamSegment->m_Specific.m_Video.m_Height);
			break;
		case SOURCE_AUDIO:
			if(firstStreamSegment->m_Specific.m_Audio.m_NbChannels != nextStreamSegment->m_Specific.m_Audio.m_NbChannels)
				throw MyError("Cannot append segment: The number of channels do not match (%d vs %d)."
					, firstStreamSegment->m_Specific.m_Audio.m_NbChannels, nextStreamSegment->m_Specific.m_Audio.m_NbChannels);
			if(firstStreamSegment->m_Specific.m_Audio.m_BitDepth != nextStreamSegment->m_Specific.m_Audio.m_BitDepth)
				throw MyError("Cannot append segment: The bit depth do not match (%d vs %d)."
					, firstStreamSegment->m_Specific.m_Audio.m_BitDepth, nextStreamSegment->m_Specific.m_Audio.m_BitDepth);
			break;
		}

		firstSegmentIt++;
		nextSegmentIt++;
	}

	// Merging accepted, notify the read handlers
	firstSegmentIt.begin(m_StreamList);
	nextSegmentIt.begin(nextSegment->GetStreamList());
	while(firstSegmentIt) {
		ModList<MatroskaStreamSegment>::iterator segmentIt((*nextSegmentIt)->m_Segments);
		while(segmentIt) {
			(*firstSegmentIt)->AppendSegment(*segmentIt);
			segmentIt++;
		}
		(*firstSegmentIt)->ReinitReadStreams();
		firstSegmentIt++;
		nextSegmentIt++;
	}

	nextSegment->Release();
	return true;
}

bool MatroskaReadHandler::getSegmentHint(const char **ppszPath) {
	if(ppszPath)
		*ppszPath = NULL;

	return false;
}

void MatroskaReadHandler::GetCommentList(ModList<ModStreamComment>& list) {
}

void MatroskaReadHandler::GetChapterList(ModChapterList& list) {
}

VDStringW MatroskaReadHandler::writeReport(void) {
	VDStringW report;
	report = VDStringW::setf(L"---== Matroska File Import filter report ==---\r\nSource file : [%s]", m_FileName.c_str());
	//report += L"\r\nNo apparent corrupted data";
	MatroskaReadNode *readNode = m_StreamList.GetFirst();
	if(readNode && (readNode->m_Segments.GetCount()>1)) {
		ModList<MatroskaStreamSegment>::iterator it(readNode->m_Segments);
		it++;
		while(it) {
			report += L"\r\n";
			report += (*it)->m_ReadHandler->writeReport();
			it++;
		}
	}
	return report;
}

void MatroskaReadHandler::_addCluster(KaxCluster *cluster) {

	VDASSERT(cluster);

	m_NbClusters++;

	MatroskaStream *stream = NULL;
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		// Only create a new Cluster if the last one was used
		if(!curStream->m_CurrentCluster || curStream->m_CurrentCluster->m_NbBlocks) {
			if(curStream->m_CurrentCluster && (curStream->m_CurrentCluster->m_NbBlocks % MATROSKA_REALLOC_BLOCKGROUP_SIZE))
				_trimClusterBlockAlloc(curStream->m_CurrentCluster);
			curStream->m_CurrentCluster = curStream->m_ClusterList.AddLast();
			curStream->m_NbClusters++;
		} else
			VDDEBUG("Cluster %I64d doesn't contain data for stream %ld\n"
				, m_NbClusters, curStream->m_Serial);
		curStream->m_CurrentCluster->m_ClusterPosition = cluster->GetElementPosition();
		m_StreamIterator++;
	}
}

void MatroskaReadHandler::_trimClusterBlockAlloc(MatroskaCluster *cluster) {
	cluster->m_Blocks = (MatroskaBlock *)_expand(cluster->m_Blocks, cluster->m_NbBlocks * sizeof(MatroskaBlock));
}

void MatroskaReadHandler::_extendClusterBlockAlloc(MatroskaCluster *cluster) {
	int nbRealloc = cluster->m_NbBlocks + MATROSKA_REALLOC_BLOCKGROUP_SIZE;
	cluster->m_Blocks = (MatroskaBlock *)realloc(cluster->m_Blocks, nbRealloc * sizeof(MatroskaBlock));
	memset(&cluster->m_Blocks[cluster->m_NbBlocks], 0, MATROSKA_REALLOC_BLOCKGROUP_SIZE * sizeof(MatroskaBlock));
}

void MatroskaReadHandler::_addBlock(MatroskaStream *stream, KaxBlock& block) {
	if(!stream)
		return;

	stream->m_NbBlocks++;
	unsigned int nbFrames = block.NumberFrames();
	stream->m_NbFrames += nbFrames;

	if(!(stream->m_CurrentCluster->m_NbBlocks % MATROSKA_REALLOC_BLOCKGROUP_SIZE)) {
		_extendClusterBlockAlloc(stream->m_CurrentCluster);
		stream->m_CurrentBlock = &stream->m_CurrentCluster->m_Blocks[stream->m_CurrentCluster->m_NbBlocks];
	} else
		stream->m_CurrentBlock++;

	stream->m_CurrentCluster->m_NbBlocks++;
	stream->m_CurrentBlock->m_BlockPosition = block.GetElementPosition();

	stream->m_CurrentBlock->m_NbFrames = nbFrames;
	stream->m_CurrentBlock->m_NbBytes = (uint32 *)malloc(stream->m_CurrentBlock->m_NbFrames * sizeof(uint32));
	for(unsigned int i=0; i< nbFrames; i++) {
		/*uint32 size = block.GetBuffer(i).Size();

		VDASSERT(size == block.GetFrameSize(i));*/

		uint32 size = block.GetFrameSize(i);
		stream->m_CurrentBlock->m_NbBytes[i] = size;
		stream->m_Bytes += size;
		if(size > stream->m_MaxFrameSize)
			stream->m_MaxFrameSize = size;
	}
	//stream->m_TotalBlockSize += size;
}

void MatroskaReadHandler::_setBlockInfo(MatroskaStream *stream, uint64 timeCode, bool bKey, bool bB) {

	VDASSERT(stream);
	VDASSERT(stream->m_CurrentBlock);
	VDASSERT(!stream->m_bIsVideo || (stream->m_CurrentBlock->m_NbFrames==1));

	//if(stream->m_CurrentBlock->m_NbFrames > 1)
	//	return;

	// for the moment, just store the time (we will compute granulepos later)
	stream->m_CurrentBlock->m_Granulepos = timeCode;
	if(stream->m_bIsVideo) {
		if(bKey)
			stream->m_CurrentBlock->m_Granulepos |= MKV_FLAG_KEY;
		if(bB)
			stream->m_CurrentBlock->m_Granulepos |= MKV_FLAG_B;
	}
}

long MatroskaReadHandler::ReadData(void *buffer, long nBytes) {
	EnterCriticalSection(&g_diskcs);
	long nread = m_File->read(buffer, nBytes);
	LeaveCriticalSection(&g_diskcs);
	return nread;
}

void MatroskaReadHandler::SeekFile(__int64 filePos) {
	m_File->setFilePointer(filePos);
}

long MatroskaReadHandler::GetSegmentDuration(int iSegment) {
	if(iSegment < 0)
		return -1;

	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		MatroskaStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_MatroskaStream;
		if(curStream->m_StreamType == SOURCE_VIDEO) {
			MatroskaStreamSegment *segment = (*m_StreamIterator)->m_Segments[iSegment];
			if(segment && segment->m_MatroskaStream->m_SampleRate.getHi())
				return segment->m_MatroskaStream->m_SampleRate.scale64ir(VD64(1000)*(segment->m_End-segment->m_Start));
			else
				return -1;
		}
		m_StreamIterator++;
	}
	return -1;
}

KaxTags *MatroskaReadHandler::GetTags(void) {
	return SegmentTags;
}

KaxChapters *MatroskaReadHandler::GetChapters(void) {
	return SegmentChapters;
}