//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MATROSKAREADHANDLER_H__
#define __MATROSKAREADHANDLER_H__

#include <vd2/system/VDString.h>
#include <vd2/system/File64.h>
#include <vd2/system/Fraction.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vector>
#include "ModClassInterfaces.h"
#include "AVIReadHandler.h"
#include "ModList.h"
#include "ModCompactList.h"
#include "ModTextUtil.h"
#include "ModTextSourceSRT.h"
#include <list>

// There are conflicting typedefs between VirtualDub and libebml
// Fool the compiler and make libebml/libmatroska typedef to
// LIBEBML_int8/16/32/64 instead of int8/16/32/64 that are already
// defined by VirtualDub
#define	uint8	LIBEBML_uint8
#define	int8	LIBEBML_int8
#define	uint16	LIBEBML_uint16
#define	int16	LIBEBML_int16
#define	uint32	LIBEBML_uint32
#define	int32	LIBEBML_int32
#define	uint64	LIBEBML_uint64
#define	int64	LIBEBML_int64

#include "ebml/EbmlHead.h"
#include "ebml/EbmlSubHead.h"
#include "ebml/EbmlStream.h"
#include "ebml/EbmlContexts.h"
#include "matroska/FileKax.h"
#include "matroska/KaxSegment.h"
#include "matroska/KaxContexts.h"
#include "matroska/KaxTracks.h"
#include "matroska/KaxTrackVideo.h"
#include "matroska/KaxTrackAudio.h"
#include "matroska/KaxInfo.h"
#include "matroska/KaxInfoData.h"
#include "matroska/KaxCluster.h"
#include "matroska/KaxClusterData.h"
#include "matroska/KaxBlock.h"
#include "matroska/KaxBlockData.h"
#include "matroska/KaxTags.h"
#include "matroska/KaxTag.h"
#include "matroska/KaxTagMulti.h"
#include "matroska/KaxChapters.h"
#include "src/platform/win32/WinIOCallback.h"
#include "matroska/KaxTags.h"

#undef uint8
#undef int8
#undef uint16
#undef int16
#undef uint32
#undef int32
#undef uint64
#undef int64

using namespace LIBMATROSKA_NAMESPACE;

#define MATROSKA_PRAGMA_PACK

#ifdef MATROSKA_PRAGMA_PACK
#pragma pack(push, 2)
#endif //MATROSKA_PRAGMA_PACK

namespace {
	enum { kVDST_MatroskaReadHandler = 151 };
	
	enum {
		kVDM_OverlappingSubtitles,
		kVDM_WeirdFrameOrder,
		kVDM_Seek_BadGranulepos,
		kVDM_Seek_BadBlock,
		kVDM_Seek_BadCluster,
		kVDM_LibError,
		kVDM_SubtitleWithoutDuration,
		kVDM_BlockWithoutTimecode,
		kVDM_OnlyForwardRef,
		kVDM_MaybeVFR
	};
}

#define MKV_FLAGS			VD64(0xC000000000000000)
#define MKV_GETFLAGS(x)		((x)&MKV_FLAGS)
#define MKV_GRANULEPOS(x)	((x)&VD64(0x3FFFFFFFFFFFFFFF))
#define MKV_FLAG_KEY		VD64(0x8000000000000000)
#define MKV_GETFLAG_KEY(x)	((x)&MKV_FLAG_KEY)
#define MKV_FLAG_B			VD64(0x4000000000000000)
#define MKV_GETFLAG_B(x)	((x)&MKV_FLAG_B)
#define MKV_BLOCKPOS(x)		((x)&VD64(0x00FFFFFFFFFFFFFF))
#define MKV_CLUSTERPOS(x)	((x)&VD64(0xFFFFFFFFFFFFFFFF))

class MatroskaBlock {
public:
	// Bit 63 of m_Granulepos will tell if this is a Keyframe
	uint64	m_Granulepos;
	// Bit 57-63 of m_BlockPosition contains the number of frames
	// Not a problem with matroska which handle only up to 2^56 in EBML currently
	union {
		uint64	m_BlockPosition;
		struct {
			uint8	m_Junk0;
			uint8	m_Junk1;
			uint8	m_Junk2;
			uint8	m_Junk3;
			uint8	m_Junk4;
			uint8	m_Junk5;
			uint8	m_Junk6;
			uint8	m_NbFrames;
		};
	};
	uint32 *m_NbBytes;

	MatroskaBlock(void) {
		m_Granulepos	= 0;
		//m_NbFrames		= 0;
		m_BlockPosition	= 0;
		m_NbBytes		= NULL;
	};

	~MatroskaBlock(void) {
		if(m_NbBytes)
			delete[] m_NbBytes;
	};
};


class MatroskaCluster {
public:
	uint64			m_ClusterPosition;
	MatroskaBlock	*m_Blocks;
	uint16			m_NbBlocks;

	MatroskaCluster(void) {
		m_ClusterPosition = 0;
		m_NbBlocks	= 0;
		m_Blocks	= NULL;
	};

	~MatroskaCluster(void) {
		if(m_Blocks) {
			for(int i=0 ; i<m_NbBlocks ; i++)
				if(m_Blocks[i].m_NbBytes)
					delete[] m_Blocks[i].m_NbBytes;
			free(m_Blocks);
		}
	};
};
#ifdef MATROSKA_PRAGMA_PACK
#pragma pack(pop)
#endif //MATROSKA_PRAGMA_PACK

class MatroskaStream {
public:
	sint32			m_RefCount;
	sint32			m_Serial;
	uint64			m_UID;
	uint32			m_StreamType;
	std::vector<ogg_packet *> m_OggHeaderPacketList;
	bool			m_bIsVideo;
	uint64			m_TimecodeScale;
	uint64			m_DefaultDuration;
	VDFraction		m_SampleRate;
	uint32			m_EstimatedSamplesInLastBlock;
	ModCompactList<MatroskaCluster>	m_ClusterList;
	ModCompactList<MatroskaCluster>::iterator	m_ClusterIterator;
	MatroskaCluster	*m_CurrentCluster;
	MatroskaBlock	*m_CurrentBlock;
	bool			m_bGranuleposComputed;
	KaxTrackEntry	*m_TrackEntry;
	KaxCodecID		*m_CodecID;
	KaxCodecPrivate	*m_CodecPrivate;
	union {
		struct {
			sint32	m_Width;
			sint32	m_Height;
		} m_Video;
		struct {
			uint32	m_NbChannels;
			uint32	m_BitDepth;
		} m_Audio;
	} m_Specific;
	uint32			m_NbClusters;
	uint32			m_NbClustersCorrupted;
	uint32			m_NbBlocks;
	uint32			m_NbFrames;
	uint32			m_MaxFrameSize;
	uint64			m_StartGranulepos;
	sint64			m_StartOffsetMS;
	sint64			m_Bytes;
	sint64			m_LastGranulepos;
	sint64			m_NbSamples;
	bool			m_bIndexBuilt;
	std::vector<uint8>	m_KeyIndex;
	std::vector<uint8>	m_BIndex;
	bool			m_bIsText;
	ModTextUtil			m_TextUtil;
	ModSubtitleProvider	m_SubtitleProvider;

	MatroskaStream(void);
	virtual ~MatroskaStream(void);
	void AddOggHeaderPacket(binary *data, uint32 dataSize);
	void AddRef(void);
	void Release(void);
	VDStringW writeReport(void);
	//void GetCommentList(ModList<ModStreamComment>& list);
	//void GetChapterList(ModChapterList& list, uint32 offsetMS);
};

class MatroskaReadHandler;
class MatroskaReadStream;

class MatroskaStreamSegment : public ModListNode<MatroskaStreamSegment> {
public:
	MatroskaReadHandler	*m_ReadHandler;
	MatroskaStream		*m_MatroskaStream;
	sint64				m_Start;
	sint64				m_End;
	sint64				m_LastFilePos;

	MatroskaStreamSegment(void);
	~MatroskaStreamSegment(void);
	void UpdateState(bool bVBRHandling=true);
	void GetCommentList(ModList<ModStreamComment>& list);
	void GetChapterList(ModChapterList& list, uint32 offsetMS);
};

class MatroskaReadNode : public ModListNode<MatroskaReadNode> {
public:
	sint64	m_Start;
	sint64	m_End;
	sint64	m_LastFilePos;
	sint64	m_NbSamples;
	sint64	m_Bytes;
	sint64	m_FirstKeyFrame;
	uint32	m_MaxFrameSize;
	sint16	m_BlockAlign;
	bool	m_bKeyFrameOnly;
	bool	m_bHasBFrames;
	ModList<MatroskaStreamSegment>	m_Segments;
	std::list<MatroskaReadStream *>	m_RegisteredReadStreams;

	MatroskaReadNode(void);
	~MatroskaReadNode(void);
	void AppendSegment(MatroskaStreamSegment *segment);
	void UpdateState(bool bVBRHandling=true);
	void AddRef(void);
	void Release(void);
	void RegisterReadStream(MatroskaReadStream *readStream);
	void UnregisterReadStream(MatroskaReadStream *readStream);
	void ReinitReadStreams(void);
};


class MatroskaReadHandler : public IAVIReadHandler {
friend class MatroskaReadStream;
protected:
	sint32		m_RefCount;
	WinIOCallback	*m_File;
	VDStringW	m_FileName;
	sint64		m_FileSize;
	//VDStringW	m_Title;
	ModList<MatroskaReadNode>	m_StreamList;
	ModList<MatroskaReadNode>::iterator	m_StreamIterator;
	MatroskaStream	*m_CurrentStream;
	uint64		m_TimecodeScale;
	sint64		m_NbClusters;
	sint64		m_BytesRead;
	sint64		m_BytesSkipped;

	EbmlStream	*m_EbmlStream;
	EbmlElement	*ElementLevel0;
	EbmlElement	*ElementLevel1;
	EbmlElement	*ElementLevel2;
	EbmlElement	*ElementLevel3;
	EbmlElement	*ElementLevel4;
	int			UpperElementLevel;
	KaxInfo		*SegmentInfo;
	KaxTags		*SegmentTags;
	KaxChapters	*SegmentChapters;
	KaxCluster	*SegmentCluster;
	MatroskaCluster	*lastCluster;

	void		_construct(const wchar_t *pszFile);
	void		_setTimecodeScale(void);
	void		_updateStreamStates(void);
	void		_parseFile(void);
	void		_destruct(void);
	MatroskaStream *_getStream(sint32 serial);
	MatroskaStream *_getStream(KaxTrackEntry *aTrack);
	MatroskaStream *_addStream(KaxTrackEntry *aTrack);
	void    _addCluster(KaxCluster *cluster);
	void	_extendClusterBlockAlloc(MatroskaCluster *cluster);
	void	_trimClusterBlockAlloc(MatroskaCluster *cluster);
	void    _addBlock(MatroskaStream *stream/*, uint32 size*/, KaxBlock& block/*, uint64 blockGroupPos*/);
	void    _setBlockInfo(MatroskaStream *stream, uint64 timeCode, bool bKey, bool bB);
	//void    _addFrame(MatroskaStream *stream, uint32 size);

public:
	MatroskaReadHandler(const wchar_t *pszFile);
	virtual ~MatroskaReadHandler(void);

	void AddRef();
	void Release();
	long ReadData(void *buffer, long nBytes);
	void SeekFile(__int64 filePos);
	IOCallback& getIOCallback(void);
	IAVIReadStream *GetStream(DWORD fccType, LONG lParam);
	IAVIReadStream *GetStream(LONG lParam);
	ModList<MatroskaReadNode>& GetStreamList(void);
	long GetStreamIndexByType(MatroskaReadNode *stream);
	long GetSegmentDuration(int iSegment);
	void EnableFastIO(bool);
	bool isOptimizedForRealtime();
	bool isStreaming();
	bool isIndexFabricated();
	bool AppendFile(const wchar_t *pszFile);
	bool getSegmentHint(const char **ppszPath);
	void GetCommentList(ModList<ModStreamComment>& list);
	void GetChapterList(ModChapterList& list);
	VDStringW writeReport(void);
	VDStringW GetFileName(void);
	KaxTags	*GetTags(void);
	KaxChapters	*GetChapters(void);
};


class ModFrameIndex {
public:
	long	m_FrameNumber;
	bool	m_bBFrame;

	ModFrameIndex(long frameNum, bool bBFrame)
	: m_FrameNumber(frameNum)
	, m_bBFrame(bBFrame) {
	};
};

class MatroskaReadStream : public IAVIReadStream {
protected:
	MatroskaReadHandler	*m_Parent;
	EbmlStream			*m_EbmlStream;
	//EbmlElement	*ElementLevel0;
	EbmlElement	*ElementLevel1;
	EbmlElement	*ElementLevel2;
	EbmlElement	*ElementLevel3;
	EbmlElement	*ElementLevel4;
	int			UpperElementLevel;
	uint64		m_TimecodeScale;
	VDFraction	m_TimeFactor;
	MatroskaStream		*m_Stream;
	MatroskaReadNode		*m_ReadNode;
	MatroskaStreamSegment	*m_CurrentStreamSegment;
	void			*m_Format;
	sint32			m_FormatSize;
	sint16			m_BlockAlign;
	ModCompactList<MatroskaCluster>::iterator	m_ClusterIterator;
	ModCompactList<MatroskaCluster>::iterator	m_TmpIterator;
	bool			m_bVBRHandling;
	uint32			m_BlockIndex;
	sint64			m_BlockByte;
	MatroskaCluster	*m_Cluster;
	sint64			m_Length;
	sint64			m_LastFilePos;
	uint8			*m_KeyFrames;
	uint8			*m_KeyFramesEnd;
	uint8			*m_BFrames;
	uint8			*m_BFramesEnd;
	bool			m_bIsVideo;
	bool			m_bCanTrimPacket;
	bool			m_bIsText;
	ModSubtitleProvider	m_SubtitleProvider;

	long			_GetIndexFromDisplayIndex(long lFrame);
	bool			_IsBFrame(long lFrame);
	long			FindStreamSegment(long lStart);
	inline void		SynchSegmentInformation(void);
	void			SeekFirstStreamSegment(void);
	void			SeekLastStreamSegment(void);
	bool			SeekPreviousSegment(void);
	bool			SeekNextSegment(void);
	HRESULT			ReadSubtitle(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples);
	HRESULT			ReadCBR(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples);

public:
	MatroskaReadStream(MatroskaReadNode *readNode);
	virtual ~MatroskaReadStream();

	void SetVBRHandling(bool bVBRHandling);
	void Reinit(void);

	HRESULT BeginStreaming(long lStart, long lEnd, long lRate);
	HRESULT EndStreaming();
	HRESULT Info(AVISTREAMINFO *pasi, long lSize);
	long GetIndexFromDisplayIndex(long lFrame);
	bool IsKeyFrame(long lFrame);
	bool IsBFrame(long lFrame);
	VDFraction getDAR(void);
	HRESULT Read(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples);
	long Start();
	long End();
	long PrevKeyFrame(long lFrame);
	long NextKeyFrame(long lFrame);
	long NearestKeyFrame(long lFrame);
	HRESULT FormatSize(long lFrame, long *plSize);
	HRESULT ReadFormat(long lFrame, void *pFormat, long *plSize);
	bool isStreaming();
	bool isKeyframeOnly();
	bool hasBFrames(void);
	void GetFramesToDecode(long lDesiredFrame, std::vector<ModFrameIndex>& listFrames, long lLastReadFrame);

	long GetStreamIndexByType(void);

	bool getVBRInfo(double& bitrate_mean, double& bitrate_stddev, double& maxdev);
	void fixVBRAudio(bool bFix, double averageBitrate);
	sint64 sampleOffset(sint64 lFrame);
	sint32 samplesLag(sint64 lFrame);
	sint32 sourceOffsetMS(void);

	void GetCommentList(ModList<ModStreamComment>& list);
	void GetChapterList(ModChapterList& list);
	uint32 GetStreamType(void);
	uint16 GetNbHeaderPackets(void);
	ogg_packet *GetHeaderPacket(uint16 index);
	std::string GetCodecID(void);
	std::vector<uint8> GetCodecPrivate(void);
	void *GetKaxTrackEntry(void);
	VDStringW writeReport(void);
	ModSubtitleProvider *GetSubtitleProvider(void);
};

#endif // __MATROSKAREADHANDLER_H__