//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <vd2/system/log.h>
#include <vd2/Dita/resources.h>
#include "crash.h"
#include "MJPEGDecoder.h"
#include <tchar.h>
#include "misc.h"
#include "StreamSourceMatroska.h"
#include "MatroskaReadHandler.h"


bool VideoSourceMatroska::QueryInterface(InterfaceID ID, void **destination) {
	/*if(ID == IOggStream::GetInterfaceID()) {
		*destination = static_cast<IOggStream *>(this);
		return true;
	} else */if(ID == IMatroskaStream::GetInterfaceID()) {
		*destination = static_cast<IMatroskaStream *>(this);
		return true;
	}
	return VideoSourceAVI::QueryInterface(ID, destination);
}

VideoSourceMatroska::VideoSourceMatroska(MatroskaReadHandler *readHandler)
: VideoSourceAVI(readHandler) {
	m_SourceType = SOURCE_MKV | SOURCE_VIDEO;
	clearSourceInfo();
	m_SourceName = L"Input Matroska File";
	m_SourceDesc = L"Video stream";

	m_ReadBufferNum = -1;
	m_ReadBufferBytes = 0;
	AllocReadBuffer(0x10000);

	m_bHasBFrames = (pAVIStream && ((MatroskaReadStream *)pAVIStream)->hasBFrames());
}

bool VideoSourceMatroska::init(void) {
	bool bResult = VideoSourceAVI::init();
	m_bHasBFrames = (pAVIStream && ((MatroskaReadStream *)pAVIStream)->hasBFrames());
	return bResult;
}

void VideoSourceMatroska::Reinit(void) {
	VideoSourceAVI::Reinit();
	m_bHasBFrames = (pAVIStream && ((MatroskaReadStream *)pAVIStream)->hasBFrames());
}

char VideoSourceMatroska::getFrameTypeChar(long lFrameNum) {
	if (lFrameNum<lSampleFirst || lFrameNum >= lSampleLast)
		return ' ';

	if(m_bHasBFrames) {
		if (_isKey(lFrameNum))
			return 'I';

		if (((MatroskaReadStream *)pAVIStream)->IsBFrame(lFrameNum))
			return 'B';

		long lBytes, lSamples;
		int err = _read(lFrameNum, 1, NULL, 0, &lBytes, &lSamples);

		if (err != AVIERR_OK)
			return ' ';

		return lBytes ? 'P' : 'D';
	} else {
		if (_isKey(lFrameNum))
			return 'K';

		long lBytes, lSamples;
		int err = _read(lFrameNum, 1, NULL, 0, &lBytes, &lSamples);

		if (err != AVIERR_OK)
			return ' ';

		return lBytes ? ' ' : 'D';
	}
}

bool VideoSourceMatroska::_isB(LONG lSample) {
	return ((MatroskaReadStream *)pAVIStream)->IsBFrame(lSample);
}

bool VideoSourceMatroska::hasBFrames(void) {
	return m_bHasBFrames/*((MatroskaReadStream *)pAVIStream)->hasBFrames()*/;
}

void VideoSourceMatroska::AllocReadBuffer(long size) {
	m_ReadBufferLen = size;
	m_ReadBuffer.resize(size);
	m_pReadBuffer = &m_ReadBuffer.front();
}

void *VideoSourceMatroska::_StreamGetFrame(void *inputBuffer, LONG data_len, BOOL is_key, BOOL is_preroll, long frame_num) {
	DWORD err;

	if (isKey(frame_num)) {
		if (mbConcealingErrors) {
			const unsigned frame = frame_num;
			VDLogAppMessage(kVDLogWarning, kVDST_VideoSource, kVDM_ResumeFromConceal, 1, &frame);
		}
		mbConcealingErrors = false;
	}

	if (!data_len || mbConcealingErrors) return getFrameBuffer();

	if (bDirectDecompress) {
		int to_copy = getImageFormat()->biSizeImage;
		if (data_len < to_copy) {
			if (mErrorMode != kErrorModeReportAll) {
				const unsigned actual = data_len;
				const unsigned expected = to_copy;
				const unsigned frame = frame_num;
				VDLogAppMessage(kVDLogWarning, kVDST_VideoSource, kVDM_FrameTooShort, 3, &frame, &actual, &expected);
				to_copy = data_len;
			} else
				throw MyError("VideoSourceMatroska: uncompressed frame %u is short (expected %d bytes, got %d)", frame_num, getImageFormat()->biSizeImage, data_len);
		}
		
		memcpy(getFrameBuffer(), inputBuffer, to_copy);
	} else if (fUseGDI) {
		if (!hbmLame)
			throw MyError("Insufficient GDI resources to convert frame.");

		SetDIBits(NULL, hbmLame, 0, getDecompressedFormat()->biHeight, inputBuffer, (BITMAPINFO *)getFormat(),
			DIB_RGB_COLORS);
		GdiFlush();
	} else if (hicDecomp && !bDirectDecompress) {
		// Asus ASV1 crashes with zero byte frames!!!

		if (data_len) {
			BITMAPINFOHEADER *bih_src = bmihTemp;
			BITMAPINFOHEADER *bih_dst = getDecompressedFormat();

			bmihTemp->biSizeImage = data_len;

			if (IsMMXState())
				throw MyInternalError("MMX state left on: %s:%d", __FILE__, __LINE__);

			VDCHECKPOINT;

			vdprotected3("decompressing video frame %d with \"%.64s\" [biCompression=%08x]", unsigned, frame_num, const char *, szCodecName, unsigned, getImageFormat()->biCompression) {
				if (use_ICDecompressEx)
					err = 	ICDecompressEx(
								hicDecomp,
		//						  (is_preroll ? ICDECOMPRESS_PREROLL : 0) |
		//						  | (data_len ? 0 : ICDECOMPRESS_NULLFRAME)
								  (is_key ? 0 : ICDECOMPRESS_NOTKEYFRAME),
								bih_src,
								inputBuffer,
								0,0,
								bih_src->biWidth, bih_src->biHeight,
								bih_dst,
								lpvBuffer,
								0,0,
								bih_dst->biWidth, bih_dst->biHeight
								);

				else
					err = 	ICDecompress(
								hicDecomp,
		//						  (is_preroll ? ICDECOMPRESS_PREROLL : 0) |
		//						  | (data_len ? 0 : ICDECOMPRESS_NULLFRAME)
								  (is_key ? 0 : ICDECOMPRESS_NOTKEYFRAME),
								bih_src,
								inputBuffer,
								bih_dst,
								lpvBuffer
								);
			}
			CheckMMX();

			VDCHECKPOINT;

			if (ICERR_OK != err) {
				if (mErrorMode == kErrorModeReportAll) {
					throw MyICError(err, "Error decompressing video frame %d:\n\n%%s\n(error code %d)", frame_num, (int)err);
				} else {
					const unsigned frame = frame_num;
					VDLogAppMessage(kVDLogWarning, kVDST_VideoSource, kVDM_DecodingError, 1, &frame);

					if (mErrorMode == kErrorModeConceal)
						mbConcealingErrors = true;
				}
			}
		}

	} else if (mdec) {

		try {
			const BITMAPINFOHEADER *pbih = getDecompressedFormat();
			unsigned long *pBuffer = (unsigned long *)getFrameBuffer();

			if (pbih->biCompression == '2YUY')
				mdec->decodeFrameYUY2(pBuffer, (unsigned char *)inputBuffer, data_len);
			else if (pbih->biCompression == 'YVYU')
				mdec->decodeFrameUYVY(pBuffer, (unsigned char *)inputBuffer, data_len);
			else if (pbih->biBitCount == 16)
				mdec->decodeFrameRGB15(pBuffer, (unsigned char *)inputBuffer, data_len);
			else if (pbih->biBitCount == 32)
				mdec->decodeFrameRGB32(pBuffer, (unsigned char *)inputBuffer, data_len);
			else
				__asm int 3
		} catch(char *s) {
			throw MyError(s);
		}

   } else {
		const BITMAPINFOHEADER *bih = getImageFormat();
		long nBytesRequired = ((bih->biWidth * bih->biBitCount + 31)>>5) * 4 * bih->biHeight;
		void *tmpBuffer = 0;

		if (data_len < nBytesRequired) {
			if (mErrorMode != kErrorModeReportAll) {
				const unsigned frame = frame_num;
				const unsigned actual = data_len;
				const unsigned expected = nBytesRequired;
				VDLogAppMessage(kVDLogWarning, kVDST_VideoSource, kVDM_FrameTooShort, 3, &frame, &actual, &expected);

				tmpBuffer = malloc(nBytesRequired);
				if (!tmpBuffer)
					throw MyMemoryError();

				memcpy(tmpBuffer, inputBuffer, data_len);

				inputBuffer = tmpBuffer;
			} else
				throw MyError("VideoSourceAVI: uncompressed frame %u is short (expected %d bytes, got %d)", frame_num, nBytesRequired, data_len);
		}

		DIBconvert(inputBuffer, getImageFormat(), getFrameBuffer(), getDecompressedFormat());

		free(tmpBuffer);
   }
//		memcpy(getFrameBuffer(), inputBuffer, getDecompressedFormat()->biSizeImage);

	lLastFrame = frame_num;

	return getFrameBuffer();
}

int VideoSourceMatroska::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {
	int result;
	++m_CSQueue;
	try {
		result = VideoSourceAVI::_read(lStart, lCount, lpBuffer, cbBuffer, lBytesRead, lSamplesRead);
	} catch(...) {
		--m_CSQueue;
		throw;
	}
	--m_CSQueue;

	return result;
}

void *VideoSourceMatroska::streamGetFrame(void *inputBuffer, LONG data_len, BOOL is_key, BOOL is_preroll, long frame_num) {
	if(!m_bHasBFrames)
		return VideoSourceAVI::streamGetFrame(inputBuffer, data_len, is_key, is_preroll, frame_num);

	void *buffer = NULL;

	++m_CSQueue;
	try {
		buffer = getFrame(frame_num);
	} catch(...) {
		--m_CSQueue;
		throw;
	}
	--m_CSQueue;

	return buffer;
}

void *VideoSourceMatroska::getFrame(LONG lFrameDesired) {
	if(!m_bHasBFrames)
		return VideoSourceAVI::getFrame(lFrameDesired);

	void *dataBuffer = NULL, *dataBuffer_Save = NULL;
	LONG dataBufferSize = 0, dataBufferSize_Save = 0;
	LONG lFrameNum, lFrameKey;
	LONG lBytesRead, lSamplesRead;
	DWORD err;
	int aviErr;

	// illegal frame number?

	if (lFrameDesired < lSampleFirst || lFrameDesired >= lSampleLast)
		throw MyError("VideoSourceAVI: bad frame # (%d not within [%d, %d])", lFrameDesired, lSampleFirst, lSampleLast-1);

	// do we already have this frame?

	if (lLastFrame == lFrameDesired)
		return getFrameBuffer();

	//

	lFrameKey = nearestKey(lFrameDesired);

	_RPT1(0,"Nearest key frame: %ld\n", lFrameKey);

	if (hicDecomp) {

		// tell VCM we're going to do a little decompression...

		if (lLastFrame == -1) {
			err = ICDecompressBegin(hicDecomp, getImageFormat(), getDecompressedFormat());
			if (err != ICERR_OK) throw MyICError("VideoSourceMatroska", err);
		}
	}

	_RPT2(0,"VideoSourceAVI: obtaining frame %ld, last was %ld\n", lFrameDesired, lLastFrame);

	try {
		std::vector<ModFrameIndex>	listFrames;

		if( ((m_ReadBufferNum == lFrameDesired) && (lLastFrame != -1) )
		|| (lFrameDesired == lFrameKey) ) {
			listFrames.push_back(ModFrameIndex(lFrameDesired, false));
			// XviD need to be given the first i-frame twice to actually decode it
			// (otherwise you get a "Warning: Nothing to output\nBFrame decoder lag")
			// XviD & DivX needs to be given other i-frames twice too
			if( (lLastFrame == -1) || (lFrameDesired == lFrameKey) )
				listFrames.push_back(ModFrameIndex(lFrameDesired, false));
		} else
			((MatroskaReadStream *)pAVIStream)->GetFramesToDecode(lFrameDesired, listFrames, m_ReadBufferNum);

		ModFrameIndex *pFrameIndex = &listFrames.front();
		int nbFrames = listFrames.size();
		bool bUseReadBuffer = false;

		while(nbFrames--) {
			lFrameNum = pFrameIndex->m_FrameNumber;
			if(bUseReadBuffer = !pFrameIndex->m_bBFrame) {
				dataBuffer = m_pReadBuffer;
				dataBufferSize = m_ReadBufferLen;
			} else {
				dataBuffer = dataBuffer_Save;
				dataBufferSize = dataBufferSize_Save;
			}
			pFrameIndex++;

			_RPT1(0, "Reading frame %ld\n", lFrameNum);

			if(m_ReadBufferNum == lFrameNum) {
				lBytesRead = m_ReadBufferBytes;
				lSamplesRead = 1;
			} else for(;;) {
				if (!dataBuffer)
					aviErr = AVIERR_BUFFERTOOSMALL;
				else
					aviErr = read(lFrameNum, 1, dataBuffer, dataBufferSize, &lBytesRead, &lSamplesRead);

				if (aviErr == AVIERR_BUFFERTOOSMALL) {
					aviErr = read(lFrameNum, 1, NULL, 0, &lBytesRead, &lSamplesRead);

					if (aviErr) throw MyAVIError("VideoSourceMatroska", aviErr);

					dataBufferSize = (lBytesRead+65535) & -65535;

					if(bUseReadBuffer) {
						AllocReadBuffer(dataBufferSize);
						dataBuffer = m_pReadBuffer;
					} else {
						void *newDataBuffer;

						if (!(newDataBuffer = reallocmem(dataBuffer, dataBufferSize)))
							throw MyMemoryError();

						dataBuffer = newDataBuffer;
						dataBuffer_Save = dataBuffer;
						dataBufferSize_Save = dataBufferSize_Save;
					}

				} else if (aviErr) {
					throw MyAVIError("VideoSourceMatroska", aviErr);
				} else {
//					if (hicDecomp || lBytesRead) break;
//					--lFrameNum;
					break;
				}
			};

			if (!lBytesRead) continue;

			if(bUseReadBuffer) {
				m_ReadBufferNum = lFrameNum;
				m_ReadBufferBytes = lBytesRead;
			}

			if (fUseGDI)
				_StreamGetFrame(dataBuffer, lBytesRead, TRUE, FALSE, lFrameNum);
			else if (hicDecomp && !bDirectDecompress) {
				bmihTemp->biSizeImage = lBytesRead;

				if (lBytesRead) {
					VDCHECKPOINT;

					if (IsMMXState())
						throw MyInternalError("MMX state left on: %s:%d", __FILE__, __LINE__);

					vdprotected3("decompressing video frame %d with \"%.64s\" [biCompression=%08x]", unsigned, lFrameNum, const char *, szCodecName, unsigned, getImageFormat()->biCompression) {
						if (ICERR_OK != (err = 	ICDecompress(
										hicDecomp,
	//									  (lFrameNum<lFrameDesired ? ICDECOMPRESS_PREROLL : 0) |
	//									  (lBytesRead ? 0 : ICDECOMPRESS_NULLFRAME) |
										  (lFrameNum > lFrameKey ? ICDECOMPRESS_NOTKEYFRAME : 0),
										bmihTemp,
										dataBuffer,
										getDecompressedFormat(),
										lpvBuffer)))

							throw MyICError(err, "Error decompressing video frame %d:\n\n%%s\n(error code %d)", lFrameNum, (int)err);
					}

					CheckMMX();

					VDCHECKPOINT;
				}
			} else if (mdec) {
				try {
					const BITMAPINFOHEADER *pbih = getDecompressedFormat();
					unsigned long *pBuffer = (unsigned long *)getFrameBuffer();

					if (pbih->biCompression == '2YUY')
						mdec->decodeFrameYUY2(pBuffer, (unsigned char *)dataBuffer, lBytesRead);
					else if (pbih->biCompression == 'YVYU')
						mdec->decodeFrameUYVY(pBuffer, (unsigned char *)dataBuffer, lBytesRead);
					else if (pbih->biBitCount == 16)
						mdec->decodeFrameRGB15(pBuffer, (unsigned char *)dataBuffer, lBytesRead);
					else if (pbih->biBitCount == 32)
						mdec->decodeFrameRGB32(pBuffer, (unsigned char *)dataBuffer, lBytesRead);
					else
						__asm int 3
				} catch(char *s) {
					throw MyError(s);
				}
			} else {
				const BITMAPINFOHEADER *bih = getImageFormat();
				long nBytesRequired = ((bih->biWidth * bih->biBitCount + 31)>>5) * 4 * bih->biHeight;

				if (lBytesRead < nBytesRequired)
					throw MyError("VideoSourceAVI: uncompressed frame is short (expected %d bytes, got %d)", nBytesRequired, lBytesRead);

				if (!bDirectDecompress)
					DIBconvert(dataBuffer, getImageFormat(), getFrameBuffer(), getDecompressedFormat());
				else
					memcpy(getFrameBuffer(), dataBuffer, getDecompressedFormat()->biSizeImage);
			}
//			} else memcpy(getFrameBuffer(), dataBuffer, getDecompressedFormat()->biSizeImage);

		};

	} catch(const MyError&) {
		if (dataBuffer_Save) freemem(dataBuffer_Save);
		ICDecompressEnd(hicDecomp);
		lLastFrame = -1;
		throw;
	}

	if (dataBuffer_Save) freemem(dataBuffer_Save);
//	if (hicDecomp) ICDecompressEnd(hicDecomp);

	lLastFrame = lFrameDesired;

	return getFrameBuffer();
}

/*void VideoSourceMatroska::streamSetDesiredFrame(long frame_num) {
	m_ReadBufferNum = stream_current_frame;
	stream_desired_frame = frame_num;

	m_ListFrames.clear();

	long lFrameKey = nearestKey(stream_desired_frame);

	if( ((m_ReadBufferNum == stream_desired_frame) && (lLastFrame != -1) )
	|| (stream_desired_frame == lFrameKey) ) {
		m_ListFrames.push_back(ModFrameIndex(stream_desired_frame, false));
		// XviD need to be given the first i-frame twice to actually decode it
		// (otherwise you get a "Warning: Nothing to output\nBFrame decoder lag")
		// XviD & DivX needs to be given other i-frames twice too
		if( (lLastFrame == -1) || (stream_desired_frame == lFrameKey) )
			m_ListFrames.push_back(ModFrameIndex(stream_desired_frame, false));
	} else
		((MatroskaReadStream *)pAVIStream)->GetFramesToDecode(frame_num, m_ListFrames, m_ReadBufferNum);

	m_NextFrame = &m_ListFrames.front();
	m_FrameRemaining = m_ListFrames.size();
}

long VideoSourceMatroska::streamGetNextRequiredFrame(BOOL *is_preroll) {
	if(!m_FrameRemaining--)
		return -1;
	stream_current_frame = m_NextFrame->m_FrameNumber;
	*is_preroll = (stream_current_frame != stream_desired_frame);
	m_NextFrame++;
	return stream_current_frame;
}*/

/*uint32 VideoSourceMatroska::GetOggStreamType(void) {
	return ((MatroskaReadStream *)pAVIStream)->GetOggStreamType();
}

uint16 VideoSourceMatroska::GetNbHeaderPackets(void) {
	return ((MatroskaReadStream *)pAVIStream)->GetNbHeaderPackets();
}

ogg_packet *VideoSourceMatroska::GetHeaderPacket(uint16 index) {
	return ((MatroskaReadStream *)pAVIStream)->GetHeaderPacket(index);
}*/

void VideoSourceMatroska::GetCommentList(ModList<ModStreamComment>& list) {
	((MatroskaReadStream *)pAVIStream)->GetCommentList(list);
}

void VideoSourceMatroska::GetChapterList(ModChapterList& list) {
	((MatroskaReadStream *)pAVIStream)->GetChapterList(list);
}

VDStringW VideoSourceMatroska::writeReport(void) {
	VDStringW report = VideoSourceAVI::writeReport();
	report += L"\r\n";
	report += ((MatroskaReadStream *)pAVIStream)->writeReport();
	return report;
}

std::string VideoSourceMatroska::GetCodecID(void) {
	return ((MatroskaReadStream *)pAVIStream)->GetCodecID();
}

std::vector<uint8> VideoSourceMatroska::GetCodecPrivate(void) {
	return ((MatroskaReadStream *)pAVIStream)->GetCodecPrivate();
}

void *VideoSourceMatroska::GetKaxTrackEntry(void) {
	return ((MatroskaReadStream *)pAVIStream)->GetKaxTrackEntry();
}

//////////////////////////////////////////////////////////////////////

bool StreamSourceMatroska::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IStreamIndex::GetInterfaceID()) {
		*destination = static_cast<IStreamIndex *>(this);
		return true;
	} else if(ID == ICommentList::GetInterfaceID()) {
		*destination = static_cast<ICommentList *>(this);
		return true;
	} else if((ID==IOggStream::GetInterfaceID()) && (getSourceType()==SOURCE_AUDIO_VORBIS)) {
		*destination = static_cast<IOggStream *>(this);
		return true;
	} else if(ID == IMatroskaStream::GetInterfaceID()) {
		*destination = static_cast<IMatroskaStream *>(this);
		return true;
	}
	return false;
}

StreamSourceMatroska::StreamSourceMatroska(MatroskaReadHandler *readHandler, long index)
: AudioSource() {
	VDASSERT(readHandler && (index>=0));
	m_ReadHandler = readHandler;
	m_ReadStream = NULL;
	m_StreamIndex = index;
	m_bVBRHandling = true;

	m_SourceType = SOURCE_MKV;
	clearSourceInfo();
	m_SourceName = L"Matroska File";
}

StreamSourceMatroska::~StreamSourceMatroska(void) {
	if(m_ReadStream)
		delete m_ReadStream;
}


bool StreamSourceMatroska::init() {
	LONG format_len;

	m_ReadStream = (MatroskaReadStream *)m_ReadHandler->GetStream(m_StreamIndex);

	if (!m_ReadStream) return false;

	if (m_ReadStream->Info(&streamInfo, sizeof streamInfo))
		return FALSE;

	m_ReadStream->FormatSize(0, &format_len);

	if (!allocFormat(format_len)) return false;

	if (m_ReadStream->ReadFormat(0, getFormat(), &format_len))
		return false;

	lSampleFirst = m_ReadStream->Start();
	lSampleLast = m_ReadStream->End();

	m_SourceType |= m_ReadStream->GetStreamType();

	VDASSERT(getSourceMainType() != SOURCE_VIDEO);

	std::string codecID = m_ReadStream->GetCodecID();
	VDStringW codecIDW(VDFastTextAToW(codecID.c_str()));

	switch(getSourceMainType()) {
	case SOURCE_TEXT:
		m_SourceDesc = VDStringW::setf(L"Text stream %d, ID [%s]"
			, GetStreamIndex()+1, codecIDW.c_str());
		break;
	case SOURCE_AUDIO:
		if(getSourceType() == SOURCE_AUDIO_VORBIS)
			m_SourceDesc = VDStringW::setf(L"Audio stream %d [Vorbis], %dch, %dHz, %.1fkbps"
				, GetStreamIndex()+1
				, getWaveFormat()->nChannels, getWaveFormat()->nSamplesPerSec, ((double)8*getWaveFormat()->nAvgBytesPerSec)/1000.);
		else {
			TCHAR szTag[5], szCodecName[1024];
			memset(szTag, 0, sizeof(szTag));
			_snprintf(szTag, sizeof(szTag)-1, _T("%04X"), getWaveFormat()->wFormatTag);
			memset(szCodecName, 0, sizeof(szCodecName));
			ModGetCodecName(_T("audio"), szTag, _T("Unknown"), szCodecName, sizeof(szCodecName)/sizeof(szCodecName[0]));
			VDStringW sTag;
			sTag = L"tag 0x";
			if(sizeof(TCHAR) == sizeof(char)) {
				sTag += VDTextAToW((char *)szTag);
				sTag += L", ";
				sTag += VDTextAToW((char *)szCodecName);
			} else {
				sTag += (wchar_t *)szTag;
				sTag += L", ";
				sTag += (wchar_t *)szCodecName;
			}
			m_SourceDesc = VDStringW::setf(L"Audio stream %d [%s], ID [%s], %dch, %dHz, %.1fkbps"
				, GetStreamIndex()+1
				, sTag.c_str()
				, codecIDW.c_str()
				, getWaveFormat()->nChannels, getWaveFormat()->nSamplesPerSec, ((double)8*getWaveFormat()->nAvgBytesPerSec)/1000.);
		}
		break;
	}

	return true;
}

void StreamSourceMatroska::extend(bool bVBRHandling, sint32 groupFrames) {
	m_bVBRHandling = bVBRHandling;
	m_ReadStream->SetVBRHandling(bVBRHandling);
	Reinit();
}

void StreamSourceMatroska::Reinit() {
	LONG formatLen = getFormatLen();
	m_ReadStream->ReadFormat(0, getFormat(), &formatLen);
	m_ReadStream->Info(&streamInfo, sizeof streamInfo);
	lSampleFirst = m_ReadStream->Start();
	lSampleLast = m_ReadStream->End();
}

bool StreamSourceMatroska::isStreaming() {
	return m_ReadStream->isStreaming();
}

void StreamSourceMatroska::streamBegin(bool fRealTime) {
	m_ReadStream->BeginStreaming(lSampleFirst, lSampleLast, fRealTime ? 1000 : 2000);
}

void StreamSourceMatroska::streamEnd() {
	m_ReadStream->EndStreaming();
}

sint64 StreamSourceMatroska::sampleOffset(sint64 lSample) {
	return m_ReadStream->sampleOffset(lSample);
}

sint32 StreamSourceMatroska::samplesLag(sint64 lStart) {
	return m_ReadStream->samplesLag(lStart);
}

sint32 StreamSourceMatroska::sourceOffsetMS(void) {
	return m_ReadStream->sourceOffsetMS();
}

int StreamSourceMatroska::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lpBytesRead, LONG *lpSamplesRead) {
	return m_ReadStream->Read(lStart, lCount, lpBuffer, cbBuffer, lpBytesRead, lpSamplesRead);
}

DWORD StreamSourceMatroska::GetStreamType(void) {
	switch(MAIN_TYPE(m_ReadStream->GetStreamType())) {
	case SOURCE_VIDEO:
		throw MyError("Accessing to a Video stream via StreamSourceMatroska!");
		return streamtypeVIDEO;
	case SOURCE_AUDIO:
		return streamtypeAUDIO;
	case SOURCE_TEXT:
		return streamtypeTEXT;
	default:
		break;
	}
	return 0;
}

long StreamSourceMatroska::GetStreamIndex(void) {
	return m_ReadStream->GetStreamIndexByType();
}

uint32 StreamSourceMatroska::GetOggStreamType(void) {
	return m_ReadStream->GetStreamType();
}

uint16 StreamSourceMatroska::GetNbHeaderPackets(void) {
	return m_ReadStream->GetNbHeaderPackets();
}

ogg_packet *StreamSourceMatroska::GetHeaderPacket(uint16 index) {
	return m_ReadStream->GetHeaderPacket(index);
}

void StreamSourceMatroska::GetCommentList(ModList<ModStreamComment>& list) {
	m_ReadStream->GetCommentList(list);
}

void StreamSourceMatroska::GetChapterList(ModChapterList& list) {
	m_ReadStream->GetChapterList(list);
}

std::string StreamSourceMatroska::GetCodecID(void) {
	return m_ReadStream->GetCodecID();
}

std::vector<uint8> StreamSourceMatroska::GetCodecPrivate(void) {
	return m_ReadStream->GetCodecPrivate();
}

void *StreamSourceMatroska::GetKaxTrackEntry(void) {
	return m_ReadStream->GetKaxTrackEntry();
}

VDStringW StreamSourceMatroska::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	report += L"\r\n";
	report += m_ReadStream->writeReport();
	return report;
}

//////////////////////////////////////////////////////////////////////

bool TextSourceMatroska::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == ISubtitleHandler::GetInterfaceID()) {
		*destination = static_cast<ISubtitleHandler *>(this);
		return true;
	}
	return StreamSourceMatroska::QueryInterface(ID, destination);
}

TextSourceMatroska::TextSourceMatroska(MatroskaReadHandler *readHandler, long index)
: StreamSourceMatroska(readHandler, index) {
}

bool TextSourceMatroska::IsASCII(void) {
#pragma message(__TODO__ "Get the correct type")
	return false;
}

bool TextSourceMatroska::HasBOM(void) {
	return false;
}

eTextType TextSourceMatroska::GetOutputType(void) {
	return kTextType_UTF8;
}

ModSubtitleProvider *TextSourceMatroska::GetSubtitleProvider(void) {
	return m_ReadStream->GetSubtitleProvider();
}