//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <windows.h>
#include <commdlg.h>
#include <vd2/system/Error.h>
#include <vd2/system/filesys.h>
#include "oshelper.h"
#include "InputFileOGM.h"
#include "VideoSource.h"
#include "StreamSourceOGM.h"
#include "OGMReadHandler.h"
#include "prefs.h"


extern HWND	g_hWnd;
extern const wchar_t fileFiltersAppendOGM[]=
		L"OGM segment (*.ogm)\0"		L"*.ogm\0"
		L"All files (*.*)\0"			L"*.*\0"
		;

static bool fileExists(const char *fn) {
	DWORD dwAttrib = GetFileAttributes(fn);

	return dwAttrib != -1 && !(dwAttrib & FILE_ATTRIBUTE_DIRECTORY);
}


class VDInputDriverOGM : public vdrefcounted<IVDInputDriver> {
public:
	const wchar_t *GetSignatureName() { return L"OGM input driver (internal)"; }

	int GetDefaultPriority() {
		return 0;
	}

	uint32 GetFlags() { return kF_Video | kF_Audio | kF_Text; }

	const wchar_t *GetFilenamePattern() {
		return L"OGM file (*.ogm)\0*.ogm\0";
	}

	bool DetectByFilename(const wchar_t *pszFilename) {
		return false;
	}

	int DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
		if (nHeaderSize >= 4) {
			if(!memcmp(pHeader, "OggS", 4))
				return 1;
		}

		return -1;
	}

	InputFile *CreateInputFile(uint32 flags) {
		InputFileOGM *pf = new_nothrow InputFileOGM;

		if (!pf)
			throw MyMemoryError();

		if (flags & kOF_Quiet)
			pf->setAutomated(true);

		if (flags & kOF_AutoSegmentScan)
			pf->EnableSegmentAutoscan();

		return pf;
	}
};

extern IVDInputDriver *VDCreateInputDriverOGM() { return new VDInputDriverOGM; }

//////////////////////////////////////////////////////////////////////

char InputFileOGM::szME[] = "Ogg Media File import filter";

InputFileOGM::InputFileOGM(void)
: InputFileAVI() {
	pAVIFile = NULL;
}

InputFileOGM::~InputFileOGM(void) {
}

void InputFileOGM::Init(const wchar_t *szFile) {
	AddFilename(szFile);

	if(!(pAVIFile = new OGMReadHandler(szFile)))
		throw MyMemoryError();

	if(!(videoSrc = new VideoSourceOGM((OGMReadHandler *)pAVIFile)))
		throw MyMemoryError();

	if(!videoSrc->init())
		throw MyError("%s: problem opening video stream", szME);

	ModList<OGMReadNode>::iterator it(((OGMReadHandler *)pAVIFile)->GetStreamList());
	int index = 0;
	while(it) {
		OggStream *aStream = (*it)->m_Segments.GetFirst()->m_OggStream;
		if(MAIN_TYPE(aStream->m_StreamType) != SOURCE_VIDEO) {
			vdrefptr<AudioSource> newSource;
			if(MAIN_TYPE(aStream->m_StreamType) == SOURCE_TEXT)
				newSource = new TextSourceOGM((OGMReadHandler *)pAVIFile, index);
			else
				newSource = new StreamSourceOGM((OGMReadHandler *)pAVIFile, index);
			if(!newSource)
				throw MyMemoryError();
			if(!newSource->init()) {
				newSource = NULL;
				break;
			}
			m_Sources.push_back(newSource);
		}
		it++;
		index++;
	}

	if(fAutoscanSegments) {
		const wchar_t *pszName = VDFileSplitPath(szFile);
		VDStringW sPathBase(szFile, pszName - szFile);
		VDStringW sPathTail(pszName);

		if (sPathTail.size() >= 7 && !wcsicmp(sPathTail.c_str() + sPathTail.size() - 7, L".00.ogm")) {
			int nSegment = 0;

			sPathTail.resize(sPathTail.size() - 7);

			VDStringW sPathPattern(VDMakePath(sPathBase, sPathTail));

			try {
				const char *pszPathHint;

				while(pAVIFile->getSegmentHint(&pszPathHint)) {

					++nSegment;

					VDStringW sPath(sPathPattern + VDswprintf(L".%02d.ogm", 1, &nSegment));

#if 0
					if (!VDDoesPathExist(sPath)) {
						if (pszPathHint && *pszPathHint) {
							sPathBase = VDTextAToW(pszPathHint) + sPathTail;
						}

						sPath = (sPathPattern + VDswprintf(L".%02d.ogm", 1, &nSegment));
					}
#endif

#if 0
					if (!VDDoesPathExist(sPath)) {
						char szPath[MAX_PATH];
						wchar_t szTitle[MAX_PATH];

						swprintf(szTitle, L"Cannot find file %s", sPath.c_str());

						strcpy(szPath, VDTextWToA(sPath).c_str());

						const VDStringW fname(VDGetLoadFileName(VDFSPECKEY_LOADVIDEOFILE, (VDGUIHandle)g_hWnd, szTitle, fileFiltersAppend, L"avi", 0, 0));

						if (fname.empty())
							throw MyUserAbortError();

						if (!Append(fname.c_str()))
							break;

						sPathPattern = VDMakePath(VDFileSplitPathLeft(fname), sPathTail);
					} else if (!Append(sPath.c_str()))
						break;
#endif
					if(!VDDoesPathExist(sPath) || !Append(sPath.c_str()))
						break;
				}
			} catch(const MyError& e) {
				char err[128];
				sprintf(err, "Cannot load video segment %02d", nSegment);

				e.post(NULL, err);
			}
		}
	}
}

bool InputFileOGM::Append(const wchar_t *szFile) {
	if(!szFile)
		return true;

	if(pAVIFile->AppendFile(szFile)) {
		if(videoSrc)
			((VideoSourceOGM *)&*videoSrc)->Reinit();
		std::vector< vdrefptr<AudioSource> >::const_iterator it;
		for(it=m_Sources.begin() ; it!=m_Sources.end() ; it++)
				((StreamSourceOGM *)&*(*it))->Reinit();

		AddFilename(szFile);

		return true;
	}

	return false;
}

VDStringW InputFileOGM::writeReport(void) {
	VDStringW report = InputFile::writeReport();
	report += L"\r\n\r\n";
	report += ((OGMReadHandler *)pAVIFile)->writeReport();
	return report;
}