//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "ModAudioStreamSourceVorbis.h"
#include "ModClassInterfaces.h"


ModAudioStreamSourceVorbis::ModAudioStreamSourceVorbis(AudioSource *source, long first_samp, long max_samples)
: AudioStream() {
	m_Source = source;
	m_BlankSamples = 0;
	m_SamplesInPCMBuffer = 0;
	m_SamplesLeftInPCMBuffer = 0;
	m_PCMBufferPos = 0;
	m_bSourceEOS = false;
	m_bDecoderInitialized = false;
	vorbis_info_init(&m_VorbisInfo);
	vorbis_comment_init(&m_VorbisComment);
	memset(&m_Packet, 0, sizeof(m_Packet));
	m_Packet.packet = (unsigned char *)m_InputBuffer;
	m_Packet.granulepos = 0;

	m_SourceRate = m_Source->getStreamInfo().dwRate;
	m_SourceScale = m_Source->getStreamInfo().dwScale;

	if (max_samples < 0)
		max_samples = 0;

	WAVEFORMATEX *iFormat = m_Source->getWaveFormat();
	WAVEFORMATEX *oFormat;

	if (!AllocFormat(sizeof(WAVEFORMATEX)))
		throw MyMemoryError();

	oFormat = (WAVEFORMATEX *)GetFormat();

	long blank_samples = 0;
	if(first_samp < 0) {
		blank_samples = -first_samp;
		first_samp = 0;
	}
	cur_samp = first_samp;
	end_samp = first_samp + max_samples;

	IOggStream *IOgg;
	if(!m_Source->QueryInterface(IOggStream::GetInterfaceID(), (void **)&IOgg))
		throw MyError("Not a valid Vorbis source (doesn't provide an IOggStream interface)");
	if(IOgg->GetNbHeaderPackets() < 3)
		throw MyError("Not a valid Vorbis source (has less than 3 header packets)");

	int i=0;
	while(i<3) {
		if(vorbis_synthesis_headerin(&m_VorbisInfo, &m_VorbisComment, IOgg->GetHeaderPacket(i++)) < 0)
			throw MyError("Not a valid Vorbis source (one or more invalid header packet)");
	}
    vorbis_synthesis_init(&m_VorbisDSP, &m_VorbisInfo);
    vorbis_block_init(&m_VorbisDSP, &m_VorbisBlock);

	m_bDecoderInitialized = true;

	oFormat->wFormatTag			= WAVE_FORMAT_PCM;
	oFormat->nChannels			= (m_VorbisInfo.channels >= 2)? 2 : 1;
	oFormat->nBlockAlign		= oFormat->nChannels*2;
	oFormat->nSamplesPerSec		= m_VorbisInfo.rate;
	oFormat->nAvgBytesPerSec	= oFormat->nSamplesPerSec * oFormat->nBlockAlign;
	oFormat->wBitsPerSample		= 16;
	oFormat->cbSize				= 0;

	stream_len = std::min<long>(max_samples, m_Source->getEnd() - (blank_samples ? -blank_samples : first_samp));

	m_BlankSamples += blank_samples;

	m_BlockAlign = oFormat->nBlockAlign;
}

ModAudioStreamSourceVorbis::~ModAudioStreamSourceVorbis(void) {
	if(m_bDecoderInitialized) {
		vorbis_block_clear(&m_VorbisBlock);
		vorbis_dsp_clear(&m_VorbisDSP);
	}
	vorbis_comment_clear(&m_VorbisComment);
	vorbis_info_clear(&m_VorbisInfo);
}

bool ModAudioStreamSourceVorbis::ReadSource(void) {
	LONG bytesRead, samplesRead;
	if((m_Source->read(cur_samp
		, AVISTREAMREAD_CONVENIENT
		, m_InputBuffer
		, sizeof(m_InputBuffer)
		, &bytesRead, &samplesRead)==AVIERR_OK) && samplesRead) {
		VDDEBUG("Read %d samples starting at %d, %d bytes\n"
			, samplesRead, cur_samp, bytesRead);
		m_Packet.bytes = bytesRead;
		cur_samp += samplesRead;
		m_Packet.granulepos += samplesRead;
		m_Packet.packetno++;
		return true;
	}
	return false;
}

bool ModAudioStreamSourceVorbis::GetNextFrame(void) {
	if( (m_SamplesInPCMBuffer=vorbis_synthesis_pcmout(&m_VorbisDSP, &m_PCMBuffer)) > 0) {
		m_SamplesLeftInPCMBuffer = m_SamplesInPCMBuffer;
		m_PCMBufferPos = 0;
		return true;
	}

	if(!ReadSource()) {
		m_bSourceEOS = true;
		return false;
	}

	if(vorbis_synthesis(&m_VorbisBlock, &m_Packet) == 0)
		vorbis_synthesis_blockin(&m_VorbisDSP, &m_VorbisBlock);

	return GetNextFrame();
}

long ModAudioStreamSourceVorbis::_Read(void *buffer, long max_samples, long *lplBytes) {
	if(!m_bDecoderInitialized) {
		if(lplBytes)
			*lplBytes = 0;
		return 0;
	}

	LONG lSamplesLeft = max_samples;
	LONG lTotalBytes = 0;

	while(lSamplesLeft > 0) {
		if(m_BlankSamples) {
			long blankSamples = m_BlankSamples;
			if(blankSamples > max_samples)
				blankSamples = max_samples;
			long blankBytes = m_BlockAlign * blankSamples;

			memset(buffer, 0, blankBytes);

			buffer = (char *)buffer + blankBytes;

			max_samples -= blankSamples;
			lTotalBytes += blankBytes;
			lSamplesLeft -= blankSamples;
			m_BlankSamples -= blankSamples;
			continue;
		}

		if(m_SamplesLeftInPCMBuffer) {
			long samplesToRead = lSamplesLeft;
			if(samplesToRead > m_SamplesLeftInPCMBuffer)
				samplesToRead = m_SamplesLeftInPCMBuffer;
			int i, j;
			for(i=0 ; i<GetFormat()->nChannels ; i++) {
				ogg_int16_t *ptr = (ogg_int16_t *)buffer + i;
				float *mono = m_PCMBuffer[i] + m_PCMBufferPos;
				for(j=0 ; j<samplesToRead ; j++) {
					int val = VDRoundToInt(mono[j] * 32767.f);
					if(val > 32767) {
						*ptr = 32767;
						VDDEBUG("Clipping in frame %ld\n",(long)(m_VorbisDSP.sequence));
					} else if(val < -32768) {
						*ptr = -32768;
						VDDEBUG("Clipping in frame %ld\n",(long)(m_VorbisDSP.sequence));
					} else
						*ptr = val;
					ptr += GetFormat()->nChannels;
				}
			}

			lSamplesLeft -= samplesToRead;
			lTotalBytes += samplesToRead * m_BlockAlign;
			m_SamplesLeftInPCMBuffer -= samplesToRead;
			m_PCMBufferPos += samplesToRead;

			// When we have finished with the buffer, inform the library
			if(!m_SamplesLeftInPCMBuffer) {
				vorbis_synthesis_read(&m_VorbisDSP, m_SamplesInPCMBuffer);
				m_SamplesInPCMBuffer = 0;
				m_PCMBufferPos = 0;
			}
			
			buffer = (char *)buffer + samplesToRead * m_BlockAlign;
			continue;
		}

		if(!GetNextFrame())
			break;
	}

	*lplBytes = lTotalBytes;

	return (max_samples - lSamplesLeft);
}


bool ModAudioStreamSourceVorbis::Skip(long samples) {
	while(samples) {
		if(samples <= m_BlankSamples) {
			m_BlankSamples -= samples;
			return true;
		}
		samples -= m_BlankSamples;
		m_BlankSamples = 0;
		if(samples <= m_SamplesLeftInPCMBuffer) {
			m_SamplesLeftInPCMBuffer -= samples;
			m_PCMBufferPos += samples;
			return true;
		}
		samples -= m_SamplesLeftInPCMBuffer;
		m_SamplesLeftInPCMBuffer = 0;
		cur_samp += samples;
		if(cur_samp > end_samp) {
			return false;
			cur_samp = end_samp;
		}
		return true;
	}
	return true;
}

BOOL ModAudioStreamSourceVorbis::_isEnd() {
	return ((cur_samp>=end_samp) && m_bSourceEOS && !m_BlankSamples && !m_SamplesLeftInPCMBuffer);
}
