//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <vd2/system/File64.h>
#include "FastWriteStream.h"
#include "OGMOutputFile.h"
#include "OGMPacketizer.h"
#include "OGMSpecs.h"
#include <limits.h>


OGMVideoOutputStream::OGMVideoOutputStream(OGMOutputFile *parent)
: AVIOutputStream(NULL) {
	m_Parent = parent;
	m_Packetizer = NULL;
}

OGMVideoOutputStream::~OGMVideoOutputStream(void) {
	if(m_Packetizer)
		delete m_Packetizer;
}

void OGMVideoOutputStream::setPacketizer(OGMPacketizer *packetizer) {
	VDASSERT(!m_Packetizer);
	if(m_Packetizer)
		delete m_Packetizer;
	m_Packetizer = packetizer;
}

void OGMVideoOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	m_Packetizer->ProcessFrame((void *)pBuffer, cbBuffer, samples, true, (flags & AVIIF_KEYFRAME ? true : false), false);
	m_Parent->ProcessPendingMuxerPages();
}

void OGMVideoOutputStream::finalize(void) {
	// No packetizer here when demuxing and saving a Vorbis stream
	if(m_Packetizer && !m_Packetizer->IsEOS())
		m_Packetizer->WriteEOS();
}

void OGMVideoOutputStream::setEndOfStream(void) {
	m_Packetizer->WriteEOS();
}

void OGMVideoOutputStream::setStreamLag(sint32 samplesLag) {
	m_Packetizer->AddLagGranulepos(samplesLag);
}

OGMPacketizer *OGMVideoOutputStream::GetPacketizer(void) {
	return m_Packetizer;
}

//////////////////////////////////////////////////////////////////////

OGMOutputStream::OGMOutputStream(OGMOutputFile *parent)
: AVIOutputStream(NULL) {
	m_Parent = parent;
	m_Packetizer = NULL;
	m_bPureData = true;
	m_bFlushData = false;
	m_bWriteEOSOnEnd = true;
}

OGMOutputStream::~OGMOutputStream(void) {
	if(m_Packetizer)
		delete m_Packetizer;
}

void OGMOutputStream::setPacketizer(OGMPacketizer *packetizer) {
	VDASSERT(!m_Packetizer);
	if(m_Packetizer)
		delete m_Packetizer;
	m_Packetizer = packetizer;
}

void OGMOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	m_Packetizer->ProcessFrame((void *)pBuffer, cbBuffer, samples, m_bPureData, (flags & AVIIF_KEYFRAME ? true : false), m_bFlushData);
	m_Parent->ProcessPendingMuxerPages();
}

void OGMOutputStream::finalize(void) {
	if(m_bWriteEOSOnEnd && !m_Packetizer->IsEOS())
		m_Packetizer->WriteEOS();
}

void OGMOutputStream::setEndOfStream(void) {
	if(m_bWriteEOSOnEnd)
		m_Packetizer->WriteEOS();
}

void OGMOutputStream::setStreamLag(sint32 samplesLag) {
	m_Packetizer->AddLagGranulepos(samplesLag);
}

//////////////////////////////////////////////////////////////////////

OGMSubtitleOutputStream::OGMSubtitleOutputStream(OGMOutputFile *parent, eTextType inputType)
: OGMOutputStream(parent)
, m_TextUtil(inputType, kTextType_ASCII, false, true)
{
	// Output as SubtiDS
}

OGMSubtitleOutputStream::~OGMSubtitleOutputStream(void) {
}

void OGMSubtitleOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	VDStringW subtitle = m_TextUtil.GetText((void *)pBuffer, cbBuffer);
	VDStringA subASCII = VDTextWToA(subtitle);
	if(subASCII.length())
		subASCII += '\r';
	m_Packetizer->ProcessFrame((void *)subASCII.c_str(), subASCII.length()+1, samples, m_bPureData, (flags & AVIIF_KEYFRAME ? true : false), m_bFlushData);
	m_Parent->ProcessPendingMuxerPages();
}

//////////////////////////////////////////////////////////////////////

char OGMOutputFile::szME[] = "OGMOutputFile";

OGMOutputFile::OGMOutputFile(bool bCreateVideo) {
	m_fastIO			= NULL;
	m_i64FilePosition	= 0;
	m_bCaching			= true;
	m_VideoStream		= NULL;
	m_VideoSerial		= 0;
	m_bEOSWritten		= false;
	m_bCreateVideo		= bCreateVideo;
}

OGMOutputFile::~OGMOutputFile(void) {
	m_PacketizerList.DeleteContents(true);
	m_PacketizerList.Clear();
	m_Outputs.DeleteContents(true);
	m_Outputs.Clear();

	if(m_File.isOpen()) {
		if (m_File.seekNT(m_i64FilePosition))
			m_File.truncateNT();

		m_File.closeNT();
	}

	if(m_fastIO)
		delete m_fastIO;
}

void OGMOutputFile::disable_os_caching() {
	m_bCaching = FALSE;
}

void OGMOutputFile::setBufferSize(sint32 iBufferSize) {
	mBufferSize = iBufferSize;
}

IVDMediaOutputStream *OGMOutputFile::createVideoStream() {
	VDASSERT(!videoOut);
	if(!m_bCreateVideo)
		return NULL;
	if (!(videoOut = new_nothrow OGMVideoOutputStream(this)))
		throw MyMemoryError();

	return videoOut;
}

IVDMediaOutputStream *OGMOutputFile::createAudioStream() {
	VDASSERT(!audioOut);
	if (!(audioOut = new_nothrow OGMOutputStream(this)))
		throw MyMemoryError();

	return audioOut;
}

bool OGMOutputFile::initOutputStreams(ModVideoStream& videoStream, ModList<ModDubStream>& streams) {
	if(m_bCreateVideo)
		if(!(videoOut = new OGMVideoOutputStream(this)))	return false;

	m_VideoStream = &videoStream;

	sint32 serial = 1;
	ModList<ModDubStream>::iterator it(streams);
	while(it) {
		ModDubStream *curInStream = *it;
		OGMOutputStreamNode *curOutStream = m_Outputs.AddLast();
		// Output subtitles in ASCII as SubtiDS does
		ISubtitleHandler *pSubH;
		if(curInStream->m_Source->QueryInterface(ISubtitleHandler::GetInterfaceID(), (void **)&pSubH)) {
			//pSubH->FormatAsSubtiDS(true);
			//pSubH->OutputInASCII();
			curOutStream->m_Output = new OGMSubtitleOutputStream(this, pSubH->GetOutputType());
		} else {
			if(curInStream->m_Source->getSourceMainType() == SOURCE_TEXT)
				throw MyError("Subtitle stream %ld doesn't expose the ISubtitleHandler interface!"
					, serial);
			curOutStream->m_Output = new OGMOutputStream(this);
		}
		if(!curOutStream->m_Output)
			return FALSE;
		curOutStream->m_Input = curInStream;
		curOutStream->m_Output->m_bPureData = ( (curInStream->m_Source->getSourceType() != SOURCE_AUDIO_VORBIS)
				|| (curInStream->m_Opts.mode > DubAudioOptions::M_NONE));
		curOutStream->m_Output->m_bFlushData = (curInStream->m_Source->getSourceMainType() == SOURCE_TEXT);
		curOutStream->m_Output->m_bWriteEOSOnEnd = (curInStream->m_Source->getSourceMainType() != SOURCE_TEXT);
		CopyModList<ModStreamComment>(curInStream->m_Comments, curOutStream->m_Comments);
		curOutStream->m_Chapters = curInStream->m_Chapters;
		curInStream->m_Serial = serial;
		curOutStream->m_Serial = serial;
		serial++;
		it++;
	}

	return true;
}

IVDMediaOutputStream *OGMOutputFile::getOutputStream(uint32 serial) {
	m_OutputIterator.begin(m_Outputs);
	while(m_OutputIterator) {
		if((*m_OutputIterator)->m_Serial == serial)
			return (*m_OutputIterator)->m_Output;
		m_OutputIterator++;
	}
	throw MyError("Attempt to access unavailable output stream (%lu) in OGM", serial);
}

bool OGMOutputFile::init(const wchar_t *szFile) {

	OGMPacketizerNode *newPacketizerNode = NULL;

	if(!m_bCaching) {
		m_File.open(szFile, nsVDFile::kWrite | nsVDFile::kDenyNone | nsVDFile::kOpenAlways | nsVDFile::kSequential);

		if (!(m_fastIO = new FastWriteStream(szFile, mBufferSize, mBufferSize/4, true)))
			throw MyMemoryError();
	} else {
		m_File.open(szFile, nsVDFile::kWrite | nsVDFile::kDenyWrite | nsVDFile::kOpenAlways | nsVDFile::kSequential | nsVDFile::kWriteThrough);
	}

	m_i64FilePosition = 0;

	if(videoOut) {
		AVIStreamHeader_fixed infos = videoOut->getStreamInfo();
		BITMAPINFOHEADER *bi = (BITMAPINFOHEADER *)videoOut->getFormat();

		// Get the framerate
		VDFraction SampleRate(infos.dwRate, infos.dwScale);
		long maxFrameSize = infos.dwSuggestedBufferSize;
		// Alright I should verify what should be the suggested size of the
		// generated Ogg stream's buffer and for our Packetizer's buffer
		// But it is of no use here : Maybe the user asked for a compressed stream
		// Moreover I don't have the input stream itself :p
		// So the solution : get the suggested buffer size
		// If the value is 0, then take 32k (should be enough)
		// Anyway if the buffer is too small at a moment the Packetizer realloc a new one :)

		if(maxFrameSize<=0)
			maxFrameSize = 32 * 1024;

		stream_header sh;
		memset(&sh, 0, sizeof(stream_header));

		strcpy(sh.streamtype, "video");
		strcpy(sh.subtype, (char *)&bi->biCompression);
		sh.size = sizeof(stream_header);
		sh.time_unit = SampleRate.scale64ir(10000000);
		sh.samples_per_unit = 1;
		sh.default_len = 1;
		sh.buffersize = maxFrameSize;
		sh.bits_per_sample = bi->biBitCount;
		sh.sh.video.width = bi->biWidth;
		sh.sh.video.height = bi->biHeight;

		int nbHeaders = 1;
		if(m_VideoStream->m_Comments.GetCount() || m_VideoStream->m_Chapters.size())
			nbHeaders = 2;

		newPacketizerNode = new OGMPacketizerNode();
		newPacketizerNode->m_Serial = m_VideoSerial;
		newPacketizerNode->m_Packetizer = new OGMPacketizer(m_VideoSerial, maxFrameSize, 0, true);
		if(!newPacketizerNode->m_Packetizer)
			throw MyMemoryError();
		((OGMVideoOutputStream *)videoOut)->setPacketizer(newPacketizerNode->m_Packetizer);
		if(!newPacketizerNode->m_Packetizer->CopyHeader(&sh, sh.size))
			return FALSE;
		if(!newPacketizerNode->m_Packetizer->WriteHeader())
			return FALSE;
		if(m_VideoStream->m_Comments.GetCount() || m_VideoStream->m_Chapters.size()) {
			if(!newPacketizerNode->m_Packetizer->WriteComments(m_VideoStream->m_Comments, m_VideoStream->m_Chapters, m_VideoStream->m_Source->asStream()->getSourceContainer() == SOURCE_OGM, true))
				return FALSE;
		}
		m_PacketizerList.Append(newPacketizerNode);
	}

	if(m_Outputs.GetCount()) {
		m_OutputIterator.begin(m_Outputs);
		while(m_OutputIterator) {
			OGMOutputStreamNode *curNode = *m_OutputIterator;
			AVISTREAMINFO infos = curNode->m_Input->m_Source->getStreamInfo();
			WAVEFORMATEX *wfe = curNode->m_Input->m_AudioStream->GetFormat();

			long maxFrameSize = infos.dwSuggestedBufferSize;

			if(maxFrameSize <= 0)
				maxFrameSize = 32 * 1024;

			std::vector<uint8>	ogmHeader;
			ogmHeader.resize(sizeof(stream_header) + wfe->cbSize);
			// Extra data of WAVEFORMATEX are appended to the header
			stream_header& sh = *((stream_header *)&ogmHeader.front());
			memset(&ogmHeader.front(), 0, ogmHeader.size());
			sh.size = sizeof(stream_header) + wfe->cbSize;
			if(wfe->cbSize)
				memcpy(&sh + 1, wfe + 1, wfe->cbSize);

			if(curNode->m_Input->m_Source->getSourceMainType() != SOURCE_TEXT) {
				strcpy(sh.streamtype, "audio");
				VDStringA format = VDStringA::setf("%04X", wfe->wFormatTag);
				memcpy(sh.subtype, format.c_str(), 4);
				sh.time_unit = 10000000;
				sh.samples_per_unit = wfe->nSamplesPerSec;
				sh.default_len = 1;
				sh.buffersize = maxFrameSize;
				sh.bits_per_sample = wfe->wBitsPerSample;
				sh.sh.audio.avgbytespersec = wfe->nAvgBytesPerSec;
				sh.sh.audio.blockalign = wfe->nBlockAlign;
				sh.sh.audio.channels = wfe->nChannels;
			} else {
				strcpy(sh.streamtype, "text");
				sh.size = sizeof(stream_header);
				sh.time_unit = 10000;
				sh.samples_per_unit = 1;
				sh.default_len = 1;
				sh.buffersize = 16 * 1024;
			}

			newPacketizerNode = new OGMPacketizerNode();
			newPacketizerNode->m_Serial = curNode->m_Serial;
			newPacketizerNode->m_bWriteEOSOnEnd = curNode->m_Output->m_bWriteEOSOnEnd;
			newPacketizerNode->m_bHasFewPackets = (curNode->m_Input->m_Source->getSourceMainType()==SOURCE_TEXT);
			newPacketizerNode->m_Packetizer = new OGMPacketizer(curNode->m_Serial, maxFrameSize, (curNode->m_Input->m_Source->getSourceMainType()==SOURCE_TEXT ? 0 : 1), (curNode->m_Input->m_Source->getSourceMainType()==SOURCE_TEXT ? false : true));
			if(!newPacketizerNode->m_Packetizer)
				throw MyMemoryError();
			curNode->m_Output->setPacketizer(newPacketizerNode->m_Packetizer);
			if(!newPacketizerNode->m_Packetizer->CopyHeader(&sh, sh.size))
				return FALSE;
			if( (curNode->m_Input->m_Source->getSourceType() == SOURCE_AUDIO_VORBIS)
				&& (curNode->m_Input->m_Opts.mode == DubAudioOptions::M_NONE)) {
				IOggStream *pOgg;
				if(!curNode->m_Input->m_Source->QueryInterface(IOggStream::GetInterfaceID(), (void **)&pOgg))
					return FALSE;
				if(pOgg->GetNbHeaderPackets() != 3)
					return FALSE;
				newPacketizerNode->m_Packetizer->ProcessPacket(pOgg->GetHeaderPacket(0), true, true);
				if(!newPacketizerNode->m_Packetizer->WriteComments(curNode->m_Input->m_Comments, curNode->m_Input->m_Chapters, true, false))
					return FALSE;
				newPacketizerNode->m_Packetizer->ProcessPacket(pOgg->GetHeaderPacket(2), true, true);
			} else {
				if(!newPacketizerNode->m_Packetizer->WriteHeader())
					return FALSE;
				if(curNode->m_Input->m_Comments.GetCount() || curNode->m_Input->m_Chapters.size()) {
					if(!newPacketizerNode->m_Packetizer->WriteComments(curNode->m_Input->m_Comments, curNode->m_Input->m_Chapters, curNode->m_Input->m_Source->getSourceContainer() == SOURCE_OGM, true))
						return FALSE;
				}
			}
			if(curNode->m_Input->m_OffsetSamples < 0)
				newPacketizerNode->m_Packetizer->AddLagGranulepos(curNode->m_Input->m_SamplesPerBlock.scale64r(-curNode->m_Input->m_OffsetSamples));
			m_PacketizerList.Append(newPacketizerNode);
			m_OutputIterator++;
		}
	} else if(audioOut) {
		delete audioOut;
		audioOut = NULL;
	}

	m_FileName = szFile;

	FlushPacketizers();
	ProcessHeaderPages();

	return TRUE;
}

void OGMOutputFile::finalize(void) {
	if(videoOut)	((OGMVideoOutputStream *)videoOut)->finalize();
	if(audioOut)	((OGMOutputStream *)audioOut)->finalize();

	m_OutputIterator.begin(m_Outputs);
	while(m_OutputIterator) {
		(*m_OutputIterator)->m_Output->finalize();
		m_OutputIterator++;
	}
	WriteEOS();
	FlushPacketizers();
	ProcessPendingMuxerPages();

	if(m_fastIO) {
		m_fastIO->Flush1();
		m_fastIO->Flush2((HANDLE)m_File.getRawHandle());
		delete m_fastIO;
		m_fastIO = NULL;
	}

	RewriteHeaderPages();
	RewriteLastPages();

	m_File.truncateNT();
	m_File.closeNT();
}

bool OGMOutputFile::PendingHeaderPage(void) {
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		if((*m_PacketizerIterator)->m_Packetizer->PendingHeaderPage())
			return true;
		m_PacketizerIterator++;
	}

	return false;
}

OGMMuxerPage *OGMOutputFile::GetHeaderPage(void) {
	sint32 maxPriority = _I32_MIN;
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curPack = *m_PacketizerIterator;
		if(curPack->m_Packetizer->PendingHeaderPage()) {
			if(curPack->m_Packetizer->PendingHeaderPage()->m_Priority > maxPriority)
				maxPriority = curPack->m_Packetizer->PendingHeaderPage()->m_Priority;
		}
		m_PacketizerIterator++;
	}

	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curPack = *m_PacketizerIterator;
		if(curPack->m_Packetizer->PendingHeaderPage()) {
			if(curPack->m_Packetizer->PendingHeaderPage()->m_Priority == maxPriority)
				return curPack->m_Packetizer->GetHeaderPage();
		}
		m_PacketizerIterator++;
	}

	return NULL;
}

void OGMOutputFile::WriteEOS(void) {
	if(m_bEOSWritten)
		return;

	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curPack = *m_PacketizerIterator;
		if(curPack->m_bWriteEOSOnEnd && !curPack->m_Packetizer->IsEOS() && !curPack->m_Packetizer->IsInvalid())
			curPack->m_Packetizer->WriteEOS();
		m_PacketizerIterator++;
	}

	m_bEOSWritten = true;
}

bool OGMOutputFile::PendingMuxerPage(void) {
	// Some packetizers could be marked as handling few packets
	// Using PrepareMuxerPages() in a normal way could result
	// in queueing a lot of Packets in the other Packetizers
	// before reaching the next available Packet
	// Hence the m_bHasFewPackets flag
	//
	// PendingMuxerPage() will verify that all valid Packetizers that
	// haven't the m_bHasFewPackets flag have at least one Page pending
	// But we must also verify that there is at least one Packetizer that
	// haven't the m_bHasFewPackets
	bool bAtLeastOneMuxerPagePending = false;
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curPack = *m_PacketizerIterator;
		if(!curPack->m_Packetizer->IsEOS() && !curPack->m_Packetizer->IsInvalid()
			&& !curPack->m_bHasFewPackets
			&& !curPack->m_Packetizer->PendingMuxerPage())
			return false;
		if(curPack->m_Packetizer->PendingMuxerPage())
			bAtLeastOneMuxerPagePending = true;
		m_PacketizerIterator++;
	}

	// Reaching this point could mean :
	//1) There is one or more valid Packetizers without the has_few_packets
	//  flag, and they all have at least one muxer_page pending
	//2) All Packetizers have the has_few_packets flag
	//
	// Case 1 is OK and means at_least_one_mpage_pending is true, so return true;
	// Case 2 make us test the at_least_one_mpage_pending flag to know
	// if it's OK or not.
	return bAtLeastOneMuxerPagePending;
}

OGMMuxerPage *OGMOutputFile::GetMuxerPage(void) {
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curPack = *m_PacketizerIterator;
		if(curPack->m_Packetizer->PendingMuxerPage())
			return curPack->m_Packetizer->GetMuxerPage();
		m_PacketizerIterator++;
	}

	VDDEBUG("No pending page available\n"); 

	return NULL;
}

OGMMuxerPage *OGMOutputFile::GetFirstMuxerPage(void) {
	OGMPacketizerNode *winnerPack	= NULL;
	sint64 smallest_timestamp	= _I64_MAX;
	sint64 timestamp			= _I64_MAX;
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curPack = *m_PacketizerIterator;
		if(curPack->m_Packetizer->PendingMuxerPage()
			&& ((timestamp=curPack->m_Packetizer->GetFirstTimestamp())<smallest_timestamp) ) {
			winnerPack = curPack;
			smallest_timestamp = timestamp;
		}
		m_PacketizerIterator++;
	}

	if(winnerPack)
		return winnerPack->m_Packetizer->GetMuxerPage();

	VDDEBUG("No pending page available\n"); 

	return NULL;
}

void OGMOutputFile::WriteMuxerPage(OGMMuxerPage *muxerPage) {
	VDASSERT(muxerPage);
	
	ogg_page *opage = muxerPage->m_Page;
	VDASSERT(opage && opage->header && (opage->header_len>0) && opage->body && (opage->body_len>0));

	OGMPacketizerNode *packNode = GetPacketizerNode(muxerPage->m_Serial);
	VDASSERT(packNode);
	if(packNode->m_FirstPageLength) {
		packNode->m_LastPageLength = opage->header_len + opage->body_len;
		packNode->m_LastPagePos = m_i64FilePosition;
	} else {
		packNode->m_FirstPageLength = opage->header_len + opage->body_len;
		packNode->m_FirstPagePos = m_i64FilePosition;
		packNode->m_LastPageLength = packNode->m_FirstPageLength;
		packNode->m_LastPagePos = packNode->m_FirstPagePos;
	}

	if(m_fastIO) {
		m_fastIO->Put(opage->header, opage->header_len);
		m_fastIO->Put(opage->body, opage->body_len);
	} else {
		DWORD n_write = 0;
		m_File.write(opage->header, opage->header_len);
		m_File.write(opage->body, opage->body_len);
	}
	m_i64FilePosition += packNode->m_LastPageLength;
}

OGMPacketizerNode *OGMOutputFile::GetPacketizerNode(sint32 serial) {
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		if((*m_PacketizerIterator)->m_Serial == serial)
			return (*m_PacketizerIterator);
		m_PacketizerIterator++;
	}
	return NULL;
}

void OGMOutputFile::ProcessPendingMuxerPages(void) {
	OGMMuxerPage *muxerPage = NULL;

	while(PendingMuxerPage()) {
		muxerPage = GetFirstMuxerPage();
		WriteMuxerPage(muxerPage);
		if(muxerPage)
			delete muxerPage;
	}
}

void OGMOutputFile::FlushPacketizers(void) {
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		(*m_PacketizerIterator)->m_Packetizer->ProcessPages(true);
		m_PacketizerIterator++;
	}
}

void OGMOutputFile::ProcessHeaderPages(void) {
	OGMMuxerPage *muxerPage = NULL;
	while(PendingHeaderPage()) {
		muxerPage = GetHeaderPage();
		WriteMuxerPage(muxerPage);
		if(muxerPage)
			delete muxerPage;
	}
}

void OGMOutputFile::RewriteHeaderPages(void) {
	__int64 curFilePos = m_i64FilePosition;
	DWORD dwError = NO_ERROR;
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curNode = *m_PacketizerIterator;
		if(curNode->m_FirstPageLength == 0) {
			m_PacketizerIterator++;
			continue;
		}
		void *lpBuffer = malloc(curNode->m_FirstPageLength);
		if(!lpBuffer) {
			m_File.seekNT(curFilePos);
			throw MyMemoryError();
		}
		if(!curNode->m_Packetizer->IsInvalid() && curNode->m_Packetizer->RewriteHeaderPage(lpBuffer, curNode->m_FirstPageLength)) {
			m_File.seekNT(curNode->m_FirstPagePos);
			m_File.write(lpBuffer, curNode->m_FirstPageLength);
		}
		free(lpBuffer);
		m_PacketizerIterator++;
	}
	m_File.seekNT(curFilePos);
}

void OGMOutputFile::RewriteLastPages(void) {
	if(!videoOut)
		return;
	OGMPacketizer *videoPack = ((OGMVideoOutputStream *)videoOut)->GetPacketizer();
	if(!videoPack)
		return;
	VDFile tmpFile;
	if(!tmpFile.open(m_FileName.c_str(), nsVDFile::kRead | nsVDFile::kOpenExisting | nsVDFile::kDenyNone))
		return;
	sint64 timestampEnd = videoPack->GetEndTimestamp();
	__int64 curFilePos = m_i64FilePosition;
	DWORD dwError = NO_ERROR;
	m_PacketizerIterator.begin(m_PacketizerList);
	while(m_PacketizerIterator) {
		OGMPacketizerNode *curNode = *m_PacketizerIterator;
		if(!curNode->m_bWriteEOSOnEnd && curNode->m_LastPageLength && !curNode->m_Packetizer->IsInvalid()) {
			m_File.seekNT(curNode->m_LastPagePos);
			void *tmpBuffer = malloc(curNode->m_LastPageLength);
			if(!tmpBuffer) {
				m_File.seekNT(curFilePos);
				throw MyMemoryError();
			}
			if(!tmpFile.seekNT(curNode->m_LastPagePos, nsVDFile::kSeekStart)
				|| (tmpFile.readData(tmpBuffer, curNode->m_LastPageLength)!=curNode->m_LastPageLength))
				dwError = GetLastError();
			else if(curNode->m_Packetizer->RewriteLastPage(tmpBuffer, curNode->m_LastPageLength, timestampEnd)) {
				m_File.write(tmpBuffer, curNode->m_LastPageLength);
			}
			free(tmpBuffer);
		}
		m_PacketizerIterator++;
	}
	m_File.seekNT(curFilePos);
	tmpFile.closeNT();
	if(dwError != NO_ERROR)
		throw MyWin32Error("%s: %%s", dwError, szME);
}
