//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <vorbis/codec.h>
#include "OGMPacketizer.h"
#include "OggComments.h"
#include "ISO-639.h"
#include <limits.h>


OGMPacketizer::OGMPacketizer(sint32 serial, uint32 bufferSize, ogg_int64_t startGranulepos, bool bGranuleposMeansEnd, bool bPriorityHeaders) {
	m_Header = NULL;
	m_HeaderLength = 0;
	m_StreamHeader = NULL;
	m_bStandardHeader = false;
	m_Serial = serial;
	m_bEOS = false;
	m_bInvalid = false;
	m_StartGranulepos = startGranulepos;
	m_bGranuleposMeansEnd = bGranuleposMeansEnd;
	m_HeaderPages = -1;
	m_bPrioritySet = false;
	m_bPriorityHeaders = bPriorityHeaders;
	if(m_bGranuleposMeansEnd)
		m_LastGranulepos = m_StartGranulepos - 1;
	else
		m_LastGranulepos = 0;
	m_Timestamp = 0;
	m_MaxFrameLength = 0;
	if(bufferSize > 0)
		m_BufferSize = bufferSize;
	else
		m_BufferSize = 32*1024;
	m_Buffer = NULL;
	ResizeBuffer();
	m_bCachedPacket = NULL;
	m_NbPackets = 0;
	memset(&m_CachePacket, 0, sizeof(m_CachePacket)); 
	memset(&m_StreamState, 0, sizeof(m_StreamState));
	if(ogg_stream_init(&m_StreamState, m_Serial) == -1) {
		m_bInvalid = true;
		throw MyError("ogg_stream_init failed");
	}
}

void OGMPacketizer::ResizeBuffer(void) {
	if(m_Buffer) {
		VirtualFree(m_Buffer, 0, MEM_RELEASE);
		m_Buffer = NULL;
	}
	if(m_BufferSize) {
		m_Buffer = (unsigned char *)VirtualAlloc(NULL, m_BufferSize + OGM_PACKET_MAXOVERHEAD, MEM_COMMIT, PAGE_READWRITE);
		if(!m_Buffer)
			throw MyMemoryError();
	}
}

OGMPacketizer::~OGMPacketizer(void) {
	if(m_Header)
		delete[] m_Header;
	uint32 pagesPending = m_PageList.GetCount();
	m_PageList.DeleteContents(true);
	m_PageList.Clear();
	m_BufferSize = 0;
	ResizeBuffer();
	ogg_stream_clear(&m_StreamState);
	if(pagesPending)
		throw MyError("Error : %lu Pages were still pending while freeing an OGMPacketizer", pagesPending);
}

uint32 OGMPacketizer::GetBufferSize(void) {
	return m_BufferSize;
}

void OGMPacketizer::AddPage(ogg_page *page) {
	if(m_bInvalid)
		throw MyError("Trying to add a Page in the pending list for an Invalid stream (%ld)", m_Serial);

	OGMMuxerPage *lastPage = m_PageList.AddLast();
	VDASSERT(lastPage);
	VDASSERT(page && page->header && page->body);

	lastPage->m_Page = CopyOggPage(page);
	VDASSERT(lastPage->m_Page);
	// At the present time timestamp equal the time_stamp of the previous page
	// So update m_TimeUS before calling ComputeTimestamp()
	ogg_int64_t granulepos = ogg_page_granulepos(page);
	if(granulepos==0)
		lastPage->m_TimeUS = 0;
	else {
		lastPage->m_TimeUS = m_Timestamp;
		ComputeTimestamp(granulepos+1);
	}

	if(!m_bGranuleposMeansEnd)
		lastPage->m_TimeUS = m_Timestamp;

	lastPage->m_Serial = m_Serial;
}

ogg_page *OGMPacketizer::CopyOggPage(ogg_page *page) {
	ogg_page *newPage = new ogg_page;
	VDASSERT(page);

	if(page->header_len > 0) {
		VDASSERT(page->header);
		newPage->header = (unsigned char *)malloc(page->header_len);
		memcpy(newPage->header, page->header, page->header_len);
		newPage->header_len = page->header_len;
	} else {
		VDDEBUG("Copying an ogg_page with NULL header\n");
		newPage->header = NULL;
		newPage->header_len = 0;
	}

	if(page->body_len > 0) {
		newPage->body = (unsigned char *)malloc(page->body_len);
		memcpy(newPage->body, page->body, page->body_len);
		newPage->body_len = page->body_len;
	} else {
		VDDEBUG("Copying an ogg_page with NULL body\n");
		newPage->body = (unsigned char *)malloc(1);
		newPage->body_len = 0;
		*(newPage->body) = 0;
	}
	
	return newPage;
}

bool OGMPacketizer::CopyHeader(void *header, uint16 headerLen) {
	if(headerLen < sizeof(stream_header))
		return false;

	if(m_Header)
		delete[] m_Header;

	m_Header = new char[headerLen];
	if(!m_Header)
		return false;

	memcpy(m_Header, header, headerLen);
	m_HeaderLength = headerLen;

	m_StreamHeader = (stream_header *)m_Header;

	m_SampleRate = VDFraction::reduce64((sint64)m_StreamHeader->samples_per_unit * 10000000, m_StreamHeader->time_unit);

	return true;
}

bool OGMPacketizer::WriteHeader(void) {
	if(!m_Header)
		return false;

	if(m_bInvalid)
		throw MyError("Trying to produce an header Packet for an Invalid stream (%ld)", m_Serial);

	if(m_NbPackets)
		throw MyError("Trying to produce an header Packet when packets have already been sent (stream %ld)", m_Serial);

	if(m_bCachedPacket)
		ProcessCachedPacket();

	ogg_packet headerPacket;
	headerPacket.b_o_s = 1;
	headerPacket.e_o_s = 0;
	headerPacket.granulepos = 0;
	headerPacket.packet = m_Buffer;

	if(m_BufferSize < m_HeaderLength + 1) {
		m_BufferSize = m_HeaderLength + 1;
		ResizeBuffer();
	}

	*headerPacket.packet = PACKET_TYPE_HEADER;
	headerPacket.bytes = m_HeaderLength + 1;
	memcpy(m_Buffer + 1, m_Header, m_HeaderLength);

	_ProcessPacket(&headerPacket, true, true);
	m_NbPackets++;

	m_bStandardHeader = true;

	return true;
}

bool OGMPacketizer::WriteComments(ModList<ModStreamComment>& comments, ModChapterList& chapters, bool bSourceIsOgg, bool bFlush) {
	if(m_bCachedPacket)
		ProcessCachedPacket();

	VDStringA sTag;
	VDStringW sValueW;
	VDStringA sValue;
	bool bConvertToUTF8 = false;

	ogg_packet commentPacket;
	commentPacket.granulepos = 0;
	commentPacket.packetno = 1;
	vorbis_comment vc;
	vorbis_comment_init(&vc);
	ModList<ModStreamComment>::iterator commentIt(comments);
	while(commentIt) {
		ModStreamComment *curComment = *commentIt;
		if((curComment->m_Type!=nsModStream::cGeneric)
			&& !GetOggCommentTag(curComment->m_Type)
			&& !bSourceIsOgg) {
			commentIt++;
			continue;
		}
		if(curComment->m_Type > cLanguage) {
			int tableIndex = (curComment->m_Type - cLanguage) - 1;

			VDASSERT(ISO639_IsValidIndex_2(tableIndex));

			sValueW = s_LanguageList[tableIndex].m_EnglishName;
			sValueW += L"[";
			sValueW += s_LanguageList[tableIndex].m_ISO639_2_Code;
			sValueW += L"]";
		} else
			sValueW = curComment->m_Value;
		if(bConvertToUTF8)
			sValue = VDTextWToU8(sValueW);
		else
			sValue = VDTextWToA(sValueW);
		const wchar_t *commentTag = GetOggCommentTag(curComment->m_Type);
		if(!commentTag && curComment->m_Tag.size())
			commentTag = curComment->m_Tag.c_str();
		if(commentTag) {
			sTag = VDTextWToA(commentTag, -1);
			vorbis_comment_add_tag(&vc, (char *)sTag.c_str(), (char *)sValue.c_str());
		} else
			vorbis_comment_add(&vc, (char *)sValue.c_str());
		commentIt++;
	}
	chapters.Sort();
	ModChapterList::iterator chapterIt;
	uint32 chapterIndex = 1;
	for(chapterIt=chapters.begin() ; chapterIt!=chapters.end() ; chapterIt++) {
		ModChapterList::tChapterNode *curChapter = &(*chapterIt);
		VDStringW chapterTime = VDStringW::setf(L"CHAPTER%02d", chapterIndex);
		VDStringW chapterName = VDStringW::setf(L"CHAPTER%02dNAME", chapterIndex);
		sTag = VDTextWToA(chapterTime);
		if(bConvertToUTF8)
			sValue = VDTextWToU8(curChapter->GetTime());
		else
			sValue = VDTextWToA(curChapter->GetTime());
		vorbis_comment_add_tag(&vc, (char *)sTag.c_str(), (char *)sValue.c_str());
		sTag = VDTextWToA(chapterName);
		if(bConvertToUTF8)
			sValue = VDTextWToU8(curChapter->GetName(), -1);
		else
			sValue = VDTextWToA(curChapter->GetName());
		vorbis_comment_add_tag(&vc, (char *)sTag.c_str(), (char *)sValue.c_str());
		chapterIndex++;
	}
	vorbis_commentheader_out(&vc, &commentPacket);

	if(m_BufferSize + OGM_PACKET_MAXOVERHEAD < commentPacket.bytes) {
		m_BufferSize = commentPacket.bytes;
		ResizeBuffer();
	}
	memcpy(m_Buffer, commentPacket.packet, commentPacket.bytes);
	// Don't free the packet buffer (handled by the library)
	commentPacket.packet = m_Buffer;
	
	_ProcessPacket(&commentPacket, bFlush, true);

	return true;
}

void OGMPacketizer::ProcessPacket(ogg_packet *packet, bool bFlush, bool bHeader) {
	if(m_bCachedPacket)
		ProcessCachedPacket();

	if(m_BufferSize + OGM_PACKET_MAXOVERHEAD < packet->bytes) {
		m_BufferSize = packet->bytes;
		ResizeBuffer();
	}
	memcpy(m_Buffer, packet->packet, packet->bytes);
	ogg_packet newPacket = *packet;
	newPacket.packet = m_Buffer;

	_ProcessPacket(&newPacket, bFlush, bHeader);
}

void OGMPacketizer::_ProcessPacket(ogg_packet *packet, bool bFlush, bool bHeader) {
	if(m_bInvalid)
		throw MyError("Trying to process a Packet for an invalid stream (%ld)", m_Serial);

	if(bHeader) {
		ogg_stream_packetin(&m_StreamState, packet);
		// First packet (header) should be alone in its own Page
		// Seems the Ogg layer do this on its own (return a Page when
		// calling QueueHeaderPages() after sending the first Packet)
		// But do it on our own too
		if(!m_NbPackets)
			ProcessHeaderPages(true);
		else
			ProcessHeaderPages(bFlush);
		m_NbPackets++;
		// Who knows ...
		if(packet->e_o_s)
			m_bEOS = true;
		return;
	}

	if(m_bCachedPacket)
		ProcessCachedPacket();
	if(m_bEOS) {
		VDDEBUG("Trying to process a new Page after EOS declared (stream %ld)\n", m_Serial);
		return;
	}

	if(packet->granulepos > 0)
		m_LastGranulepos = packet->granulepos;

	// Keep the new packet in cache (usefull when you reach the end of a file and
	// call the ProduceEOSPacket() on a Packetizer, because the Packetizer will then
	// just set the eos flag in the cached packet and then send it to the ogg stream
	if(bFlush) {
		// If flushing asked, don't use the buffer
		ogg_stream_packetin(&m_StreamState, packet);
		ProcessPages(true);
		if(packet->e_o_s)
			m_bEOS = true;
	} else {
		m_CachePacket = *packet;
		m_bCachedPacket = true;
		if(packet->e_o_s)
			ProcessCachedPacket();
	}

	m_NbPackets++;
}

void OGMPacketizer::ProcessCachedPacket(void) {
	if(m_bInvalid)
		throw MyError("Trying to process a cached Packet for an invalid stream (%ld)", m_Serial);
	if(m_bEOS) {
		VDDEBUG("Trying to process a cached Packet after EOS declared (stream %ld)\n", m_Serial);
		return;
	}
#ifdef _DEBUG
	if(!m_bCachedPacket)
		throw MyError("Trying to process cached Packet when there are none (stream %ld)", m_Serial);
#endif

	if(m_CachePacket.e_o_s)
		m_bEOS = true;

	ogg_stream_packetin(&m_StreamState, &m_CachePacket);

	m_bCachedPacket = false;

	ProcessPages(false);
}

void OGMPacketizer::ProcessPages(bool bFlush) {
	ogg_page new_page;

	if(bFlush) {
		while(ogg_stream_flush(&m_StreamState, &new_page))
			AddPage(&new_page);
	} else {
		while(ogg_stream_pageout(&m_StreamState, &new_page))
			AddPage(&new_page);
	}
}

void OGMPacketizer::ProcessHeaderPages(bool bFlush) {
	ogg_page new_page;

	if(bFlush) {
		while(ogg_stream_flush(&m_StreamState, &new_page)) {
			if(m_HeaderPages == -1)
				m_HeaderPages = 1;
			else
				m_HeaderPages++;
			AddPage(&new_page);
			SetMuxerPagePriority();
		}
	} else {
		while(ogg_stream_pageout(&m_StreamState, &new_page)) {
			if(m_HeaderPages == -1)
				m_HeaderPages = 1;
			else
				m_HeaderPages++;
			AddPage(&new_page);
			SetMuxerPagePriority();
		}
	}
}

void OGMPacketizer::SetMuxerPagePriority(void) {
	OGMMuxerPage *lastPage = m_PageList.GetLast();
	VDASSERT(lastPage);
	if(!m_bPrioritySet)
		lastPage->m_Priority = 3;
	else if(m_bPriorityHeaders)
		lastPage->m_Priority = 2;
	else
		lastPage->m_Priority = 1;
	m_bPrioritySet = true;
}

void OGMPacketizer::WriteEOS(void) {
	if(m_bInvalid)
		throw MyError("Trying to process an EOS for an invalid stream (%ld)", m_Serial);

	if(m_bEOS) {
		VDDEBUG("Trying to process an EOS after EOS declared (stream %ld)\n", m_Serial);
		return;
	}

	if(m_bCachedPacket) {
		m_CachePacket.e_o_s = 1;
		ProcessCachedPacket();
	} else {
		if(m_NbPackets) {
			*m_Buffer = PACKET_TYPE_DATA;
			m_CachePacket.b_o_s = 0;
			m_CachePacket.bytes = 1;
			m_CachePacket.e_o_s = 1;
			m_CachePacket.granulepos = m_LastGranulepos;
			m_CachePacket.packet = m_Buffer;
			m_CachePacket.packetno = 0;
			m_bCachedPacket = true;
			ProcessCachedPacket();
		}
		m_bEOS = true;
	}
}

void OGMPacketizer::ComputeTimestamp(ogg_int64_t granulepos) {
	m_Timestamp = m_SampleRate.scale64ir(1000000 * (granulepos - m_StartGranulepos));
}

void OGMPacketizer::AddLagGranulepos(sint64 granuleposDiff) {
	m_LastGranulepos += granuleposDiff;
}

void OGMPacketizer::AddLagTimeUS(sint64 timeUSDiff) {
	AddLagGranulepos(m_SampleRate.scale64r(timeUSDiff) / 1000000);
}

OGMMuxerPage *OGMPacketizer::PendingMuxerPage(void) {
	return m_PageList.GetFirst();
}

OGMMuxerPage *OGMPacketizer::GetMuxerPage(void) {
	if(m_bInvalid)
		throw MyError("Trying to get a Page for an invalid stream (%ld)", m_Serial);

	OGMMuxerPage *firstPage = m_PageList.GetFirst();
	VDASSERT(firstPage);
	if(firstPage)
		m_PageList.DetachNode(firstPage);
	return firstPage;
}

OGMMuxerPage *OGMPacketizer::PendingHeaderPage(void) {
	return (m_HeaderPages > 0 ? PendingMuxerPage() : NULL);
}

OGMMuxerPage *OGMPacketizer::GetHeaderPage(void) {
	if(m_HeaderPages > 0) {
		m_HeaderPages--;
		return GetMuxerPage();
	}
	return NULL;
}

sint64 OGMPacketizer::GetFirstTimestamp(void) {
	OGMMuxerPage *firstPage = m_PageList.GetFirst();
	VDASSERT(firstPage);
	return (firstPage ? firstPage->m_TimeUS : _I64_MAX);
}

void OGMPacketizer::ProcessFrame(void *frame, uint32 frameLength, ogg_int64_t frameSize, bool bPureData, bool bKeyFrame, bool bFlush) {
	if(m_bInvalid)
		throw MyError("Trying to process data for an invalid stream (%ld)", m_Serial);

	if(m_bCachedPacket)
		ProcessCachedPacket();

	if(m_bEOS) {
		VDDEBUG("Trying to process data after EOS declared (stream %ld)\n", m_Serial);
		return;
	}

	m_CachePacket.b_o_s = 0;
	m_CachePacket.e_o_s = 0;
	if(m_bGranuleposMeansEnd)
		m_CachePacket.granulepos = m_LastGranulepos + frameSize;
	else
		m_CachePacket.granulepos = m_LastGranulepos;

	if(m_CachePacket.granulepos < 0) {
		if(!m_bGranuleposMeansEnd && (m_CachePacket.granulepos + frameSize > 0))
			m_CachePacket.granulepos = 0;
		else {
			VDDEBUG("Trying to create a Packet with negative granulepos (stream %ld), skipped\n", m_Serial);
			return;
		}
	}

	if(frameLength > m_BufferSize) {
		VDDEBUG("Size of data (%d) exceeds predefined buffer size (%d).\n", frameLength, m_BufferSize);
		m_BufferSize = frameLength;
		ResizeBuffer();
	}
	if(frameLength > m_MaxFrameLength)
		m_MaxFrameLength = frameLength;

	m_CachePacket.packet = m_Buffer;

	if(bPureData) {
		*m_CachePacket.packet = PACKET_TYPE_DATA;
		uint8 hdrlen = 0;
		if(frameSize > 1) {
			ogg_int64_t nbSamples = frameSize;
			while(nbSamples) {
				*(m_CachePacket.packet + 1 + hdrlen) = nbSamples & 0xFF;
				nbSamples >>= 8;
				hdrlen++;
			}
		}
		*m_CachePacket.packet |= ((hdrlen>>1) & PACKET_LEN_BITS2) | ((hdrlen<<6) & PACKET_LEN_BITS01);
		if(bKeyFrame)
			*m_CachePacket.packet |= PACKET_IS_SYNCPOINT;
		m_CachePacket.bytes = frameLength + hdrlen + 1;
		memcpy(m_Buffer + hdrlen + 1, frame, frameLength);
	} else {
		m_CachePacket.bytes = frameLength;
		memcpy(m_Buffer, frame, frameLength);
	}

	if(bFlush) {
		ogg_stream_packetin(&m_StreamState, &m_CachePacket);
		ProcessPages(true);
	} else
		m_bCachedPacket = true;

	m_LastGranulepos += frameSize;
	m_NbPackets++;
}

sint64 OGMPacketizer::GetEndTimestamp(void) {
	return m_SampleRate.scale64ir(1000000 * (m_LastGranulepos + 1 - m_StartGranulepos));
}

bool OGMPacketizer::RewriteHeaderPage(void *lpBuffer, uint32 pageLength) {
	if(!m_bStandardHeader)
		return false;

	if(!m_StreamHeader)
		return false;

	if(!lpBuffer)
		return false;

	if(m_bInvalid)
		throw MyError("Trying to rewrite header Page for an Invalid stream (%ld)", m_Serial);

	m_StreamHeader->buffersize = m_MaxFrameLength;

	VDASSERT(!m_bCachedPacket);

	ogg_packet headerPacket;
	headerPacket.b_o_s = 1;
	headerPacket.e_o_s = 0;
	headerPacket.granulepos = 0;
	headerPacket.packet = m_Buffer;

	if(m_BufferSize < m_HeaderLength + 1) {
		m_BufferSize = m_HeaderLength + 1;
		ResizeBuffer();
	}

	*headerPacket.packet = PACKET_TYPE_HEADER;
	headerPacket.bytes = m_HeaderLength + 1;
	memcpy(m_Buffer + 1, m_Header, m_HeaderLength);

	ogg_stream_state tmpStreamState;
	if(ogg_stream_init(&tmpStreamState, m_Serial) == -1)
		return false;

	ogg_stream_packetin(&tmpStreamState, &headerPacket);

	ogg_page page;

	if(!ogg_stream_flush(&tmpStreamState, &page))
		return false;

	if(page.header_len+page.body_len != pageLength)
		return false;

	memcpy(lpBuffer, page.header, page.header_len);
	memcpy((char *)lpBuffer+page.header_len, page.body, page.body_len);

	ogg_stream_clear(&tmpStreamState);

	return true;
}

bool OGMPacketizer::RewriteLastPage(void *lpBuffer, uint32 pageLength, sint64 timestampEndUS) {
	if(!lpBuffer)
		return false;

	// We goes here _only_ for text streams
	if(m_bGranuleposMeansEnd)
		return false;

	if(m_bInvalid)
		throw MyError("Trying to rewrite header Page for an Invalid stream (%ld)", m_Serial);

	VDASSERT(!m_bCachedPacket);

	ogg_stream_state tmpStreamStateIn;
	if(ogg_stream_init(&tmpStreamStateIn, m_Serial) == -1)
		return false;
	ogg_stream_reset(&tmpStreamStateIn);
	ogg_stream_state tmpStreamStateOut;
	if(ogg_stream_init(&tmpStreamStateOut, m_Serial) == -1)
		return false;

	ogg_packet	packet;
	ogg_page	page_in;
	ogg_page	page_out;

	memset(&packet, 0, sizeof(packet));

	ogg_sync_state syncState;
	ogg_sync_init(&syncState);

	bool bPageRewritten = false;

	char *buffer = ogg_sync_buffer(&syncState, pageLength);
	if(!buffer)
		throw MyError("ogg_sync_buffer failed");
	memcpy(buffer, lpBuffer, pageLength);
	ogg_sync_wrote(&syncState, pageLength);
	if((ogg_sync_pageseek(&syncState, &page_in)==pageLength)
		&& (ogg_page_serialno(&page_in)==m_Serial)
		&& !(ogg_page_continued(&page_in))) {
		unsigned char charZero = 0;
		sint64 pageNumber = ogg_page_pageno(&page_in);
		ogg_int64_t maxGranulepos = m_SampleRate.scale64r(timestampEndUS) / 1000000;
		ogg_int64_t pageGranulepos = ogg_page_granulepos(&page_in);
		packet.bytes = 1;
		packet.packet = &charZero;
		// First Page is header
		sint64 curPageNumber = 0;
		while(curPageNumber < pageNumber - 1) {
			ogg_stream_packetin(&tmpStreamStateOut, &packet);
			while(ogg_stream_flush(&tmpStreamStateOut, &page_out))
				curPageNumber = ogg_page_pageno(&page_out);
		}
		int nbPackets = 0;
		bool bSizeCorrected = false;
		bool bNeedRewrite = false;
		if((curPageNumber==pageNumber-1)
			&& (ogg_stream_pagein(&tmpStreamStateIn, &page_in)==0)
			&& ((nbPackets=ogg_page_packets(&page_in))==1) ) {
			if((ogg_stream_packetout(&tmpStreamStateIn, &packet)==1)
				&& ((*packet.packet & 0x01)==PACKET_TYPE_DATA)) {
				sint32 hdrlen, i;
				uint64 nbSamples = 0;
				hdrlen = ((*packet.packet & PACKET_LEN_BITS01) >> 6)
					| ((*packet.packet & PACKET_LEN_BITS2) << 1);
				unsigned char *pos = (unsigned char *)packet.packet + hdrlen;
				i = hdrlen;
				while(i--) {
					nbSamples <<= 8;
					nbSamples += *pos;
					pos--;
				}
				if(!nbSamples)
					nbSamples = 1;
				if(pageGranulepos + nbSamples > maxGranulepos) {
					bSizeCorrected = true;
					bNeedRewrite = true;
					nbSamples = maxGranulepos - pageGranulepos;
					uint64 nbSamplesTmp = nbSamples;
					pos = (unsigned char *)packet.packet + 1;
					i = hdrlen;
					while(i--) {
						*pos = nbSamplesTmp & 0xFF;
						nbSamplesTmp >>= 8;
						pos++;
					}
				}
				if(!--nbPackets) {
					if(!packet.e_o_s)
						bNeedRewrite = true;
					packet.e_o_s = 1;
				}
				pageGranulepos += nbSamples;
				ogg_stream_packetin(&tmpStreamStateOut, &packet);
				ogg_page newPage;
				if(bNeedRewrite && ogg_stream_flush(&tmpStreamStateOut, &newPage)
					&& !ogg_stream_flush(&tmpStreamStateOut, &page_out)
					&& (ogg_page_pageno(&newPage) == pageNumber)
					&& (newPage.header_len+newPage.body_len==pageLength)) {
					memcpy(lpBuffer, newPage.header, newPage.header_len);
					memcpy((char *)lpBuffer+newPage.header_len, newPage.body, newPage.body_len);
					bPageRewritten = true;
				}
			}
		}
	}

	ogg_stream_clear(&tmpStreamStateIn);
	ogg_stream_clear(&tmpStreamStateOut);
	ogg_sync_clear(&syncState);

	return bPageRewritten;
}