//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/VDString.h>
#include <vd2/system/log.h>
#include <vd2/Dita/resources.h>
#include <vd2/system/text.h>
#include <vd2/system/Error.h>
#include "ProgressDialog.h"
#include "OGMReadHandler.h"
#include "DubSource.h"
#include "OggDS.h"
#include "OggComments.h"
#include "ISO-639.h"

#if 0
#undef VDDEBUG
#define VDDEBUG VDDEBUG2
#endif

namespace {
	enum { kVDST_OGMReadHandler = 153 };
	
	enum {
		kVDM_FileErrors
	};
}

extern CRITICAL_SECTION g_diskcs;
extern bool g_disklockinited;

#define OGG_BLOCKSIZE	65535


OGMReadStream::OGMReadStream(OGMReadNode *readNode) {

	VDASSERT(readNode);

	m_ReadNode = readNode;
	m_ReadNode->AddRef();
	m_ReadNode->RegisterReadStream(this);
	m_bVBRHandling = true;
	m_BlockAlign = 1;

	SeekFirstStreamSegment();

	m_Format = NULL;
	m_FormatSize = 0;
	m_LastFilePos = m_ReadNode->m_LastFilePos;

	if(!(long)m_Stream->m_SampleRate)
		throw MyError("Ogg stream %ld doesn't have a samplerate", m_Stream->m_Serial);

	m_bIsVideo = (MAIN_TYPE(m_Stream->m_StreamType) == SOURCE_VIDEO);
	m_bCanTrimPacket = !m_bIsVideo;

	ogg_sync_init(&m_SyncState);

	m_Length = m_ReadNode->m_End - m_ReadNode->m_Start;

	double duration = (double)m_ReadNode->m_NbSamples / (double)m_Stream->m_SampleRate;

	sint32 byteRate = VDRoundToLong((double)m_ReadNode->m_Bytes / duration);

	if(m_Stream->m_StreamHeader->size != sizeof(stream_header))
		VDDEBUG("Header of Ogg stream %ld is %d bytes longer than stream_header (%d bytes)\n"
		, m_Stream->m_Serial, m_Stream->m_StreamHeader->size - sizeof(stream_header)
		, sizeof(stream_header));

	switch(m_Stream->m_StreamType) {
	case SOURCE_VIDEO:
		{
			VDASSERT(m_Stream->m_StreamHeader->size == sizeof(stream_header));

			m_FormatSize = sizeof(BITMAPINFOHEADER);
			m_Format = malloc(m_FormatSize);
			if(!m_Format)
				throw MyMemoryError();

			memset(m_Format, 0, m_FormatSize);

			BITMAPINFOHEADER *bi = (BITMAPINFOHEADER *)m_Format;
			bi->biBitCount = m_Stream->m_StreamHeader->bits_per_sample;
			memcpy(&bi->biCompression, m_Stream->m_StreamHeader->subtype, 4);
			bi->biHeight = m_Stream->m_StreamHeader->sh.video.height;
			bi->biPlanes = 1;
			bi->biSize = m_FormatSize;
			bi->biWidth = m_Stream->m_StreamHeader->sh.video.width;
		}
		break;

	case SOURCE_AUDIO:
		{
			bool bSetMP3 = false;
			char tag[5];
			memcpy(tag, m_Stream->m_StreamHeader->subtype, 4);
			tag[4] = 0;
			WORD wFormatTag = strtoul(tag, NULL, 16);
			if( (wFormatTag == WAVE_FORMAT_MPEGLAYER3) && (m_Stream->m_StreamHeader->size - sizeof(stream_header) != sizeof(MPEGLAYER3WAVEFORMAT) - sizeof(WAVEFORMATEX)) ) {
				bSetMP3 = true;
				m_FormatSize = sizeof(MPEGLAYER3WAVEFORMAT);
			} else
				m_FormatSize = sizeof(WAVEFORMATEX) + m_Stream->m_StreamHeader->size - sizeof(stream_header);

			m_Format = malloc(m_FormatSize);
			if(!m_Format)
				throw MyMemoryError();
			memset(m_Format, 0, m_FormatSize);

			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;
			wf->cbSize = m_FormatSize - sizeof(WAVEFORMATEX);
			if(wf->cbSize > 0)
				memcpy(wf + 1, m_Stream->m_StreamHeader + 1, wf->cbSize);

			VDASSERT( (double)(byteRate-m_Stream->m_StreamHeader->sh.audio.avgbytespersec) / byteRate < 5);
			VDDEBUG("Ogg stream %ld : %ld kB/s in header, %ld kB/s computed\n"
				, m_Stream->m_Serial
				, m_Stream->m_StreamHeader->sh.audio.avgbytespersec
				, byteRate);

			if( ((double)(byteRate-m_Stream->m_StreamHeader->sh.audio.avgbytespersec) / byteRate < 5)
				&& (m_Stream->m_StreamHeader->sh.audio.avgbytespersec>0))
				wf->nAvgBytesPerSec = m_Stream->m_StreamHeader->sh.audio.avgbytespersec;
			else
				wf->nAvgBytesPerSec = byteRate;
			if(m_ReadNode->m_BlockAlign > 0)
				wf->nBlockAlign = m_BlockAlign = m_ReadNode->m_BlockAlign;
			else
				wf->nBlockAlign = m_BlockAlign = 1;
			wf->nChannels = m_Stream->m_StreamHeader->sh.audio.channels;
			wf->nSamplesPerSec = (long)m_Stream->m_SampleRate;
			wf->wBitsPerSample = m_Stream->m_StreamHeader->bits_per_sample;
			wf->wFormatTag = wFormatTag;
			if(bSetMP3) {
				MPEGLAYER3WAVEFORMAT *mp3format = (MPEGLAYER3WAVEFORMAT *)m_Format;
				mp3format->wID = MPEGLAYER3_ID_MPEG;
				mp3format->fdwFlags = MPEGLAYER3_FLAG_PADDING_ISO;
				mp3format->nFramesPerBlock = 1;
				mp3format->nCodecDelay = /*1393*/0;
				if(wf->nSamplesPerSec < 32000)	// MPEG-2 ?
					mp3format->nBlockSize = (m_ReadNode->m_Bytes * 576) / m_ReadNode->m_NbSamples;
				else	// MPEG-1
					mp3format->nBlockSize = (m_ReadNode->m_Bytes * 1152) / m_ReadNode->m_NbSamples;
				if(mp3format->nBlockSize <= 0)
					mp3format->nBlockSize = 1;
			}
		}
		break;

	case SOURCE_AUDIO_VORBIS:
		{
			m_FormatSize = sizeof(WAVEFORMATEXTENSIBLE);
			m_Format = malloc(m_FormatSize);
			if(!m_Format)
				throw MyMemoryError();

			memset(m_Format, 0, m_FormatSize);

			WAVEFORMATEXTENSIBLE *wfe = (WAVEFORMATEXTENSIBLE *)m_Format;
			wfe->Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
			wfe->Format.cbSize = sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX);
			
			VDASSERT( (double)(byteRate-m_Stream->m_StreamHeader->sh.audio.avgbytespersec) / byteRate < 5);
			VDDEBUG("Ogg Vorbis stream %ld : %ld kB/s in header, %ld kB/s computed\n"
				, m_Stream->m_Serial
				, m_Stream->m_StreamHeader->sh.audio.avgbytespersec
				, byteRate);

			if( ((double)(byteRate-m_Stream->m_StreamHeader->sh.audio.avgbytespersec) / byteRate < 5)
				&& (m_Stream->m_StreamHeader->sh.audio.avgbytespersec>0))
				wfe->Format.nAvgBytesPerSec = m_Stream->m_StreamHeader->sh.audio.avgbytespersec;
			else
				wfe->Format.nAvgBytesPerSec = byteRate;
			if(m_ReadNode->m_BlockAlign > 0)
				wfe->Format.nBlockAlign = m_BlockAlign = m_ReadNode->m_BlockAlign;
			else
				wfe->Format.nBlockAlign = m_BlockAlign = 1;
			wfe->Format.nChannels = m_Stream->m_StreamHeader->sh.audio.channels;
			wfe->Format.nSamplesPerSec = (long)m_Stream->m_SampleRate;
			wfe->Samples.wReserved = 0;
			// Sorry I don't know aything about multiple speakers :p
			// If someone know what should be set here, please tell me
			// (anyway I don't think anything will use those settings afterwards ....)
			// Nando used that : (?)
			//	wfe->dwChannelMask = (1<<vi.channels)-1;
			switch(wfe->Format.nChannels) {
				case 1:
					wfe->dwChannelMask = SPEAKER_FRONT_CENTER;
					break;
				case 2:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT;
					break;
				case 4:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT
						| SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT;
					break;
				case 5:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT
						| SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT
						| SPEAKER_LOW_FREQUENCY;
					break;
				case 6:
					wfe->dwChannelMask = SPEAKER_FRONT_LEFT | SPEAKER_FRONT_RIGHT
						| SPEAKER_BACK_LEFT | SPEAKER_BACK_RIGHT
						| SPEAKER_LOW_FREQUENCY | SPEAKER_FRONT_CENTER;
					break;
				default:
					VDNEVERHERE;
					break;
			}
			wfe->SubFormat = MEDIASUBTYPE_Vorbis;
		}
		break;

	case SOURCE_TEXT:
		{
			VDASSERT(m_Stream->m_StreamHeader->size == sizeof(stream_header));

			m_bCanTrimPacket = true;
			// Let's use this text stream as an audio one :p
			m_FormatSize = sizeof(WAVEFORMATEX);
			m_Format = malloc(m_FormatSize);
			if(!m_Format)
				throw MyMemoryError();

			memset(m_Format, 0, m_FormatSize);

			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;
			wf->cbSize = 0;
			// Since we may need to convert from ASCII to Unicode/UTF8, compensate here
			if(m_Stream->m_MaxPacketSize > byteRate)
				wf->nAvgBytesPerSec = 40 * m_Stream->m_MaxPacketSize;
			else if(byteRate > 0)
				wf->nAvgBytesPerSec = 40 * byteRate;
			else
				wf->nAvgBytesPerSec = 40;
			wf->nBlockAlign = 1;
			wf->nChannels = 1;
			wf->nSamplesPerSec = 1000;
			wf->wBitsPerSample = 0;
			wf->wFormatTag = 0x0000;
		}
		break;

	default:
		VDNEVERHERE;
		break;
	}
}

OGMReadStream::~OGMReadStream() {
	m_ReadNode->UnregisterReadStream(this);
	m_ReadNode->Release();
	if(m_Format)
		delete m_Format;
}

void OGMReadStream::Reinit(void) {
	SeekFirstStreamSegment();

	m_LastFilePos = m_ReadNode->m_LastFilePos;
	m_Length = m_ReadNode->m_End - m_ReadNode->m_Start;

	double duration = (double)m_ReadNode->m_NbSamples / (double)m_Stream->m_SampleRate;

	sint32 byteRate = VDRoundToLong((double)m_ReadNode->m_Bytes / duration);

	switch(m_Stream->m_StreamType) {
	case SOURCE_AUDIO:
	case SOURCE_AUDIO_VORBIS:
		{
			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;
			wf->nAvgBytesPerSec = byteRate;
			if(m_ReadNode->m_BlockAlign > 0)
				wf->nBlockAlign = m_BlockAlign = m_ReadNode->m_BlockAlign;
			else
				wf->nBlockAlign = m_BlockAlign = 1;
			if(wf->wFormatTag == WAVE_FORMAT_MPEGLAYER3) {
				MPEGLAYER3WAVEFORMAT *mp3format = (MPEGLAYER3WAVEFORMAT *)m_Format;
				mp3format->nBlockSize = (m_ReadNode->m_Bytes * 1152) / m_ReadNode->m_NbSamples;
				if(mp3format->nBlockSize <= 0)
					mp3format->nBlockSize = 1;
			}
		}
		break;

	case SOURCE_TEXT:
		{
			WAVEFORMATEX *wf = (WAVEFORMATEX *)m_Format;

			// Since we may need to convert from ASCII to Unicode/UTF8, compensate here
			if(m_Stream->m_MaxPacketSize > byteRate)
				wf->nAvgBytesPerSec = 40 * m_Stream->m_MaxPacketSize;
			else if(byteRate > 0)
				wf->nAvgBytesPerSec = 40 * byteRate;
			else
				wf->nAvgBytesPerSec = 40;
		}
		break;
	}
}

HRESULT OGMReadStream::BeginStreaming(long lStart, long lEnd, long lRate) {
	return 0;
}

HRESULT OGMReadStream::EndStreaming() {
	return 0;
}

VDFraction OGMReadStream::getDAR(void) {
	return VDFraction(0,1);
}

HRESULT OGMReadStream::Info(AVISTREAMINFO *pasi, long lSize) {
	AVISTREAMINFO asi;

	memset(&asi, 0, sizeof asi);

	double duration = (double)m_ReadNode->m_NbSamples / (double)m_Stream->m_SampleRate;

	sint32 byteRate = VDRoundToLong((double)m_ReadNode->m_Bytes / duration);

	asi.fccType				= (m_bIsVideo ? streamtypeVIDEO
		: m_Stream->m_StreamType==SOURCE_TEXT ? streamtypeTEXT
		: streamtypeAUDIO);
	if(m_bIsVideo) {
		memcpy(&asi.fccHandler, m_Stream->m_StreamHeader->subtype, 4);
		// The XviD codec is selected as decompressor when fccHandler is set to DX50
		// but the XviD codec can't decode all versions of DivX (resulting in error -100
		// in VirtualDub(Mod) and no video in the preview windows).
		// So force the 'correct' decoder handle.
		switch(asi.fccHandler) {
		// DivX 4.x/5.x
		case 'xvid':
		case 'XVID':
		case '05xd':
		case '05XD':
			// Use the DivX decoder
			asi.fccHandler = 'xvid';
			break;
		// XviD
		case 'divx':
		case 'DIVX':
			// Use the XviD decoder
			asi.fccHandler = 'divx';
			break;
		}
	}
	asi.dwFlags				= 0;
	asi.wPriority			= 0;
	asi.wLanguage			= 0;
	asi.dwSampleSize		= 0;
	if(m_bIsVideo) {
		asi.dwScale				= m_Stream->m_StreamHeader->time_unit;
		VDASSERT(m_Stream->m_StreamHeader->samples_per_unit == 1);
		asi.dwRate				= 10000000 * m_Stream->m_StreamHeader->samples_per_unit;
	} else if(m_Stream->m_StreamType == SOURCE_TEXT) {
		asi.dwScale				= 1;
		asi.dwRate				= 1000;
	} else {
		if(m_bVBRHandling) {
			asi.dwSampleSize		= 0;
			asi.dwScale				= 1;
			asi.dwRate				= m_Stream->m_StreamHeader->samples_per_unit;
			asi.dwLength			= m_Length;
		} else {
			asi.dwSampleSize		= m_BlockAlign;
			asi.dwScale				= m_BlockAlign;
			asi.dwRate				= byteRate;
			if(m_BlockAlign)
				asi.dwLength			= m_Length / m_BlockAlign;
			else
				asi.dwLength			= m_Length;
		}
	}
	//asi.dwStart				= 0;
	asi.dwStart = m_ReadNode->m_Start;
	asi.dwInitialFrames		= 0;
	// If the max size we found in the stream is greater than the buffersize
	// suggested in the header, take it
	if(m_ReadNode->m_MaxPacketSize > m_Stream->m_StreamHeader->buffersize) {
		VDDEBUG("Suggested buffersize of stream %ld (%d) below max size found (%d)\n"
			, m_Stream->m_Serial, m_Stream->m_StreamHeader->buffersize
			, m_ReadNode->m_MaxPacketSize);
		asi.dwSuggestedBufferSize = m_ReadNode->m_MaxPacketSize;
	} else
		asi.dwSuggestedBufferSize = m_Stream->m_StreamHeader->buffersize;
	asi.dwQuality			= 0xFFFFFFFF;
	if(m_bIsVideo) {
		asi.rcFrame.top			= 0;
		asi.rcFrame.left		= 0;
		asi.rcFrame.right		= m_Stream->m_StreamHeader->sh.video.width;
		asi.rcFrame.bottom		= m_Stream->m_StreamHeader->sh.video.height;
	}

	if (lSize < sizeof asi)
		memcpy(pasi, &asi, lSize);
	else {
		memcpy(pasi, &asi, sizeof asi);
		memset((char *)pasi + sizeof asi, 0, lSize - sizeof asi);
	}

	return 0;
}

void OGMReadStream::SetVBRHandling(bool bVBRHandling) {
	if(m_Stream->m_StreamType == SOURCE_AUDIO_VORBIS)
		bVBRHandling = true;
	m_bVBRHandling = bVBRHandling;
	m_ReadNode->UpdateState(bVBRHandling);
	Reinit();
}

inline void OGMReadStream::SynchSegmentInformation(void) {

	VDASSERT(m_CurrentStreamSegment);
	VDASSERT(m_Stream != m_CurrentStreamSegment->m_OggStream);

	m_Parent = m_CurrentStreamSegment->m_ReadHandler;
	m_Stream = m_CurrentStreamSegment->m_OggStream;
	m_KeyFrames = m_Stream->m_KeyFrames;
	m_KeyFramesEnd = m_Stream->m_KeyFramesEnd;
	m_PageIterator = m_CurrentStreamSegment->m_FirstDataPage;
	m_PageByte = 0;
}

bool OGMReadStream::SeekPreviousSegment(void) {
	OggStreamSegment *previousSegment = m_CurrentStreamSegment->GetPrevious();
	if(previousSegment) {
		m_CurrentStreamSegment = previousSegment;
		SynchSegmentInformation();
		return true;
	} else
		return false;
}

bool OGMReadStream::SeekNextSegment(void) {
	OggStreamSegment *nextSegment = m_CurrentStreamSegment->GetNext();
	if(nextSegment) {
		m_CurrentStreamSegment = nextSegment;
		SynchSegmentInformation();
		return true;
	} else
		return false;
}

void OGMReadStream::SeekFirstStreamSegment(void) {
	m_CurrentStreamSegment = m_ReadNode->m_Segments.GetFirst();
	SynchSegmentInformation();
}

void OGMReadStream::SeekLastStreamSegment(void) {
	m_CurrentStreamSegment = m_ReadNode->m_Segments.GetLast();
	SynchSegmentInformation();
}

long OGMReadStream::FindStreamSegment(long lFrame) {
	if( (lFrame<0) || (lFrame>=m_Length) )
		return false;
	if(lFrame < m_CurrentStreamSegment->m_Start) {
		do {
			m_CurrentStreamSegment = m_CurrentStreamSegment->GetPrevious();

			VDDEBUG("Accessing previous OGM segment\n");
			VDASSERT(m_CurrentStreamSegment);

		} while(lFrame < m_CurrentStreamSegment->m_Start);
		SynchSegmentInformation();
	} else if(lFrame >= m_CurrentStreamSegment->m_End) {
		do {
			m_CurrentStreamSegment = m_CurrentStreamSegment->GetNext();

			VDDEBUG("Accessing next OGM segment\n");
			VDASSERT(m_CurrentStreamSegment);

		} while(lFrame >= m_CurrentStreamSegment->m_End);
		SynchSegmentInformation();
	}

	VDASSERT(m_CurrentStreamSegment && (lFrame>=m_CurrentStreamSegment->m_Start) && (lFrame<m_CurrentStreamSegment->m_End) );

	return (lFrame - m_CurrentStreamSegment->m_Start);
}

bool OGMReadStream::IsKeyFrame(long lFrame) {
	if((lFrame=FindStreamSegment(lFrame))<0)
		return false;
	if(!m_KeyFrames)
		return true;
	// & 7 == % 8
	return !!(m_KeyFrames[lFrame>>3] & (0x80>>(lFrame&7)));
}

sint64 OGMReadStream::sampleOffset(sint64 lStart) {
	if(lStart < 0)
		return 0;
	if(lStart >= m_Length)
		return m_LastFilePos;
	if(!m_bVBRHandling)
		return lStart * m_BlockAlign;

	if((lStart=FindStreamSegment(lStart))<0)
		return 0;

	while(m_PageIterator.previous() && ((*m_PageIterator)->m_Granulepos>lStart) )
		m_PageIterator--;
	while(m_PageIterator.next() && (m_PageIterator.next()->m_Granulepos<=lStart) )
		m_PageIterator++;

	m_Page = *m_PageIterator;
	sint64 pos = OGM_FILEPOS(m_Page->m_FilePos);
	lStart -= m_Page->m_Granulepos;
	if((lStart>0) && (lStart<=m_Page->m_NbPackets))
		while(lStart--)
			pos += m_Page->m_NbBytes[lStart];

	return pos;
}

HRESULT OGMReadStream::Read(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples) {
	if(!m_bVBRHandling && !m_bIsVideo)
		return ReadCBR(lStart, lSamples, lpBuffer, cbBuffer, plBytes, plSamples);

	VDDEBUG("Stream %ld : Read request, start=%ld, samples=%ld, buffer=%08X, size=%ld\n"
		, m_Stream->m_Serial
		, lStart
		, lSamples
		, lpBuffer
		, cbBuffer);

	if((lStart=FindStreamSegment(lStart))<0) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return 0;
	}

	//VDDEBUG("Seeking correct PageIterator");
	// Doing this way we are sure that the current page now contains
	// at least one Packet, and especially contains the Packet we
	// want
	while(m_PageIterator.previous() && ((*m_PageIterator)->m_Granulepos>lStart) )
		m_PageIterator--;
	while(m_PageIterator.next() && (m_PageIterator.next()->m_Granulepos<=lStart) )
		m_PageIterator++;

	m_Page = *m_PageIterator;
	//VDDEBUG("Checking continued flag");
	// m_Page contains the Packet corresponding to lStart. If the granulepos of
	// the Packet following the first one in this Page exceed lStart, and that
	// we can trim Packets, this means we want the first Packet.
	// In this case we must verify if previous pages do not contain this (possibly)
	// spanned packet the 64th bit of m_FilePos tell if this is a continued page
	if( (m_Page->m_Granulepos+m_Page->m_NbSamples[0]>lStart)
		&& (m_bCanTrimPacket || m_bIsVideo)
		&& OGM_GETFLAG_CONT(m_Page->m_FilePos) ) {

		VDASSERT(m_PageIterator.previous());

		m_PageIterator--;
		m_Page = *m_PageIterator;
		while(OGM_GETFLAG_CONT(m_Page->m_FilePos) && (!m_Page->m_NbPackets)) {

			VDASSERT(m_PageIterator.previous());

			m_PageIterator--;
			m_Page = *m_PageIterator;
		}
	}

	int samplesToRead = lSamples;
	if(samplesToRead == AVISTREAMREAD_CONVENIENT)
		samplesToRead = 1;

	sint16 packetsToSkip = 0;
	// Determines how many packets we must skip before reaching the wanted one
	ogg_int64_t granulepos = m_Page->m_Granulepos;
	// For the first sample, the packet may not be the first in the page
	// The problem : Page's granulepos = 0, but first Packet may be header or comment :(
	// Fortunately we prepared ourself for this case :)
	if(!lStart)
		packetsToSkip = m_CurrentStreamSegment->m_FirstPacketIndex;
	//VDDEBUG("Define how many packets to skip");
	if(m_bCanTrimPacket) {
		// We can "cut" the Packet (i.e. send data but lower the number of
		// samples it contains according to where the Packet start and lStart)
		// So we must get the Packet that contains lStart
		while((packetsToSkip<m_Page->m_NbPackets) && (granulepos+m_Page->m_NbSamples[packetsToSkip]<=lStart)) {
			granulepos += m_Page->m_NbSamples[packetsToSkip];
			packetsToSkip++;
		}
	} else {
		// We cannot cut, so we must get the first Packet just after lStart
		while( (granulepos < lStart) && (packetsToSkip<m_Page->m_NbPackets) ) {
			granulepos += m_Page->m_NbSamples[packetsToSkip];
			packetsToSkip++;
		}
	}

	// With a video stream if we couldn't reach the exact granulepos, it means
	// that frame lStart is in the middle of a Packet, i.e. it is a dropped frame
	if( m_bIsVideo && (granulepos != lStart)) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 1;
		return 0;
	}

	int samplesShorten = 0;
	if(m_bCanTrimPacket)
		samplesShorten = lStart - granulepos;

	// Get the total length of data
	sint16 packetsSkipped = packetsToSkip;
	uint8 packetIndex = 0;
	m_TmpIterator = m_PageIterator;
	sint32 nbBytes = 0;
	// Start by skipping unwanted packets
	while(packetsSkipped--) {
		if(packetIndex < (*m_TmpIterator)->m_NbPackets)
			packetIndex++;
		else {
			packetIndex = 0;
			m_TmpIterator++;
		}
	}
	int samplesToRead2 = samplesToRead;
	m_Page = *m_TmpIterator;
	//VDDEBUG("Determine size of data");
	while(samplesToRead2>0) {
		if(packetIndex < m_Page->m_NbPackets) {
			nbBytes += m_Page->m_NbBytes[packetIndex];
			samplesToRead2 -= m_Page->m_NbSamples[packetIndex];
			// One Packet at a time
			if(m_Page->m_NbSamples[packetIndex])
				break;
			packetIndex++;
		} else {
			packetIndex = 0;
			if(m_TmpIterator.next()) {
				m_TmpIterator++;
				m_Page = *m_TmpIterator;
			}
			while(!m_Page->m_NbPackets && m_TmpIterator.next()) {
				m_TmpIterator++;
				m_Page = *m_TmpIterator;
			}
			if(!m_Page->m_NbPackets) {
				//VDNEVERHERE;
				break;
			}
		}
	}

	// If we wanted to know the size of the data or if the buffer is too small
	if(!lpBuffer || (cbBuffer < nbBytes) ) {
		if(plBytes)
			*plBytes = nbBytes;
		if(plSamples) {
			// This is a video stream, and VDub asked us to read the convenient number of frames ...
			// We may have found dropped frames, but VDub do not have to know that yet (it wiil
			// know that later anyway), so we say we read 1 sample (which is true)
			if( m_bIsVideo && (lSamples==AVISTREAMREAD_CONVENIENT) )
				*plSamples = 1;
			else
				//samplesToRead-samplesToRead2 give exactly how many samples we read
				*plSamples = (samplesToRead-samplesToRead2) - samplesShorten;
		}
		if(!lpBuffer)
			return AVIERR_OK;
		return AVIERR_BUFFERTOOSMALL;
	}

	m_Page = *m_PageIterator;

	// If we must skip packets, verify that the first one will be
	// outputed (it is not the case if this packet is a spanned one)
	packetIndex = 0;
	if(packetsToSkip && OGM_GETFLAG_CONT(m_Page->m_FilePos)) {
		packetsToSkip--;
		packetIndex = 1;
	}
	// Now we must read the samples, let's start
	// First reset the ogg_sync_state (because we will seek)
	ogg_sync_reset(&m_SyncState);
	ogg_stream_reset(&m_CurrentStreamSegment->m_StreamState);
	// Set position in file
	m_Parent->SeekFile(OGM_FILEPOS(m_Page->m_FilePos));

	VDDEBUG("Stream %ld : Reading data at byte %I64u (granulepos %I64d, %d packets in Page), skipping %d packets\n"
		, m_Stream->m_Serial
		, OGM_FILEPOS(m_Page->m_FilePos)
		, m_Page->m_Granulepos
		, (int)m_Page->m_NbPackets
		, packetsToSkip);

	long bytesRead = 0;
	int samplesRead = 0;
	char *lpBufferPos = (char *)lpBuffer;
	int hdrlen;
	int iResult;
	while(true) {
		if(!_getNextPage(&m_OggPage, m_Page->m_Size)) {
			//VDNEVERHERE;
			VDDEBUG("ERROR: Couldn't reconstruct a new Page");
			break;
		}

		// If this page belongs to another stream, skip it
		if(ogg_page_serialno(&m_OggPage) != m_Stream->m_Serial)
			continue;

		if(ogg_stream_pagein(&m_CurrentStreamSegment->m_StreamState, &m_OggPage))
			throw MyError("ogg_stream_pagein failed");

		//while(ogg_stream_packetout(&m_CurrentStreamSegment->m_StreamState, &m_OggPacket)==1) {
		while((iResult=ogg_stream_packetout(&m_CurrentStreamSegment->m_StreamState, &m_OggPacket))!=0) {
			if(iResult==-1)
				continue;
			// If we have packets to skip ...
			if(samplesRead >= samplesToRead)
				break;
			else if(packetsToSkip)
				packetsToSkip--;
			else {
				//VDDEBUG("Found the data to fetch packet %ld/%ld", packetIndex+1, m_Page->m_NbPackets);
				if(m_Stream->m_StreamType == SOURCE_AUDIO_VORBIS) {
					memcpy(lpBufferPos, m_OggPacket.packet, m_OggPacket.bytes);
					samplesRead += m_Page->m_NbSamples[packetIndex];
					bytesRead += m_OggPacket.bytes;
					// Only one packet at a time
					break;
				} else {
					// We should believe the information gathered before
					// But who knows ... at least verify lenght of data
					hdrlen = ((*m_OggPacket.packet & PACKET_LEN_BITS01) >> 6)
						| ((*m_OggPacket.packet & PACKET_LEN_BITS2) << 1);
					memcpy(lpBufferPos, m_OggPacket.packet+1+hdrlen, m_OggPacket.bytes-1-hdrlen);
					samplesRead += m_Page->m_NbSamples[packetIndex];
					bytesRead += (m_OggPacket.bytes-1-hdrlen);
					// Only one packet at a time
					//lpBufferPos += (packet.bytes-1-hdrlen);
					break;
				}
			}
			packetIndex++;
		}

		// Only one packet at a time
		if(samplesRead)
			break;
		/*if(samplesRead >= samplesToRead)
			break;*/

		// 2 cases when corrupted data have been found :
		//  - a continued Packet could be rebuilt, thus all the intermediate Pages
		//    (which should have expose Packets) have the 'Continued' flag
		//  - no continued Packet could be rebuilt, and the flag isn't set
		// In any case we never reach directly one of those intermediate Pages here.
		if(OGM_GETFLAG_RESETED(m_Page->m_FilePos) && m_Page->m_NbPackets) {
			// If we are on the Page following the intermdiate ones (where all retained
			// Packets have been flushed) we either want to read
			//  - one Packet of the intermediate Pages
			//    So we musn't get the next Page in our list nor reset packetIndex to 0
			// or
			//  - the continued Packet of the next Page
			//    We must get the next Page in our list and reset packetIndex to 0
			//    when we finished with the intermediate Pages (i.e. packetsToSkip==0)
			//VDDEBUG("Processing next RESETED data");
			if((packetIndex + packetsToSkip >= m_Page->m_NbPackets) && !packetsToSkip) {

				VDASSERT(m_PageIterator.next());

				if(!m_PageIterator.next())
					break;
				m_PageIterator++;
				m_Page = *m_PageIterator;
				packetIndex = 0;
			}
		} else {
			// Just in case : verify we haven't reached the end of the stream
			// (maybe some stupid guy asked us more samples than what remains after lStart ;))
			VDASSERT(m_PageIterator.next());

			if(!m_PageIterator.next())
				break;
			m_PageIterator++;
			m_Page = *m_PageIterator;
			// If this Page appears to be an intermediate Page this means we want to read
			// the continued Packet that could be rebuilt after the corrupted data
			//VDDEBUG("Processing current RESETED data");
			while(OGM_GETFLAG_RESETED(m_Page->m_FilePos) && !m_Page->m_NbPackets && m_PageIterator.next()) {
				// There shouldn't be any packets to skip
				VDASSERT(!packetsToSkip);

				m_PageIterator++;
				m_Page = *m_PageIterator;
			}
			packetIndex = 0;
		}
	}

	if(plBytes)
		*plBytes = bytesRead;
	if(plSamples)
		*plSamples = samplesRead - samplesShorten;

	VDASSERT(samplesRead - samplesShorten > 0);

	//VDDEBUG("All done :)");

	return AVIERR_OK;
}

HRESULT OGMReadStream::ReadCBR(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples) {
	if((lStart=FindStreamSegment(lStart*m_BlockAlign))<0) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return 0;
	}

	m_Page = *m_PageIterator;
	uint8 count;
	sint64 nextPageByte;
	if(m_PageByte > lStart) {
		while(m_PageIterator.previous() && (m_PageByte>lStart) ) {
			m_PageIterator--;
			m_Page = *m_PageIterator;
			if(count = m_Page->m_NbPackets) {
				while(count--)
					m_PageByte -= m_Page->m_NbBytes[count];
			}
		}
	} else {
		nextPageByte = m_PageByte;
		do {
			if(count = m_Page->m_NbPackets) {
				while(count--)
					nextPageByte += m_Page->m_NbBytes[count];
			}
			if(!m_PageIterator.next() || (nextPageByte>lStart))
				break;
			m_PageIterator++;
			m_Page = *m_PageIterator;
			m_PageByte = nextPageByte;
		} while(true);
	}

	if( (m_PageByte+m_Page->m_NbBytes[0]>lStart)
		&& OGM_GETFLAG_CONT(m_Page->m_FilePos) ) {

		VDASSERT(m_PageIterator.previous());

		m_PageIterator--;
		m_Page = *m_PageIterator;
		if(count = m_Page->m_NbPackets) {
			while(count--)
				m_PageByte -= m_Page->m_NbBytes[count];
		}
		while(OGM_GETFLAG_CONT(m_Page->m_FilePos) && (!m_Page->m_NbPackets)) {

			VDASSERT(m_PageIterator.previous());

			m_PageIterator--;
			m_Page = *m_PageIterator;
			if(count = m_Page->m_NbPackets) {
				while(count--)
					m_PageByte -= m_Page->m_NbBytes[count];
			}
		}
	}

	sint16 packetsToSkip = 0;
	ogg_int64_t bytePos = m_PageByte;
	if(!lStart)
		packetsToSkip = m_CurrentStreamSegment->m_FirstPacketIndex;
	while((packetsToSkip<m_Page->m_NbPackets) && (bytePos+m_Page->m_NbBytes[packetsToSkip]<=lStart)) {
		bytePos += m_Page->m_NbBytes[packetsToSkip];
		packetsToSkip++;
	}

	long bytesShorten = lStart - bytePos;

	VDASSERT((bytesShorten >= 0) && !(bytesShorten % m_BlockAlign));

	int bytesToRead = lSamples * m_BlockAlign;

	sint16 packetsSkipped = packetsToSkip;
	uint8 packetIndex = 0;
	m_TmpIterator = m_PageIterator;
	sint32 nbBytes = 0;
	while(packetsSkipped--) {
		if(packetIndex < (*m_TmpIterator)->m_NbPackets)
			packetIndex++;
		else {
			packetIndex = 0;
			m_TmpIterator++;
		}
	}
	int bytesToRead2 = bytesToRead;
	m_Page = *m_TmpIterator;
	if(lSamples == AVISTREAMREAD_CONVENIENT) {
		bytesToRead2 = 0;
		while(true) {
			if(packetIndex < m_Page->m_NbPackets) {
				nbBytes += m_Page->m_NbBytes[packetIndex];
				bytesToRead = m_Page->m_NbBytes[packetIndex];
				packetIndex++;
				break;
			} else {
				packetIndex = 0;
				if(m_TmpIterator.next()) {
					m_TmpIterator++;
					m_Page = *m_TmpIterator;
				}
				while(!m_Page->m_NbPackets && m_TmpIterator.next()) {
					m_TmpIterator++;
					m_Page = *m_TmpIterator;
				}
				if(!m_Page->m_NbPackets)
					break;
			}
		}
	} else {
		while(bytesToRead2>0) {
			if(packetIndex < m_Page->m_NbPackets) {
				long iTmpBytes = m_Page->m_NbBytes[packetIndex];
				if(iTmpBytes > bytesToRead2)
					iTmpBytes = bytesToRead2;
				nbBytes += iTmpBytes;
				bytesToRead2 -= iTmpBytes;
				packetIndex++;
			} else {
				packetIndex = 0;
				if(m_TmpIterator.next()) {
					m_TmpIterator++;
					m_Page = *m_TmpIterator;
				}
				while(!m_Page->m_NbPackets && m_TmpIterator.next()) {
					m_TmpIterator++;
					m_Page = *m_TmpIterator;
				}
				if(!m_Page->m_NbPackets)
					break;
			}
		}
	}

	if(!lpBuffer || (cbBuffer < nbBytes) ) {
		if(plBytes)
			*plBytes = nbBytes;
		if(plSamples)
			*plSamples = VDRoundToLong(double((bytesToRead-bytesToRead2) - bytesShorten) / m_BlockAlign);
		if(!lpBuffer)
			return AVIERR_OK;
		return AVIERR_BUFFERTOOSMALL;
	}

	m_Page = *m_PageIterator;

	packetIndex = 0;
	if(packetsToSkip && OGM_GETFLAG_CONT(m_Page->m_FilePos)) {
		packetsToSkip--;
		packetIndex = 1;
	}
	ogg_sync_reset(&m_SyncState);
	ogg_stream_reset(&m_CurrentStreamSegment->m_StreamState);
	m_Parent->SeekFile(OGM_FILEPOS(m_Page->m_FilePos));

	long bytesLeft = bytesToRead;

	VDDEBUG("Stream %ld : Reading %ld bytes starting at byte %I64u+%ld (granulepos %I64d, %d packets in Page), skipping %d packets\n"
		, m_Stream->m_Serial
		, bytesLeft
		, OGM_FILEPOS(m_Page->m_FilePos)
		, bytesShorten
		, m_Page->m_Granulepos
		, (int)m_Page->m_NbPackets
		, packetsToSkip);

	long bytesRead = 0;
	long toRead;
	char *lpBufferPos = (char *)lpBuffer;
	int hdrlen;
	int iResult;
	while(true) {
		if(!_getNextPage(&m_OggPage, m_Page->m_Size))
			break;

		if(ogg_page_serialno(&m_OggPage) != m_Stream->m_Serial)
			continue;

		if(ogg_stream_pagein(&m_CurrentStreamSegment->m_StreamState, &m_OggPage))
			throw MyError("ogg_stream_pagein failed");

		//while(ogg_stream_packetout(&m_CurrentStreamSegment->m_StreamState, &m_OggPacket)==1) {
		while((iResult=ogg_stream_packetout(&m_CurrentStreamSegment->m_StreamState, &m_OggPacket))!=0) {
			if(iResult==-1)
				continue;
			if(bytesLeft <= 0)
				break;
			else if(packetsToSkip)
				packetsToSkip--;
			else {
				if(m_Stream->m_StreamType == SOURCE_AUDIO_VORBIS) {
					// We shouldn't be here !!!
					VDASSERT(false);

					toRead = m_OggPacket.bytes;
					if(toRead > bytesLeft)
						toRead = bytesLeft;
					
					memcpy(lpBufferPos, m_OggPacket.packet+bytesShorten, toRead);
					
					bytesShorten = 0;
					bytesRead += toRead;
					lpBufferPos += toRead;
					bytesLeft -= toRead;
				} else {
					hdrlen = ((*m_OggPacket.packet & PACKET_LEN_BITS01) >> 6)
						| ((*m_OggPacket.packet & PACKET_LEN_BITS2) << 1);

					toRead = m_OggPacket.bytes-1-hdrlen-bytesShorten;
					if(toRead > bytesLeft)
						toRead = bytesLeft;

					VDASSERT(toRead > 0);
					
					memcpy(lpBufferPos, m_OggPacket.packet+1+hdrlen+bytesShorten, toRead);

					bytesShorten = 0;
					bytesRead += toRead;
					lpBufferPos += toRead;
					bytesLeft -= toRead;
				}
			}
			packetIndex++;
		}

		if(bytesLeft <= 0)
			break;

		if(!m_PageIterator.next()) {
			if(!SeekNextSegment())
				break;
		} else {
			if(count = m_Page->m_NbPackets) {
				while(count--)
					m_PageByte += m_Page->m_NbBytes[count];
			}
			m_PageIterator++;
		}

		m_Page = *m_PageIterator;
		packetIndex = 0;
	}

	VDASSERT(bytesLeft == 0);

	if(plBytes)
		*plBytes = bytesRead;
	if(plSamples) {
		*plSamples = VDRoundToLong(double(bytesRead - bytesShorten) / m_BlockAlign);

		VDASSERT(*plSamples > 0);
	}

	return AVIERR_OK;
}

long OGMReadStream::Start() {
	return 0;
}

long OGMReadStream::End() {
	return m_Length;
}

long OGMReadStream::PrevKeyFrame(long lFrame) {
	if(lFrame <= 0)
		return -1;
	if(lFrame >= m_Length) {
		SeekLastStreamSegment();
		if(!m_KeyFrames)
			return m_Length-1;
		lFrame = m_CurrentStreamSegment->m_End - m_CurrentStreamSegment->m_Start;
	} else if((lFrame=FindStreamSegment(lFrame))<0)
		return -1;

	if(!m_KeyFrames)
		return (m_CurrentStreamSegment->m_Start+lFrame-1);

	do {
		uint8 *keys = m_KeyFrames + (lFrame>>3);
		sint8 offset = lFrame & 7;
		offset--;
		while(keys>=m_KeyFrames) {
			while(*keys && (offset>=0)) {
				if(((*keys)<<offset)&0x80)
					return (m_CurrentStreamSegment->m_Start + ((keys-m_KeyFrames)<<3) + offset);
				offset--;
			}
			offset = 7;
			keys--;
		}
		if(!SeekPreviousSegment())
			break;
		lFrame = m_CurrentStreamSegment->m_End;
	} while(true);

	return -1;
}

long OGMReadStream::NextKeyFrame(long lFrame) {
	if(lFrame >= m_Length)
		return -1;
	if(lFrame < 0) {
		return m_ReadNode->m_FirstKeyFrame;
		/*SeekFirstStreamSegment();
		if((*m_KeyFrames) & 0x80)
			return 0;
		lFrame = 0;*/
	} else if((lFrame=FindStreamSegment(lFrame))<0)
		return -1;

	if(!m_KeyFrames)
		return (m_CurrentStreamSegment->m_Start+lFrame+1);

	do {
		uint8 *keys = m_KeyFrames + (lFrame>>3);
		sint8 offset = lFrame & 7;
		offset++;
		while(keys < m_KeyFramesEnd) {
			while(*keys && (offset<=7)) {
				if(((*keys)<<offset)&0x80)
					return (m_CurrentStreamSegment->m_Start + ((keys-m_KeyFrames)<<3) + offset);
				offset++;
			}
			offset = 0;
			keys++;
		}
		if(!SeekNextSegment())
			return -1;
		if((*m_KeyFrames) & 0x80)
			return m_CurrentStreamSegment->m_Start;
		lFrame = 0;
	} while(true);
	return -1;
}

// This function must return the nearest (previous) key-frame (and not the absolute
// nearest one)
long OGMReadStream::NearestKeyFrame(long lFrame) {
	if(lFrame <= 0) {
		if(m_ReadNode->m_FirstKeyFrame)
			return -1;
		return 0;
	}
	if(IsKeyFrame(lFrame))
		return lFrame;

	long lprev = PrevKeyFrame(lFrame);

	if(lprev < 0)
		return -1;
	else
		return lprev;
}

HRESULT OGMReadStream::FormatSize(long lFrame, long *plSize) {
	if(plSize)
		*plSize = m_FormatSize;
	return 0;
}

HRESULT OGMReadStream::ReadFormat(long lFrame, void *pFormat, long *plSize) {
	if(!pFormat)
		*plSize = m_FormatSize;

	if(*plSize < m_FormatSize)
		memcpy(pFormat, m_Format, *plSize);
	else {
		memcpy(pFormat, m_Format, m_FormatSize);
		*plSize = m_FormatSize;
	}

	return 0;
}

bool OGMReadStream::isStreaming() {
	return false;
}

bool OGMReadStream::isKeyframeOnly() {
	return m_ReadNode->m_bKeyFrameOnly;
}

bool OGMReadStream::getVBRInfo(double& bitrate_mean, double& bitrate_stddev, double& maxdev) {
	// No there are no VBR stream in this file :p
	return false;
}

void OGMReadStream::fixVBRAudio(bool fix, double _bitrate_mean) {
}

sint32 OGMReadStream::_getNextPage(ogg_page *page, uint16 size) {
	sint32 np, nread;
	char *buffer;
	while(true) {
		np = ogg_sync_pageseek(&m_SyncState, page);
		if(np>0)
			return np;
		else if(np==0) {
			buffer = ogg_sync_buffer(&m_SyncState, size);
			if(!buffer)
				throw MyError("ogg_sync_buffer failed");
			if((nread = m_Parent->ReadData(buffer, size)) <= 0)
				return 0;
			ogg_sync_wrote(&m_SyncState, nread);
#ifndef _DEBUG
		}
#else
		} else
			VDDEBUG("ogg_sync_pageseek failed, %d bytes skipped\n", -np);
#endif
	}
	return np;
}

sint32 OGMReadStream::samplesLag(sint64 lStart) {
	if(!m_bVBRHandling)
		return 0;
	if((lStart=FindStreamSegment(lStart))<0)
		return 0;
	if(m_Stream->m_StreamType==SOURCE_TEXT) {
		// For text stream, if lStart!=0, return 1, so that
		// first Packet start @ granulepos 1
		// This make the text allways shown by OggDS (not allways
		// the case if granulepos = 0)
		if(lStart)
			return 1;
		else
			return 0;
	}

	while(m_PageIterator.next() && (m_PageIterator.next()->m_Granulepos<lStart) )
		m_PageIterator++;
	while(m_PageIterator.previous() && ((*m_PageIterator)->m_Granulepos>lStart) )
		m_PageIterator--;

	m_Page = *m_PageIterator;
	if( (m_Page->m_Granulepos+m_Page->m_NbSamples[0]>lStart)
		&& (m_bCanTrimPacket || m_bIsVideo)
		&& OGM_GETFLAG_CONT(m_Page->m_FilePos) ) {

		VDASSERT(m_PageIterator.previous());

		m_PageIterator--;
		m_Page = *m_PageIterator;
		while(OGM_GETFLAG_CONT(m_Page->m_FilePos) && (!m_Page->m_NbPackets)) {

			VDASSERT(m_PageIterator.previous());

			m_PageIterator--;
			m_Page = *m_PageIterator;
		}
	}

	sint16 packetsToSkip = 0;
	ogg_int64_t granulepos = m_Page->m_Granulepos;
	if(!lStart)
		packetsToSkip = m_CurrentStreamSegment->m_FirstPacketIndex;
	if(m_bCanTrimPacket) {
		while((packetsToSkip<m_Page->m_NbPackets) && (granulepos+m_Page->m_NbSamples[packetsToSkip]<=lStart)) {
			granulepos += m_Page->m_NbSamples[packetsToSkip];
			packetsToSkip++;
		}
	} else {
		while( (granulepos < lStart) && (packetsToSkip<m_Page->m_NbPackets) ) {
			granulepos += m_Page->m_NbSamples[packetsToSkip];
			packetsToSkip++;
		}
	}

	//granulepos = beginning of the Packet (that we would have read)
	// If lStart == gpos, there is no lag
	// If not we would have skipped gpos-lStart samples
	return granulepos-lStart;
}

sint32 OGMReadStream::sourceOffsetMS(void) {
	return m_Stream->m_StartOffsetMS;
}

void OGMReadStream::GetCommentList(ModList<ModStreamComment>& list) {
	OggStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	while(streamNode) {
		streamNode->m_OggStream->GetCommentList(list);
		streamNode = streamNode->GetNext();
	}
	ModList<ModStreamComment> copyList;
	copyList = list;
	list.DeleteContents(true);
	list.Clear();
	ModList<ModStreamComment>::iterator srcIt(copyList);
	ModList<ModStreamComment>::iterator dstIt;
	while(srcIt) {
		bool bAddComment = true;
		dstIt.begin(list);
		while(dstIt) {
			if((**srcIt) == (**dstIt)) {
				bAddComment = false;
				break;
			}
			dstIt++;
		}
		if(bAddComment) {
			ModStreamComment *newComment = list.AddLast();
			*newComment = *(*srcIt);
		}
		srcIt++;
	}
	copyList.DeleteContents(true);
	copyList.Clear();
}

void OGMReadStream::GetChapterList(ModChapterList& list) {
	OggStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	while(streamNode) {
		streamNode->m_OggStream->GetChapterList(list, (unsigned long)streamNode->m_OggStream->m_SampleRate.scale64ir(1000 * streamNode->m_Start));
		streamNode = streamNode->GetNext();
	}
	ModChapterList copyList;
	copyList = list;
	list.clear();
	ModChapterList::iterator srcIt;
	ModChapterList::iterator dstIt;
	for(srcIt=copyList.begin() ; srcIt!=copyList.end() ; srcIt++) {
		ModChapterList::tChapterNode& srcChapter = (*srcIt);
		bool bAddChapter = true;
		for(dstIt=list.begin() ; dstIt!=list.end() ; dstIt++) {
			ModChapterList::tChapterNode& dstChapter = (*dstIt);
			if( (srcChapter==dstChapter) || (srcChapter.SeemsSimilar(dstChapter)) ) {
				bAddChapter = false;
				break;
			}
		}
		if(bAddChapter)
			list.AddChapter(srcChapter);
	}
	list.Sort();
	copyList.clear();
}

uint32 OGMReadStream::GetOggStreamType(void) {
	OggStreamSegment *firstNode = m_ReadNode->m_Segments.GetFirst();
	return firstNode->m_OggStream->m_StreamType;
}

uint16 OGMReadStream::GetNbHeaderPackets(void) {
	OggStreamSegment *firstNode = m_ReadNode->m_Segments.GetFirst();
	return firstNode->m_OggStream->m_HeaderPacketList.size();
}

ogg_packet *OGMReadStream::GetHeaderPacket(uint16 index) {
	OggStreamSegment *firstNode = m_ReadNode->m_Segments.GetFirst();
	if(index >= firstNode->m_OggStream->m_HeaderPacketList.size())
		return NULL;
	return firstNode->m_OggStream->m_HeaderPacketList[index];
}

long OGMReadStream::GetStreamIndexByType(void) {
	OggStreamSegment *firstNode = m_ReadNode->m_Segments.GetFirst();
	return firstNode->m_ReadHandler->GetStreamIndexByType(m_ReadNode);
}

VDStringW OGMReadStream::writeReport(void) {
	SeekFirstStreamSegment();
	VDStringW report;
	bool bWriteStreamHeader = true;
	bool bWriteVorbisInfo = false;
	switch(m_Stream->m_StreamType) {
	case SOURCE_VIDEO:
		report = L"Video stream";
		break;
	case SOURCE_TEXT:
		report = L"Text stream";
		break;
	case SOURCE_AUDIO:
		report = L"Audio stream";
		break;
	case SOURCE_AUDIO_VORBIS:
		report = L"Vorbis stream";
		bWriteStreamHeader = false;
		bWriteVorbisInfo = true;
		break;
	default:
		report = L"Other stream";
		break;
	}
	if(bWriteStreamHeader) {
		char tmpBuf[9];
		VDStringW tmpStr;
		report += VDStringW::setf(L"\r\n Header Size :    %ld bytes", m_Stream->m_StreamHeader->size);
		memset(tmpBuf, 0, sizeof(tmpBuf));
		memcpy(tmpBuf, m_Stream->m_StreamHeader->streamtype, 8);
		tmpStr = VDTextAToW(tmpBuf, -1);
		report += VDStringW::setf(L"\r\n Type :           [%s]", tmpStr.c_str());
		memset(tmpBuf, 0, sizeof(tmpBuf));
		memcpy(tmpBuf, m_Stream->m_StreamHeader->subtype, 4);
		tmpStr = VDTextAToW(tmpBuf, -1);
		report += VDStringW::setf(L"\r\n SubType :        [%s]", tmpStr.c_str());
		report += VDStringW::setf(L"\r\n BufferSize :     %ld bytes", m_Stream->m_StreamHeader->buffersize);
		report += VDStringW::setf(L"\r\n Time Unit :      %I64d", m_Stream->m_StreamHeader->time_unit);
		report += VDStringW::setf(L"\r\n SamplesPerUnit : %I64d", m_Stream->m_StreamHeader->samples_per_unit);
		report += VDStringW::setf(L"\r\n Default Len :    %ld", m_Stream->m_StreamHeader->default_len);
		report += VDStringW::setf(L"\r\n BitsPerSample :  %ld", (sint32)m_Stream->m_StreamHeader->bits_per_sample);
		switch(MAIN_TYPE(m_Stream->m_StreamType)) {
		case SOURCE_VIDEO:
			report += VDStringW::setf(L"\r\n Width :          %ld pixels", m_Stream->m_StreamHeader->sh.video.width);
			report += VDStringW::setf(L"\r\n Height :         %ld pixels", m_Stream->m_StreamHeader->sh.video.height);
			report += VDStringW::setf(L"\r\n SampleRate :     %.3f", (double)m_Stream->m_SampleRate);
			break;
		case SOURCE_AUDIO:
			report += VDStringW::setf(L"\r\n Channels :       %ld", (sint32)m_Stream->m_StreamHeader->sh.audio.channels);
			report += VDStringW::setf(L"\r\n Byterate :       %ld Bytes Per Second", m_Stream->m_StreamHeader->sh.audio.avgbytespersec);
			report += VDStringW::setf(L"\r\n BlockAlign :     %ld bytes", (sint32)m_Stream->m_StreamHeader->sh.audio.blockalign);
		default:
			report += VDStringW::setf(L"\r\n SampleRate :     %ld", (long)m_Stream->m_SampleRate);
			break;
		}
	}
	if(bWriteVorbisInfo) {
		report += L"\r\nVorbis information :";
		report += VDStringW::setf(L"\r\n Version :         %ld", m_Stream->m_VorbisInfo.version);
		report += VDStringW::setf(L"\r\n Channels :        %ld", m_Stream->m_VorbisInfo.channels);
		report += VDStringW::setf(L"\r\n SampleRate :      %ld", m_Stream->m_VorbisInfo.rate);
		report += VDStringW::setf(L"\r\n BitRate Lower :   %ld bps", m_Stream->m_VorbisInfo.bitrate_lower);
		report += VDStringW::setf(L"\r\n BitRate Nominal : %ld bps", m_Stream->m_VorbisInfo.bitrate_nominal);
		report += VDStringW::setf(L"\r\n BitRate Upper :   %ld bps", m_Stream->m_VorbisInfo.bitrate_upper);
		report += VDStringW::setf(L"\r\n BitRate Window :  %ld", m_Stream->m_VorbisInfo.bitrate_window);
	}
	OggStreamSegment *streamNode = m_ReadNode->m_Segments.GetFirst();
	while(streamNode) {
		report += VDStringW::setf(L"\r\nStream report for file \"%s\"", streamNode->m_ReadHandler->GetFileName().c_str());
		report += streamNode->m_OggStream->writeReport();
		streamNode = streamNode->GetNext();
	}
	return report;
}

//////////////////////////////////////////////////////////////////////

OggStream::OggStream(void) {
	m_RefCount = 1;
	m_Serial = -1;
	m_SampleRate = VDFraction(0);
	m_Header = NULL;
	m_StreamHeader = NULL;
	vorbis_info_init(&m_VorbisInfo);
	vorbis_comment_init(&m_VorbisComment);
	m_bHasComments = false;
	m_PreviousW = 0;
	m_PacketIndex = 0;
	m_MaxPacketSize = 0;
	m_StartOffsetMS = 0;
	m_FirstGranulepos = 0;
	m_FirstGranuleposPage = 0;
	m_Bytes = 0;
	m_NbPackets = 0;
	m_NbSamples = 0;
	m_bKeyIndexBuilt = false;
	m_KeyFrames = NULL;
	m_KeyFramesEnd = NULL;
	m_TmpPageGranulepos_0 = 0;
	m_TmpPageGranulepos_1 = 0;
	m_CheckPage = 0;
	m_PacketsExpected = 0;
	m_SplittedPages = 0;
}

OggStream::~OggStream(void) {
	ogg_stream_clear(&m_StreamState);
	m_PageList.Clear();
	if(m_Header)
		delete m_Header;
	if(m_KeyFrames)
		delete[] m_KeyFrames;
	while(!m_HeaderPacketList.empty()) {
		ogg_packet *packet = m_HeaderPacketList.back();
		m_HeaderPacketList.pop_back();
		if(packet->packet) {
			delete[] packet->packet;
			packet->packet = NULL;
		}
		delete packet;
	}
	m_WeirdPageList.Clear();
	m_WeirdPacketList.Clear();
}

void OggStream::AddRef(void) {
	m_RefCount++;
}

void OggStream::Release(void) {
	if(!--m_RefCount)
		delete this;
}

void OggStream::SetHeaderLen(uint16 size) {
	size = std::max<uint16>(size, sizeof(stream_header));
	m_Header = malloc(size);
	m_StreamHeader = (stream_header *)m_Header;
	memset(m_Header, 0, size);
	m_StreamHeader->size = size;
}

void OggStream::AddHeaderPacket(ogg_packet *packet) {
	ogg_packet *new_packet = new ogg_packet;
	*new_packet = *packet;
	if(packet->bytes > 0) {
		new_packet->packet = new unsigned char[packet->bytes];
		memcpy(new_packet->packet, packet->packet, packet->bytes);
	} else {
		new_packet->bytes = 0;
		new_packet->packet = NULL;
	}
	m_HeaderPacketList.push_back(new_packet);
}

void OggStream::AddWeirdPage(uint64 filePos, uint64 timePosMS, uint8 nbPacketsExpected, uint16 nbPacketsReceived) {
	if(nbPacketsExpected == nbPacketsReceived)
		return;
	OggWeirdPage *lastPage = m_WeirdPageList.GetLast();
	if(lastPage && (lastPage->m_FilePos==filePos) )
		lastPage->m_NbPacketsReceived = nbPacketsReceived;
	else {
		lastPage = m_WeirdPageList.AddLast();
		lastPage->m_FilePos = filePos;
		lastPage->m_NbPacketsExpected = nbPacketsExpected;
		lastPage->m_NbPacketsReceived = nbPacketsReceived;
		lastPage->m_TimePosMS = timePosMS;
	}
}

void OggStream::SetLastWeirdPageLostPackets(uint16 nbPacketsLostInAction) {
	if(!nbPacketsLostInAction)
		return;
	OggWeirdPage *lastPage = m_WeirdPageList.GetLast();
	if(!lastPage)
		return;
	lastPage->m_NbPacketsLostInAction = nbPacketsLostInAction;
}

void OggStream::AddWeirdPacket(uint64 filePos, uint64 timePosMS, uint8 packetIndex, uint64 nbSamplesAnnounced, uint32 nbSamplesCorrected) {
	OggWeirdPacket *lastPacket = m_WeirdPacketList.AddLast();
	lastPacket->m_FilePos = filePos;
	lastPacket->m_TimePosMS = timePosMS;
	lastPacket->m_PacketIndex = packetIndex;
	lastPacket->m_NbSamplesAnnounced = nbSamplesAnnounced;
	lastPacket->m_NbSamplesCorrected = nbSamplesCorrected;
}

VDStringW OggStream::writeReport(void) {
	VDStringW report;
	ModCompactList<OggWeirdPage>::iterator weirdPageIt(m_WeirdPageList);
	OggWeirdPage *aWeirdPage;
	ModCompactList<OggWeirdPacket>::iterator weirdPacketIt(m_WeirdPacketList);
	OggWeirdPacket *aWeirdPacket;
	if(weirdPacketIt || weirdPageIt)
		report += L"\r\nSome problems have been encountered while parsing the stream :";
	else
		report += L"\r\nNo apparent problem have been encountered while parsing the stream.";
	uint64 timeMS;
	sint32 nHour, nMinute, nSecond, nMillisecond;
	while(weirdPacketIt) {
		aWeirdPacket = *weirdPacketIt;
		if( OGM_GETFLAG_CONT(aWeirdPacket->m_FilePos)
		&& (aWeirdPacket->m_NbSamplesAnnounced == -1)
		&& (aWeirdPacket->m_NbSamplesCorrected == -1) ) {
			// Special case : a header (stream header, comment, or codec setup)
			// has been defined a second time in the middle of the stream
			timeMS = aWeirdPacket->m_TimePosMS;
			nMillisecond = timeMS % 1000;
			nSecond = (timeMS / 1000) % 60;
			nMinute = (timeMS / 1000 / 60) % 60;
			nHour = (timeMS / 1000 / 60 / 60);
			report += VDStringW::setf(L"\r\nPacket %d in Page at byte %I64u (%02d:%02d:%02d.%03d) appears to be a header that should have already been defined.\r\nPacket skipped."
				, (int)aWeirdPacket->m_PacketIndex, OGM_FILEPOS(aWeirdPacket->m_FilePos)
				, nHour, nMinute, nSecond, nMillisecond);
		} else {
			while(weirdPageIt && (OGM_FILEPOS((*weirdPageIt)->m_FilePos) <= OGM_FILEPOS(aWeirdPacket->m_FilePos))) {
				aWeirdPage = *weirdPageIt;
				timeMS = aWeirdPage->m_TimePosMS;
				nMillisecond = timeMS % 1000;
				nSecond = (timeMS / 1000) % 60;
				nMinute = (timeMS / 1000 / 60) % 60;
				nHour = (timeMS / 1000 / 60 / 60);
				report += VDStringW::setf(L"\r\nPage at byte %I64u (%02d:%02d:%02d.%03d) exposed %d Packets instead of %d expected."
					, OGM_FILEPOS(aWeirdPage->m_FilePos), nHour, nMinute, nSecond, nMillisecond
					, (int)aWeirdPage->m_NbPacketsReceived, (int)aWeirdPage->m_NbPacketsExpected);
				if(aWeirdPage->m_NbPacketsLostInAction)
					report += VDStringW::setf(L"\r\nSeems like %d Packets have been lost in action."
						,(int)aWeirdPage->m_NbPacketsLostInAction);
				weirdPageIt++;
			}
			timeMS = aWeirdPacket->m_TimePosMS;
			nMillisecond = timeMS % 1000;
			nSecond = (timeMS / 1000) % 60;
			nMinute = (timeMS / 1000 / 60) % 60;
			nHour = (timeMS / 1000 / 60 / 60);
			report += VDStringW::setf(L"\r\nInformation about Packet %d in Page at byte %I64u (%02d:%02d:%02d.%03d) has been corrected.\r\nNow containing %u samples instead of %I64u."
				, (int)aWeirdPacket->m_PacketIndex, OGM_FILEPOS(aWeirdPacket->m_FilePos)
				, nHour, nMinute, nSecond, nMillisecond
				, aWeirdPacket->m_NbSamplesCorrected, aWeirdPacket->m_NbSamplesAnnounced);
		}
		weirdPacketIt++;
	}
	while(weirdPageIt) {
		aWeirdPage = *weirdPageIt;
		timeMS = aWeirdPage->m_TimePosMS;
		nMillisecond = timeMS % 1000;
		nSecond = (timeMS / 1000) % 60;
		nMinute = (timeMS / 1000 / 60) % 60;
		nHour = (timeMS / 1000 / 60 / 60);
		report += VDStringW::setf(L"\r\nPage at byte %I64u (%02d:%02d:%02d.%03d) exposed %d Packets instead of %d expected."
			, OGM_FILEPOS(aWeirdPage->m_FilePos), nHour, nMinute, nSecond, nMillisecond
			, (int)aWeirdPage->m_NbPacketsReceived, (int)aWeirdPage->m_NbPacketsExpected);
		if(aWeirdPage->m_NbPacketsLostInAction)
			report += VDStringW::setf(L"\r\nSeems like %d Packets have been lost in action."
				,aWeirdPage->m_NbPacketsLostInAction);
		weirdPageIt++;
	}
	return report;
}

void OggStream::GetCommentList(ModList<ModStreamComment>& list) {
	int i;
	for(i=0 ; i<m_VorbisComment.comments ; i++) {
		char *curComment = m_VorbisComment.user_comments[i];
		if(!curComment)
			continue;
		int length = strlen(curComment);
		if(length <= 0)
			continue;
		if(!strnicmp(curComment, "CHAPTER", 7) && (
			( (length>=10) && isdigit(curComment[7]) && isdigit(curComment[8]) && (curComment[9]=='=') )
			|| ( (length>=14) && isdigit(curComment[7]) && isdigit(curComment[8]) && !strnicmp(curComment+9, "NAME=", 5) )
			) )
			continue;
		char *equalSign = curComment;
		while( (*equalSign) && (*equalSign != '=')) {
			equalSign++;
			VDASSERT(equalSign-curComment<=m_VorbisComment.comment_lengths[i]);
		}
		ModStreamComment *newModComment = list.AddLast();
		char *tmpBuffer = new char[length+1];
		if(!tmpBuffer)
			throw MyMemoryError();
		memset(tmpBuffer, 0, length+1);
		newModComment->m_Type = cGeneric;
		if(*equalSign == '=') {
			memcpy(tmpBuffer, curComment, equalSign-curComment);
			newModComment->m_Type = cSpecific;
			GetOggCommentType(VDFastTextAToW(tmpBuffer), newModComment->m_Type);
			if(GetOggCommentTag(newModComment->m_Type))
				newModComment->m_Tag = GetOggCommentTag(newModComment->m_Type);
			else
				newModComment->m_Tag = VDFastTextAToW(tmpBuffer);
			equalSign++;
		} else
			equalSign = curComment;
		bool bValidUTF8 = true;
		newModComment->m_Value = VDTextU8ToW(equalSign, -1, bValidUTF8);
		if(!bValidUTF8)
			newModComment->m_Value = VDTextAToW(equalSign);
		if(newModComment->m_Type == cLanguage) {
			sint32 langIndex = 0;
			VDStringW tmpValue(newModComment->m_Value.c_str());
			wchar_t *startBr = NULL;
			wchar_t *pos = (wchar_t *)tmpValue.c_str();
			while(*pos) {
				if((*pos) == L'[') {
					startBr = pos;
					break;
				}
				pos++;
			}
			if(startBr) {
				*startBr = 0;
				startBr++;
				pos = startBr;
				while(*pos) {
					if((*pos) == L']') {
						*pos = 0;
						break;
					}
					pos++;
				}
				langIndex = ISO639_GetIndex_2(startBr) + 1;
				if(!langIndex)
					langIndex = ISO639_GetIndex_1(startBr) + 1;
			}
			if(!langIndex)
				langIndex = ISO639_GetIndex(tmpValue.c_str()) + 1;
			newModComment->m_Type = (eCommentType)(cLanguage + langIndex);
		}
		delete[] tmpBuffer;
	}
}

 void OggStream::GetChapterList(ModChapterList& list, uint32 offsetMS) {
	int i;
	for(i=0 ; i<m_VorbisComment.comments ; i++) {
		char *curComment = m_VorbisComment.user_comments[i];
		if(!curComment)
			continue;
		int length = strlen(curComment);
		if(length <= 0)
			continue;
		if(!strnicmp(curComment, "CHAPTER", 7)) {
			if( (length>=10) && isdigit(curComment[7]) && isdigit(curComment[8]) && (curComment[9]=='=') ) {
				char *equalSign = curComment + 9;
				char *tmpBuffer = new char[length+1];
				if(!tmpBuffer)
					throw MyMemoryError();
				memset(tmpBuffer, 0, length+1);
				memcpy(tmpBuffer, curComment, 9);
				_snprintf(tmpBuffer+9, length-9, "NAME=");
				int j;
				char *chapterName;
				bool bNameFound = false;
				for(j=0 ; j<m_VorbisComment.comments ; j++) {
					chapterName = m_VorbisComment.user_comments[j];
					if(!chapterName)
						continue;
					int length2 = strlen(chapterName);
					if(length2 <= 0)
						continue;
					if(!strnicmp(chapterName, tmpBuffer, 14)) {
						chapterName += 14;
						bNameFound = true;
						break;
					}
				}
				if(bNameFound) {
					ModChapterList::tChapterNode *newChapter = new ModChapterList::tChapterNode();
					if(newChapter->SetTime(equalSign+1)) {
						newChapter->SetTimeMS(uint32(newChapter->GetTimeMS()) + offsetMS);
						bool bValidUTF8 = true;
						newChapter->SetName(VDTextU8ToW(chapterName, -1, bValidUTF8).c_str());
						if(!bValidUTF8)
							newChapter->SetName(VDTextAToW(chapterName).c_str());
						list.AddChapter(*newChapter);
					}
					delete newChapter;
				}
				delete[] tmpBuffer;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////

OggStreamSegment::OggStreamSegment(void) {
	m_ReadHandler	= NULL;
	m_OggStream		= NULL;
	m_FirstPacketIndex = 0;
	m_Start			= 0;
	m_End			= 0;
	m_LastFilePos	= 0;
}

OggStreamSegment::~OggStreamSegment(void) {
	if(m_OggStream)
		m_OggStream->Release();
	ogg_stream_clear(&m_StreamState);
}

void OggStreamSegment::UpdateState(bool bVBRHandling) {
	m_Start = 0;
	bool bIsVideo = (MAIN_TYPE(m_OggStream->m_StreamType) == SOURCE_VIDEO);

	m_LastFilePos =
		(m_OggStream && m_OggStream->m_PageList.GetCount() ? OGM_FILEPOS(m_OggStream->m_PageList.GetLast()->m_FilePos) : 0);

	if(ogg_stream_init(&m_StreamState, m_OggStream->m_Serial))
		throw MyError("Could not initialize the ogg_stream_state");

	// Find the first Page containing data
	m_FirstDataPage.begin(m_OggStream->m_PageList);
	OggPage *page = (*m_FirstDataPage);
	VDASSERT(page);
	if(!page)
		return;
	sint32 nbPackets_1 = page->m_NbPackets;
	sint32 nbPackets_2 = 0;
	sint32 pageNum = 0;
	VDASSERT(m_FirstDataPage.next());
	while((nbPackets_1<=m_OggStream->m_HeaderPacketList.size()) && m_FirstDataPage.next()) {
		nbPackets_2 += page->m_NbPackets;
		m_FirstDataPage++;
		pageNum++;
		page = (*m_FirstDataPage);
		nbPackets_1 += page->m_NbPackets;

		VDASSERT(m_FirstDataPage.next());
	}

	// Oops the first data packet may not be the first packet in the Page :(
	m_FirstPacketIndex = m_OggStream->m_HeaderPacketList.size() - nbPackets_2;

	ModCompactList<OggPage>::iterator m_PageIterator;
	m_PageIterator = m_FirstDataPage;

	if(nbPackets_1 <= m_OggStream->m_HeaderPacketList.size())
		throw MyError("Invalid Ogg Media Stream");

	// Try to compute an eventual offset for the stream
	sint64 granulepos = m_OggStream->m_FirstGranulepos;
	if(bIsVideo)
		granulepos++;
	// For text streams, a negative first granulepos == offset
	// same if the granulepos is the one of the first Page containing data
	if( (m_OggStream->m_StreamType == SOURCE_TEXT) && ((granulepos < 0) || (m_OggStream->m_FirstGranuleposPage == pageNum)) )
		m_OggStream->m_StartOffsetMS = m_OggStream->m_SampleRate.scale64ir((sint64)1000 * granulepos);
	else {
		for(int i=m_FirstPacketIndex ; i<page->m_NbPackets ; i++)
			granulepos -= page->m_NbSamples[i];
		// Try to deal with some special cases where first non-0 granulepos isn't declared
		// in the first Page containing data
		// Doesn't apply to text streams
		if(m_OggStream->m_StreamType != SOURCE_TEXT) {
			while(m_OggStream->m_FirstGranuleposPage > pageNum) {
				VDASSERT(m_PageIterator.next());
				m_PageIterator++;
				page = *m_PageIterator;
				pageNum++;
				for(int i=0 ; i<page->m_NbPackets ; i++)
					granulepos -= page->m_NbSamples[i];
			}
		}
		if(m_OggStream->m_StreamType == SOURCE_AUDIO_VORBIS) {
			// For Vorbis streams, accept 32, 64, 128, 256, 512 and 1024 as non offsets
			sint64 iGPos = granulepos;
			if(iGPos < 0)
				iGPos = -iGPos;
			switch(iGPos) {
			case 32:
			case 64:
			case 128:
			case 256:
			case 512:
			case 1024:
				break;
			default:
				m_OggStream->m_StartOffsetMS = m_OggStream->m_SampleRate.scale64ir((sint64)1000 * granulepos);
			}
		} else
			m_OggStream->m_StartOffsetMS = m_OggStream->m_SampleRate.scale64ir((sint64)1000 * granulepos);
	}

	if(bVBRHandling)
		m_End = m_OggStream->m_NbSamples;
	else
		m_End = m_OggStream->m_Bytes;


	// Build a compact key index :)
	if(bIsVideo && !m_OggStream->m_bKeyIndexBuilt) {
		uint32 nbChars = (m_OggStream->m_NbSamples-m_Start+7)>>3;
		m_OggStream->m_KeyFrames = new uint8[nbChars];
		if(!m_OggStream->m_KeyFrames)
			throw MyMemoryError();

		memset(m_OggStream->m_KeyFrames, 0, nbChars);

		bool onlyKeys = true;
		uint8 *curPos = m_OggStream->m_KeyFrames;
		m_OggStream->m_KeyFramesEnd = m_OggStream->m_KeyFrames + nbChars;
		uint8 counter = 8;
		uint8 keys = 0;
		m_PageIterator = m_FirstDataPage;
		while(m_PageIterator) {
			page = *m_PageIterator;
			for(uint8 packetNum=0 ; packetNum<page->m_NbPackets ; packetNum++) {
				// Take care of dropped frames
				uint32 nbSamples = page->m_NbSamples[packetNum];
#ifdef _DEBUG
				VDASSERT(nbSamples > 0);
				if(nbSamples > 1)
					VDDEBUG("Found %ld dropped frames after frame %ld\n"
					, nbSamples - 1, ((curPos-m_OggStream->m_KeyFrames)>>3)+(8-counter));
#endif
				while(nbSamples--) {
					keys <<= 1;
					if(page->m_NbBytes[packetNum] >> 30) {
						keys |= 1;
						// Reset the flag
						page->m_NbBytes[packetNum] &= 0x7FFFFFFF;
					} else if(onlyKeys)
						onlyKeys = false;
					if(!--counter) {
						*curPos = keys;
						keys = 0;
						curPos++;
						counter = 8;
						VDASSERT( (curPos<m_OggStream->m_KeyFramesEnd)
							|| ((packetNum==page->m_NbPackets-1) && !m_PageIterator.next()));
					}
				}
			}
			m_PageIterator++;
		}
		if(counter < 8) {
			keys <<= counter;
			*curPos = keys;
			VDASSERT( (curPos==m_OggStream->m_KeyFramesEnd-1) && (counter==8-((m_End-m_Start)%8)) );
		} else
			VDASSERT(curPos==m_OggStream->m_KeyFramesEnd);
		if(onlyKeys) {
			VDDEBUG("Video stream only contains keyframes\n");
			delete[] m_OggStream->m_KeyFrames;
			m_OggStream->m_KeyFrames = NULL;
			m_OggStream->m_KeyFramesEnd = NULL;
		}
		m_OggStream->m_bKeyIndexBuilt = true;
	}
}

//////////////////////////////////////////////////////////////////////

OGMReadNode::OGMReadNode(void) {
	m_Start			= 0;
	m_End			= 0;
	m_LastFilePos	= 0;
	m_NbSamples		= 0;
	m_Bytes			= 0;
	m_FirstKeyFrame	= 0;
	m_BlockAlign	= 1;
	m_MaxPacketSize	= 0;
	m_bKeyFrameOnly	= false;
}

OGMReadNode::~OGMReadNode(void) {
	m_Segments.DeleteContents(true);
	m_Segments.Clear();
}

void OGMReadNode::UpdateState(bool bVBRHandling) {
	m_Start			= 0;
	m_End			= 0;
	m_LastFilePos	= 0;
	m_NbSamples		= 0;
	m_Bytes			= 0;
	m_FirstKeyFrame	= 0;
	m_BlockAlign	= 0;
	m_MaxPacketSize	= 0;
	m_bKeyFrameOnly	= !!m_Segments.GetCount();
	bool bHaveSameBlockAlign = true;
	ModList<OggStreamSegment>::iterator it(m_Segments);
	OggStreamSegment *curSegment = NULL;
	while(it) {
		curSegment = *it;
		curSegment->UpdateState(bVBRHandling);

		if(MAIN_TYPE(curSegment->m_OggStream->m_StreamType) == SOURCE_AUDIO) {
			if(!m_BlockAlign && (curSegment->m_OggStream->m_StreamHeader->sh.audio.blockalign>0))
				m_BlockAlign = curSegment->m_OggStream->m_StreamHeader->sh.audio.blockalign;
			else if( (!curSegment->m_OggStream->m_StreamHeader->sh.audio.blockalign)
			|| (m_BlockAlign != curSegment->m_OggStream->m_StreamHeader->sh.audio.blockalign) )
				bHaveSameBlockAlign = false;
		}

		sint64 segmentLength = curSegment->m_End - curSegment->m_Start;

		curSegment->m_Start = m_End;
		curSegment->m_End =
		m_End = curSegment->m_Start + segmentLength;
		m_LastFilePos += curSegment->m_LastFilePos;
		m_NbSamples += curSegment->m_OggStream->m_NbSamples;
		m_Bytes += curSegment->m_OggStream->m_Bytes;
		if(curSegment->m_OggStream->m_MaxPacketSize > m_MaxPacketSize)
			m_MaxPacketSize = curSegment->m_OggStream->m_MaxPacketSize;

		if(curSegment->m_OggStream->m_KeyFrames)
			m_bKeyFrameOnly = false;

		it++;
	}

	if(!bHaveSameBlockAlign)
		m_BlockAlign = 1;

	curSegment = m_Segments.GetFirst();
	bool bFound = false;
	while(curSegment && curSegment->m_OggStream->m_KeyFrames && !bFound) {
		if((*curSegment->m_OggStream->m_KeyFrames) & 0x80) {
			bFound = true;
			break;
		}
		uint8 *keys = curSegment->m_OggStream->m_KeyFrames;
		sint8 offset = 0;
		offset++;
		while(!bFound && (keys < curSegment->m_OggStream->m_KeyFramesEnd) ) {
			while(*keys && (offset<=7)) {
				if(((*keys)<<offset)&0x80) {
					m_FirstKeyFrame = (curSegment->m_Start + ((keys-curSegment->m_OggStream->m_KeyFrames)<<3) + offset);
					bFound = true;
					break;
				}
				offset++;
			}
			offset = 0;
			keys++;
		}
		if(bFound)
			break;
		if(!(curSegment = curSegment->GetNext()))
			break;
		if((*curSegment->m_OggStream->m_KeyFrames) & 0x80) {
			m_FirstKeyFrame = curSegment->m_Start;
			bFound = true;
			break;
		}
	}
}

void OGMReadNode::AppendSegment(OggStreamSegment *segment) {
	if(!segment)
		return;

	OggStreamSegment *newSegment = m_Segments.AddLast();

	newSegment->m_ReadHandler = segment->m_ReadHandler;
	newSegment->m_OggStream = segment->m_OggStream;
	newSegment->m_Start = m_End;
	newSegment->m_End =
	m_End = newSegment->m_Start + (segment->m_End - segment->m_Start);
	newSegment->m_FirstPacketIndex = segment->m_FirstPacketIndex;
	newSegment->m_LastFilePos = segment->m_LastFilePos;
	//m_LastFilePos += newSegment->m_LastFilePos;
	//m_NbSamples += newSegment->m_OggStream->m_NbSamples;
	//m_Bytes += newSegment->m_OggStream->m_Bytes;
	if(ogg_stream_init(&newSegment->m_StreamState, newSegment->m_OggStream->m_Serial))
		throw MyError("Could not initialize the ogg_stream_state");
	newSegment->m_FirstDataPage = segment->m_FirstDataPage;
	newSegment->m_ReadHandler->AddRef();
	newSegment->m_OggStream->AddRef();

	UpdateState(true);
}

void OGMReadNode::AddRef(void) {
	ModList<OggStreamSegment>::iterator it(m_Segments);
	while(it) {
		(*it)->m_ReadHandler->AddRef();
		it++;
	}
}

void OGMReadNode::Release(void) {
	// One of the read handlers own this list
	// Better not process the list while releasing the read handler
	// that may clear it ^^'
	std::list<OGMReadHandler *> tmpList;
	ModList<OggStreamSegment>::iterator it(m_Segments);
	while(it) {
		tmpList.push_back((*it)->m_ReadHandler);
		it++;
	}
	std::list<OGMReadHandler *>::const_iterator listIt;
	OGMReadHandler *readHandler;
	while(!tmpList.empty()) {
		readHandler = tmpList.front();
		readHandler->Release();
		tmpList.pop_front();
	}
}

void OGMReadNode::RegisterReadStream(OGMReadStream *readStream) {
	std::list<OGMReadStream *>::const_iterator it;
	for(it=m_RegisteredReadStreams.begin() ; it!=m_RegisteredReadStreams.end() ; it++) {
		if((*it) == readStream)
			return;
	}
	m_RegisteredReadStreams.push_back(readStream);
}

void OGMReadNode::UnregisterReadStream(OGMReadStream *readStream) {
	m_RegisteredReadStreams.remove(readStream);
}

void OGMReadNode::ReinitReadStreams(void) {
	std::list<OGMReadStream *>::const_iterator it;
	for(it=m_RegisteredReadStreams.begin() ; it!=m_RegisteredReadStreams.end() ; it++)
		(*it)->Reinit();
}

//////////////////////////////////////////////////////////////////////

OGMReadHandler::OGMReadHandler(const wchar_t *pszFile) {
	m_RefCount = 1;

	if (!g_disklockinited) {
		g_disklockinited=true;
		InitializeCriticalSection(&g_diskcs);
	}

	try {
		_construct(pszFile);
	} catch(...) {
		_destruct();
		throw;
	}
}

OGMReadHandler::~OGMReadHandler(void) {
	_destruct();
}

void OGMReadHandler::AddRef(void) {
	m_RefCount++;
}

void OGMReadHandler::Release(void) {
	if(!--m_RefCount)
		delete this;
}

VDStringW OGMReadHandler::GetFileName(void) {
	return m_FileName;
}

void OGMReadHandler::_destruct(void) {
	m_File.close();
	m_StreamList.DeleteContents(true);
	m_StreamList.Clear();
}

void OGMReadHandler::_updateStreamStates(void) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		//OggStreamSegment *segment = (*m_StreamIterator)->m_Segments.GetFirst();
		//segment->UpdateState();
		//(*m_StreamIterator)->m_End = segment->m_End - segment->m_Start;
		//(*m_StreamIterator)->m_LastFilePos = segment->m_LastFilePos;
		//(*m_StreamIterator)->m_NbSamples = segment->m_OggStream->m_NbSamples;
		//(*m_StreamIterator)->m_Bytes = segment->m_OggStream->m_Bytes;
		(*m_StreamIterator)->UpdateState(true);
		m_StreamIterator++;
	}
}

void OGMReadHandler::_construct(const wchar_t *pszFile) {
	m_CurrentStream = NULL;
	m_BytesRead = 0;
	m_BytesSkipped = 0;

	m_FileName = pszFile;

	if(!m_File.open(pszFile))
		throw MyWin32Error("Couldn't open %s: %%s", GetLastError(), VDTextWToA(pszFile).c_str());

	_parseFile();

	_updateStreamStates();

	// Warn when errors were found
	bool bWarnErrors = false;
	if(m_CorruptedDataList.GetCount())
		bWarnErrors = true;
	else {
		m_StreamIterator.begin(m_StreamList);
		while(m_StreamIterator) {
			OGMReadNode *readNode = *m_StreamIterator;
			ModList<OggStreamSegment>::iterator segmentIt(readNode->m_Segments);
			while(segmentIt) {
				OggStream *stream = (*segmentIt)->m_OggStream;
				segmentIt++;
				if(!stream)
					continue;
				if(stream->m_WeirdPageList.GetCount() || stream->m_WeirdPacketList.GetCount()) {
					bWarnErrors = true;
					break;
				}
			}
			if(bWarnErrors)
				break;
			m_StreamIterator++;
		}
	}
	if(bWarnErrors)
		VDLogAppMessage(kVDLogError, kVDST_OGMReadHandler, kVDM_FileErrors);
}

void OGMReadHandler::_addStream(ogg_page *page) {
	VDASSERT(ogg_page_pageno(page)==0);

	sint32 serial = ogg_page_serialno(page);

	OGMReadNode *lastNode = m_StreamList.AddLast();
	OggStreamSegment *lastSegment = lastNode->m_Segments.AddLast();
	lastSegment->m_ReadHandler = this;
	m_CurrentStream = lastSegment->m_OggStream = new OggStream();

	m_CurrentStream->m_FirstGranulepos = 0;
	m_CurrentStream->m_FirstGranuleposPage = 0;

	// This is a new stream, create its first Page now
	OggPage *lastPage = m_CurrentStream->m_PageList.AddLast();

	// Beginning => granulepos = 0;
	lastPage->m_Granulepos = 0;
	lastPage->m_FilePos = m_BytesRead;
	
	VDASSERT(!ogg_page_continued(page));
	
	//if(ogg_page_continued(page))
	//	lastPage->m_FilePos |= OGM_FLAG_CONT;
	lastPage->m_Size = page->header_len + page->body_len;
	lastPage->SetNbPackets(ogg_page_packets(page));
	m_BytesRead += page->header_len + page->body_len;
	m_CurrentStream->m_PacketIndex = 0;

	m_CurrentStream->m_Serial = ogg_page_serialno(page);

	if(ogg_stream_init(&m_CurrentStream->m_StreamState, m_CurrentStream->m_Serial))
		throw MyError("Could not initialize the ogg_stream_state");

	if(ogg_stream_pagein(&m_CurrentStream->m_StreamState, page))
		throw MyError("ogg_stream_pagein failed");

	ogg_packet packet;
	int iResult;
	//while(ogg_stream_packetout(&m_CurrentStream->m_StreamState, &packet)==1) {
	while((iResult=ogg_stream_packetout(&m_CurrentStream->m_StreamState, &packet))!=0) {
		if(iResult==-1)
			continue;
		if(packet.packetno==0) {
			// This is the header packet
			if((packet.bytes >= 7) && !_strnicmp((const char *)&packet.packet[1], "vorbis", 6)) {
				// Vorbis stream
				m_CurrentStream->m_StreamType = SOURCE_AUDIO_VORBIS;
				vorbis_info_init(&m_CurrentStream->m_VorbisInfo);
				vorbis_comment_init(&m_CurrentStream->m_VorbisComment);
				m_CurrentStream->SetHeaderLen(sizeof(stream_header));
				m_CurrentStream->m_StreamHeader->time_unit = 10000000;
				if(vorbis_synthesis_headerin(&m_CurrentStream->m_VorbisInfo, &m_CurrentStream->m_VorbisComment, &packet) < 0) {
					VDDEBUG("Invalid Vorbis stream\n");
					m_CurrentStream->m_StreamType = SOURCE_UNKNOWN;
					return;
				}
				m_CurrentStream->m_StreamHeader->samples_per_unit = m_CurrentStream->m_VorbisInfo.rate;
				m_CurrentStream->m_StreamHeader->default_len = 1;
				m_CurrentStream->m_StreamHeader->buffersize = -1;
				m_CurrentStream->m_StreamHeader->bits_per_sample = m_CurrentStream->m_VorbisInfo.bitrate_nominal / m_CurrentStream->m_VorbisInfo.rate;
				m_CurrentStream->m_StreamHeader->sh.audio.channels = m_CurrentStream->m_VorbisInfo.channels;
				m_CurrentStream->m_StreamHeader->sh.audio.blockalign = 1;
				m_CurrentStream->m_StreamHeader->sh.audio.avgbytespersec = m_CurrentStream->m_VorbisInfo.bitrate_nominal / 8;
				m_CurrentStream->m_SampleRate = VDFraction(m_CurrentStream->m_VorbisInfo.rate);
			} else if( (packet.bytes >= 142)
				&& !_strnicmp((const char *)&packet.packet[1],"Direct Show Samples embedded in Ogg", 35) ) {
				//VDNEVERHERE;
				m_CurrentStream->m_StreamType = SOURCE_UNKNOWN;
				return;
			} else if( (*packet.packet & PACKET_TYPE_BITS ) == PACKET_TYPE_HEADER) {
				
				VDASSERT(packet.bytes >= sizeof(stream_header) + 1);
				
				stream_header *header = (stream_header *)&packet.packet[1];
				
				VDASSERT(header->size <= packet.bytes-1);
				
				if(header->size != packet.bytes-1) {
					//VDNEVERHERE;
					VDDEBUG("Ogg Stream header (%d bytes) doesn't fill the Packet (%d bytes)\n"
					, header->size, packet.bytes-1);
				}
				long headerSize = std::min<long>(header->size, packet.bytes-1);
				m_CurrentStream->SetHeaderLen(headerSize);
				memcpy(m_CurrentStream->m_Header, &packet.packet[1], headerSize);

				VDASSERT(m_CurrentStream->m_StreamHeader->time_unit);
				
				if(m_CurrentStream->m_StreamHeader->time_unit)
					m_CurrentStream->m_SampleRate = VDFraction::reduce64((sint64)10000000 * m_CurrentStream->m_StreamHeader->samples_per_unit, m_CurrentStream->m_StreamHeader->time_unit);
				// What type of stream ?
				if(!_strnicmp(header->streamtype, "audio", 6))
					m_CurrentStream->m_StreamType = SOURCE_AUDIO;
				else if(!_strnicmp(header->streamtype, "video", 6))
					m_CurrentStream->m_StreamType = SOURCE_VIDEO;
				else if(!_strnicmp(header->streamtype, "text", 5))
					m_CurrentStream->m_StreamType = SOURCE_TEXT;
				else {
					//VDNEVERHERE;
					m_CurrentStream->m_StreamType = SOURCE_UNKNOWN;
					return;
				}
			} else {
				VDDEBUG("Invalid ogg header\n");
				//VDNEVERHERE;
				m_CurrentStream->m_StreamType = SOURCE_UNKNOWN;
				return;
			}
			if(m_CurrentStream->m_StreamType == SOURCE_AUDIO_VORBIS)
				m_CurrentStream->m_bHasComments = true;
			m_CurrentStream->AddHeaderPacket(&packet);
		}
		_addPacket(&packet);
	}
}

OggStream *OGMReadHandler::_getStream(sint32 serial) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		OggStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_OggStream;
		if(curStream->m_Serial == serial)
			return curStream;
		m_StreamIterator++;
	}
	return NULL;
}

void OGMReadHandler::_addPage(ogg_page *page) {
	VDASSERT(!ogg_page_version(page));
	if(ogg_page_bos(page)) {
		_addStream(page);
		return;
	}

	m_CurrentStream = _getStream(ogg_page_serialno(page));
	if(!m_CurrentStream) {
		VDDEBUG("Stream %d not handled\n", ogg_page_serialno(page));
		m_BytesRead += page->header_len + page->body_len;
		return;
	}

	if(!m_CurrentStream->m_FirstGranulepos) {
		m_CurrentStream->m_FirstGranulepos = ogg_page_granulepos(page);
		m_CurrentStream->m_FirstGranuleposPage = ogg_page_pageno(page);
	}

	if(m_CurrentStream->m_CheckPage) {
		// Uh oh there were corrupted data in the stream before this Page
		OggPage *aPage = m_CurrentStream->m_PageList.GetLast();
		int nbPackets = aPage->m_NbPackets;
		if(nbPackets && ogg_page_continued(page) && (nbPackets-1 == m_CurrentStream->m_PacketIndex)) {
			//VDDEBUG("One continued packet lost");
			aPage->m_NbPackets--;
		} else if(nbPackets && (m_CurrentStream->m_PacketIndex==0)) {
			//VDDEBUG("RESETING kept data");
			// We have a serious problem here
			// It seems that sometimes, due to corrupted data, the ogg library
			// keep Packets and burst them on the next Page :/
			// Example :
			//  Corrupted data
			//  Page : header say 18 packets, but none goes out :/
			//  Page : header say 16 packets, but 34 (18+16) goes out :/
			// What to do : reset the number of Packets to 0 here and treat
			// this problem when the expected Packets come (i.e. next Page)
			m_CurrentStream->m_PacketsExpected += nbPackets;
			uint64 timePosMS = 0;
			_getCurrentTimeMS(false, timePosMS);
			m_CurrentStream->AddWeirdPage(OGM_FILEPOS(aPage->m_FilePos), timePosMS, aPage->m_NbPackets, 0);
			aPage->ResetNbPackets();
			aPage->m_FilePos |= OGM_FLAG_RESETED;
			if(m_CurrentStream->m_CheckPage >= 2)
				m_CurrentStream->m_CheckPage++;
		} else if(m_CurrentStream->m_CheckPage >= 2) {
			// We processed the Page coming after the corrupted data
			// We also saved the granulepos of the Page just before
			// the corrupted part, and have the one coming after
			// Time to check nothing went too wrong in-between
			ModCompactList<OggPage>::iterator pageIt;
			pageIt.end(m_CurrentStream->m_PageList);
			OggPage *lastPage = *pageIt;
			sint64 addedGranulepos = 0;
			int i = lastPage->m_NbPackets;
			// If this page doesn't contain the first Packet rebuild after having
			// passed the corrupted data, then just wait a bit more :p
			if(i) {
				int j = m_CurrentStream->m_CheckPage - 2;
				if(j) {
					// If inbetween Pages have been reseted, fix their file position
					// so that for all packets coming out later we actually start
					// reading on the first reseted Page (otherwise we would miss
					// the retained Packets)
					int h = 0;
					while(j && pageIt.previous() && OGM_GETFLAG_RESETED(pageIt.previous()->m_FilePos)) {
						j--;
						h++;
						pageIt--;
					}
					if(h) {
						uint64 fixedFilePos = (*pageIt)->m_FilePos;
						// If packets have been lost, this should mean that if the first Packet
						// was a continued one then you won't get it, so reset the flag
						fixedFilePos = OGM_FILEPOS(fixedFilePos)
							| ((OGM_GETFLAG_CONT(fixedFilePos) && !m_CurrentStream->m_PacketsExpected) ? OGM_FLAG_CONT : 0);
						while(h && pageIt.next()) {
							h--;
							pageIt++;
							(*pageIt)->m_FilePos = fixedFilePos | OGM_FLAG_RESETED;
						}
					}
				}
				// If we splitted data over various OggPages (because we got more
				// than 255 Packets in a Page) go to the one that received the first
				// Packets
				while(m_CurrentStream->m_SplittedPages && pageIt.previous()) {
					i = lastPage->m_NbPackets;
					while(i--)
						addedGranulepos += lastPage->m_NbSamples[i];
					pageIt--;
					lastPage = *pageIt;
					m_CurrentStream->m_SplittedPages--;
				}
				m_CurrentStream->m_SplittedPages = 0;
				i = lastPage->m_NbPackets;
				while(i--)
					addedGranulepos += lastPage->m_NbSamples[i];
				if(m_CurrentStream->m_TmpPageGranulepos_0 + addedGranulepos != m_CurrentStream->m_TmpPageGranulepos_1) {
					VDDEBUG("May have found corrupted header in stream %ld : jumping from granulepos %I64d to %I64d while %I64d (at most) should have been reached. Modifying information in memory.\n"
						, m_CurrentStream->m_Serial, m_CurrentStream->m_TmpPageGranulepos_0, m_CurrentStream->m_TmpPageGranulepos_0 + addedGranulepos, m_CurrentStream->m_TmpPageGranulepos_1);
					addedGranulepos -= lastPage->m_NbSamples[0];
					VDASSERT(m_CurrentStream->m_TmpPageGranulepos_0 + addedGranulepos>m_CurrentStream->m_TmpPageGranulepos_0);
					VDASSERT((m_CurrentStream->m_TmpPageGranulepos_0 + addedGranulepos - m_CurrentStream->m_TmpPageGranulepos_0)<=0xFFFFFFFF);
					uint64 nbSamplesAnnounced = lastPage->m_NbSamples[0];
					uint32 nbSamplesCorrected = 0;
					// For non video stream use at least 1 sample (otherwise the stream would be cut :()
					if(MAIN_TYPE(m_CurrentStream->m_StreamType) != SOURCE_VIDEO)
						nbSamplesCorrected = 1;
					uint64 timePosMS = 0;
					_getCurrentTimeMS(MAIN_TYPE(m_CurrentStream->m_StreamType) == SOURCE_VIDEO, timePosMS);
					if(m_CurrentStream->m_TmpPageGranulepos_0 + addedGranulepos <= m_CurrentStream->m_TmpPageGranulepos_1)
						nbSamplesCorrected = m_CurrentStream->m_TmpPageGranulepos_1 - m_CurrentStream->m_TmpPageGranulepos_0 - addedGranulepos;
					m_CurrentStream->m_NbSamples -= (nbSamplesAnnounced - nbSamplesCorrected);
					m_CurrentStream->AddWeirdPacket(OGM_FILEPOS(lastPage->m_FilePos), timePosMS, 0, nbSamplesAnnounced, nbSamplesCorrected);
					lastPage->m_NbSamples[0] = nbSamplesCorrected;
				}
				if(m_CurrentStream->m_PacketsExpected) {
					m_CurrentStream->SetLastWeirdPageLostPackets(m_CurrentStream->m_PacketsExpected);
					m_CurrentStream->m_PacketsExpected = 0;
				}
				// Readjust the granulepos if necessary
				while(pageIt.next()) {
					uint64 granulepos = (*pageIt)->m_Granulepos;
					int count = (*pageIt)->m_NbPackets;
					while(count--)
						granulepos += (*pageIt)->m_NbSamples[count];
					pageIt++;
					VDASSERT((*pageIt)->m_Granulepos != granulepos);
					(*pageIt)->m_Granulepos = granulepos;
				}
				m_CurrentStream->m_CheckPage = 0;
			}
		}
		if(m_CurrentStream->m_CheckPage == 1) {
			// Save the granulepos of the previous Page in the stream
			m_CurrentStream->m_TmpPageGranulepos_0 = m_CurrentStream->m_TmpPageGranulepos_1;
			m_CurrentStream->m_CheckPage = 2;
		}
	}

	m_CurrentStream->m_TmpPageGranulepos_1 = ogg_page_granulepos(page);

	if(ogg_stream_pagein(&m_CurrentStream->m_StreamState, page))
		throw MyError("ogg_stream_pagein failed");

	// We will estimate the absolute granulepos of the first sample in the Page
	// For that we need the granulepos of the previous Page, and the number
	// of samples in this previous Page
	// Nb : here granulepos starts @ 0 (convenient way since VirtualDub do the same)
	OggPage *lastPage = m_CurrentStream->m_PageList.GetLast();
	ogg_int64_t granulepos = lastPage->m_Granulepos;
	int i = lastPage->m_NbPackets;
	while(i--)
		granulepos += lastPage->m_NbSamples[i];

	lastPage = m_CurrentStream->m_PageList.AddLast();

	lastPage->m_Granulepos = granulepos;
	lastPage->m_FilePos = m_BytesRead;
	lastPage->m_Size = page->header_len + page->body_len;
	if(ogg_page_continued(page))
		lastPage->m_FilePos |= OGM_FLAG_CONT;
	lastPage->SetNbPackets(ogg_page_packets(page));
	m_BytesRead += page->header_len + page->body_len;
	m_CurrentStream->m_PacketIndex = 0;

	int iResult;
	//while(ogg_stream_packetout(&m_CurrentStream->m_StreamState, &m_Packet)==1) {
	while((iResult=ogg_stream_packetout(&m_CurrentStream->m_StreamState, &m_Packet))!=0) {
		if(iResult==-1)
			continue;
		_addPacket(&m_Packet);
	}
}

void OGMReadHandler::_addPacket(ogg_packet *packet) {
	OggPage *lastPage = m_CurrentStream->m_PageList.GetLast();
	uint16 packetIndex = m_CurrentStream->m_PacketIndex;

	if(!lastPage)
		throw MyError("OGMReadHandler::_addPacket : no Page created yet for the stream %ld!", m_CurrentStream->m_Serial);

	if(packetIndex>=lastPage->m_NbPackets) {
		VDASSERT(packetIndex==lastPage->m_NbPackets);
		if(m_CurrentStream->m_CheckPage)
			m_CurrentStream->m_PacketsExpected -= (packetIndex + 1 - lastPage->m_NbPackets);
		VDASSERT(m_CurrentStream->m_PacketsExpected >= 0);
		uint64 timePosMS = 0;
		_getCurrentTimeMS(MAIN_TYPE(m_CurrentStream->m_StreamType) == SOURCE_VIDEO, timePosMS);
		m_CurrentStream->AddWeirdPage(OGM_FILEPOS(lastPage->m_FilePos), timePosMS, lastPage->m_NbPackets, packetIndex+1);
		if(packetIndex > 255) {
			VDDEBUG("Created a new OggPage for stream %ld because library returned more than 255 Packets.\n"
				, m_CurrentStream->m_Serial);
			if(m_CurrentStream->m_CheckPage)
				m_CurrentStream->m_SplittedPages++;
			OggPage *newLastPage = m_CurrentStream->m_PageList.AddLast();
			newLastPage->m_FilePos = lastPage->m_FilePos;
			newLastPage->m_Size = lastPage->m_Size;
			newLastPage->SetNbPackets(1);
			uint64 granulepos = lastPage->m_Granulepos;
			int i = lastPage->m_NbPackets;
			while(i--)
				granulepos += lastPage->m_NbSamples[i];
			newLastPage->m_Granulepos = granulepos;
			lastPage = newLastPage;
			packetIndex = m_CurrentStream->m_PacketIndex = 0;
		} else {
			VDDEBUG("Invalid Page found in stream %ld : stated %d packets but more found\r"
			, m_CurrentStream->m_Serial, (int)lastPage->m_NbPackets);
			lastPage->UpdateNbPackets(packetIndex+1);
		}
	}

	VDASSERT(packet->bytes>0);

	if(m_CurrentStream->m_StreamType == SOURCE_AUDIO_VORBIS) {
		// Vorbis Packet
		if(packet->packetno<3)  {
			lastPage->m_NbBytes[packetIndex] = packet->bytes;
			// First packet already processed here
			if(packet->packetno!=0) {
				if(vorbis_synthesis_headerin(&m_CurrentStream->m_VorbisInfo, &m_CurrentStream->m_VorbisComment, packet)<0) {
					VDDEBUG("Stream %ld is not a valid Vorbis stream\n", m_CurrentStream->m_Serial);
					m_CurrentStream->m_StreamType = SOURCE_UNKNOWN;
					return;
				}
				m_CurrentStream->AddHeaderPacket(packet);
			}
		} else {
			if((*packet->packet) & 0x1) {
				// It doesn't seem to be a data packet :/
				uint64 timePosMS = 0;
				_getCurrentTimeMS(MAIN_TYPE(m_CurrentStream->m_StreamType) == SOURCE_VIDEO, timePosMS);
				m_CurrentStream->AddWeirdPacket(lastPage->m_FilePos | OGM_FLAG_CONT, timePosMS, packetIndex, -1, -1);
			} else {
				sint32 thisW = vorbis_packet_blocksize(&m_CurrentStream->m_VorbisInfo, packet);
				// Found in vcut (part of vorbis-tools : see www.vorbis.org)
				sint32 nbSamples = (thisW+m_CurrentStream->m_PreviousW)/4;
				lastPage->m_NbSamples[packetIndex] = nbSamples;
				lastPage->m_NbBytes[packetIndex] = packet->bytes;
				m_CurrentStream->m_Bytes += packet->bytes;
				m_CurrentStream->m_NbSamples += nbSamples;
				m_CurrentStream->m_NbPackets++;
				m_CurrentStream->m_PreviousW = thisW;
			}
		}
	} else if( (*packet->packet & 0x01) == PACKET_TYPE_DATA) {
		// Data packet
		sint32 hdrlen, i;
		uint64 nbSamples = 0;
		hdrlen = ((*packet->packet & PACKET_LEN_BITS01) >> 6)
			| ((*packet->packet & PACKET_LEN_BITS2) << 1);
		unsigned char *pos = (unsigned char *)packet->packet + hdrlen;
		i = hdrlen;
		while(i--) {
			nbSamples <<= 8;
			nbSamples += *pos;
			pos--;
		}
		if(!nbSamples) {
			nbSamples = m_CurrentStream->m_StreamHeader->default_len;
		} else if(nbSamples >> 32)
			nbSamples = 0xFFFFFFFF;
		lastPage->m_NbSamples[packetIndex] = nbSamples;
		lastPage->m_NbBytes[packetIndex] = packet->bytes-1-hdrlen;
		m_CurrentStream->m_Bytes += packet->bytes-1-hdrlen;
		m_CurrentStream->m_NbSamples += nbSamples;
		m_CurrentStream->m_NbPackets++;
		// If this is a keyframe, add it to the list
		if( (m_CurrentStream->m_StreamType == SOURCE_VIDEO) && (*packet->packet & PACKET_IS_SYNCPOINT))
			lastPage->m_NbBytes[packetIndex] |= 0x80000000;
	} else if((*packet->packet & PACKET_TYPE_BITS) == PACKET_TYPE_COMMENT) {
		// Comment Packet, but not a Vorbis stream
		VDASSERT(!m_CurrentStream->m_bHasComments);
		lastPage->m_NbBytes[packetIndex] = packet->bytes;
		m_CurrentStream->AddHeaderPacket(packet);
		m_CurrentStream->m_VorbisInfo.channels = 2;
		m_CurrentStream->m_VorbisInfo.rate = 48000;
		if(vorbis_synthesis_headerin(&m_CurrentStream->m_VorbisInfo, &m_CurrentStream->m_VorbisComment, packet)>=0)
			m_CurrentStream->m_bHasComments = true;
		else
			VDDEBUG("Stream %ld doesn't contain a valid Comment Packet\n", m_CurrentStream->m_Serial);
	} else {
		// Probably a Header Packet (already processed), or a corrupted one
		VDASSERT((*packet->packet & PACKET_TYPE_BITS) == PACKET_TYPE_HEADER);
		lastPage->m_NbBytes[packetIndex] = packet->bytes;
	}

	if(lastPage->m_NbBytes[packetIndex] > m_CurrentStream->m_MaxPacketSize)
		m_CurrentStream->m_MaxPacketSize = lastPage->m_NbBytes[packetIndex];

	m_CurrentStream->m_PacketIndex++;
}

void OGMReadHandler::_addCorruptedData(uint64 filePos, uint32 length) {
	if(!length)
		return;
	OggCorruptedData *lastData = m_CorruptedDataList.GetLast();
	if(lastData && (lastData->m_FilePos+lastData->m_Length==filePos) )
		lastData->m_Length += length;
	else {
		lastData = m_CorruptedDataList.AddLast();
		lastData->m_FilePos = filePos;
		lastData->m_Length = length;
		_getCurrentTimeMS(false, lastData->m_TimePosMS);
	}
}

bool OGMReadHandler::_getCurrentTimeMS(bool bDiscardLastPage, uint64& nMillisecond) {
	nMillisecond = 0;
	ModList<OGMReadNode>::iterator it(m_StreamList);
	while(it) {
		OggStream *curStream = (*it)->m_Segments.GetFirst()->m_OggStream;
		if(MAIN_TYPE(curStream->m_StreamType) == SOURCE_VIDEO) {
			VDASSERT((long)curStream->m_SampleRate);
			ModCompactList<OggPage>::iterator pageIt;
			pageIt.end(curStream->m_PageList);
			if(pageIt && bDiscardLastPage)
				pageIt--;
			OggPage *lastPage = (pageIt ? *pageIt : NULL);
			if(!lastPage || !(long)curStream->m_SampleRate)
				return false;
			sint64 granulepos = lastPage->m_Granulepos;
			int i = lastPage->m_NbPackets;
			while(i--)
				granulepos += lastPage->m_NbSamples[i];
			nMillisecond = curStream->m_SampleRate.scale64ir(1000 * granulepos);
			return true;
		}
		it++;
	}
	return false;
}

void OGMReadHandler::_setCorruptedFlags(void) {
	ModList<OGMReadNode>::iterator it(m_StreamList);
	while(it) {
		OggStream *curStream = (*it)->m_Segments.GetFirst()->m_OggStream;
		VDASSERT(!curStream->m_CheckPage || (curStream->m_CheckPage==1));
		curStream->m_CheckPage = 1;
		it++;
	}
}

void OGMReadHandler::_parseFile() {
	sint64 fileSize = m_File.size();

	sint64 pageLimit = fileSize / (500 * 4096);
	if(pageLimit <= 0)
		pageLimit = 1;

	ProgressDialog pd(NULL, "Ogg Media File Import Filter", "Parsing Ogg Media File", (long)(fileSize/1024), true);

	pd.setValueFormat("%ldK of %ldK");

	ogg_sync_init(&m_SyncState);
	sint64 pageProcessed = 0;
	while(_getNextPage(&m_Page)>0) {
		_addPage(&m_Page);
		if( (pageProcessed % pageLimit) == 0) {
			pd.advance((long)(m_BytesRead/1024));
			pd.check();
		}
		pageProcessed++;
	}
	ogg_sync_clear(&m_SyncState);
}

sint32 OGMReadHandler::_getNextPage(ogg_page *page) {
	sint32 np, nread;
	char *buffer;
	while(true) {
		np = ogg_sync_pageseek(&m_SyncState, page);
		if(np>0)
			return np;
		else if(np==0) {
			buffer = ogg_sync_buffer(&m_SyncState, OGG_BLOCKSIZE);
			if(!buffer)
				throw MyError("ogg_sync_buffer failed");
			//EnterCriticalSection(&g_diskcs);
			if((nread = m_File.readData(buffer, OGG_BLOCKSIZE)) <= 0) {
				//LeaveCriticalSection(&g_diskcs);
				return 0;
			}
			//LeaveCriticalSection(&g_diskcs);
			ogg_sync_wrote(&m_SyncState, nread);
		} else {
			VDDEBUG("ogg_sync_pageseek failed, %ld bytes skipped\n", -np);
			_addCorruptedData(m_BytesRead, -np);
			_setCorruptedFlags();
			m_BytesRead += (-np);
			m_BytesSkipped += (-np);
		}
	}
	return np;
}

long OGMReadHandler::ReadData(void *buffer, long nBytes) {
	EnterCriticalSection(&g_diskcs);
	long nread = m_File.readData(buffer, nBytes);
	LeaveCriticalSection(&g_diskcs);
	return nread;
}

void OGMReadHandler::SeekFile(sint64 filePos) {
	m_File.seek(filePos);
}

IAVIReadStream *OGMReadHandler::GetStream(DWORD fccType, LONG lParam) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		OggStream *curStream = (*m_StreamIterator)->m_Segments.GetFirst()->m_OggStream;
		switch(curStream->m_StreamType) {
		case SOURCE_VIDEO:
			if( (fccType==streamtypeVIDEO) && (!lParam--) )
				return new OGMReadStream(*m_StreamIterator);
			break;
		case SOURCE_AUDIO:
		case SOURCE_AUDIO_VORBIS:
			if( (fccType==streamtypeAUDIO) && (!lParam--) )
				return new OGMReadStream(*m_StreamIterator);
			break;
		case SOURCE_TEXT:
			if( (fccType==streamtypeAUDIO) && (!lParam--) )
				return new OGMReadStream(*m_StreamIterator);
			break;
		}
		m_StreamIterator++;
	}
	return NULL;
}

IAVIReadStream *OGMReadHandler::GetStream(LONG lParam) {
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		if(!lParam--)
			return new OGMReadStream(*m_StreamIterator);
		m_StreamIterator++;
	}
	return NULL;
}

long OGMReadHandler::GetStreamIndexByType(OGMReadNode *stream) {
	
	VDASSERT(stream);
	
	long index = 0;
	m_StreamIterator.begin(m_StreamList);
	while(m_StreamIterator) {
		OGMReadNode *curStream = *m_StreamIterator;
		if(curStream == stream)
			return index;
		if(MAIN_TYPE(curStream->m_Segments.GetFirst()->m_OggStream->m_StreamType) == MAIN_TYPE(stream->m_Segments.GetFirst()->m_OggStream->m_StreamType) )
			index++;
		m_StreamIterator++;
	}
	return -1;
}

ModList<OGMReadNode>& OGMReadHandler::GetStreamList(void) {
	return m_StreamList;
}

void OGMReadHandler::EnableFastIO(bool bEnable) {
}

bool OGMReadHandler::isOptimizedForRealtime(void) {
	return false;
}

bool OGMReadHandler::isStreaming(void) {
	return false;
}

bool OGMReadHandler::isIndexFabricated(void) {
	return false;
}

bool OGMReadHandler::AppendFile(const wchar_t *pszFile) {
	if(!pszFile)
		return false;
	OGMReadHandler *nextSegment = new OGMReadHandler(pszFile);
	if(!nextSegment)
		throw MyMemoryError();

	ModList<OGMReadNode>::iterator firstSegmentIt;
	ModList<OGMReadNode>::iterator nextSegmentIt;
	try {
		if(m_StreamList.GetCount() != nextSegment->GetStreamList().GetCount())
			throw MyError("Cannot append segment: The segment has a different number of streams.");

		firstSegmentIt.begin(m_StreamList);
		nextSegmentIt.begin(nextSegment->GetStreamList());
		while(firstSegmentIt) {
			OggStream *firstStreamSegment = (*firstSegmentIt)->m_Segments.GetFirst()->m_OggStream;
			OggStream *nextStreamSegment = (*nextSegmentIt)->m_Segments.GetFirst()->m_OggStream;

			if(firstStreamSegment->m_StreamType != nextStreamSegment->m_StreamType)
				throw MyError("Cannot append segment: The stream types do not match.");

			if(firstStreamSegment->m_StreamHeader->size != nextStreamSegment->m_StreamHeader->size)
				throw MyError("Cannot append segment: The header sizes do not match (%d vs %d)."
					, firstStreamSegment->m_StreamHeader->size
					, nextStreamSegment->m_StreamHeader->size);

			if(firstStreamSegment->m_SampleRate != nextStreamSegment->m_SampleRate)
				throw MyError("Cannot append segment: The samplerates do not match (%.5f vs %.5f)."
					, (double)firstStreamSegment->m_SampleRate
					, (double)nextStreamSegment->m_SampleRate);

			switch(MAIN_TYPE(firstStreamSegment->m_StreamType)) {
			case SOURCE_VIDEO:
				if( (firstStreamSegment->m_StreamHeader->sh.video.width!=nextStreamSegment->m_StreamHeader->sh.video.width)
				|| (firstStreamSegment->m_StreamHeader->sh.video.height!=nextStreamSegment->m_StreamHeader->sh.video.height) )
					throw MyError("Cannot append segment: The video size do not match (%ldx%ld vs %ldx%ld)."
						, firstStreamSegment->m_StreamHeader->sh.video.width, firstStreamSegment->m_StreamHeader->sh.video.height
						, nextStreamSegment->m_StreamHeader->sh.video.width, nextStreamSegment->m_StreamHeader->sh.video.height);
				break;
			case SOURCE_AUDIO:
				if(firstStreamSegment->m_StreamHeader->sh.audio.channels != nextStreamSegment->m_StreamHeader->sh.audio.channels)
					throw MyError("Cannot append segment: The number of chanels do not match (%d vs %d)."
						, firstStreamSegment->m_StreamHeader->sh.audio.channels, nextStreamSegment->m_StreamHeader->sh.audio.channels);
				break;
			}

			firstSegmentIt++;
			nextSegmentIt++;
		}
	} catch(...) {
		nextSegment->Release();
		throw;
	}

	// Merging accepted, notify the read handlers
	firstSegmentIt.begin(m_StreamList);
	nextSegmentIt.begin(nextSegment->GetStreamList());
	while(firstSegmentIt) {
		ModList<OggStreamSegment>::iterator segmentIt((*nextSegmentIt)->m_Segments);
		while(segmentIt) {
			(*firstSegmentIt)->AppendSegment(*segmentIt);
			segmentIt++;
		}
		(*firstSegmentIt)->ReinitReadStreams();
		firstSegmentIt++;
		nextSegmentIt++;
	}

	nextSegment->Release();
	return true;
}

bool OGMReadHandler::getSegmentHint(const char **ppszPath) {
	if(ppszPath)
		*ppszPath = NULL;
	else
		return false;

	return true;
}

void OGMReadHandler::GetCommentList(ModList<ModStreamComment>& list) {
}

void OGMReadHandler::GetChapterList(ModChapterList& list) {
}

VDStringW OGMReadHandler::writeReport(void) {
	VDStringW report;
	report = VDStringW::setf(L"---== Ogg Media File Import filter report ==---\r\nSource file : [%s]", m_FileName.c_str());
	if(m_CorruptedDataList.GetCount()) {
		report += L"\r\nSome bytes have been skipped in the file due to corrupted data :";
		ModCompactList<OggCorruptedData>::iterator it(m_CorruptedDataList);
		while(it) {
			OggCorruptedData *data = *it;
			sint32 ms = data->m_TimePosMS % 1000;
			sint32 s = (data->m_TimePosMS / 1000) % 60;
			sint32 m = (data->m_TimePosMS / 1000 / 60) % 60;
			sint32 h = (data->m_TimePosMS / 1000 / 60 / 60);
			report += VDStringW::setf(L"\r\n %lu bytes starting at byte %I64u (%02d:%02d:%02d.%03d)"
				, data->m_Length, data->m_FilePos, h, m, s, ms);
			it++;
		}
	} else
		report += L"\r\nNo apparent corrupted data";
	OGMReadNode *readNode = m_StreamList.GetFirst();
	if(readNode && (readNode->m_Segments.GetCount()>1)) {
		ModList<OggStreamSegment>::iterator it(readNode->m_Segments);
		it++;
		while(it) {
			report += L"\r\n";
			report += (*it)->m_ReadHandler->writeReport();
			it++;
		}
	}
	return report;
}
