//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <tchar.h>
#include "misc.h"
#include "StreamSourceOGM.h"
#include "OGMReadHandler.h"


bool VideoSourceOGM::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IOggStream::GetInterfaceID()) {
		*destination = static_cast<IOggStream *>(this);
		return true;
	}
	return VideoSourceAVI::QueryInterface(ID, destination);
}

VideoSourceOGM::VideoSourceOGM(OGMReadHandler *readHandler)
: VideoSourceAVI(readHandler) {
	m_SourceType = SOURCE_OGM | SOURCE_VIDEO;
	clearSourceInfo();
	m_SourceName = L"Input Ogg Media File";
	m_SourceDesc = L"Video stream";
}

uint32 VideoSourceOGM::GetOggStreamType(void) {
	return ((OGMReadStream *)pAVIStream)->GetOggStreamType();
}

uint16 VideoSourceOGM::GetNbHeaderPackets(void) {
	return ((OGMReadStream *)pAVIStream)->GetNbHeaderPackets();
}

ogg_packet *VideoSourceOGM::GetHeaderPacket(uint16 index) {
	return ((OGMReadStream *)pAVIStream)->GetHeaderPacket(index);
}

void VideoSourceOGM::GetCommentList(ModList<ModStreamComment>& list) {
	((OGMReadStream *)pAVIStream)->GetCommentList(list);
}

void VideoSourceOGM::GetChapterList(ModChapterList& list) {
	((OGMReadStream *)pAVIStream)->GetChapterList(list);
}

VDStringW VideoSourceOGM::writeReport(void) {
	VDStringW report = VideoSourceAVI::writeReport();
	report += L"\r\n";
	report += ((OGMReadStream *)pAVIStream)->writeReport();
	return report;
}

//////////////////////////////////////////////////////////////////////

bool StreamSourceOGM::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IStreamIndex::GetInterfaceID()) {
		*destination = static_cast<IStreamIndex *>(this);
		return true;
	} else if(ID == ICommentList::GetInterfaceID()) {
		*destination = static_cast<ICommentList *>(this);
		return true;
	} else if(ID == IOggStream::GetInterfaceID()) {
		*destination = static_cast<IOggStream *>(this);
		return true;
	}
	return false;
}

StreamSourceOGM::StreamSourceOGM(OGMReadHandler *readHandler, long index)
: AudioSource() {
	VDASSERT(readHandler && (index>=0));
	m_ReadHandler = readHandler;
	m_ReadStream = NULL;
	m_StreamIndex = index;
	m_bVBRHandling = true;

	m_SourceType = SOURCE_OGM;
	clearSourceInfo();
	m_SourceName = L"Ogg Media File";
}

StreamSourceOGM::~StreamSourceOGM(void) {
	if(m_ReadStream)
		delete m_ReadStream;
}


bool StreamSourceOGM::init() {
	LONG format_len;

	m_ReadStream = (OGMReadStream *)m_ReadHandler->GetStream(m_StreamIndex);

	if (!m_ReadStream) return false;

	if (m_ReadStream->Info(&streamInfo, sizeof streamInfo))
		return false;

	m_ReadStream->FormatSize(0, &format_len);

	if (!allocFormat(format_len)) return false;

	if (m_ReadStream->ReadFormat(0, getFormat(), &format_len))
		return false;

	lSampleFirst = m_ReadStream->Start();
	lSampleLast = m_ReadStream->End();

	m_SourceType |= m_ReadStream->GetOggStreamType();

	VDASSERT(getSourceMainType() != SOURCE_VIDEO);

	switch(getSourceMainType()) {
	case SOURCE_TEXT:
		m_SourceDesc = VDStringW::setf(L"Ogg Media File text stream %d", GetStreamIndex()+1);
		break;
	case SOURCE_AUDIO:
		if(getSourceType() == SOURCE_AUDIO_VORBIS)
			m_SourceDesc = VDStringW::setf(L"OGM audio stream %d [Vorbis], %dch, %dHz, %.1fkbps"
				, GetStreamIndex()+1
				, getWaveFormat()->nChannels, getWaveFormat()->nSamplesPerSec, ((double)8*getWaveFormat()->nAvgBytesPerSec)/1000.);
		else {
			TCHAR szTag[5], szCodecName[1024];
			memset(szTag, 0, sizeof(szTag));
			_snprintf(szTag, sizeof(szTag)-1, _T("%04X"), getWaveFormat()->wFormatTag);
			memset(szCodecName, 0, sizeof(szCodecName));
			ModGetCodecName(_T("audio"), szTag, _T("Unknown"), szCodecName, sizeof(szCodecName)/sizeof(szCodecName[0]));
			VDStringW sTag;
			sTag = L"tag 0x";
			if(sizeof(TCHAR) == sizeof(char)) {
				sTag += VDTextAToW((char *)szTag);
				sTag += L", ";
				sTag += VDTextAToW((char *)szCodecName);
			} else {
				sTag += (wchar_t *)szTag;
				sTag += L", ";
				sTag += (wchar_t *)szCodecName;
			}
			m_SourceDesc = VDStringW::setf(L"OGM audio stream %d [%s], %dch, %dHz, %.1fkbps"
				, GetStreamIndex()+1
				, sTag.c_str()
				, getWaveFormat()->nChannels, getWaveFormat()->nSamplesPerSec, ((double)8*getWaveFormat()->nAvgBytesPerSec)/1000.);
		}
		break;
	}

	return true;
}

void StreamSourceOGM::extend(bool bVBRHandling, sint32 groupFrames) {
	m_bVBRHandling = bVBRHandling;
	m_ReadStream->SetVBRHandling(bVBRHandling);
	Reinit();
}

void StreamSourceOGM::Reinit() {
	LONG formatLen = getFormatLen();
	m_ReadStream->ReadFormat(0, getFormat(), &formatLen);
	m_ReadStream->Info(&streamInfo, sizeof streamInfo);
	lSampleFirst = m_ReadStream->Start();
	lSampleLast = m_ReadStream->End();
}

bool StreamSourceOGM::isStreaming() {
	return m_ReadStream->isStreaming();
}

void StreamSourceOGM::streamBegin(bool fRealTime) {
	m_ReadStream->BeginStreaming(lSampleFirst, lSampleLast, fRealTime ? 1000 : 2000);
}

void StreamSourceOGM::streamEnd() {
	m_ReadStream->EndStreaming();
}

sint64 StreamSourceOGM::sampleOffset(sint64 lSample) {
	return m_ReadStream->sampleOffset(lSample);
}

sint32 StreamSourceOGM::samplesLag(sint64 lStart) {
	return m_ReadStream->samplesLag(lStart);
}

sint32 StreamSourceOGM::sourceOffsetMS(void) {
	return m_ReadStream->sourceOffsetMS();
}

int StreamSourceOGM::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lpBytesRead, LONG *lpSamplesRead) {
	return m_ReadStream->Read(lStart, lCount, lpBuffer, cbBuffer, lpBytesRead, lpSamplesRead);
}

DWORD StreamSourceOGM::GetStreamType(void) {
	switch(MAIN_TYPE(m_ReadStream->GetOggStreamType())) {
	case SOURCE_VIDEO:
		throw MyError("Accessing to a Video stream via StreamSourceOGM!");
		return streamtypeVIDEO;
	case SOURCE_AUDIO:
		return streamtypeAUDIO;
	case SOURCE_TEXT:
		return streamtypeTEXT;
	default:
		break;
	}
	return 0;
}

long StreamSourceOGM::GetStreamIndex(void) {
	return m_ReadStream->GetStreamIndexByType();
}

uint32 StreamSourceOGM::GetOggStreamType(void) {
	return m_ReadStream->GetOggStreamType();
}

uint16 StreamSourceOGM::GetNbHeaderPackets(void) {
	return m_ReadStream->GetNbHeaderPackets();
}

ogg_packet *StreamSourceOGM::GetHeaderPacket(uint16 index) {
	return m_ReadStream->GetHeaderPacket(index);
}

void StreamSourceOGM::GetCommentList(ModList<ModStreamComment>& list) {
	m_ReadStream->GetCommentList(list);
}

void StreamSourceOGM::GetChapterList(ModChapterList& list) {
	m_ReadStream->GetChapterList(list);
}

VDStringW StreamSourceOGM::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	report += L"\r\n";
	report += m_ReadStream->writeReport();
	return report;
}

//////////////////////////////////////////////////////////////////////

bool TextSourceOGM::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == ISubtitleHandler::GetInterfaceID()) {
		*destination = static_cast<ISubtitleHandler *>(this);
		return true;
	}
	return StreamSourceOGM::QueryInterface(ID, destination);
}

TextSourceOGM::TextSourceOGM(OGMReadHandler *readHandler, long index)
: StreamSourceOGM(readHandler, index) {
}

bool TextSourceOGM::IsASCII(void) {
	return true;
}

bool TextSourceOGM::HasBOM(void) {
	return false;
}

eTextType TextSourceOGM::GetOutputType(void) {
	return kTextType_ASCII;
}

ModSubtitleProvider *TextSourceOGM::GetSubtitleProvider(void) {
	return NULL;
}