//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <crtdbg.h>

#include <vd2/system/Error.h>
#include <vd2/system/fraction.h>
#include "ModAudio.h"
#include "ModDubStream.h"


#define __MOD_AUDIO_OVERLAP__


long ModAudioTranslateVideoSubset(FrameSubset& dst, FrameSubset& src, const VDFraction& videoFrameRate, WAVEFORMATEX *pwfex, ModDubStream *pDubStream) {
	const long lSampPerSec = pwfex->nSamplesPerSec;
	const long nBytesPerSec = pwfex->nAvgBytesPerSec;
	int nBlockAlign = pwfex->nBlockAlign;
	long total = 0;
	bool bNandubMP3 = pDubStream && pDubStream->m_bIsNandubMP3VBR
		&& ((pwfex->wFormatTag==WAVE_FORMAT_MPEGLAYER3) || (pwfex->wFormatTag==WAVE_FORMAT_MPEG))
		&& (nBlockAlign==1152);
	if(bNandubMP3 && (pDubStream->m_Source->getStreamInfo().dwScale>0))
		nBlockAlign = pDubStream->m_Source->getStreamInfo().dwScale;

	// I like accuracy, so let's strive for accuracy.  Accumulate errors as we go;
	// use them to offset the starting points of subsequent segments, never being
	// more than 1/2 segment off.
	//
	// The conversion equation is in units of (1000000*nBlockAlign).

	sint64 nError		= 0;
	sint64 nMultiplier	= (sint64)videoFrameRate.getLo() * lSampPerSec;
	sint64 nDivisor;
	if(bNandubMP3)
		nDivisor		= (sint64)videoFrameRate.getHi() * nBlockAlign;
	else
		nDivisor		= (sint64)videoFrameRate.getHi();
	sint64 nRound		= nDivisor/2;
	long nTotalFramesAccumulated = 0;

	for(FrameSubset::iterator it = src.begin(), itEnd = src.end(); it != itEnd; ++it) {
		long start, end;

		// Compute error.
		//
		// Ideally, we want the audio and video streams to be of the exact length.
		//
		// Audiolen = (videolen * usPerFrame * nBytesPerSec) / (1000000*nBlockAlign);

		nError = total*nDivisor - (nTotalFramesAccumulated * nMultiplier);

		VDDEBUG("nError = %I64d\n", nError);

		// Add a block.

		start = ((__int64)it->start * nMultiplier + nRound + nError) / nDivisor;
		end = ((__int64)(it->start + it->len) * nMultiplier + nRound) / nDivisor;

		nTotalFramesAccumulated += it->len;

		dst.addRange(start, end-start, false);

		total += end-start;
	}

	return total;
}

ModAudioSubset::ModAudioSubset(AudioStream *src, FrameSubset *pfs, const VDFraction& videoFrameRate, sint64 preskew, bool isText, ModDubStream *pDubStream)
: AudioSubset() {
	memcpy(AllocFormat(src->GetFormatLen()), src->GetFormat(), src->GetFormatLen());

	m_bText = isText;

	SetSource(src);

	long total = ModAudioTranslateVideoSubset(subset, *pfs, videoFrameRate, src->GetFormat(), pDubStream);

	const WAVEFORMATEX *pSrcFormat = src->GetFormat();
	bool bNandubMP3 = pDubStream && pDubStream->m_bIsNandubMP3VBR
		&& ((pSrcFormat->wFormatTag==WAVE_FORMAT_MPEGLAYER3) || (pSrcFormat->wFormatTag==WAVE_FORMAT_MPEG))
		&& (pSrcFormat->nBlockAlign==1152);

	if(bNandubMP3) {
		sint64 nScale;
		if(pDubStream->m_Source->getStreamInfo().dwScale > 0)
			nScale = pDubStream->m_Source->getStreamInfo().dwScale;
		else
			nScale = pSrcFormat->nBlockAlign;
		subset.deleteRange(0, (preskew*pSrcFormat->nSamplesPerSec) / ((sint64)1000000*nScale));
	} else
		subset.deleteRange(0, (preskew*pSrcFormat->nSamplesPerSec) / (sint64)1000000);

	stream_len = total;

	SetLimit(total);

	pfsnCur = subset.begin();
	iOffset = 0;
	lSrcPos = 0;
	// Arrgh we may treat VBR streams here, so lSkipSize is not really accurate
	// I will divide this value by 3 (i.e. I assume the stream won't exceed
	// 3*average_bitrate as instant bitrate in a particular time)
	// Well anyway normally this value will be of no use when treating OGM
	// but just in case ...
	// If no average bitrate provided, assume 128kbps (=> max of 384kbps with the /3)
	lSkipSize = (sizeof skipBuffer * src->GetFormat()->nSamplesPerSec)
		/ (src->GetFormat()->nAvgBytesPerSec > 0 ? src->GetFormat()->nAvgBytesPerSec : 48000);
	lSkipSize /= 3;
	if(lSkipSize <= 0)
		lSkipSize = 1;
}

long ModAudioSubset::_Read(void *buffer, long samples, long *lplBytes) {
	long offset, actual;

	if (pfsnCur == subset.end()) {
		*lplBytes = 0;
		return 0;
	}

	const FrameSubsetNode& node = *pfsnCur;

	/* Modification for OGM
	 *
	 * VirtualDub native routines read exactly "samples" samples in the
	 * stream, no more no less. But for OGM we read as many samples as
	 * we can in a frame (AC3 or MP3 stream) or Packet (Ogg/Ogm stream).
	 *
	 * That's why I added a boolean in AudioStreamSource::Read to know
	 * if we have to test the number of samples read.
	 *
	 * But here this is another thing : the user deleted parts of the video
	 * (and so audio&text too). Of course we want all the stuffs still
	 * synched between each parts. With original routines, there is no
	 * problem since we read exactly "samples" samples, but it is not the case
	 * for OGM :( : depending how many samples we will read at a time coming
	 * from the input stream, we may end up with a lag of 0 (ideal case indeed)
	 * up to a certain amount of samples (example : one MPeg3-Layer1 frame
	 * contains 1152 samples, an AC3 contains 1536 samples, and in general
	 * onr Packet of Vorbis stream contains 1024 samples ... but in an OGM
	 * file the orignal stream may have cut the MP3/AC3 frames into various
	 * Packets, or even merged various frames into one Packet ....)
	 *
	 * To keep synchronisation, I thought of two possibilities here :
	 * 1. When we reach the end of a part and that the number of samples
	 *    read exceed the end of this part, assume we in fact read
	 *    enough samples to just reach the part (so we lower the number
	 *    of samples really read)
	 * 2. Keep in my mind by how many samples we exeeded the end of the part
	 *    , reporting this number to the beginning of the next part (thanks
	 *    to the iOffset variable gracefully provided by Avery Lee :) )
	 *
	 * In any case we must take into account the type of the stream (because
	 * here we may treat audio stream, but also text stream).
	 * And for a text stream, only rule 1. is acceptable (unless you don't
	 * care seeing a subtitle continuing on a next part and thus replacing
	 * the subtitle you should see as beginning of this next part ;) ).
	 */
#ifdef __MOD_AUDIO_OVERLAP__
	while((offset = node.start - lSrcPos) > 0) {
#else
	while((offset = (node.start - lSrcPos) + iOffset) > 0) {
#endif
		long t;

		if(source->Skip(offset)) {
			lSrcPos += offset;
			break;
		}

		if(offset > lSkipSize) offset = lSkipSize;

		actual = source->Read(skipBuffer, offset, &t, false);
		
		if(!actual) {
			*lplBytes = 0;
			return 0;
		}

		lSrcPos += actual;
	}

	if(samples > node.len - iOffset)
		samples = node.len - iOffset;

	samples = source->Read(buffer, samples, lplBytes, false);

	// Update state of lSrcPos now : even if we state as having read less
	// samples than in reality, lSrcPos must reflect the state of the input
	// stream
	// Special case for text streams : we can make as if ...
	if(m_bText && (samples > node.len - iOffset)) {
		// take our real position
		lSrcPos += node.len - iOffset;
		// and make the source take it too
		source->Seek(lSrcPos);
	} else
		lSrcPos += samples;

#ifdef __MOD_AUDIO_OVERLAP__
	// All streams treated the same way
	// At the end of a part, iOffset will return to 0
	if(samples > node.len - iOffset)
		samples = node.len - iOffset;
#else
	// Treat text streams separately
	if(m_bText && (samples > node.len - iOffset) )
		samples = node.len - iOffset;
#endif

	// Update iOffset accordingly to the method we used
	iOffset += samples;

	while (pfsnCur != subset.end() && iOffset >= pfsnCur->len) {
		iOffset -= node.len;
		++pfsnCur;
	}

	return samples;
}
