//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "ModAudioSourceAC3.h"
#include <vd2/system/Error.h>
#include <vd2/system/math.h>
#include "ProgressDialog.h"
#include "prefs.h"


static const unsigned char AC3_SYNCWORD[AC3_SYNCWORD_LEN] =
{ 0x0B, 0x77};
static const unsigned char AC3_SYNCWORD_BITS[AC3_SYNCWORD_LEN] =
{ 0xFF, 0xFF};


ModAudioSourceAC3::ModAudioSourceAC3(const wchar_t *szFileName, AC3Options options)
: m_AC3Options(options) {
	m_AC3Info.m_Version = 8;
	m_AC3Info.m_SampleRateFlags = 0;
	m_AC3Info.m_SampleRate = 0;
	m_AC3Info.m_Channels = 0;
	m_AC3Info.m_NbChannels = 0;
	m_AC3Info.m_LFE = 0;
	m_AC3Info.m_Frames = 0;
	m_AC3Info.m_FrameList.Clear();
	m_bHasRIFFHeader = false;
	m_RIFFHeaderLength = 0;
	m_FileSize = 0;
	memset(m_AC3Info.m_Buffer, 0, AC3_HEADER_LEN);
	m_AC3Info.m_Bytes = 0;
	m_AC3Info.m_BytesSkipped = 0;
	m_BlockAlign = 1;
	m_NodeSample = 0;
	m_NodeFrame = 0;
	m_NodeByte = 0;
	m_FrameCount = 0;
	m_LastSample = 0;
	m_TotalBytes = 0;
	m_MaxFrameLength = 0;
	m_bVariableFrameLength = false;
	// Output to AVI assumed : NO VBR, BYTE ACCURACY
	m_bVBRHandling = false;
	m_GroupFrames = 1;

	m_SourceType = SOURCE_STD_FILE | SOURCE_AUDIO_AC3;
	clearSourceInfo();
	m_SourceName = szFileName;
	m_SourceDesc = L"AC3 file";
}

ModAudioSourceAC3::~ModAudioSourceAC3(void) {
	m_AC3Info.m_FrameList.DeleteContents(true);
	m_AC3Info.m_FrameList.Clear();
	m_VDFile.close();
}

bool ModAudioSourceAC3::QueryInterface(InterfaceID ID, void** destination) {
	if(ID == IFramedSource::GetInterfaceID()) {
		*destination = static_cast<IFramedSource *>(this);
		return true;
	}
	return false;
}

bool ModAudioSourceAC3::UseFrameMode(void) {
	return m_AC3Options.m_bFrameAccuracy;
}

bool ModAudioSourceAC3::init(void) {
	// open file
	if(!m_VDFile.open(m_SourceName.c_str(),
	nsVDFile::kRead | nsVDFile::kDenyWrite | nsVDFile::kOpenExisting | nsVDFile::kSequential))
		return FALSE;

	m_FileSize = m_VDFile.size();

	ProgressDialog pd(NULL, "AC3 File Import Filter", "Parsing File", (sint32)(m_FileSize/1024), true);
	pd.setValueFormat("%ldK of %ldK");

	CheckTags();

	if(!m_VDFile.seekNT((m_bHasRIFFHeader ? m_RIFFHeaderLength : 0), nsVDFile::kSeekStart))
		return FALSE;

	sint64 bytesSkipped;
	sint64 startSkipped = m_RIFFHeaderLength;
	
	// Get the first valid frame (or at least the first that seem valid ;))
	if(!GetNextFrame(false, bytesSkipped))
		return FALSE;
	if(bytesSkipped) {
		AddCorruptedPart(startSkipped, bytesSkipped);
		startSkipped += bytesSkipped;
	}

	AddAC3Frame();

	// Being unable to read the first frame is not good ;)
	if(!m_VDFile.seekNT(m_AC3Info.m_FrameLength - AC3_HEADER_LEN, nsVDFile::kSeekCur))
		return FALSE;

	m_AC3Info.m_Frames = 1;

	sint32 frame_length = m_AC3Info.m_FrameLength;
	startSkipped += frame_length;
	m_AC3Info.m_Bytes = frame_length;

	while(GetNextFrame(true, bytesSkipped)) {
		// Verify that this frame wasn't cut at the end of the file
		if(m_VDFile.tell()-AC3_HEADER_LEN+m_AC3Info.m_FrameLength > m_FileSize) {
			// The last frame was cut :(
			bytesSkipped = m_FileSize-startSkipped;
			break;
		}

		if(bytesSkipped) {
			AddCorruptedPart(startSkipped, bytesSkipped);
			startSkipped += bytesSkipped;
			// If the skipped data aren't a multiple of frame sizes, we
			// shouldn't use frame mode
			if(m_AC3Options.m_bKeepCorruptedData && (bytesSkipped % frame_length))
				m_bVariableFrameLength = true;
			if(m_AC3Info.m_FrameLength == frame_length)
				AddAC3Frame();
		}

		// We got a new frame

		// If its length differs from the previous ones, add it to the list
		if(m_AC3Info.m_FrameLength != frame_length) {
			AddAC3Frame();
			frame_length = m_AC3Info.m_FrameLength;
			m_bVariableFrameLength = true;
			if(frame_length > m_MaxFrameLength)
				m_MaxFrameLength = frame_length;
		}

		m_AC3Info.m_Frames++;
		startSkipped += frame_length;

		// Skip the frame
		m_VDFile.seek(m_AC3Info.m_FrameLength - AC3_HEADER_LEN, nsVDFile::kSeekCur);

		m_AC3Info.m_Bytes += frame_length;

		pd.advance((sint32)(m_VDFile.tell()/1024));
		pd.check();
	}
	if(bytesSkipped) {
		AddCorruptedPart(startSkipped, bytesSkipped);
		startSkipped += bytesSkipped;
		// If the skipped data aren't a multiple of frame sizes, we
		// shouldn't use frame mode
		if(m_AC3Options.m_bKeepCorruptedData && (bytesSkipped % frame_length))
			m_bVariableFrameLength = true;
	}

	if(	(m_AC3Info.m_Frames <= 0)
	|| (m_AC3Info.m_Version > 8)
	|| (m_AC3Info.m_SampleRate <= 0)
	|| ( (m_AC3Info.m_NbChannels<1) || (m_AC3Info.m_NbChannels>6)) )
		return FALSE;

	// Update the frame count taking into account partial/corrupted data
	UpdateFrameCount();

	m_TotalBytes = m_AC3Info.m_Bytes;
	if(m_AC3Options.m_bKeepCorruptedData)
		m_TotalBytes += m_AC3Info.m_BytesSkipped;

	double fBitrate = (double)m_TotalBytes * 8 * m_AC3Info.m_SampleRate / ((double)1000 * AC3_SAMPLESPERFRAME * m_AC3Info.m_Frames);

	if(!allocFormat(sizeof WAVEFORMATEX/*TENSIBLE*/))
		return FALSE;

	if(m_bVariableFrameLength)
		m_AC3Options.m_bFrameAccuracy = false;
	// By default assume we will be outputed to an AVI => NO VBR, BYTE ACCURACY when reading data
	extend(false, g_VDMPrefs.m_GroupFrames);

	m_SourceDesc = VDStringW::setf(L"AC3 %ldch %ldHz %.1fkbps (bitstream v%d)"
		, m_AC3Info.m_NbChannels, m_AC3Info.m_SampleRate, fBitrate, (int)m_AC3Info.m_Version);

	return TRUE;
}

VDStringW ModAudioSourceAC3::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	m_TotalBytes = m_AC3Info.m_Bytes;
	if(m_AC3Options.m_bKeepCorruptedData)
		m_TotalBytes += m_AC3Info.m_BytesSkipped;
	double fBitrate = (double)m_TotalBytes * 8 * m_AC3Info.m_SampleRate / ((double)1000 * AC3_SAMPLESPERFRAME * m_AC3Info.m_Frames);
	report += VDStringW::setf(L"\r\nAC3 stream : %ld channels, %ldHz, %.1fkbps, bitstream version %d"
		, m_AC3Info.m_NbChannels, m_AC3Info.m_SampleRate, fBitrate, (int)m_AC3Info.m_Version);
	if(m_bHasRIFFHeader)
		report += VDStringW::setf(L"\r\nFound a RIFF header at the beginning of the file, %ld bytes", m_RIFFHeaderLength);
	if(m_AC3Info.m_BytesSkipped) {
		report += VDStringW::setf(L"\r\nA total of %I64d bytes have been seen as corrupted :"
			, m_AC3Info.m_BytesSkipped);
		m_Iterator.begin(m_AC3Info.m_FrameList);
		while(m_Iterator) {
			AC3Frame *curFrame = *m_Iterator;
			if(curFrame->m_Flags == CORRUPTED_AC3_FRAME) {
				sint64 filePos = curFrame->m_FileOffset;
				sint64 bytesSkipped = (m_Iterator.next() ? m_Iterator.next()->m_FileOffset : m_FileSize) - filePos;
				report += VDStringW::setf(L"\r\n %I64d bytes starting at byte %I64d"
					, bytesSkipped, filePos);
			}
			m_Iterator++;
		}
		m_Iterator.begin(m_AC3Info.m_FrameList);
	} else
		report += L"\r\nNo apparent corrupted data";

	return report;
}

void ModAudioSourceAC3::extend(bool bVBRHandling, sint32 groupFrames) {
	m_bVBRHandling = bVBRHandling;
	m_GroupFrames = groupFrames;

	m_TotalBytes = m_AC3Info.m_Bytes;
	if(m_AC3Options.m_bKeepCorruptedData)
		m_TotalBytes += m_AC3Info.m_BytesSkipped;

	sint32 dByterate = VDRoundToLong((double)m_TotalBytes * m_AC3Info.m_SampleRate / (AC3_SAMPLESPERFRAME * m_AC3Info.m_Frames));

	WAVEFORMATEX/*TENSIBLE*/ *wfex = (WAVEFORMATEX/*TENSIBLE*/ *)getWaveFormat();
	memset(wfex, 0, sizeof(WAVEFORMATEX/*TENSIBLE*/));
	wfex->/*Format.*/cbSize = sizeof WAVEFORMATEX/*TENSIBLE*/ - sizeof WAVEFORMATEX;
	//wfex->Samples.wReserved = 0;
	wfex->/*Format.*/nAvgBytesPerSec = (DWORD)VDRoundToLong(dByterate);
	wfex->/*Format.*/nChannels = m_AC3Info.m_NbChannels;
	wfex->/*Format.*/nSamplesPerSec = m_AC3Info.m_SampleRate;
	wfex->/*Format.*/wBitsPerSample = 0;
	wfex->/*Format.*/wFormatTag = 0x2000;
	// Set SubFormat and dwChannelMask ?
	if(!m_bVariableFrameLength &&
	(m_bVBRHandling || (!m_bVBRHandling && m_AC3Options.m_bFrameAccuracy)) ) {
		wfex->/*Format.*/nBlockAlign = m_AC3Info.m_FrameLength;
		streamInfo.dwSampleSize = m_AC3Info.m_FrameLength;
	} else {
		wfex->/*Format.*/nBlockAlign = 1;
	}

	lSampleFirst = 0;

	streamInfo.fccType         = streamtypeAUDIO;
	streamInfo.fccHandler      = 0;
	streamInfo.dwFlags         = 0;
	streamInfo.wPriority       = 0;
	streamInfo.wLanguage       = 0;
	streamInfo.dwInitialFrames = 0;
	streamInfo.dwStart         = 0;
	streamInfo.dwQuality       = 0xFFFFFFFF;
	streamInfo.dwSuggestedBufferSize = 0;

	sint32 bytesPerSample = wfex->/*Format.*/nBlockAlign;
	if(m_bVBRHandling) {
		lSampleLast = m_AC3Info.m_Frames * AC3_SAMPLESPERFRAME;

		streamInfo.dwScale         = 1;
		streamInfo.dwRate          = m_AC3Info.m_SampleRate;
		streamInfo.dwSampleSize    = bytesPerSample;
	} else {	// Handle as a standard CBR stream
		lSampleLast = m_TotalBytes / bytesPerSample;

		streamInfo.dwScale         = bytesPerSample;
		streamInfo.dwRate          = wfex->/*Format.*/nAvgBytesPerSec;
		streamInfo.dwSampleSize    = bytesPerSample;
	}
	streamInfo.dwLength        = lSampleLast - lSampleFirst;

	m_LastSample = lSampleLast;

	m_BlockAlign = wfex->/*Format.*/nBlockAlign;

	m_Iterator.begin(m_AC3Info.m_FrameList);
	m_NodeSample = 0;
	m_NodeFrame = 0;
	m_NodeByte = 0;
	ComputeFramesInNode();
}

int ModAudioSourceAC3::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {
	if(lBytesRead)
		*lBytesRead = 0;
	if(lSamplesRead)
		*lSamplesRead = 0;
	if(!lCount)
		return AVIERR_OK;
	sint32 framesLength, frameIndex, frameLength;
	sint32 readLength = 0;
	if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
		// Too bad
		return AVIERR_OK;

	// Treat in my order of preference :
	if(m_bVBRHandling) {	// Correctly handle this stream in any case (CBR / VBR)
		// Request by sample number
		if(m_GroupFrames > 0) {
			if( (lCount==AVISTREAMREAD_CONVENIENT) || (lCount<m_GroupFrames*AC3_SAMPLESPERFRAME) )
				lCount = m_GroupFrames * AC3_SAMPLESPERFRAME;
		} else
			lCount = AC3_SAMPLESPERFRAME;

		while(true) {
			if(lpBuffer) {
				if(cbBuffer < frameLength) {
					// If buffer too short, tell the one that should always work
					if(lBytesRead)
						*lBytesRead += ((lCount+AC3_SAMPLESPERFRAME-1)/AC3_SAMPLESPERFRAME) * AC3_MAX_LENGTH;
					return AVIERR_BUFFERTOOSMALL;
				}

				m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
				if((readLength=m_VDFile.readData(lpBuffer, frameLength))!=frameLength) {
					if(lBytesRead)
						*lBytesRead += readLength;
					return AVIERR_FILEREAD;
				}

				lpBuffer = (char *)lpBuffer + readLength;
			}

			if(lBytesRead)
				*lBytesRead += frameLength;
			if(lSamplesRead)
				*lSamplesRead += AC3_SAMPLESPERFRAME;

			cbBuffer -= frameLength;
			lCount -= AC3_SAMPLESPERFRAME;
			lStart += AC3_SAMPLESPERFRAME;

			if(lCount <= 0)
				return AVIERR_OK;

			if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
				return AVIERR_OK;
		}
	} else if(!m_AC3Options.m_bFrameAccuracy) {	// Handle as a standard CBR stream
		// Request by block number (=> number of bytes)
		if(lCount == AVISTREAMREAD_CONVENIENT)
			lCount = 1;

		if(lBytesRead)
			*lBytesRead = frameLength * lCount;
		if(lSamplesRead)
			*lSamplesRead = lCount;
		if(!lpBuffer)
			return AVIERR_OK;

		if(cbBuffer < frameLength * lCount)
			return AVIERR_BUFFERTOOSMALL;

		m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
		if((readLength=m_VDFile.readData(lpBuffer, frameLength*lCount))!=frameLength*lCount) {
			if(lBytesRead)
				*lBytesRead = readLength;
			return AVIERR_FILEREAD;
		}

		return AVIERR_OK;
	} else {
		// Request by frame number
		if(m_GroupFrames > 0) {
			if( (lCount==AVISTREAMREAD_CONVENIENT) || (lCount<m_GroupFrames) )
				lCount = m_GroupFrames;
		} else
			lCount = 1;

		while(true) {
			if(lpBuffer) {
				if(cbBuffer < frameLength) {
					// If buffer too short, tell the one that should always work
					if(lBytesRead)
						*lBytesRead += lCount * AC3_MAX_LENGTH;
					return AVIERR_BUFFERTOOSMALL;
				}

				m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
				if((readLength=m_VDFile.readData(lpBuffer, frameLength))!=frameLength) {
					if(lBytesRead)
						*lBytesRead += readLength;
					return AVIERR_FILEREAD;
				}

				lpBuffer = (char *)lpBuffer + readLength;
			}

			if(lBytesRead)
				*lBytesRead += frameLength;
			if(lSamplesRead)
				*lSamplesRead += 1;

			cbBuffer -= frameLength;
			lCount--;
			lStart++;

			if(lCount <= 0)
				return AVIERR_OK;

			if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
				return AVIERR_OK;
		}
	}
}

bool ModAudioSourceAC3::SeekSample(LONG lStart, sint32& framesLength, sint32& frameIndex, sint32& frameLength) {
	if(lStart >= m_LastSample)
		return false;

	if(m_bVBRHandling) {
		while((lStart < m_NodeSample) && (m_Iterator.previous())) {
			m_Iterator--;
			ComputeFramesInNode();
			m_NodeSample -= m_FrameCount*AC3_SAMPLESPERFRAME;
		}
		while(m_NodeSample + m_FrameCount*AC3_SAMPLESPERFRAME <= lStart) {
			m_Iterator++;
			m_NodeSample += m_FrameCount*AC3_SAMPLESPERFRAME;
			ComputeFramesInNode();
		}
		framesLength = ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
		frameIndex = (lStart - m_NodeSample) / AC3_SAMPLESPERFRAME;
		if( ((*m_Iterator)->m_Flags==CORRUPTED_AC3_FRAME) && (frameIndex==m_FrameCount-1) ) {
			AC3Frame *next = m_Iterator.next();
			frameLength = (next?next->m_FileOffset:m_FileSize)-((*m_Iterator)->m_FileOffset+frameIndex*framesLength);
		} else
			frameLength = framesLength;

		return true;
	} else if(!m_AC3Options.m_bFrameAccuracy) {
		while(lStart < m_NodeByte / m_BlockAlign) {
			sint64 pos = (*m_Iterator)->m_FileOffset;
			m_Iterator--;
			m_NodeByte -= pos-(*m_Iterator)->m_FileOffset;
		}
		AC3Frame *next = m_Iterator.next();
		while( (m_NodeByte + (next?next->m_FileOffset:m_FileSize)-(*m_Iterator)->m_FileOffset) / m_BlockAlign <= lStart) {
			m_NodeByte += (next?next->m_FileOffset:m_FileSize) - (*m_Iterator)->m_FileOffset;
			m_Iterator++;
			next = m_Iterator.next();
		}
		framesLength = 1;
		frameIndex = lStart * m_BlockAlign - m_NodeByte;
		frameLength = m_BlockAlign;

		return true;
	} else {
		while(lStart < m_NodeFrame) {
			m_Iterator--;
			ComputeFramesInNode();
			m_NodeFrame -= m_FrameCount;
		}
		while(m_NodeFrame + m_FrameCount <= lStart) {
			m_Iterator++;
			m_NodeFrame += m_FrameCount;
			ComputeFramesInNode();
		}
		framesLength = ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
		frameIndex = lStart - m_NodeFrame;
		if( ((*m_Iterator)->m_Flags==CORRUPTED_AC3_FRAME) && (frameIndex==m_FrameCount-1) ) {
			AC3Frame *next = m_Iterator.next();
			frameLength = (next?next->m_FileOffset:m_FileSize)-((*m_Iterator)->m_FileOffset+frameIndex*framesLength);
		} else
			frameLength = framesLength;

		return true;
	}

	return false;
}

sint64 ModAudioSourceAC3::sampleOffset(sint64 lSample) {
	if(lSample >= m_LastSample)
		return m_TotalBytes;
	sint32 framesLength, frameIndex, frameLength;
	if(!SeekSample(lSample, framesLength, frameIndex, frameLength))
		return 0;
	return ((*m_Iterator)->m_FileOffset + frameIndex*framesLength);
}

sint32 ModAudioSourceAC3::samplesLag(sint64 lStart) {
	sint32 framesLength, frameIndex, frameLength;
	if((lStart<=0) || !m_bVBRHandling || !SeekSample(lStart, framesLength, frameIndex, frameLength))
		return 0;

	return ( (m_NodeSample + frameIndex * AC3_SAMPLESPERFRAME) - lStart);
}

void ModAudioSourceAC3::CheckTags(void) {
	char buffer[1024];

	// Is there a RIFF header at the beginning ?
	m_VDFile.seek(0, nsVDFile::kSeekStart);
	if(m_VDFile.readData(buffer, 12) != 12)
		return;
	if( (*(long *)(buffer+0)=='FFIR') && (*(long *)(buffer+8)=='EVAW') ) {
		while(true) {
			if(m_VDFile.readData(buffer, 8) != 8)
				return;
			if(*(long *)(buffer+0) == 'atad') {
				m_bHasRIFFHeader = true;
				m_RIFFHeaderLength = m_VDFile.tell();
				break;
			} else {
				sint64 chunkLength = (buffer[7]<<24)
					+ (buffer[6]<<16)
					+ (buffer[5]<<8)
					+ (buffer[4]);
				if(!m_VDFile.skipNT(chunkLength))
					return;
			}
		}
	}

	m_VDFile.seek(0, nsVDFile::kSeekStart);
}

void ModAudioSourceAC3::AddAC3Frame(void) {
	AC3Frame *frame = m_AC3Info.m_FrameList.AddLast();
	frame->m_FileOffset = m_VDFile.tell() - AC3_HEADER_LEN;
	// Masks :
	// 0xC0 -> samplerate flags
	// 0x3F -> bitrate flags
	frame->m_Flags = m_AC3Info.m_Buffer[4];
}

void ModAudioSourceAC3::AddCorruptedPart(sint64 startCorrupted, sint64 bytesSkipped) {
	m_AC3Info.m_BytesSkipped += bytesSkipped;
	AC3Frame *frame = m_AC3Info.m_FrameList.AddLast();
	frame->m_FileOffset = startCorrupted;
	frame->m_Flags = CORRUPTED_AC3_FRAME;
}

void ModAudioSourceAC3::UpdateFrameCount(void) {
	ModList<AC3Frame>::iterator it(m_AC3Info.m_FrameList);
	AC3Frame *prev = NULL;
	AC3Frame *current = *it;
	AC3Frame *next = NULL;

	while(current) {
		next = it.next();
		if( (current->m_Flags == CORRUPTED_AC3_FRAME) && (m_AC3Options.m_bKeepCorruptedData) ) {
			sint64 sizeData = (next ? next->m_FileOffset : m_FileSize) - current->m_FileOffset;
			m_AC3Info.m_Frames += VDRoundToLong((double)sizeData / ComputeFrameLength(prev, current, next));
		}
		prev = current;
		current = next;
		it++;
	}
}

sint32 ModAudioSourceAC3::ComputeFrameLength(AC3Frame *previous, AC3Frame *current, AC3Frame *next) {
	if(!current)
		return 0;

	sint32 sizeFrame = GetFrameLength(current);
	if(sizeFrame <= 0) {
		sint32 sizePrev = GetFrameLength(previous);
		sint32 sizeNext = GetFrameLength(next);
		if(sizePrev <= 0)
			sizeNext *= 2;
		if(sizeNext <= 0)
			sizePrev *= 2;
		return VDRoundToLong(((double)sizePrev + sizeNext)/2);
	} else
		return sizeFrame;
}

sint32 ModAudioSourceAC3::GetFrameLength(AC3Frame *node) {
	if(!node)
		return 0;

	// Reminder :
	// flags & 0xC0 -> samplerate flags
	// flags & 0x3F -> bitrate flags
	sint32 fscod = (node->m_Flags & 0xC0)>>6;
	sint32 frmsizecod = (node->m_Flags & 0x3F);
	if(fscod == 3)
		return 0;
	if(frmsizecod >= 38)
		return 0;

	return (2 * AC3_FrameLength[frmsizecod][fscod]);
}

bool ModAudioSourceAC3::GetNextFrame(bool ensureConsistancy, sint64& bytesSkipped) {
	bytesSkipped = 0;
	sint64 skipped;
	while(true) {
		if(!m_VDFile.synchOn(AC3_SYNCWORD, AC3_SYNCWORD_BITS, AC3_SYNCWORD_LEN, 2*AC3_MAX_LENGTH, skipped)) {
			bytesSkipped += skipped;
			return false;
		}
		bytesSkipped += skipped;
		if(GetHeader(ensureConsistancy))
			return true;
		bytesSkipped += 1;
		m_VDFile.seek(-AC3_HEADER_LEN+1, nsVDFile::kSeekCur);
	}
}

bool ModAudioSourceAC3::GetHeader(bool ensureConsistancy) {
	if(m_VDFile.tell() >= m_FileSize)
		return false;
	char *buffer = m_AC3Info.m_Buffer;
	// Read the header (7 bytes)
	if(m_VDFile.readData(buffer, AC3_HEADER_LEN) != AC3_HEADER_LEN)
		return false;

	// Test the syncword
	if( ((buffer[0] & AC3_SYNCWORD_BITS[0]) != AC3_SYNCWORD[0])
	|| ((buffer[1] & AC3_SYNCWORD_BITS[1]) != AC3_SYNCWORD[1]) )
		return false;

	sint32 fscod;
	sint32 frmsizecod;
	sint32 bsid;
	sint32 acmod;
	sint32 lfeon;

	fscod = (buffer[4] & 0xC0)>>6;
	frmsizecod = (buffer[4] & 0x3F);
	bsid = (buffer[5] & 0xF8)>>3;
	acmod = (buffer[6] & 0xE0)>>5;
	switch(acmod) {
	case 0x1:	// nothing
		lfeon = (buffer[6] & 0x10)>>4;
		break;
	case 0x5:	// surmixlvl & cmixlvl
	case 0x7:	// surmixlvl & cmixlvl
		lfeon = (buffer[6] & 0x01);
		break;
	default:
	//case 0x2:	// surmod
	//case 0x3:	// cmixlvl
	//case 0x4:	// surmixlvl
	//case 0x6:	// surmixlvl
		lfeon = (buffer[6] & 0x04)>>2;
		break;
	}

	sint32 samplerate = AC3_SampleRate[fscod];

	if(bsid > 8)
		return false;

	if(samplerate <= 0)
		return false;

	if(frmsizecod >= 38)
		return false;

	sint32 nb_channels = AC3_NbChannels[acmod] + lfeon;
	sint32 bitrate = AC3_Bitrate[frmsizecod];
	sint32 frame_length = AC3_FrameLength[frmsizecod][fscod];

	// If asked, verify the stream remain consistant
	if(ensureConsistancy && (
		(bsid != m_AC3Info.m_Version)
		|| (fscod != m_AC3Info.m_SampleRateFlags)
		|| (acmod != m_AC3Info.m_Channels)
		|| (lfeon != m_AC3Info.m_LFE) ) )
		return false;

	m_AC3Info.m_Version = bsid;
	m_AC3Info.m_Bitrate = bitrate;
	m_AC3Info.m_SampleRateFlags = fscod;
	m_AC3Info.m_SampleRate = samplerate;
	m_AC3Info.m_Channels = acmod;
	m_AC3Info.m_NbChannels = nb_channels;
	m_AC3Info.m_LFE = lfeon;

	m_AC3Info.m_FrameLength = frame_length * 2;

	return true;
}

void ModAudioSourceAC3::ComputeFramesInNode(void) {
	AC3Frame *next = m_Iterator.next();
	if( ((*m_Iterator)->m_Flags == CORRUPTED_AC3_FRAME) && !m_AC3Options.m_bKeepCorruptedData)
		m_FrameCount = 0;
	else
		m_FrameCount = VDRoundToLong( ((double)(next?next->m_FileOffset:m_FileSize)-(*m_Iterator)->m_FileOffset) / ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next()));
}
