//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "ModAudioSourceDTS.h"
#include <vd2/system/Error.h>
#include <vd2/system/math.h>
#include "ProgressDialog.h"
#include "prefs.h"


static const unsigned char DTS_SYNCWORD[DTS_SYNCWORD_LEN] =
{ 0x7F, 0xFE, 0x80, 0x01};
static const unsigned char DTS_SYNCWORD_BITS[DTS_SYNCWORD_LEN] =
{ 0xFF, 0xFF, 0xFF, 0xFF};


ModAudioSourceDTS::ModAudioSourceDTS(const wchar_t *szFileName, DTSOptions options)
: m_DTSOptions(options) {
	m_DTSInfo.m_SampleRateFlags = 0;
	m_DTSInfo.m_SampleRate = 0;
	m_DTSInfo.m_TotalSamples = 0;
	m_DTSInfo.m_Channels = 0;
	m_DTSInfo.m_NbChannels = 0;
	m_DTSInfo.m_Frames = 0;
	m_DTSInfo.m_FrameList.Clear();
	m_bHasRIFFHeader = false;
	m_RIFFHeaderLength = 0;
	m_FileSize = 0;
	memset(m_DTSInfo.m_Buffer, 0, DTS_HEADER_LEN);
	m_DTSInfo.m_Bytes = 0;
	m_DTSInfo.m_BytesSkipped = 0;
	m_BlockAlign = 1;
	m_NodeSample = 0;
	m_NodeFrame = 0;
	m_NodeByte = 0;
	m_FrameCount = 0;
	m_LastSample = 0;
	m_TotalBytes = 0;
	m_MaxFrameLength = 0;
	m_bVariableFrameLength = false;
	// Output to AVI assumed : NO VBR, BYTE ACCURACY
	m_bVBRHandling = false;
	m_GroupFrames = 1;

	m_SourceType = SOURCE_STD_FILE | SOURCE_AUDIO_DTS;
	clearSourceInfo();
	m_SourceName = szFileName;
	m_SourceDesc = L"DTS file";
}

ModAudioSourceDTS::~ModAudioSourceDTS(void) {
	m_DTSInfo.m_FrameList.DeleteContents(true);
	m_DTSInfo.m_FrameList.Clear();
	m_VDFile.close();
}

bool ModAudioSourceDTS::QueryInterface(InterfaceID ID, void** destination) {
	if(ID == IFramedSource::GetInterfaceID()) {
		*destination = static_cast<IFramedSource *>(this);
		return true;
	}
	return false;
}

bool ModAudioSourceDTS::UseFrameMode(void) {
	return m_DTSOptions.m_bFrameAccuracy;
}

bool ModAudioSourceDTS::init(void) {
	// open file
	if(!m_VDFile.open(m_SourceName.c_str(),
	nsVDFile::kRead | nsVDFile::kDenyWrite | nsVDFile::kOpenExisting | nsVDFile::kSequential))
		return FALSE;

	m_FileSize = m_VDFile.size();

	ProgressDialog pd(NULL, "DTS File Import Filter", "Parsing File", (sint32)(m_FileSize/1024), true);
	pd.setValueFormat("%ldK of %ldK");

	CheckTags();

	if(!m_VDFile.seekNT((m_bHasRIFFHeader ? m_RIFFHeaderLength : 0), nsVDFile::kSeekStart))
		return FALSE;

	sint64 bytesSkipped;
	sint64 startSkipped = m_RIFFHeaderLength;

	// Get the first valid frame (or at least the first that seem valid ;))
	if(!GetNextFrame(false, bytesSkipped))
		return FALSE;
	if(bytesSkipped) {
		AddCorruptedPart(startSkipped, bytesSkipped);
		startSkipped += bytesSkipped;
	}

	AddDTSFrame();

	// Being unable to read the first frame is not good ;)
	if(!m_VDFile.seekNT(m_DTSInfo.m_FrameLength - DTS_HEADER_LEN, nsVDFile::kSeekCur))
		return FALSE;

	m_DTSInfo.m_Frames = 1;

	sint32 frame_length = m_DTSInfo.m_FrameLength;
	sint32 frame_size = m_DTSInfo.m_Samples;
	startSkipped += frame_length;
	m_DTSInfo.m_Bytes = frame_length;
	m_DTSInfo.m_TotalSamples = frame_size;

	while(GetNextFrame(true, bytesSkipped)) {
		// Verify that this frame wasn't cut at the end of the file
		if(m_VDFile.tell()-DTS_HEADER_LEN+m_DTSInfo.m_FrameLength > m_FileSize) {
			// The last frame was cut :(
			bytesSkipped = m_FileSize-startSkipped;
			break;
		}

		if(bytesSkipped) {
			AddCorruptedPart(startSkipped, bytesSkipped);
			startSkipped += bytesSkipped;
			// If the skipped data aren't a multiple of frame sizes, we
			// shouldn't use frame mode
			if(m_DTSOptions.m_bKeepCorruptedData && (bytesSkipped % frame_length))
				m_bVariableFrameLength = true;
			if( (m_DTSInfo.m_FrameLength==frame_length) && (m_DTSInfo.m_Samples==frame_size) )
				AddDTSFrame();
		}

		// We got a new frame

		// If its length differs from the previous ones, add it to the list
		if( (m_DTSInfo.m_FrameLength != frame_length) || (m_DTSInfo.m_Samples != frame_size) ) {
			AddDTSFrame();
			frame_length = m_DTSInfo.m_FrameLength;
			frame_size = m_DTSInfo.m_Samples;
			m_bVariableFrameLength = true;
			if(frame_length > m_MaxFrameLength)
				m_MaxFrameLength = frame_length;
		}

		m_DTSInfo.m_Frames++;
		startSkipped += frame_length;

		// Skip the frame
		m_VDFile.seek(m_DTSInfo.m_FrameLength - DTS_HEADER_LEN, nsVDFile::kSeekCur);

		m_DTSInfo.m_Bytes += frame_length;
		m_DTSInfo.m_TotalSamples += frame_size;

		pd.advance((sint32)(m_VDFile.tell()/1024));
		pd.check();
	}
	if(bytesSkipped) {
		AddCorruptedPart(startSkipped, bytesSkipped);
		startSkipped += bytesSkipped;
		// If the skipped data aren't a multiple of frame sizes, we
		// shouldn't use frame mode
		if(m_DTSOptions.m_bKeepCorruptedData && (bytesSkipped % frame_length))
			m_bVariableFrameLength = true;
	}

	if(	(m_DTSInfo.m_Frames <= 0)
	|| (m_DTSInfo.m_SampleRate <= 0)
	|| ( (m_DTSInfo.m_NbChannels<1) || (m_DTSInfo.m_NbChannels>5)) )
		return FALSE;

	// Update the frame count taking into account partial/corrupted data
	UpdateFrameCount();

	m_TotalBytes = m_DTSInfo.m_Bytes;
	if(m_DTSOptions.m_bKeepCorruptedData)
		m_TotalBytes += m_DTSInfo.m_BytesSkipped;

	double fBitrate = (double)m_TotalBytes * 8 * m_DTSInfo.m_SampleRate / ((double)1000 * m_DTSInfo.m_TotalSamples);

	if(!allocFormat(sizeof WAVEFORMATEX/*TENSIBLE*/))
		return FALSE;

	if(m_bVariableFrameLength)
		m_DTSOptions.m_bFrameAccuracy = false;
	// By default assume we will be outputed to an AVI => NO VBR, BYTE ACCURACY when reading data
	extend(false, g_VDMPrefs.m_GroupFrames);

	m_SourceDesc = VDStringW::setf(L"DTS %ldch %ldHz %.1fkbps"
		, m_DTSInfo.m_NbChannels, m_DTSInfo.m_SampleRate, fBitrate);

	return TRUE;
}

VDStringW ModAudioSourceDTS::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	m_TotalBytes = m_DTSInfo.m_Bytes;
	if(m_DTSOptions.m_bKeepCorruptedData)
		m_TotalBytes += m_DTSInfo.m_BytesSkipped;
	double fBitrate = (double)m_TotalBytes * 8 * m_DTSInfo.m_SampleRate / ((double)1000 * m_DTSInfo.m_TotalSamples);
	report += VDStringW::setf(L"\r\nDTS stream : %ld channels, %ldHz, %.1fkbps"
		, m_DTSInfo.m_NbChannels, m_DTSInfo.m_SampleRate, fBitrate);
	if(m_bHasRIFFHeader)
		report += VDStringW::setf(L"\r\nFound a RIFF header at the beginning of the file, %ld bytes", m_RIFFHeaderLength);
	if(m_DTSInfo.m_BytesSkipped) {
		report += VDStringW::setf(L"\r\nA total of %I64d bytes have been seen as corrupted :"
			, m_DTSInfo.m_BytesSkipped);
		m_Iterator.begin(m_DTSInfo.m_FrameList);
		while(m_Iterator) {
			DTSFrame *curFrame = *m_Iterator;
			if(curFrame->m_Short == CORRUPTED_DTS_FRAME) {
				sint64 filePos = curFrame->m_FileOffset;
				sint64 bytesSkipped = (m_Iterator.next() ? m_Iterator.next()->m_FileOffset : m_FileSize) - filePos;
				report += VDStringW::setf(L"\r\n %I64d bytes starting at byte %I64d"
					, bytesSkipped, filePos);
			}
			m_Iterator++;
		}
		m_Iterator.begin(m_DTSInfo.m_FrameList);
	} else
		report += L"\r\nNo apparent corrupted data";

	return report;
}

void ModAudioSourceDTS::extend(bool bVBRHandling, sint32 groupFrames) {
	m_bVBRHandling = bVBRHandling;
	m_GroupFrames = groupFrames;

	m_TotalBytes = m_DTSInfo.m_Bytes;
	if(m_DTSOptions.m_bKeepCorruptedData)
		m_TotalBytes += m_DTSInfo.m_BytesSkipped;

	sint32 dByterate = VDRoundToLong((double)m_TotalBytes * m_DTSInfo.m_SampleRate / m_DTSInfo.m_TotalSamples);

	WAVEFORMATEX/*TENSIBLE*/ *wfex = (WAVEFORMATEX/*TENSIBLE*/ *)getWaveFormat();
	memset(wfex, 0, sizeof(WAVEFORMATEX/*TENSIBLE*/));
	wfex->/*Format.*/cbSize = sizeof WAVEFORMATEX/*TENSIBLE*/ - sizeof WAVEFORMATEX;
	//wfex->Samples.wReserved = 0;
	wfex->/*Format.*/nAvgBytesPerSec = (DWORD)VDRoundToLong(dByterate);
	wfex->/*Format.*/nChannels = m_DTSInfo.m_NbChannels;
	wfex->/*Format.*/nSamplesPerSec = m_DTSInfo.m_SampleRate;
	wfex->/*Format.*/wBitsPerSample = 0;
	wfex->/*Format.*/wFormatTag = 0x2001;
	// Set SubFormat and dwChannelMask
	if(!m_bVariableFrameLength &&
	(m_bVBRHandling || (!m_bVBRHandling && m_DTSOptions.m_bFrameAccuracy)) ) {
		wfex->/*Format.*/nBlockAlign = m_DTSInfo.m_FrameLength;
		streamInfo.dwSampleSize = m_DTSInfo.m_FrameLength;
	} else {
		wfex->/*Format.*/nBlockAlign = 1;
	}

	lSampleFirst = 0;

	streamInfo.fccType         = streamtypeAUDIO;
	streamInfo.fccHandler      = 0;
	streamInfo.dwFlags         = 0;
	streamInfo.wPriority       = 0;
	streamInfo.wLanguage       = 0;
	streamInfo.dwInitialFrames = 0;
	streamInfo.dwStart         = 0;
	streamInfo.dwQuality       = 0xFFFFFFFF;
	streamInfo.dwSuggestedBufferSize = 0;

	sint32 bytesPerSample = wfex->/*Format.*/nBlockAlign;
	if(m_bVBRHandling) {
		lSampleLast = m_DTSInfo.m_TotalSamples;

		streamInfo.dwScale         = 1;
		streamInfo.dwRate          = m_DTSInfo.m_SampleRate;
		streamInfo.dwSampleSize    = bytesPerSample;
	} else {	// Handle as a standard CBR stream
		lSampleLast = m_TotalBytes / bytesPerSample;

		streamInfo.dwScale         = bytesPerSample;
		streamInfo.dwRate          = wfex->/*Format.*/nAvgBytesPerSec;
		streamInfo.dwSampleSize    = bytesPerSample;
	}
	streamInfo.dwLength        = lSampleLast - lSampleFirst;

	m_LastSample = lSampleLast;

	m_BlockAlign = wfex->/*Format.*/nBlockAlign;

	m_Iterator.begin(m_DTSInfo.m_FrameList);
	m_NodeSample = 0;
	m_NodeFrame = 0;
	m_NodeByte = 0;
	ComputeFramesInNode();
}

int ModAudioSourceDTS::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {
	if(lBytesRead)
		*lBytesRead = 0;
	if(lSamplesRead)
		*lSamplesRead = 0;
	if(!lCount)
		return AVIERR_OK;
	sint32 framesLength, frameIndex, frameLength;
	sint32 readLength = 0;
	if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
		// Too bad
		return AVIERR_OK;

	// Treat in my order of preference :
	if(m_bVBRHandling) {	// Correctly handle this stream in any case (CBR / VBR)
		// Request by sample number
		if(m_GroupFrames == 0)
			lCount = 1;

		long group = m_GroupFrames;

		while(true) {
			if(lpBuffer) {
				if(cbBuffer < frameLength) {
					// If buffer too short, tell the one that should work
					if(lBytesRead)
						*lBytesRead += ((lCount+m_SampleCount-1)/m_SampleCount) * DTS_MAX_LENGTH;
					return AVIERR_BUFFERTOOSMALL;
				}

				m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
				if((readLength=m_VDFile.readData(lpBuffer, frameLength))!=frameLength) {
					if(lBytesRead)
						*lBytesRead += readLength;
					return AVIERR_FILEREAD;
				}

				lpBuffer = (char *)lpBuffer + readLength;
			}

			if(lBytesRead)
				*lBytesRead += readLength;
			if(lSamplesRead)
				*lSamplesRead += m_SampleCount;

			cbBuffer -= readLength;
			lCount -= m_SampleCount;
			lStart += m_SampleCount;

			if( (lCount <= 0) && (group <= 0 ) )
				return AVIERR_OK;

			group--;

			if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
				return AVIERR_OK;
		}
	} else if(!m_DTSOptions.m_bFrameAccuracy) {	// Handle as a standard CBR stream
		// Request by block number (=> number of bytes)
		if(lCount == AVISTREAMREAD_CONVENIENT)
			lCount = 1;

		if(lBytesRead)
			*lBytesRead = frameLength * lCount;
		if(lSamplesRead)
			*lSamplesRead = lCount;
		if(!lpBuffer)
			return AVIERR_OK;

		if(cbBuffer < frameLength * lCount)
			return AVIERR_BUFFERTOOSMALL;

		m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
		if((readLength=m_VDFile.readData(lpBuffer, frameLength*lCount))!=frameLength*lCount) {
			if(lBytesRead)
				*lBytesRead = readLength;
			return AVIERR_FILEREAD;
		}

		return AVIERR_OK;
	} else {
		// Request by frame number
		if(m_GroupFrames > 0) {
			if( (lCount==AVISTREAMREAD_CONVENIENT) || (lCount<m_GroupFrames) )
				lCount = m_GroupFrames;
		} else
			lCount = 1;

		while(true) {
			if(lpBuffer) {
				if(cbBuffer < frameLength) {
					// If buffer too short, tell the one that should always work
					if(lBytesRead)
						*lBytesRead += lCount * DTS_MAX_LENGTH;
					return AVIERR_BUFFERTOOSMALL;
				}

				m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
				if((readLength=m_VDFile.readData(lpBuffer, frameLength))!=frameLength) {
					if(lBytesRead)
						*lBytesRead += readLength;
					return AVIERR_FILEREAD;
				}

				lpBuffer = (char *)lpBuffer + readLength;
			}

			if(lBytesRead)
				*lBytesRead += frameLength;
			if(lSamplesRead)
				*lSamplesRead += 1;

			cbBuffer -= frameLength;
			lCount--;
			lStart++;

			if(lCount <= 0)
				return AVIERR_OK;

			if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
				return AVIERR_OK;
		}
	}
}

bool ModAudioSourceDTS::SeekSample(LONG lStart, sint32& framesLength, sint32& frameIndex, sint32& frameLength) {
	if(lStart >= m_LastSample)
		return false;

	if(m_bVBRHandling) {
		while((lStart < m_NodeSample) && (m_Iterator.previous())) {
			m_Iterator--;
			ComputeFramesInNode();
			m_NodeSample -= m_FrameCount*m_SampleCount;
		}
		while(m_NodeSample + m_FrameCount*m_SampleCount <= lStart) {
			m_Iterator++;
			m_NodeSample += m_FrameCount*m_SampleCount;
			ComputeFramesInNode();
		}
		framesLength = ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
		frameIndex = (lStart - m_NodeSample) / m_SampleCount;
		if( ((*m_Iterator)->m_Short==CORRUPTED_DTS_FRAME) && (frameIndex==m_FrameCount-1) ) {
			DTSFrame *next = m_Iterator.next();
			frameLength = (next?next->m_FileOffset:m_FileSize)-((*m_Iterator)->m_FileOffset+frameIndex*framesLength);
		} else
			frameLength = framesLength;

		return true;
	} else if(!m_DTSOptions.m_bFrameAccuracy) {
		while(lStart < m_NodeByte / m_BlockAlign) {
			sint64 pos = (*m_Iterator)->m_FileOffset;
			m_Iterator--;
			m_NodeByte -= pos-(*m_Iterator)->m_FileOffset;
		}
		DTSFrame *next = m_Iterator.next();
		while( (m_NodeByte + (next?next->m_FileOffset:m_FileSize)-(*m_Iterator)->m_FileOffset) / m_BlockAlign <= lStart) {
			m_NodeByte += (next?next->m_FileOffset:m_FileSize) - (*m_Iterator)->m_FileOffset;
			m_Iterator++;
			next = m_Iterator.next();
		}
		framesLength = 1;
		frameIndex = lStart * m_BlockAlign - m_NodeByte;
		frameLength = m_BlockAlign;

		return true;
	} else {
		while(lStart < m_NodeFrame) {
			m_Iterator--;
			ComputeFramesInNode();
			m_NodeFrame -= m_FrameCount;
		}
		while(m_NodeFrame + m_FrameCount <= lStart) {
			m_Iterator++;
			m_NodeFrame += m_FrameCount;
			ComputeFramesInNode();
		}
		framesLength = ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
		frameIndex = lStart - m_NodeFrame;
		if( ((*m_Iterator)->m_Short==CORRUPTED_DTS_FRAME) && (frameIndex==m_FrameCount-1) ) {
			DTSFrame *next = m_Iterator.next();
			frameLength = (next?next->m_FileOffset:m_FileSize)-((*m_Iterator)->m_FileOffset+frameIndex*framesLength);
		} else
			frameLength = framesLength;

		return true;
	}

	return false;
}

sint64 ModAudioSourceDTS::sampleOffset(sint64 lSample) {
	if(lSample >= m_LastSample)
		return m_TotalBytes;
	sint32 framesLength, frameIndex, frameLength;
	if(!SeekSample(lSample, framesLength, frameIndex, frameLength))
		return 0;
	return ((*m_Iterator)->m_FileOffset + frameIndex*framesLength);
}

sint32 ModAudioSourceDTS::samplesLag(sint64 lStart) {
	sint32 framesLength, frameIndex, frameLength;
	if((lStart<=0) || !m_bVBRHandling || !SeekSample(lStart, framesLength, frameIndex, frameLength))
		return 0;

	return ( (m_NodeSample + frameIndex * m_SampleCount) - lStart);
}

void ModAudioSourceDTS::CheckTags(void) {
	char buffer[1024];

	// Is there a RIFF header at the beginning ?
	m_VDFile.seek(0, nsVDFile::kSeekStart);
	if(m_VDFile.readData(buffer, 12) != 12)
		return;
	if( (*(long *)(buffer+0)=='FFIR') && (*(long *)(buffer+8)=='EVAW') ) {
		while(true) {
			if(m_VDFile.readData(buffer, 8) != 8)
				return;
			if(*(long *)(buffer+0) == 'atad') {
				m_bHasRIFFHeader = true;
				m_RIFFHeaderLength = m_VDFile.tell();
				break;
			} else {
				sint64 chunkLength = (buffer[7]<<24)
					+ (buffer[6]<<16)
					+ (buffer[5]<<8)
					+ (buffer[4]);
				if(!m_VDFile.skipNT(chunkLength))
					return;
			}
		}
	}

	m_VDFile.seek(0, nsVDFile::kSeekStart);
}

void ModAudioSourceDTS::AddDTSFrame(void) {
	DTSFrame *frame = m_DTSInfo.m_FrameList.AddLast();
	frame->m_FileOffset = m_VDFile.tell() - DTS_HEADER_LEN;
	frame->m_Short = m_DTSInfo.m_SampleDeficit-1;
	frame->m_Blocks = m_DTSInfo.m_NbBlocks-1;
	frame->m_Length = m_DTSInfo.m_FrameLength-1;
}

void ModAudioSourceDTS::AddCorruptedPart(sint64 startCorrupted, sint64 bytesSkipped) {
	m_DTSInfo.m_BytesSkipped += bytesSkipped;
	DTSFrame *frame = m_DTSInfo.m_FrameList.AddLast();
	frame->m_FileOffset = startCorrupted;
	frame->m_Short = CORRUPTED_DTS_FRAME;
}

void ModAudioSourceDTS::UpdateFrameCount(void) {
	ModList<DTSFrame>::iterator it(m_DTSInfo.m_FrameList);
	DTSFrame *prev = NULL;
	DTSFrame *current = *it;
	DTSFrame *next = NULL;

	while(current) {
		next = it.next();
		if( (current->m_Short == CORRUPTED_DTS_FRAME) && (m_DTSOptions.m_bKeepCorruptedData) ) {
			sint64 sizeData = (next ? next->m_FileOffset : m_FileSize) - current->m_FileOffset;
			long nbFrames = VDRoundToLong((double)sizeData / ComputeFrameLength(prev, current, next));
			m_DTSInfo.m_Frames += nbFrames;
			m_DTSInfo.m_TotalSamples += nbFrames * ComputeFrameSize(prev, current, next);
		}
		prev = current;
		current = next;
		it++;
	}
}

sint32 ModAudioSourceDTS::ComputeFrameLength(DTSFrame *previous, DTSFrame *current, DTSFrame *next) {
	if(!current)
		return 0;

	sint32 sizeFrame = GetFrameLength(current);
	if(sizeFrame <= 0) {
		sint32 sizePrev = GetFrameLength(previous);
		sint32 sizeNext = GetFrameLength(next);
		if(sizePrev <= 0)
			sizeNext *= 2;
		if(sizeNext <= 0)
			sizePrev *= 2;
		return VDRoundToLong(((double)sizePrev + sizeNext)/2);
	} else
		return sizeFrame;
}

sint32 ModAudioSourceDTS::GetFrameLength(DTSFrame *node) {
	if(!node)
		return 0;

	if(node->m_Short == CORRUPTED_DTS_FRAME)
		return 0;

	return node->m_Length + 1;
}

sint32 ModAudioSourceDTS::ComputeFrameSize(DTSFrame *previous, DTSFrame *current, DTSFrame *next) {
	if(!current)
		return 0;

	sint32 sizeFrame = GetFrameSize(current);
	if(sizeFrame <= 0) {
		sint32 sizePrev = GetFrameSize(previous);
		sint32 sizeNext = GetFrameSize(next);
		if(sizePrev <= 0)
			sizeNext *= 2;
		if(sizeNext <= 0)
			sizePrev *= 2;
		return VDRoundToLong(((double)sizePrev + sizeNext)/2);
	} else
		return sizeFrame;
}

sint32 ModAudioSourceDTS::GetFrameSize(DTSFrame *node) {
	if(!node)
		return 0;

	if(node->m_Short == CORRUPTED_DTS_FRAME)
		return 0;

	return node->m_Blocks * DTS_SAMPLESPERBLOCK + node->m_Short + 1;
}

bool ModAudioSourceDTS::GetNextFrame(bool ensureConsistancy, sint64& bytesSkipped) {
	bytesSkipped = 0;
	sint64 skipped;
	while(true) {
		if(!m_VDFile.synchOn(DTS_SYNCWORD, DTS_SYNCWORD_BITS, DTS_SYNCWORD_LEN, 2*DTS_MAX_LENGTH, skipped)) {
			bytesSkipped += skipped;
			return false;
		}
		bytesSkipped += skipped;
		if(GetHeader(ensureConsistancy))
			return true;
		bytesSkipped += 1;
		m_VDFile.seek(-DTS_HEADER_LEN+1, nsVDFile::kSeekCur);
	}
}

bool ModAudioSourceDTS::GetHeader(bool ensureConsistancy) {
	if(m_VDFile.tell() >= m_FileSize)
		return false;
	char *buffer = m_DTSInfo.m_Buffer;
	// Read the header (7 bytes)
	if(m_VDFile.readData(buffer, DTS_HEADER_LEN) != DTS_HEADER_LEN)
		return false;

	// Test the syncword
	if( ((buffer[0] & DTS_SYNCWORD_BITS[0]) != DTS_SYNCWORD[0])
	|| ((buffer[1] & DTS_SYNCWORD_BITS[1]) != DTS_SYNCWORD[1])
	|| ((buffer[2] & DTS_SYNCWORD_BITS[2]) != DTS_SYNCWORD[2])
	|| ((buffer[3] & DTS_SYNCWORD_BITS[3]) != DTS_SYNCWORD[3]) )
		return false;

	bool normalFrame = true;
	sint32 nbBlocks;
	sint32 nbShort;
	uint32 frmsize;
	sint16 sample_rate_flags;
	sint32 sample_rate;
	sint32 bitrate;
	uint8  acmod;

	normalFrame = (buffer[4]&0x80 ? true : false);
	nbShort = ((buffer[4]&0x7C)>>2) + 1;

	if(normalFrame && (nbShort!=32))
		return false;

	nbBlocks = (((buffer[4]&0x01)<<6) | ((buffer[5]&0xFC)>>2)) + 1;
	frmsize = (((buffer[5]&0x03)<<12) | ((buffer[6]&0xFF)<<4) | ((buffer[7]&0xF0)>>4)) + 1;
	acmod = ((buffer[7]&0x0F)<<2) | ((buffer[8]&0xC0)>>6);

	if(acmod&0x30)
		return false;

	sample_rate_flags = (buffer[8]&0x3C)>>2;
	sample_rate = DTS_SampleRate[sample_rate_flags];
	bitrate = DTS_Bitrate[((buffer[8]&0x03)<<3) | ((buffer[9]&0xE0)>>5)];
	// Verify the sample rate value is not Invalid (in this case we cannot do anything)
	if(sample_rate == -1)
		return false;

	// If asked, verify the stream remain consistant
	if(ensureConsistancy && (
		(sample_rate_flags != m_DTSInfo.m_SampleRateFlags)
		|| (acmod != m_DTSInfo.m_Channels) ) )
		return false;

	m_DTSInfo.m_Bitrate = bitrate;
	m_DTSInfo.m_SampleRateFlags = sample_rate_flags;
	m_DTSInfo.m_SampleRate = sample_rate;
	m_DTSInfo.m_Channels = acmod;
	m_DTSInfo.m_NbChannels = DTS_NbChannels[acmod];
	m_DTSInfo.m_SampleDeficit = nbShort;
	m_DTSInfo.m_NbBlocks = nbBlocks;
	m_DTSInfo.m_Samples = normalFrame ? nbBlocks * DTS_SAMPLESPERBLOCK : (nbBlocks-1)*DTS_SAMPLESPERBLOCK + nbShort;
	m_DTSInfo.m_FrameLength = frmsize;

	return true;
}

void ModAudioSourceDTS::ComputeFramesInNode(void) {
	DTSFrame *next = m_Iterator.next();
	if( ((*m_Iterator)->m_Short == CORRUPTED_DTS_FRAME) && !m_DTSOptions.m_bKeepCorruptedData) {
		m_FrameCount = 0;
		m_SampleCount = 0;
	} else {
		m_FrameCount = VDRoundToLong( ((double)(next?next->m_FileOffset:m_FileSize)-(*m_Iterator)->m_FileOffset) / ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next()));
		m_SampleCount = ComputeFrameSize(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
	}
}
