//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "ModAudioSourceMP3.h"
#include <vd2/system/Error.h>
#include <vd2/system/log.h>
#include <vd2/Dita/resources.h>
#include <vd2/system/math.h>
#include "ProgressDialog.h"
#include "gui.h"
#include "prefs.h"
#include "resource.h"

namespace {
	enum { kVDST_AudioSourceMPA = 155 };
	
	enum {
		kVDM_InformationChange,
		kVDM_VBR_NotLayer3
	};
}

extern HWND			g_hWnd;

static const unsigned char MPA_SYNCWORD[MPA_SYNCWORD_LEN] =
{ 0xFF, 0xE0};
static const unsigned char MPA_SYNCWORD_BITS[MPA_SYNCWORD_LEN] =
{ 0xFF, 0xE0};


ModAudioSourceMP3::ModAudioSourceMP3(const wchar_t *szFileName, MPAOptions options, bool bShutUp)
: m_MPAOptions(options) {
	m_MPAInfo.m_Version		= MPEG_RESERVED;
	m_MPAInfo.m_Layer		= LAYER_RESERVED;
	m_MPAInfo.m_SampleRate	= 0;
	m_MPAInfo.m_Channels	= SINGLE_CHANNEL;
	m_MPAInfo.m_NbChannels	= 0;
	m_MPAInfo.m_HeadMode	= 0;
	m_MPAInfo.m_HeadModeExt	= 0;
	m_MPAInfo.m_Emphasis	= 0;
	m_MPAInfo.m_bProtected	= false;
	m_MPAInfo.m_bPrivate	= false;
	m_MPAInfo.m_bCopyrighted	= false;
	m_MPAInfo.m_bOriginal	= false;
	m_MPAInfo.m_Frames = 0;
	m_MPAInfo.m_FrameList.Clear();
	m_bHasRIFFHeader = false;
	m_RIFFHeaderLength = 0;
	m_bHasTagV1 = false;
	m_bHasTagV2_Start = false;
	m_bHasTagV2_End = false;
	m_TagV2StartLength = 0;
	m_TagV2EndLength = 0;
	m_FileSize = 0;
	m_EndData = 0;
	memset(m_MPAInfo.m_Buffer, 0, MPA_HEADER_LEN);
	m_MPAInfo.m_Bytes = 0;
	m_MPAInfo.m_BytesSkipped = 0;
	m_BlockAlign = 1;
	m_NodeSample = 0;
	m_NodeFrame = 0;
	m_NodeByte = 0;
	m_FrameCount = 0;
	m_LastSample = 0;
	m_TotalBytes = 0;
	m_MaxFrameLength = 0;
	m_bVariableFrameLength = false;
	// Output to AVI assumed : NO VBR, BYTE ACCURACY
	m_bVBRHandling = false;
	m_GroupFrames = 1;
	m_bIsVBR = false;
	m_bShutUp = bShutUp;
	m_bEmphasisChanged = false;
	m_bProtectedChanged = false;
	m_bPrivateChanged = false;
	m_bCopyrightedChanged = false;
	m_bOriginalChanged = false;

	m_SourceType = SOURCE_STD_FILE | SOURCE_AUDIO_MP3;
	clearSourceInfo();
	m_SourceName = szFileName;
	m_SourceDesc = L"MPEG Audio file";
}

ModAudioSourceMP3::~ModAudioSourceMP3(void) {
#ifndef _USE_COMPACTLIST_MPA
	m_MPAInfo.m_FrameList.DeleteContents(true);
#endif
	m_MPAInfo.m_FrameList.Clear();
	m_VDFile.close();
}

bool ModAudioSourceMP3::QueryInterface(InterfaceID ID, void** destination) {
	if(ID == IMP3Source::GetInterfaceID()) {
		*destination = static_cast<IMP3Source *>(this);
		return true;
	}
	return false;
}

bool ModAudioSourceMP3::UseVBRalaNandub(void) {
	return m_MPAOptions.m_bNandubTrick;
}

bool ModAudioSourceMP3::IsVBR(void) {
	return m_bIsVBR;
}

bool ModAudioSourceMP3::FixVBR(bool bFix) {
	m_MPAOptions.m_bNandubTrick = !bFix;
	extend(false, m_GroupFrames);
	return true;
}

bool ModAudioSourceMP3::init(void) {
	// open file
	if(!m_VDFile.open(m_SourceName.c_str(),
	nsVDFile::kRead | nsVDFile::kDenyWrite | nsVDFile::kOpenExisting | nsVDFile::kSequential))
		return FALSE;

	m_FileSize = m_VDFile.size();

	ProgressDialog pd(NULL, "MPA File Import Filter", "Parsing File", (sint32)(m_FileSize/1024), true);
	pd.setValueFormat("%ldK of %ldK");

	// Check if there are Tags in the file
	CheckTags();

	m_EndData = m_FileSize - m_TagV2EndLength - (m_bHasTagV1 ? 128 : 0);

	// Skip the Beginning Tag if any
	if(!m_VDFile.seekNT((m_bHasTagV2_Start ? m_TagV2StartLength : 0) + (m_bHasRIFFHeader ? m_RIFFHeaderLength : 0), nsVDFile::kSeekStart))
		return FALSE;

	sint64 bytesSkipped;
	sint64 startSkipped = m_TagV2StartLength + m_RIFFHeaderLength;

	// Get the first valid frame (or at least the first that seem valid ;))
	if(!GetNextFrame(false, bytesSkipped))
		return FALSE;
	if(bytesSkipped) {
		AddCorruptedPart(startSkipped, bytesSkipped);
		startSkipped += bytesSkipped;
	}

	AddMPAFrame();

	// Being unable to read the first frame is not good ;)
	if(!m_VDFile.seekNT(m_MPAInfo.m_FrameLength - MPA_HEADER_LEN, nsVDFile::kSeekCur))
		return FALSE;

	m_MPAInfo.m_Frames = 1;

	sint32 frame_length = m_MPAInfo.m_FrameLength;
	startSkipped += frame_length;
	m_MPAInfo.m_Bytes = frame_length;

	while(GetNextFrame(true, bytesSkipped)) {
		// Verify that this frame wasn't cut at the end of the file
		if(m_VDFile.tell()-MPA_HEADER_LEN+m_MPAInfo.m_FrameLength > m_EndData) {
			// The last frame was cut :(
			bytesSkipped = m_EndData-startSkipped;
			break;
		}

		if(bytesSkipped) {
			AddCorruptedPart(startSkipped, bytesSkipped);
			startSkipped += bytesSkipped;
			// If the skipped data aren't a multiple of frame sizes, we
			// shouldn't use frame mode
			if(m_MPAOptions.m_bKeepCorruptedData && (bytesSkipped % frame_length))
				m_bVariableFrameLength = true;
			if(m_MPAInfo.m_FrameLength == frame_length)
				AddMPAFrame();
		}

		// We got a new frame

		// If its length differs from the previous ones, add it to the list
		if(m_MPAInfo.m_FrameLength != frame_length) {
			AddMPAFrame();
			frame_length = m_MPAInfo.m_FrameLength;
			m_bVariableFrameLength = true;
			if(frame_length > m_MaxFrameLength)
				m_MaxFrameLength = frame_length;
		}

		m_MPAInfo.m_Frames++;
		startSkipped += frame_length;

		// Skip the frame
		m_VDFile.seek(m_MPAInfo.m_FrameLength - MPA_HEADER_LEN, nsVDFile::kSeekCur);

		m_MPAInfo.m_Bytes += frame_length;

		pd.advance((sint32)(m_VDFile.tell()/1024));
		pd.check();
	}
	if(bytesSkipped) {
		AddCorruptedPart(startSkipped, bytesSkipped);
		startSkipped += bytesSkipped;
		// If the skipped data aren't a multiple of frame sizes, we
		// shouldn't use frame mode
		if(m_MPAOptions.m_bKeepCorruptedData && (bytesSkipped % frame_length))
			m_bVariableFrameLength = true;
	}

	if(	(m_MPAInfo.m_Frames <= 0)
	|| (m_MPAInfo.m_Version == MPEG_RESERVED)
	|| (m_MPAInfo.m_Layer == LAYER_RESERVED)
	|| (m_MPAInfo.m_SampleRate <= 0)
	|| ( (m_MPAInfo.m_NbChannels!=1) && (m_MPAInfo.m_NbChannels!=2)) )
		return FALSE;

	// Update the frame count taking into account partial/corrupted data
	UpdateFrameCount();

	m_TotalBytes = m_MPAInfo.m_Bytes;
	if(m_MPAOptions.m_bKeepCorruptedData)
		m_TotalBytes += m_MPAInfo.m_BytesSkipped;
	if(m_bVBRHandling && m_MPAOptions.m_bKeepTags)
		m_TotalBytes += m_TagV2StartLength + m_TagV2EndLength + (m_bHasTagV1 ? 128 : 0);

	double fBitrate = (double)m_TotalBytes * 8 * m_MPAInfo.m_SampleRate / ((double)m_MPAInfo.m_SamplesPerFrame * m_MPAInfo.m_Frames);

	switch(m_MPAInfo.m_Layer) {
	case LAYER_3:
		if(!allocFormat(sizeof MPEGLAYER3WAVEFORMAT))
			return FALSE;
		break;
	case LAYER_1:
	case LAYER_2:
		if(!allocFormat(sizeof MPEG1WAVEFORMAT))
			return FALSE;
		break;
	}

	m_Iterator.begin(m_MPAInfo.m_FrameList);
	MPAFrame *curFrame = *m_Iterator;
	sint64 bitsProcessed = 0;
	if(curFrame && curFrame->m_FileOffset && m_MPAOptions.m_bKeepCorruptedData)
		bitsProcessed = (curFrame->m_FileOffset - m_TagV2StartLength) * 8;
	MPAFrame *nextFrame = NULL;
	sint64 nbFrames = 0;
	sint64 maxBitsDev = 0;
	sint64 sumSquareBits = 0;
	sint64 bitsDev = 0;
	while(m_Iterator) {
		curFrame = *m_Iterator;
		nextFrame = m_Iterator.next();
		bitsDev = (fBitrate * nbFrames * m_MPAInfo.m_SamplesPerFrame) / m_MPAInfo.m_SampleRate - bitsProcessed;
		if(bitsDev < 0)
			bitsDev = -bitsDev;
		if(bitsDev > maxBitsDev)
			maxBitsDev = bitsDev;
		ComputeFramesInNode();
		sint64 blockLength = (nextFrame ? nextFrame->m_FileOffset : m_EndData) - curFrame->m_FileOffset;
		double frameLength = m_FrameCount ? (double)blockLength / m_FrameCount : 0;
		bitsProcessed += m_FrameCount * 8 * frameLength;
		sumSquareBits += m_FrameCount * 64 * frameLength * frameLength;
		nbFrames += m_FrameCount;
		m_Iterator++;
	}
	bitsDev = (fBitrate * nbFrames * m_MPAInfo.m_SamplesPerFrame) / m_MPAInfo.m_SampleRate - bitsProcessed;
	if(bitsDev < 0)
		bitsDev = -bitsDev;
	if(bitsDev > maxBitsDev)
		maxBitsDev = bitsDev;
	double maxDevSec = maxBitsDev / fBitrate;
	double stdBitsDev = sqrt(double(nbFrames*sumSquareBits-bitsProcessed*bitsProcessed)) * m_MPAInfo.m_SampleRate / (nbFrames * m_MPAInfo.m_SamplesPerFrame);
	if(m_bIsVBR && !m_bShutUp) {
		if(guiMessageBoxF(g_hWnd, (!m_MPAOptions.m_bNandubTrick ? IDS_WARN_VBR_MP3FILE_FIX : IDS_WARN_VBR_MP3FILE_NOFIX), IDS_WARN_VBR_DETECTED_CAP, MB_YESNO|MB_TASKMODAL|MB_SETFOREGROUND|MB_ICONQUESTION
			, VDTextWToA(m_SourceName).c_str(), maxDevSec*1000.0, fBitrate*0.001, stdBitsDev*0.001) == IDNO)
			m_MPAOptions.m_bNandubTrick = !m_MPAOptions.m_bNandubTrick;
		if(m_MPAInfo.m_Layer != LAYER_3) {
			// Layer I-II decoders aren't supposed to handle VBR (on the contrary to Layer III ones)
			// So warn the user we don't really like VBR here :p
			VDLogAppMessage(kVDLogWarning, kVDST_AudioSourceMPA, kVDM_VBR_NotLayer3);
		}
	} else if(!m_bIsVBR && m_MPAOptions.m_bNandubTrick)
		m_MPAOptions.m_bNandubTrick = false;

	// By default assume we will be outputed to an AVI => NO VBR, BYTE ACCURACY when reading data
	extend(false, g_VDMPrefs.m_GroupFrames);

	m_SourceDesc = VDStringW::setf(L"MPEG%s-L%d %s %ldHz %.1f%.1fkbps"
		, m_MPAInfo.m_Version==MPEG_V1 ? L"1" : m_MPAInfo.m_Version==MPEG_V2 ? L"2" : m_MPAInfo.m_Version==MPEG_V25 ? L"2.5" : L"?"
		, (int)m_MPAInfo.m_Layer+1
		, m_MPAInfo.m_Channels==SINGLE_CHANNEL ? L"Single Channel" : m_MPAInfo.m_Channels==DUAL_CHANNEL ? L"Dual Channel"
		: m_MPAInfo.m_Channels==JOINT_STEREO ? L"Joint Stereo" : m_MPAInfo.m_Channels==STEREO ? L"Stereo" : L"??"
		, m_MPAInfo.m_SampleRate, fBitrate / 1000, stdBitsDev / 1000
		);

	return TRUE;
}

VDStringW ModAudioSourceMP3::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	m_TotalBytes = m_MPAInfo.m_Bytes;
	if(m_MPAOptions.m_bKeepCorruptedData)
		m_TotalBytes += m_MPAInfo.m_BytesSkipped;
	if(m_bVBRHandling && m_MPAOptions.m_bKeepTags)
		m_TotalBytes += m_TagV2StartLength + m_TagV2EndLength + (m_bHasTagV1 ? 128 : 0);
	double fBitrate = (double)m_TotalBytes * 8 * m_MPAInfo.m_SampleRate / ((double)1000 * m_MPAInfo.m_SamplesPerFrame * m_MPAInfo.m_Frames);
	report += VDStringW::setf(L"\r\nMPEG Audio stream : MPEG Version %s, Layer %s, %ld channels %s, %ldHz, %.1fkbps"
		, m_MPAInfo.m_Version==MPEG_V1 ? L"1" : m_MPAInfo.m_Version==MPEG_V2 ? L"2" : m_MPAInfo.m_Version==MPEG_V25 ? L"2.5" : L"?"
		, m_MPAInfo.m_Layer==LAYER_1 ? L"I" : m_MPAInfo.m_Layer==LAYER_2 ? L"II" : m_MPAInfo.m_Layer==LAYER_3 ? L"III" : L"?"
		, m_MPAInfo.m_NbChannels
		, m_MPAInfo.m_Channels==SINGLE_CHANNEL ? L"Single Channel" : m_MPAInfo.m_Channels==DUAL_CHANNEL ? L"Dual Channel"
		: m_MPAInfo.m_Channels==JOINT_STEREO ? L"Joint Stereo" : m_MPAInfo.m_Channels==STEREO ? L"Stereo" : L"??"
		, m_MPAInfo.m_SampleRate, fBitrate);
	if(m_bHasRIFFHeader)
		report += VDStringW::setf(L"\r\nFound a RIFF header at the beginning of the file, %ld bytes", m_RIFFHeaderLength);
	if(m_bHasTagV2_Start)
		report += VDStringW::setf(L"\r\nFound an ID Tag v2 at the beginning of the file, %ld bytes", m_TagV2StartLength);
	if(m_bHasTagV1)
		report += VDStringW::setf(L"\r\nFound an ID Tag v1(.1) (128 bytes at the end of the file)");
	if(m_bHasTagV2_End)
		report += VDStringW::setf(L"\r\nFound an ID Tag v2 at the end of the file, %ld bytes", m_TagV2EndLength);
	if(m_MPAInfo.m_BytesSkipped) {
		report += VDStringW::setf(L"\r\nA total of %I64d bytes have been seen as corrupted :"
			, m_MPAInfo.m_BytesSkipped);
		m_Iterator.begin(m_MPAInfo.m_FrameList);
		while(m_Iterator) {
			MPAFrame *curFrame = *m_Iterator;
			if(curFrame->m_Flags == CORRUPTED_MPA_FRAME) {
				sint64 filePos = curFrame->m_FileOffset;
				sint64 bytesSkipped = (m_Iterator.next() ? m_Iterator.next()->m_FileOffset : m_EndData) - filePos;
				report += VDStringW::setf(L"\r\n %I64d bytes starting at byte %I64d"
					, bytesSkipped, filePos);
			}
			m_Iterator++;
		}
		m_Iterator.begin(m_MPAInfo.m_FrameList);
	} else
		report += L"\r\nNo apparent corrupted data";

	return report;
}

void ModAudioSourceMP3::extend(bool bVBRHandling, sint32 groupFrames) {
	m_bVBRHandling = bVBRHandling;
	m_GroupFrames = groupFrames;

	m_TotalBytes = m_MPAInfo.m_Bytes;
	if(m_MPAOptions.m_bKeepCorruptedData)
		m_TotalBytes += m_MPAInfo.m_BytesSkipped;
	if(m_bVBRHandling && m_MPAOptions.m_bKeepTags)
		m_TotalBytes += m_TagV2StartLength + m_TagV2EndLength + (m_bHasTagV1 ? 128 : 0);

	double fByterate = (double)m_TotalBytes * m_MPAInfo.m_SampleRate / (m_MPAInfo.m_SamplesPerFrame * m_MPAInfo.m_Frames);
	sint32 dByterate = VDRoundToLong(fByterate);
	sint32 mpegBitrate = 0;
	if(!m_bIsVBR)
		mpegBitrate = m_MPAInfo.m_Bitrate;

	WAVEFORMATEX *wfx = getWaveFormat();
	switch(m_MPAInfo.m_Layer) {
	case LAYER_3:
		{
			MPEGLAYER3WAVEFORMAT *mp3format = (MPEGLAYER3WAVEFORMAT *)getWaveFormat();
			memset(mp3format, 0, sizeof(MPEGLAYER3WAVEFORMAT));
			mp3format->wfx.cbSize			= MPEGLAYER3_WFX_EXTRA_BYTES;// sizeof MPEGLAYER3WAVEFORMAT - sizeof WAVEFORMATEX
			mp3format->wfx.wFormatTag		= WAVE_FORMAT_MPEGLAYER3;
			mp3format->wfx.nAvgBytesPerSec	= (DWORD)dByterate;
			mp3format->wfx.nChannels		= m_MPAInfo.m_NbChannels;
			mp3format->wfx.nSamplesPerSec	= m_MPAInfo.m_SampleRate;
			mp3format->wfx.wBitsPerSample	= 0;
			mp3format->wID					= MPEGLAYER3_ID_MPEG;
			mp3format->nFramesPerBlock = 1;
			mp3format->nCodecDelay		= /*1393*/0;
			if(!m_bVariableFrameLength) {
				// For constant frame length no padding was used
				mp3format->fdwFlags			= MPEGLAYER3_FLAG_PADDING_OFF;
				mp3format->wfx.nBlockAlign	= m_MPAInfo.m_FrameLength;
				mp3format->nBlockSize		= m_MPAInfo.m_FrameLength;
				streamInfo.dwSampleSize		= m_MPAInfo.m_FrameLength;
			} else {
				// Attempt to detect the padding mode and block size for the stream.
				double fAverageFrameSize = double(m_TotalBytes) / m_MPAInfo.m_Frames;
				int estimated_bitrate = (int)floor(0.5 + fByterate * (1.0/1000.0)) * 8;
				double fEstimatedFrameSizeISO = 144000.0 * estimated_bitrate / mp3format->wfx.nSamplesPerSec;
				if (mp3format->wfx.nSamplesPerSec < 32000)	// MPEG-2?
					fEstimatedFrameSizeISO *= 0.5;
				double fEstimatedFrameSizePaddingOff = floor(fEstimatedFrameSizeISO);
				double fEstimatedFrameSizePaddingOn = fEstimatedFrameSizePaddingOff + 1.0;
				double fErrorISO = fabs(fEstimatedFrameSizeISO - fAverageFrameSize);
				double fErrorPaddingOn = fabs(fEstimatedFrameSizePaddingOn - fAverageFrameSize);
				double fErrorPaddingOff = fabs(fEstimatedFrameSizePaddingOff - fAverageFrameSize);
				if (fErrorISO <= fErrorPaddingOn) {				// iso < on
					if (fErrorISO <= fErrorPaddingOff)			// iso < on, off
						mp3format->fdwFlags	= MPEGLAYER3_FLAG_PADDING_ISO;
					else										// off < iso < on
						mp3format->fdwFlags	= MPEGLAYER3_FLAG_PADDING_OFF;
				} else {										// on < iso
					if (fErrorPaddingOn <= fErrorPaddingOff)	// on < iso, off
						mp3format->fdwFlags	= MPEGLAYER3_FLAG_PADDING_ON;
					else										// off < on < iso
						mp3format->fdwFlags	= MPEGLAYER3_FLAG_PADDING_OFF;
				}
				mp3format->wfx.nBlockAlign	= 1;
				mp3format->nBlockSize		= ceil((double)m_TotalBytes / m_MPAInfo.m_Frames);
			}
		}
		break;
	case LAYER_1:
	case LAYER_2:
		{
			MPEG1WAVEFORMAT *mpaformat		= (MPEG1WAVEFORMAT *)getWaveFormat();
			memset(mpaformat, 0, sizeof(MPEG1WAVEFORMAT));
			mpaformat->wfx.cbSize			= sizeof MPEG1WAVEFORMAT - sizeof WAVEFORMATEX;
			mpaformat->wfx.wFormatTag		= WAVE_FORMAT_MPEG;
			mpaformat->wfx.nAvgBytesPerSec	= (DWORD)dByterate;
			mpaformat->wfx.nChannels		= m_MPAInfo.m_NbChannels;
			mpaformat->wfx.nSamplesPerSec	= m_MPAInfo.m_SampleRate;
			mpaformat->wfx.wBitsPerSample	= 0;
			mpaformat->wfx.nBlockAlign		= 1;
			if(!m_bVariableFrameLength) {
				mpaformat->wfx.nBlockAlign	= m_MPAInfo.m_FrameLength;
				streamInfo.dwSampleSize		= m_MPAInfo.m_FrameLength;
			}

			if(m_MPAInfo.m_Layer == LAYER_2)
				mpaformat->fwHeadLayer		= ACM_MPEG_LAYER2;
			else
				mpaformat->fwHeadLayer		= ACM_MPEG_LAYER1;

			mpaformat->dwHeadBitrate        = mpegBitrate;

			mpaformat->fwHeadMode			= m_MPAInfo.m_HeadMode;
			// fwHeadModeExt should only be set for streams using stereo
			if(mpaformat->fwHeadMode & ACM_MPEG_STEREO)
				mpaformat->fwHeadModeExt	= m_MPAInfo.m_HeadModeExt;
			mpaformat->wHeadEmphasis		= m_MPAInfo.m_Emphasis;
			if(m_MPAInfo.m_Version == MPEG_V1)
				mpaformat->fwHeadFlags		|= ACM_MPEG_ID_MPEG1;
			if(m_MPAInfo.m_bProtected)
				mpaformat->fwHeadFlags		|= ACM_MPEG_PROTECTIONBIT;
			if(m_MPAInfo.m_bPrivate)
				mpaformat->fwHeadFlags		|= ACM_MPEG_PRIVATEBIT;
			if(m_MPAInfo.m_bCopyrighted)
				mpaformat->fwHeadFlags		|= ACM_MPEG_COPYRIGHT;
			if(m_MPAInfo.m_bOriginal)
				mpaformat->fwHeadFlags		|= ACM_MPEG_ORIGINALHOME;

			mpaformat->dwPTSLow = 0;
			mpaformat->dwPTSHigh = 0;
		}
		break;
	}

	lSampleFirst = 0;

	streamInfo.fccType         = streamtypeAUDIO;
	streamInfo.fccHandler      = 0;
	streamInfo.dwFlags         = 0;
	streamInfo.wPriority       = 0;
	streamInfo.wLanguage       = 0;
	streamInfo.dwInitialFrames = 0;
	streamInfo.dwStart         = 0;
	streamInfo.dwQuality       = 0xFFFFFFFF;
	streamInfo.dwSuggestedBufferSize = 0;

	sint32 bytesPerSample = wfx->nBlockAlign;
	if(m_bVBRHandling) {
		lSampleLast = m_MPAInfo.m_Frames * m_MPAInfo.m_SamplesPerFrame;

		streamInfo.dwScale         = 1;
		streamInfo.dwRate          = m_MPAInfo.m_SampleRate;
		streamInfo.dwSampleSize    = bytesPerSample;
	} else if(!m_MPAOptions.m_bNandubTrick) {	// Handle as a standard CBR stream (i.e. no Nandub trick)
		lSampleLast = m_TotalBytes / bytesPerSample;

		streamInfo.dwScale         = bytesPerSample;
		streamInfo.dwRate          = wfx->nAvgBytesPerSec;
		streamInfo.dwSampleSize    = bytesPerSample;
	} else {
		// Who the hell is still using f$!# VBR MP3 in AVI !!
		wfx->nBlockAlign = 1152;

		lSampleLast = m_MPAInfo.m_Frames;

		streamInfo.dwScale         = MPA_FrameSize[m_MPAInfo.m_Layer][m_MPAInfo.m_Version];
		streamInfo.dwRate          = m_MPAInfo.m_SampleRate;
		streamInfo.dwSampleSize    = 0;
	}
	streamInfo.dwLength        = lSampleLast - lSampleFirst;

	m_LastSample = lSampleLast;

	m_BlockAlign = wfx->nBlockAlign;

	m_Iterator.begin(m_MPAInfo.m_FrameList);
	m_NodeSample = 0;
	m_NodeFrame = 0;
	m_NodeByte = 0;
	ComputeFramesInNode();
}

int ModAudioSourceMP3::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {
	if(lBytesRead)
		*lBytesRead = 0;
	if(lSamplesRead)
		*lSamplesRead = 0;
	if(!lCount)
		return AVIERR_OK;
	sint32 framesLength, frameIndex, frameLength;
	sint32 readLength = 0;
	if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
		// Too bad
		return AVIERR_OK;

//#pragma message(__TODO__ "Send the Tags when asking first or last sample")

	// Treat in my order of preference :
	if(m_bVBRHandling) {	// Correctly handle this stream in any case (CBR / VBR)
		// Request by sample number
		if(m_GroupFrames > 0) {
			if( (lCount==AVISTREAMREAD_CONVENIENT) || (lCount<m_GroupFrames*m_MPAInfo.m_SamplesPerFrame) )
				lCount = m_GroupFrames * m_MPAInfo.m_SamplesPerFrame;
		} else
			lCount = m_MPAInfo.m_SamplesPerFrame;

		while(true) {
			if(lpBuffer) {
				if(cbBuffer < frameLength) {
					// If buffer too short, tell the one that should always work
					if(lBytesRead)
						*lBytesRead += ((lCount+m_MPAInfo.m_SamplesPerFrame-1)/m_MPAInfo.m_SamplesPerFrame) * MPA_MAX_LENGTH;
					return AVIERR_BUFFERTOOSMALL;
				}

				m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
				if((readLength=m_VDFile.readData(lpBuffer, frameLength))!=frameLength) {
					if(lBytesRead)
						*lBytesRead += readLength;
					return AVIERR_FILEREAD;
				}

				lpBuffer = (char *)lpBuffer + readLength;
			}

			if(lBytesRead)
				*lBytesRead += frameLength;
			if(lSamplesRead)
				*lSamplesRead += m_MPAInfo.m_SamplesPerFrame;

			cbBuffer -= frameLength;
			lCount -= m_MPAInfo.m_SamplesPerFrame;
			lStart += m_MPAInfo.m_SamplesPerFrame;

			if(lCount <= 0)
				return AVIERR_OK;

			if(!SeekSample(lStart, framesLength, frameIndex, frameLength))
				return AVIERR_OK;
		}
	} else if(!m_MPAOptions.m_bNandubTrick) {	// Handle as a standard CBR stream
		// Request by block number (=> number of bytes)
		if(lCount == AVISTREAMREAD_CONVENIENT)
			lCount = 1;

		if(lBytesRead)
			*lBytesRead = frameLength * lCount;
		if(lSamplesRead)
			*lSamplesRead = lCount;
		if(!lpBuffer)
			return AVIERR_OK;

		if(cbBuffer < frameLength * lCount)
			return AVIERR_BUFFERTOOSMALL;

		m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
		if((readLength=m_VDFile.readData(lpBuffer, frameLength*lCount))!=frameLength*lCount) {
			if(lBytesRead)
				*lBytesRead = readLength;
			return AVIERR_FILEREAD;
		}

		return AVIERR_OK;
	} else {	// Use this good old (*cough*) Nandub trick
		// Nandub : request by frame number
		if(lBytesRead)
			*lBytesRead = frameLength;
		if(lSamplesRead)
			*lSamplesRead = 1;
		if(!lpBuffer)
			return AVIERR_OK;

		if(cbBuffer < frameLength)
			return AVIERR_BUFFERTOOSMALL;

		m_VDFile.seek((*m_Iterator)->m_FileOffset + frameIndex*framesLength, nsVDFile::kSeekStart);
		if((readLength=m_VDFile.readData(lpBuffer, frameLength))!=frameLength) {
			if(lBytesRead)
				*lBytesRead = readLength;
			return AVIERR_FILEREAD;
		}

		return AVIERR_OK;
	}
}

bool ModAudioSourceMP3::SeekSample(LONG lStart, sint32& framesLength, sint32& frameIndex, sint32& frameLength) {
	if(lStart >= m_LastSample)
		return false;

	if(m_bVBRHandling) {
		while((lStart < m_NodeSample) && (m_Iterator.previous())) {
			m_Iterator--;
			ComputeFramesInNode();
			m_NodeSample -= m_FrameCount*m_MPAInfo.m_SamplesPerFrame;
		}
		while(m_NodeSample + m_FrameCount*m_MPAInfo.m_SamplesPerFrame <= lStart) {
			m_Iterator++;
			m_NodeSample += m_FrameCount*m_MPAInfo.m_SamplesPerFrame;
			ComputeFramesInNode();
		}
		framesLength = ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
		frameIndex = (lStart - m_NodeSample) / m_MPAInfo.m_SamplesPerFrame;
		if( ((*m_Iterator)->m_Flags==CORRUPTED_MPA_FRAME) && (frameIndex==m_FrameCount-1) ) {
			MPAFrame *next = m_Iterator.next();
			frameLength = (next?next->m_FileOffset:m_EndData)-((*m_Iterator)->m_FileOffset+frameIndex*framesLength);
		} else
			frameLength = framesLength;

		return true;
	} else if(!m_MPAOptions.m_bNandubTrick) {
		while(lStart < m_NodeByte / m_BlockAlign) {
			sint64 pos = (*m_Iterator)->m_FileOffset;
			m_Iterator--;
			m_NodeByte -= pos-(*m_Iterator)->m_FileOffset;
		}
		MPAFrame *next = m_Iterator.next();
		while( (m_NodeByte + (next?next->m_FileOffset:m_EndData)-(*m_Iterator)->m_FileOffset) / m_BlockAlign <= lStart) {
			m_NodeByte += (next?next->m_FileOffset:m_EndData) - (*m_Iterator)->m_FileOffset;
			m_Iterator++;
			next = m_Iterator.next();
		}
		framesLength = 1;
		frameIndex = lStart * m_BlockAlign - m_NodeByte;
		frameLength = m_BlockAlign;

		return true;
	} else {
		while(lStart < m_NodeFrame) {
			m_Iterator--;
			ComputeFramesInNode();
			m_NodeFrame -= m_FrameCount;
		}
		while(m_NodeFrame + m_FrameCount <= lStart) {
			m_Iterator++;
			m_NodeFrame += m_FrameCount;
			ComputeFramesInNode();
		}
		framesLength = ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next());
		frameIndex = lStart - m_NodeFrame;
		if( ((*m_Iterator)->m_Flags==CORRUPTED_MPA_FRAME) && (frameIndex==m_FrameCount-1) ) {
			MPAFrame *next = m_Iterator.next();
			frameLength = (next?next->m_FileOffset:m_EndData)-((*m_Iterator)->m_FileOffset+frameIndex*framesLength);
		} else
			frameLength = framesLength;

		return true;
	}

	return false;
}

sint64 ModAudioSourceMP3::sampleOffset(sint64 lSample) {
	if(lSample >= m_LastSample)
		return m_TotalBytes;
	sint32 framesLength, frameIndex, frameLength;
	if(!SeekSample(lSample, framesLength, frameIndex, frameLength))
		return 0;
	return ((*m_Iterator)->m_FileOffset + frameIndex*framesLength);
}

sint32 ModAudioSourceMP3::samplesLag(sint64 lStart) {
	sint32 framesLength, frameIndex, frameLength;
	if((lStart<=0) || !m_bVBRHandling || !SeekSample(lStart, framesLength, frameIndex, frameLength))
		return 0;

	return ( (m_NodeSample + frameIndex * m_MPAInfo.m_SamplesPerFrame) - lStart);
}

void ModAudioSourceMP3::CheckTags(void) {
	// Check if there are tags in this MP3 file
	char buffer[1024];

	// Is there a RIFF header at the beginning ?
	m_VDFile.seek(0, nsVDFile::kSeekStart);
	if(m_VDFile.readData(buffer, 12) != 12)
		return;
	if( (*(long *)(buffer+0)=='FFIR') && (*(long *)(buffer+8)=='EVAW') ) {
		while(true) {
			if(m_VDFile.readData(buffer, 8) != 8)
				return;
			if(*(long *)(buffer+0) == 'atad') {
				m_bHasRIFFHeader = true;
				m_RIFFHeaderLength = m_VDFile.tell();
				break;
			} else {
				sint64 chunkLength = (buffer[7]<<24)
					+ (buffer[6]<<16)
					+ (buffer[5]<<8)
					+ (buffer[4]);
				if(!m_VDFile.skipNT(chunkLength))
					return;
			}
		}
	} else {
		// Is there an ID3v2 Tag at the beginning ?
		if(!memcmp(buffer, "ID3", 3)) {
			// Seems so :P
			if(!(buffer[5]&0x0F) && !(buffer[6]&0x80) && !(buffer[7]&0x80) && !(buffer[8]&0x80) && !(buffer[9]&0x80)) {
				// Yeah it's really a flag
				m_bHasTagV2_Start = true;
				bool bHasFooter = !!(buffer[5] & 0x10);
				m_TagV2StartLength = (bHasFooter ? 20 : 10);
				// The length of the tag is put in 4 bytes which bit 7 is set to 0
				// to prevent any parser to see it as the synch part of a header
				m_TagV2StartLength += ((sint32)buffer[6])<<21;
				m_TagV2StartLength += ((sint32)buffer[7])<<14;
				m_TagV2StartLength += ((sint32)buffer[8])<<7;
				m_TagV2StartLength += (sint32)buffer[9];
			}
		}
	}

	// Is there an ID3v2 Tag at the end ?
	m_VDFile.seek(-10, nsVDFile::kSeekEnd);
	if(m_VDFile.readData(buffer, 3) != 3)
		return;
	if(!memcmp(buffer, "3DI", 3)) {
		if(m_VDFile.readData(buffer, 7) != 7)
			return;

		if(!(buffer[2]&0x0F) && !(buffer[3]&0x80) && !(buffer[4]&0x80) && !(buffer[5]&0x80) && !(buffer[6]&0x80)) {
			m_bHasTagV2_End = true;
			bool bHasFooter = /*!!(buffer[2] & 0x10)*/true; // this is the footer !!
			m_TagV2EndLength = (bHasFooter ? 20 : 10);
			m_TagV2EndLength += ((sint32)buffer[3])<<21;
			m_TagV2EndLength += ((sint32)buffer[4])<<14;
			m_TagV2EndLength += ((sint32)buffer[5])<<7;
			m_TagV2EndLength += (sint32)buffer[6];
		}
	}

	if(!m_bHasTagV2_End) {
		// Is there an ID3v1(.1) Tag at the end ?
		m_VDFile.seek(-128, nsVDFile::kSeekEnd);
		if(m_VDFile.readData(buffer, 3) != 3)
			return;
		if(!memcmp(buffer, "TAG", 3))
			m_bHasTagV1 = true;
	}

	m_VDFile.seek(0, nsVDFile::kSeekStart);
}

void ModAudioSourceMP3::AddMPAFrame(void) {
	MPAFrame *frame = m_MPAInfo.m_FrameList.AddLast();
	frame->m_FileOffset = m_VDFile.tell() - MPA_HEADER_LEN;
	// Masks :
	// 0xF0 -> bitrate
	// 0x02 -> padding
	frame->m_Flags = m_MPAInfo.m_Buffer[2] & 0xF2;
}

void ModAudioSourceMP3::AddCorruptedPart(sint64 startCorrupted, sint64 bytesSkipped) {
	m_MPAInfo.m_BytesSkipped += bytesSkipped;
	MPAFrame *frame = m_MPAInfo.m_FrameList.AddLast();
	frame->m_FileOffset = startCorrupted;
	// Set the bitrate as 'bad' == -1
	frame->m_Flags = CORRUPTED_MPA_FRAME;
}

void ModAudioSourceMP3::UpdateFrameCount(void) {
#ifdef _USE_COMPACTLIST_MPA
	ModCompactList<MPAFrame>::iterator it(m_MPAInfo.m_FrameList);
#else // !_USE_COMPACTLIST_MPA
	ModList<MPAFrame>::iterator it(m_MPAInfo.m_FrameList);
#endif _USE_COMPACTLIST_MPA
	MPAFrame *prev = NULL;
	MPAFrame *current = *it;
	MPAFrame *next = NULL;

	while(current) {
		next = it.next();
		if( (current->m_Flags == CORRUPTED_MPA_FRAME) && (m_MPAOptions.m_bKeepCorruptedData) ) {
			sint64 sizeData = (next ? next->m_FileOffset : m_EndData) - current->m_FileOffset;
			m_MPAInfo.m_Frames += VDRoundToLong((double)sizeData / ComputeFrameLength(prev, current, next));
		}
		prev = current;
		current = next;
		it++;
	}
}

sint32 ModAudioSourceMP3::ComputeFrameLength(MPAFrame *previous, MPAFrame *current, MPAFrame *next) {
	if(!current)
		return 0;

	sint32 sizeFrame = GetFrameLength(current);
	if(sizeFrame <= 0) {
		sint32 sizePrev = GetFrameLength(previous);
		sint32 sizeNext = GetFrameLength(next);
		if(sizePrev <= 0)
			sizeNext *= 2;
		if(sizeNext <= 0)
			sizePrev *= 2;
		return VDRoundToLong(((double)sizePrev + sizeNext)/2);
	} else
		return sizeFrame;
}

sint32 ModAudioSourceMP3::GetFrameLength(MPAFrame *node) {
	if(!node)
		return 0;

	// Reminder :
	// flags & 0xF0 -> bitrate
	// flags & 0x02 -> padding
	sint32 bitrate = 0;
	if(m_MPAInfo.m_Version == MPEG_V1)
		bitrate = 1000 * MPA_Bitrate_V1[m_MPAInfo.m_Layer][(node->m_Flags & 0xF0)>>4];
	else
		bitrate = 1000 * MPA_Bitrate_V2[m_MPAInfo.m_Layer][(node->m_Flags & 0xF0)>>4];

	if(bitrate <= 0)
		return 0;

	if(m_MPAInfo.m_Layer == LAYER_1)
		return (12 * bitrate / m_MPAInfo.m_SampleRate + ((node->m_Flags & 0x02)>>1)) * 4;
	else {
		if((m_MPAInfo.m_Version == MPEG_V1) || (m_MPAInfo.m_Layer == LAYER_2))
			return 144 * bitrate / m_MPAInfo.m_SampleRate + ((node->m_Flags & 0x02)>>1);
		else
			return 72 * bitrate / m_MPAInfo.m_SampleRate + ((node->m_Flags & 0x02)>>1);
	}
}

bool ModAudioSourceMP3::GetNextFrame(bool ensureConsistancy, sint64& bytesSkipped) {
	bytesSkipped = 0;
	sint64 skipped;
	while(true) {
		if(m_VDFile.tell() >= m_EndData)
			return false;
		if(!m_VDFile.synchOn(MPA_SYNCWORD, MPA_SYNCWORD_BITS, MPA_SYNCWORD_LEN, 2*MPA_MAX_LENGTH, skipped)) {
			bytesSkipped += skipped;
			if(m_VDFile.tell() > m_EndData)
				bytesSkipped -= (m_VDFile.tell() - m_EndData);
			return false;
		}
		bytesSkipped += skipped;
		if(GetHeader(ensureConsistancy))
			return true;
		bytesSkipped += 1;
		m_VDFile.seek(-MPA_HEADER_LEN+1, nsVDFile::kSeekCur);
	}
}

bool ModAudioSourceMP3::GetHeader(bool ensureConsistancy) {
	if(m_VDFile.tell() >= m_EndData)
		return false;
	char *buffer = m_MPAInfo.m_Buffer;
	// Read the header (4 bytes)
	if(m_VDFile.readData(buffer, MPA_HEADER_LEN) != MPA_HEADER_LEN)
		return false;

	// Test the syncword
	if( ((buffer[0] & MPA_SYNCWORD_BITS[0]) != MPA_SYNCWORD[0])
	|| ((buffer[1] & MPA_SYNCWORD_BITS[1]) != MPA_SYNCWORD[1]) )
		return false;

	MPA_Version version = MPEG_RESERVED;
	MPA_Layer layer = LAYER_RESERVED;
	sint32 bitrate = 0;
	sint32 samplerate = 0;
	sint32 padding = 0;
	MPA_Channels channels = SINGLE_CHANNEL;
	sint32 nb_channels = 0;

	// Get the MPEG version
	switch((buffer[1] & 0x18)>>3) {
		case 3:
			version = MPEG_V1;
			break;
		case 2:
			version = MPEG_V2;
			break;
		case 0:
			version = MPEG_V25;
			break;
		case 1:
		default:
			version = MPEG_RESERVED;
			break;
	}

	if(version == MPEG_RESERVED)
		return false;

	// Get the Layer version
	switch((buffer[1] & 0x06)>>1) {
		case 3:
			layer = LAYER_1;
			break;
		case 2:
			layer = LAYER_2;
			break;
		case 1:
			layer = LAYER_3;
			break;
		case 0:
		default:
			layer = LAYER_RESERVED;
			break;
	}

	if(layer == LAYER_RESERVED)
		return false;

	// Get the bitrate
	if(version == MPEG_V1)
		bitrate = 1000 * MPA_Bitrate_V1[layer][(buffer[2]&0xF0)>>4];
	else
		bitrate = 1000 * MPA_Bitrate_V2[layer][(buffer[2]&0xF0)>>4];

	if(bitrate <= 0)
		return false;

	// Get the samplerate
	samplerate = MPA_SampleRate[(buffer[2]&0x0C)>>2][version];

	if(samplerate <= 0)
		return false;

	// Get the padding
	padding = (buffer[2]&0x02)>>1;

	switch((buffer[3]&0xC0)>>6) {
		case 0:
			channels = STEREO;
			break;
		case 1:
			channels = JOINT_STEREO;
			break;
		case 2:
			channels = DUAL_CHANNEL;
			break;
		case 3:
		default:
			channels = SINGLE_CHANNEL;
			break;
	}

	nb_channels = MPA_NbChannels[channels];

	// If asked, verify the stream remain consistant
	// i.e. same mpeg version, layer, samplerate, channels number
	if(ensureConsistancy && (
		(version != m_MPAInfo.m_Version)
		|| (layer != m_MPAInfo.m_Layer)
		|| (samplerate != m_MPAInfo.m_SampleRate)
		|| (nb_channels != m_MPAInfo.m_NbChannels)
		/*|| (channels != m_MPAInfo.m_Channels)*/ ) )
		return false;

	if(!m_bIsVBR && (bitrate != m_MPAInfo.m_Bitrate)) {
		if(m_MPAInfo.m_Frames)
			m_bIsVBR = true;
	}

	m_MPAInfo.m_Version = version;
	m_MPAInfo.m_Layer = layer;
	m_MPAInfo.m_Bitrate = bitrate;
	m_MPAInfo.m_SampleRate = samplerate;
	m_MPAInfo.m_Padding = padding;

	if(!m_MPAInfo.m_Emphasis) {
		m_MPAInfo.m_Channels = channels;
		m_MPAInfo.m_NbChannels = nb_channels;
		m_MPAInfo.m_Emphasis = (buffer[3] & 0x03) + 1;
		m_MPAInfo.m_bProtected = !(buffer[1] & 0x01);
		m_MPAInfo.m_bPrivate = !!(buffer[2] & 0x01);
		m_MPAInfo.m_bCopyrighted = !!((buffer[3]&0x08)>>3);
		m_MPAInfo.m_bOriginal = !!((buffer[3]&0x04)>>2);
	} else {
		if(!m_bEmphasisChanged && (m_MPAInfo.m_Emphasis != (buffer[3] & 0x03) + 1) ) {
			m_bEmphasisChanged = true;
			char *name = "emphasis";
			VDLogAppMessage(kVDLogWarning, kVDST_AudioSourceMPA, kVDM_InformationChange, 1, &name);
		}
		if(!m_bProtectedChanged && (m_MPAInfo.m_bProtected != !(buffer[1] & 0x01)) ) {
			m_bProtectedChanged = true;
			// This bit must be set if any of the frames has it
			// Since its value changed it must be set
			m_MPAInfo.m_bProtected = true;
			char *name = "protected";
			VDLogAppMessage(kVDLogWarning, kVDST_AudioSourceMPA, kVDM_InformationChange, 1, &name);
		}
		if(!m_bPrivateChanged && (m_MPAInfo.m_bPrivate != !!(buffer[2] & 0x01)) ) {
			m_bPrivateChanged = true;
			m_MPAInfo.m_bPrivate = true;
			char *name = "private";
			VDLogAppMessage(kVDLogWarning, kVDST_AudioSourceMPA, kVDM_InformationChange, 1, &name);
		}
		if(!m_bCopyrightedChanged && (m_MPAInfo.m_bCopyrighted != !!((buffer[3]&0x08)>>3)) ) {
			m_bCopyrightedChanged = true;
			m_MPAInfo.m_bCopyrighted = true;
			char *name = "copyrighted";
			VDLogAppMessage(kVDLogWarning, kVDST_AudioSourceMPA, kVDM_InformationChange, 1, &name);
		}
		if(!m_bOriginalChanged && (m_MPAInfo.m_bOriginal != !!((buffer[3]&0x04)>>2)) ) {
			m_bOriginalChanged = true;
			m_MPAInfo.m_bOriginal = true;
			char *name = "original";
			VDLogAppMessage(kVDLogWarning, kVDST_AudioSourceMPA, kVDM_InformationChange, 1, &name);
		}
	}
	m_MPAInfo.m_HeadMode |= 0x1 << ((buffer[3]&0xC0)>>6);
	m_MPAInfo.m_HeadModeExt |= 0x1 << ((buffer[3]&0x30)>>4);

	m_MPAInfo.m_SamplesPerFrame = MPA_FrameSize[layer][version];
	if(layer == LAYER_1)
		m_MPAInfo.m_FrameLength = (12 * bitrate / samplerate + padding) * 4;
	else {
		if((version == MPEG_V1) || (layer == LAYER_2))
			m_MPAInfo.m_FrameLength = 144 * bitrate / samplerate + padding;
		else
			m_MPAInfo.m_FrameLength = 72 * bitrate / samplerate + padding;
	}

	return true;
}

void ModAudioSourceMP3::ComputeFramesInNode(void) {
	MPAFrame *next = m_Iterator.next();
	if( ((*m_Iterator)->m_Flags == CORRUPTED_MPA_FRAME) && !m_MPAOptions.m_bKeepCorruptedData)
		m_FrameCount = 0;
	else
		m_FrameCount = VDRoundToLong( ((double)(next?next->m_FileOffset:m_EndData)-(*m_Iterator)->m_FileOffset) / ComputeFrameLength(m_Iterator.previous(), (*m_Iterator), m_Iterator.next()));
}
