//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <vd2/system/cpuaccel.h>
#include "ModAudioStreamSourceAC3.h"
#include "ModAudioSourceAC3.h"


static const char AC3_SYNCWORD[AC3_SYNCWORD_LEN] =
{ 0x0B, 0x77};

static void __declspec(naked) __cdecl a52_convert_stereo(void *src, void *dst) {
	__asm {
			push    ebx
			push    ecx
			push    esi
			push    edi
			mov     esi, [esp+4+16]     ; src
			mov     edi, [esp+8+16]     ; dst
			mov     ecx, 256
loop1:      mov     eax, [esi]
			mov     ebx, [esi+256*4]
//  43BF8000 is min: 01000011 10111111 10000000 00000000
//  43C07FFF is max: 01000011 11000000 01111111 11111111
			cmp     eax, 43C07FFFh
			ja      clamp1a
			cmp     eax, 43BF8000h
			jb      clamp1b
			shl     eax, 16
clamp1done: cmp     ebx, 43C07FFFh
			ja      clamp2a
			cmp     ebx, 43BF8000h
			jb      clamp2b
clamp2done: shld    ebx, eax, 16
			add     esi, 4
			mov     [edi], ebx
			add     edi, 4
			loop    loop1
			pop     edi
			pop     esi
			pop     ecx
			pop     ebx
			ret
clamp1a:    mov     eax, 7FFF0000h
			jmp     clamp1done
clamp1b:    mov     eax, 80000000h
			jmp     clamp1done
clamp2a:    mov     ebx, 00007FFFh
			jmp     clamp2done
clamp2b:    mov     ebx, 00008000h
			jmp     clamp2done
	}
}

static void __declspec(naked) __cdecl a52_convert_mono(void *src, void *dst) {
	__asm {
			push    ecx
			push    esi
			push    edi
			mov     esi, [esp+4+12]     ; src
			mov     edi, [esp+8+12]     ; dst
			mov     ecx, 256
loop1:      mov     eax, [esi]
			add     esi, 4
			cmp     eax, 43C07FFFh
			ja      clamp1a
			cmp     eax, 43BF8000h
			jb      clamp1b
clamp1done: mov     [edi], ax
			add     edi, 2
			loop    loop1
			pop     edi
			pop     esi
			pop     ecx
			ret
clamp1a:    mov     eax, 00007FFFh
			jmp     clamp1done
clamp1b:    mov     eax, 00008000h
			jmp     clamp1done
	}
}

static inline void MoveData(char *dst, const char *src, int length) {
	while(--length >= 0) {
		*dst = *src;
		dst++;
		src++;
	}
}

ModAudioStreamSourceAC3::ModAudioStreamSourceAC3(AudioSource *source, long first_samp, long max_samples)
: AudioStream() {
	m_Source = source;
	m_BlankSamples = 0;
	m_BytesInInputBuffer = 0;
	m_BytesInOutputBuffer = 0;
	m_LastFrameLength = 768;
	m_bSourceEOS = false;
	m_bSeemsVBR = false;

	m_SourceBlockAlign = m_Source->getWaveFormat()->nBlockAlign;
	m_SourceRate = m_Source->getStreamInfo().dwRate;
	m_SourceScale = m_Source->getStreamInfo().dwScale;
	// Is this a VBR stream ?
	// Simple test : request a small amount of samples (1 and 2 for example)
	// If we get exactly what we requested, in the expected number of bytes
	// then it is more likely to be a CBR stream
	LONG bytesRead, samplesRead;
	if( (m_Source->_read(0, 1, NULL, 0x7FFFFFFF, &bytesRead, &samplesRead)==AVIERR_OK)
		&& ((samplesRead!=1) || (bytesRead!=m_SourceBlockAlign)) )
		m_bSeemsVBR = true;
	if( !m_bSeemsVBR && (m_Source->_read(0, 2, NULL, 0x7FFFFFFF, &bytesRead, &samplesRead)==AVIERR_OK)
		&& ((samplesRead!=2) || (bytesRead!=2*m_SourceBlockAlign)) )
		m_bSeemsVBR = true;
	VDASSERT( (m_bSeemsVBR && ((m_SourceBlockAlign!=m_SourceScale) || (m_Source->getWaveFormat()->nAvgBytesPerSec!=m_SourceRate)))
		|| (!m_bSeemsVBR && (m_SourceBlockAlign==m_SourceScale) && (m_Source->getWaveFormat()->nAvgBytesPerSec==m_SourceRate)) );

	if (max_samples < 0)
		max_samples = 0;

	WAVEFORMATEX *iFormat = m_Source->getWaveFormat();
	WAVEFORMATEX *oFormat;

	if (!AllocFormat(sizeof(WAVEFORMATEX)))
		throw MyMemoryError();

	oFormat = (WAVEFORMATEX *)GetFormat();

	int a52flags, a52srate, a52brate;
	a52state = a52_init((MMX_enabled)? MM_ACCEL_X86_MMX: 0);

	long blank_samples = 0;
	if(first_samp < 0) {
		blank_samples = -first_samp;
		first_samp = 0;
	}
	cur_samp = first_samp;
	end_samp = first_samp + max_samples;

	bool bSuccess = false;
	while(a52state && GetNextFrame(false)) {
		if(a52_syncinfo((uint8_t *)m_InputBuffer, &a52flags, &a52srate, &a52brate) >= 8) {
			bSuccess = true;
			break;
		}
		GetNextFrame(false, true);
	}
	if(bSuccess) {
		oFormat->wFormatTag			= WAVE_FORMAT_PCM;
		oFormat->nChannels			= (a52flags == A52_MONO)? 1: 2;
		oFormat->nBlockAlign		= oFormat->nChannels*2;
		oFormat->nSamplesPerSec		= a52srate;
		oFormat->nAvgBytesPerSec	= oFormat->nSamplesPerSec * oFormat->nBlockAlign;
		oFormat->wBitsPerSample		= 16;
		oFormat->cbSize				= 0;
	} else if(a52state) {
		a52_free(a52state);
		a52state = NULL;
	}

	stream_len = std::min<long>(max_samples, m_Source->getEnd() - (blank_samples ? -blank_samples : first_samp));
	stream_len = MulDiv(stream_len
		, GetFormat()->nSamplesPerSec * m_SourceScale
		, m_SourceRate);

	m_BlankSamples += MulDiv(blank_samples
		, GetFormat()->nSamplesPerSec * m_SourceScale
		, m_SourceRate);

	m_BlockAlign = oFormat->nBlockAlign;

}

ModAudioStreamSourceAC3::~ModAudioStreamSourceAC3(void) {
	if(a52state) {
		a52_free(a52state);
		a52state = NULL;
	}
}

bool ModAudioStreamSourceAC3::ReadSource(long bytes) {
	if(m_BytesInInputBuffer >= bytes)
		return true;
	LONG bytesRead, samplesRead;
	if(!m_bSeemsVBR) {
		if(m_Source->read(cur_samp
			, (bytes - m_BytesInInputBuffer + m_SourceBlockAlign-1) / m_SourceBlockAlign
			, m_InputBuffer + m_BytesInInputBuffer
			, sizeof(m_InputBuffer)-m_BytesInInputBuffer
			, &bytesRead, &samplesRead) != AVIERR_OK)
			return false;
		cur_samp += samplesRead;
		m_BytesInInputBuffer += bytesRead;
	} else {
		int res;
		while( (m_BytesInInputBuffer<bytes) && ((res=m_Source->read(cur_samp
			, AVISTREAMREAD_CONVENIENT
			, m_InputBuffer + m_BytesInInputBuffer
			, sizeof(m_InputBuffer)-m_BytesInInputBuffer
			, &bytesRead, &samplesRead))==AVIERR_OK) && samplesRead) {
			VDDEBUG("Read %d samples starting at %d, %d bytes\n"
				, samplesRead, cur_samp, bytesRead);
			cur_samp += samplesRead;
			m_BytesInInputBuffer += bytesRead;
		}
	}
	return (m_BytesInInputBuffer>=bytes);
}

bool ModAudioStreamSourceAC3::GetNextFrame(bool bDecode, bool bSkip) {
	if(!ReadSource(AC3_HEADER_LEN)) {
		m_bSourceEOS = true;
		return false;
	}

	if( (m_InputBuffer[0]==AC3_SYNCWORD[0]) && (m_InputBuffer[1]==AC3_SYNCWORD[1]) ) {
		m_LastFrameLength = 2*AC3_FrameLength[m_InputBuffer[4] & 0x3F][(m_InputBuffer[4] & 0xC0)>>6];
		if(!ReadSource(m_LastFrameLength)) {
			m_bSourceEOS = true;
			return false;
		}
	} else {
		long skippedBytes = 0;
		int synchMatch = 0;
		while(true) {
			if(!m_bSeemsVBR)
				ReadSource(sizeof(m_InputBuffer)-m_SourceBlockAlign);
			if(!ReadSource(AC3_HEADER_LEN)) {
				m_bSourceEOS = true;
				return false;
			}
			for(int i=0 ; i<m_BytesInInputBuffer ; i++) {
				skippedBytes++;
				if(m_InputBuffer[i] == AC3_SYNCWORD[synchMatch]) {
					synchMatch++;
					if(synchMatch == AC3_SYNCWORD_LEN) {
						i -= (AC3_SYNCWORD_LEN-1);
						MoveData(m_InputBuffer, m_InputBuffer+i, m_BytesInInputBuffer-i);
						m_BytesInInputBuffer -= i;
						skippedBytes -= AC3_SYNCWORD_LEN;
						m_BlankSamples += MulDiv(skippedBytes, AC3_SAMPLESPERFRAME, m_LastFrameLength);
						return GetNextFrame(bDecode, bSkip);
					}
				} else
					synchMatch = 0;
			}
			if(synchMatch) {
				MoveData(m_InputBuffer, m_InputBuffer + m_BytesInInputBuffer - synchMatch, synchMatch);
				m_BytesInInputBuffer = synchMatch;
			} else
				m_BytesInInputBuffer = 0;
		}
	}

	if(!bDecode) {
		if(bSkip) {
			MoveData(m_InputBuffer, m_InputBuffer+m_LastFrameLength, m_BytesInInputBuffer-m_LastFrameLength);
			m_BytesInInputBuffer -= m_LastFrameLength;
		}
		return true;
	}

	m_OutputBufferPos = NULL;
	m_BytesInOutputBuffer = 0;

	try {
		int a52flags = A52_DOLBY | A52_ADJUST_LEVEL;
		sample_t a52level = 1.0;
		a52_frame(a52state, (uint8_t *)m_InputBuffer, &a52flags, &a52level, 384.0);

		for (int i = 0; i < 6; i++) {
			a52_block(a52state);
			if (GetFormat()->nChannels > 1) {
				a52_convert_stereo(a52_samples(a52state),
					(long *)m_OutputBuffer+i*256);
			} else {
				a52_convert_mono(a52_samples(a52state),
					(short *)m_OutputBuffer+i*256);
			}
		}
	} catch(...) {
		m_bSourceEOS = true;
		throw MyError("AC3 decoder error");
		return false;
	}

	MoveData(m_InputBuffer, m_InputBuffer+m_LastFrameLength, m_BytesInInputBuffer-m_LastFrameLength);
	m_BytesInInputBuffer -= m_LastFrameLength;

	m_OutputBufferPos = m_OutputBuffer;
	m_BytesInOutputBuffer = m_BlockAlign * AC3_SAMPLESPERFRAME;
	return true;
}

long ModAudioStreamSourceAC3::_Read(void *buffer, long max_samples, long *lplBytes) {
	if(!a52state) {
		if(lplBytes)
			*lplBytes = 0;
		return 0;
	}

	LONG lBytesLeft = max_samples * m_BlockAlign;
	LONG lTotalBytes = 0;

	while(lBytesLeft > 0) {
		if(m_BlankSamples) {
			long blankBytes = m_BlockAlign * m_BlankSamples;
			if(blankBytes > lBytesLeft)
				blankBytes = lBytesLeft;
			long blankSamples = blankBytes / m_BlockAlign;

			memset(buffer, 0, blankBytes);

			buffer = (char *)buffer + blankBytes;

			lTotalBytes += blankBytes;
			lBytesLeft -= blankBytes;
			m_BlankSamples -= blankSamples;
			continue;
		}

		if(m_BytesInOutputBuffer) {
			long toRead = lBytesLeft;
			if(toRead > m_BytesInOutputBuffer)
				toRead = m_BytesInOutputBuffer;
			memcpy(buffer, m_OutputBufferPos, toRead);
			buffer = (void *)((char *)buffer + toRead);
			lBytesLeft -= toRead;
			lTotalBytes += toRead;
			m_BytesInOutputBuffer -= toRead;
			m_OutputBufferPos += toRead;
			continue;
		}

		if(!GetNextFrame(true))
			break;
	}

	*lplBytes = lTotalBytes;

	return lTotalBytes / m_BlockAlign;
}


bool ModAudioStreamSourceAC3::Skip(long samples) {
	while(samples) {
		if(samples <= m_BlankSamples) {
			m_BlankSamples -= samples;
			return true;
		}
		samples -= m_BlankSamples;
		m_BlankSamples = 0;
		long samplesInBuffer = m_BytesInOutputBuffer / m_BlockAlign;
		if(samples <= samplesInBuffer) {
			m_BytesInOutputBuffer -= samples * m_BlockAlign;
			m_OutputBufferPos += samples * m_BlockAlign;
			return true;
		}
		m_BytesInOutputBuffer = 0;
		samples -= samplesInBuffer;
		if(!GetNextFrame(false))
			return false;
	}
	return true;
}

BOOL ModAudioStreamSourceAC3::_isEnd() {
	return ((cur_samp>=end_samp) && m_bSourceEOS && !m_BlankSamples && !m_BytesInOutputBuffer);
}
