//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <vd2/system/cpuaccel.h>
#include "ModAudioStreamSourceMP3.h"
#include "ModAudioSourceMP3.h"


static const unsigned char MPA_SYNCWORD[MPA_SYNCWORD_LEN] =
{ 0xFF, 0xE0};
static const unsigned char MPA_SYNCWORD_BITS[MPA_SYNCWORD_LEN] =
{ 0xFF, 0xE0};

static inline void MoveData(char *dst, const char *src, int length) {
	while(--length >= 0) {
		*dst = *src;
		dst++;
		src++;
	}
}

static bool IsValidMPAHeader(const char *pBuffer, int& version, int& layer, sint32& bitrate, sint32& samplerate, sint16& nb_channels) {
	if( ((pBuffer[0]&MPA_SYNCWORD_BITS[0])!=MPA_SYNCWORD[0])
		|| ((pBuffer[1]&MPA_SYNCWORD_BITS[1])!=MPA_SYNCWORD[1]) )
		return false;
	switch((pBuffer[1] & 0x18)>>3) {
		case 3:
			version = MPEG_V1;
			break;
		case 2:
			version = MPEG_V2;
			break;
		case 0:
			version = MPEG_V25;
			break;
		case 1:
		default:
			version = MPEG_RESERVED;
			break;
	}
	switch((pBuffer[1] & 0x06)>>1) {
		case 3:
			layer = LAYER_1;
			break;
		case 2:
			layer = LAYER_2;
			break;
		case 1:
			layer = LAYER_3;
			break;
		case 0:
		default:
			layer = LAYER_RESERVED;
			break;
	}
	if( (version == MPEG_RESERVED) || (layer==LAYER_RESERVED) )
		return false;

	if(version == MPEG_V1)
		bitrate = 1000 * MPA_Bitrate_V1[layer][(pBuffer[2]&0xF0)>>4];
	else
		bitrate = 1000 * MPA_Bitrate_V2[layer][(pBuffer[2]&0xF0)>>4];
	if(bitrate <= 0)
		return false;

	samplerate = MPA_SampleRate[(pBuffer[2]&0x0C)>>2][version];
	if(samplerate <= 0)
		return false;
	int channels;
	switch((pBuffer[3]&0xC0)>>6) {
		case 0:
			channels = STEREO;
			break;
		case 1:
			channels = JOINT_STEREO;
			break;
		case 2:
			channels = DUAL_CHANNEL;
			break;
		case 3:
		default:
			channels = SINGLE_CHANNEL;
			break;
	}
	nb_channels = MPA_NbChannels[channels];
	return true;
}

ModAudioStreamSourceMP3::ModAudioStreamSourceMP3(AudioSource *source, long first_samp, long max_samples)
: AudioStream() {
	m_Source = source;
	m_BlankSamples = 0;
	m_BytesInInputBuffer = 0;
	m_BytesInOutputBuffer = 0;
	m_SamplesPerFrame = 0;
	m_BytesSent = 0;
	m_DecodeStart = 0;
	m_bSourceEOS = false;
	m_bSeemsVBR = false;

	m_SourceBlockAlign = m_Source->getWaveFormat()->nBlockAlign;
	m_SourceRate = m_Source->getStreamInfo().dwRate;
	m_SourceScale = m_Source->getStreamInfo().dwScale;
	// Is this a VBR stream ?
	// Simple test : request a small amount of samples (1 and 2 for example)
	// If we get exactly what we requested, in the expected number of bytes
	// then it is more likely to be a CBR stream
	LONG bytesRead, samplesRead;
	if( (m_Source->_read(0, 1, NULL, 0x7FFFFFFF, &bytesRead, &samplesRead)==AVIERR_OK)
		&& ((samplesRead!=1) || (bytesRead!=m_SourceBlockAlign)) )
		m_bSeemsVBR = true;
	if( !m_bSeemsVBR && (m_Source->_read(0, 2, NULL, 0x7FFFFFFF, &bytesRead, &samplesRead)==AVIERR_OK)
		&& ((samplesRead!=2) || (bytesRead!=2*m_SourceBlockAlign)) )
		m_bSeemsVBR = true;
	VDASSERT( (m_bSeemsVBR && ((m_SourceBlockAlign!=m_SourceScale) || (m_Source->getWaveFormat()->nAvgBytesPerSec!=m_SourceRate)))
		|| (!m_bSeemsVBR && (m_SourceBlockAlign==m_SourceScale) && (m_Source->getWaveFormat()->nAvgBytesPerSec==m_SourceRate)) );

	if (max_samples < 0)
		max_samples = 0;

	WAVEFORMATEX *iFormat = m_Source->getWaveFormat();
	WAVEFORMATEX *oFormat;

	if (!AllocFormat(sizeof(WAVEFORMATEX)))
		throw MyMemoryError();

	m_bNandubVBR = (((iFormat->wFormatTag==WAVE_FORMAT_MPEGLAYER3) || (iFormat->wFormatTag==WAVE_FORMAT_MPEG))
		&& (iFormat->nBlockAlign==1152));
	// There are 3 cases in which we can find an nBlockalign of 1152 for an MP3 stream :
	// 1. MP3 VBR stream coming from ModAudioSourceMP3 and using the Nandub trick
	// 2. MP3 stream coming from an AVI file, and using the Nandub trick
	// 3. MP3 stream coming from an AVI file (chunks length is a multiple of 1152 then)
	// Case 3 is quite rare (and weird) but I once found such a file (apparently made with
	// Nandub). Since applying Nandub tricks to a stream that doesn't use them lead to serious
	// out-of-synch issues, we do an extra test here :
	// IsVBR() returns true in Case 1&2, and false in case 3
	IMP3Source *sMP3;
	if(m_bNandubVBR && source->QueryInterface(IMP3Source::GetInterfaceID(), (void **)&sMP3))
		m_bNandubVBR = sMP3->IsVBR();
	VDASSERT(!m_bNandubVBR || (m_bNandubVBR && m_bSeemsVBR));

	oFormat = (WAVEFORMATEX *)GetFormat();

	iad = CreateAMPDecoder();

	long blank_samples = 0;
	if(first_samp < 0) {
		blank_samples = -first_samp;
		first_samp = 0;
	}
	cur_samp = first_samp;
	end_samp = first_samp + max_samples;

	try {
		bool bSuccess = false;
		while(iad && GetNextFrame(false)) {
			int version;
			int layer;
			sint32 bitrate;
			sint32 samplerate;
			sint16 nb_channels;
			if(!IsValidMPAHeader(m_InputBuffer, version, layer, bitrate, samplerate, nb_channels)) {
				GetNextFrame(false, true);
				continue;
			}

			m_SamplesPerFrame = MPA_FrameSize[layer][version];

			if(layer == LAYER_3) {
				if(version == MPEG_V1)
					m_DecodeStart = 512;
				else
					m_DecodeStart = 255;
			} else if(version == MPEG_V1)
				m_DecodeStart = 32;

			oFormat->wFormatTag			= WAVE_FORMAT_PCM;
			oFormat->nChannels			= nb_channels;
			oFormat->nBlockAlign		= oFormat->nChannels*2;
			oFormat->nSamplesPerSec		= samplerate;
			oFormat->nAvgBytesPerSec	= oFormat->nSamplesPerSec * oFormat->nBlockAlign;
			oFormat->wBitsPerSample		= 16;
			oFormat->cbSize				= 0;

			bSuccess = true;

			iad->Init();
			iad->setSource(this);
			iad->Reset();

			break;
		}
		if(!bSuccess && iad) {
			delete iad;
			iad = NULL;
		}
	} catch(int i) {
		throw MyError(iad->getErrorString(i));
	}

	stream_len = std::min<long>(max_samples, m_Source->getEnd() - (blank_samples ? -blank_samples : first_samp));
	if(m_bNandubVBR) {
		stream_len = stream_len * m_SamplesPerFrame;
		m_BlankSamples += blank_samples * m_SamplesPerFrame;
	} else {
		stream_len = MulDiv(stream_len
			, GetFormat()->nSamplesPerSec * m_SourceScale
			, m_SourceRate);
		m_BlankSamples += MulDiv(blank_samples
			, GetFormat()->nSamplesPerSec * m_SourceScale
			, m_SourceRate);
	}

	m_BlockAlign = oFormat->nBlockAlign;

}

ModAudioStreamSourceMP3::~ModAudioStreamSourceMP3(void) {
	if(iad)
		delete iad;
}

bool ModAudioStreamSourceMP3::ReadSource(long bytes) {
	if(m_BytesInInputBuffer >= bytes)
		return true;
	LONG bytesRead, samplesRead;
	if(!m_bSeemsVBR) {
		if(m_Source->read(cur_samp
			, (bytes - m_BytesInInputBuffer + m_SourceBlockAlign-1) / m_SourceBlockAlign
			, m_InputBuffer + m_BytesInInputBuffer
			, sizeof(m_InputBuffer)-m_BytesInInputBuffer
			, &bytesRead, &samplesRead) != AVIERR_OK)
			return false;
		cur_samp += samplesRead;
		m_BytesInInputBuffer += bytesRead;
	} else {
		int res;
		while( (m_BytesInInputBuffer<bytes) && ((res=m_Source->read(cur_samp
			, AVISTREAMREAD_CONVENIENT
			, m_InputBuffer + m_BytesInInputBuffer
			, sizeof(m_InputBuffer)-m_BytesInInputBuffer
			, &bytesRead, &samplesRead))==AVIERR_OK) && samplesRead) {
			cur_samp += samplesRead;
			m_BytesInInputBuffer += bytesRead;
		}
	}
	return (m_BytesInInputBuffer>=bytes);
}

bool ModAudioStreamSourceMP3::GetNextFrame(bool bDecode, bool bSkip) {
	if(!ReadSource(MPA_HEADER_LEN)) {
		m_bSourceEOS = true;
		return false;
	}

	if( ((m_InputBuffer[0]&MPA_SYNCWORD_BITS[0])==MPA_SYNCWORD[0])
		&& ((m_InputBuffer[1]&MPA_SYNCWORD_BITS[1])==MPA_SYNCWORD[1]) ) {
	} else {
		int synchMatch = 0;
		while(true) {
			if(!m_bSeemsVBR)
				ReadSource(sizeof(m_InputBuffer)-m_SourceBlockAlign);
			if(!ReadSource(MPA_HEADER_LEN)) {
				m_bSourceEOS = true;
				return false;
			}
			for(int i=0 ; i<m_BytesInInputBuffer ; i++) {
				if( (m_InputBuffer[i]&MPA_SYNCWORD_BITS[synchMatch]) == MPA_SYNCWORD[synchMatch]) {
					synchMatch++;
					if(synchMatch == MPA_SYNCWORD_LEN) {
						i -= (MPA_SYNCWORD_LEN-1);
						MoveData(m_InputBuffer, m_InputBuffer+i, m_BytesInInputBuffer-i);
						m_BytesInInputBuffer -= i;
						if(!ReadSource(MPA_HEADER_LEN)) {
							m_bSourceEOS = true;
							return false;
						}
						int version;
						int layer;
						sint32 bitrate;
						sint32 samplerate;
						sint16 nb_channels;
						if(IsValidMPAHeader(m_InputBuffer, version, layer, bitrate, samplerate, nb_channels))
							return GetNextFrame(bDecode);
						synchMatch = 0;
						// Make sure we won't get the first bytes as synch bytes again
						m_InputBuffer[0] = 0;
						break;
					}
				} else
					synchMatch = 0;
			}
			if(synchMatch) {
				MoveData(m_InputBuffer, m_InputBuffer + m_BytesInInputBuffer - synchMatch, synchMatch);
				m_BytesInInputBuffer = synchMatch;
			} else
				m_BytesInInputBuffer = 0;
		}
	}

	if(!bDecode) {
		if(bSkip) {
			MoveData(m_InputBuffer, m_InputBuffer+2, m_BytesInInputBuffer-2);
			m_BytesInInputBuffer -= 2;
		}
		return true;
	}

	m_OutputBufferPos = NULL;
	m_BytesInOutputBuffer = 0;

	try {
		iad->setDestination((short *)m_OutputBuffer);
		iad->ReadHeader();
		// Hmm prereading frames give me some clicks at the beginning in debugging builds :/
		if(m_BytesSent < m_DecodeStart)
			iad->PrereadFrame();
		else
			iad->DecodeFrame();
	} catch(int err) {
		m_bSourceEOS = true;
		throw MyError("MP3 decoder error : %s", iad->getErrorString(err));
		return false;
	}

	m_OutputBufferPos = m_OutputBuffer;
	m_BytesInOutputBuffer = m_BlockAlign * m_SamplesPerFrame;
	return true;
}

long ModAudioStreamSourceMP3::_Read(void *buffer, long max_samples, long *lplBytes) {
	if(!iad) {
		if(lplBytes)
			*lplBytes = 0;
		return 0;
	}

	LONG lBytesLeft = max_samples * m_BlockAlign;
	LONG lTotalBytes = 0;

	while(lBytesLeft > 0) {
		if(m_BlankSamples) {
			long blankBytes = m_BlockAlign * m_BlankSamples;
			if(blankBytes > lBytesLeft)
				blankBytes = lBytesLeft;
			long blankSamples = blankBytes / m_BlockAlign;

			memset(buffer, 0, blankBytes);

			buffer = (char *)buffer + blankBytes;

			lTotalBytes += blankBytes;
			lBytesLeft -= blankBytes;
			m_BlankSamples -= blankSamples;
			continue;
		}

		if(m_BytesInOutputBuffer) {
			long toRead = lBytesLeft;
			if(toRead > m_BytesInOutputBuffer)
				toRead = m_BytesInOutputBuffer;
			memcpy(buffer, m_OutputBufferPos, toRead);
			buffer = (void *)((char *)buffer + toRead);
			lBytesLeft -= toRead;
			lTotalBytes += toRead;
			m_BytesInOutputBuffer -= toRead;
			m_OutputBufferPos += toRead;
			continue;
		}

		if(!GetNextFrame(true))
			break;
	}

	*lplBytes = lTotalBytes;

	return lTotalBytes / m_BlockAlign;
}

bool ModAudioStreamSourceMP3::Skip(long samples) {
	iad->Reset();
	m_BytesSent = 0;
	if(m_BlankSamples) {
		if(samples <= m_BlankSamples) {
			m_BlankSamples -= samples;
			return true;
		}
		samples -= m_BlankSamples;
		m_BlankSamples = 0;
	}
	long samplesInBuffer = m_BytesInOutputBuffer / m_BlockAlign;
	if(samples <= samplesInBuffer) {
		m_BytesInOutputBuffer -= samples * m_BlockAlign;
		m_OutputBufferPos += samples * m_BlockAlign;
		return true;
	}
	m_BytesInOutputBuffer = 0;
	samples -= samplesInBuffer;
	if(m_bNandubVBR) {
		long delta = (samples + m_SamplesPerFrame - 1) / m_SamplesPerFrame;
		m_BlankSamples = delta * m_SamplesPerFrame - samples;
		cur_samp += delta;
	} else {
		long delta = MulDiv(samples
			, m_SourceRate
			, GetFormat()->nSamplesPerSec * m_SourceScale);
		long lag = m_Source->samplesLag(cur_samp + delta);
		if(lag > 0)
			m_BlankSamples = lag;
		else if(lag < 0) {
			m_Source->_read(cur_samp + delta, AVISTREAMREAD_CONVENIENT, NULL, 0x7FFFFFFF, NULL, &m_BlankSamples);
			m_BlankSamples += lag;
			lag = m_BlankSamples;
			m_BlankSamples = MulDiv(m_BlankSamples
				, GetFormat()->nSamplesPerSec * m_SourceScale
				, m_SourceRate);
		}
		cur_samp += delta + lag;
	}
	return true;
}

int ModAudioStreamSourceMP3::read(void *buffer, int bytes) {
	int bytesToRead = bytes;
	int toRead;
	bool eos;
	while(bytesToRead > 0) {
		toRead = bytesToRead;
		if(toRead > m_BytesInInputBuffer)
			toRead = m_BytesInInputBuffer;
		if(toRead) {
			memcpy(buffer, m_InputBuffer, toRead);
			m_BytesInInputBuffer -= toRead;
			bytesToRead -= toRead;
			buffer = (char *)buffer + toRead;
			MoveData(m_InputBuffer, m_InputBuffer+toRead, m_BytesInInputBuffer);
		}
		if(bytesToRead > 0) {
			if(!m_bSeemsVBR)
				eos = !ReadSource(std::min<int>(bytesToRead, sizeof(m_InputBuffer) - m_SourceBlockAlign));
			else {
				if(!ReadSource(1))
					break;
			}
			toRead = bytesToRead;
			if(toRead > m_BytesInInputBuffer)
				toRead = m_BytesInInputBuffer;
			VDASSERT(toRead > 0);
			memcpy(buffer, m_InputBuffer, toRead);
			m_BytesInInputBuffer -= toRead;
			bytesToRead -= toRead;
			buffer = (char *)buffer + toRead;
			MoveData(m_InputBuffer, m_InputBuffer+toRead, m_BytesInInputBuffer);
			if(eos)
				break;
		}
	}
	m_BytesSent += bytes - bytesToRead;
	return bytes - bytesToRead;
}

BOOL ModAudioStreamSourceMP3::_isEnd() {
	return ((cur_samp>=end_samp) && m_bSourceEOS && !m_BlankSamples && !m_BytesInOutputBuffer);
}
