//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "ModTextSource.h"

#include <vd2/system/Error.h>


struct sBOMInformation {
	eTextType	m_TextType;
	char		*m_BOM;
	uint8		m_BOMLength;
};

static sBOMInformation s_BOMList[] = {
	{kTextType_UTF32BE,	"\x00\x00\xFE\xFF",	4},
	{kTextType_UTF32LE,	"\xFF\xFE\x00\x00",	4},
	{kTextType_UTF16LE,	"\xFF\xFE",			2},
	{kTextType_UTF16BE,	"\xFE\xFF",			2},
	{kTextType_UTF8,	"\xEF\xBB\xBF",		3},
	{kTextType_ASCII,	"",					0},
	{kTextType_Unknown,	NULL,				0}
};

bool GetBOMInformation(eTextType textType, void * & BOM, uint8& BOMLength) {
	BOM = NULL;
	BOMLength = 0;
	int i;
	for(i=0
		;s_BOMList[i].m_TextType!=kTextType_Unknown || s_BOMList[i].m_BOM || s_BOMList[i].m_BOMLength
		;i++) {
		if(s_BOMList[i].m_TextType == textType) {
			BOM = s_BOMList[i].m_BOM;
			BOMLength = s_BOMList[i].m_BOMLength;
			return true;
		}
	}
	return false;
}

ModTextSource::ModTextSource(void) {
	m_TextType = kTextType_Unknown;
	m_bHasBOM = false;

	m_Buffer.resize(4*1024);
	m_pBufferReadPos = &m_Buffer.front();
	m_iBufferLevel = 0;
}

ModTextSource::~ModTextSource(void) {
	Close();
}

void ModTextSource::Close(void) {
	m_VDFile.close();
	m_TextBuffer.clear();
	m_TextBufferPos = m_TextBuffer.c_str();
	m_pBufferReadPos = &m_Buffer.front();
	m_iBufferLevel = 0;
}

bool ModTextSource::Open(VDStringW fileName) {
	m_FileName = fileName;
	if(m_FileName.empty())
		return false;

	if(!m_VDFile.open(m_FileName.c_str(),
	nsVDFile::kRead | nsVDFile::kDenyWrite | nsVDFile::kOpenExisting | nsVDFile::kSequential))
		return false;

	m_TextType = kTextType_Unknown;
	m_bHasBOM = false;
	sint32 fileSize = m_VDFile.size();
	if(fileSize <= 4)
		return false;

	if(m_VDFile.readData(m_U8, 4) != 4)
		return false;

	bool bHandled = false;
	int sizeWCHAR = sizeof(wchar_t);
	switch(m_U8[0]) {
	case 0xEF:	// maybe UTF-8
		if( (m_U8[1]==0xBB) && (m_U8[2]==0xBF) ) {
			m_TextType = kTextType_UTF8;
			bHandled = true;
			m_bHasBOM = true;
		}
		break;

	case 0xFE:	// maybe UTF-16 big endian
		if((sizeWCHAR==2) && (m_U8[1]==0xFF)) {
			m_TextType = kTextType_UTF16BE;
			bHandled = true;
			m_bHasBOM = true;
		}
		break;

	case 0xFF:	// maybe UTF-16 little endian or UTF-32 little endian
		if(m_U8[1]==0xFE) {
			if( (m_U8[2]==0x00) && (m_U8[3]==0x00) ) {
				if(sizeWCHAR == 4) {
					m_TextType = kTextType_UTF32LE;
					bHandled = true;
					m_bHasBOM = true;
				}
			} else if(sizeWCHAR == 2) {
				m_TextType = kTextType_UTF16LE;
				bHandled = true;
				m_bHasBOM = true;
			}
		}
		break;

	case 0x00:	// maybe UTF-32 big endian
		if( (m_U8[1]==0x00) && (m_U8[2]==0xFE) && (m_U8[3]==0xFF) ) {
			m_TextType = kTextType_UTF32BE;
			bHandled = true;
			m_bHasBOM = true;
		}
		break;

	default:	// maybe ASCII
		if( (m_U8[0]>=0x20) && (m_U8[1]!=0x00) ) {
			m_TextType = kTextType_ASCII;
			bHandled = true;
		}
		break;
	}

	if(!bHandled) {
		// Hmm, it seems there is no BOM in the file, try to guess the format
		bool bGuessed = false;
		m_VDFile.seek(0, nsVDFile::kSeekStart);
		bool bHas00 = false;
		bool bHas000A = false;
		bool bHas0A00 = false;
		bool bHas0000000A = false;
		bool bHas0A000000 = false;
		while(m_VDFile.readData(m_U8, 4) == 4) {
			if( !m_U8[0]
				|| !m_U8[1]
				|| !m_U8[2]
				|| !m_U8[3] )
				bHas00 = true;
			if( (m_U16[0] == 0x000A)
				|| (m_U16[1] == 0x000A) )
				bHas000A = true;
			if( (m_U16[0] == 0x0A00)
				|| (m_U16[1] == 0x0A00) )
				bHas0A00 = true;
			if(m_U32 == 0x0000000A) {
				bHas0000000A = true;
				break;
			}
			if(m_U32 == 0x0A000000) {
				bHas0A000000 = true;
				break;
			}
		}
		if(bHas0000000A && (sizeWCHAR==4)) {
			bGuessed = true;
			m_TextType = kTextType_UTF32LE;
		} else if(bHas0A000000 && (sizeWCHAR==4)) {
			bGuessed = true;
			m_TextType = kTextType_UTF32BE;
		} else if(bHas000A && (sizeWCHAR==2)) {
			bGuessed = true;
			m_TextType = kTextType_UTF16LE;
		} else if(bHas0A00 && (sizeWCHAR==2)) {
			bGuessed = true;
			m_TextType = kTextType_UTF16BE;
		} else if(!bHas00) {
			bGuessed = true;
			m_TextType = kTextType_ASCII;
		}
		if(!bGuessed)
			return false;
	}

	switch(m_TextType) {
	case kTextType_ASCII:
		m_VDFile.seek(0, nsVDFile::kSeekStart);
		break;

	case kTextType_UTF8:
		{
			char *buffer = (char *)malloc(fileSize);
			if(!buffer)
				throw MyMemoryError();
			memset(buffer, 0, fileSize);
			m_VDFile.seek(3, nsVDFile::kSeekStart);
			if(m_VDFile.readData(buffer, fileSize-3) != fileSize-3) {
				free(buffer);
				return false;
			}
			m_TextBuffer = VDTextU8ToW(buffer, -1);
			m_TextBufferPos = m_TextBuffer.c_str();
			free(buffer);
		}
		break;

	case kTextType_UTF16BE:
	case kTextType_UTF16LE:
		m_VDFile.seek(m_bHasBOM ? 2 : 0, nsVDFile::kSeekStart);
		break;

	case kTextType_UTF32BE:
	case kTextType_UTF32LE:
		m_VDFile.seek(m_bHasBOM ? 4 : 0, nsVDFile::kSeekStart);
		break;
	}

	m_U32 = 0xFFFFFFFF;

	return true;
}

bool ModTextSource::eof(void) {
	if(!m_U32)
		return true;

	if(m_TextType == kTextType_UTF8)
		return !(*m_TextBufferPos);

	return false;
}

bool ModTextSource::getLine(VDStringW& line) {
	if(m_TextType == kTextType_UTF8)
		return getLine(m_TextBufferPos, line);

	VDStringA lineANSI;

	line.clear();

	switch(m_TextType) {
	case kTextType_ASCII:
		if(!m_U8[0])
			return false;
		while(m_U8[0]) {
			if(/*m_VDFile.*/readData(m_U8, 1) != 1) {
				m_U32 = 0;
				break;
			}
			if(m_U8[0] == '\n')
				break;
			if(m_U8[0] != '\r')
				lineANSI += m_U8[0];
				//line += VDTextAToW((char *)m_U8, 1);
		}
		line = VDTextAToW(lineANSI);
		return true;
		break;

	case kTextType_UTF16BE:
		if(!m_WCHAR)
			return false;
		while(m_WCHAR) {
			if(/*m_VDFile.*/readData(&m_WCHAR, 2) != 2) {
				m_U32 = 0;
				break;
			}
			m_U8[4] = m_U8[0];
			m_U8[0] = m_U8[1];
			m_U8[1] = m_U8[4];
			if(m_WCHAR == L'\n')
				return true;
			if(m_WCHAR != L'\r')
				line += m_WCHAR;
		}
		return true;
		break;

	case kTextType_UTF16LE:
		if(!m_WCHAR)
			return false;
		while(m_WCHAR) {
			if(/*m_VDFile.*/readData(&m_WCHAR, 2) != 2) {
				m_U32 = 0;
				break;
			}
			if(m_WCHAR == L'\n')
				return true;
			if(m_WCHAR != L'\r')
				line += m_WCHAR;
		}
		return true;
		break;

	case kTextType_UTF32BE:
		if(!m_WCHAR)
			return false;
		while(m_WCHAR) {
			if(/*m_VDFile.*/readData(&m_WCHAR, 4) != 4) {
				m_U32 = 0;
				break;
			}
			m_U8[4] = m_U8[0];
			m_U8[0] = m_U8[3];
			m_U8[3] = m_U8[4];
			m_U8[4] = m_U8[1];
			m_U8[1] = m_U8[2];
			m_U8[2] = m_U8[4];
			if(m_WCHAR == L'\n')
				return true;
			if(m_WCHAR != L'\r')
				line += m_WCHAR;
		}
		return true;
		break;

	case kTextType_UTF32LE:
		if(!m_WCHAR)
			return false;
		while(m_WCHAR) {
			if(/*m_VDFile.*/readData(&m_WCHAR, 4) != 4) {
				m_U32 = 0;
				break;
			}
			if(m_WCHAR == L'\n')
				return true;
			if(m_WCHAR != L'\r')
				line += m_WCHAR;
		}
		return true;
		break;

	}
	return false;
}

bool ModTextSource::getLine(const wchar_t * & buffer, VDStringW& line) {
	line.clear();
	if(!*buffer)
		return false;
	while(*buffer) {
		if(*buffer == L'\n') {
			buffer++;
			return true;
		}
		if(*buffer != L'\r')
			line += *buffer;
		buffer++;
	}
	return true;
}

sint32 ModTextSource::readData(void *pBuffer, sint32 iBytes) {
	sint32 iBytesRead = 0;

	while(iBytes > 0) {
		if(m_iBufferLevel == 0) {
			/* we need to read from the source */
			m_pBufferReadPos = &m_Buffer.front();
			m_iBufferLevel = 0;
			sint32 iRead = m_VDFile.readData(m_pBufferReadPos, m_Buffer.size());

			if(iRead <= 0)
				break;

			m_iBufferLevel = iRead;
		}

		sint32 iToCopy = iBytes;
		if(iToCopy > m_iBufferLevel)
			iToCopy = m_iBufferLevel;

		memcpy(pBuffer, m_pBufferReadPos, iToCopy);
		iBytesRead += iToCopy;
		m_iBufferLevel -= iToCopy;
		iBytes -= iToCopy;
		m_pBufferReadPos += iToCopy;
		pBuffer = (uint8 *)pBuffer + iToCopy;
	}

	return iBytesRead;
}

bool ModTextSource::HasBOM(void) {
	return m_bHasBOM;
}

eTextType ModTextSource::GetTextType(void) {
	return m_TextType;
}