//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "ModTextSourceSRT.h"

#include <vd2/system/Error.h>
#include <vd2/system/log.h>
#include <vd2/Dita/resources.h>

namespace {
	enum { kVDST_TextSourceSRT = 154 };
	
	enum {
		kVDM_FileErrors
	};
}


ModTextSourceSRT::ModTextSourceSRT(const wchar_t *szFileName) {
	m_LastSubtitleEndMS = 0;

	clearSourceInfo();
	m_SourceType = SOURCE_STD_FILE | SOURCE_TEXT_SRT;
	m_SourceName = szFileName;
	m_SourceDesc = L"SRT (text) file";
}

ModTextSourceSRT::~ModTextSourceSRT(void) {
	m_TextSource.Close();
}

VDStringW ModTextSourceSRT::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	if(m_FileErrors.length() > 0) {
		report += L"\r\nSome errors were found :";
		report += m_FileErrors;
	}
	return report;
}

bool ModTextSourceSRT::init(void) {
	if(m_SourceName.empty())
		return false;

	if(!allocFormat(sizeof(WAVEFORMATEX)))
		return false;

	// open file
	if(!m_TextSource.Open(m_SourceName))
		return false;

	m_TextType = m_TextSource.GetTextType();

	bool bWarnErrors = false;

	VDStringW line;
	VDStringW subNumber;
	int lineNumber = 0;
	int lineNumber_SubtitleDeclaration;
	VDStringW subtitle;
	sint32 startMS, durationMS;
	bool bOverlap;
	bool bValidTime = true;
	while(m_TextSource.getLine(line)) {
		lineNumber++;

		// Skip the void lines
		if(ModTextUtil::IsBlank(line))
			continue;

		// Current line should be the subtitle number, skip to get the start and end times
		lineNumber_SubtitleDeclaration = lineNumber;
		subNumber = line;
		ModTextUtil::Trim(subNumber);
		if(!m_TextSource.getLine(line)) {
			// oops, the file ends on a subtitle number ?
			m_FileErrors += VDStringW::setf(L"\r\nFile ends while declaring subtitle [%s] (line %d)."
				, subNumber.c_str(), lineNumber_SubtitleDeclaration);
			bWarnErrors = true;
			break;
		}
		lineNumber++;
		if(ModTextUtil::IsBlank(line)) {
			// oops, the subtitle declaration is b0rked ?
			m_FileErrors += VDStringW::setf(L"\r\nUnexpected blank line while declaring subtitle [%s] (line %d)."
				, subNumber.c_str(), lineNumber_SubtitleDeclaration);
			continue;
		}

		// Add the times
		ModTextUtil::Trim(line);
		bValidTime = getTimes(line, startMS, durationMS, bOverlap);
		if(bValidTime && bOverlap) {
			// oops, subtitle overlaps with previous one
			m_FileErrors += VDStringW::setf(L"\r\nSubtitle [%s] (line %d) collides with the previous one."
				, subNumber.c_str(), lineNumber_SubtitleDeclaration);
			bWarnErrors = true;
		}

		// Add the subs
		bool bGotLine = false;
		while((bGotLine=m_TextSource.getLine(line)) && !ModTextUtil::IsBlank(line)) {
			lineNumber++;
			if(subtitle.length() > 0)
				subtitle += VDStringW::setf(L"\r\n%s", line.c_str());
			else
				subtitle = line;
		}

		if(subtitle.length() > 0) {
			lineNumber++;
			if(!bValidTime || !m_SubtitleProvider.AddSubtitle(subtitle, startMS, durationMS)) {
				// oops, the time information is b0rked
				m_FileErrors += VDStringW::setf(L"\r\nStart and End times of subtitle [%s] (line %d) aren't correct. Subtitle skipped."
					, subNumber.c_str(), lineNumber_SubtitleDeclaration);
				bWarnErrors = true;
			}
			subtitle.clear();
		} else if(!bGotLine) {
			// oops, the file ends on a time format ?
			m_FileErrors += VDStringW::setf(L"\r\nFile ends while declaring subtitle [%s] (line %d)."
				, subNumber.c_str(), lineNumber_SubtitleDeclaration);
			bWarnErrors = true;
			break;
		} else {
			// oops, it is a blank subtitle
			m_FileErrors += VDStringW::setf(L"\r\nSubtitle [%s] (line %d) is blank. Subtitle skipped."
				, subNumber.c_str(), lineNumber_SubtitleDeclaration);
			bWarnErrors = true;
		}
	}

	if(bWarnErrors)
		VDLogAppMessage(kVDLogError, kVDST_TextSourceSRT, kVDM_FileErrors);

	if(m_SubtitleProvider.GetSourceSubtitles().size() <= 0)
		return false;

	//How to use a text stream within audio routines of VirtualDub :p
	getWaveFormat()->nChannels = 1;
	getWaveFormat()->wFormatTag = 0x0000;
	getWaveFormat()->nSamplesPerSec = 1000;
	getWaveFormat()->cbSize = 0;
	getWaveFormat()->nAvgBytesPerSec = 1024;
	getWaveFormat()->nBlockAlign = 1;
	getWaveFormat()->wBitsPerSample = 1;

	lSampleFirst	= 0;
	lSampleLast		= m_SubtitleProvider.GetSourceSubtitles().back().m_StartMS + m_SubtitleProvider.GetSourceSubtitles().back().m_DurationMS;

	streamInfo.fccType					= streamtypeTEXT;
	streamInfo.fccHandler				= 0;
	streamInfo.dwFlags					= 0;
	streamInfo.wPriority				= 0;
	streamInfo.wLanguage				= 0;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwScale					= 1;
	streamInfo.dwRate					= getWaveFormat()->nSamplesPerSec;
	streamInfo.dwStart					= 0;
	streamInfo.dwLength					= lSampleLast;
	streamInfo.dwSuggestedBufferSize	= 16*1024;
	streamInfo.dwQuality				= 0xffffffff;
	streamInfo.dwSampleSize				= getWaveFormat()->nBlockAlign;

	m_SourceDesc = VDStringW::setf(L"SRT (%s%s) file"
		, m_TextType==kTextType_ASCII ? L"ASCII"
		: m_TextType==kTextType_UTF8 ? L"UTF-8"
		: m_TextType==kTextType_UTF16LE ? L"UTF-16 Little Endian"
		: m_TextType==kTextType_UTF16BE ? L"UTF-16 Big Endian"
		: m_TextType==kTextType_UTF32LE ? L"UTF-32 Little Endian"
		: m_TextType==kTextType_UTF32BE ? L"UTF-32 Big Endian"
		: L"?"
		, (m_TextSource.HasBOM() || (m_TextType==kTextType_ASCII)) ? L"" : L" without BOM"
		);

	return true;
}

bool ModTextSourceSRT::getTimes(VDStringW line, sint32& startMS, sint32& durationMS, bool& bOverlap) {
	bOverlap = false;

	int h1, m1, s1, ms1;
	int h2, m2, s2, ms2;
	if(swscanf(line.c_str(), L"%02d:%02d:%02d,%03d --> %02d:%02d:%02d,%03d"
		,&h1, &m1, &s1, &ms1
		,&h2, &m2, &s2, &ms2) != 8)
		return false;
	sint32 subStartMS = ms1
		+ (1000 * s1)
		+ (1000 * 60 * m1)
		+ (1000 * 60 * 60 * h1);
	sint32 subEndMS = ms2
		+ (1000 * s2)
		+ (1000 * 60 * m2)
		+ (1000 * 60 * 60 * h2);

	startMS = subStartMS;
	durationMS = subEndMS - subStartMS;

	if(durationMS <= 0)
		return false;

	if(startMS < m_LastSubtitleEndMS)
		bOverlap = true;


	m_LastSubtitleEndMS = subEndMS;

	return true;
}

int ModTextSourceSRT::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *plBytes, LONG *plSamples) {
	VDStringW subtitle;
	sint32 durationMS = 0;

	if(!m_SubtitleProvider.GetSubtitle_Linear(subtitle, lStart, durationMS)) {
		if(plBytes)
			*plBytes = 0;
		if(plSamples)
			*plSamples = 0;

		return AVIERR_OK;
	}

	VDStringA subUTF8 = VDTextWToU8(subtitle);

	if(plBytes)
		*plBytes = subUTF8.length() + 1;
	if(plSamples)
		*plSamples = durationMS;

	if(lpBuffer && (cbBuffer < subUTF8.length() + 1) )
		return AVIERR_BUFFERTOOSMALL;

	if(!lpBuffer)
		return AVIERR_OK;

	memcpy(lpBuffer, subUTF8.c_str(), subUTF8.length() + 1);

	return AVIERR_OK;
}

bool ModTextSourceSRT::QueryInterface(InterfaceID ID, void** destination) {
	if(ID == ISubtitleHandler::GetInterfaceID()) {
		*destination = static_cast<ISubtitleHandler *>(this);
		return true;
	}
	return false;
}

bool ModTextSourceSRT::IsASCII(void) {
	return (m_TextType == kTextType_ASCII);
}

bool ModTextSourceSRT::HasBOM(void) {
	return m_TextSource.HasBOM();
}

eTextType ModTextSourceSRT::GetOutputType(void) {
	return kTextType_UTF8;
}

ModSubtitleProvider *ModTextSourceSRT::GetSubtitleProvider(void) {
	if(m_SubtitleProvider.HasCollisions())
		return &m_SubtitleProvider;
	return NULL;
}