//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "ModTextUtil.h"
#include "misc.h"


ModTextUtil::ModTextUtil(eTextType inputType, eTextType outputType, bool bWriteBOM, bool bFormatAsSubtiDS)
: m_TextType(inputType)
, m_OutputInType(outputType)
, m_bPrefixBOMInOutput(bWriteBOM)
, m_bFormatAsSubtiDS(bFormatAsSubtiDS)
{
}

ModTextUtil::~ModTextUtil(void) {
}

void ModTextUtil::SetInputType(eTextType type) {
	m_TextType = type;
}

void ModTextUtil::SetOutputType(eTextType type, bool bPrefixBOM, bool bAsSubtiDS) {
	m_OutputInType = type;
	m_bPrefixBOMInOutput = bPrefixBOM;
	m_bFormatAsSubtiDS = bAsSubtiDS;
}

VDStringW ModTextUtil::GetText(void *pBuffer, long iBufferLen) {
	// Remove any unwanted characters at the end of the text
	// Make sure the string ends with 0x00
	void *tmpBuffer = malloc(iBufferLen + 2*sizeof(wchar_t));
	if(!tmpBuffer)
		throw MyMemoryError();
	memset((char *)tmpBuffer + iBufferLen, 0, 2*sizeof(wchar_t));
	//((char *)tmpBuffer)[iBufferLen] = 0;
	memcpy(tmpBuffer, pBuffer, iBufferLen);
	VDStringW subtitle;
	switch(m_TextType) {
	case kTextType_UTF16LE:
		subtitle = (wchar_t *)tmpBuffer;
		break;
	case kTextType_UTF8:
		subtitle = VDTextU8ToW((const char *)tmpBuffer, -1);
		break;
	case kTextType_ASCII:
		subtitle = VDTextAToW((const char *)tmpBuffer, -1);
		break;
	}
	free(tmpBuffer);

	Shorten(subtitle);

	return subtitle;
}

bool ModTextUtil::IsBlank(VDStringW& text) {
	if(!text.length())
		return true;
	wchar_t *pos = (wchar_t *)text.c_str();
	wchar_t *end = pos + text.length();
	while(pos < end) {
		switch(*pos++) {
		case 0:
			return true;
		case L' ':
		case L'\r':
		case L'\n':
			break;
		default:
			return false;
		}
	}
	return true;
}

void ModTextUtil::Trim(VDStringW& text) {
	wchar_t *subtitleStart = (wchar_t *)text.c_str();
	wchar_t *pos = subtitleStart + text.length();
	while(pos >= subtitleStart) {
		switch(*pos) {
		case L' ':
		case L'\r':
		case L'\n':
			*pos = 0;
			break;
		}
		if(*pos)
			break;
		pos--;
	}

	// void text
	if(pos < subtitleStart) {
		if(text.length())
			text.resize(0);
		return;
	}

	wchar_t *p_src = subtitleStart;
	wchar_t *p_dst = subtitleStart;
	while(p_src < pos) {
		switch(*p_src) {
		case L' ':
		case L'\r':
		case L'\n':
			*p_src = 0;
			break;
		}
		if(*p_src)
			break;
		p_src++;
	}
	if(p_src != p_dst) {
		while(p_src <= pos)
			*p_dst++ = *p_src++;
	} else
		p_dst = pos + 1;

	if(p_dst - subtitleStart != text.length())
		text.resize(p_dst - subtitleStart);
}

void ModTextUtil::Shorten(VDStringW& text) {
	wchar_t *subtitleStart = (wchar_t *)text.c_str();
	wchar_t *pos = subtitleStart + text.length();

	while(pos >= subtitleStart) {
		switch(*pos) {
		case L'\n':
		case L'\r':
			*pos = 0;
			break;
		}
		if(*pos)
			break;
		pos--;
	}

	// void text
	if(pos < subtitleStart) {
		if(text.length())
			text.resize(0);
		return;
	}

	pos++;
	if(pos - subtitleStart != text.length())
		text.resize(pos - subtitleStart);
}

int ModTextUtil::SetText(VDStringW text, void *pBuffer, long iBufferLen, long *iBytesWritten) {
	void *dataToWrite = NULL;
	int dataLength = 0;
	void *BOMToWrite = NULL;
	uint8 BOMLength = 0;

	VDStringW srcString;
	
	Shorten(text);

	// There are some subtitle engines that seem to expect the exact
	// SubtiDS format ("...\r" and "...\r\n...\r") and have the bad habit
	// to just eat the last char when you give them "..." and "...\r\n..."
	if(m_bFormatAsSubtiDS)
		text += L'\r';

	if(m_bPrefixBOMInOutput)
		GetBOMInformation(m_OutputInType, BOMToWrite, BOMLength);

	VDStringA convString;
	switch(m_OutputInType) {
	case kTextType_UTF16LE:
		dataToWrite = (void *)srcString.c_str();
		dataLength = sizeof(wchar_t) * (srcString.length() + 1);
		break;

	case kTextType_UTF8:
		convString = VDTextWToU8(srcString);
		dataToWrite = (void *)convString.c_str();
		dataLength = convString.length() + 1;
		break;

	default:
		convString = VDTextWToA(srcString);
		dataToWrite = (void *)convString.c_str();
		dataLength = convString.length() + 1;
		break;
	}

	if(iBytesWritten)
		*iBytesWritten = dataLength + BOMLength;

	if(!pBuffer)
		return AVIERR_OK;
	if(iBufferLen < dataLength + BOMLength)
		return AVIERR_BUFFERTOOSMALL;

	if(BOMToWrite && (BOMLength>0)) {
		memcpy(pBuffer, BOMToWrite, BOMLength);
		pBuffer = (uint8 *)pBuffer + BOMLength;
	}
	if(dataToWrite && (dataLength>0))
		memcpy(pBuffer, dataToWrite, dataLength);

	return AVIERR_OK;
}

//////////////////////////////////////////////////////////////////////

ModSubtitleProvider::ModSubtitleProvider(void)
: m_bIteratorReady(false)
, m_bCollisions(false)
{
}

ModSubtitleProvider::~ModSubtitleProvider(void) {
	Clear();
}

std::list<SubtitleFrame>& ModSubtitleProvider::GetSourceSubtitles(void) {
	return m_SourceSubtitles;
}

std::list<SubtitleFrame>& ModSubtitleProvider::GetAdjustedSubtitles(void) {
	return m_AdjustedSubtitles;
}

bool ModSubtitleProvider::HasCollisions(void) {
	return m_bCollisions;
}

bool ModSubtitleProvider::AddSubtitle(VDStringW subtitle, sint32 startMS, sint32 durationMS) {
	if( (startMS < 0) || (durationMS < 0) )
		return false;
	if(ModTextUtil::IsBlank(subtitle))
		return false;

	SubtitleFrame *lastFrame = NULL;
	if(m_SourceSubtitles.size())
		lastFrame = &(m_SourceSubtitles.back());
	if(lastFrame && (startMS < lastFrame->m_StartMS))
		return false;

	SubtitleFrame newFrame;

	if(lastFrame && (startMS < lastFrame->m_StartMS + lastFrame->m_DurationMS))
		m_bCollisions = true;

	ModTextUtil::Shorten(subtitle);
	newFrame.m_Subtitle = subtitle;
	newFrame.m_StartMS = startMS;
	newFrame.m_DurationMS = durationMS;

	m_SourceSubtitles.push_back(newFrame);
	m_AdjustedSubtitles.push_back(newFrame);

	return true;
}

bool ModSubtitleProvider::GetSubtitle_Linear(VDStringW& subtitle, sint32 startMS, sint32& durationMS) {
	if(!m_bIteratorReady) {
		m_SubtitleIterator = m_AdjustedSubtitles.begin();
		m_bIteratorReady = true;
	}
	if(m_SubtitleIterator == m_AdjustedSubtitles.end()) {
		subtitle.clear();
		durationMS = 0;
		return false;
	}

	while((m_SubtitleIterator!=m_AdjustedSubtitles.begin()) && ((*m_SubtitleIterator).m_StartMS>startMS))
		m_SubtitleIterator--;

	while((*m_SubtitleIterator).m_StartMS + (*m_SubtitleIterator).m_DurationMS <= startMS) {
		m_SubtitleIterator++;
		if(m_SubtitleIterator == m_AdjustedSubtitles.end()) {
			subtitle.clear();
			durationMS = 0;
			return false;
		}
	}

	SubtitleFrame *subFrame = &(*m_SubtitleIterator);

	if( (subFrame->m_StartMS<=startMS) && (subFrame->m_StartMS+subFrame->m_DurationMS>startMS)) {
		// Got the right frame
		subtitle = subFrame->m_Subtitle;
		durationMS = subFrame->m_StartMS + subFrame->m_DurationMS - startMS;
		return true;
	} else {
		// No frame where we want, use a blank fake frame
		subtitle.clear();
		durationMS = subFrame->m_StartMS - startMS;
		return true;
	}
}

void ModSubtitleProvider::ResetIterator(void) {
	m_bIteratorReady = false;
}

bool ModSubtitleProvider::GetNextSubtitle(VDStringW& subtitle, sint32& startMS, sint32& durationMS) {
	if(!m_bIteratorReady) {
		m_SubtitleIterator = m_AdjustedSubtitles.begin();
		m_bIteratorReady = true;
	} else if(m_SubtitleIterator != m_AdjustedSubtitles.end())
		m_SubtitleIterator++;
	if(m_SubtitleIterator == m_AdjustedSubtitles.end()) {
		subtitle.clear();
		startMS = 0;
		durationMS = 0;
		return false;
	}

	SubtitleFrame *subFrame = &(*m_SubtitleIterator);

	subtitle = subFrame->m_Subtitle;
	startMS = subFrame->m_StartMS;
	durationMS = subFrame->m_DurationMS;
	return true;
}

bool ModSubtitleProvider::IsEnd(void) {
	return ((m_AdjustedSubtitles.size()==0) || (m_SubtitleIterator==m_AdjustedSubtitles.end()));
}

ModSubtitleProvider& ModSubtitleProvider::operator =(ModSubtitleProvider& subProvider) {
	Clear();
	ResetIterator();

	m_bCollisions = subProvider.m_bCollisions;
	m_SourceSubtitles = subProvider.m_SourceSubtitles;
	m_AdjustedSubtitles = subProvider.m_AdjustedSubtitles;

	return *this;
}

void ModSubtitleProvider::Clear(void) {
	m_SourceSubtitles.clear();
	m_AdjustedSubtitles.clear();
	m_bCollisions = false;
}

bool AdjustSubtitle(SubtitleFrame& sub, sint32 iTimeStartMS, sint32 iTimeEndMS, sint64 iOffsetMS) {
	sint32 iTimeSubStartMS = sub.m_StartMS + iOffsetMS;
	sint32 iSubDurationMS = sub.m_DurationMS;
	if((iTimeSubStartMS<iTimeStartMS) && (iTimeSubStartMS+iSubDurationMS>iTimeStartMS)) {
		iSubDurationMS -= (iTimeStartMS-iTimeSubStartMS);
		iTimeSubStartMS = iTimeStartMS;
	}
	if((iTimeSubStartMS<iTimeEndMS) && (iTimeSubStartMS+iSubDurationMS>iTimeEndMS))
		iSubDurationMS = iTimeEndMS - iTimeSubStartMS;
	if((iTimeSubStartMS>=iTimeStartMS) && (iTimeSubStartMS<iTimeEndMS)) {
		sub.m_StartMS = iTimeSubStartMS - iTimeStartMS;
		sub.m_DurationMS = iSubDurationMS;
		return true;
	}
	return false;
}

bool AdjustSubtitle(SubtitleFrame& sub, sint32 iTimeStartMS, sint32 iTimeEndMS, FrameSubset::reference subset, sint64 iOffsetSubsetFrames, VDFraction& frameRate) {
	sint64 iTimeSubsetStartMS = frameRate.scale64ir(VD64(1000) * subset.start);
	sint64 iTimeSubsetEndMS = frameRate.scale64ir(VD64(1000) * (subset.start + subset.len));
	sint64 iOffsetSubsetMS = frameRate.scale64ir(VD64(1000) * iOffsetSubsetFrames);

	bool bResult = AdjustSubtitle(sub, iTimeSubsetStartMS, iTimeSubsetEndMS, 0);
	if(bResult)
		bResult = AdjustSubtitle(sub, iTimeStartMS, iTimeEndMS, iOffsetSubsetMS);
	return bResult;
}

void ModSubtitleProvider::ApplySubsets(long iTimeStartMS, long iTimeEndMS, FrameSubset *pSubset, VDFraction frameRate) {
	std::list<SubtitleFrame>::iterator subIt;
	std::list<SubtitleFrame>::reverse_iterator subIt_2;
	m_AdjustedSubtitles.clear();

	if(pSubset) {
		sint64 iTotalFrames = 0;
		FrameSubset::iterator subsetIt;
		for(subsetIt=pSubset->begin() ; subsetIt!=pSubset->end() ; subsetIt++) {
			for(subIt=m_SourceSubtitles.begin() ; subIt!=m_SourceSubtitles.end() ; subIt++) {
				SubtitleFrame sub = (*subIt);

				// Let the fun begin :D
				// Verify the subtitle can fit in the current subset (and thus adjust its timing)
				if(AdjustSubtitle(sub, iTimeStartMS, iTimeEndMS, *subsetIt, iTotalFrames, frameRate)) {
					bool bPreviouslyStarted = false;
					// Iterate (reverse order to go faster) in the previously added subtitles
					for(subIt_2=m_AdjustedSubtitles.rbegin() ; subIt_2!=m_AdjustedSubtitles.rend() ; subIt_2++) {
						SubtitleFrame& otherSub = (*subIt_2);
						// If we already have this subtitle, and that timings correspond
						//, update the duration of the already added subtitle
						if((otherSub.m_Subtitle==sub.m_Subtitle) && (otherSub.m_StartMS<sub.m_StartMS) && (otherSub.m_StartMS+otherSub.m_DurationMS>=sub.m_StartMS)) {

							VDASSERT(otherSub.m_StartMS + otherSub.m_DurationMS == sub.m_StartMS);

							bPreviouslyStarted = true;
							if(otherSub.m_StartMS + otherSub.m_DurationMS < sub.m_StartMS + sub.m_DurationMS) {
								if(sub.m_StartMS < otherSub.m_StartMS + otherSub.m_DurationMS) {
									sub.m_DurationMS -= (otherSub.m_StartMS+otherSub.m_DurationMS-sub.m_StartMS);
									sub.m_StartMS = otherSub.m_StartMS + otherSub.m_DurationMS;
								}
								otherSub.m_DurationMS += sub.m_DurationMS;
								break;
							}
						}
					}
					// If this is a brand new subtitle (or if there is a gap after the previous one)
					//, add it to the list
					if(!bPreviouslyStarted)
						m_AdjustedSubtitles.push_back(sub);
				}
			}
			iTotalFrames += (*subsetIt).len;
		}
	} else {
		for(subIt=m_SourceSubtitles.begin() ; subIt!=m_SourceSubtitles.end() ; subIt++) {
			SubtitleFrame sub = (*subIt);

			sint32 iTimeSubStartMS = sub.m_StartMS;
			sint32 iSubDurationMS = sub.m_DurationMS;
			if((iTimeSubStartMS<iTimeStartMS) && (iTimeSubStartMS+iSubDurationMS>iTimeStartMS)) {
				iSubDurationMS -= (iTimeStartMS-iTimeSubStartMS);
				iTimeSubStartMS = iTimeStartMS;
			}
			if((iTimeSubStartMS<iTimeEndMS) && (iTimeSubStartMS+iSubDurationMS>iTimeEndMS))
				iSubDurationMS = iTimeEndMS - iTimeSubStartMS;
			if((iTimeSubStartMS>=iTimeStartMS) && (iTimeSubStartMS<iTimeEndMS)) {
				sub.m_StartMS = iTimeSubStartMS;
				sub.m_DurationMS = iSubDurationMS;
				m_AdjustedSubtitles.push_back(sub);
			}
		}
	}
}