//  VirtualDub - Video processing and capture application
//  Copyright (C) 1998-2001 Avery Lee
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//  This version has been modified by fccHandler, December 2003

#include "stdafx.h"
#include <process.h>
#include <windows.h>
#include <commctrl.h>
#include <fcntl.h>
#include <io.h>

#include "AudioSource.h"
#include "VideoSource.h"
#include "FastReadStream.h"
#include <vd2/system/error.h>
#include <vd2/system/fraction.h>
#include <vd2/system/log.h>
#include <vd2/system/file.h>
#include <vd2/Dita/resources.h>

#include "misc.h"
#include "mpeg.h"
#include "mpeg_decode.h"
#include "resource.h"
#include "gui.h"
#include <vd2/system/cpuaccel.h>
#include <vd2/Priss/decoder.h>

//////////////////////////////////////////////////////////////////////////

extern HINSTANCE g_hInst;

//////////////////////////////////////////////////////////////////////////

namespace {
	enum { kVDST_Mpeg = 5 };

	enum {
		kVDM_AudioConcealingError,		// MPEGAudio: Concealing decoding error on frame %lu: %hs.
		kVDM_OpeningFile,				// MPEG: Opening file "%hs"
		kVDM_OOOTimestamp,				// MPEG: Anachronistic or discontinuous timestamp found in %ls stream %d at byte position %lld, from %lld to %lld. This may indicate an improper join.
		kVDM_Incomplete,				// MPEG: File ended unexpectedly during parsing -- file may be damaged or incomplete.
		kVDM_UnknownFramerate			// MPEG: Unknown frame rate code (%d)
	};
}

//////////////////////////////////////////////////////////////////////////

//#define ALLOW_AUDIO_DSC
//#define USE_FIELD_MAP		// WORK IN PROGRESS

#define VIDPKT_TYPE_SEQUENCE_START      0xB3
#define VIDPKT_TYPE_SEQUENCE_END        0x1B7
#define VIDPKT_TYPE_GROUP_START         0xB8    // not used
#define VIDPKT_TYPE_PICTURE_START       0x00
#define VIDPKT_TYPE_SLICE_START_MIN     0x01
#define VIDPKT_TYPE_SLICE_START_MAX     0xAF
#define VIDPKT_TYPE_EXT_START           0xB5
#define VIDPKT_TYPE_USER_START          0xB2

//////////////////////////////////////////////////////////////////////////
//
//
//                          DataVector
//
//
//////////////////////////////////////////////////////////////////////////

class DataVectorBlock {
public:
    enum { BLOCK_SIZE = 4096 };

    DataVectorBlock *next;

    char heap[BLOCK_SIZE];
    int index;

    DataVectorBlock() {
        next = NULL;
        index = 0;
    }
};

class DataVector {
private:
    DataVectorBlock *first, *last;
    int item_size;
    int count;

    void _Add(void *);

public:
    DataVector(int item_size);
    ~DataVector();

    void Add(void *pp) {
        if (!last || last->index >= DataVectorBlock::BLOCK_SIZE - item_size) {
            _Add(pp);
            return;
        }

        memcpy(last->heap + last->index, pp, item_size);
        last->index += item_size;
        ++count;
    }

    void *MakeArray();

    int Length() { return count; }
};

DataVector::DataVector(int _item_size) : item_size(_item_size) {
    first = last = NULL;
    count = 0;
}

DataVector::~DataVector() {
    DataVectorBlock *i, *j;

    j = first;
    while(i=j) {
        j = i->next;
        delete i;
    }
}

void DataVector::_Add(void *pp) {
    if (!last || last->index > DataVectorBlock::BLOCK_SIZE - item_size) {
        DataVectorBlock *ib = new DataVectorBlock();

        if (!ib) throw MyMemoryError();

        if (last)       last->next = ib;
        else            first = ib;

        last = ib;
    }

    memcpy(last->heap + last->index, pp, item_size);
    last->index += item_size;
    ++count;
}

void *DataVector::MakeArray() {
    char *array = new char[count * item_size], *ptr = array;
    DataVectorBlock *dvb = first;

    if (!array) throw MyMemoryError();

    while(dvb) {
        memcpy(ptr, dvb->heap, dvb->index);
        ptr += dvb->index;

        dvb=dvb->next;
    }

    return array;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                  InputFileMPEGOptions
//
//
//////////////////////////////////////////////////////////////////////////

class InputFileMPEGOptions : public InputFileOptions {
public:
    enum {
        DECODE_NO_B = 1,
        DECODE_NO_P = 2,
    };

    struct InputFileMPEGOpts {
        int len;
        int iDecodeMode;
        bool fAcceptPartial;
#ifdef ALLOW_AUDIO_DSC
        bool fAllowDSC;
#endif
    } opts;


    ~InputFileMPEGOptions();

    bool read(const char *buf);
    int write(char *buf, int buflen);

    static BOOL APIENTRY SetupDlgProc( HWND hDlg, UINT message, UINT wParam, LONG lParam);
};

InputFileMPEGOptions::~InputFileMPEGOptions() {
}

bool InputFileMPEGOptions::read(const char *buf) {
    const InputFileMPEGOpts *pp = (const InputFileMPEGOpts *)buf;

    if (pp->len != sizeof(InputFileMPEGOpts))
        return false;

    opts = *pp;

    return true;
}

int InputFileMPEGOptions::write(char *buf, int buflen) {
    InputFileMPEGOpts *pp = (InputFileMPEGOpts *)buf;

    if (buflen<sizeof(InputFileMPEGOpts))
        return 0;

    opts.len = sizeof(InputFileMPEGOpts);
    *pp = opts;

    return sizeof(InputFileMPEGOpts);
}

///////

BOOL APIENTRY InputFileMPEGOptions::SetupDlgProc( HWND hDlg, UINT message, UINT wParam, LONG lParam) {
    InputFileMPEGOptions *thisPtr = (InputFileMPEGOptions *)GetWindowLong(hDlg, DWL_USER);

    switch(message) {
        case WM_INITDIALOG:
            SetWindowLong(hDlg, DWL_USER, lParam);
            CheckDlgButton(hDlg, IDC_MPEG_ALL_FRAMES, TRUE);
			//CheckDlgButton(hDlg, IDC_MPEG_ACCEPTPARTIAL, TRUE);
            return TRUE;

        case WM_COMMAND:
            if (LOWORD(wParam) == IDCANCEL) {
                if (IsDlgButtonChecked(hDlg, IDC_MPEG_I_FRAMES_ONLY ))
                    thisPtr->opts.iDecodeMode = InputFileMPEGOptions::DECODE_NO_B | InputFileMPEGOptions::DECODE_NO_P;

                if (IsDlgButtonChecked(hDlg, IDC_MPEG_IP_FRAMES_ONLY))
                    thisPtr->opts.iDecodeMode = InputFileMPEGOptions::DECODE_NO_B;

                if (IsDlgButtonChecked(hDlg, IDC_MPEG_ALL_FRAMES    ))
                    thisPtr->opts.iDecodeMode = 0;

				//thisPtr->opts.fAcceptPartial = !!IsDlgButtonChecked(hDlg, IDC_MPEG_ACCEPTPARTIAL);
				thisPtr->opts.fAcceptPartial = true;
#ifdef ALLOW_AUDIO_DSC
				thisPtr->opts.fAllowDSC = !!IsDlgButtonChecked(hDlg, IDC_MPEG_ALLOW_DSC);
#endif

                EndDialog(hDlg, 0);
                return TRUE;
            }
            break;
    }

    return FALSE;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                  MPEGPacketInfo/MPEGSampleInfo
//
//
//////////////////////////////////////////////////////////////////////////


struct MPEGPacketInfo {
    __int64     file_pos;
    __int64     stream_pos;
	__int64		PTS;
};

// flags for MPEGSampleInfo.bitflags
#define PICTUREFLAG_RFF         1       // repeat_first_field
#define PICTUREFLAG_TFF         2       // top_field_first
#define PICTUREFLAG_PF          4       // progressive_frame
#define PICTUREFLAG_ATTACH      8       // sequence_header attached

struct MPEGSampleInfo {
    __int64     stream_pos;
	__int64		PTS;
    int         size;
    union {
        struct {
            char    frame_type;         // I,P,B
            char    picture_structure;  // FRAME/FIELD
            char    bitflags;           // MPEG-2 rff/tff, etc.
            char    fields;             // # of fields
        };

        struct {
            long    header;
        };
    };
};

//////////////////////////////////////////////////////////////////////////
//
//
//                  InputFileMPEG
//
//
//////////////////////////////////////////////////////////////////////////

#define ATYPE_NONE 0
#define ATYPE_MPEG 1
#define ATYPE_AC3  2
#define ATYPE_LPCM 3

class AudioSourceMPEG;
class VideoSourceMPEG;
class AudioSourceAC3;
class AudioSourcePCM;
class MPEGAudioParser;

class InputFileMPEG : public InputFile {
friend VideoSourceMPEG;
friend AudioSourceMPEG;
friend AudioSourceAC3;
friend AudioSourcePCM;
private:
    char *video_packet_buffer;
    MPEGPacketInfo *video_packet_list;
    MPEGSampleInfo *video_sample_list;
    MPEGPacketInfo *audio_packet_list;
    MPEGSampleInfo *audio_sample_list;
#ifdef USE_FIELD_MAP
	long *video_field_map;
#endif
    int packets, apackets;
    int frames, aframes;
	__int64 firstPTS, afirstPTS;
    int last_packet[2];
    int width, height;
    int aspect_ratio;       // aspect_ratio_information
	int chroma_format;
    VDFraction mFrameRate;
    bool fInterleaved, fIsVCD;
    bool progressive_sequence;
    int AudioType;
    bool fIsMPEG2;

    long audio_first_header;

    int iDecodeMode;
    bool fAcceptPartial;
    bool fAllowDSC;

    FastReadStream *pFastRead;
    VDFile  mFile;
//    int fh;

    static const char szME[];

    enum {
        RESERVED_STREAM     = 0xbc,
        PRIVATE_STREAM1     = 0xbd,
        PADDING_STREAM      = 0xbe,
        PRIVATE_STREAM2     = 0xbf,
    };

    void    ReadStream(void *buffer, __int64 pos, long len, bool fAudio);

    bool PostInit();

    void setOptions(InputFileOptions *);
    InputFileOptions *createOptions(const char *buf);
    InputFileOptions *promptForOptions(HWND);
public:
    InputFileMPEG();
    ~InputFileMPEG();

    void Init(const wchar_t *szFile);
    static void _InfoDlgThread(void *pvInfo);
    static BOOL APIENTRY _InfoDlgProc( HWND hDlg, UINT message, UINT wParam, LONG lParam);
    void InfoDialog(HWND hwndParent);
};


//////////////////////////////////////////////////////////////////////////
//
//
//                      VideoSourceMPEG
//
//
//////////////////////////////////////////////////////////////////////////

// Chroma formats
#define CHROMA420       1
#define CHROMA422       2
#define CHROMA444       3

class VideoSourceMPEG : public VideoSource {
private:
    InputFileMPEG *parentPtr;

    long frame_forw, frame_back, frame_bidir;
	int chroma_format;

    bool fFBValid;

    long prev_IP(long f);
    long next_IP(long f);

public:
    VideoSourceMPEG(InputFileMPEG *);
    ~VideoSourceMPEG();

    bool init();
    char getFrameTypeChar(long lFrameNum);
    bool _isKey(LONG lSample);
    virtual LONG nearestKey(LONG lSample);
    virtual LONG prevKey(LONG lSample);
    virtual LONG nextKey(LONG lSample);
    bool setDecompressedFormat(int depth);
    bool setDecompressedFormat(BITMAPINFOHEADER *pbih);
    void invalidateFrameBuffer();
    bool isFrameBufferValid();
    void streamBegin(bool fRealTime);
    void streamSetDesiredFrame(long frame_num);
    long streamGetNextRequiredFrame(BOOL *is_preroll);
    int streamGetRequiredCount(long *);
    void *streamGetFrame(void *inputBuffer, long data_len, BOOL is_key, BOOL is_preroll, long frame_num);
    void *getFrame(LONG frameNum);
    eDropType getDropType(long);
    int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);

    long streamToDisplayOrder(long sample_num) {
        return sample_num;
    }

    long displayToStreamOrder(long display_num) {
        return display_num;
    }

    bool isDecodable(long sample_num);
};

VideoSourceMPEG::VideoSourceMPEG(InputFileMPEG *parent) {
    parentPtr = parent;

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	m_SourceType = SOURCE_MPEG | SOURCE_VIDEO;
	m_SourceName = L"Input MPEG file";
	m_SourceDesc = L"MPEG video stream";
// END ****************************************************************
}

bool VideoSourceMPEG::init() {
    BITMAPINFOHEADER *bmih;
    int w, h;

    fFBValid = false;

    lSampleFirst = 0;
    lSampleLast = parentPtr->frames;
    if (lSampleLast <= lSampleFirst) return false;

    // Strictly enforce the following rules:
    //   prev_IP(x) and next_IP(x) cannot fail if x is a B-frame
    //   prev_IP(x) cannot fail if x is a P-frame
    // This means that the first frame of the reordered sequence
    // must be an I-frame, and the last frame can't be a B-frame:

    if (parentPtr->video_sample_list[lSampleFirst]
        .frame_type != MPEG_FRAME_TYPE_I) return false;

    if (parentPtr->video_sample_list[lSampleLast-1]
        .frame_type == MPEG_FRAME_TYPE_B) return false;

	// Check width and height, chroma format
	w = parentPtr->chroma_format;
	if (w == CHROMA420 || w == CHROMA422) {
		if (parentPtr->height & 1) return false;
		if (w == CHROMA420) {
			if (parentPtr->width & 1) return false;
		}
	} else if (w != CHROMA444) {
		return false;
	}
	this->chroma_format = w;

    w = (parentPtr->width+15) & -16;
    h = (parentPtr->height+1) & -2;

	if (w<=0 || h<=0) return false;

    if (!AllocFrameBuffer(w * h * 4 + 4))
        throw MyMemoryError();

    if (!(bmih = (BITMAPINFOHEADER *)allocFormat(sizeof(BITMAPINFOHEADER))))
        throw MyMemoryError();

    if (!(bmihDecompressedFormat = (BITMAPINFOHEADER *)allocmem(getFormatLen()))) throw MyMemoryError();

    bmih->biSize        = sizeof(BITMAPINFOHEADER);
    bmih->biWidth       = w;
    bmih->biHeight      = h;
    bmih->biPlanes      = 1;
    bmih->biBitCount    = 32;
    bmih->biCompression = 0xffffffff;
    bmih->biSizeImage   = 0;
    bmih->biXPelsPerMeter   = 0;
    bmih->biYPelsPerMeter   = 0;
    bmih->biClrUsed     = 0;
    bmih->biClrImportant    = 0;

    streamInfo.fccType                  = streamtypeVIDEO;
    streamInfo.fccHandler               = 0;
    streamInfo.dwFlags                  = 0;
    streamInfo.wPriority                = 0;
    streamInfo.wLanguage                = 0;
    streamInfo.dwScale                  = parentPtr->mFrameRate.getLo();
    streamInfo.dwRate                   = parentPtr->mFrameRate.getHi();
    streamInfo.dwStart                  = 0;
    streamInfo.dwLength                 = parentPtr->frames;
    streamInfo.dwInitialFrames          = 0;
    streamInfo.dwSuggestedBufferSize    = 0;
    streamInfo.dwQuality                = 0xffffffffL;
    streamInfo.dwSampleSize             = 0;
    streamInfo.rcFrame.left             = 0;
    streamInfo.rcFrame.top              = 0;
    streamInfo.rcFrame.right            = w;
    streamInfo.rcFrame.bottom           = h;

    return true;
}

VideoSourceMPEG::~VideoSourceMPEG() {
}

bool VideoSourceMPEG::setDecompressedFormat(int depth) {
    if (depth != 32 && depth != 24 && depth != 16) return FALSE;

//#pragma warning("don't ship with this!!!!!!");
//  if (depth != 32 && depth != 16) return FALSE;

    return VideoSource::setDecompressedFormat(depth);
}

bool VideoSourceMPEG::setDecompressedFormat(BITMAPINFOHEADER *pbih) {
    if (pbih->biCompression == BI_RGB)
        return setDecompressedFormat(pbih->biBitCount);

    // Sanity-check format.

    BITMAPINFOHEADER *pbihInput = getImageFormat();

    if (pbih->biWidth != pbihInput->biWidth
        || pbih->biHeight != pbihInput->biHeight
        || pbih->biPlanes != 1)

        return false;

    do {
		// Accept UYVY, YUYV, or YUY2 at 16-bits per pel.
		// Accept YV12 only if chroma_format is 4:2:0

#if 1
		if (pbih->biBitCount == 12 && chroma_format == CHROMA420) {
			if (isEqualFOURCC(pbih->biCompression, '21VY')) break;
		} else if (pbih->biBitCount == 16 && chroma_format != CHROMA444) {
			if (isEqualFOURCC(pbih->biCompression, 'YVYU')) break;
			if (isEqualFOURCC(pbih->biCompression, 'VYUY')) break;
			if (isEqualFOURCC(pbih->biCompression, '2YUY')) break;
		}
#else
        // Allow YUY2 only (for testing)
        if ((pbih->biBitCount == 16) &&
          isEqualFOURCC(pbih->biCompression, '2YUY'))
            break;
#endif

        // Damn.
        return false;

    } while(0);

    // Looks good!

    memcpy(bmihDecompressedFormat, pbih, sizeof(BITMAPINFOHEADER));
    invalidateFrameBuffer();

    return true;
}

void VideoSourceMPEG::invalidateFrameBuffer() {
    fFBValid = false;
}

bool VideoSourceMPEG::isFrameBufferValid() {
    return fFBValid;
}

char VideoSourceMPEG::getFrameTypeChar(long lFrameNum) {
    if (lFrameNum >= lSampleFirst && lFrameNum < lSampleLast) {
        int i = parentPtr->video_sample_list[lFrameNum].frame_type;
        if (i == MPEG_FRAME_TYPE_I) return 'I';
        if (i == MPEG_FRAME_TYPE_P) return 'P';
        if (i == MPEG_FRAME_TYPE_B) return 'B';
    }
    return ' ';
}

VideoSource::eDropType VideoSourceMPEG::getDropType(long lFrameNum) {
    if (lFrameNum >= lSampleFirst && lFrameNum < lSampleLast) {
        int i = parentPtr->video_sample_list[lFrameNum].frame_type;
        if (i == MPEG_FRAME_TYPE_I) return kIndependent;
        if (i == MPEG_FRAME_TYPE_P) return kDependant;
    }
    return kDroppable;
}

bool VideoSourceMPEG::isDecodable(long sample_num) {
    if (sample_num >= lSampleFirst && sample_num < lSampleLast) {
        switch (parentPtr->video_sample_list[sample_num].frame_type) {
            case MPEG_FRAME_TYPE_B:
                if (mpeg_lookup_frame(next_IP(sample_num)) < 0) return false;
            case MPEG_FRAME_TYPE_P:
                if (mpeg_lookup_frame(prev_IP(sample_num)) < 0) return false;
        }
        return true;
    }
    return false;
}

bool VideoSourceMPEG::_isKey(LONG lSample) {
    if (lSample >= lSampleFirst && lSample < lSampleLast) {
        if (parentPtr->video_sample_list[lSample]
          .frame_type == MPEG_FRAME_TYPE_I) return true;
    }
    return false;
}

/*
LONG VideoSourceMPEG::nearestKey(LONG lSample) {
    LONG f;
    if (_isKey(lSample)) {
        f = lSample;
    } else {
        f = nextKey(lSample);
        if (f < 0) {
            f = prevKey(lSample);
        } else {
            LONG b = prevKey(lSample);
            if (b >= 0)
                if ((lSample - b) <= (f - lSample)) f = b;
        }
    }
    return f;
}
*/

LONG VideoSourceMPEG::nearestKey(LONG lSample) {
    if (_isKey(lSample)) return lSample;
	return prevKey(lSample);
}

LONG VideoSourceMPEG::prevKey(LONG lSample) {
	if (lSample < lSampleLast) {
		while (--lSample >= lSampleFirst) {
			if (parentPtr->video_sample_list[lSample].frame_type
			  == MPEG_FRAME_TYPE_I) return lSample;
		}
	}
    return -1;
}

LONG VideoSourceMPEG::nextKey(LONG lSample) {
	if (lSample >= lSampleFirst) {
		while (++lSample < lSampleLast) {
			if (parentPtr->video_sample_list[lSample].frame_type
			  == MPEG_FRAME_TYPE_I) return lSample;
		}
	}
    return -1;
}

void VideoSourceMPEG::streamBegin(bool fRealTime) {

//	_RPT0(0, "MPEG::streamBegin\n");

	frame_forw = frame_back = frame_bidir = -1;
	mpeg_reset();
}

// For field-based video in which the top field and the bottom
// field come from different MPEG frames, consider these rules:
//
// There are 9 possible combinations for a 2-field frame:
//		II IP IB, PI PP PB, BI BP BB
//
// II:	No searching is needed.
//
// IP:	If the P-frame is after the I-frame in display order, then
//		no searching is needed.  Otherwise we search backward to
//		find the forward reference for the P-frame.
//
// IB:	If the B-frame is after the I-frame in display order, then
//		the I-frame is its forward reference, and we search forward
//		to find the backward reference for the B-frame.  Otherwise
//		the I-frame is the backward reference, so we search backward
//		to find the forward reference for the B-frame.
//
// PI:	Same rule as for IP.
//
// PP:	We search backward to find the forward reference for the
//		P-frame which is earlier in display order.
//
// PB:	If the B-frame is before the P-frame in display order,
//		then the P-frame is its backward reference, and we only need
//		to search backward to find the forward reference for the
//		B-frame.  Otherwise the the P-frame is the forward reference,
//		so we must search backward to find the forward reference for
//		the P-frame, and forward to find the backward reference for
//		the B-frame (ugh).
//
// BI:	Same rule as for IB.
//
// BP:	Same rule as for PB.
//
// BB:	We search backward to find the forward reference for both
//		B-frames, and forward to find the backward reference for
//		both B-frames.


//  streamSetDesiredFrame() Remarks:
//      This function is called once for every frame that
//      the Dub engine wants to get from the decoder.

// Important: We must not modify the buffer tracking
// variables (frame_forw, frame_back, and frame_bidir)!

void VideoSourceMPEG::streamSetDesiredFrame(long frame_num) {
	long f, b;

//	_RPT1(0,"MPEG::streamSetDesiredFrame %d\n", frame_num);

	stream_current_frame = stream_desired_frame = frame_num;

	switch(parentPtr->video_sample_list[frame_num].frame_type) {

	case MPEG_FRAME_TYPE_P:
		if (frame_forw != stream_desired_frame && frame_back != stream_desired_frame) {
			// Back to previous I/P
			// Given: prev_IP(x) cannot fail when x is a P-frame
			do {
				b = prev_IP(stream_current_frame);
				if (frame_back == b) break;
				stream_current_frame = b;
			} while (parentPtr->video_sample_list[b].frame_type != MPEG_FRAME_TYPE_I);
		}
		break;

	case MPEG_FRAME_TYPE_B:
		if (frame_bidir != stream_desired_frame) {
			// Given: prev_IP(x) cannot fail when x is a B-frame
			// Given: next_IP(x) cannot fail when x is a B-frame
			f = next_IP(stream_desired_frame);
			b = prev_IP(stream_desired_frame);
			if (frame_forw != b || frame_back != f) {
				if (frame_back == b) {
					// We only need the next I/P
					stream_current_frame = f;
				} else {
					// Back to previous I/P
					do {
						stream_current_frame = b;
						if (parentPtr->video_sample_list[b].frame_type
						  == MPEG_FRAME_TYPE_I) break;

						b = prev_IP(stream_current_frame);
					} while (frame_back != b);
				}
			}
		}
	}
}

//	streamGetNextRequiredFrame() remarks:
//
//	This function is called repeatedly until it returns -1.
//
//	With each call, the return value is the frame number to read
//	from the file and decode. The reading and decoding stage is
//	skipped (I think) if this function returns -1 the first time
//	it's called after the dubber specifies its "desired frame."
//
//	The frame numbers returned are queued by the dubber, and
//	eventually passed in sequence to streamGetFrame(). However,
//	streamGetFrame() seems to be called from a different thread,
//	and it WILL NOT necessarily be called with the same frame #
//	that "stream_desired_frame" refers to at any given moment!
//
//	"is_preroll" should be set to TRUE when the frame number
//	returned equals the "desired frame".  In all other cases
//	it should be returned as FALSE.

long VideoSourceMPEG::streamGetNextRequiredFrame(BOOL *is_preroll) {
	long current = stream_current_frame;

//	_RPT2(0,"MPEG::streamGetNextRequiredFrame %d, %d,",
//	  stream_current_frame, stream_desired_frame);
//	_RPT3(0," %d, %d, %d\n", frame_forw, frame_back, frame_bidir);

	if (frame_forw == stream_desired_frame
	 || frame_back == stream_desired_frame
	 || frame_bidir == stream_desired_frame) {
		*is_preroll = FALSE;
		return -1;
	}

	if (current < stream_desired_frame) {
		stream_current_frame = next_IP(current);
	} else {
		stream_current_frame = stream_desired_frame;
	}

	if (parentPtr->video_sample_list[current].frame_type != MPEG_FRAME_TYPE_B) {
		frame_forw = frame_back;
		frame_back = current;
	} else frame_bidir = current;

	*is_preroll = (current != stream_desired_frame);
	return current;
}

// As far as I can tell, this function is never used for anything:
int VideoSourceMPEG::streamGetRequiredCount(long *pSize) {

// Return int = the number of frames which need to be decoded
// Return *pSize = the amount of memory needed to decode them
    long size;
    int needed = 1;
    int current = stream_current_frame;

    if (frame_forw == stream_desired_frame
      || frame_back == stream_desired_frame
      || frame_bidir == stream_desired_frame) {
        if (pSize) *pSize = 0;
        return 0;
    }

    size = parentPtr->video_sample_list[stream_desired_frame].size;

    while (current < stream_desired_frame) {
        if (parentPtr->video_sample_list[current]
          .frame_type != MPEG_FRAME_TYPE_B) {
            size += parentPtr->video_sample_list[current].size;
            needed++;
        }
        current = next_IP(current);
        if (current < 0) break;
    }
    if (current > stream_desired_frame) {
        size += parentPtr->video_sample_list[current].size;
        needed++;
    }

    if (pSize) *pSize = size;
    return needed;
}

void *VideoSourceMPEG::streamGetFrame(void *inputBuffer, long data_len, BOOL is_key, BOOL is_preroll, long frame_num) {

	int buffer = mpeg_lookup_frame(frame_num);

//	_RPT2(0,"MPEG::streamGetFrame %d %d\n", frame_num, is_preroll);

	if (is_preroll || buffer < 0) {

		// the "read" function gave us the extra 4 bytes we need
		if (data_len < 4) {
			is_preroll = TRUE;	// skips the next part
		} else {
			mpeg_decode_frame(inputBuffer, data_len, frame_num);
			buffer = mpeg_lookup_frame(frame_num);
		}
	}

	if (!is_preroll) {
		if (bmihDecompressedFormat->biCompression == 'YVYU')
			mpeg_convert_frameUYVY16(lpvBuffer, buffer);
		else if (bmihDecompressedFormat->biCompression == '2YUY')
			mpeg_convert_frameYUY216(lpvBuffer, buffer);
		else if (bmihDecompressedFormat->biCompression == '21VY')
			mpeg_convert_frameYV12(lpvBuffer, buffer);
		else if (bmihDecompressedFormat->biBitCount == 16)
			mpeg_convert_frame16(lpvBuffer, buffer);
		else if (bmihDecompressedFormat->biBitCount == 24)
			mpeg_convert_frame24(lpvBuffer, buffer);
		else
			mpeg_convert_frame32(lpvBuffer, buffer);

		fFBValid = true;
	}

	//if (MMX_enabled) __asm emms

	return lpvBuffer;
}

void *VideoSourceMPEG::getFrame(LONG frameNum) {
	int buffer, i;
	long lCurrent, f;
	MPEGSampleInfo *msi;

	if (frameNum < lSampleFirst || frameNum >= lSampleLast)
		throw MyError("VideoSourceMPEG->getFrame(): Invalid frameNum");

	// Do we have the buffer stored somewhere?
	buffer = mpeg_lookup_frame(frameNum);
	if (buffer < 0) {	// nope
		lCurrent = frameNum;

		switch (parentPtr->video_sample_list[frameNum].frame_type) {

        case MPEG_FRAME_TYPE_B:
			// Look for prediction frames and swap accordingly
			lCurrent = prev_IP(frameNum);
			i = mpeg_lookup_frame(lCurrent);
			if (i >= 0) {
				// Got the prev reference, how about next?
				lCurrent = next_IP(frameNum);
				if (mpeg_lookup_frame(lCurrent) >= 0) {
					mpeg_swap_buffers(i, MPEG_BUFFER_FORWARD);
					lCurrent = frameNum;
				} else {
					mpeg_swap_buffers(i, MPEG_BUFFER_BACKWARD);
					// lCurrent is next_IP
				}
			} else {
				// Back to previous reference, or I-frame
				while (lCurrent > lSampleFirst) {
					if (parentPtr->video_sample_list[lCurrent].frame_type
					  == MPEG_FRAME_TYPE_I) break;
					f = prev_IP(lCurrent);
					i = mpeg_lookup_frame(f);
					if (i >= 0) {
						mpeg_swap_buffers(i, MPEG_BUFFER_BACKWARD);
						break;
					}
					lCurrent = f;
				}
			}
			buffer = MPEG_BUFFER_BIDIRECTIONAL;
			break;

		case MPEG_FRAME_TYPE_P:
			// Look for forward prediction frame
			lCurrent = prev_IP(frameNum);
			i = mpeg_lookup_frame(lCurrent);
			if (i >= 0) {
				mpeg_swap_buffers(i, MPEG_BUFFER_BACKWARD);
				lCurrent = frameNum;
			} else {
				// Back to previous reference, or I-frame
				while (lCurrent > lSampleFirst) {
					if (parentPtr->video_sample_list[lCurrent].frame_type
					  == MPEG_FRAME_TYPE_I) break;
					f = prev_IP(lCurrent);
					i = mpeg_lookup_frame(f);
					if (i >= 0) {
						mpeg_swap_buffers(i, MPEG_BUFFER_BACKWARD);
						break;
					}
					lCurrent = f;
				}
			}
			// fall through

		case MPEG_FRAME_TYPE_I:
			buffer = MPEG_BUFFER_BACKWARD;
        }

		while (lCurrent >= lSampleFirst) {
			msi = &parentPtr->video_sample_list[lCurrent];
			parentPtr->ReadStream(parentPtr->video_packet_buffer, msi->stream_pos, msi->size, FALSE);
			mpeg_decode_frame(parentPtr->video_packet_buffer, msi->size+4, lCurrent);
			if (lCurrent > frameNum) {
				lCurrent = frameNum;
			} else {
				if (lCurrent == frameNum) break;
				lCurrent = next_IP(lCurrent);
			}
		}
	}

	if (bmihDecompressedFormat->biCompression == 'YVYU')
		mpeg_convert_frameUYVY16(lpvBuffer, buffer);
	else if (bmihDecompressedFormat->biCompression == '2YUY')
		mpeg_convert_frameYUY216(lpvBuffer, buffer);
	else if (bmihDecompressedFormat->biCompression == '21VY')
		mpeg_convert_frameYV12(lpvBuffer, buffer);
	else if (bmihDecompressedFormat->biBitCount == 16)
		mpeg_convert_frame16(lpvBuffer, buffer);
	else if (bmihDecompressedFormat->biBitCount == 24)
		mpeg_convert_frame24(lpvBuffer, buffer);
	else
		mpeg_convert_frame32(lpvBuffer, buffer);

	//if (MMX_enabled) __asm emms

	fFBValid = true;
	return lpvBuffer;
}

int VideoSourceMPEG::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {

    if (lStart < lSampleFirst || lStart >= lSampleLast)
        return AVIERR_BADPARAM;

    MPEGSampleInfo *msi = &parentPtr->video_sample_list[lStart];
    long len = msi->size;

    // Check to see if this is a frame type we're omitting

    switch(msi->frame_type) {
        case MPEG_FRAME_TYPE_P:
            if (parentPtr->iDecodeMode & InputFileMPEGOptions::DECODE_NO_P) {
                if (lSamplesRead) *lSamplesRead = 1;
                if (lBytesRead) *lBytesRead = 1;
                return AVIERR_OK;
            }
            break;
        case MPEG_FRAME_TYPE_B:
            if (parentPtr->iDecodeMode & InputFileMPEGOptions::DECODE_NO_B) {
                if (lSamplesRead) *lSamplesRead = 1;
                if (lBytesRead) *lBytesRead = 1;
                return AVIERR_OK;
            }
            break;
    }

    if (!lpBuffer) {
        if (lSamplesRead) *lSamplesRead = 1;
        if (lBytesRead) *lBytesRead = len;
        return AVIERR_OK;
    }

    if (len > cbBuffer) {
        if (lSamplesRead) *lSamplesRead = 0;
        if (lBytesRead) *lBytesRead = 0;
        return AVIERR_BUFFERTOOSMALL;
    }

    parentPtr->ReadStream(lpBuffer, msi->stream_pos, len, FALSE);

    if (lSamplesRead) *lSamplesRead = 1;
    if (lBytesRead) *lBytesRead = len;

    return AVIERR_OK;
}

///////

long VideoSourceMPEG::prev_IP(long lSample) {
	if (lSample < lSampleLast) {
		while (--lSample >= lSampleFirst) {
			if (parentPtr->video_sample_list[lSample].frame_type
			  != MPEG_FRAME_TYPE_B) return lSample;
		}
	}
    return -1;
}

long VideoSourceMPEG::next_IP(long lSample) {
	if (lSample >= lSampleFirst) {
		while (++lSample < lSampleLast) {
			if (parentPtr->video_sample_list[lSample].frame_type
			  != MPEG_FRAME_TYPE_B) return lSample;
		}
	}
    return -1;
}

//////////////////////////////////////////////////////////////////////////
//
//
//						AudioSourceMPEG
//
//
//////////////////////////////////////////////////////////////////////////

// ........ ........ 111..... 11111111 syncword
// ........ ........ ...xx... ........ version (11=1, 10=2, 00=2.5)
// ........ ........ .....xx. ........ layer (01=III, 10=II, 11=I)
// ........ ........ .......x ........ crc (0=yes, 1=no)
// xx...... ........ ........ ........ mode (00=stereo, 10=joint, 01=dual, 11=mono)
// ..xx.... ........ ........ ........ mode ext (only for joint stereo)
// ....x... ........ ........ ........ copyright (0=no, 1=yes)
// .....x.. ........ ........ ........ original (0=orig, 1=copy)
// ......xx ........ ........ ........ emphasis (not 10)
// ........ xxxx.... ........ ........ bitrate (not 0000 nor 1111)
// ........ ....xx.. ........ ........ sampling rate (not 11)
// ........ ......x. ........ ........ padded (0=no, 1=yes)
// ........ .......x ........ ........ private bit

static const int bitrate[2][3][16] = {
		{
			{ 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0 },	// MPEG-1 layer I
			{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, 0 },	// MPEG-1 layer II
			{ 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 0 },	// MPEG-1 layer III
		},
		{
			{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, 0 },	// MPEG-2 layer I
			{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 },	// MPEG-2 layer II
			{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 },	// MPEG-2 layer III
		}
};

static const long samp_freq[2][2][4] = {
	{
		{ 44100, 48000, 32000, 0 },		// MPEG-1
		{ 22050, 24000, 16000, 0 },		// MPEG-2
	},
	{
		{     0,     0,     0, 0 },		// impossible
		{ 11025, 12000,  8000, 0 },		// MPEG-2.5
	}
};

struct MPEGAudioHeader {
	enum {
		kMaskNone		= 0x00000000,
		kMaskSync		= 0x0000E0FF,
		kMaskMPEG25		= 0x00001000,
		kMaskVersion	= 0x00000800,
		kMaskLayer		= 0x00000600,
		kMaskCRC		= 0x00000100,
		kMaskBitrate	= 0x00F00000,
		kMaskSampleRate	= 0x000C0000,
		kMaskPadding	= 0x00020000,
		kMaskPrivate	= 0x00010000,
		kMaskMode		= 0xC0000000,
		kMaskModeExt	= 0x30000000,
		kMaskCopyright	= 0x08000000,
		kMaskOriginal	= 0x04000000,
		kMaskEmphasis	= 0x03000000,
		kMaskAll		= 0xFFFFFFFF
	};

	const uint32 mHeader;

	MPEGAudioHeader(uint32 hdr) : mHeader(hdr) {}

	bool		IsSyncValid() const				{ return (mHeader & kMaskSync) == kMaskSync; }
	bool		IsMPEG2() const					{ return !(mHeader & kMaskVersion); }
	bool		IsMPEG25() const				{ return IsMPEG2() && !(mHeader & kMaskMPEG25); }
	unsigned	GetLayer() const				{ return 4 - ((mHeader >> 9)&3); }
	bool		IsCRCProtected() const			{ return !(mHeader & kMaskCRC); }
	unsigned	GetBitrateIndex() const			{ return (mHeader >> 20) & 15; }
	unsigned	GetSamplingRateIndex() const	{ return (mHeader >> 18) & 3; }
	bool		IsPadded() const				{ return 0!=(mHeader & kMaskPadding); }
//    unsigned    GetModeIndex() const            { return (mHeader >> 28) & 3; }
	bool		IsStereo() const				{ return (mHeader & kMaskMode) != kMaskMode; }

	unsigned	GetBitrateKbps() const			{ return bitrate[IsMPEG2()][GetLayer()-1][GetBitrateIndex()]; }
	unsigned	GetSamplingRateHz() const		{ return samp_freq[IsMPEG25()][IsMPEG2()][GetSamplingRateIndex()]; }

// Added these:
    bool        IsCopyrighted() const           { return ((mHeader & kMaskCopyright)!=0); }
    bool        IsOriginal() const              { return ((mHeader & kMaskOriginal)==0); }
    bool        IsPrivate() const               { return ((mHeader & kMaskPrivate)!=0); }
    unsigned    GetModeExt() const              { return (mHeader >> 28) & 3; }
    unsigned    GetMode() const                 { return (mHeader >> 30) & 3; }
    unsigned    GetEmphasis() const             { return (mHeader >> 24) & 3; }

	bool IsValid() const;
	bool IsConsistent(uint32 hdr) const;
	unsigned GetFrameSize() const;
	unsigned GetPayloadSizeL3() const;
};

bool MPEGAudioHeader::IsValid() const {
	return IsSyncValid()					// need at least 11 bits for sync mark
		&& (!IsMPEG25() || IsMPEG2())		// either twelfth bit is set or it's MPEG-2.5
		&& GetLayer() != 4					// layer IV invalid
		&& GetBitrateIndex() != 15			// bitrate=1111 invalid
		&& GetSamplingRateIndex() != 3		// sampling_rate=11 reserved
		;
}

bool MPEGAudioHeader::IsConsistent(uint32 hdr) const {
	uint32 headerdiff = (hdr ^ mHeader);

	// do not allow MPEG version, layer, or sampling rate to change
	if (headerdiff & (kMaskSampleRate | kMaskVersion | kMaskLayer | kMaskMPEG25))
		return false;

	// only layer III decoders must accept VBR
	if (GetLayer() != 3 && (headerdiff & kMaskBitrate))
		return false;

	return true;
}

unsigned MPEGAudioHeader::GetFrameSize() const {
	const bool		is_mpeg2	= IsMPEG2();
	const unsigned	layer		= GetLayer();
	const unsigned	bitrate		= GetBitrateKbps();
	const unsigned	freq		= GetSamplingRateHz();
	const unsigned	padding		= IsPadded();

	if (GetLayer() == 1)
		return 4*(12000*bitrate/freq + padding);
	else {
		if (is_mpeg2)
			return (72000*bitrate/freq + padding);
		else
			return (144000*bitrate/freq + padding);
	}
}

unsigned MPEGAudioHeader::GetPayloadSizeL3() const {
	VDASSERT(GetLayer() == 3);

	static const unsigned sideinfo_size[2][2]={17,32,9,17};

	unsigned size = GetFrameSize() - sideinfo_size[IsMPEG2()][IsStereo()];

	if (IsCRCProtected())
		size -= 2;

	VDASSERT((signed)size > 0);

	return size;
}

// ********************************************************************
// *** VirtualDubMod	 											***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
class AudioSourceMPEG : public AudioSource, IAMPBitsource, IStreamIndex {
private:
	InputFileMPEG *parentPtr;
	IAMPDecoder *iad;
	void *pkt_buffer;
	short sample_buffer[1152*2][2];
	char *pDecoderPoint;
	char *pDecoderLimit;

	long lCurrentPacket;
	int layer;
	int mSamplesPerFrame;
    bool fAllowDSC;

	bool	mbIsMPEG2;
	ErrorMode	mErrorMode;

	bool _isKey(LONG lSample);

public:
	AudioSourceMPEG(InputFileMPEG *);
	~AudioSourceMPEG();

	// AudioSource methods

	bool init();
	int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);

	void setDecodeErrorMode(ErrorMode mode);
	bool isDecodeErrorModeSupported(ErrorMode mode);

	// IAMPBitsource methods

	int read(void *buffer, int bytes);

	bool QueryInterface(InterfaceID ID, void **destination);
	DWORD GetStreamType(void);
	long GetStreamIndex(void);
};

bool AudioSourceMPEG::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IStreamIndex::GetInterfaceID()) {
		*destination = static_cast<IStreamIndex *>(this);
		return true;
	}
	return false;
}

DWORD AudioSourceMPEG::GetStreamType(void) {
	return streamtypeAUDIO;
}

long AudioSourceMPEG::GetStreamIndex(void) {
	return 0;
}

AudioSourceMPEG::AudioSourceMPEG(InputFileMPEG *pp)
	: AudioSource()
	, mErrorMode(kErrorModeReportAll)
{
	parentPtr = pp;

	if (!(pkt_buffer = new char[8192]))
		throw MyMemoryError();

	if (!(iad = CreateAMPDecoder())) {
		delete pkt_buffer;
		throw MyMemoryError();
	}

    this->fAllowDSC = false;
	lCurrentPacket = -1;

	m_SourceType = SOURCE_MPEG | SOURCE_AUDIO;
	m_SourceName = L"Input MPEG file";
	m_SourceDesc = L"MPEG audio stream (MPEG)";
}
// END ****************************************************************

AudioSourceMPEG::~AudioSourceMPEG() {
	delete pkt_buffer;
	iad->Destroy();
}

bool AudioSourceMPEG::_isKey(LONG lSample) {
	return true;
}

bool AudioSourceMPEG::init() {
	WAVEFORMATEX *wfex;

	MPEGAudioHeader	header(parentPtr->audio_first_header);

	layer = header.GetLayer();

	mSamplesPerFrame = (layer==1 ? 384 : layer == 3 && header.IsMPEG2() ? 576 : 1152);

	lSampleFirst = 0;

    this->fAllowDSC = (parentPtr->fAllowDSC &&
      (layer == 3 || !header.IsMPEG2()));

    // Allow Direct Stream Copy of audio stream?
    if (this->fAllowDSC) {    
        int brate = header.GetBitrateKbps();
        long srate = header.GetSamplingRateHz();
        int version = header.IsMPEG2()? 2: 1;

        lSampleLast = (parentPtr->aframes > 0)?
            parentPtr->audio_sample_list[parentPtr->aframes-1].stream_pos
          + parentPtr->audio_sample_list[parentPtr->aframes-1].size: 0;

        if (layer == 3) {
            MPEGLAYER3WAVEFORMAT *wfex3;

            if (!(wfex3 = (MPEGLAYER3WAVEFORMAT *)allocFormat(sizeof(MPEGLAYER3WAVEFORMAT))))
                return false;

            wfex3->wfx.wFormatTag       = WAVE_FORMAT_MPEGLAYER3;
            wfex3->wfx.nChannels        = header.IsStereo()? 2: 1;
            wfex3->wfx.nSamplesPerSec   = srate;
            wfex3->wfx.nBlockAlign      = 1;
            wfex3->wfx.wBitsPerSample   = 0;
            wfex3->wfx.cbSize           = MPEGLAYER3_WFX_EXTRA_BYTES;
            wfex3->wID                  = MPEGLAYER3_ID_MPEG;
            wfex3->fdwFlags             = MPEGLAYER3_FLAG_PADDING_ISO;  //?
            wfex3->nBlockSize           = ((144000 / version) * brate) / srate;
            wfex3->nFramesPerBlock      = 1;
            wfex3->nCodecDelay          = 1393;
            wfex3->wfx.nAvgBytesPerSec  = (srate * version * wfex3->nBlockSize) / 1152;

            streamInfo.fccType			= streamtypeAUDIO;
	        streamInfo.fccHandler		= 0;
	        streamInfo.dwCaps			= 0;
	        streamInfo.wPriority		= 0;
	        streamInfo.wLanguage		= 0;
	        streamInfo.dwFlags			= 0;
            streamInfo.dwScale          = 1;
            streamInfo.dwRate           = wfex3->wfx.nAvgBytesPerSec;
	        streamInfo.dwStart			= 0;
            streamInfo.dwLength         = lSampleLast;
	        streamInfo.dwInitialFrames	        = 1;
	        streamInfo.dwSuggestedBufferSize	= 0;
	        streamInfo.dwQuality				= 0xffffffffL;
	        streamInfo.dwSampleSize				= 1;
	        streamInfo.rcFrame.left				= 0;
	        streamInfo.rcFrame.top				= 0;
	        streamInfo.rcFrame.right			= 0;
	        streamInfo.rcFrame.bottom			= 0;

        } else {    // Layer I or II, MPEG-1 only
            MPEG1WAVEFORMAT *wfex1;
			double total_br;
			int i;

			bool pad = header.IsPadded();

			int layermask[4] = {
				ACM_MPEG_LAYER1,
				ACM_MPEG_LAYER2,
				ACM_MPEG_LAYER3,
				0};

			int headmode[4] = {
				ACM_MPEG_STEREO,
				ACM_MPEG_JOINTSTEREO,
				ACM_MPEG_DUALCHANNEL,
				ACM_MPEG_SINGLECHANNEL};

            if (!(wfex1 = (MPEG1WAVEFORMAT *)allocFormat(sizeof(MPEG1WAVEFORMAT))))
                return false;

			wfex1->wfx.wFormatTag		= WAVE_FORMAT_MPEG;
			wfex1->wfx.nChannels		= header.IsStereo()? 2: 1;
			wfex1->wfx.nSamplesPerSec	= srate;
			wfex1->wfx.nBlockAlign		= 1;
			wfex1->wfx.wBitsPerSample	= 0;
			wfex1->wfx.cbSize			= 22;

			wfex1->dwHeadBitrate = brate * 1000;

			total_br = 0.0;
			wfex1->fwHeadMode = 0;
            wfex1->fwHeadFlags = 0;
			wfex1->fwHeadLayer = 0;

			// Assume a fixed frame size
			if (layer == 1) {
				wfex1->wfx.nBlockAlign =
				  4 * (int)(12000 * brate / srate);
			} else {	// Layer II
				wfex1->wfx.nBlockAlign =
				  (int)(144000 * brate / srate);
			}

			for (i=0; i<parentPtr->aframes; i++) {
				MPEGAudioHeader	hdr(parentPtr->audio_sample_list[i].header);

				// Channel mode may change
				int k = headmode[hdr.GetMode()];
				wfex1->fwHeadMode |= k;
				if (k == ACM_MPEG_JOINTSTEREO) {
					wfex1->fwHeadModeExt |= (1 << hdr.GetModeExt());
				}

				// Layer may change
				wfex1->fwHeadLayer |= layermask[hdr.GetLayer()-1];

				// Sampling rate may change
				if (hdr.GetSamplingRateHz() != srate)
					wfex1->wfx.nSamplesPerSec = 0;

				// Bitrate may change
				k = hdr.GetBitrateKbps();
				total_br += k;
				if (k != brate) {
					// Variable Bit Rate
					wfex1->dwHeadBitrate = 0;
					wfex1->wfx.nBlockAlign = 1;
				} else if (hdr.IsPadded() != pad) {
					// Variable frame size
					wfex1->wfx.nBlockAlign = 1;
				}

				// Flags may change
				if (!hdr.IsMPEG2()) wfex1->fwHeadFlags |= ACM_MPEG_ID_MPEG1;
				if (hdr.IsPrivate()) wfex1->fwHeadFlags |= ACM_MPEG_PRIVATEBIT;
				if (hdr.IsCopyrighted()) wfex1->fwHeadFlags |= ACM_MPEG_COPYRIGHT;
				if (hdr.IsOriginal()) wfex1->fwHeadFlags |= ACM_MPEG_ORIGINALHOME;
				if (hdr.IsCRCProtected()) wfex1->fwHeadFlags |= ACM_MPEG_PROTECTIONBIT;
			}

			wfex1->wfx.nAvgBytesPerSec = (parentPtr->aframes<=0)? 0:
			    VDRoundToLong((total_br * 125.0) / (double)parentPtr->aframes);

            wfex1->wHeadEmphasis = header.GetEmphasis()+1;

			// FIXME: store the real PTS?
            wfex1->dwPTSLow = 0;
            wfex1->dwPTSHigh = 0;

            streamInfo.fccType			= streamtypeAUDIO;
	        streamInfo.fccHandler		= 0;
	        streamInfo.dwCaps			= 0;
	        streamInfo.wPriority		= 0;
	        streamInfo.wLanguage		= 0;
	        streamInfo.dwFlags			= 0;
            streamInfo.dwScale          = wfex1->wfx.nBlockAlign;
            streamInfo.dwRate           = wfex1->wfx.nAvgBytesPerSec;
	        streamInfo.dwStart			= 0;
            streamInfo.dwLength         = lSampleLast/wfex1->wfx.nBlockAlign;
	        streamInfo.dwInitialFrames	        = 0;
	        streamInfo.dwSuggestedBufferSize	= 0;
	        streamInfo.dwQuality				= 0xffffffffL;
	        streamInfo.dwSampleSize				= wfex1->wfx.nBlockAlign;
	        streamInfo.rcFrame.left				= 0;
	        streamInfo.rcFrame.top				= 0;
	        streamInfo.rcFrame.right			= 0;
	        streamInfo.rcFrame.bottom			= 0;
        }
        return true;
    }

	lSampleLast = parentPtr->aframes * mSamplesPerFrame;

    if (!(wfex = (WAVEFORMATEX *)allocFormat(sizeof(WAVEFORMATEX))))
        return false;

    // [?] * [bits/sec] / [samples/sec] = [bytes/frame]
    // [?] * [bits/sample] = [bytes/frame]
    // [?] * [.125 byte/bit] * [bits/sample] = [bytes/frame]
    // [samples/frame]

    wfex->wFormatTag        = WAVE_FORMAT_PCM;
    wfex->nChannels         = header.IsStereo()? 2: 1;
    wfex->nSamplesPerSec    = header.GetSamplingRateHz();
    wfex->nBlockAlign       = wfex->nChannels*2;
    wfex->nAvgBytesPerSec   = wfex->nSamplesPerSec * wfex->nBlockAlign;
    wfex->wBitsPerSample    = 16;
    wfex->cbSize            = 0;

    streamInfo.fccType					= streamtypeAUDIO;
    streamInfo.fccHandler				= 0;
    streamInfo.dwCaps					= 0;
    streamInfo.wPriority				= 0;
    streamInfo.wLanguage				= 0;
    streamInfo.dwFlags					= 0;
    streamInfo.dwScale					= wfex->nBlockAlign;
    streamInfo.dwRate					= wfex->nSamplesPerSec * wfex->nBlockAlign;
    streamInfo.dwStart					= 0;
    streamInfo.dwLength					= parentPtr->aframes * mSamplesPerFrame;
    streamInfo.dwInitialFrames			= 0;
    streamInfo.dwSuggestedBufferSize	= 0;
    streamInfo.dwQuality				= 0xffffffffL;
    streamInfo.dwSampleSize				= wfex->nBlockAlign;
    streamInfo.rcFrame.left				= 0;
    streamInfo.rcFrame.top				= 0;
    streamInfo.rcFrame.right			= 0;
    streamInfo.rcFrame.bottom			= 0;

	try {
		iad->Init();
		iad->setSource(this);
	} catch(int i) {
		throw MyError(iad->getErrorString(i));
	}

	return true;
}

int AudioSourceMPEG::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {
    long lAudioPacket;
    MPEGSampleInfo *msi;
    long samples, len;

	long ba = getWaveFormat()->nBlockAlign;

    if (this->fAllowDSC) {
		if (lCount == AVISTREAMREAD_CONVENIENT)	{
			lCount = (ba > 1)? 1:
			  MPEGAudioHeader(parentPtr->audio_first_header).GetFrameSize();
		}
        if (!lpBuffer || lCount <= 0) {
			if (lCount < 0) lCount = 0;
            if (lSamplesRead) *lSamplesRead = lCount;
            if (lBytesRead) *lBytesRead = lCount;
            return AVIERR_OK;
        }
		len = lCount*ba;
		if (len > cbBuffer) {
			if (lSamplesRead) *lSamplesRead = 0;
			if (lBytesRead) *lBytesRead = 0;
			return AVIERR_BUFFERTOOSMALL;
		}
        parentPtr->ReadStream(lpBuffer, lStart*ba, len, true);
        if (lSamplesRead) *lSamplesRead = lCount;
        if (lBytesRead) *lBytesRead = len;
        return AVIERR_OK;
    }


	lAudioPacket = lStart/mSamplesPerFrame;
	samples = mSamplesPerFrame - (lStart % mSamplesPerFrame);

	if (lCount != AVISTREAMREAD_CONVENIENT)
		if (samples > lCount) samples = lCount;

#if 0
	if (samples*ba > cbBuffer) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return AVIERR_BUFFERTOOSMALL;
	}
#else
	if (samples*ba > cbBuffer && lpBuffer) {
		samples = cbBuffer / ba;

		if (samples <= 0) {
			if (lSamplesRead) *lSamplesRead = 0;
			if (lBytesRead) *lBytesRead = 0;
			return AVIERR_BUFFERTOOSMALL;
		}
	}
#endif

	if (!lpBuffer) {
		if (lSamplesRead) *lSamplesRead = samples;
		if (lBytesRead) *lBytesRead = samples * ba;
		return AVIERR_OK;
	}

	// Because of the overlap from the subband synthesis window, we must
	// decode the previous packet in order to avoid glitches in the output.
	// For layer III, we must also ensure that at least 511 bytes are present
	// in the bit reservoir, as packets may dig that far back using
	// main_data_begin.  An MPEG-1 stereo stream may use up to 32 bytes of
	// the data payload for sideband information.

	if (lCurrentPacket != lAudioPacket) {
		try {
			long nDecodeStart;

			if (layer!=3 || lCurrentPacket<0 || lCurrentPacket+1 != lAudioPacket) {
//				_RPT0(0,"Resetting...\n");
				iad->Reset();

				lCurrentPacket = lAudioPacket;
				--lCurrentPacket;
				if (lCurrentPacket < 0)
					lCurrentPacket = 0;

				nDecodeStart = lCurrentPacket;

				// layer III: add packets to preload bit reservoir
				if (layer == 3) {
					long nReservoirDelay = 511;		// main_data_start is 9 bits (0..511)
					
					while(lCurrentPacket > 0 && nReservoirDelay > 0) {
						--lCurrentPacket;

						nReservoirDelay -= MPEGAudioHeader(parentPtr->audio_sample_list[lCurrentPacket].header).GetPayloadSizeL3();
					}
				}
			} else {
				lCurrentPacket	= lAudioPacket;
				nDecodeStart	= lAudioPacket;
			}

			do {
//				_RPT1(0,"Decoding packet: %d\n", lCurrentPacket);

				msi = &parentPtr->audio_sample_list[lCurrentPacket];
				len = msi->size;

				parentPtr->ReadStream(pkt_buffer, msi->stream_pos, len, TRUE);

				pDecoderPoint = (char *)pkt_buffer;
				pDecoderLimit = (char *)pkt_buffer + len;

				if ((unsigned char)pDecoderPoint[0] != 0xff || ((unsigned char)pDecoderPoint[1]&0xe0)!=0xe0)
					throw MyInternalError("MPEG audio header in sample list has bad sync mark.\n(%s:%d)", __FILE__, __LINE__);

				iad->setDestination((short *)sample_buffer);
				iad->ReadHeader();

				try {
					if (lCurrentPacket < nDecodeStart)
						iad->PrereadFrame();
					else
						iad->DecodeFrame();
				} catch(int i) {
					if (mErrorMode != kErrorModeReportAll) {
						const char *pszError = iad->getErrorString(i);
						unsigned frame = lAudioPacket;
						VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_AudioConcealingError, 2, &lAudioPacket, &pszError);

						if (lCurrentPacket >= nDecodeStart)
							memset(sample_buffer, 0, sizeof sample_buffer);

						iad->ConcealFrame();
					} else
						throw;
				}

			} while(++lCurrentPacket <= lAudioPacket);
			--lCurrentPacket;
		} catch(int i) {
			char buf[64];
			long badpacket = lCurrentPacket;

			uint32 ticks = (uint32)(0.5 + (badpacket * 1000.0 * mSamplesPerFrame) / getWaveFormat()->nSamplesPerSec);
			ticks_to_str(buf, ticks);

			lCurrentPacket = -1;

			throw MyError("Error decoding MPEG audio frame %lu (%s): %s", (unsigned long)badpacket, buf, iad->getErrorString(i));
		}
	}

	memcpy(lpBuffer, (short *)sample_buffer + (lStart%mSamplesPerFrame)*(ba/2), samples*ba);

	if (lSamplesRead) *lSamplesRead = samples;
	if (lBytesRead) *lBytesRead = samples*ba;

	return AVIERR_OK;
}

int AudioSourceMPEG::read(void *buffer, int bytes) {
	if (pDecoderPoint+bytes > pDecoderLimit)
		throw MyError("Incomplete audio frame");

	memcpy(buffer, pDecoderPoint, bytes);
	pDecoderPoint += bytes;

	return bytes;
}

void AudioSourceMPEG::setDecodeErrorMode(ErrorMode mode) {
	if (mode == kErrorModeDecodeAnyway)
		mode = kErrorModeConceal;
	mErrorMode = mode;
}

bool AudioSourceMPEG::isDecodeErrorModeSupported(ErrorMode mode) {
	return mode == kErrorModeReportAll || mode == kErrorModeConceal;
}


// Derived from liba52
static int ac3_syncinfo(void *src, long *srate, int *brate, int *chans) {

	static int rate[19] = {32,40,48,56,64,80,96,112,
		128,160,192,224,256,320,384,448,512,576,640};
	static char lfeon[8] = {16,16,4,4,4,1,4,1};
	static char chan[8] = {2,1,2,3,3,4,4,5};

	const unsigned char *buf = (unsigned char *)src;

	if (*(short *)src == 0x770B && buf[4] < 0xC0) {
		int frmsizecod = buf[4] & 63;
		if (frmsizecod < 38) {
			int bsid = buf[5] >> 3;
			if (bsid < 12) {
				int acmod = buf[6] >> 5;
				int bitrate = rate[frmsizecod >> 1];

				if (bsid > 8) bsid -= 8; else bsid = 0;
				if (chans) *chans = chan[acmod] + !!(buf[6] & lfeon[acmod]);
				if (brate) *brate = bitrate >> bsid;

				switch (buf[4] & 0xC0) {
					case 0:
						if (srate) *srate = 48000 >> bsid;
						return (4 * bitrate);
					case 0x40:
						if (srate) *srate = 44100 >> bsid;
						return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));
					case 0x80:
						if (srate) *srate = 32000 >> bsid;
						return (6 * bitrate);
				}
			}
		}
	}
	return 0;
}

// ********************************************************************
// *** VirtualDubMod	 											***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
class AudioSourceAC3 : public AudioSource, public IStreamIndex {
private:
    InputFileMPEG *parentPtr;
	int nBlockAlign;

    bool _isKey(LONG lSample);

public:
    AudioSourceAC3(InputFileMPEG *);
    ~AudioSourceAC3();

    // AudioSource methods
    bool init();
    int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);

	bool QueryInterface(InterfaceID ID, void **destination);
	DWORD GetStreamType(void);
	long GetStreamIndex(void);
};

bool AudioSourceAC3::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IStreamIndex::GetInterfaceID()) {
		*destination = static_cast<IStreamIndex *>(this);
		return true;
	}
	return false;
}

DWORD AudioSourceAC3::GetStreamType(void) {
	return streamtypeAUDIO;
}

long AudioSourceAC3::GetStreamIndex(void) {
	return 0;
}

AudioSourceAC3::AudioSourceAC3(InputFileMPEG *pp) : AudioSource() {
    parentPtr = pp;

	m_SourceType = SOURCE_MPEG | SOURCE_AUDIO;
	m_SourceName = L"Input MPEG file";
	m_SourceDesc = L"MPEG audio stream (AC3)";
}
// END ****************************************************************

AudioSourceAC3::~AudioSourceAC3() {
}

bool AudioSourceAC3::_isKey(LONG lSample) {
    return true;
}

// For AudioSourceAC3, audio_first_header is of this form:
//      ........ ........ ........ ..xxxxxx     fscode
//      ........ ........ ........ xx......     srcode
//      ........ ........ .....xxx ........     audio service
//      ........ ........ xxxxx... ........     version
//      ........ xxx..... ........ ........     audio coding mode

bool AudioSourceAC3::init() {
    WAVEFORMATEX *wfex;

	unsigned long syncinfo[2];
	int frmsize, chans, brate;
	long srate;

    lSampleFirst = lSampleLast = 0;

    wfex = (WAVEFORMATEX *)allocFormat(sizeof(WAVEFORMATEX)*2);
    if (!wfex) return false;

	syncinfo[0] = 0x770B;
	syncinfo[1] = parentPtr->audio_first_header;

	frmsize = ac3_syncinfo(syncinfo, &srate, &brate, &chans);
	if (!frmsize) return false;

	// Construct WAVEFORMATEX for AC-3
    wfex->wFormatTag = 0x2000;
	wfex->nChannels = chans;
	wfex->nSamplesPerSec = srate;
	wfex->nAvgBytesPerSec = brate * 125UL;
	wfex->wBitsPerSample = 0;

	if (srate == 44100) {
		if (parentPtr->aframes > 0) {
			lSampleLast = parentPtr->audio_sample_list[parentPtr->aframes-1].stream_pos
				+ parentPtr->audio_sample_list[parentPtr->aframes-1].size;
		}
		wfex->nBlockAlign = 1;
	} else {
		if (parentPtr->aframes > 0) lSampleLast = parentPtr->aframes;
		wfex->nBlockAlign = frmsize;
	}

	this->nBlockAlign = wfex->nBlockAlign;

	// I think someone goofed this
	wfex->cbSize = sizeof(WAVEFORMATEX);
	memset((char *)wfex + sizeof(WAVEFORMATEX), 0, sizeof(WAVEFORMATEX));

    streamInfo.fccType                  = streamtypeAUDIO;
    streamInfo.fccHandler               = 0;
    streamInfo.dwCaps                   = 0;
    streamInfo.wPriority                = 0;
    streamInfo.wLanguage                = 0;
    streamInfo.dwFlags                  = 0;
    streamInfo.dwScale                  = wfex->nBlockAlign;
    streamInfo.dwRate                   = wfex->nAvgBytesPerSec;
    streamInfo.dwStart                  = 0;
    streamInfo.dwLength                 = lSampleLast;
    streamInfo.dwInitialFrames          = 1;
    streamInfo.dwSuggestedBufferSize    = 0;
    streamInfo.dwQuality                = 0xffffffffL;
    streamInfo.dwSampleSize             = wfex->nBlockAlign;
    streamInfo.rcFrame.left             = 0;
    streamInfo.rcFrame.top              = 0;
    streamInfo.rcFrame.right            = 0;
    streamInfo.rcFrame.bottom           = 0;

    return true;
}

int AudioSourceAC3::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {

	if (lCount == AVISTREAMREAD_CONVENIENT)
		lCount = (nBlockAlign == 1)? 2048: 1;

	if (!lpBuffer || lCount <= 0) {
		if (lCount < 0) lCount = 0;
		if (lSamplesRead) *lSamplesRead = lCount;
		if (lBytesRead) *lBytesRead = lCount * nBlockAlign;
		return AVIERR_OK;
	}

	if (lStart < lSampleFirst || lStart >= lSampleLast) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return AVIERR_OK;
	}

	if (lSamplesRead) *lSamplesRead = lCount;

	if (nBlockAlign != 1) {
		lStart = parentPtr->audio_sample_list[lStart].stream_pos;
		lCount *= nBlockAlign;
	}

	parentPtr->ReadStream(lpBuffer, lStart, lCount, true);
	if (lBytesRead) *lBytesRead = lCount;

	return AVIERR_OK;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                      AudioSourcePCM
//
//
//////////////////////////////////////////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod	 											***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
class AudioSourcePCM : public AudioSource, public IStreamIndex {
private:
    InputFileMPEG *parentPtr;
    void *pkt_buffer;
    short sample_buffer[1536*2][2];

    long lCurrentPacket;

    bool _isKey(LONG lSample);

public:
    AudioSourcePCM(InputFileMPEG *);
    ~AudioSourcePCM();

    // AudioSource methods
    bool init();
    int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);

	bool QueryInterface(InterfaceID ID, void **destination);
	DWORD GetStreamType(void);
	long GetStreamIndex(void);
};

bool AudioSourcePCM::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == IStreamIndex::GetInterfaceID()) {
		*destination = static_cast<IStreamIndex *>(this);
		return true;
	}
	return false;
}

DWORD AudioSourcePCM::GetStreamType(void) {
	return streamtypeAUDIO;
}

long AudioSourcePCM::GetStreamIndex(void) {
	return 0;
}

AudioSourcePCM::AudioSourcePCM(InputFileMPEG *pp) : AudioSource() {
    parentPtr = pp;

    if (!(pkt_buffer = new char[8192]))
        throw MyMemoryError();

    lCurrentPacket = -1;

	m_SourceType = SOURCE_MPEG | SOURCE_AUDIO;
	m_SourceName = L"Input MPEG file";
	m_SourceDesc = L"MPEG audio stream (PCM)";
}
// END ****************************************************************

AudioSourcePCM::~AudioSourcePCM() {
    delete pkt_buffer;
}

bool AudioSourcePCM::_isKey(LONG lSample) {
    return true;
}

bool AudioSourcePCM::init() {

    WAVEFORMATEX *wfex;
    int hdr;

    if (!(wfex = (WAVEFORMATEX*)allocFormat(sizeof(WAVEFORMATEX))))
        return false;

    hdr = parentPtr->audio_first_header;

    wfex->wFormatTag = WAVE_FORMAT_PCM;
    wfex->nChannels = (hdr & 7) + 1;
    wfex->nSamplesPerSec = (hdr & 0x30)? 96000: 48000;
    wfex->nBlockAlign = wfex->nChannels*2;
    wfex->nAvgBytesPerSec = wfex->nSamplesPerSec * wfex->nBlockAlign;
    wfex->wBitsPerSample = 16;

    lSampleFirst = 0;
    lSampleLast = parentPtr->aframes * 1536;

    streamInfo.fccType                  = streamtypeAUDIO;
    streamInfo.fccHandler               = 0;
    streamInfo.dwCaps                   = 0;
    streamInfo.wPriority                = 0;
    streamInfo.wLanguage                = 0;
    streamInfo.dwFlags                  = 0;
    streamInfo.dwScale                  = wfex->nBlockAlign;
    streamInfo.dwRate                   = wfex->nAvgBytesPerSec;
    streamInfo.dwStart                  = 0;
    streamInfo.dwLength                 = lSampleLast;
    streamInfo.dwInitialFrames          = 0;
    streamInfo.dwSuggestedBufferSize    = 0;
    streamInfo.dwQuality                = 0xffffffffL;
    streamInfo.dwSampleSize             = wfex->nBlockAlign;
    streamInfo.rcFrame.left             = 0;
    streamInfo.rcFrame.top              = 0;
    streamInfo.rcFrame.right            = 0;
    streamInfo.rcFrame.bottom           = 0;

    return true;
}

static __inline void pcm_byte_swap(void *src, void *dst, int len) {
    __asm {
        mov     esi, src
        mov     edi, dst
        mov     ecx, len
        sar     ecx, 2
        js      done        ; len < 0 not allowed
        jz      part2       ; less than 4 bytes left
        sub     edi, 4
        align 16
loop1:  mov     eax, [esi]  ; [3][2][1][0]
        add     edi, 4
        bswap   eax         ; [0][1][2][3]
        add     esi, 4
        rol     eax, 16     ; [2][3][0][1]
        dec     ecx
        mov     [edi], eax
        jg      loop1
part2:  test    len, 2
        jz      done        ; less than 2 bytes left
        mov     ax, word ptr [esi]
        xchg    ah, al
        mov     word ptr [edi], ax
done:
    }
}

int AudioSourcePCM::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) {
    long lAudioPacket;
    MPEGSampleInfo *msi;
    long len,samples;

    WAVEFORMATEX *wfex = getWaveFormat();
    long ba = wfex->nBlockAlign;

    lAudioPacket = lStart/1536;
    samples = 1536 - (lStart % 1536);

    if (lCount != AVISTREAMREAD_CONVENIENT)
        if (samples > lCount) samples = lCount;

    if (samples*ba > cbBuffer && lpBuffer) {
        samples = cbBuffer / ba;

        if (samples <= 0) {
            if (lSamplesRead) *lSamplesRead = 0;
            if (lBytesRead) *lBytesRead = 0;
            return AVIERR_BUFFERTOOSMALL;
        }
    }

    if (!lpBuffer) {
        if (lSamplesRead) *lSamplesRead = samples;
        if (lBytesRead) *lBytesRead = samples * ba;
        return AVIERR_OK;
    }

    if (lCurrentPacket != lAudioPacket) {
        try {
            msi = &parentPtr->audio_sample_list[lAudioPacket];
            len = msi->size;

            parentPtr->ReadStream(pkt_buffer, msi->stream_pos, len, TRUE);
            pcm_byte_swap(pkt_buffer, sample_buffer, len);
            lCurrentPacket = lAudioPacket;

        } catch(...) {
            throw MyError("LPCM read error");
        }
    }

    memcpy(lpBuffer, (short *)sample_buffer + (lStart % 1536)*(ba/2), samples*ba);

    if (lSamplesRead) *lSamplesRead = samples;
    if (lBytesRead) *lBytesRead = samples*ba;

    return AVIERR_OK;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                  MPEGAudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class MPEGAudioParser {
protected:
    unsigned long lFirstHeader;
    MPEGSampleInfo msi;
    __int64 bytepos;
    int hstate, skip;
private:
    unsigned long header;

public:
    MPEGAudioParser();

    virtual void Parse(const void *, int, DataVector *);
    unsigned long getHeader();
};

MPEGAudioParser::MPEGAudioParser() {
    lFirstHeader = 0;
    header = 0;
    hstate = 0;
    skip = 0;
    bytepos = 0;
}

unsigned long MPEGAudioParser::getHeader() {
    return lFirstHeader;
}

void MPEGAudioParser::Parse(const void *pData, int len, DataVector *dst) {
    unsigned char *src = (unsigned char *)pData;

    while(len>0) {
        if (skip) {
            int tc = skip;

            if (tc > len)
                tc = len;

            len -= tc;
            skip -= tc;
            src += tc;

            // Audio frame finished?

            if (!skip) {
				msi.PTS = -1;
                dst->Add(&msi);
            }

            continue;
        }

        // Collect header bytes.

        ++hstate;
        header = (header>>8) | ((long)*src++ << 24);
        --len;

        MPEGAudioHeader hdr(header);
        if (hstate>=4 && hdr.IsValid()) {

            if (lFirstHeader && !hdr.IsConsistent(lFirstHeader))
                continue;

            // Okay, we like the header.

            hstate = 0;

            // Must be a frame start.

            if (!lFirstHeader)
                lFirstHeader = header;

            // Setup the sample information.  Don't add the sample, in case it's incomplete.

            msi.stream_pos      = bytepos + (src - (unsigned char *)pData) - 4;
            msi.header          = header;
            msi.size            = hdr.GetFrameSize();

            // Now skip the remainder of the sample.

            skip = msi.size-4;
        }
    }

    bytepos += src - (unsigned char *)pData;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                  AC3AudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class AC3AudioParser : public MPEGAudioParser {
private:
    unsigned long syncinfo[2];
public:
    AC3AudioParser();
    void Parse(const void *, int, DataVector *);
};

AC3AudioParser::AC3AudioParser() {
    lFirstHeader = 0;
    hstate = 0;
    skip = 0;
    bytepos = 0;
    syncinfo[0] = 0;
}

void AC3AudioParser::Parse(const void *pData, int len, DataVector *dst) {
    unsigned char *src = (unsigned char *)pData;

    while(len>0) {
        if (skip) {
            int tc = skip;
            if (tc > len) tc = len;

            len -= tc;
            skip -= tc;
            src += tc;

            // Audio frame finished?
            if (!skip) {
				msi.PTS = -1;
				dst->Add(&msi);
			}
            continue;
        }

        // Collect header bytes.
        if (syncinfo[0] == 0x770B) {

//  0B 77 mm ll fs va cm ??
//          0B 77 = syncword
//          mm ll = MSB, LSB of CRC (ignore)
//          fs    = fscode, srcode
//          va    = version, audio service
//          cm    = audio coding mode
//  store audio_first_header as [?? cm va fs]

            syncinfo[1] = (syncinfo[1] >> 8) | ((*src++) << 24);
            --len;
            if (++hstate == 6) {
                hstate = 0;

                if (skip = ac3_syncinfo(syncinfo, NULL, NULL, NULL)) {
                    if (!lFirstHeader) lFirstHeader = syncinfo[1];
                    msi.stream_pos = bytepos + (src - (unsigned char *)pData) - 8;
                    msi.header = syncinfo[1];
                    msi.size = skip;
                    skip -= 8;
                }

                syncinfo[0] = 0;
            }
        } else {
            syncinfo[0] = (syncinfo[0] >> 8) | ((*src++) << 8);
            --len;
        }
    }

    bytepos += src - (unsigned char *)pData;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                  PCMAudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class PCMAudioParser : public MPEGAudioParser {
public:
    PCMAudioParser();
    void Parse(const void *pData, int, DataVector *);
};

PCMAudioParser::PCMAudioParser() {
    lFirstHeader = 0;
    bytepos = 0;
    skip = 0;
}

void PCMAudioParser::Parse(const void *pData, int len, DataVector *dst) {

    if (pData) {
    // pData = LPCM header (6 bytes):
    //  0:  xxxxxxxx    number of frames
    //  1:  xxxxxxxx    offset of first frame (MSB)
    //  2:  xxxxxxxx    offset of first frame (LSB)
    //  3:  x.......    emphasis
    //      .x......    mute
    //      ..x.....    reserved
    //      ...xxxxx    frame number % 20
    //  4:  xx......    00=16-,01=20-,10=24-bit (11=forbidden)
    //      ..xx....    00=48,01=96 KHz (others reserved?)
    //      ....x...    reserved
    //      .....xxx    number of channels - 1
    //  5:  xxxxxxxx    dynamic range control

        unsigned char *src = (unsigned char *)pData;
        lFirstHeader = (src[3]<<8) + src[4] + (src[5]<<16);
        lFirstHeader |= 0xFF000000;     // ensure a non-zero lFirstHeader
        skip = ((lFirstHeader&7)+1) * 1536*2;
        return;
    }

    while(len > 0) {
        int tc = skip;
        if (tc > len) tc = len;
        len -= tc;
        skip -= tc;
        if (!skip) {
            skip = ((lFirstHeader&7)+1) * 1536*2;
            msi.stream_pos = bytepos;
            bytepos += skip;
            msi.size = skip;
			msi.PTS = -1;
            dst->Add(&msi);
        }
    }
}

//////////////////////////////////////////////////////////////////////////
//
//
//                  MPEGVideoParser
//
//
//////////////////////////////////////////////////////////////////////////

#define SEQUENCE_EXTENSION_ID           1
#define SEQUENCE_DISPLAY_EXTENSION_ID   2
#define PICTURE_CODING_EXTENSION_ID     8

// Picture structures
#define TOP_FIELD       1
#define BOTTOM_FIELD    2
#define FRAME_PICTURE   3

// Fix for Kylie: limit what res we will support
#define MPEG_MAX_WIDTH      1920
#define MPEG_MAX_HEIGHT     1080
// 1920*1080*4 = 8MB buffer

class MPEGVideoParser {
private:
    unsigned char buf[8];

    int idx, bytes;

    MPEGSampleInfo msi;
    __int64 bytepos;
    long header;

    bool fPicturePending;
    bool fFoundSequenceStart;
    bool fFoundSequenceExt;
    bool fInitialized;
    bool fSeqAttach;
    int ReferenceCount;

public:
    VDFraction mFrameRate;
    int width, height;
	int chroma_format;
    int dwidth, dheight;
    bool fFoundDisplayExt;
    int aspect_ratio_info;
    bool progressive_sequence;
    bool fIsMPEG2;

    MPEGVideoParser();

    void setPos(__int64);
    void Parse(const void *, int, DataVector *);
};

MPEGVideoParser::MPEGVideoParser() {
    bytepos = 0;
    header = -1;

    fPicturePending = false;
    fFoundSequenceStart = false;
    fFoundSequenceExt = false;
    fFoundDisplayExt = false;
    fInitialized = false;
    fSeqAttach = true;
    fIsMPEG2 = false;
    ReferenceCount = 0;
	chroma_format = CHROMA420;

    idx = bytes = 0;
    msi.frame_type = 0;
}

void MPEGVideoParser::setPos(__int64 pos) {
    bytepos = pos;
}

void MPEGVideoParser::Parse(const void *pData, int len, DataVector *dst) {
    unsigned char *src = (unsigned char *)pData;

    // bytes is zero when the class is created,
    // thus (idx<bytes) is false the first time
    while(len>0) {
        if (idx<bytes) {
            int tc = bytes - idx;
            if (tc > len) tc = len;
            memcpy(buf+idx, src, tc);
            len -= tc; idx += tc; src += tc;

            if (idx>=bytes) {   // Finished?
                switch(header) {
                    case VIDPKT_TYPE_PICTURE_START:
                        msi.frame_type = (buf[1]>>3)&7;
                        msi.picture_structure = FRAME_PICTURE;
                        msi.bitflags = PICTUREFLAG_PF;
                        msi.fields = 2;
                        fPicturePending = true;
                        header = 0xFFFFFFFF;
                        break;

                    case VIDPKT_TYPE_SEQUENCE_START:
                        if (!fInitialized) {
    // xxxxxxxx xxxx.... ........ ........  horizontal_size
    // ........ ....xxxx xxxxxxxx ........  vertical_size
    // ........ ........ ........ xxxx....  aspect_ratio
    // ........ ........ ........ ....xxxx  frame_rate
							int frc = buf[3] & 15;	// frame rate code

                            width = (buf[0]<<4) + (buf[1]>>4);
                            height = ((buf[1]<<8)&0xf00) + buf[2];

							if (width && height && frc && width <= MPEG_MAX_WIDTH && height <= MPEG_MAX_HEIGHT) {
								aspect_ratio_info = buf[3] >> 4;

								switch(frc) {
								case 1:	mFrameRate = VDFraction(24000, 1001);   break;	// 1 (23.976) - NTSC FILM interlaced
								case 2:	mFrameRate = VDFraction(24   , 1   );   break;	// 2 (24.000) - FILM
								case 3:	mFrameRate = VDFraction(25   , 1   );   break;	// 3 (25.000) - PAL interlaced
								case 4:	mFrameRate = VDFraction(30000, 1001);   break;	// 4 (29.970) - NTSC color interlaced
								case 5:	mFrameRate = VDFraction(30   , 1   );   break;	// 5 (30.000) - NTSC b&w progressive
								case 6:	mFrameRate = VDFraction(50   , 1   );   break;	// 6 (50.000) - PAL progressive
								case 7:	mFrameRate = VDFraction(60000, 1001);   break;	// 7 (59.940) - NTSC color progressive
								case 8:	mFrameRate = VDFraction(60   , 1   );   break;	// 8 (60.000) - NTSC b&w progressive
								default:
//									throw MyError("MPEG video stream contains an invalid frame rate (%d).", buf[3] & 15);
									VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_UnknownFramerate, 1, &frc);
									mFrameRate = VDFraction(30000, 1001);
								}

                                // Initialize MPEG video decoder.
                                mpeg_initialize(width, height, NULL, NULL, 0);
                                dwidth = width;
                                dheight = height;
                                fInitialized = true;
                            }
                        }
                        header = 0xFFFFFFFF;
                        break;

                    case VIDPKT_TYPE_EXT_START:

                        if (bytes == 1) {
                            int ext = buf[0] >> 4;
                            if (ext == SEQUENCE_EXTENSION_ID) {
                                if (fInitialized && (!fFoundSequenceExt)) {
                                    bytes = 3;
                                    break;
                                }
                            } else if (ext == PICTURE_CODING_EXTENSION_ID) {
                                if (fPicturePending) {
                                    bytes = 5;
                                    break;
                                }
                            } else if (ext == SEQUENCE_DISPLAY_EXTENSION_ID) {
                                if (fInitialized && (!fFoundDisplayExt)) {
                                    bytes = 8;
                                    break;
                                }
                            }

                        } else if (bytes == 3) {
                            fFoundSequenceExt = true;
                            fIsMPEG2 = true;

            // xxxx.... ........ ........ SEQUENCE_EXTENSION_ID
            // ....x... ........ ........ escape_bit
            // .....xxx xxxx.... ........ MPEG-2 profile / level
            // ........ ....x... ........ progressive_sequence
            // ........ .....xx. ........ chroma_format
            // ........ .......x x....... horizontal_size_extension
            // ........ ........ .xx..... vertical_size_extension

                            progressive_sequence = !!(buf[1]&8);
                            width |= ((buf[1]&1)<<13) | ((buf[2]&0x80)<<5);
                            height |= (buf[2]&0x60)<<7;
							chroma_format = (buf[1]&6)>>1;
                        // frame rate extensions are not allowed
                        // in any defined MPEG-2 profile
                            mpeg_initialize(width,height,(char *)buf,NULL,0);

                        } else if (bytes == 5) {

// xxxx.... ........ ........ ........ ........ PICTURE_CODING_EXTENSION_ID
// ....xxxx xxxxxxxx xxxx.... ........ ........ f_codes
// ........ ........ ....xx.. ........ ........ intra_dc_precision
// ........ ........ ......xx ........ ........ picture_structure
// ........ ........ ........ x....... ........ top_field_first
// ........ ........ ........ .x...... ........ frame_pred_frame_dct
// ........ ........ ........ ..x..... ........ concealment_vectors
// ........ ........ ........ ...x.... ........ q_scale_type
// ........ ........ ........ ....x... ........ intra_vlc_format
// ........ ........ ........ .....x.. ........ alternate_scan
// ........ ........ ........ ......x. ........ repeat_first_field
// ........ ........ ........ .......x ........ chroma_420_type
// ........ ........ ........ ........ x....... progressive_frame

                            msi.picture_structure = (buf[2] & 3);
                            if (buf[3] & 0x80) msi.bitflags |= PICTUREFLAG_TFF;
                            if (buf[3] & 0x02) msi.bitflags |= PICTUREFLAG_RFF;
                            if (!(buf[4] & 0x80)) msi.bitflags &= ~PICTUREFLAG_PF;

                            // this is overly complicated
                            if ((msi.bitflags & PICTUREFLAG_RFF)
                              && (msi.picture_structure == 3)
                              && (msi.bitflags & PICTUREFLAG_PF)) {
                                if (this->progressive_sequence) {
                                    msi.fields = (msi.bitflags
                                      & PICTUREFLAG_TFF)? 6: 4;
                                } else {
                                    msi.fields = 3;
                                }
                            } else if (msi.picture_structure != FRAME_PICTURE) {
                                msi.fields = 1;
                            } else {
                                msi.fields = 2;
                            }

                        } else if (bytes == 8) {
// xxxx.... (3?) ........ ........ ........ ........ SEQUENCE_DISPLAY_EXTENSION_ID
// ....xxx. (3?) ........ ........ ........ ........ video_format
// .......? (3?) ........ ........ ........ ........ color_description
// ........ (3?) xxxxxxxx xxxxxx.. ........ ........ display_horizontal_size
// ........ (3?) ........ ......1. ........ ........ marker_bit
// ........ (3?) ........ .......x xxxxxxxx xxxxx... display_vertical_size
                            int i = (buf[0] & 1)? 4: 1;
                            dwidth = (buf[i] << 6) | (buf[i+1] >> 2);
                            dheight = ((buf[i+1] & 1) << 13) |
                              (buf[i+2] << 5) | (buf[i+3] >> 3);
                            fFoundDisplayExt = true;
                        }

                        header = 0xFFFFFFFF;
                        break;
                }
            }
            continue;
        }

        // Look for a valid MPEG header
        header = (header<<8) + *src++;
        len--;

        if ((header&0xffffff00) == 0x00000100) {
            header &= 0xff;

            if (fPicturePending && (
              header<VIDPKT_TYPE_SLICE_START_MIN ||
              header>VIDPKT_TYPE_SLICE_START_MAX) &&
              header != VIDPKT_TYPE_USER_START &&
              header != VIDPKT_TYPE_EXT_START) {
                // Don't add pictures until we've parsed the sequence!
                // Fix for Shakira: don't add picture if it is corrupt
                if (fFoundSequenceStart) {
					bool valid;

					if (ReferenceCount >= 2) {
						valid = (msi.frame_type == MPEG_FRAME_TYPE_I
							  || msi.frame_type == MPEG_FRAME_TYPE_P
							  || msi.frame_type == MPEG_FRAME_TYPE_B);

					} else {
						switch (msi.frame_type) {

							case MPEG_FRAME_TYPE_I:
								ReferenceCount++;
								valid = true;
								break;

							case MPEG_FRAME_TYPE_P:
								valid = (ReferenceCount >= 1);
								if (valid) ReferenceCount++;
								break;

							default:
								valid = false;
						}
					}

                    if (valid) {
                        if (fSeqAttach) msi.bitflags |= PICTUREFLAG_ATTACH;
                        msi.size = (int)(bytepos + (src - (unsigned char *)pData) - 4 - msi.stream_pos);
						msi.PTS = -1;
                        dst->Add(&msi);
                    }
                }
                fPicturePending = false;
                fSeqAttach = false;
            }

            switch(header) {
            case VIDPKT_TYPE_SEQUENCE_START:
                fSeqAttach = true;
                msi.stream_pos = bytepos + (src - (unsigned char *)pData) - 4;
                if (!fFoundSequenceStart) {
                    fFoundSequenceStart = true;
                    bytes = 4; idx = 0;
                } else header = 0xFFFFFFFF;
                break;

            case VIDPKT_TYPE_PICTURE_START:
                if (!fSeqAttach) {
                    msi.stream_pos = bytepos + (src - (unsigned char *)pData) - 4;
                }
                bytes = 2; idx = 0;
                break;

            case VIDPKT_TYPE_EXT_START:
                bytes = 1; idx = 0;
                break;

            default:
                header = 0xFFFFFFFF;
                break;
            }
        }
    }

    bytepos += src - (unsigned char *)pData;
}

//////////////////////////////////////////////////////////////////////////
//
//
//                         MPEGFileParser
//
//
//////////////////////////////////////////////////////////////////////////

extern HWND g_hWnd;

class MPEGFileParser {
private:
    static const unsigned short MPEG_AUDIO_DIALOG[];
    static const char szME[];

    __int64 video_stream_pos;
    __int64 audio_stream_pos;

    sint64 last_dts[52];

    enum {
        SCAN_BUFFER_SIZE    = 65536,
        SEARCH_SYSTEM_LIMIT = 200000, //(70*1024),
        SEARCH_AUDIO_LIMIT  = 256
    };

    unsigned char *packet_buffer;
    unsigned char *pScanBuffer;
    unsigned char *pScan;
    unsigned char *pScanLimit;
    __int64 i64ScanCpos;

    VDFile  *mFile;
//    int     fh;
    int     substream;
	int		vidstream;
    bool    fIsVCD;
    bool    fAbort;
    bool    fIsValidMPEG;
    long    lFirstCode;
    HWND    hWndStatus;

    __int64 file_len;

    void    Rewind();
    int     NextStartCode();
    bool    Fill();
    bool    Skip(int bytes);

    int     Read();
    int     Read(void *, int);

    __int64 Tell();

    long    ScanAudio();
    bool    ParsePack(int stream_id);

    static BOOL CALLBACK MpegAudioDlg(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
    static BOOL CALLBACK ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
    MPEGFileParser();
    ~MPEGFileParser();

    DataVector *video_stream_blocks;
    DataVector *video_stream_samples;
    DataVector *audio_stream_blocks;
    DataVector *audio_stream_samples;

    MPEGVideoParser *videoParser;
    MPEGAudioParser *audioParser;

    int width, height;
    VDFraction frame_rate;
    bool fIsMPEG2;
    bool fIsPartial;
    bool fInterleaved;
    bool fIsScrambled;
    bool progressive_sequence;
    int aspect_ratio_info;
    long audio_first_header;
    int AudioType;

    void Parse(VDFile *handle);
};

const char MPEGFileParser::szME[]="MPEG File Parser";

MPEGFileParser::MPEGFileParser() {
    // Clear variables
    pScan = pScanLimit = pScanBuffer = NULL;
    video_stream_blocks = NULL;
    video_stream_samples = NULL;
    audio_stream_blocks = NULL;
    audio_stream_samples = NULL;
    video_stream_pos = 0;
    audio_stream_pos = 0;
    packet_buffer = NULL;
    i64ScanCpos = 0;
    videoParser = NULL;
    audioParser = NULL;
    hWndStatus = NULL;
	vidstream = 0;
    fIsScrambled = false;
    fIsPartial = true;
    fIsVCD = false;
    fAbort = false;
	mFile = NULL;
//    fh = -1;
}

MPEGFileParser::~MPEGFileParser() {
	if (hWndStatus) {
		EnableWindow(GetParent(hWndStatus), TRUE);
		DestroyWindow(hWndStatus);
	}
    delete videoParser;
    delete audio_stream_samples;
    delete audio_stream_blocks;
    delete video_stream_samples;
    delete video_stream_blocks;
    delete audioParser;
    delete packet_buffer;
    delete pScanBuffer;
}

void MPEGFileParser::Rewind() {
    pScan = pScanLimit = pScanBuffer;
    i64ScanCpos = (fIsVCD)? 40: 0;
//    if (fh >= 0) _lseeki64(fh, i64ScanCpos, SEEK_SET);
    mFile->seek(i64ScanCpos);
}

int MPEGFileParser::NextStartCode() {
    int i;

    do {
        if ((i = Read()) == 0) {
            if ((i = Read()) == 0) {
                do {
                    i = Read();
                } while (i == 0);
                if (i == 1) {
                    if ((i = Read()) != EOF) {
                        return (i | 0x100);
                    }
                }
            }
        }
    } while (i != EOF);

    return 0;
}

bool MPEGFileParser::Fill() {
    // If EOF, pScanLimit == pScanBuffer
    // If disk error return false
    int actual;

    i64ScanCpos += (pScanLimit - pScanBuffer);
    pScan = pScanLimit = pScanBuffer;

    if (fIsVCD) {
        actual = mFile->readData(pScanBuffer, 20);
        //actual = _read(fh, pScanBuffer, 20);
        if (actual >= 0) {
            if (actual >= 20) {
                i64ScanCpos += 20;
                //actual = _read(fh, pScanBuffer, 2332);
                actual = mFile->readData(pScanBuffer, 2332);
            } else actual = 0;
        }
    } else {
        actual = mFile->readData(pScanBuffer, SCAN_BUFFER_SIZE);
//        actual = _read(fh, pScanBuffer, SCAN_BUFFER_SIZE);
    }
    if (actual < 0) return false;

    pScanLimit = pScanBuffer + actual;
    return true;
}

bool MPEGFileParser::Skip(int bytes) {
    if (bytes > 0) {
        int tc = pScanLimit - pScan;
        do {
            if (tc >= bytes) {
                pScan += bytes;
                bytes = 0;
                break;
            }
            bytes -= tc;
            if (!Fill()) break;
            tc = pScanLimit - pScanBuffer;
        } while (tc > 0);
    }
    return (bytes <= 0);
}

int MPEGFileParser::Read() {
    if (pScan < pScanLimit) return *pScan++;
    unsigned char c;
    if (Read(&c, 1) == 1) return c;
    return EOF;
}

int MPEGFileParser::Read(void *buffer, int bytes) {
    int tc, actual;
    int total = 0;

    if (bytes > 0) {

        // Start by emptying the buffer
        if (pScan < pScanLimit) {
            tc = pScanLimit - pScan;
            if (tc >= bytes) {
                memcpy(buffer, pScan, bytes);
                pScan += bytes;
                return bytes;
            }
            memcpy(buffer, pScan, tc);
            buffer = (char *)buffer + tc;
            bytes -= tc;
            total = tc;
        }

        // Now:
        //  bytes is greater than zero
        //  the scan buffer is empty
        i64ScanCpos += (pScanLimit - pScanBuffer);
        pScan = pScanLimit = pScanBuffer;

        // Read chunks directly until bytes <= chunksize
        tc = (fIsVCD)? 2332: SCAN_BUFFER_SIZE;

        while (bytes > tc) {
            if (fIsVCD) {
                actual = mFile->readData(pScanBuffer, 20);
//                actual = _read(fh, pScanBuffer, 20);
                if (actual < 20) {
                    if (actual < 0)
                        throw MyError("%s: read error", szME);
                    return total;
                }
                i64ScanCpos += 20;
            }
            actual = mFile->readData(buffer, tc);
            //actual = _read(fh, buffer, tc);
            if (actual < tc) {
                if (actual < 0)
                    throw MyError("%s: read error", szME);
                pScan = pScanLimit = (pScanBuffer + actual);
                return (total + actual);
            }
            buffer = (char *)buffer + tc;
            i64ScanCpos += tc;
            total += tc;
            bytes -= tc;
        }

        // Now, bytes <= chunksize
        if (fIsVCD) {
            actual = mFile->readData(pScanBuffer, 20);
            //actual = _read(fh, pScanBuffer, 20);
            if (actual < 20) {
                if (actual < 0)
                    throw MyError("%s: read error", szME);
                return total;
            }
            i64ScanCpos += 20;
        }
        actual = mFile->readData(pScanBuffer, tc);
        //actual = _read(fh, pScanBuffer, tc);
        if (actual < 0)
            throw MyError("%s: read error", szME);

        if (actual > 0) {
            pScanLimit = (pScanBuffer + actual);
            if (actual > bytes) actual = bytes;
            memcpy(buffer, pScanBuffer, actual);
            pScan += actual;
            total += actual;
        }
    }
    return total;
}

__int64 MPEGFileParser::Tell() {
    return i64ScanCpos + (pScan - pScanBuffer);
}

// Create MPEG audio dialog in memory
const unsigned short MPEGFileParser::MPEG_AUDIO_DIALOG[] = {
    // DLGTEMPLATE
    0x0040, 0x00C0,     // style (DS_SETFONT | WS_CAPTION)
    0, 0,               // dwExtendedStyle
    3,                  // cdit
    0, 0, 256, 96,      // x, y, cx, cy
    0, 0,               // menu, class
    'M','u','l','t','i','p','l','e',' ','A','u','d','i','o',' ',
    'S','t','r','e','a','m','s',' ','F','o','u','n','d',' ',0,
    8,                  // pointsize
    'M','S',' ','S','a','n','s',' ','S','e','r','i','f',0,

    // DLGITEMTEMPLATE
    0x0001, 0x5002,     // style (WS_CHILD | WS_VISIBLE | WS_GROUP | SS_CENTER)
    0, 0,               // dwExtendedStyle
    80, 4, 96, 8,       // x, y, cx, cy
    -1,                 // id (IDC_STATIC)
    0xFFFF, 0x0082,     // class (Static)
    'C','h','o','o','s','e',' ','a','n',' ',
    'a','u','d','i','o',' ','s','t','r','e','a','m',':',0,
    0,                  // data

    // DLGITEMTEMPLATE
    0x0103, 0x50A1,     // style (WS_CHILD | WS_VISIBLE | WS_BORDER |
    // WS_VSCROLL | WS_TABSTOP | LBS_NOINTEGRALHEIGHT | LBS_SORT | LBS_NOTIFY)
    0, 0,               // dwExtendedStyle
    4, 16, 248, 56,     // x, y, cx, cy
    IDOK+1,             // id
    0xFFFF, 0x0083,     // class (List Box)
    'L','i','s','t','1',0,
    0,                  // data

    // DLGITEMTEMPLATE
    0x0001, 0x5001,     // style (WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_DEFPUSHBUTTON)
    0, 0,               // dwExtendedStyle
    100, 76, 56, 16,    // x, y, cx, cy
    IDOK,               // id
    0xFFFF, 0x0080,     // class (Button)
    'O','K',0,
    0                   // data
};

BOOL CALLBACK MPEGFileParser::MpegAudioDlg(HWND hDlg, UINT uMsg,
  WPARAM wParam, LPARAM lParam) {

    switch (uMsg) {

    case WM_INITDIALOG:
        {
            TCHAR buf[64];
            unsigned char *src = (unsigned char *)lParam;
            HWND hwndList = GetDlgItem(hDlg, IDOK+1);

            for (int i=0; i<64; i++) {
                if (src[i] != ATYPE_NONE) {
                    bool valid = true;
                    switch (src[i]) {
                        case ATYPE_MPEG:
                            wsprintf(buf, "MPEG Audio Stream 0x%X", 0xC0+i);
                            break;
                        case ATYPE_AC3:
                            wsprintf(buf, "Dolby AC-3 Audio (substream 0x8%X)", i-32);
                            break;
                        case ATYPE_LPCM:
                            wsprintf(buf, "Linear PCM Audio (substream 0xA%X)", i-48);
                            break;
                        default:
                            valid = false;
                    }
                    if (valid)
                        SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
                }
            }
            SendMessage(hwndList,LB_SETCURSEL,0,0);
            return TRUE;
        }

    case WM_COMMAND:
        if (LOWORD(wParam) == IDOK) {
            HWND hwndList = GetDlgItem(hDlg, IDOK+1);
            int i = SendMessage(hwndList, LB_GETCURSEL, 0, 0);
            EndDialog(hDlg, i);
            return TRUE;
        }
        break;
    }

    return FALSE;
}

long MPEGFileParser::ScanAudio() {
    long i,count,pkSize;
    unsigned long hdr;
    unsigned char audlist[64];

    // audlist[0  .. 31] = MPEG Audio
    // audlist[32 .. 47] = AC3
    // audlist[48 .. 63] = LPCM

    for (i=0; i<64; i++) audlist[i] = ATYPE_NONE;

    hdr = 0xFFFFFFFF;
    fIsVCD = false;
    fIsMPEG2 = false;
    fInterleaved = false;
    fIsValidMPEG = false;

    Rewind();

    // hdr = .. .. .. 00, count = 0
    // hdr = .. .. 00 00, count = 1
    // hdr = .. 00 00 01, count = 2
    // hdr = 00 00 01 BA, count = 3

    // Search for system layer
    // Emotion.m2v has empty PES headers (0x1E0) but no system!

    lFirstCode = -1;

    count = 4;
    Read(&hdr, 4);
    if (hdr == 'FFIR') {
        fIsVCD = true;
        Rewind();
        Read(&hdr, 4);
    }

    while (count < SEARCH_SYSTEM_LIMIT) {

        if (hdr == 0xBA010000) {
            // Verify packet_header
            Read(&hdr, 4);
            count += 4;
            if ((hdr & 0x000400C4) == 0x00040044) {
                //  MPEG-2 (bswapped):
                // 0: ........ .....1.. ........ 01...1..
                // 4: ........ ........ .......1 .....1..
                // 8: .....ppp ......11
                Read(&hdr, 4);
                count += 4;
                if ((hdr & 0x00000104) == 0x00000104) {
                    hdr = (hdr >> 8) | (Read() << 24);
                    hdr = (hdr >> 8) | (Read() << 24);
                    count += 2;
                    if ((hdr & 0x00030000) == 0x00030000) {
                        Skip((hdr >> 24) & 7);
                        fInterleaved = true;
						fIsMPEG2 = true;
                        lFirstCode = count-14;
                        break;
                    }
                }

            } else if ((hdr & 0x000100F1) == 0x00010021) {
            //  MPEG-1 (bswapped):
            // 0: ........ .......1 ........ 0010...1
            // 4: .......1 ........ 1....... .......1
                Read(&hdr, 4);
                count += 4;
                if ((hdr & 0x01008001) == 0x01008001) {
                    fInterleaved = true;
                    lFirstCode = count-12;
                    break;
                }
            }

        } else {
            // if we hit a sequence_header, mark it for later
            if (hdr == 0xB3010000) {
                if (lFirstCode < 0) lFirstCode = count-4;
            }
            hdr = (hdr >> 8) | (Read() << 24);
            ++count;
        }
    }

    if (lFirstCode < 0) return -1;

    if (!fInterleaved) {
        // No system found; check for MPEG2
        if (!fIsVCD) {      // can't be VCD
            Rewind();
            Skip(lFirstCode+4);
            // aaaaffff hhhhhhhh wwwwhhhh wwwwwwww
            Read(&hdr, 4);
            if ((hdr & 0x0000F0FF) && (hdr & 0x00FF0F00)
             && (hdr & 0xF0000000) && (hdr & 0x0F000000)) {
                fIsValidMPEG = true;
				if (NextStartCode() == 0x1B5) {
					fIsMPEG2 = ((Read() & 0xF0) == 0x10);
				}
            }
        }
        return -1;
    }

    fIsValidMPEG = true;

    // Search for audio packets
    for (count=0; count < SEARCH_AUDIO_LIMIT; count++) {

        hdr = 0xFFFFFFFF;
        for (pkSize = 0; pkSize < SEARCH_SYSTEM_LIMIT; pkSize++) {
            if ((i = Read()) == EOF) {
                pkSize = SEARCH_SYSTEM_LIMIT;
                break;
            }
            hdr = (hdr << 8) + i;
            if ((hdr & 0xFFFFFF00) == 0x100)
                if (hdr >= 0x1B9) break;
        }
        if (pkSize >= SEARCH_SYSTEM_LIMIT) break;

        if (hdr == 0x1BA) {
            if ((i = Read()) == EOF) break;
            if ((i & 0xC0) == 0x40) {
                this->fIsMPEG2 = true;
                Skip(8);
                if ((i = Read()) == EOF) break;
                Skip(i & 7);
            } else Skip(7);
        } else {
            if (hdr == 0x1B9) break;

            pkSize = Read();
            pkSize = (pkSize << 8) | Read();
            if ((i = Read()) == EOF) break;
            pkSize--;

            if (hdr == 0x1BD) {
                // private_stream_1
                if ((i & 0xC0) == 0x80) {
                    // MPEG-2
                    if (i & 0x30) fIsScrambled = true;
                    Read();
                    i = Read();
                    Skip(i);
                    pkSize -= (i+2);
                } else {
                    // MPEG-1
                    while (i & 0x80) {
                        if (--pkSize <= 0) break;
                        i = Read();
                    }
                    if (i == EOF) break;
                    if (i & 0x40) {
                        Read();
                        i = Read();
                        pkSize -= 2;
                    }
                    if (i < 0x40) {
                        if (i & 0x20) {
                            if (i & 0x10) i = 9; else i = 4;
                            pkSize -= i;
                            Skip(i);
                        }
                    }
                }
                int ss = Read();	// substream
                if (ss == EOF) break;
                pkSize--;

                if ((ss & 0xF8) == 0x80) {
                    // Probably AC3, but let's not assume...
                    // (DTS Audio has a syncword of 7FFF8001)
                    // 1 byte = # of syncwords (access units) in packet
                    // 2 bytes = 1-based offset of first syncword
                    pkSize--;
                    if (Read() > 0) {   // # of access units
                        i = Read();
                        i = (i<<8) + Read();    // offset
                        pkSize -=2;
            // Say pkSize is 7 here (# of bytes remaining in pack)
            //      ### ohi olo xxx xxx xxx xxx xxx xxx xxx
            //     offset:       1   2   3   4   5   6   7
            // Thus offset must be greater than 0 and less than 7
                        if (i > 0 && i < pkSize) {
                          if (i>1) Skip(i-1);
                          pkSize -= i;
                          if (Read()==0x0B) {
                            pkSize--;
                            if (Read()==0x77)
                              // verified
                              audlist[(ss & 15) + 32] = ATYPE_AC3;
                            }
                        }
                    }
                } else if ((ss & 0xF8) == 0xA0) {
        // LPCM
        // 1 byte = # of frames in packet
        // 2 bytes = 1-based offset of first frame
        // 1 byte:  x.......    emphasis
        //          .x......    mute
        //          ..x.....    reserved
        //          ...xxxxx    frame number % 20
        // 1 byte:  xx......    00=16-,01=20-,10=24-bit (11=forbidden)
        //          ..xx....    00=48,01=96 KHz (others reserved?)
        //          ....x...    reserved
        //          .....xxx    number of channels - 1
        // 1 byte:  xxxxxxxx    dynamic range control

                    Skip(4);
                    pkSize -=5;
                    if ((i = Read()) == EOF) break;
        // For now, we can only support 16-bit 48K mono or stereo
                    if ((i & 0xF6) == 0)
                        audlist[(ss & 15) + 48] = ATYPE_LPCM;
                }

            } else if (hdr >= 0x1C0 && hdr <= 0x1DF) {
                // MPEG Audio
                if ((i & 0x30) && fIsMPEG2) fIsScrambled = true;
                audlist[hdr & 31] = ATYPE_MPEG;
            }
            Skip(pkSize);
        }

    }

    // If only one audio stream, return that
    hdr = -1;
    count = 0;
    for (i=0; i<64; i++) {
        if (audlist[i] != ATYPE_NONE) {
            if (hdr == -1) hdr = i;
            count++;
        }
    }
    if (count <= 1) return (long)hdr;

    // Let the user select an audio stream
    hdr = DialogBoxIndirectParam(GetModuleHandle(NULL),
        (LPCDLGTEMPLATE)MPEG_AUDIO_DIALOG, g_hWnd,
        MpegAudioDlg, (LPARAM)audlist);

    count = 0;
    for (i=0; i<64; i++) {
        if (audlist[i] != ATYPE_NONE) {
            if ((long)hdr == count) break;
            count++;
        }
    }
    if (i >= 64) i = -1;
    return i;
}

bool MPEGFileParser::ParsePack(int stream_id) {
    // Parse one elementary stream packet
    // Return "false" if packet is incomplete
    int i, actual, ptsflags;
    MPEGPacketInfo mpi;
    int pack_length = 0;
    sint64 pts, dts, tagpos;
    bool complete = false;

    switch (stream_id) {

    case 0x1BA:
        pack_length = 0;
        if ((i = Read()) != EOF) {
            if ((i & 0xC0) == 0x40) {
                // MPEG-2
                if (Skip(8)) {
                    if ((i = Read()) != EOF) {
                        if (Skip(i & 7)) complete = true;
                    }
                }
            } else if ((i & 0xF0) == 0x20) {
                // MPEG-1
                if (Skip(7)) complete = true;
            }
        }
        break;

    default:
        tagpos = Tell();
		mpi.PTS = -1;

        if (fInterleaved) {
            pack_length = Read();
            pack_length = (pack_length << 8) + Read();
            if (pack_length == 0) {
                complete = true;
                break;
            }

            if ((stream_id >= 0x1C0 && stream_id <= 0x1EF) || stream_id == 0x1BD) {
                // parse stream header
                i = Read();
                --pack_length;

                if ((i & 0xC0) == 0x80) {

                    // MPEG-2 packet
                    if (i & 0x30) fIsScrambled = true;
                    ptsflags = Read();  // PTS/DTS flags
                    i = Read();         // PES_header_data_length
                    pack_length -= 2;

                    if (ptsflags & 0x80) {  // PTS follows
                        //  0010ppp1 pppppppp ppppppp1 pppppppp ppppppp1
                        pts = (Read() & 0x0E) >> 1;
                        pts = (pts << 8) + Read();
                        pts = (pts << 7) + (Read() >> 1);
                        pts = (pts << 8) + Read();
                        pts = (pts << 7) + (Read() >> 1);
                        pack_length -= 5;
                        i -= 5;
						mpi.PTS = pts;
                        if (ptsflags & 0x40) {  // DTS follows
                            // 0001ddd1 dddddddd ddddddd1 dddddddd ddddddd1
                            dts = (Read() & 0x0E) >> 1;
                            dts = (dts << 8) + Read();
                            dts = (dts << 7) + (Read() >> 1);
                            dts = (dts << 8) + Read();
                            dts = (dts << 7) + (Read() >> 1);
                            pack_length -= 5;
                            i -= 5;
                        } else dts = pts;
                    }
                    if (i > 0) {
                        Skip(i);
                        pack_length -= i;
                    }

                } else {
                    // MPEG-1 packet
                    while (i & 0x80) {
                        if (--pack_length <= 0) break;
                        i = Read();
                    }
                    if (pack_length <= 0) break;
                    if (i & 0x40) {
                        Skip(1);    // P-STD_buffer size
                        i = Read();
                        pack_length -= 2;
                    }
                    ptsflags = i << 2;
                    if (ptsflags & 0x80) {  // PTS follows
                        //  0010ppp1 pppppppp ppppppp1 pppppppp ppppppp1
                        pts = (i & 0x0E) >> 1;
                        pts = (pts << 8) + Read();
                        pts = (pts << 7) + (Read() >> 1);
                        pts = (pts << 8) + Read();
                        pts = (pts << 7) + (Read() >> 1);
                        pack_length -= 4;
						mpi.PTS = pts;
                        if (ptsflags & 0x40) {  // DTS follows
                            // 0001ddd1 dddddddd ddddddd1 dddddddd ddddddd1
                            dts = (Read() & 0x0E) >> 1;
                            dts = (dts << 8) + Read();
                            dts = (dts << 7) + (Read() >> 1);
                            dts = (dts << 8) + Read();
                            dts = (dts << 7) + (Read() >> 1);
                            pack_length -= 5;
                        } else dts = pts;
                    }
                }

                if (ptsflags & 0x80) {
                    sint64 last_stream_dts = last_dts[stream_id - 0x1BD];
                    if (last_stream_dts != -1) {
                        // decoding timestamps should never run backwards
                        // and must occur at least every 0.7s (90KHz clock)
                        if ((dts - last_stream_dts) > 63000) {
                            const wchar_t *pStreamType = (stream_id < 0x1E0)? L"audio" : L"video";
                            const int nStream = (stream_id - 0x1C0) & 0x1F;
                            const sint64 nPos = tagpos;
                            VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_OOOTimestamp, 5, &pStreamType, &nStream, &tagpos, &last_stream_dts, &dts);
                        }
                    }
                    last_dts[stream_id - 0x1BD] = dts;
                }
            }
            if (pack_length <= 0) {
                if (pack_length == 0) complete = true;
                break;
            }

        } else {
            // not interleaved
            stream_id = 0x1E0;
            pack_length = 65536;
        }
        
        if (stream_id == 0x1BD) {
            // private_stream_1, AC3 or LPCM
            i = Read();
            --pack_length;
            if ((i & 0xF0) == 0x80) {
                if (audioParser == NULL) {
                    if (audioParser = new AC3AudioParser()) {
                        AudioType = ATYPE_AC3;
                        substream = i;
                    }
                }

                if (i == substream && audioParser) {
    // 1 byte  = # of syncwords (i.e., access units) in packet
    // 2 bytes = 1-based offset of first syncword (0x0B77)
                    Skip(3);
                    pack_length -= 3;

                    mpi.file_pos = Tell();
                    mpi.stream_pos = audio_stream_pos;
                    audio_stream_blocks->Add(&mpi);

                    actual = Read(packet_buffer, pack_length);
                    if (actual >= pack_length) complete = true;
                    audioParser->Parse(packet_buffer, actual, audio_stream_samples);
                    audio_stream_pos += actual;
                    pack_length = 0;
                }
            } else if ((i & 0xF0) == 0xA0) {
                if (audioParser == NULL) {
                    if (audioParser = new PCMAudioParser()) {
                        AudioType = ATYPE_LPCM;
                        substream = i;
                    }
                }
                if (i == substream && audioParser) {
                    if (audioParser->getHeader() == 0) {
                        char buf[6];
                        Read(buf,6);
                        audioParser->Parse(buf, 0, NULL);
                    } else Skip(6);
                    pack_length -= 6;

                    mpi.file_pos = Tell();
                    mpi.stream_pos = audio_stream_pos;
                    audio_stream_blocks->Add(&mpi);

                    actual = Read(packet_buffer, pack_length);
                    if (actual >= pack_length) complete = true;
                    audioParser->Parse(NULL, actual, audio_stream_samples);
                    audio_stream_pos += actual;
                    pack_length = 0;
                }
            }

        } else if (stream_id >= 0x1E0 && stream_id <= 0x1F0) {
            // MPEG video (only the first stream found)
			if (!vidstream) vidstream = stream_id;
			if (stream_id == vidstream) {
				if (fInterleaved) {
					mpi.file_pos = Tell();
					mpi.stream_pos = video_stream_pos;
					video_stream_blocks->Add(&mpi);
				}
				actual = Read(packet_buffer, pack_length);
				if (actual >= pack_length) complete = true;
				videoParser->Parse(packet_buffer, actual, video_stream_samples);
				video_stream_pos += actual;
				pack_length = 0;
			}

        } else if (stream_id >= 0x1C0 && stream_id <= 0x1DF) {
            // MPEG audio
            i = (stream_id & 0xFF);
            if (audioParser == NULL) {
                if (audioParser = new MPEGAudioParser()) {
                    AudioType = ATYPE_MPEG;
                    substream = i;
                }
            }
            if (i == substream && audioParser) {
                mpi.file_pos = Tell();
                mpi.stream_pos = audio_stream_pos;
                audio_stream_blocks->Add(&mpi);

                actual = Read(packet_buffer, pack_length);
                if (actual >= pack_length) complete = true;
                audioParser->Parse(packet_buffer, actual, audio_stream_samples);
                audio_stream_pos += actual;
                pack_length = 0;
            }
        }
    }

    if (pack_length > 0) return Skip(pack_length);
    return complete;
}

void MPEGFileParser::Parse(VDFile *handle) {
    int i;

    if ((this->mFile = handle) == NULL)
        throw MyError("%s: invalid file handle", szME);

    // Determine the file's size
    file_len = mFile->size();

    if (file_len < 0)
        throw MyError("%s: couldn't determine file size", szME);

    pScanBuffer = new unsigned char[SCAN_BUFFER_SIZE];
    if (pScanBuffer == NULL) throw MyMemoryError();

    // Clear all DTS to -1 (an impossible number)
    memset(last_dts, 255, 52*sizeof(sint64));

    // Detect Audio type...
    // (fIsValidMPEG, fInterleaved, and fIsVCD are initialized here)
    substream = ScanAudio();

    if (!fIsValidMPEG)
        throw MyError("%s: couldn't determine MPEG type", szME);

    if (substream >= 48) {
        AudioType = ATYPE_LPCM;
        audioParser = new PCMAudioParser();
        substream += (0xA0 - 48);
    } else if (substream >= 32) {
        AudioType = ATYPE_AC3;
        audioParser = new AC3AudioParser();
        substream += (0x80 - 32);
    } else if (substream >= 0) {
        AudioType = ATYPE_MPEG;
        audioParser = new MPEGAudioParser();
        substream += 0xC0;
    } else {
        AudioType = ATYPE_NONE;
        audioParser = NULL;
        substream = -1;
    }

    // Begin file parsing!  This is a royal bitch!

    try {
        packet_buffer = new unsigned char[65536];

        video_stream_blocks = new DataVector(sizeof MPEGPacketInfo);
        video_stream_samples = new DataVector(sizeof MPEGSampleInfo);
        video_stream_pos = 0;

        audio_stream_blocks = new DataVector(sizeof MPEGPacketInfo);
        audio_stream_samples = new DataVector(sizeof MPEGSampleInfo);
        audio_stream_pos = 0;

        videoParser = new MPEGVideoParser();

        bool end_of_file = false;

        try {
            Rewind();

            // pop up the dialog...
            if (!(hWndStatus = CreateDialogParam(GetModuleHandle(NULL),
              MAKEINTRESOURCE(IDD_PROGRESS), g_hWnd, ParseDialogProc, (LPARAM)this)))
                throw MyMemoryError();

            Skip(lFirstCode);

            do {
                guiDlgMessageLoop(hWndStatus);
                if (fAbort) break;

                if (fInterleaved)
                    i = NextStartCode();
                else i = 0x1E0;

                switch (i) {
                    // One for audio and for video?

                    case VIDPKT_TYPE_SEQUENCE_END:
                    case 0x1B9:     // ISO 11172 end code
                        fIsPartial = false;
                    case 0:         // no more start codes
                        end_of_file = true;
                        break;

                    default:
                        if (i > 0x1B9)
                            end_of_file = !ParsePack(i);
                }
                
            } while (!end_of_file);

        } catch(const MyError&) {
            throw;
        }

        // If the last frame is incomplete, this start
        // code will prevent it from being included...
        i = (VIDPKT_TYPE_USER_START << 24) | 0x00010000;
        videoParser->Parse(&i, 4, video_stream_samples);

        if (fInterleaved) {
            MPEGPacketInfo mpi;

            mpi.file_pos = 0;
            mpi.stream_pos = video_stream_pos;
            video_stream_blocks->Add(&mpi);

            if (audioParser) {
                mpi.file_pos = 0;
                mpi.stream_pos = audio_stream_pos;
                audio_stream_blocks->Add(&mpi);
            }
        }

    } catch(const MyError&) {
        if (hWndStatus) {
			EnableWindow(GetParent(hWndStatus), TRUE);
            DestroyWindow(hWndStatus);
            hWndStatus = NULL;
        }
        throw;
    }

    if (hWndStatus) {
		EnableWindow(GetParent(hWndStatus), TRUE);
        DestroyWindow(hWndStatus);
        hWndStatus = NULL;
    }
}

BOOL CALLBACK MPEGFileParser::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    MPEGFileParser *thisPtr = (MPEGFileParser *)GetWindowLong(hDlg, DWL_USER);

    switch(uMsg) {

    case WM_INITDIALOG:
        {
            TCHAR buf[64];

            SetWindowLong(hDlg, DWL_USER, lParam);
            thisPtr = (MPEGFileParser *)lParam;

            SendMessage(hDlg, WM_SETTEXT, 0, (LPARAM)"MPEG Import Filter");

            if (thisPtr->fIsVCD) {
                wsprintf(buf, "Parsing VideoCD stream");
            } else {
                if (thisPtr->fInterleaved) {
                    wsprintf(buf, "Parsing interleaved MPEG-%ld file",
                      (thisPtr->fIsMPEG2)? 2L: 1L);
                } else {
                    wsprintf(buf, "Parsing MPEG-%ld video file",
                      (thisPtr->fIsMPEG2)? 2L: 1L);
                }
            }
            SetDlgItemText(hDlg, IDC_STATIC_MESSAGE, (LPCTSTR)buf);

            SendMessage(GetDlgItem(hDlg, IDC_PROGRESS), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
            SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);

			EnableWindow(GetParent(hDlg), FALSE);
            return TRUE;
        }

    case WM_TIMER:
        {
            TCHAR buf[64];
            __int64 file_cpos = thisPtr->Tell();

            SendMessage(GetDlgItem(hDlg, IDC_PROGRESS), PBM_SETPOS,
              (WPARAM)((file_cpos*16384) / thisPtr->file_len), 0);

            if (thisPtr->fIsVCD) {
                wsprintf(buf, "%ld of %ld sectors",
                  (long)(file_cpos/2352),
                  (long)(thisPtr->file_len/2352));
            } else {
                wsprintf(buf, "%ldK of %ldK",
                  (long)(file_cpos>>10),
                  (long)((thisPtr->file_len+1023)>>10));
            }

            SendDlgItemMessage(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
        }
        return TRUE;

    case WM_COMMAND:
        if (LOWORD(wParam) == IDCANCEL)
            thisPtr->fAbort = true;
        return TRUE;
    }

    return 0;
}


//////////////////////////////////////////////////////////////////////////
//
//
//                          InputFileMPEG
//
//
//////////////////////////////////////////////////////////////////////////

const char InputFileMPEG::szME[]="MPEG Import Filter";

InputFileMPEG::InputFileMPEG() {
    // clear variables

//    fh = -1;
#ifdef USE_FIELD_MAP
	video_field_map = NULL;
#endif
    video_packet_buffer = NULL;
    video_packet_list = NULL;
    video_sample_list = NULL;
    audio_packet_list = NULL;
    audio_sample_list = NULL;
    audio_first_header = 0;

	firstPTS = -1;
	afirstPTS = -1;

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	//audioSrc = NULL;
// END ****************************************************************
    videoSrc = NULL;

    AudioType = ATYPE_NONE;

    iDecodeMode = 0;
    fAcceptPartial = true;
    fAllowDSC = false;
    fIsMPEG2 = false;

    pFastRead = NULL;

    last_packet[0] = last_packet[1] = 0;
}

InputFile *CreateInputFileMPEG() {
    return new InputFileMPEG();
}

void InputFileMPEG::Init(const wchar_t *szFile) {
    VDLogAppMessage(kVDLogMarker, kVDST_Mpeg, kVDM_OpeningFile, 1, &szFile);

    AddFilename(szFile);

    // see if we can open the file

    try {
        mFile.open(szFile, nsVDFile::kRead | nsVDFile::kDenyWrite | nsVDFile::kOpenExisting | nsVDFile::kSequential);

        pFastRead = new FastReadStream(mFile.getRawHandle(), 24, 32768);

        MPEGFileParser fileParser;
        fileParser.Parse(&mFile);

        if (fileParser.fIsPartial && !fAcceptPartial)
            //throw MyError("%s: incomplete MPEG", szME);
            VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_Incomplete);

        this->width = fileParser.videoParser->width;
        this->height = fileParser.videoParser->height;
		this->chroma_format = fileParser.videoParser->chroma_format;
        this->mFrameRate = fileParser.videoParser->mFrameRate;
        this->fIsMPEG2 = fileParser.videoParser->fIsMPEG2;
        this->fInterleaved = fileParser.fInterleaved;
        this->progressive_sequence = fileParser.videoParser->progressive_sequence;
        this->aspect_ratio = fileParser.videoParser->aspect_ratio_info;
        this->AudioType = fileParser.AudioType;

        // Construct stream and packet lookup tables.

        if (fileParser.fInterleaved) {
            video_packet_list = (MPEGPacketInfo *)
                fileParser.video_stream_blocks->MakeArray();
            packets = fileParser.video_stream_blocks->Length() - 1;

			if (packets <= 0)
				throw MyError("%s: no video stream!", szME);

            if (fileParser.audioParser) {
                audio_packet_list = (MPEGPacketInfo *)
                    fileParser.audio_stream_blocks->MakeArray();
                apackets = fileParser.audio_stream_blocks->Length() - 1;

                audio_sample_list = (MPEGSampleInfo *)
                    fileParser.audio_stream_samples->MakeArray();
                aframes = fileParser.audio_stream_samples->Length();

                audio_first_header = fileParser.audioParser->getHeader();
            }
        }

        video_sample_list = (MPEGSampleInfo *)
            fileParser.video_stream_samples->MakeArray();
        frames = fileParser.video_stream_samples->Length();

    } catch(const MyError&) {
        throw;
    }

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
    // initialize DubSource pointers
    //audioSrc = NULL;

	vdrefptr<AudioSource> audioSrc;
    switch (AudioType) {
        case ATYPE_MPEG:
            audioSrc = new AudioSourceMPEG(this);
            break;
        case ATYPE_AC3:
            audioSrc = new AudioSourceAC3(this);
            break;
        case ATYPE_LPCM:
            audioSrc = new AudioSourcePCM(this);
            break;
        default:
            audioSrc = NULL;
            break;
    }

    if (audioSrc) audioSrc->init();
    m_Sources.push_back(audioSrc);
// END ****************************************************************

    if (!PostInit()) throw MyMemoryError();

    videoSrc = new VideoSourceMPEG(this);
    videoSrc->init();
}

bool InputFileMPEG::PostInit() {
    long i, j, count;
    MPEGSampleInfo m;

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2004											***
// BEGIN **************************************************************
	vdrefptr<AudioSource> audioSrc;
	if(m_Sources.size() > 0) {

		VDASSERT(m_Sources.size() == 1);

		audioSrc = m_Sources.front();
	}
// END ****************************************************************

	// Assign Video Presentation Time Stamps
	if (fInterleaved) {
		count = 0;
		for (i = 0; i < packets; i++) {
			if (video_packet_list[i].PTS != -1) {
				for (j = count; j < frames; j++) {
					if (video_sample_list[j].stream_pos >= video_packet_list[i].stream_pos) {
						video_sample_list[j].PTS = video_packet_list[i].PTS;
						count = j + 1;
						break;
					}
				}
			}
		}

		if (audioSrc) {
			// Assign Audio Presentation Time Stamps
			count = 0;
			for (i = 0; i < apackets; i++) {
				if (audio_packet_list[i].PTS != -1) {
					for (j = count; j < aframes; j++) {
						if (audio_sample_list[j].stream_pos >= audio_packet_list[i].stream_pos) {
							audio_sample_list[j].PTS = audio_packet_list[i].PTS;
							count = j + 1;
							break;
						}
					}
				}
			}
		}
	}

    // Deal with non-progressive sequences as best we can

    if (!this->progressive_sequence) {

        // Count the number of fields, and check for field pictures
        j = count = 0;
        for (i = 0; i < frames; i++) {
            if (video_sample_list[i].picture_structure != FRAME_PICTURE) {
                j = 1;  // has field pictures = true
            }
            count += video_sample_list[i].fields;
        }

        if (j) {    // has field pictures

            // Combine fields into frames
            i = j = 0;
            while (i < frames) {
                video_sample_list[j++] = video_sample_list[i];

                if (i < frames - 1) {
                    /*** Gad this is such a hack ***/
                    if ((video_sample_list[i].picture_structure == TOP_FIELD &&
                         video_sample_list[i+1].picture_structure == BOTTOM_FIELD) ||
                        (video_sample_list[i].picture_structure == BOTTOM_FIELD &&
                         video_sample_list[i+1].picture_structure == TOP_FIELD)) {

                        if (video_sample_list[i].frame_type ==
                            video_sample_list[i+1].frame_type ||
                           (video_sample_list[i].frame_type == MPEG_FRAME_TYPE_I &&
                            video_sample_list[i+1].frame_type == MPEG_FRAME_TYPE_P)) {

                            video_sample_list[j-1].size =
                              (video_sample_list[i+1].stream_pos -
                               video_sample_list[i].stream_pos) +
                               video_sample_list[i+1].size;

                            i++;
                        }
                    }
                }
                i++;
            }
            frames = j;

        } else {
            // Detect 3:2 pulldown and adjust the frame rate
            if ((mFrameRate.getHi() == 30000) && (mFrameRate.getLo() == 1001)) {
                if (count > (long)(frames * 2.4))
                    mFrameRate = VDFraction(24000,1001);
            }
        }
    }

    // Bubble-sort MPEG frames into display order
    for (i = 1; i < frames; i++) {
        if (video_sample_list[i].frame_type == MPEG_FRAME_TYPE_B) {
            m = video_sample_list[i];
            video_sample_list[i] = video_sample_list[i-1];
            video_sample_list[i-1] = m;
        }
    }

	// Now that the video frames are in display order,
	// find the earliest video and audio timestamps
	if (fInterleaved) {
		for (i = 0; i < frames; i++) {
			if (video_sample_list[i].PTS != -1) {
				firstPTS = video_sample_list[i].PTS;
				break;
			}
		}
		afirstPTS = firstPTS;
		if (audioSrc) {
			for (i = 0; i < aframes; i++) {
				if (audio_sample_list[i].PTS != -1) {
					afirstPTS = audio_sample_list[i].PTS;
					break;
				}
			}
		}
	} else {
		firstPTS = afirstPTS = 0;
	}

#if 0
	{
		FILE *f = fopen("C:\\vdebug.txt", "wt");
		if (f != NULL) {
			for (i = 0; i < frames; i++) {
				fprintf(f, "frame %ld (", i);
				switch (video_sample_list[i].frame_type) {
					case MPEG_FRAME_TYPE_I:
						fprintf(f, "I");
						break;
					case MPEG_FRAME_TYPE_P:
						fprintf(f, "P");
						break;
					case MPEG_FRAME_TYPE_B:
						fprintf(f, "B");
						break;
					case MPEG_FRAME_TYPE_D:
						fprintf(f, "D");
						break;
					default:
						fprintf(f, "%ld", video_sample_list[i].frame_type);
						break;
				}
				fprintf(f, ") PTS = %I64d\n", video_sample_list[i].PTS);
			}
			fclose(f);
		}
	}
#endif

    // Set video_packet_buffer to size of the largest frame + 16
    count = 65520;
    for (i = 0; i < frames; i++) {
        if (video_sample_list[i].size > count)
            count = video_sample_list[i].size;
    }
    count = (count + 31) & 0xFFFFFFF0;

    if (video_packet_buffer = new char[count]) return true;
    return false;
}

InputFileMPEG::~InputFileMPEG() {
    _RPT0(0,"Destructor called\n");

//    delete videoSrc;
//    delete audioSrc;
	videoSrc = NULL;
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	//audioSrc = NULL;
// END ****************************************************************

    mpeg_deinitialize();
#ifdef USE_FIELD_MAP
	delete video_field_map;
#endif
    delete video_packet_buffer;
    delete video_packet_list;
    delete video_sample_list;
    delete audio_packet_list;
    delete audio_sample_list;

    if (pFastRead)
        delete pFastRead;

//    if (fh >= 0)
//        _close(fh);
}

//////////////////////

void InputFileMPEG::ReadStream(void *buffer, __int64 pos, long len, bool fAudio) {
    int pkt = 0;
    long delta;
    char *ptr = (char *)buffer;
    MPEGPacketInfo *packet_list = fAudio ? audio_packet_list : video_packet_list;
    int pkts = fAudio ? apackets : packets;

//  _RPT2(0,"Stream read request: pos %ld, len %ld\n", pos, len);

    if (!fInterleaved) {

        if (pFastRead)
            pFastRead->Read(0, pos, buffer, len);
        else {
            mFile.seek(pos);
            mFile.read(buffer, len);
//            if (_lseeki64(fh, pos, SEEK_SET)<0) throw MyError("seek error");
//            if (len != _read(fh, buffer, len)) throw MyError("read error");
        }
        return;
    }

    // find the packet containing the data start using a binary search

    do {
        int l = 0;
        int r = pkts-1;

        // Check current packet

        pkt = last_packet[fAudio?1:0];

        if (pkt>=0 && pkt<pkts) {
            if (pos < packet_list[pkt].stream_pos)
                r = pkt-1;
            else if (pos < packet_list[pkt+1].stream_pos)
                break;
            else if (pkt+1 < pkts && pos < packet_list[pkt+2].stream_pos) {
                ++pkt;
                break;
            } else
                l = pkt+2;
        }

        for(;;) {
            if (l > r) {
                pkt = l;
                break;
            }

            pkt = (l+r)/2;

            if (pos < packet_list[pkt].stream_pos)
                r = pkt-1;
            else if (pos >= packet_list[pkt+1].stream_pos)
                l = pkt+1;
            else
                break;
        }

        if (pkt<0 || pkt >= pkts) throw MyError("MPEG Internal error: Invalid stream read position (%ld)", pos);

    } while(false);

    // offset into packet for this stream position:
    delta = (long)(pos - packet_list[pkt].stream_pos);

    // read data from packets

    while(len) {
        if (pkt >= pkts) throw MyError("Attempt to read past end of stream (pos %ld)", pos);

        long tc = (long)((packet_list[pkt+1].stream_pos - packet_list[pkt].stream_pos) - delta);

        if (tc>len) tc=len;

//              if (!fAudio)
//                      _RPT3(0,"Reading %ld bytes at %08lx+%ld\n", tc, video_packet_list[pkt].file_pos,delta);

        if (pFastRead) {
            pFastRead->Read(fAudio ? 1 : 0, packet_list[pkt].file_pos + delta, ptr, tc);
        } else {
            mFile.seek(packet_list[pkt].file_pos + delta);
            mFile.read(ptr, tc);
//            if (_lseeki64(fh, packet_list[pkt].file_pos + delta, SEEK_SET)<0)
//                throw MyError("seek error");

//            if (tc!=_read(fh, ptr, tc)) throw MyError("read error");
        }
        len -= tc;
        ptr += tc;
        ++pkt;
        delta = 0;
    }

    last_packet[fAudio?1:0] = pkt-1;
}

///////////////////////////////////////////////////////////////////////////

void InputFileMPEG::setOptions(InputFileOptions *_ifo) {
    InputFileMPEGOptions *ifo = (InputFileMPEGOptions *)_ifo;

    iDecodeMode = ifo->opts.iDecodeMode;
    fAcceptPartial = ifo->opts.fAcceptPartial;
#ifdef ALLOW_AUDIO_DSC
    fAllowDSC = ifo->opts.fAllowDSC;
#endif

    if (iDecodeMode & InputFileMPEGOptions::DECODE_NO_P)
        iDecodeMode |= InputFileMPEGOptions::DECODE_NO_B;
}

InputFileOptions *InputFileMPEG::createOptions(const char *buf) {
    InputFileMPEGOptions *ifo = new InputFileMPEGOptions();

    if (!ifo) throw MyMemoryError();

    if (!ifo->read(buf)) {
        delete ifo;
        return NULL;
    }

    return ifo;
}

InputFileOptions *InputFileMPEG::promptForOptions(HWND hwnd) {
    InputFileMPEGOptions *ifo = new InputFileMPEGOptions();

    if (!ifo) throw MyMemoryError();

    DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_EXTOPENOPTS_MPEG), hwnd, InputFileMPEGOptions::SetupDlgProc, (LPARAM)ifo);

    return ifo;
}

///////////////////////////////////////////////////////////////////////////

typedef struct MyFileInfo {
    InputFileMPEG *thisPtr;

    volatile HWND hWndAbort;
    UINT statTimer;

    long lFrames;
    long lTotalSize;
    long lFrameCnt[3];
    long lFrameMinSize[3];
    long lFrameMaxSize[3];
    long lFrameTotalSize[3];
    long lAudioSize;
    long lAudioAvgBitrate;
    const char *lpszAudioMode;
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	vdrefptr<AudioSource> audioSrc;
// END ****************************************************************

} MyFileInfo;

void InputFileMPEG::_InfoDlgThread(void *pvInfo) {
    MyFileInfo *pInfo = (MyFileInfo *)pvInfo;
    InputFileMPEG *thisPtr = pInfo->thisPtr;
    long i;
    VideoSourceMPEG *vSrc = (VideoSourceMPEG *)&*pInfo->thisPtr->videoSrc;
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	AudioSourceMPEG *aSrc = (AudioSourceMPEG *)&*pInfo->audioSrc;
// END ****************************************************************
    MPEGSampleInfo *msi;

	const VDPosition videoFrameStart	= vSrc->getStart();
	const VDPosition videoFrameEnd		= vSrc->getEnd();

    for(i=0; i<3; i++)
        pInfo->lFrameMinSize[i] = 0x7FFFFFFF;

    msi = thisPtr->video_sample_list;
    for(i = videoFrameStart; i < videoFrameEnd; ++i) {
        int iFrameType = msi->frame_type;

        if (iFrameType) {
            long lSize = msi->size;
            --iFrameType;

            ++pInfo->lFrameCnt[iFrameType];
            pInfo->lFrameTotalSize[iFrameType] += lSize;

            if (lSize < pInfo->lFrameMinSize[iFrameType])
                pInfo->lFrameMinSize[iFrameType] = lSize;

            if (lSize > pInfo->lFrameMaxSize[iFrameType])
                pInfo->lFrameMaxSize[iFrameType] = lSize;

            pInfo->lTotalSize += lSize;

        }
        ++pInfo->lFrames;

        ++msi;
        if (pInfo->hWndAbort) {
            SendMessage(pInfo->hWndAbort, WM_USER+256, 0, 0);
            return;
        }
    }

    ///////////////////////////////////////////////////////////////////////

    if (aSrc) {
        static const char *szModes[8] =
			{"stereo", "joint stereo", "dual channel", "mono",
			 "3 channels", "4 channels", "5 channels", "5:1"};

        double lTotalBitrate = 0;

        switch (thisPtr->AudioType) {

        case ATYPE_MPEG: {
            bool fAudioMono = false;
            bool fAudioMixedMode = false;
            unsigned int audio_first_mode = MPEGAudioHeader(thisPtr->audio_first_header).GetMode();
            fAudioMono = !MPEGAudioHeader(thisPtr->audio_first_header).IsStereo();
            msi = thisPtr->audio_sample_list;

            for(i = thisPtr->aframes; i > 0; i--) {
                MPEGAudioHeader fAudioHeader(msi->header);

                if (!fAudioMixedMode) {
                    if (fAudioHeader.GetMode() != audio_first_mode)
                        fAudioMixedMode = true;
                }

                lTotalBitrate += fAudioHeader.GetBitrateKbps();
                pInfo->lAudioSize += msi->size;
                ++msi;

                if (pInfo->hWndAbort) {
                    SendMessage(pInfo->hWndAbort, WM_USER+256, 0, 0);
                    return;
                }
            }

            if (thisPtr->aframes)
                lTotalBitrate /= thisPtr->aframes;
            else lTotalBitrate = 0;
            pInfo->lAudioAvgBitrate = (long)lTotalBitrate;

            if (fAudioMixedMode) {
                if (fAudioMono)
                    pInfo->lpszAudioMode = "mixed mode";
                else
                    pInfo->lpszAudioMode = "mixed stereo";
            } else
                pInfo->lpszAudioMode = szModes[(thisPtr->audio_first_header>>30) & 3];

            }
            break;

        case ATYPE_AC3: {
			WAVEFORMATEX *wfex = aSrc->getWaveFormat();
			pInfo->lAudioAvgBitrate = wfex->nAvgBytesPerSec / 125;
			if (wfex->nChannels > 2) {
				pInfo->lpszAudioMode = szModes[wfex->nChannels+1];
			} else if (wfex->nChannels > 1) {
				pInfo->lpszAudioMode = szModes[0];
			} else {
				pInfo->lpszAudioMode = szModes[3];
			}
			pInfo->lAudioSize = aSrc->getEnd() * wfex->nBlockAlign;
            } break;

        case ATYPE_LPCM: {
            WAVEFORMATEX *wfex = aSrc->getWaveFormat();
            pInfo->lAudioAvgBitrate = wfex->nAvgBytesPerSec / 125;
            pInfo->lAudioSize = thisPtr->aframes * 1536*2 * wfex->nChannels;
            pInfo->lpszAudioMode = szModes[(wfex->nChannels > 1)? 0: 3];
            } break;
        }
    }

    pInfo->hWndAbort = (HWND)1;
}

BOOL APIENTRY InputFileMPEG::_InfoDlgProc( HWND hDlg, UINT message, UINT wParam, LONG lParam) {
    MyFileInfo *pInfo = (MyFileInfo *)GetWindowLong(hDlg, DWL_USER);
    InputFileMPEG *thisPtr;

    if (pInfo)
        thisPtr = pInfo->thisPtr;

    switch (message)
    {
        case WM_INITDIALOG:
            SetWindowLong(hDlg, DWL_USER, lParam);
            pInfo = (MyFileInfo *)lParam;
            thisPtr = pInfo->thisPtr;

            if (thisPtr->videoSrc) {
				char *s;
				char buf[128];

				sprintf(buf, "MPEG-%ld Information",
					thisPtr->fIsMPEG2? 2: 1);
				SetWindowText(hDlg, buf);

				sprintf(buf, "%dx%d, %.3f fps (%ld s)",
							thisPtr->width,
							thisPtr->height,
							thisPtr->videoSrc->getRate().asDouble(),
							VDRoundToLong(1000000.0 / thisPtr->videoSrc->getRate().asDouble()));
				SetDlgItemText(hDlg, IDC_VIDEO_FORMAT, buf);

				s = buf + sprintf(buf, "%lu (", (unsigned)thisPtr->videoSrc->getLength());
				ticks_to_str(s, VDRoundToLong(1000.0*thisPtr->videoSrc->getLength()/thisPtr->videoSrc->getRate().asDouble()));
				strcat(s,")");
				SetDlgItemText(hDlg, IDC_VIDEO_NUMFRAMES, buf);
			}
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
			if (pInfo->audioSrc) {
				WAVEFORMATEX *fmt = pInfo->audioSrc->getWaveFormat();
// END ****************************************************************
                char buf[128];

                sprintf(buf, "%ldHz, %s", fmt->nSamplesPerSec, fmt->nChannels>1 ? "Stereo" : "Mono");
                SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);

                sprintf(buf, "%ld", thisPtr->aframes);
                SetDlgItemText(hDlg, IDC_AUDIO_NUMFRAMES, buf);
            }
            _beginthread(_InfoDlgThread, 10000, pInfo);

            pInfo->statTimer = SetTimer(hDlg, 1, 250, NULL);

            return (TRUE);

        case WM_COMMAND:
            if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
            {
                if (pInfo->hWndAbort == (HWND)1)
                    EndDialog(hDlg, TRUE);
                return TRUE;
            }
            break;

        case WM_DESTROY:
            if (pInfo->statTimer) KillTimer(hDlg, pInfo->statTimer);
            break;

        case WM_TIMER:
            {
                char buf[128];

                sprintf(buf, "%ld", pInfo->lFrames);
                SetDlgItemText(hDlg, IDC_VIDEO_NUMKEYFRAMES, buf);

                sprintf(buf, "%ld / %ld / %ld", pInfo->lFrameCnt[0], pInfo->lFrameCnt[1], pInfo->lFrameCnt[2]);
                SetDlgItemText(hDlg, IDC_VIDEO_FRAMETYPECNT, buf);

                UINT uiCtlIds[]={ IDC_VIDEO_IFRAMES, IDC_VIDEO_PFRAMES, IDC_VIDEO_BFRAMES };

                for(int i=0; i<3; i++) {

                    if (pInfo->lFrameCnt[i])
                        sprintf(buf, "%ld / %ld / %ld (%ldK)"
                                    ,pInfo->lFrameMinSize[i]
                                    ,pInfo->lFrameTotalSize[i]/pInfo->lFrameCnt[i]
                                    ,pInfo->lFrameMaxSize[i]
                                    ,(pInfo->lFrameTotalSize[i]+1023)>>10);
                    else
                        sprintf(buf,"(no %c-frames)", "IPB"[i]);

                    SetDlgItemText(hDlg, uiCtlIds[i], buf);
                }

                if (pInfo->lTotalSize) {
                    long lKBitsPerSec, lKBytesPerSec;

                    // bits * (frames/sec) / frames = bits/sec
                    // (a / b) / c  =  a / (b * c)

					double d = ((double)pInfo->lTotalSize *
						thisPtr->videoSrc->getRate().asDouble())
						/ (double)pInfo->lFrames;

                    lKBytesPerSec = VDRoundToLong(d / 1024.0);
                    lKBitsPerSec = VDRoundToLong(d * 0.008);

                    //sprintf(g_msgBuf, "%ld Kbps (%ldKB/s)", (lBytesPerSec+124)/125, (lBytesPerSec+1023)/1024);
                    sprintf(buf, "%ld Kbps (%ldKB/s)", lKBitsPerSec, lKBytesPerSec);
                    SetDlgItemText(hDlg, IDC_VIDEO_AVGBITRATE, buf);
                }

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
				if (pInfo->lpszAudioMode && pInfo->audioSrc) {
					static const char *szLayers[]={"I","II","III"};
					WAVEFORMATEX *fmt = pInfo->audioSrc->getWaveFormat();
// END ****************************************************************

                    switch (thisPtr->AudioType) {
                        case ATYPE_MPEG:
                            sprintf(buf, "%ldKHz %s, %ldKbps layer %s",
                              fmt->nSamplesPerSec/1000,
                              pInfo->lpszAudioMode,
                              pInfo->lAudioAvgBitrate,
                              //szLayers[((thisPtr->audio_first_header>>17)&3)^3]
                              szLayers[((thisPtr->audio_first_header>>9)&3)^3]
                            );
                            break;

                        case ATYPE_AC3:
                            sprintf(buf, "%ldKHz %s, %ldKbps AC-3",
                              fmt->nSamplesPerSec/1000,
                              pInfo->lpszAudioMode,
                              pInfo->lAudioAvgBitrate
                            );
                            break;

                        case ATYPE_LPCM:
                            sprintf(buf, "%ldKHz %s, %ldKbps LPCM",
                              fmt->nSamplesPerSec/1000,
                              pInfo->lpszAudioMode,
                              pInfo->lAudioAvgBitrate
                            );
                            break;
                    }
                    SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);

                    sprintf(buf, "%ldK", (pInfo->lAudioSize + 1023) / 1024);
                    SetDlgItemText(hDlg, IDC_AUDIO_SIZE, buf);

					sprintf(buf, "%ld ms", VDRoundToLong((thisPtr->afirstPTS - thisPtr->firstPTS) / 90.0));
					SetDlgItemText(hDlg, IDC_AUDIO_SKEW, buf);
                }
            }

            /////////

            if (pInfo->hWndAbort) {
                KillTimer(hDlg, pInfo->statTimer);
                return TRUE;
            }

            break;

        case WM_USER+256:
            EndDialog(hDlg, TRUE);
            break;
    }
    return FALSE;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void InputFileMPEG::InfoDialog(HWND hwndParent) {
    MyFileInfo mai;

    memset(&mai, 0, sizeof mai);
    mai.thisPtr = this;

	mai.audioSrc = (AudioSource *)GetSource(streamtypeAUDIO, 0);

    DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_MPEG_INFO), hwndParent, _InfoDlgProc, (LPARAM)&mai);
}
// END ****************************************************************

/////////////////////////////////////////////////////////////////////

class VDInputDriverMPEG : public vdrefcounted<IVDInputDriver> {
public:
	const wchar_t *GetSignatureName() { return L"MPEG input driver (internal)"; }

	int GetDefaultPriority() {
		return 0;
	}

	uint32 GetFlags() { return kF_Video | kF_Audio; }

	const wchar_t *GetFilenamePattern() {
		return L"MPEG video/systems stream (*.mpg,*.mpeg,*.mpv,*.m1v,*.m2v,*.m2p,*.vob,*.dat)\0*.mpg;*.mpeg;*.mpv;*.m1v;*.m2v;*.m2p;*.vob;*.dat\0";
	}

	bool DetectByFilename(const wchar_t *pszFilename) {
		return false;
	}

	int DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
		if (nHeaderSize >= 12) {
			if (!memcmp(pHeader, "RIFF", 4) && !memcmp((char*)pHeader+8, "CDXA", 4))
				return 1;

			if (*(const uint32 *)pHeader == 0xba010000 || *(const uint32 *)pHeader==0xb3010000)
				return 1;

			// Second pass for MPEG.  This time, scan the first 64 bytes for 00 00 01 BA.

			int i;
			int limit = nHeaderSize - 3;

			if (limit > 60)
				limit = 60;

			for(i=1; i<limit; ++i)
				if (*(const uint32 *)((const uint8 *)pHeader+i) == 0xba010000 || *(const uint32 *)((const uint8 *)pHeader+i)==0xb3010000)
					break;

            if (i < limit)
				return 0;

		}

		return -1;
	}

	InputFile *CreateInputFile(uint32 flags) {
		return new InputFileMPEG;
	}
};

extern IVDInputDriver *VDCreateInputDriverMPEG() { return new VDInputDriverMPEG; }